/************************************************************************
 *	Copyright (C) 1998. Bernoulli Project, Cornell University.	*
 ************************************************************************/
// -*- C++ -*-

#ifndef BRN_UTILS_H
#define BRN_UTILS_H

#pragma interface


void crs_to_ccs(int m, int n, int nzs,
		   int *in_rowptr, int *in_colind, double *in_value,
		   int *out_colptr, int *out_rowind, double *out_value);
void ccs_to_crs(int m, int n, int nzs,
		int *in_colptr, int *in_rowind, double *in_value,
		int *out_rowptr, int *out_colind, double *out_value);

class BaseArray {
public:
    typedef mxArray *mxArrayp;
    virtual mxArray *copy() = 0;
};

class MatlabArray : public BaseArray {
protected:
    const mxArray *a;
public:
    MatlabArray(const mxArray *a_) : a(a_) { }
    virtual mxArray *copy() { return mxDuplicateArray(a); }
};

class DenseMatrix : public MatlabArray {
public:
    DenseMatrix(const mxArray *_a) : MatlabArray(_a) { }
    bool valid_p() { 
	return (bool)(mxGetClassID(a) == mxDOUBLE_CLASS && !mxIsComplex(a));
    }
    int m() { return mxGetM(a); }
    int n() { return mxGetN(a); }
    double *value() { return mxGetPr(a); }
};

class DenseVector : public MatlabArray {
public:
    DenseVector(const mxArray *_a) : MatlabArray(_a) { }
    bool valid_p() { 
	return (bool)(mxGetClassID(a) == mxDOUBLE_CLASS && !mxIsComplex(a)
	    && mxGetN(a) == 1);
    }
    int n() { return mxGetM(a); }
    double *value() { return mxGetPr(a); }
};


class CCS : public MatlabArray {
public:
    CCS(const mxArray *_a) : MatlabArray(_a) { }
    bool valid_p() { 
	return (bool)(mxGetClassID(a) == mxSPARSE_CLASS && !mxIsComplex(a));
    }
    int m() { return mxGetM(a); }
    int n() { return mxGetN(a); }
    int nzmax() { return mxGetNzmax(a); }
    int *colptr() { return mxGetJc(a); }
    int *rowind() { return mxGetIr(a); }
    double *value() { return mxGetPr(a); }
    int nzs() { return colptr()[n()]; }
};

class CRS : public BaseArray {
private:
    bool valid_p_;
    int m_, n_, nzs_, *rowptr_, *colind_;
    double *value_;
public:
    CRS(const mxArray *a) { 
	CCS A(a);
	valid_p_ = A.valid_p();
	if (!valid_p_) return;
	m_ = A.n(); n_ = A.m();	nzs_ = A.nzs();
	rowptr_ = new int[m_+1];
	colind_ = new int[nzs_];
	value_ = new double[nzs_];
	ccs_to_crs(A.m(),A.n(),A.nzs(), 
		   A.colptr(),A.rowind(),A.value(),
		   rowptr_, colind_, value_);
    }
    bool valid_p() { return valid_p_; }
    int m() { return m_; }
    int n() { return n_; }
    int nzs() { return nzs_; }
    int *rowptr() { return rowptr_; }
    int *colind() { return colind_; }
    double *value() { return value_; }
    virtual mxArray *copy() {
	mxArray *a = mxCreateSparse(m(),n(),nzs(),mxREAL);
	CCS A(a);
	crs_to_ccs(m(),n(),nzs(),
		   rowptr(),colind(),value(),
		   A.colptr(),A.rowind(),A.value());
	return a;
    }
};

#endif BRN_UTILS_H
