(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Bsct_base

module type A =
  sig
    type atom
    type expr
    exception Non_linear_form

    val debug_on : unit -> unit
    val debug_off : unit -> unit

    module Map :
        sig
          type t
          val rows : t -> int
          val cols : t -> int
          val new_map : int -> int -> t
          val compose : t -> t -> t
          val is_zero : t -> bool
          val is_ident : t -> bool
          val destructure : t -> Alpha.matrix * Alpha.vector
          val construct : Alpha.matrix -> Alpha.vector -> t
          val identity : int -> t
          val sub_rows : t -> int -> int -> t
          val extend2 : t -> t -> t * t
          val stack : t -> t -> t
          val compare_maps : t -> t -> int
          val equal_maps : t -> t -> bool
          val destructure_linear_map : t -> Alpha.matrix
          val destructure_affine_map :
              t -> ((int * int) * int array array) * (int * int array)
          val make_linear_map : Alpha.matrix -> t
          val to_string : t -> string
        end

    module Form :
        sig
          type t
          type term = | Term of atom * int
          type builder = | Builder_const of int | Builder_term of term
          val new_form : builder -> t
          val get_const : t -> int
          val get_term_list : t -> term list
          val coeff4atom : atom -> t -> int
          val update : t -> builder -> t
          val update_list : t -> builder list -> t
          val destructure_form_list : t -> term list * int
          val collect_atoms : t -> atom list
          val add : t -> t -> t
          val scale : int -> t -> t
          val saxpy : int -> t -> int -> t -> t
          val mult : t -> t -> t
          val of_expr : expr -> t
          val to_expr : t -> expr
          val to_array : atom list -> t -> int array
          val to_string : t -> string
        end

    val normalize_eq : expr -> expr
    val normalize_ineq : expr -> expr

    val form_list2map : atom list -> Form.t list -> Map.t
    val map2form_list : atom list -> Map.t -> Form.t list
    val expr_list2map : atom list -> expr list -> Map.t
    val relation2expr_list :
      atom list -> Map.t -> atom list -> Map.t -> expr list

    type relation_analysis_result =
	Overconstrained
      |	Bound_x of 
	  (*t:*) atom list * (*u:*) atom list * (*f:*) expr list * (*g:*)expr list 
      |	Universal_x of 
	  (*u:*) atom list * (*f:*) expr list

    val analyze_relation :
      (int -> atom list) ->
      atom list ->
      atom list ->
      atom list -> expr list -> relation_analysis_result
  end


open Alpha
open Ssisl
open Printf

module Make =
  functor (EX : Expr.E) -> 
    struct
      (* OT gives us the atom type *)

      type atom = EX.atom
      type expr = EX.expr
      exception Non_linear_form

      module ATOM = EX.ATOM
      module U = Map'.Make(ATOM)
      module S = Map.Make(ATOM)
      module Z = Set.Make(ATOM)
      module E = EX

      let compare = ATOM.compare

      let debug = ref false
      let debug_on () = 
        debug := true
      let debug_off () =
        debug := false

      (****************************************)
      (*   Affine maps                        *)
      (****************************************)
      module Map =
        struct
          type t = Map of matrix * vector

          let rows (Map(a,b)) = matrix_rows a
          let cols (Map(a,b)) = matrix_cols a     

          let new_map m n =
            Map (create_matrix m n, make_vector m)

          (* a*(c*x + d) + b --> (a*c)*x + (a*d+b) *)
          let compose (Map(a,b)) (Map(c,d)) =
            Map(mmm a c, saxpy 1 (mvm a d) 1 b)

          let is_zero (Map(a,b)) =
            ((is_zero_matrix a) & (is_zero_vector b))
          
          let is_ident (Map(a,b)) =
            ((is_ident_matrix a) & (is_zero_vector b))

          let destructure (Map(a,b)) = (a,b)

          let construct (a:matrix) (b:vector) =
            assert 
              ((vector_length b) = (matrix_rows a))
              (* (Failure "construct: invalid sizes") *);
            Map(a,b)

          let to_string (Map(a,b)) = 
            (matrix2string a) ^ " " ^ 
            (vector2string b)

          let identity n = 
            Map((create_identity n), (make_vector n))

          let sub_rows (Map(a,b)) i1 i2 =
            let n = matrix_cols a in
            Map((extract_matrix a i1 i2 1 n),
                (extract_vector b i1 i2))

          let rec extend2 a1 a2 = 
            let n1 = (cols a1)
            and n2 = (cols a2) in
            if (n1 = n2) then (a1,a2)
            else
              if (n1 > n2) then
                let (b2,b1) = (extend2 a2 a1) in
                (b1,b2)
              else
                match a1 with
                  (Map(a,b)) ->
                    (Map((add_cols a (n2-n1)),b), a2)
                      
          let stack (Map(a1,b1)) (Map(a2,b2)) =
            (* "stack" one map "on top of" another *)
            Map((stack_matrices a1 a2), 
                (concat_vectors b1 b2))

          let compare_maps (Map(a1,b1)) (Map(a2,b2)) =
            compare_pair 
              compare_matrices 
              compare_vectors
              (a1,b1) (a2,b2)

          let equal_maps a1 a2 =
            (0 = (compare_maps a1 a2))

          let destructure_linear_map (Map(a,b)) =
            assert (is_zero_vector b) 
              (* (Failure("destructure_linear_map: b != 0")) *);
            a

          let destructure_affine_map (Map(a,b)) =
            (destructure_matrix a, destructure_vector b)

          let make_linear_map (a:matrix) =
            (Map(a, make_vector (matrix_rows a)))

          let verify (Map(a,b) as m) =
            if ((matrix_rows a) <> (vector_length b))
            then failwith "affine_map.verify: bad affine_map"
            else m

          let to_string (Map(a,b)) =
            "Affine.Map("^(matrix2string a)^","^
            (vector2string b)^")"

        end (* end of Affine.Map *)
          

      (****************************************)
      (*   Affine forms                       *)
      (****************************************)
      module Form =
        struct

          (* int is the result type
             the key type is atom *)
          type term_map = int S.t 
            
          type t = Form of term_map * int

          type term = Term of atom * int


          type builder =
              Builder_const of int
            | Builder_term of term

          let new_form bld =
            match bld with
              Builder_const(c) -> Form(S.empty, c)
            | Builder_term(Term(id,c)) ->
                Form(S.add id c S.empty, 0)

          let get_const f = 
            match f with 
              Form(_,c) -> c
                  
          let get_term_list f = 
            match f with
              (Form(m,_)) ->
                (List.map
                   (fun (a,b) -> Term(a,b))
                   (U.list_of_bindings m))

          let coeff4atom (a:atom) (f:t) =
            match f with
              (Form(m,_)) ->
                try (S.find a m) with
                  Not_found -> 0

          let update (f:t) (bld:builder) =
            match (f,bld) with
              (Form(m,c),Builder_const(c')) -> Form(m,c+c')
            | (Form(m,c),Builder_term(Term(a,x))) ->
                Form(U.replace a (x + (coeff4atom a f)) m, c)

          let update_list (f:t) (l:builder list) =
            (List.fold_left update f l)

          let destructure_form_list f = 
            (get_term_list f, get_const f)

          let destructure_form f =
            match f with
              (Form(m,c)) -> (m,c)

          let collect_atoms f =
            (List.map (fun (Term(a,b)) -> a)
               (get_term_list f))
              
          let add f1 f2 = (
            match (f1,f2) with
              (Form(m1,c1),Form(m2,c2)) ->
                let g a x m = (
                  try
                    (U.replace a (x+(S.find a m)) m)
                  with
                    Not_found -> S.add a x m) in
            (* this will enumerate over m2 and add into m1 *)
                Form(S.fold g m1 m2, c1+c2))

          let scale c f =
            let l = 
              (List.map 
                 (fun (Term(atm,x)) -> (atm,c*x))
                 (get_term_list f)) in
            Form(U.absorb_list S.empty l, c * (get_const f))

          let saxpy a1 f1 a2 f2 =
            match (f1,f2) with
              (Form(m1,c1),Form(m2,c2)) ->
                add (scale a1 f1) (scale a2 f2)
                  
          let mult f1 f2 =
            match (f1,f2) with
              (Form(m1,c1),Form(m2,c2)) -> 
                if ((U.map_size m1) = 0)
                then scale c1 f2
                else 
                  if ((U.map_size m2) = 0)
                  then scale c2 f1
                  else raise Non_linear_form

          let to_string f =
            let g str t =
              let (Term(a,i)) = t in
              str^" + "^(string_of_int i)^"*"^(Sx.to_string (ATOM.to_sexpr a)) in
            "Affine.Form("^
            (List.fold_left g "" (get_term_list f))^
            " + "^(string_of_int (get_const f))^")"

          (****************************************)
          (*  Expressions to affine forms         *)
          (****************************************)

          let of_expr (e:expr) =
            let t = E.expr_type e in
            if ((E.Type.type_lub t E.Type.Int) <> E.Type.Int) then
              raise Non_linear_form
            else
              let rec g e =
                match e with
                  (E.Atom(a)) -> new_form (Builder_term(Term(a,1)))
                | (E.Const(E.Int(i))) -> 
                    new_form (Builder_const(i))
                | (E.Op(E.Plus,l)) ->
                    let l' = List.map g l in
                    List.fold_left add (new_form (Builder_const(0))) l'
                | (E.Op(E.Times,l)) ->
                    let l' = List.map g l in
                    List.fold_left mult (new_form (Builder_const(1))) l'
                | (E.Op(E.Neg,l)) ->
                    scale (-1) (g (E.op_arg e))
                | (E.Op(E.Div,_)) -> raise Non_linear_form
                | (E.Op(E.Apply,_)) -> raise Non_linear_form
                | _ -> failwith "of_expr: should not be here"
              in (g e)        

          let to_expr f =
            let term2expr (Term(a,x)) = 
              if (x = 1) then 
                E.Atom(a)
              else if (x = (-1)) then
                E.Op(E.Neg, [E.Atom(a)])
              else
                E.Op(E.Times, [E.Const(E.Int(x)); E.Atom(a)]) 
            in
            let l = List.map term2expr (get_term_list f) in
            (E.make_plus 
               ((E.Const(E.Int(get_const f)))::l))

          let to_array (order:atom list) (f:t) =
            let fill_array order (l:term list) (v:(int array ref)) =
              try
                (List.iter 
                   (fun (Term(a,x)) -> 
                     let i = 
		       1 + (List'.long_position
			      (fun x y -> compare x y = 0) identity
			      a order)
                     in 
                     (!v).(i-1) <- x)
                   l);
              with Not_found -> failwith "form2vector: bad order list"
            in 
            let v = Array.create (List.length order) 0 in
            fill_array order (get_term_list f) (ref v);
            v
                
        end (* end Affine.Form *)


      (********************************************************
       *              Expression normalization
       *********************************************************)

      let normalize_eq e = 
	match e with
	  E.Op(E.Eq,[e1;e2]) -> E.Op(E.Eq,
				     [ (E.make_minus e1 e2) ;
				       E.Const(E.Int 0) ])
	| _ -> failwith "normalize_eq: not an equality"

      let normalize_ineq e =
	let make_leq e = E.Op(E.Leq, [e; E.Const(E.Int 0)])
	and add_one e = E.make_plus [e;E.Const(E.Int 1)] in
	match e with
	  E.Op(E.Leq,[e1;e2]) -> make_leq (E.make_minus e1 e2)
	| E.Op(E.Lt,[e1;e2]) -> make_leq (add_one (E.make_minus e1 e2))
	| E.Op(E.Geq,[e1;e2]) -> make_leq (E.make_minus e2 e1)
	| E.Op(E.Gt,[e1;e2]) -> make_leq (add_one  (E.make_minus e2 e1))
	| _ -> failwith "normalize_eq: not an inequality"


      (********************************************************)
      (*   Various converters                                 *)
      (********************************************************)

      let form_list2map order (l:Form.t list) =
        let m = List.length l
        and n = List.length order in
        if (m = 0) then
          Map.Map(create_matrix 0 n, make_vector 0)
        else
          Map.Map((caml_to_alpha_matrix 
                     (Array.of_list (List.map (Form.to_array order) l))),
                  (caml_to_alpha_vector
                     (Array.of_list (List.map Form.get_const l))))

      let map2form_list order mp = 
        let (Map.Map(a,b)) = (Map.verify mp) in
        let m = matrix_rows a 
        and n = matrix_cols a
        and l = ref [] 
        and order_array = Array.of_list order in
        begin
          for i = 1 to m do
            let f = ref 
                (Form.new_form 
                   (Form.Builder_const (vector_ref b i)))
            in
            for j = 1 to n do
              if (matrix_ref a i j <> 0) then
                f := (Form.update !f
                        (Form.Builder_term(Form.Term(order_array.(j-1),
                                                     (matrix_ref a i j))
                                           )))
              else ()
            done;
            l := !l @ [(!f)];
          done;
          (!l)
        end

      let expr_list2map order (l:E.expr list) =
        let g e = Form.of_expr e
        in form_list2map order (List.map g l)

      let relation2expr_list order1 mp1 order2 mp2 =
        let l1 = (map2form_list order1 mp1)
        and l2 = (map2form_list order2 mp2)
        in
        if ((List.length l1) <> (List.length l2)) then
          failwith "relation2expr_list: Bad affine relation"
        else
          let combine_form f1 f2 = 
            E.Op(E.Eq,
                 [ (Form.to_expr f1) ;
                   (Form.to_expr f2) 
                 ])
          in
          (List.map2 combine_form l1 l2)

      let set_of_list l =
        List.fold_right Z.add l Z.empty

      let list_of_set s =
        Z.fold (fun a l -> a::l) s []

      let is_subset s t =
        (Z.equal s (Z.inter s t))
          
      let are_disjoint s t =
        (Z.is_empty (Z.inter s t))

      let reform_to_string w x y a =
	(Sx.to_string
	   (Sx.from_list
	      [ (Sx.from_list (List.map ATOM.to_sexpr w)) ;
		(Sx.from_list (List.map ATOM.to_sexpr x)) ;
		(Sx.from_list (List.map ATOM.to_sexpr y)) ;
		(Sx.from_list (List.map E.to_sexpr a)) ]))
		

      let make_affine_product_expr_list x a b =
	(* make a list of expressions for 
	   every element in the product a*x + b *)
	let (m,n) = (matrix_rows a, matrix_cols a) in
	(zassert (m = (vector_length b))
	   (Invalid_argument "make_matrix_product_expr_list: b"));
	(zassert (n = (List.length x))
	   (Invalid_argument "make_matrix_product_expr_list: x"));
	(List.map Form.to_expr
	   (map2form_list x (Map.construct a b)))

      let make_linear_product_expr_list x a =
	(*  make a list of expressions for 
	    every element in the product a*x *)
	(make_affine_product_expr_list x a
	   (make_vector (matrix_rows a)))


      type relation_analysis_result =
	  Overconstrained
      	|	Bound_x of 
	  (*t:*) atom list * (*u:*) atom list * (*f:*) expr list * (*g:*)expr list 
      	|	Universal_x of 
	  (*u:*) atom list * (*f:*) expr list

      let analyze_relation genatoms w x y a =
	if !debug then begin
	  print_string "=====================================\n";
	  print_string "Affine.analyze trace:\n\n";
	  print_string "Inputs: ";
	  print_string (reform_to_string w x y a);
	  print_newline ();
	end
	else ();

        let w_set = set_of_list w
        and x_set = set_of_list x
        and y_set = set_of_list y in
        (zassert (is_subset x_set w_set)
           (Invalid_argument "Affine.analyze_relation"));
        (zassert (is_subset y_set w_set)
           (Invalid_argument "Affine.analyze_relation"));
        (zassert (are_disjoint x_set y_set)
           (Invalid_argument "Affine.analyze_relation"));
        let z = list_of_set (Z.diff w_set (Z.union x_set y_set)) in
        let w = x@y@z 
        and (nx,ny,nz) = (List.length x,
                          List.length y,
                          List.length z) in
        let (a_matrix,b_vector) = 
          Map.destructure (expr_list2map w a) in
        if !debug then begin
          print_string "before a.reform"; print_newline ();
          printf "a=%s b=%s\n" 
            (matrix2string a_matrix)
            (vector2string b_vector);
          print_newline ()
        end else ();
	try
	  let (x0,y0,l_matrix,m_matrix,n_matrix) = 
	    (Alpha.analyze_relation nx ny a_matrix b_vector) in
	  try
	    let (y0_new, f, g) = 
	      (Alpha.reformulate_block_echelon 
		 x0 y0 l_matrix m_matrix n_matrix) in
	    let u_lst = (genatoms (matrix_cols g)) in
	    let f_times_x = 
	      (make_affine_product_expr_list x f y0_new)
	    and g_times_t = 
	      (make_linear_product_expr_list u_lst g) in
	    let y_expr_list =
	      (List.map (fun (e1,e2) -> E.make_plus [e1;e2])
		   (List.combine f_times_x g_times_t)) in
	    (Universal_x (u_lst, y_expr_list))
	    
	  with
	    Reformulate_no_solution -> 
	      let t_lst = genatoms (matrix_cols l_matrix)
	      and u_lst = genatoms (matrix_cols n_matrix)
	      and zero_v k = (make_vector k) in
	      let x_expr_list = 
		(make_affine_product_expr_list t_lst l_matrix x0)
	      and m_times_t =
		(make_affine_product_expr_list t_lst m_matrix y0)
	      and n_times_u = 
		(make_linear_product_expr_list u_lst n_matrix) in
	      let y_expr_list = 
		(List.map (fun (e1,e2) -> E.make_plus [e1;e2])
		   (List.combine m_times_t n_times_u)) in
	      (Bound_x (t_lst, u_lst, x_expr_list, y_expr_list))
	with No_integer_solution -> 
	    Overconstrained

     (*****************
        let (y0,f,g) = 
	  (Alpha.reformulate_relation nx ny a_matrix b_vector) in
        if !debug then begin
          print_string "after a.reform"; 
          print_newline ()
        end else ();
        let t = genatoms (matrix_cols g)
        and t0 = make_vector (matrix_rows g) in
        let f_form_list = 
          map2form_list x (Map.construct f y0)
        and g_form_list =
          map2form_list t (Map.construct g t0) in
        let f_expr_list =
          List.map Form.to_expr f_form_list 
        and g_expr_list =
          List.map Form.to_expr g_form_list in
	zassert 
	  ((List.length f_expr_list)==(List.length g_expr_list))
	  (Failure("reformulate_relation:|f|<>|g|"));
        (t, f_expr_list, g_expr_list)
	***********************)

    end


