(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
module type A =
  sig
    type atom
    type expr
    exception Non_linear_form

    val debug_on : unit -> unit
    val debug_off : unit -> unit

    module Map :
	sig
	  type t
	  val rows : t -> int
	  val cols : t -> int
	  val new_map : int -> int -> t
	  val compose : t -> t -> t
	  val is_zero : t -> bool
	  val is_ident : t -> bool
	  val destructure : t -> Alpha.matrix * Alpha.vector
	  val construct : Alpha.matrix -> Alpha.vector -> t
	  val identity : int -> t
	  val sub_rows : t -> int -> int -> t
	  val extend2 : t -> t -> t * t
	  val stack : t -> t -> t
	  val compare_maps : t -> t -> int
	  val equal_maps : t -> t -> bool
	  val destructure_linear_map : t -> Alpha.matrix
	  val destructure_affine_map :
              t -> ((int * int) * int array array) * (int * int array)
	  val make_linear_map : Alpha.matrix -> t
	  val to_string : t -> string
	end

    module Form :
	sig
	  type t
	  type term = | Term of atom * int
	  type builder = | Builder_const of int | Builder_term of term
	  val new_form : builder -> t
	  val get_const : t -> int
	  val get_term_list : t -> term list
	  val coeff4atom : atom -> t -> int
	  val update : t -> builder -> t
	  val update_list : t -> builder list -> t
	  val destructure_form_list : t -> term list * int
	  val collect_atoms : t -> atom list
	  val add : t -> t -> t
	  val scale : int -> t -> t
	  val saxpy : int -> t -> int -> t -> t
	  val mult : t -> t -> t
	  val of_expr : expr -> t
	  val to_expr : t -> expr
	  val to_array : atom list -> t -> int array
	  val to_string : t -> string
	end

    val normalize_eq : expr -> expr
    val normalize_ineq : expr -> expr

    val form_list2map : atom list -> Form.t list -> Map.t
    val map2form_list : atom list -> Map.t -> Form.t list
    val expr_list2map : atom list -> expr list -> Map.t
    val relation2expr_list :
      atom list -> Map.t -> atom list -> Map.t -> expr list

    type relation_analysis_result =
	Overconstrained
      |	Bound_x of 
	  (*t:*) atom list * (*u:*) atom list * (*f:*) expr list * (*g:*)expr list 
      |	Universal_x of 
	  (*u:*) atom list * (*f:*) expr list

    (* For given an affine system F(x,y,others) = 0 one of the
       following is true:
       -- the system can be without solution (Overconstrained)
       -- x is not bound by the system. Then we can express y
          as: y = f(x,t) where t is a free param
       -- x is bound by the system to some non-empty affine set.
          the we can express the solution in echelon form:
          x = f(t)
          y = g(t,u)
          where t and u are free params *)

    val analyze_relation :
      (int -> atom list) ->
      atom list ->
      atom list ->
      atom list -> expr list -> relation_analysis_result
  end


module Make (EX : Expr.E ) : A with type atom = EX.atom and type expr = EX.expr

