(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)

(*

value expr_set2builder_set : bmf_expr -> affine_set_builder;;
value builder_set2expr_set : affine_set_builder -> bmf_expr;;
value intersect_sets : bmf_expr -> bmf_expr -> bmf_expr;;
value simplify_set : bmf_expr -> bmf_expr;;
value gist_sets : bmf_expr -> bmf_expr -> bmf_expr;;
value fm_code_gen : bmf_expr -> (bmf_expr list * bmf_expr list) list;;

*)

(************************************************************************)
(* Interface to Fourier-Motzkin						*)
(************************************************************************)

let expr_set2builder_set (orig_e: bmf_expr) = (
    walk_set orig_e
	where rec walk_set e =
	    match e with
	      expr_set(frame(frm),e') ->
		check_all_int_bindings frm;
		let (exists,constrs) = walk_exist e'
		in
		    builder_set(map bmf_ident_name frm, exists, constrs)
	    | expr_bool(false) -> builder_no_solution
	    | _ -> raise (Invalid_expr_set orig_e)
	and walk_exist e =
	    match e with
	      expr_exist(frame(frm),e') ->
		check_all_int_bindings frm;
		(map bmf_ident_name frm,
		 walk_constraints e')
	    | _ -> ([],walk_constraints e)
	and check_all_int_bindings ids =
	    do_list
		(fun id ->
		    match id with
		      ident_valid(_,type_scalar(type_int),_,_) -> ()
		    | _ -> raise (Invalid_expr_set orig_e))
		ids
	and walk_constraints e =
	    flat_map walk_constraint
		(bmf_expr_args_by_op op_and e)
	and walk_constraint e =
	    match e with
	      expr_op(op_not,[e']) ->
		let (builder_constraint(kind,left,right)) =
		    walk_simple_constraint e'
		in
		    (match kind with
		      builder_eq ->
			[builder_constraint(builder_lt,left,right);
			 builder_constraint(builder_gt,left,right)]
		    | builder_le ->
			[builder_constraint(builder_gt,left,right)]
		    | _ -> failwith "Shouldn't have happened.")
	    | expr_bool(true) -> []
	    | expr_bool(false) -> raise (Invalid_expr_set orig_e)
	    | _ -> [walk_simple_constraint e]
	and walk_simple_constraint e =
	    match e with
	      expr_op(op_eq,[e1;e2]) ->
		builder_constraint(builder_eq,
		    walk_terms e1, walk_terms e2)
	    | expr_op(op_leq,[e1;e2]) ->
		builder_constraint(builder_le,
		    walk_terms e1, walk_terms e2)
	    | _ -> raise (Invalid_expr_set orig_e)
	and walk_terms e =
	    match e with
	      expr_op(op_plus,args) -> flat_map walk_terms args
	    | expr_op(op_neg,[e]) ->
		map (fun t ->
			match t with
			  builder_const(c) ->
			    builder_const(-1*c)
			| builder_var(c,v) ->
			    builder_var(-1*c, v))
		    (walk_terms e)
	    | _ -> [walk_term e]
	and walk_term e =
	    match e with
	      expr_int(c) -> builder_const(c)
	    | expr_var(v) -> builder_var(1,v)
	    | expr_op(op_times,[expr_int(c);expr_var(v)]) ->
		builder_var(c,v)
	    | expr_op(op_times,[expr_var(v);expr_int(c)]) ->
		builder_var(c,v)
	    | _ -> raise (Invalid_expr_set orig_e)
: affine_set_builder);;

let builder_set2expr_set (orig_s: affine_set_builder) = (
    walk_set orig_s
	where rec walk_set s =
	    match s with
	      builder_set(set_vars,exist_vars,constrs) ->
		let set_ids = map make_ident set_vars
		and exist_ids = map make_ident exist_vars
		and body = walk_constrs constrs
		in
		    if exist_ids = [] then
			expr_set(frame(set_ids),body)
		    else
			expr_set(frame(set_ids),
			    expr_exist(frame(exist_ids), body))
	    | builder_no_solution -> expr_bool(false)
	and make_ident name =
	    ident_valid(name,type_scalar(type_int),ident_bind,an_list[])
	and walk_constrs constrs =
	    match constrs with
	      [] -> expr_bool(true)
	    | [x] -> walk_constr x
	    | _ -> expr_op(op_and,(map walk_constr constrs))
	and walk_constr (builder_constraint(kind,left,right)) =
	    let left' = walk_terms left
	    and right' = walk_terms right
	    in
    		match kind with
		  builder_eq -> expr_op(op_eq,[left';right'])
		| builder_le -> expr_op(op_leq,[left';right'])
		| builder_ge -> expr_op(op_leq,[right';left'])
		| builder_lt ->
		    expr_op(op_not,[expr_op(op_leq,[right';left'])])
		| builder_gt ->
		    expr_op(op_not,[expr_op(op_leq,[left';right'])])
	and walk_terms terms =
	    match terms with
	      [] -> failwith "Empty term list?"
	    | [x] -> walk_term x
	    | _ -> expr_op(op_plus,(map walk_term terms))
	and walk_term t =
	    match t with
	      builder_const(c) -> expr_int(c)
	    | builder_var(1,v) -> expr_var(v)
	    | builder_var(c,v) ->
		expr_op(op_times,[expr_int(c);expr_var(v)])
: bmf_expr);;

let intersect_sets s1 s2 =
    let (s1_bound,s1_existential,s1_pred) =
	match s1 with
	  expr_set(frame(s1_bound),
		expr_exist(frame(s1_existential),s1_pred)) ->
	    (s1_bound,s1_existential,s1_pred)
	| expr_set(frame(s1_bound),s1_pred) ->
	    (s1_bound,[],s1_pred)
	| _ -> raise (Invalid_expr_set s1)
    and (s2_bound,s2_existential,s2_pred) =
	match s2 with
	  expr_set(frame(s2_bound),
		expr_exist(frame(s2_existential),s2_pred)) ->
	    (s2_bound,s2_existential,s2_pred)
	| expr_set(frame(s2_bound),s2_pred) ->
	    (s2_bound,[],s2_pred)
	| _ -> raise (Invalid_expr_set s2)
    in let bound = 
	zassert (map bmf_ident_name s1_bound =
		    map bmf_ident_name s2_bound)
	    Incompatible_sets;
	s1_bound
    and existential = union s1_existential s2_existential
    and pred =
	make_and ((bmf_expr_args_by_op op_and s1_pred)@
		  (bmf_expr_args_by_op op_and s2_pred))
    in
	if existential = [] then
	    expr_set(frame(bound),pred)
	else
	    expr_set(frame(bound),
		expr_exist(frame(existential),pred))
;;

let simplify_set s =
    builder_set2expr_set (simplify (expr_set2builder_set s))
;;

let gist_sets s1 s2 =
    builder_set2expr_set
	(gist (expr_set2builder_set s1)
	    (expr_set2builder_set s2))
;;

let rec fm_code_gen orig_e =
    let simplified_e = simplify_set orig_e
    (* Destruct the set expr. Collect all of the inequalities *)
    in let (vars,constrs,orig_constrs) =
        match simplified_e with
	  expr_set(frame(frm),pred) ->
	    let vars = map bmf_ident_name frm
	    and constrs = 
		flat_map
		    (fun conj ->
			match conj with
			  expr_bool(true) -> []
			| expr_op(op_leq,[e1;e2]) -> [(e1,e2)]
			| expr_op(op_eq,[e1;e2]) -> [(e1,e2);(e2,e1)]
			| _ -> raise (Invalid_expr_set orig_e))
		    (bmf_expr_args_by_op op_and pred)
	    in
		(vars,constrs,pred)
	| _ -> failwith "Can't happen"
    in let do_one_step var results_for_vars' =
	(* Constraints of the form e1 <= e2 ->
	    affines of the form e1-e2 <= 0 *)
	let affines =
	    map (fun (e1,e2) ->
		    affine_add (expr2affine e1)
			(affine_scale (-1) (expr2affine e2)))
		constrs
	(* Partition the constrs in
	    pos_var <- coeff of `var' is > 0.
	    neg_var <- coeff of `var' is < 0. *)
	in let (pos_var,neg_var) =
	    it_list
		(fun (pos_var,neg_var) aff ->
		    let coef = affine_coeff4var var aff
		    in
			if coef > 0 then
			    (aff::pos_var,neg_var)
			else if coef < 0 then
			    (pos_var,aff::neg_var)
			else
			    (pos_var,neg_var))
		([],[]) affines
	and remove_var c aff =
	    let coef = affine_coeff4var var aff
	    in
		assert (coef = c) (* (Failure("Wrong coef?!?")) *);
		affine2expr
		    (affine_scale (-1*c)
			(affine_add aff
			    (new_affine_form
				(builder_var(-1*c,var)))))
	in let lbs = map (remove_var (-1)) neg_var
	and ubs = map (remove_var 1) pos_var
	in let rec remove_dups lst =
	    match lst with
	      [] -> []
	    | x::lst' ->
		let lst'' = remove_dups lst'
		in
		    if mem x lst' then
			lst''
		    else
			x::lst''
	in
	    results_for_vars'
		@[(remove_dups lbs,remove_dups ubs)]
    in
	match rev vars with
	  [] -> failwith "Can't code gen with no vars"
	| [var] -> do_one_step var []
	| var::srav' ->
	    let vars' = rev srav'
	    and make_ident name =
		ident_valid(name,type_scalar(type_int),ident_bind,an_list[])
	    in let results_for_vars' =
		fm_code_gen
		    (expr_set(frame(map make_ident vars'),
			expr_exist(frame[make_ident var],
			    orig_constrs)))
	    in
		do_one_step var results_for_vars'
;;
