(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Printf

(*open Alpha_private;;
open Alpha_types;;

type arena = Alpha_types.arena;;
type vector = Alpha_types.vector;;
type matrix = Alpha_types.matrix;;
type affine_set_builder = Alpha_types.affine_set_builder ;;
type affine_constraint_builder = Alpha_types.affine_constraint_builder;;
type affine_constraint_kind = Alpha_types.affine_constraint_kind;;
type affine_term_builder = Alpha_types.affine_term_builder;;
*)

type arena;;
type vector;;
type matrix;;

type affine_set_builder =
    Builder_no_solution
  | Builder_set of string list * string list * affine_constraint_builder list
and affine_constraint_builder =
  Builder_constraint of affine_constraint_kind * affine_term_builder list
      * affine_term_builder list
(* These must occur in the same order as in alpha-private.c *)
and affine_constraint_kind =
  Builder_eq
  | Builder_le
  | Builder_lt
  | Builder_ge
  | Builder_gt
and affine_term_builder =
    Builder_const of int
  | Builder_var of int * string
;;

exception Reformulate_no_solution;;

let debug = ref false
let debug_on () =
  debug := true
let debug_off () =
  debug := false

(************************************************************************)
(* alpha_private.c							*)
(************************************************************************)

external a_make_vector : arena -> int -> vector
    =  "ma_make_vector";;
external a_length_vec : vector -> int
    =  "ma_length_vec";;
external a_elt_vec_ref : vector -> int -> int
    =  "ma_elt_vec_ref";;
external a_elt_vec_set : vector -> int -> int -> unit
    =  "ma_elt_vec_set";;

external a_make_matrix : arena -> int -> int -> matrix
    =  "ma_make_matrix";;
external a_rows_mx : matrix -> int
    =  "ma_rows_mx";;
external a_cols_mx : matrix -> int
    =  "ma_cols_mx";;
external a_elt_mx_ref : matrix -> int -> int -> int
    =  "ma_elt_mx_ref";;
external a_elt_mx_set : matrix -> int -> int -> int -> unit
    =  "ma_elt_mx_set";;

external a_set_vec : vector -> int -> unit
    =  "ma_set_vec";;
external a_neg_vec : vector -> unit
    =  "ma_neg_vec";;
external a_saxpy : (*a:*)int -> (*x:*)vector -> (*b:*)int -> (*y:*)vector ->
	(*z:*)vector -> unit
    =  "ma_saxpy";;
external a_copy_vec : (*dst:*)vector -> (*src*)vector -> unit
    =  "ma_copy_vec";;
external a_resize_vec : arena -> vector -> int -> vector
    =  "ma_resize_vec";;
external a_is_zero_vec : vector -> bool
    =  "ma_is_zero_vec";;
external a_gcd_vec : vector -> int
    =  "ma_gcd_vec";;
external a_extract_vec : vector -> int -> int -> vector -> unit
    =  "ma_extract_vec";;

external a_is_zero_matrix : matrix -> bool
    =  "ma_is_zero_matrix";;
external a_is_ident_matrix : matrix -> bool
    =  "ma_is_ident_matrix";;

external a_copy_mx : (*dst:*)matrix -> (*src:*)matrix -> unit
    =  "ma_copy_mx";;
external a_swap_col : matrix -> int -> int -> unit
    =  "ma_swap_col";;
external a_swap_row : matrix -> int -> int -> unit
    =  "ma_swap_row";;
external a_set2ident : matrix -> unit
    =  "ma_set2ident";;
external a_matmult : arena -> (*A:*)matrix -> (*B:*)matrix -> (*C:*)matrix
	    -> unit
    =  "ma_matmult";;
external a_mvm : arena -> (*A:*)matrix -> (*x:*)vector -> (*y:*)vector
	    -> unit
    =  "ma_mvm";;
external a_transpose : (*A:*)matrix -> (*At:*)matrix -> unit
    =  "ma_transpose";;
(* B = A[i1:i2, j1:j2] *)
external a_extract : (*A:*)matrix -> (*i1:*)int -> (*i2:*)int ->
	(*j1:*)int -> (*j2:*)int -> (*B:*)matrix -> unit
    =  "ma_extract";;
(* A[i1:i2, j1:j2] = B *)
external a_replace : (*A:*)matrix -> (*i1:*)int -> (*i2:*)int ->
	(*j1:*)int -> (*j2:*)int -> (*B:*)matrix -> unit
    =  "ma_replace";;
external a_add_cols : arena -> matrix -> int -> matrix
    =  "ma_add_cols";;
external a_trunc_cols : arena -> matrix -> int -> matrix
    =  "ma_trunc_cols";;
external a_lower_echelon : arena -> (*A:*)matrix -> (*U:*)matrix
	    -> (*rank:*)int
    =  "ma_lower_echelon";;
external a_snf : arena -> (*A:*)matrix -> (*U:*)matrix -> (*V:*)matrix
	    -> (*rank:*)int
    =  "ma_snf";;
external a_solve_linear : arena -> (*A:*)matrix -> (*B:*)vector
	    -> (*X0:*)vector -> (*H:*)matrix -> (*null_dim:*)int
    =  "ma_solve_linear";;
external a_solve_echelon : (*L:*)matrix -> (*B:*)vector -> (*X:*)vector
	    -> bool
    =  "ma_solve_echelon";;

external a_unit_backsolve : (*U:*)matrix -> (*B*)vector -> (*X*)vector -> unit
    =  "ma_unit_backsolve"
    ;;

external a_block_unit_backsolve : (*U:*)matrix -> (*B*)matrix -> (*X*)matrix -> unit
    =  "ma_block_unit_backsolve"
    ;;

(************************************************************************)
(* Fourier-Motzkin							*)
(************************************************************************)
 
external ma_simplify_set : affine_set_builder -> affine_set_builder
    =  "ma_simplify_set";;
external ma_satisfiable_p : affine_set_builder -> bool
    =  "ma_satisfiable_p";;


(************************************************************************)
(* ML layer starts here							*)
(************************************************************************)

exception Out_of_bounds of string;;
exception Matrix_mismatch;;

let bounds_check i n str =
  zassert (1 <= i && i <= n) (Out_of_bounds str)
;;

(************************************************************************)
(* Arenas								*)
(************************************************************************)

external a_make_arena : unit -> arena =  "ma_make_arena";;

(* let make_arena = a_make_arena *)

(* todo: There is a bug in alpha_private.c that is only triggered after
   many arena allocations and deallocations. The current workaround is to
   only use one arena from the ML code. This is what is implemented by the
   code below. The original code that tripped the bug is commented out
   above. *)

let the_arena = a_make_arena ();;
let make_arena () = the_arena;;

(************************************************************************)
(* Integers								*)
(************************************************************************)

external a_gcd : int -> int -> int
    =  "ma_gcd";;

let gcd x y =
  zassert (x >= 0) (Invalid_argument "gcd");
  zassert (y >= 0) (Invalid_argument "gcd");
  a_gcd x y
    ;;

(* a*u + b*v == d == (a,b) *)
external a_egcd : (*a:*)int -> (*b:*)int -> (*d:*)int * (*u:*)int * (*v:*)int
    =  "ma_egcd";;

let egcd x y =
  a_egcd x y
    ;;

external a_lcm : int -> int -> int
    =  "ma_lcm";;

let lcm x y =
  zassert (x > 0) (Invalid_argument "lcm");
  zassert (y > 0) (Invalid_argument "lcm");
  a_lcm x y
    ;;

(************************************************************************)
(* Vectors								*)
(************************************************************************)

let make_vector n =
  a_make_vector (make_arena ()) n;;

let vector_length = a_length_vec;;
let vector_ref v i =
  bounds_check i (vector_length v) "vector_ref";
  a_elt_vec_ref v i
    ;;

let vector_set v i x =
  bounds_check i (vector_length v) "vector_set";
  a_elt_vec_set v i x
    ;;

let is_zero_vector = a_is_zero_vec;;

let extract_vector v i1 i2 =
  let n = vector_length v
  and n' = (i2-i1+1) in
  (zassert (1 <= i1 && i2 <= n && n' >= 0)
     (Invalid_argument "extract_vector"));
  let w = make_vector (i2-i1+1) in
  (a_extract_vec v i1 i2 w); w
    ;;

let alpha_vector_to_caml (v:vector) = (
  let n = a_length_vec v
  in let v' = Array.create n 0
  in
  for i = 1 to n do
    v'.(i-1) <- vector_ref v i
  done;
  v'
    : int array)
    ;;

let caml_to_alpha_vector (v:(int array)) = (
  let n = Array.length v
  in let v' = make_vector n
  in
  for i = 1 to n do
    vector_set v' i (v.(i-1))
  done;
  v'
    : vector)
  ;;

let destructure_vector v =
  (vector_length v, alpha_vector_to_caml v)
    ;;
    
let copy_vector v =
  let v' = make_vector (vector_length v)
  in
  a_copy_vec v' v;
  v'
    ;;
(************************************************************************)
(* Matrices								*)
(************************************************************************)

let create_matrix rs cs =
  a_make_matrix (make_arena ()) rs cs;;
let matrix_rows = a_rows_mx;;
let matrix_cols = a_cols_mx;;
let matrix_size a = (matrix_rows a, matrix_cols a);;
    
let matrix_ref m i j =
  bounds_check i (matrix_rows m) "matrix_ref";
  bounds_check j (matrix_cols m) "matrix_ref";
  a_elt_mx_ref m i j
    ;;

let matrix_set m i j x =
  bounds_check i (matrix_rows m) "matrix_set";
  bounds_check j (matrix_cols m) "matrix_set";
  a_elt_mx_set m i j x
    ;;

let alpha_matrix_to_caml (m:matrix) = (
  let rs = a_rows_mx m
  and cs = a_cols_mx m
  in let m' = Array.create_matrix rs cs 0
  in
  for i = 1 to rs do
    for j = 1 to cs do
      m'.(i-1).(j-1) <- matrix_ref m i j
    done
  done;
  m'
    : int array array)
  ;;
  
let caml_to_alpha_matrix (m:int array array) = (
  let rs = Array.length m
  in 
  if (rs = 0) then
    failwith "caml_to_alpha_matrix: zero rows";
  let cs = Array.length (m.(0))
  in let m' = create_matrix rs cs
  in
  for i = 1 to rs do
    for j = 1 to cs do
      matrix_set m' i j (m.(i-1).(j-1))
    done
  done;
  m'
    : matrix)
  ;;

let destructure_matrix a =
  (matrix_size a, alpha_matrix_to_caml a)
;;

let copy_matrix m =
  let m' = create_matrix (matrix_rows m) (matrix_cols m)
  in
  a_copy_mx m' m;
  m'
    ;;

let add_cols m c = 
  zassert (c >=0) (Invalid_argument("add_cols"));
  if c > 0 then
    a_add_cols (make_arena ()) m c
  else
    copy_matrix m
      ;;

let truncate_cols m c =
  zassert (c >=0 && c <= (matrix_cols m))
    (Invalid_argument("truncate_cols"));
  if c > 0 then
    a_trunc_cols (make_arena ()) m c
  else
    copy_matrix m
      ;;

let is_zero_matrix  = a_is_zero_matrix;;
let is_ident_matrix = a_is_ident_matrix;;
    
let create_identity n =
  let i = create_matrix n n
  in
  a_set2ident i;
  i
    ;;

(* we allow i2 and j2 to be 0 *)
let extract_matrix a i1 i2 j1 j2 =
  let m = matrix_rows a
  and n = matrix_cols a
  and m' = (i2-i1+1)
  and n' = (j2-j1+1) in
  if !debug then begin
    printf "Extract_matrix: %d %d %d %d -- %d %d\n" i1 i2 j1 j2 m n;
    print_newline ();
  end else ();
  (zassert (1 <= i1 && i2 <= m && m' >= 0)
     (Invalid_argument "extract_matrix"));
  (zassert (1 <= j1 && j2 <= n && n' >= 0)
     (Invalid_argument "extract_matrix"));
  let b = (create_matrix m' n') in
  (a_extract a i1 i2 j1 j2 b); b
    ;;

(************************************************************************)
(* Printers                                                             *)
(************************************************************************)

let rows_printer fmt f ws n =
  Format.pp_open_hbox fmt ();
  Format.pp_print_string fmt "[";
  for i = 1 to n do
    if i <> 1 then
      Format.pp_print_string fmt ",";
    let s = string_of_int (f i)
    in
    Format.pp_print_string fmt " ";
    for j = (String.length s)+1 to (ws.(i-1)) do
      Format.pp_print_string fmt " "
    done;
    Format.pp_print_string fmt s
  done;
  Format.pp_print_string fmt " ]";
  Format.pp_close_box fmt ();
    ;;

let compute_width f ws n =
  for i = 1 to n do
    let s = string_of_int (f i)
    in
    ws.(i-1) <- max (ws.(i-1)) (String.length s)
  done
    ;;

let format_vector fmt (v:vector) =
  let n = (vector_length v)
  in let ws = Array.create n 0
  in
  compute_width (vector_ref v) ws n;
  rows_printer fmt (vector_ref v) ws n
    ;;

let output_vector outch (v:vector) =
  let fmt = Format.make_formatter (output outch) (fun () -> flush outch) in
  Format.pp_set_margin fmt 99999;
  format_vector fmt v;
  Format.pp_print_flush fmt ()

let print_vector (v:vector) =
  let fmt = Format.make_formatter (output stdout) (fun () -> flush stdout) in
  format_vector fmt v;
  Format.pp_print_flush fmt ()

let format_matrix fmt (m:matrix) =
  let rs = matrix_rows m
  and cs = matrix_cols m
  in let ws = Array.create cs 0
  in
  for i = 1 to rs do
    compute_width (matrix_ref m i) ws cs
  done;
  Format.pp_open_hvbox fmt 1;
  Format.pp_print_string fmt "[";
  for i = 1 to rs do
    rows_printer fmt (matrix_ref m i) ws cs;
    if i <> rs then begin
      Format.pp_print_string fmt ",";
      Format.pp_print_space fmt ()
    end
  done;
  Format.pp_print_string fmt "]";
  Format.pp_close_box fmt ();
    ;;

let output_matrix outch (m:matrix) =
  let fmt = Format.make_formatter (output outch) (fun () -> flush outch) in
  Format.pp_set_margin fmt 999999;
  format_matrix fmt m;
  Format.pp_print_flush fmt ()

let print_matrix (m:matrix) =
  let fmt = Format.make_formatter (output stdout) (fun () -> flush stdout) in
  Format.pp_set_margin fmt 1;
  format_matrix fmt m;
  Format.pp_print_flush fmt ()

let matrix_row2string a i =
  let m = matrix_rows a
  and n = matrix_cols a
  in
  bounds_check i m "matrix_row2string";

  let s = ref "["
  in 
  if (n = 0) then "[ ]"
  else (
    for j = 1 to n-1 do 
      s := (!s ^ (string_of_int (matrix_ref a i j)) ^ ",")
    done ;
    !s ^ (string_of_int (matrix_ref a i n)) ^ "]")
      ;;

let matrix2string a =
  let m = matrix_rows a
  and n = matrix_cols a
  in let s = ref "["
  in
  if (m = 0) then
    (for j = 1 to n do
      s := !s ^ ". "
    done ;
     (!s ^ "]"))
  else
    (for i = 1 to m-1 do
      s := !s ^ ((matrix_row2string a i) ^ ",")
    done ;
     !s ^ ((matrix_row2string a m) ^ "]"))
      ;;

let vector2string v = 
  let n = vector_length v
  in
  let s = ref "["
  in 
  if (n = 0) then "[ ]"
  else (
    for i = 1 to n-1 do 
      s := (!s ^ (string_of_int (vector_ref v i)) ^ ",")
    done ;
    !s ^ (string_of_int (vector_ref v n)) ^ "]")
      ;;


(************************************************************************)
(* The interesting routines                                             *)
(************************************************************************)

exception No_integer_solution;;

let vector_gcd = a_gcd_vec;;

let saxpy a x b y =
  zassert ((vector_length x) = (vector_length y))
    (Matrix_mismatch);
  let n = (vector_length x)
  in let z = (make_vector n)
  in
  a_saxpy a x b y z;
  z
    ;;

let sax a x =
  let n = (vector_length x)
  in let y = (make_vector n)
  in for i = 1 to n do
    (vector_set y i (a*(vector_ref x i)))
  done;
  y
    ;;

let mvm (a:matrix) (x:vector) = (
  zassert (matrix_cols a = vector_length x) Matrix_mismatch;
  let y = make_vector (matrix_rows a)
  in
  a_mvm (make_arena ()) a x y;
  y
    : vector);;

let mmm (a:matrix) (b:matrix) = (
  zassert (matrix_cols a = matrix_rows b) Matrix_mismatch;
  let c = create_matrix (matrix_rows a) (matrix_cols b)
  in
  a_matmult (make_arena ()) a b c;
  c
    : matrix);;

let transpose (a:matrix) = (
  let at = create_matrix (matrix_cols a) (matrix_rows a)
  in
  a_transpose a at;
  at
    : matrix);;

let lower_echelon (a:matrix) =
  let a' = copy_matrix a
  and rs = matrix_rows a
  and cs = matrix_cols a
  in let u = create_matrix cs cs
  in let rank = a_lower_echelon (make_arena ()) a' u
  in
  (a',u,rank)
    ;;

let snf a =
  let a' = copy_matrix a
  and rs = matrix_rows a
  and cs = matrix_cols a
  in let u = create_matrix rs rs
  and v = create_matrix cs cs
  in let rank = a_snf (make_arena ()) a' u v
  in (a',u,v,rank)
    ;;

let solve_echelon l b =
  zassert (matrix_rows l = vector_length b) Matrix_mismatch;
  let x = make_vector (matrix_cols l)
  in
  let return_value = (a_solve_echelon l b x) in
  zassert return_value No_integer_solution;
  x
    ;;


let solve_linear a b =
  zassert (matrix_rows a = vector_length b) Matrix_mismatch;
  let rs = matrix_rows a
  and cs = matrix_cols a
  in let h = create_matrix cs cs
  and x0 = make_vector cs
  in let rank = a_solve_linear (make_arena ()) a b x0 h
  in if rank >= 0 then
    (truncate_cols h (cs-rank), x0)
    (* else
      (create_matrix cs 1, x0) *)
  else
    raise No_integer_solution
      ;;



let last_non_zero a i1 i2 =
  (* find the last non-zero in a row *)
  (* undefined if the matrix has no columns *)
  let (m,n) = (matrix_rows a, matrix_cols a) in
  (zassert (1<=i1 && i1<= m)
     (Invalid_argument "Alpha.last_non_zero: i1"));
  (zassert (1<=i2 && i2<= m)
     (Invalid_argument "Alpha.last_non_zero: i2"));
  let pos = ref 0 in
  for i = i1 to i2 do
    for j = 1 to n do
      let v = (matrix_ref a i j) in
      if ((v <> 0) && (j > !pos)) then
	pos := j
      else ()
    done;
  done;
  !pos
    ;;

let analyze_relation nx ny a b =
  (***************
    start with an equation system:
    a*(x y z)^T = b

    and relate x and y in the form:
    x = x0 + L*t
    y = y0 + M*t + N*u
    where L and N are lower-echelon of full
    column rank

    output is a tuple (x0,y0,L,M,N)
    *****************)
  zassert ((nx >= 0) && (ny >= 0)) 
    (Invalid_argument "reformulate");
  zassert (nx + ny <= (matrix_cols a))
    (Invalid_argument "reformulate");
  zassert (matrix_rows a = vector_length b)
    (Invalid_argument "reformulate");
  let (h,w0) = (solve_linear a (sax (-1) b)) in
               (* this can raise No_integer_solution *)
  let (m,n) = (matrix_rows h, matrix_cols h)
  and (h',_,r) = (lower_echelon h) in
  (zassert (r = n) (Failure "Alpha.analyze_relation: H not full rank"));
  let limit_x = 
    (if (nx > 0) 
    then (last_non_zero h' 1 nx)
    else 0) in
  let limit_y =
    (* it is possible that y depends only on t *)
    max 
      limit_x
      (if ((nx+ny) > 0)
      then (last_non_zero h' (nx+1) (nx+ny))
      else 0) in
  if !debug then begin
    (printf "h=%s w0=%s h_prime=%s r=%d limit_x=%d limit_y=%d"
       (matrix2string h) 
       (vector2string w0) 
       (matrix2string h') r limit_x limit_y);
    print_newline ();
  end
  else ()
      ;
  (***********
     L = h'(1:nx, 1:limit_x)
     M = h'((nx+1):(nx+ny), 1:limit_x)
     N = h'((nx+1):(nx+ny), (limit_x+1):limit_y)
     x0 = w0(1:nx)
     y0 = w0((nx+1):(nx+ny))
   **************)
  let l_matrix = (extract_matrix h' 1 nx 1 limit_x)
  and m_matrix = (extract_matrix h' (nx+1) (nx+ny) 1 limit_x)
  and n_matrix = (extract_matrix h' (nx+1) (nx+ny) (limit_x+1) limit_y) 
  and x0 = (extract_vector w0 1 nx)
  and y0 = (extract_vector w0 (nx+1) (nx+ny)) in
  if !debug then begin
    (printf "l=%s m=%s n=%s\n" 
       (matrix2string l_matrix) 
       (matrix2string m_matrix) 
       (matrix2string n_matrix));
    print_newline ();
  end else ();
  (x0, y0, l_matrix, m_matrix, n_matrix)
    ;;

let reformulate_block_echelon x0 y0 l m n =  
  zassert ((matrix_rows l) = (vector_length x0))
    (Invalid_argument "reformulate_block_echelon: l, x0");
  zassert ((matrix_rows m) = (vector_length y0))
    (Invalid_argument "reformulate_block_echelon: m, y0");
  zassert ((matrix_rows n) = (vector_length y0))
    (Invalid_argument "reformulate_block_echelon: n, y0");
  zassert ((matrix_cols l) = (matrix_cols m))
    (Invalid_argument "reformulate_block_echelon: l, m");
  (***************************************************
    We have the system in the form:
     x - x0 = L * w
     y - y0 = M * w + N * z
     where w and z are free parameters

    x is unconstrained if and only if L is square and 
    has only +-1 on the diagonals. Then we can express y as:
    y = y0 + M * L^{-1}*(x-x0) + N*z

    If we set F = M * L^{-1} and G = N the answer is:
    y = y0_new + F*x + G*z =
      = (y0 - F*x0) + F*x + G*z

    if L is not square or does not have only +-1 on the diagonal,
    then x is constrained
    
   ****************************************************)
  if !debug then begin
    (printf "reform_block: x0=%s y0=%s l=%s m=%s n=%s"
       (vector2string x0) (vector2string y0)
       (matrix2string l) (matrix2string m) (matrix2string n));
    (print_newline ())
  end
  else ();
  if ((matrix_cols l) <> (matrix_rows l)) then
    raise Reformulate_no_solution
  else begin
    for j = 1 to (matrix_cols l) do
      if (1 <> (abs(matrix_ref l j j))) then
	raise Reformulate_no_solution
      else ()
    done;
    (* at this point we know that x is uncondstrained *)
    let f_transpose = 
      (* has transposed dimenstions of M *)
      (create_matrix (matrix_cols m) (matrix_rows m)) in
    (a_block_unit_backsolve 
       (transpose l) (transpose m) f_transpose);
    let f = (transpose f_transpose) in
    let y0_new = (saxpy 1 y0 (-1) (mvm f x0)) in
    if !debug then begin
      (printf "reform_block: y0_new=%s f=%s"
	 (vector2string y0_new)
	 (matrix2string f));
      print_newline ();
    end
    else ();
    (y0_new,f,n)
  end

let reformulate_relation nx ny a b =
  let (x0,y0,l,m,n) = analyze_relation nx ny a b in
  reformulate_block_echelon x0 y0 l m n
  
  
(*********************************************************)
(*
let reformulate (nx,ny,nz) a b = 
  zassert ((nx >= 0) && (ny >= 0) && (nz >= 0)) 
    (Invalid_argument "reformulate");
  zassert (nx + ny + nz = (matrix_cols a))
    (Invalid_argument "reformulate");
  zassert (matrix_rows a = vector_length b)
    (Invalid_argument "reformulate");
  let (s,w0) = 
    try
      solve_linear a (sax (-1) b) 
    with 
      No_integer_solution -> raise (Reformulate_no_solution) 
  in
  let (m,n) = (matrix_rows s, matrix_cols s) in
  if !debug then begin
    printf "m=%d n=%d nx=%d ny=%d nz=%d\n" 
      m n nx ny nz;
    printf "a=%s b=%s s=%s w0=%s\n" 
      (matrix2string a) (vector2string b)
      (matrix2string s) (vector2string w0);
    print_newline ()
  end else ();
  let sx = extract_matrix s 1 nx 1 n
  and sy = extract_matrix s (nx+1) (nx+ny) 1 n
  and x0 = extract_vector w0 1 nx
  and y0 = extract_vector w0 (nx+1) (nx+ny) in
  if !debug then 
    (print_string "done extract 1"; print_newline ())
  else ();
  let (sigma,u,v,r) = snf sx in
  assert (r <= nx) (* (Failure "reformulate") *);
  if !debug then begin
    printf "sigma=%s u=%s v=%s r=%d\n"
      (matrix2string sigma)
      (matrix2string u)
      (matrix2string v) r;
    print_newline ()
  end else ();
  if (r <> nx) then raise Reformulate_no_solution
  else
    let k = mmm sy v in
    if !debug then begin
      printf "r=%d m_k=%d n_k=%d"
      	r (matrix_rows k) (matrix_cols k);
      print_newline ()
    end else ();
    let (k1,k2) = 
      (extract_matrix k 1 ny 1 r,
       extract_matrix k 1 ny (r+1) n) in
    for j = 1 to r do
      for i = 1 to ny do
	let k_val = (matrix_ref k1 i j)
	and s_val = (matrix_ref sigma j j) in
	if (0 <> (k_val mod s_val)) then raise Reformulate_no_solution
	else (matrix_set k1 i j (k_val/s_val) )
      done
    done;
    let f = mmm k1 u in
    let x0 = mvm f x0 in
    let y0 = saxpy 1 y0 (-1) x0 in
    if !debug then begin
      printf "f=%s k2=%s y0=%s\n"
      	(matrix2string f) (matrix2string k2) (vector2string y0);
      print_newline ()
    end else ();
    (f,k2,y0)
      ;;

*)

(*
  Here is how reformulate() works:

  we solve the equation A * w + b = 0 
  (notice that it is not A*x=b)
  to obtain w = S*t + w_0
  w is, of course a block vector: [x y z]^T
  This gives us the expressions for x and y

  x = x0 + sx * t
  y = y0 + sy * t

  sx must have full row rank -- otherwise there is no formula
  that would give us a y for EVERY x

  SNF of sx is:  u*sx*v = [sigma 0]

  let z = v^{-1}*t
  let z1 be the first nx elements of t, then
  u*(x - x0) = sigma*z1 and z1 = sigma^{-1} * u * (x-x0)

  we can write: y = y0 + sy*v*z = y0 + k*z
  let k1 be the first r=nx columns of k, let k2 be the rest
  Then:

  y = y0 + k1*z1 + k2*z2 =
    = y0 + k1*sigma^{-1}*u*(x-x0) + k2*z2

  Now, if k1*sigma^{-1} is an integer matrix, then for every integral x we 
  can obtain an integral y as

  y = f*x + g*z2 + y0'

  where f = k1*sigma^{-1}*u
        y0' = y0 - f*x0
        g = k2
        z2 is the free parameter
*)

(***********************************************************************
 *               Routines not involving calls to Alpha
 ***********************************************************************)

let stack_matrices a b =
    (* A on top of B *)
  let ma = matrix_rows a
  and na = matrix_cols a    
  and mb = matrix_rows b
  and nb = matrix_cols b
  in
  zassert (na = nb) (Invalid_argument "stack_matrices");
  if ((ma+mb) = 0) then
    a (* does not matter which one is returned *)
  else
    caml_to_alpha_matrix
  (Array.append 
     (alpha_matrix_to_caml a)
     (alpha_matrix_to_caml b))
  ;;

let concat_vectors v w =
  caml_to_alpha_vector
  (Array.append
     (alpha_vector_to_caml v)
     (alpha_vector_to_caml w))
  ;;

let compare_matrices a b =
  Pervasives.compare
    (matrix_size a, alpha_matrix_to_caml a)
    (matrix_size b, alpha_matrix_to_caml b)
    ;;

let compare_vectors x y =
  Pervasives.compare
    (alpha_vector_to_caml x)
    (alpha_vector_to_caml y)
    ;;  

let get_column a j =
  let (m,n) = matrix_size a
  in let c = make_vector m
  in
  bounds_check j n "get_column";
  for i = 1 to m do
    (vector_set c i (matrix_ref a i j))
  done ;
  c
    ;;

let list_of_int_range i j  =
    (* [i;i+1;...;j] if i <= j
       [] otherwise *)
  if i <= j then
    let n = (j-i+1)
    in let v = (Array.create n 0)
    in 
    for k = i to j do (v.(k-i) <- k) done ;
    Array.to_list v
  else
    []
      ;;

let list_of_matrix_cols a =
  (List.map (get_column a) 
     (list_of_int_range 1 (matrix_cols a)))
    ;;

let matrix_of_cols_list lst =
  if (List.length lst = 0) then
    create_matrix 0 0
  else
    let caml_vec_vec = 
      (Array.of_list (List.map alpha_vector_to_caml lst))
    in transpose (caml_to_alpha_matrix caml_vec_vec)
      ;;  


let make_pos a =
    (* this is a hack to make basis "look nice"
       in "most" cases this will produce matrices with
       positive entries *)

  let (m,n) = matrix_size a
  and cpos = ref 0
  and cneg = ref 0
  in
  for i = 1 to m do
    for j = 1 to n do
      let a0 = matrix_ref a i j
      in if a0 > 0 then incr cpos
      else (if a0 < 0 then incr cneg
      else ())
    done
  done ;
  if (cneg > cpos) then (
    for i = 1 to m do
      for j = 1 to n do
        matrix_set a i j 
          ((-1)*(matrix_ref a i j))
      done
    done ;
    a )
  else a
      ;;

let solve_matrix_eqn a b =
    (* A * X = B where B is a matrix
       a particular solution X0 and the basis for null(A) 
       are returned *)

  let (m,n) = matrix_size a
  and (m',n') = matrix_size b

  in let b_cols = 
    zassert (m = m') (Invalid_argument "solve_matrix_eqn");
    zassert (m >= 0) (Invalid_argument "solve_matrix_eqn: m < 0");
    zassert (n >= 0) (Invalid_argument "solve_matrix_eqn: n < 0");
    zassert (n' >= 0) (Invalid_argument "solve_matrix_eqn: n' < 0");
    list_of_matrix_cols b

  in let (h,x_cols) =
    List.fold_left
      (fun (_,lst) b0 -> 
        let (h,x0) = solve_linear a b0
        in (h, lst@[x0]))
      (create_matrix 0 0, [])
      b_cols

  in
  if (is_zero_matrix h) then (matrix_of_cols_list x_cols, h)
  else
    let h = make_pos h
    and nH = matrix_cols h
    in let j = ref 0
    in let upgrade x = (
      incr j; (* cycle through columns of H *)
      (if is_zero_vector x then
        saxpy 1 (get_column h (1+((!j) mod nH))) 1 x
      else x))
    in ((matrix_of_cols_list (List.map upgrade x_cols)), h)
      ;;

let matrix_saxpy x a y b =
  zassert ((matrix_size a) = (matrix_size b))
    (Invalid_argument "matrix_saxpy");
  let (m,n) = (matrix_size a)
  in let c = create_matrix m n
  in
  for i = 1 to m do
    for j = 1 to n do
      (matrix_set c i j
         (x*(matrix_ref a i j) + 
            y*(matrix_ref b i j)))
    done
  done ;
  c
    ;;

let matrix_sax x a =
  let (m,n) = matrix_size a
  in let b = create_matrix m n
  in
  for i = 1 to m do
    for j = 1 to n do
      (matrix_set b i j
         (x*(matrix_ref a i j)))
    done
  done ;
  b
    ;;
      
let gensym_num = ref 0;;
    
let gensym str =
  let n = !gensym_num in
  incr gensym_num;
  str ^ "." ^ (string_of_int n)
		;;

(* Fourier-Motzkin *)


exception Free_variable_in_set of string;;
exception Incompatible_sets;;
exception Gist_existentials;;

let check_set s =
  match s with
    Builder_no_solution -> ()
  | Builder_set(set_vars,exist_vars,constrs) -> (
      let rec check_constraint (Builder_constraint(k,left,right)) =
	List.iter check_term left;
	List.iter check_term right
      and check_term t =
	match t with
	  Builder_const(_) -> ()
	| Builder_var(_,v) ->
	    if (not ((List.mem v set_vars || List.mem v exist_vars))) then begin
	      raise (Free_variable_in_set(v))
	    end
      in
      List.iter check_constraint constrs)
	  ;;
		
let simplify s = 
  check_set s;
  ma_simplify_set s
;;

let satisfiable_p s =
  check_set s;
  ma_satisfiable_p s
;;

let redundant_constr_p (Builder_constraint(k,left,right)) s =
  match s with
    Builder_no_solution -> false
  | Builder_set(set_vars,exist_vars,constrs) ->
      let neg_constrs =
	match k with
	  Builder_eq ->
	    [Builder_constraint(Builder_lt, left, right);
	      Builder_constraint(Builder_gt, left, right)]
	| Builder_le ->
	    [Builder_constraint(Builder_gt, left, right)]
	| Builder_lt ->
	    [Builder_constraint(Builder_ge, left, right)]
	| Builder_ge ->
	    [Builder_constraint(Builder_lt, left, right)]
	| Builder_gt ->
	    [Builder_constraint(Builder_le, left, right)]
      in
      (not
	 ((ma_satisfiable_p
	     (Builder_set(set_vars,exist_vars,
			  neg_constrs@constrs)))))
	;;

let gist s1 s2 =
  check_set s1;
  check_set s2;
  match (s1,s2) with
    (Builder_no_solution,_) -> Builder_no_solution
  | (s1,Builder_no_solution) -> s1
  | (Builder_set(s1_set_vars,s1_exist_vars,s1_constrs),
     Builder_set(s2_set_vars,s2_exist_vars,s2_constrs)) ->
       if (s1_exist_vars <> []) then begin
	 raise Gist_existentials
       end;
       if (List'.difference s1_set_vars s2_set_vars <> []) then begin
	 raise Incompatible_sets
       end;
       let s1_constrs' =
	 List.fold_left
	   (fun constrs' test_constr ->
	     if redundant_constr_p test_constr s2 then
	       constrs'
	     else
	       test_constr::constrs')
	   [] s1_constrs
       in 
       Builder_set(s1_set_vars, [], s1_constrs')
	 ;;

