(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)

type arena;;
type vector;;
type matrix;;

type affine_set_builder =
    Builder_no_solution
  | Builder_set of string list * string list * affine_constraint_builder list
and affine_constraint_builder =
  Builder_constraint of affine_constraint_kind * affine_term_builder list
      * affine_term_builder list
(* These must occur in the same order as in alpha-private.c *)
and affine_constraint_kind =
  Builder_eq
  | Builder_le
  | Builder_lt
  | Builder_ge
  | Builder_gt
and affine_term_builder =
    Builder_const of int
  | Builder_var of int * string
;;

(* Matrices and vectors *)

exception Out_of_bounds of string;;
exception Matrix_mismatch;;

val debug_on : unit -> unit;;
val debug_off : unit -> unit;;

val make_arena : unit -> arena;;

val gcd : int -> int -> int;;
(* a*u + b*v == d == (a,b) *)
val egcd : (*a:*)int -> (*b:*)int -> (*d:*)int * (*u:*)int * (*v:*)int;;
val lcm : int -> int -> int;;

val make_vector : int -> vector;;
val vector_length : vector -> int;;
val vector_ref : vector -> int -> int;;
val vector_set : vector -> int -> int -> unit;;

val is_zero_vector : vector -> bool;;
val extract_vector : vector -> int -> int -> vector;;
val copy_vector : vector -> vector;;

val create_matrix : int -> int -> matrix;;
val matrix_rows : matrix -> int;;
val matrix_cols : matrix -> int;;
val matrix_ref : matrix -> int -> int -> int;;
val matrix_set : matrix -> int -> int -> int -> unit;;

val vector_gcd : vector -> int;;

val saxpy : int -> vector -> int -> vector -> vector;;
val sax : int -> vector -> vector;;
val mvm : matrix -> vector -> vector;;
val mmm : matrix -> matrix -> matrix;;
val transpose : matrix -> matrix;;

val is_zero_matrix : matrix -> bool;;
val is_ident_matrix : matrix -> bool;;
val create_identity : int -> matrix;;
val extract_matrix : matrix -> int -> int -> int -> 
                       int -> matrix;;
val copy_matrix : matrix -> matrix;;
val add_cols : matrix -> int -> matrix;;
val truncate_cols : matrix -> int -> matrix;;


(* lower_echelon A -> L*U^-1 *)
val lower_echelon : (*A:*)matrix -> 
	(*L:*)matrix * (*U:*)matrix * (*rank:*)int;;
(* snf A -> U^-1*Sigma*V^-1, first `rank' entries of Sigma's diagonal are
    non-zero, and the rest of Sigma is zero. *)
val snf : (*A:*)matrix ->
	(*Sigma:*)matrix * (*U:*)matrix * (*V:*)matrix * (*rank:*)int;;
(* A*x = b -> x = H*t + x0 *)
val solve_linear : (*A:*)matrix -> (*b:*)vector ->
	(*H:*)matrix * (*x0:*)vector;;
(* L^-1*b = x *)
exception No_integer_solution;;
val solve_echelon : (*L:*)matrix -> (*b:*)vector ->
    (*x:*)vector;;

val alpha_vector_to_caml : vector -> int array;;
val caml_to_alpha_vector : int array -> vector;;
val alpha_matrix_to_caml : matrix -> int array array;;
val caml_to_alpha_matrix : int array array -> matrix;;

val output_vector : out_channel -> vector -> unit;;
val print_vector : vector -> unit;;
val output_matrix : out_channel -> matrix -> unit;;
val print_matrix : matrix -> unit;;
val matrix2string : matrix -> string;;
val vector2string : vector -> string;;

val stack_matrices : matrix -> matrix -> matrix;;
val concat_vectors : vector -> vector -> vector;;

val matrix_size : matrix -> (int * int);;
val compare_matrices : matrix -> matrix -> int;;
val compare_vectors : vector -> vector -> int;;

val get_column : matrix -> int -> vector;;
val list_of_matrix_cols : matrix -> vector list;;
val matrix_of_cols_list : vector list -> matrix;;
val solve_matrix_eqn : matrix -> matrix -> (matrix *matrix);;
val make_pos : matrix -> matrix;;

val matrix_saxpy : int -> matrix -> int -> matrix -> matrix;;
val matrix_sax : int -> matrix -> matrix;;

val destructure_matrix : matrix -> (int * int) * (int array array);;
val destructure_vector : vector -> int * (int array);;

(*************************************************************
   The following routines analyze linear systems.
   We start with a linear relations of the form:

     [x]
   A*[y] = b
     [z]

   (x0,y0,L,M,N) = (analyze_relation nx ny a b) 
   returns the solutions for x and y in block echelon form:
   x = x0 + L*t
   y = y0 + M*t + N*u
   where L and N are in echelon form and t and u are free params

   (F,G,y0) = (reformulate_block_echelon x0 y0 L M N)
   checks if there is a valid y for every x. That is if 
   for all x we can write y = y0 + F*x + G*u, where u is a free param.
   reformulate_block_echelon raises Reformulate_no_solution exception
   if this is not possible

   reformulate_relation is just a composition of the above two functions

**************************************************************)
exception Reformulate_no_solution;;

val analyze_relation : (*nx*) int -> (*ny*) int -> (*A*) matrix ->
  (* b *) vector -> 
    (* x0,y0,L,M,N *) vector*vector*matrix*matrix*matrix;;

val reformulate_block_echelon :
    (* x0 *) vector -> (* y0 *) vector ->
      (* L *) matrix -> (* M *) matrix -> (* N *) matrix ->
	(* y0_new, F, G *) vector * matrix * matrix

val reformulate_relation : (*nx*) int -> (*ny*) int -> (*A*) matrix ->
  (* b *) vector -> 
    (* y0_new, F, G *) vector * matrix * matrix


(****************************** Fourier-Motzkin ***********************)

exception Free_variable_in_set of string;;
exception Incompatible_sets;;
exception Gist_existentials;;

val simplify : affine_set_builder -> affine_set_builder;;
val satisfiable_p : affine_set_builder -> bool;;
(* gist s1 s2 is s1 with all redundent constraints (wrt s2) removed. *)
val gist : 
  (*s1*)affine_set_builder -> (*s2*)affine_set_builder -> 
    affine_set_builder;;

(*
type arena = Alpha_types.arena;;
type vector = Alpha_types.vector;;
type matrix = Alpha_types.matrix;;
type affine_set_builder = Alpha_types.affine_set_builder ;;
type affine_constraint_builder = Alpha_types.affine_constraint_builder;;
type affine_constraint_kind = Alpha_types.affine_constraint_kind;;
type affine_term_builder = Alpha_types.affine_term_builder;;
*)


