/************************************************************************
 *	Copyright (C) 1998. Bernoulli Project, Cornell University.	*
 ************************************************************************/
/* Vladimir Kotlyar  February 1996
 *
 *
 *
 */

#include "./config.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "alpha_int.h"
#include "alpha_ineq.h"
#include "alpha_fmve.h"
#include "alpha_io.h"

#ifndef XA_TRACE_FMVE
#define XA_TRACE_FMVE  (0)
#endif

static int 		decide_ineq_list (arena_info *arena, int nvars, AlphaIneqList C);
static int      	decide_ineq_and_eq_list_matrix (arena_info *arena, int nvars, AlphaIneqList Cineq,
							AlphaMatrix A, AlphaVector B);
static int      	decide_ineq_and_eq_list (arena_info *arena, 
						 int nvars, AlphaIneqList Cineq, 
						 AlphaIneqList Ceq);
static int      	reduce_ineq_list (arena_info *arena, int n, AlphaIneqList C, AlphaIneqList Ceq, int expensive);
static void      	reduce_ineq_list_expensive (arena_info *arena, 
						    int n, AlphaIneqList C, AlphaIneqList Ceq);

static int		splinter (arena_info *arena, int nvars, int v_elim, int amax, 
				  AlphaIneqList C, AlphaIneqList Clower);

static xa_proj_list  	symbolic_projection (arena_info *arena, 
					     xa_proj P, xa_proj_list accumulator);
static xa_proj_list   	project_simplify_eq (arena_info *arena, 
					     AlphaVector X0, AlphaMatrix H, 
					     int nvars, 
					     AlphaIneqList C, AlphaIneqList Ceq, 
					     xa_proj_list accumulator);

static void	eliminate_variable (arena_info *arena, 
				    int v_elim,
				    AlphaIneqList C, 
				    AlphaIneqList Creal, AlphaIneqList Cdark, 
				    AlphaIneqList Clower, AlphaIneqList Cupper, 
				    int *pexact);

static xa_proj_list 	splinter_projection (arena_info *arena, 
					     int nvars, int nprot, int v_elim, 
					     int amax, 
					     AlphaVector X0, AlphaMatrix H,
					     AlphaIneqList C, AlphaIneqList Clower, 
					     xa_proj_list accumulator);


static void xa_eliminate_redundant_ineq (arena_info *arena, int n, AlphaIneqList C);

static int xa_check_redundant (arena_info *arena, 
			       int n, AlphaIneq ineq, AlphaIneqList C);

static void compress_projection (arena_info *arena, xa_proj P);

int xa_splinter_symbolic = 0;

/***********************************************************************
 *		     DECISION PROCEDURE ENTRY POINTS
 ***********************************************************************/

/*
 * A * X == B
 * F * X <= G
 */
int
a_decide_eq_and_ineq (arena_info *arena, AlphaMatrix A, AlphaVector B,
		      AlphaMatrix F, AlphaVector G)
{
  int		result;
  AlphaIneqList C;
  AlphaIneq	ineq;
  AlphaMatrix	Ft;
  int		m_eq, m_ineq;
  int		n;

  m_eq = a_rows_mx (A);
  m_ineq = a_rows_mx (F);
  n = a_cols_mx (A);

  assert (n == a_cols_mx (F));
  assert (m_eq == a_length_vec (B));
  assert (m_ineq == a_length_vec (G));

  assert (m_eq > 0);
  assert (m_ineq > 0);
  assert (n > 0);

  Ft = a_make_matrix (arena, n, m_ineq);
  a_transpose (F, Ft);

  C = a_make_ineq_list (arena);

  A_FOR_COLS_MX (Ft,col,j) {
    ineq = a_make_ineq_from_dense (arena, col, a_elt_vec(G,j));
    a_add_ineq_list (arena, C, ineq);
  }
  A_END_COLS_MX;

  result = decide_ineq_and_eq_list_matrix (arena, n, C, A, B);

  return result;
}

/*
 *  F * x <= G
 *
 */
int
a_decide_ineq (arena_info *arena, AlphaMatrix F, AlphaVector G)
{
  int		result;
  int		i, m, n;
  AlphaMatrix	Ft;
  AlphaIneq 	ineq;
  AlphaIneqList C;

  m = a_rows_mx (F);
  n = a_cols_mx (F);
  
  assert (m > 0);
  assert (n >= 0);
  assert (m == a_length_vec (G));

  if (n == 0) {
    /* just verify that G >= 0 */
    for (i = 1; i <= m; i++) {
      if (a_elt_vec (G, i) < 0) {
	return 0;
      }
    }
    return 1;
  }
  Ft = a_make_matrix (arena, n, m);
  a_transpose (F, Ft);

  result = 1;
  C = a_make_ineq_list (arena);
  A_FOR_COLS_MX (Ft,c,j) {
    ineq = a_make_ineq_from_dense (arena, c, a_elt_vec(G,j));
    a_add_ineq_list (arena, C, ineq);
  }
  A_END_COLS_MX;

  result = decide_ineq_list (arena, n, C);

  return result;
}

/***********************************************************************
 *			   ELIMINATION ENGINES
 ***********************************************************************/

#define	XA_FMVE_HT_STEP		(229)    /* some prime */

static int
reduce_ineq_list (arena_info *arena, int n, AlphaIneqList C, AlphaIneqList Ceq, int expensive)
{
  int		n_in = a_length_ineq_list (C);
  int		tsize;
  int		pos;
  int		tag;
  AlphaIneq	*table;
  AlphaIneq	ineq2;
  int		done;
  int		c, c2;
  int		contradiction;
  int		inserted;

  if (n_in == 0) {
    return 1;
  }

  tsize = n_in;
  table = a_nalloc0(arena, AlphaIneq, tsize);

  contradiction = 0;

  A_FOR_INEQ_IN_LIST(C,ineq,elt) {
    /*
     * The invariant: at each step C contains the set of inequalities which
     * is equivalen to the original one, but is possibly smaller
     */

    if (contradiction) {
      break;
    }

    tag = a_tag_ineq (ineq);
    pos = a_abs(tag) % tsize;

    if (a_nterm_ineq (ineq) == 0) {
      if (a_const_ineq (ineq) >= 0) {
	a_delete_ineq_list_elt (C, elt);
      }
      else {
	contradiction = 1;
	break;
      }
    }
    else {
      done = 0;
      inserted = 0;
      while ( !done ) {
	if (table[pos] == 0) {
	  if (!inserted) {
	    table[pos] = ineq;
	  }
	  done = 1;
	}
	else {
	  ineq2 = table[pos];
	  if ((a_tag_ineq (ineq2) == tag) && 
	      a_is_same_coeff_ineq (ineq, ineq2)) {
	    if (a_const_ineq (ineq) >= a_const_ineq (ineq2)) {
	      /* this inequality is redundant ==> remove it */
	      a_delete_ineq_list_elt (C, elt);
	      done = 1;
	    }
	    else {
	      /* this inequality tightens the existing one */
	      /* fake deleting the other one and inserting this one */
	      a_const_ineq (ineq2) = a_const_ineq(ineq);
	      a_delete_ineq_list_elt (C, elt);
	      
	      /* but continue checking into the table, 'cause might get an
	       * equality */
	      inserted = 1;
	    }
	  }
	  else if ((a_tag_ineq (ineq2) == -tag) && 
		   a_is_opposite_coeff_ineq (ineq, ineq2)){
	    c = a_const_ineq (ineq);
	    c2 = a_const_ineq (ineq2);
	    if (c > -c2) {
	      /* just add the new inequality */
	      /* for this we do nothing here */
	    }
	    else if (c == -c2) {
	      /* this creates an equality */
	      a_delete_ineq_list_elt (C, elt);
	      a_add_ineq_list (arena, Ceq, ineq);
	      done = 1;
	    }
	    else {
	      /* contradiction */
	      contradiction = 1;
	      done = 1;
	    }
	  }
	}
	pos = (pos + XA_FMVE_HT_STEP) % tsize;
      }
    }
  }
  A_END_INEQ_IN_LIST;

  if (!contradiction && expensive) {
    reduce_ineq_list_expensive (arena, n, C, Ceq);
  }

  return !contradiction;
}

static int
decide_ineq_and_eq_list (arena_info *arena, int nvars, 
			 AlphaIneqList Cineq,  /* cleared on exit */
			 AlphaIneqList Ceq     /* cleared on exit */
			 )
{
  int		meq = a_length_ineq_list (Ceq);
  AlphaMatrix	A;
  AlphaVector   B;
  int		i;
  int		result;

  assert (meq > 0);

#if XA_TRACE_FMVE
  a_log ("-----------------------------------------------------\n");
  a_log ("decide_ineq_and_eq_list:\n");
  a_log ("nvars == %d   Cineq == \n", nvars);
  a_dump_ineq_list (a_log_file, Cineq);
  a_log ("Ceq == \n", nvars);
  a_dump_ineq_list (a_log_file, Ceq);
  a_log ("-----------------------------------------------------\n");
#endif

  A = a_make_matrix (arena, meq, nvars);
  B = a_make_vector (arena, meq);

  i = 1;
  A_FOR_INEQ_IN_LIST (Ceq, eq, elt) {
    A_FOR_TERMS_IN_INEQ (eq, varind, coeff) {
      a_elt_mx (A, i, varind) = coeff;
    }
    A_END_TERMS_IN_INEQ;
    a_elt_vec (B,i) = a_const_ineq (eq);
    i++;
  }
  A_END_INEQ_IN_LIST;

  result = decide_ineq_and_eq_list_matrix (arena, nvars, Cineq, A, B);
  
  a_clear_ineq_list (Cineq);
  a_clear_ineq_list (Ceq);

  return result;
}

static int
decide_ineq_and_eq_list_matrix (arena_info *arena, int nvars, 
				AlphaIneqList Cineq, /* cleared on exit*/
				AlphaMatrix A, AlphaVector B)
{
  int		m_eq;
  int		null_dim;
  AlphaMatrix	H, H1;
  AlphaVector	X0;
  AlphaIneqList	Cnew;
  AlphaIneq	inew;
  int		rc;
  int		result;
  
  m_eq = a_rows_mx (A);

  assert (m_eq == a_length_vec (B));
  assert (m_eq > 0);
  assert (nvars == a_cols_mx (A));

#if XA_TRACE_FMVE
  a_log ("-----------------------------------------------------\n");
  a_log ("decide_ineq_and_eq_list_matrix:\n");
  a_log ("nvars == %d   Cineq == \n", nvars);
  a_dump_ineq_list (a_log_file, Cineq);
  a_log ("A == %M \nB==%V\n", A, B);
  a_log ("-----------------------------------------------------\n");
#endif

  X0 = a_make_vector (arena, nvars);
  H = a_make_matrix (arena, nvars, nvars);

  rc = a_solve_linear (arena, A, B, X0, H, &null_dim);

  if (rc == 0) {
    result = 0;
  }
  else {
    if (null_dim != 0) {
      Cnew = a_make_ineq_list (arena);
      H1 = a_make_matrix (arena, nvars, null_dim);
      a_extract (H, 1, nvars, 1, null_dim, H1);
      A_FOR_INEQ_IN_LIST (Cineq, ineq, elt) {
	inew = a_make_ineq_var_transform (arena, ineq, X0, H1);
	a_add_ineq_list (arena, Cnew, inew);
      }
      A_END_INEQ_IN_LIST;

      result = decide_ineq_list (arena, null_dim, Cnew);

      a_clear_ineq_list (Cnew);
    }
    else {
      result = 1;
    }
  }

  a_clear_ineq_list (Cineq);
  
  return result;
}

static int
decide_ineq_list (arena_info *arena, int nvars, 
		  AlphaIneqList C  /* cleared on exit */
		  )
{
  int			m;
  int			meq;
  xa_elim_info		ei;
  int			rc;
  AlphaIneqList		Ceq;
  AlphaIneqList		Cdark, Creal, Clower, Cupper;
  int			result;
  int			v_elim;
  int			exact;

#if XA_TRACE_FMVE
  a_log ("-----------------------------------------------------\n");
  a_log ("decide_ineq_list:\n");
  a_log ("nvars == %d   C == \n", nvars);
  a_dump_ineq_list (a_log_file, C);
  a_log ("\n\n");
#endif

  a_normalize_ineq_list (C);

#if XA_TRACE_FMVE
  a_log ("Normalized:\n");
  a_log ("C == \n");
  a_dump_ineq_list (a_log_file, C);
  a_log ("-----------------------------------------------------\n");
#endif


  m = a_length_ineq_list (C);
  if (m == 0) {
    return 1;
  }
  
  ei = xa_compile_einfo (arena, nvars, C);
  xa_elim_unbound (nvars, C, ei, 0);
  
  Ceq = a_make_ineq_list (arena);
  rc = reduce_ineq_list (arena, nvars, C, Ceq, 0);
  if (rc == 0) {
    result = 0;
  }
  else {
    m = a_length_ineq_list (C);
    meq = a_length_ineq_list (Ceq);
    if (meq > 0) {
      result = decide_ineq_and_eq_list (arena, nvars, C, Ceq);
    }
    else if (m == 0) {
      /* a tautology */
      result = 1;
    }
    else {
      /* we have a non-trivial and non-contradictoty set of inequalities
       * and no equalities, yet
       */
      v_elim = xa_var2elim (nvars, ei, 0);

      assert (v_elim > 0);

      Creal = a_make_ineq_list (arena);
      Cdark = a_make_ineq_list (arena);
      Clower = a_make_ineq_list (arena);
      Cupper = a_make_ineq_list (arena);

      eliminate_variable (arena, v_elim, C, Creal, Cdark, Clower, Cupper, &exact);

      if (exact) {
	result = decide_ineq_list (arena, nvars, Creal);
      }
      else {
	if (decide_ineq_list (arena, nvars, Cdark)) {
	  result = 1;
	}
	else {
	  if (!decide_ineq_list (arena, nvars, Creal)) {
	    result = 0;
	  }
	  else {
	    result = splinter (arena,
			       nvars, v_elim, xa_umax_einfo(ei,v_elim,nvars), 
			       C, Clower);
	  }
	}
      }
      
      a_clear_ineq_list (Cdark);
      a_clear_ineq_list (Creal);
      a_clear_ineq_list (Clower);
      a_clear_ineq_list (Cupper);
      a_clear_ineq_list (C);

    }
  }

  a_clear_ineq_list (Ceq);
  
  return result;
}

static int	
splinter (arena_info *arena, int nvars, int v_elim, int amax, AlphaIneqList C, AlphaIneqList Clower)
{
  AlphaIneqList	Ccopy;
  AlphaIneqList	Ceq;
  AlphaIneq	icopy;
  int		b;
  int		c;
  int		limit;
  int		i;
  int		result;

  Ceq = a_make_ineq_list (arena);

#if XA_TRACE_FMVE
  a_log ("*************** SPLINTERING *****************\n");
  a_log ("C == \n");
  a_dump_ineq_list (a_log_file, C);
  a_log ("Clower == \n");
  a_dump_ineq_list (a_log_file, Clower);
  a_log ("\n\n");
#endif

  result = 0;
  A_FOR_INEQ_IN_LIST (Clower, ineq, elt) {
    if (result != 0) {
      break;
    }

    b = a_coeff4var_ineq (ineq, v_elim);
    c = a_const_ineq (ineq);
    assert (b > 0);

    limit = a_floor_div(amax * b - amax - b, amax);

    for (i = 0; (result == 0) && (i <= limit); i++) {

#if XA_TRACE_FMVE
      a_log ("lower bound: %I  i = %d\n", ineq, i);
#endif

      Ccopy = a_make_ineq_list_copy (arena, C);
      icopy = a_make_ineq_copy (arena, ineq);
      a_const_ineq (icopy) = c - i;
      a_add_ineq_list (arena, Ceq, icopy);
      result = decide_ineq_and_eq_list (arena, nvars, Ccopy, Ceq);

      assert(0 == a_length_ineq_list (Ccopy));
      assert(0 == a_length_ineq_list (Ceq));
      
#if XA_TRACE_FMVE
      a_log ("------------------->>>>>>>>>>> result %d\n", result);
#endif
    }

  }
  A_END_INEQ_IN_LIST;

  return result;  
}

static void
eliminate_variable (arena_info *arena, int v_elim,
		    AlphaIneqList C, 
		    AlphaIneqList Creal, AlphaIneqList Cdark, 
		    AlphaIneqList Clower, AlphaIneqList Cupper, 
		    int *pexact)
{
  AlphaIneq	ineq2;
  AlphaIneq	inew;
  AlphaIneq	inew_dark;
  int 		a, b, c_prime, c_lower, c_upper, g;
  int		exact;

  A_FOR_INEQ_IN_LIST (C, ineq, elt) {
    a = a_coeff4var_ineq (ineq, v_elim);
    if (a == 0) {
      a_add_ineq_list (arena, Creal, ineq);
      ineq2 = a_make_ineq_copy (arena, ineq);
      a_add_ineq_list (arena, Cdark, ineq2);
    }
    else if (a > 0) {
      a_add_ineq_list (arena, Cupper, ineq);
    }
    else {
      a_add_ineq_list (arena, Clower, ineq);
    }
  }
  A_END_INEQ_IN_LIST;

  /* at this point Creal and Cdark have inequalities that do not
   * involve v_elim. In fact they have the same inequalities */
      
  exact = 1;

  A_FOR_INEQ_IN_LIST (Clower, ilower, elt_lower) {
    b = -a_coeff4var_ineq (ilower, v_elim);
    assert (b > 0);
    c_lower = a_const_ineq (ilower);
    A_FOR_INEQ_IN_LIST (Cupper, iupper, elt_upper) {
      a = a_coeff4var_ineq (iupper, v_elim);
      assert (a > 0);
      c_upper = a_const_ineq (iupper);
      /*
       * The correction for the dark shadow constant part it 
       *  -(a-1)*(b-1)
       */
      
      inew = a_make_ineq_saxpy (arena, b, iupper, a, ilower);
      c_prime = a_const_ineq(inew) - (a-1)*(b-1);
      g = a_normalize_ineq (inew);
      c_prime = a_floor_div (c_prime, g);

      a_add_ineq_list (arena, Creal, inew);
      
      exact = exact && (c_prime == a_const_ineq(inew));

      inew_dark = a_make_ineq_copy (arena, inew);
      a_const_ineq (inew_dark) = c_prime;
      a_add_ineq_list (arena, Cdark, inew_dark);

    }
    A_END_INEQ_IN_LIST;
  }
  A_END_INEQ_IN_LIST;

  *pexact = exact;
}

/***********************************************************************
 *     FUNCTIONS TO COMPILE INFORMATION ABOUT ELIMINATED VARIABLES
 ***********************************************************************/

xa_elim_info 	
xa_compile_einfo (arena_info *arena, int nvars, AlphaIneqList C)
{
  xa_elim_info	ei = xa_make_einfo (arena, nvars);

  A_FOR_INEQ_IN_LIST (C, ineq, elt) {
    A_FOR_TERMS_IN_INEQ (ineq, varind, coeff) {
      if (coeff > 0) { /* upper bound */
	xa_umax_einfo(ei,varind,nvars) = 
	  a_max(xa_umax_einfo(ei,varind,nvars), coeff);

	xa_n_upper_einfo(ei,varind,nvars) ++;
      }
      else if (coeff < 0) { /* lower bound */
	xa_lmax_einfo(ei,varind,nvars) = 
	  a_max(xa_lmax_einfo(ei,varind,nvars), -coeff);

	xa_n_lower_einfo(ei,varind,nvars) ++;
      }
    }
    A_END_TERMS_IN_INEQ;
  }
  A_END_INEQ_IN_LIST;

  return ei;
}

void
xa_elim_unbound (int nvars, AlphaIneqList C, 
		 xa_elim_info ei, int nprot)
{
  int		remove_this;

  A_FOR_INEQ_IN_LIST (C, ineq, elt) {
    remove_this = 0;
    A_FOR_TERMS_IN_INEQ (ineq, varind, coeff) {
      remove_this = remove_this || 
	((varind > nprot) && xa_is_unbound_einfo (ei, varind, nvars));
    }
    A_END_TERMS_IN_INEQ;
    if (remove_this) {
      a_delete_ineq_list_elt (C, elt);
    }
  }
  A_END_INEQ_IN_LIST;
}

/*
 * returns 0 if no more unprotected variables to eliminate
 */
int
xa_var2elim (int nvars, xa_elim_info ei, int nprot)
{
  int		v;
  int		v_elim;
  int		n_proj = -1;
  int		c, min_c = -1;

  v_elim = 0;
  for (v = nprot+1; v <= nvars; v++) {
    if (xa_is_exact_einfo (ei,v,nvars)) {
      if ((v_elim == 0) || (n_proj > xa_n_proj_einfo(ei,v,nvars))) {
	v_elim = v;
	n_proj = xa_n_proj_einfo(ei,v,nvars);
      }
    }
  }

  if (v_elim != 0) {
    return v_elim;
  }

  for (v = nprot+1; v <= nvars; v++) {
    if (!xa_is_unbound_einfo (ei,v,nvars)) {
      c = xa_lmax_einfo(ei,v,nvars) + xa_umax_einfo(ei,v,nvars);
      if ((v_elim == 0) || (c < min_c)) {
	v_elim = v;
	min_c = c;
      }
    }
  }

  return v_elim;
}


/***********************************************************************
 *		   SYMBOLIC PROJECTION ENTRY POINTS
 ***********************************************************************/

xa_proj_list
a_symbolic_projection (arena_info *arena, int nprot, 
		       AlphaMatrix A, AlphaVector B,
		       AlphaMatrix F, AlphaVector G)
{
  xa_proj_list		accumulator;
  AlphaIneqList		C, Ceq;
  AlphaIneq		ineq;
  AlphaVector		X0;
  AlphaMatrix		H;
  int			nvars = a_cols_mx (A);
  int			m_eq = a_rows_mx (A);
  int			m_ineq = a_rows_mx (F);
  AlphaProjList		l;
  AlphaProjection	P;

  assert (m_eq == a_length_vec (B));
  assert (m_ineq == a_length_vec (G));
  assert (nprot >= 0);
  assert (nprot <= nvars);

  C = a_make_ineq_list (arena);
  a_matrix_form2ineq_list (arena, F, G, C);

  Ceq = a_make_ineq_list (arena);
  a_matrix_form2ineq_list (arena, A, B, Ceq);

  X0 = a_make_vector (arena, nprot); /* initialized to 0 */
  H = a_make_matrix (arena, nprot, nprot);
  a_set2ident (H);

  accumulator = project_simplify_eq (arena, X0, H, nvars, C, Ceq, 0);

  for (l = accumulator; l ; l = a_cdr_proj_list (l)) {
    P = a_car_proj_list (l);
    if (a_ineq_list_proj (P)) {
      compress_projection (arena, P);
      xa_eliminate_redundant_ineq (arena, nvars, (a_ineq_list_proj (P)));
    }
  }

  return accumulator;
}



/***********************************************************************
 *		     SYMBOLIC PROJECTION ENGINES
 ***********************************************************************/

xa_proj_list 
symbolic_projection (arena_info *arena, 
		     xa_proj P, /* destroyed or added to accumulator on exit */
		     xa_proj_list accumulator)
{
  AlphaVector		X0 = xa_X0_proj (P);
  AlphaMatrix		H = xa_H_proj (P);
  AlphaIneqList		C = xa_ineq_list_proj (P);
  int			nvars = xa_nvars_proj (P);
  int			nprot = xa_nprot_proj (P);
  xa_proj		Q;
  xa_elim_info		ei;
  int			rc;
  AlphaIneqList		Ceq;
  AlphaIneqList		Cdark, Creal, Clower, Cupper, Ccopy;
  int			m, meq;
  int			v_elim;
  int			exact;
  
  assert (nprot <= nvars);
  assert (nprot >= 0);
  assert (nprot == a_cols_mx (H));
  assert (a_length_vec (X0) == a_rows_mx (H));


  ei = xa_compile_einfo (arena, nvars, C);
  xa_elim_unbound (nvars, C, ei, nprot);
  Ceq = a_make_ineq_list (arena);
  rc = reduce_ineq_list (arena, nvars, C, Ceq, 1);
  if (rc != 0) {
    m = a_length_ineq_list (C);
    meq = a_length_ineq_list (Ceq);
    if (meq > 0) {
      accumulator = project_simplify_eq (arena, X0, H, nvars, C, Ceq, accumulator);
    }
    else if (m == 0) {
      /* we have conjunct that says: X = X0 + H*t and t is unbound */
      Q = xa_make_hyperplane_proj (arena, X0, H);
      xa_add_to_acc_proj (arena, Q, accumulator);
    }
    else {
      /* we have a non-trivial and non-contradictoty set of inequalities
       * and no equalities, yet
       */
      
      if (nprot == nvars) {
	Ccopy = a_make_ineq_list_copy (arena, C);
	if (decide_ineq_list (arena, nvars, Ccopy)) {
	  xa_add_to_acc_proj (arena, P, accumulator);
	}
      }
      else {
	v_elim = xa_var2elim (nvars, ei, nprot);
	if (v_elim > 0) {
	  Creal = a_make_ineq_list (arena);
	  Cdark = a_make_ineq_list (arena);
	  Clower = a_make_ineq_list (arena);
	  Cupper = a_make_ineq_list (arena);

	  eliminate_variable (arena, v_elim, C, Creal, Cdark, Clower, Cupper, &exact);

	  if (exact) {
	    Q = xa_make_proj_build (arena, nvars, nprot, X0, H, Creal);
	    accumulator = symbolic_projection (arena, Q, accumulator);
	  }
	  else {
	    /*
	     * And perform splintering 
	     */
	    if (xa_splinter_symbolic) {
	      accumulator = splinter_projection (arena, nvars, nprot, v_elim, 
						 xa_umax_einfo(ei,v_elim,nvars),
						 X0, H,
						 C, Clower, accumulator);
	    }
	    else {
	      assert (nprot > 0);
	      Q = xa_make_proj_build (arena, nvars, nprot, X0, H, C);
	      xa_add_to_acc_proj (arena, Q, accumulator);
	    }
	  }

	  a_clear_ineq_list (Cdark);
	  a_clear_ineq_list (Creal);
	  a_clear_ineq_list (Clower);
	  a_clear_ineq_list (Cupper);
	}
	else {
	  /* after all cleaning up we were left with no variables to
	   * project away: just report the solution */
	  xa_add_to_acc_proj (arena, P, accumulator);
	}
      }
    }
  }

  a_clear_ineq_list (Ceq);

  a_clear_ineq_list (C);
  
  return accumulator;
}

xa_proj_list
project_simplify_eq (arena_info *arena, AlphaVector X0, AlphaMatrix H, 
		     int nvars,
 		     AlphaIneqList C, AlphaIneqList Ceq,
		     xa_proj_list accumulator)
{
  int		nprot = a_cols_mx (H);
  int		nx0 = a_length_vec (X0);
  AlphaVector   T0, T0_1, X0prime, HT0, HT0_1;
  AlphaMatrix   G_1, G_1part, Gsquare, G, HG;
  AlphaMatrix   U, F;
  AlphaMatrix	A;
  AlphaIneqList Cnew;
  AlphaIneq	inew;
  AlphaVector   B;
  int		meq = a_length_ineq_list (Ceq);
  int		i;
  int		null_dim, r;
  int		rc;
  xa_proj	R;

  assert (nvars > 0);
  assert (nprot >= 0);
  assert (nx0 >= 0);
  
  if (nprot == 0) {
    /*
     * We have a conjunct of the form
     *   x = x0
     *     +
     *   inequalities on unprotected t
     * 
     * So we just decide the inequalities and
     * if they are satisfiable, we report x0
     */
    if (decide_ineq_and_eq_list (arena, nvars, C, Ceq)) {
      R = xa_make_single_value_proj (arena, X0);
      xa_add_to_acc_proj (arena, R, accumulator);
    }
    else {
      /* do nothing, this conjunct is empty */
    }
  }
  else {
    assert (nprot <= nvars);

    A = a_make_matrix (arena, meq, nvars);
    B = a_make_vector (arena, meq);
    T0 = a_make_vector (arena, nvars);
    Gsquare = a_make_matrix (arena, nvars, nvars);

    a_ineq_list2matrix_form (Ceq, A, B);
    
    rc = a_solve_linear (arena, A, B, T0, Gsquare, &null_dim);
    
    if (rc == 0) { /* no solutions */
      /* nothing to do here */
      /* the accumulator stays the same */
    }
    else {
      if (null_dim == 0) {
	/* a unique solution found */
	/* so we have 
	 *   x == x0 + H*t
	 *   t == t0
	 *   P*t <= Q
	 *
	 * We just check if t0 satisfies the inequalities
	 * and then report the solution or not
	 */
	if (a_is_satisfied_ineq_list (C, T0)) {
	  /* report X = X0 + H*T0 as the solution */
	  X0prime = a_make_vector (arena, nx0);
	  HT0 = a_make_vector (arena, nx0);
	  a_mvm (arena, H, T0, HT0);
	  a_saxpy (1, X0, 1, HT0, X0prime);

	  R = xa_make_single_value_proj (arena, X0prime);
	  xa_add_to_acc_proj (arena, R, accumulator);

	}
	else {
	  /* do nothing the accumulator is unchanged */
	}
      }
      else {
	G = a_make_matrix (arena, nvars, null_dim);
	a_extract (Gsquare, 1, nvars, 1, null_dim, G);
	
	if (nprot == nvars) {
	  /* we just eliminate the equalities 
	   * X0 becomes X0 + H*T0
	   * H becomes H*G
	   * and the inequalities get transformed accordingly
	   */
	  X0prime = a_make_vector (arena, nx0);
	  HT0 = a_make_vector (arena, nx0);
	  a_mvm (arena, H, T0, HT0);
	  a_saxpy (1, X0, 1, HT0, X0);

	  HG = a_make_matrix (arena, nx0, null_dim);
	  a_matmult (arena, H, G, HG);

	  Cnew = a_make_ineq_list (arena);
	  a_transform_ineq_list (arena, C, T0, G, Cnew);

	  R = xa_make_proj_build (arena, null_dim, null_dim, X0, HG, Cnew);
	  xa_add_to_acc_proj (arena, R, accumulator);

	}
	else {
	  G_1 = a_make_matrix (arena, nprot, null_dim);
	  T0_1 = a_make_vector (arena, nprot);
	  a_extract (G, 1, nprot, 1, null_dim, G_1);
	  a_extract_vec (T0, 1, nprot, T0_1);
	  U = a_make_matrix (arena, null_dim, null_dim);
	  a_lower_echelon (arena, G_1, U, &r);
	  G_1part = a_make_matrix (arena, nprot, r);
	  if (r > 0) {
	    /* extract fails when r == 0 */
	    a_extract (G_1, 1, nprot, 1, r, G_1part);
	  }
	  
	  /* r is the number of protected vars in the
	   * new problem 
	   */

	  F = a_make_matrix (arena, nvars, null_dim);

	  a_matmult (arena, G, U, F);

	  /*
	   * The new substitution is:
	   *  X = X0 + H*T = X0 + H(T0_1 + L*W_1) = 
	   *    = (X0 + H*T0_1) + H*G_1part*W_1
	   *
	   * The transformation for the inequalities is:
	   *  T ---> T0 + G*U*W
	   *       or
	   *  T ---> T0 + F*W
	   */

	  X0prime = a_make_vector (arena, nx0);
	  HT0_1 = a_make_vector (arena, nx0);
	  a_mvm (arena, H, T0_1, HT0_1);
	  a_saxpy (1, X0, 1, HT0_1, X0prime);
	  
	  Cnew = a_make_ineq_list (arena);
	  a_transform_ineq_list (arena, C, T0, F, Cnew);

	  HG = a_make_matrix (arena, nx0, r);
	  a_matmult (arena, H, G_1part, HG);
	  
	  R = xa_make_proj_build (arena, null_dim, r, X0prime, HG, Cnew);

	  accumulator = symbolic_projection (arena, R, accumulator);

	}
      }
    }

  }

  a_clear_ineq_list (Ceq);
  a_clear_ineq_list (C);

  return accumulator;
}

static xa_proj_list 	
splinter_projection (arena_info *arena, int nvars, int nprot, int v_elim, 
		     int amax, 
		     AlphaVector X0, AlphaMatrix H,
		     AlphaIneqList C, AlphaIneqList Clower, 
		     xa_proj_list accumulator)
{
  int		b;
  int		c;
  int		limit;
  int		i;
  AlphaIneqList	Ccopy;
  AlphaIneqList Ceq;
  AlphaIneq	icopy;

  Ceq = a_make_ineq_list (arena);

  A_FOR_INEQ_IN_LIST (Clower, ineq, elt) {
    b = -a_coeff4var_ineq (ineq, v_elim);
    c = a_const_ineq (ineq);
    assert (b > 0);

    limit = a_floor_div (amax * b - amax - b, amax);

    for (i = 0; i <= limit; i++) {
      Ccopy = a_make_ineq_list_copy (arena, C);
      icopy = a_make_ineq_copy (arena, ineq);
      a_const_ineq (icopy) = c - i;
      a_add_ineq_list (arena, Ceq, icopy);

      accumulator = project_simplify_eq (arena, X0, H, nvars, Ccopy, Ceq, accumulator);

      assert(0 == a_length_ineq_list(Ceq));
    }
  }
  A_END_INEQ_IN_LIST;

  return accumulator;
}



/***********************************************************************
 *		      PROJECTION DATA STRUCTURES
 ***********************************************************************/


xa_proj 
xa_make_proj (arena_info *arena)
{
  xa_proj P;

  P = a_alloc0 (arena, xa_proj_t);
  
  return P;
}

xa_proj 
xa_make_proj_build (arena_info *arena, int nvars, int nprot, AlphaVector X0, AlphaMatrix H, AlphaIneqList C)
{
  xa_proj P;

  P = xa_make_proj(arena);

  assert (X0);
  assert (H);
  assert (C);

  xa_X0_proj (P) = a_make_vector (arena, a_length_vec (X0));
  a_copy_vec (xa_X0_proj (P), X0);

  assert (a_cols_mx (H) == nprot);
  assert (a_rows_mx (H) == a_length_vec (X0));
  xa_H_proj (P) = a_make_matrix (arena, a_rows_mx(H), a_cols_mx(H));
  a_copy_mx (xa_H_proj (P), H);
   
  xa_ineq_list_proj (P) = a_make_ineq_list_copy (arena, C);
  A_FOR_INEQ_IN_LIST(xa_ineq_list_proj (P), ineq, elt) {
    if (!a_is_normalized_ineq (ineq)) {
      a_normalize_ineq (ineq);
    }
  }
  A_END_INEQ_IN_LIST;

  xa_nprot_proj (P) = nprot;
  xa_nvars_proj (P) = nvars;

  xa_next_proj (P) = 0;

  return P;
}

xa_proj
xa_make_proj_copy (arena_info *arena, xa_proj P)
{
  xa_proj Q;

  Q = xa_make_proj_build (arena,
			  xa_nvars_proj (P),
			  xa_nprot_proj (P),
			  xa_X0_proj (P),
			  xa_H_proj (P),
			  xa_ineq_list_proj(P));

  return Q;
}

xa_proj 
xa_make_single_value_proj (arena_info *arena, AlphaVector X0)
{
  AlphaMatrix H = a_make_matrix (arena, a_length_vec(X0), 0);
  AlphaIneqList C = a_make_ineq_list(arena);
  xa_proj P = xa_make_proj_build (arena, 0, 0, X0, H, C);

  return P;
}

xa_proj 
xa_make_hyperplane_proj (arena_info *arena, AlphaVector X0, AlphaMatrix H)
{
  xa_proj P;
  AlphaIneqList C = a_make_ineq_list(arena);
  P = xa_make_proj_build (arena, a_cols_mx(H), a_cols_mx(H), X0, H, C);
  return P;
}

void
xa_dump_proj (FILE *fp, xa_proj P)
{
  fprintf (fp, "PROJECTION P==%p  nvars==%d  nprot==%d\n",
	   P, xa_nvars_proj (P), xa_nprot_proj (P));

  a_fprintf (fp, "X0 == %V\n", xa_X0_proj (P));
  a_fprintf (fp, "H == %M\n", xa_H_proj (P));
  a_fprintf (fp, "C == \n");
  a_dump_ineq_list (fp, xa_ineq_list_proj (P));

  fprintf (fp, "END PROJECTION\n");
}

void 
a_print_proj (FILE *fp, xa_proj P)
{
  fprintf (fp, "[[ ");

  a_fprintf (fp, "X0 == %V  ", xa_X0_proj (P));
  a_fprintf (fp, "H == %M  ", xa_H_proj (P));
  a_fprintf (fp, "C == ");
  a_print_ineq_list (fp, xa_ineq_list_proj (P));

  assert (xa_nvars_proj (P) >= xa_nprot_proj (P));

  fprintf (fp, " protected = %d nvars = %d ", xa_nprot_proj (P), xa_nvars_proj (P));

  fprintf (fp, "]]\n");  
}

void
xa_dump_proj_list (FILE *fp, xa_proj_list l)
{
  fprintf (fp, "PROJECTION LIST %p\n", l);

  while (l) {
    xa_dump_proj (fp, xa_car_proj_list (l));
    l = xa_cdr_proj_list (l);
  }

  fprintf (fp, "END PROJECTION LIST\n");  
}

void
a_print_proj_list (FILE *fp, AlphaProjList l)
{
  if (0 == l) {
    fprintf (fp, "(PROJ LIST)0\n");
  }
  else {
    fprintf (fp, "<< ");
    while (l) {
      a_print_proj (fp, a_car_proj_list (l));
      fprintf (fp, " , ");
      l = a_cdr_proj_list (l);
    }
    fprintf (fp, " >>");
  }
}

/***************************************************************************
 * Simplification engines
 ***************************************************************************/

static void
xa_eliminate_redundant_ineq (arena_info *arena, int n, AlphaIneqList C)
{
  AlphaIneqList Ccopy = a_make_ineq_list(arena);

#if XA_TRACE_FMVE
  a_log ("-------------------------\n");
  a_log ("Checking redundant list: \n");
  a_dump_ineq_list (a_log_file, C);
#endif
  
  A_FOR_INEQ_IN_LIST (C,ineq,elt) {
    A_FOR_INEQ_IN_LIST (C,i2,elt2) {
      if (! a_is_equal_ineq_list_elt(elt, elt2)) {
	a_add_ineq_list (arena, Ccopy, a_make_ineq_copy (arena, i2));
      }
    }
    A_END_INEQ_IN_LIST;
    if (xa_check_redundant (arena, n, ineq, Ccopy)) {
      a_delete_ineq_list_elt (C, elt);
    }

    a_clear_ineq_list (Ccopy);
  }
  A_END_INEQ_IN_LIST;    

#if XA_TRACE_FMVE
  a_log ("Simplified: \n");
  a_dump_ineq_list (a_log_file, C);
#endif

}

static void
xa_find_equalities_expensive (arena_info *arena, int n, AlphaIneqList C, AlphaIneqList Ceq)
{
  AlphaIneqList Ccopy = a_make_ineq_list(arena);
  AlphaIneq  ineq2;

#if XA_TRACE_FMVE
  a_log ("-------------------------\n");
  a_log ("Checking looking for equalities in: \n");
  a_dump_ineq_list (a_log_file, C);
#endif
  
  A_FOR_INEQ_IN_LIST (C,ineq,elt) {
    A_FOR_INEQ_IN_LIST (C,i2,elt2) {
      if (! a_is_equal_ineq_list_elt(elt, elt2)) {
	a_add_ineq_list (arena, Ccopy, a_make_ineq_copy (arena, i2));
      }
    }
    A_END_INEQ_IN_LIST;

    ineq2 = a_make_ineq_copy (arena, ineq);
    a_const_ineq (ineq2) -= 1;

    a_add_ineq_list (arena, Ccopy, ineq2);
    if (! decide_ineq_list (arena, n, Ccopy)) {
      /* actually is an equality */
      a_add_ineq_list (arena, Ceq, ineq);
      a_delete_ineq_list_elt (C, elt);
    }

    a_clear_ineq_list (Ccopy);
  }
  A_END_INEQ_IN_LIST;    

#if XA_TRACE_FMVE
  a_log ("Simplified: \n");
  a_dump_ineq_list (a_log_file, C);
  a_dump_ineq_list (a_log_file, Ceq);
#endif

}


static int
xa_check_redundant  (arena_info *arena, int n, AlphaIneq ineq, AlphaIneqList C)
{
  AlphaIneq	ineg = a_make_ineq_neg (arena, ineq);

  a_add_ineq_list (arena, C, ineg);

  return ! decide_ineq_list (arena, n, C); /* ineg is deallocated along with the rest of C */
}

static void
reduce_ineq_list_expensive (arena_info *arena, int n, AlphaIneqList C, AlphaIneqList Ceq)
{
  AlphaIneqList Ccopy = a_make_ineq_list(arena);

  xa_find_equalities_expensive (arena, n, C, Ceq);
  xa_eliminate_redundant_ineq (arena, n, C);
}

static void
compress_projection (arena_info *arena, xa_proj P)
{
  int		n = xa_nvars_proj (P);
  int		n2;
  int		nprot = xa_nprot_proj (P);
  AlphaIneqList C = xa_ineq_list_proj (P);
  AlphaVector	perm = a_make_vector (arena, n);
  int 		i, ii;
  int		v;

  assert (C);

  /*
   * first mark all used variables 
   */
  for (i = 1; i <= nprot; i++) { /* protected vars are all marked */
    a_elt_vec(perm, i) = 1;
  }

  A_FOR_INEQ_IN_LIST (C, ineq, elt) {
    A_FOR_TERMS_IN_INEQ (ineq, v, c) {
      a_elt_vec (perm, v) = 1;
    }
    A_END_TERMS_IN_INEQ;
  }
  A_END_INEQ_IN_LIST;

  /*
   * assign new numbers to the used variables 
   */
  for (i = 1, ii=1; i <= n; i++) {
    if (1 == a_elt_vec (perm,i)) {
      a_elt_vec (perm,i) = ii;
      ii++;
    }
  }
  
  n2 = ii - 1;  /* number of new vars */
  assert (n2 >= nprot);

  /*
   * now renumber the inequalities
   */
  A_FOR_INEQ_IN_LIST (C, ineq, elt) {
    for (i = 1; i < a_nterm_ineq(ineq); i++) {
      v = a_elt_vec(perm, a_varind_ineq(ineq,i));
      a_varind_ineq(ineq,i) = v;
    }
  }
  A_END_INEQ_IN_LIST;

  xa_nvars_proj (P) = n2;
}



