(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers

open Bmf;;

type black_box_maker = keyword_map -> black_box
and keyword_list = (string * annotation) list
and keyword_map = annotation String_map.t
and black_box = {
    bb_index_fields : string list;
    bb_value_field : string;
    bb_bounds_maker : bounds_method;
    bb_access_methods : access_method list;
    bb_ordered_access : (string * string) list
}
and bounds_method = bmf_expr list -> bmf_expr
and access_method = {
    am_name : string;
    am_in_fields : string list;
    am_out_field : string;
    am_info : access_method_kind
}
and access_method_kind =
    Am_fun of am_fun_type
|   Am_rel of am_rel_type
and am_fun_type =
    am_fun_cost    (* cost of access *)
  * am_fun_kind    (* func for generating code *)

and am_fun_cost = O_1 | O_logn | O_n

and am_fun_kind =
    Am_fun_search of am_search_type
|   Am_fun_lookup of am_lookup_type
and am_search_type =
    bmf_expr list               (* args *)
    -> ( bmf_expr -> bmf_stmt ) (* found_f *)
    -> ( unit -> bmf_stmt )     (* not_found_f *)
    -> bmf_stmt
and am_lookup_type =
    bmf_expr list               (* args *)
    -> ( bmf_expr -> bmf_stmt)  (* found_f *)
    -> bmf_stmt
and am_rel_type =
    Am_rel_interval of am_range_type
|   Am_rel_general of am_stream_type
and am_range_type =
    bmf_expr list
    -> ( bmf_expr(*lb*) -> bmf_expr(*ub*) -> bmf_stmt )
    -> bmf_stmt
and am_stream_type =
    bmf_expr list
    -> am_handler_record
and am_handler_record = {
    decl_ids : bmf_ident list;
    init_st : bmf_stmt;
    incr_st : bmf_stmt;
    close_st : bmf_stmt;
    valid_ex : bmf_expr;
    deref_ex : bmf_expr }
;;

exception Insane_black_box of string;;

let ht = String_hashtbl.create 101;;

let register (name: string) (maker: black_box_maker) = (
    String_hashtbl.add ht name maker
: unit);;

let is_registered (name: string) = (
    try
      (String_hashtbl.find ht name); true
    with
      Not_found -> false
: bool);;

(* find_registered returns a "wrapper" to the black_box_ maker that checks
   the sanity of the generated black-box. *)

let rec find_registered (name: string) = (
    maker_wrapper name (String_hashtbl.find ht name)
: black_box_maker)

and maker_wrapper name maker keywords =
    let bb = maker keywords
    in
	zassert (sane_bb_p bb) (Insane_black_box(name));
	bb

and sane_bb_p ({
	bb_index_fields = index_fields;
	bb_value_field = value_field;
	bb_access_methods = access_methods }: black_box) =

    (* All fields that appear in the access methods. *)
    let mthd_fields =
	List.fold_left
	    (fun results { am_in_fields = in_fields;
			   am_out_field = out_field } ->
		List'.union results (out_field::in_fields))
	    [] access_methods
    in

    (* Check that all fields mentioned anywhere in the bb are reachable
       from a traversal of the access methods. *)
    sane_bb_iter
	(List'.union index_fields (List'.union [value_field] mthd_fields))
	access_methods

and sane_bb_iter unbound_fields access_methods =

    let bound_field_p f = (not ((List.mem f unbound_fields)))
    in let ready_method_p am =
	(List.for_all bound_field_p am.am_in_fields)
	&& (not ((bound_field_p am.am_out_field)))
    in let ready_methods = List'.remove_if_not ready_method_p access_methods
    in
    if ready_methods = [] then
	(* There are no more ready methods. The bb is sane if all of
	   its fields have been bound. *)
	unbound_fields = []
    else
	(* Remove the out fields of the ready methods from the list of
	   unbound fields, and iterate. *)
	sane_bb_iter
	    (List'.difference unbound_fields
		(List.fold_left List'.union []
		    (List.map (fun am -> [am.am_out_field])
			ready_methods)))
	    access_methods

;;

