(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers
open Bmf;;

(************************************************************************)
(* Overview								*)
(************************************************************************)

(************************************************************************
The concrete array type is the only kind of array for which the compiler
has a built-in storage implementation. All other storage implementations
are provided to the compiler as modules that implement a defined interface,
the black-box protocol.

Here are the steps that occurs when the compiler processes a Array_bb into a
black-box object.

1. The black-box name and list of keywords and values, hereafter called the   
   keyword arguments, are extracted from the Array_bb.                     

2. The compiler searches for a black-box maker that is registered under
   the given black-box name. If not such maker is found, then the compiler 
   signals an error.                                                      

3. The keyword arguments are inserted into an associate map, which has 
   type [[keyword_map]]. The black-box maker is called with this map as its
   sole argument.                                                         

4. The black-box maker, using the keyword map, constructs a black-box 
   and returns it to the compiler.                                        

************************************************************************)

(************************************************************************)
(* Registering the black box						*)
(************************************************************************)

(************************************************************************
A module registers a function of type [[black_box_maker]] with the       
compiler, and it is intended to construct black-boxes that describe the  
implementation for a particular abstract array.                          
************************************************************************)
type black_box_maker = keyword_map -> black_box
and keyword_list = (string * annotation) list
and keyword_map = annotation String_map.t

(************************************************************************)
(* Black-box objects							*)
(************************************************************************)

(************************************************************************
A black-box object contains all of the information about a particular   
abstract array that the compiler needs to schedule, optimize, and generate
references to the array. The abstraction that is presented to the compiler
is that the abstract array is a relation, with                          

+ a set of fields, and 
+ a set of of methods for accessing relation's tuples (the
  \textit{access methods}).                                             

The fields of a [[black_box]] record contain the following information, 

o [[bb_index_fields]] and [[bb_value_field]]: These two values      
  constitute the mapping, as described in Section~\ref{s-163}, between the
  positions of an array reference and the fields of the relation        
  representing the array. [[bb_index_fields]] are the fields that match the
  array reference indices that appear in the user's source code, and    
  [[bb_value_field]] is the field that contains values of the reference.
  That is, if array [[A]] has index fields [[A.i]] and [[A.j]], and value
  field [[A.v]], then the array reference [[A[x,y]]] will match the tuple
  in [[A]] whose [[A.i]] field is [[x]] and whose [[A.j]] field is [[y]],
  and will refer to the location of the [[A.v]] field within that tuple.

o [[bb_bounds_maker]]: This is a function for generating an         
  expression that describes the bounds on the index fields of the relation.
  More specifically, this field contains a function that takes a list of
  expressions, one for each index field, and returns a set expression that
  describes the contraints on the index fields in terms of the parameter
  expressions. For instance, if [[A]] is a square [[n]]$\times$[[n]]    
  matrix, and if its bounds maker is called with the expressions [[i]] and
  [[j]], then it might return the BMF expression,                       

    { [i,j] | 0 <= i && i <= n-1 && 0 <= j && j <= n-1  }                   

  Note that a bounds maker does not have to return rectangular bounds. If
  [[B]] was only defined for the lower triangle of [[A]]'s bounds, and its
  bounds maker is called with the expressions [[p]] and [[q]], it might 
  return the BMF expression,                                            

    { [p,q] | 0 <= p && p <= n-1 && 0 <= q && q <= n-1                      
		&& p >= q }                                                 

  There is nothing within the protocol that ensures that these contrains
  are even linear. However, in practice, they must be.                  

o [[bb_access_methods]]: A list of the methods that                 
  are available for accessing the tuples of the relation, which are     
  described below                                                       
************************************************************************)

and black_box = {
    bb_index_fields : string list;
    bb_value_field : string;
    bb_bounds_maker : bounds_method;
    bb_access_methods : access_method list;
    bb_ordered_access : (string * string) list
}
and bounds_method = bmf_expr list -> bmf_expr


(************************************************************************)
(* Access methods							*)
(************************************************************************)

(************************************************************************
The access methods are the mechanisms by which tuples of a relation are
accessed.

The [[am_name]] field is the name of the access method. This name, which
has no connection with either the array's name or the names of the array's
fields, is simply a tag for refering to a particular access method. When a
reference to a access method appears in a BMF program, the access method's
name appears as the reference's first argument. For instance,

    A[["lookup_i", i, jj]]                                                   

refers to the result of invoking the access method within [[A]]'s        
black-box whose [[am_name]] field is [["lookup_i"]], with the arguments  
[[i]] and [[jj]].                                                        

When an access method is invoked, the method returns all of the values of 
the [[am_out_field]] of all of the tuples in which the [[am_in_fields]]   
match the arguments to the access method. This can be expressed more     
clearly using the relational algebra: the reference                      
[[A[["lookup_i",i,jj]]]], where the [["lookup_i"]] method of [[A]] has   
infields [[A1]] and [[A2]], and output field [[A3]], is equivalent to    

    \Project[{[[A3]]}]\Select[{[[A1]]=[[i]],[[A2]]=[[jj]]}][[A]]           
                                                                         
The [[am_info]] field indicates the cardinality of the result of the access
method. In general, an access method will return a stream of all of the
tuples that match the values of the input fields. This is indicated by an
[[am_rel]] value in the [[am_info]] field. But sometimes it is the case
that the values of the input fields uniquely determine the value of the
output field. That is, only one tuple will be found in the relation for the
any combination of values of the infields. In this case, for a set of
values of the input fields, the access method could only return a single
value for the output field. this single result is refered to as a
singleton, and this type of constraint on the values of the fields of the
relation is refered to as a \textit{functional dependency} in the database
literature. It indicated by a [[am_fun]] value in the [[am_info]].

It is important to realized that, although a functional dependency implies 
that there cannot be more than one tuple for a particular set of values of 
the input fields, it does \emph{not} imply that a tuple with those values 
exists in the relation. It simply implies that, if there do any tuples with 
those values, then there is only one.                                    
************************************************************************)

and access_method = {
    am_name : string;
    am_in_fields : string list;
    am_out_field : string;
    am_info : access_method_kind
}

and access_method_kind =
    Am_fun of am_fun_type
|   Am_rel of am_rel_type

(************************************************************************)
(* Functional Access Methods						*)
(************************************************************************)


(************************************************************************
A functional access method is used when there is a singleton result that
can match the given values of the input fields. In this case, the compiler
needs to know the cost of accessing the tuple, and how to generate code for
accessing the tuple. The [[am_fun_cost]] of a functional access method is
the expected cost of execution the code associated with the access method.
There are two ways of specifying how the code for the method can be
generated.
************************************************************************)

and am_fun_type =
    am_fun_cost    (* cost of access *)
  * am_fun_kind    (* func for generating code *)

and am_fun_cost = O_1 | O_logn | O_n

and am_fun_kind =
    Am_fun_search of am_search_type
|   Am_fun_lookup of am_lookup_type


(************************************************************************
The first interface is used to describe access methods that can either
succeed or fail in finding an entry for a particular set of infield values.
An example of this sort of access method is one that searches for a
particular index within the entries stored in the storage format. If the
indices stored are space, then the search may or may not succeed. A
discussion of how this interface is used can be found in
Section~\ref{s-113}.
************************************************************************)

and am_search_type =
    bmf_expr list               (* args *)
    -> ( bmf_expr -> bmf_stmt ) (* found_f *)
    -> ( unit -> bmf_stmt )     (* not_found_f *)
    -> bmf_stmt

(************************************************************************
The second interface is used to describe access methods that can only
succeed in finding an entry for a particular set of infield values. An
example of this sort of access method is one that performs an array
dereference. If the infields' values were correctly obtained, then the
array dereference should always return a result. A discussion of how this
interface is used can be found in Section~\ref{sec:1}.
************************************************************************)

and am_lookup_type =
    bmf_expr list               (* args *)
    -> ( bmf_expr -> bmf_stmt)  (* found_f *)
    -> bmf_stmt

(************************************************************************)
(* Relational Access Methods						*)
(************************************************************************)

(************************************************************************ 
A relational access method is used when an arbitrary number of tuples can
match the given values of the input fields. In this case, the compiler
needs code for enumerating these tuples. Unlike an [[am_fun_type]], an
[[am_rel_type]] does not have any cost information: the current protocol
assumes, perhaps incorrectly, that the cost of enumerating a set of tuples
is proportional to the number of tuples. In practice, this appears to be a
reasonable assumption.
************************************************************************)
and am_rel_type =
    Am_rel_interval of am_range_type
|   Am_rel_general of am_stream_type


(************************************************************************
There are two ways of specifying how code can be generated for
relational access methods. The first, the [[am_range_type]], is used
when the values of the output field are exactly all integer values in
closed interval, $[\verb|lb|\ldots\verb|ub|]$. A discussion of how this
interface is used can be found in Section~\ref{sec:1}.
************************************************************************)
and am_range_type =
    bmf_expr list
    -> ( bmf_expr(*lb*) -> bmf_expr(*ub*) -> bmf_stmt )
    -> bmf_stmt


(************************************************************************
The second mechanism, the [[am_stream_type]], allows a more flexible
means  of enumeration. A discussion of how this interface is used can be
found in  Section~\ref{s-114}.
************************************************************************)

and am_stream_type =
    bmf_expr list
    -> am_handler_record

and am_handler_record = {
    decl_ids : bmf_ident list;
    init_st : bmf_stmt;
    incr_st : bmf_stmt;
    close_st : bmf_stmt;
    valid_ex : bmf_expr;
    deref_ex : bmf_expr }

;;

(************************************************************************)
(* Registering the black box						*)
(************************************************************************)

(************************************************************************
At the beginning of the compiler's execution, each of the modules invokes
the function \verb|register| in order to enter their name and function for
making black-boxes with the compiler.
************************************************************************)
val register : string -> black_box_maker -> unit;;

(************************************************************************
Then, when the compiler encounters a Array_bb, it looks up the storage name
to find the corrisponding black-box. The function \verb|is_registered|
returns true or false, depending upon whether a particular name has been
registered. The function \verb|find_registered| returns the black-box that
has been registered under a particular name. \verb|find_registered| throws
the exception \verb|Not_found|, if no black-box has been registered under
that name.
************************************************************************)
val is_registered : string -> bool;; val find_registered : string ->
black_box_maker;;

(************************************************************************
Exception thrown when a black-box maker generated an insane black-box
object.
************************************************************************)
exception Insane_black_box of string;;
