(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers

open Bmf;;
open Bb;;

exception Not_a_bb_decl of string;;
exception No_such_bb_registered of string;;
type black_box_map = black_box String_map.t;;
let empty_black_box_map = String_map.empty

(************************************************************************)

let extract_bb_directive (id: bmf_ident) = (
    match id with
      Ident_valid(name,Type_array(Array_bb(_,_,bb_name,bb_args)),_,_) ->
	(bb_name,bb_args)
    | Ident_valid(name,_,_,_) -> raise (Not_a_bb_decl(name))
    | _ -> raise (Not_a_bb_decl("ident_null"))
: string * keyword_list);;

let all_fields_in_black_box bbi =
    List.fold_left
	(fun results { am_in_fields = in_fields; 
		       am_out_field = out_field } ->
	    List'.union results (out_field::in_fields))
	[] bbi.bb_access_methods
;;

let rec black_box_of_bb_decl id = 
    let (bb_name,bb_args) = extract_bb_directive id
    in 
    make_black_box_obj bb_name bb_args

and make_black_box_obj bb_name bb_args =
  let bb_maker =
    (try
      find_registered bb_name
    with Not_found ->
      raise (No_such_bb_registered(bb_name)))
  and args_map =
    List.fold_left
      (fun m (tag,val0) -> String_map.add tag val0 m)
      String_map.empty
      bb_args
  in
  bb_maker args_map

;;

let find_method name (bbi: black_box) = (
    let rec loop acc_lst =
	match acc_lst with
	  ({am_name=name'} as acc)::acc_lst' when name = name' -> acc
	| _::acc_lst' -> loop acc_lst'
	| [] -> raise Not_found
    in
	loop bbi.bb_access_methods
: access_method)
;;

(************************************************************************)
(* Fake a bbi for a dense array declaration				*)
(************************************************************************)

let dense_index_fields = [ "i"; "j"; "k"; "l"; "m"; "n" ];;

let black_box_of_dense_decl id =
    let (id_name,lb,ubs,maj,sc_ty,init,ann) =
	match id with
	  Ident_valid(id_name,Type_array(Array_dense(lb,ubs,maj,sc_ty)),
		Ident_var(init),ann) ->
	    (id_name,lb,ubs,maj,sc_ty,init,ann)
	| _ -> failwith "black_box_of_dense_decl -- Not a dense decl"
    in let lb_e = (Expr_int(if lb = Dense_lb_0 then 0 else 1))
    and dimens = List.length ubs
    in let check_dimens lst =
	if List.length lst <> dimens then
	    failwith ("Expected a length "^
			(string_of_int dimens)^" list")
    in let index_fields =
	let rec recurs i index_fields =
	    match (i,index_fields) with
	      (_,[]) -> failwith "Ridiculous number of dimens"
	    | (0,_) -> []
	    | (i,field::index_fields') ->
		field::(recurs (i-1) index_fields')
	in
	    recurs dimens dense_index_fields
    and value_field = "v"
    in let make_index_enumerator index_field ub_e =
	let enum_m prereq body_f =
	    if prereq <> [] then begin
		failwith "Expected a length 0 list."
	    end;
	    body_f lb_e ub_e
	in
	    ("enum_"^index_field,[],index_field,
		Am_rel(Am_rel_interval(enum_m)))
    and make_value_lookup () =
	let value_m prereq body_f =
	    check_dimens prereq;
	    body_f (Expr_access(Access_strong,Expr_var(id_name),
			"",prereq))
	in
	    ("",index_fields,value_field,
		Am_fun(O_1,Am_fun_lookup(value_m)))
    and bounds_gen vars =
	check_dimens vars;
	make_and
	    (List.map2 (fun var ub_e ->
		    Expr_op(Op_and,[
			Expr_op(Op_leq,[lb_e;var]);
			Expr_op(Op_leq,[var;ub_e])]))
		vars ubs)
    in let acc_mthd (name,in_fs,out_f,info) = {
	am_name = name; am_in_fields = in_fs;
	am_out_field = out_f; am_info = info
    } 
    in let acc_mthds =
	(make_value_lookup ())::
	    (List.map2 make_index_enumerator index_fields ubs)
    in {
	bb_index_fields = index_fields;
	bb_value_field = value_field;
	bb_access_methods = List.map acc_mthd acc_mthds;
	bb_bounds_maker = bounds_gen;
  bb_ordered_access = []
    }
;;

(************************************************************************)
let black_box_of_view_decl id =
  failwith "*ARRAY* not yet implemented"

(************************************************************************)

let rec add_array_decls (decls: bmf_ident list) (bbm: black_box_map) = (
    List.fold_right add_array_decl decls bbm
: black_box_map)

and add_array_decl (decl: bmf_ident) (bbm: black_box_map) = (
  match decl with
    Ident_valid(_,Type_array(atyp),_,_) ->
      (match atyp with
    	Array_bb _ ->
	  String_map.add (bmf_ident_name decl)
	    (black_box_of_bb_decl decl) bbm
      | Array_dense _ ->
	  String_map.add (bmf_ident_name decl)
	    (black_box_of_dense_decl decl) bbm
      | Array_view _ ->
	  String_map.add (bmf_ident_name decl)
	    (black_box_of_view_decl decl) bbm)
  | _ -> bbm
: black_box_map)

