(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers

open Bmf;;
open Bmf_parser_aux;;
open Bb;;

(************************************************************************)
(* maker argument processing						*)
(************************************************************************)

let an_expr_arg m str default_f =
    try
	let an = String_map.find str m 
	in
	    match an with
	      An_expr(e) -> e
	    | _ -> failwith ("\"" ^ str ^ "\" arg must be an expr")
    with Not_found ->
	default_f ()
;;


let an_expr_var_arg m str default_f =
    try
        let e = an_expr_arg m str (fun () -> raise Not_found)
	in
	    match e with
	      Expr_var(name) -> name
	    | _ -> failwith ("\"" ^ str ^ "\" arg must be a var expr")
    with Not_found ->
	default_f ()
;;

let an_string_arg m str default_f =
    try
	let an = String_map.find str m 
	in
	    match an with
	      An_string(s) -> s
	    | _ -> failwith ("\"" ^ str ^ "\" arg must be a string")
    with Not_found ->
	default_f ()
;;

let an_type_list_arg m str default_f =
    let an_type2type an = (
        match an with
        | (An_type t) -> t
        | _ -> failwith "an_type_list_arg: not a type_ann")
    in 
       try
          let an = String_map.find str m
          in (
             match an with
             | (An_list l) -> List.map an_type2type l
             | _ -> failwith 
               "an_type_list_arg: list of types expected")
       with Not_found -> 
           default_f ()
;;

let an_type_arg m str default_f =
    try (
      match (String_map.find str m) with
      | (An_type t) -> t
      | _ -> failwith "an_type_arg: not a type_ann")
    with Not_found -> 
        default_f ()
;;

let required f m str =
    f m str (fun () -> failwith ("\""^str^"\" arg required"))
;;

(************************************************************************)
(* Gen code for binary search of a vector				*)
(************************************************************************)

let gen_binary_search i lb ub idx found_f not_found_f =
    let low = Misc.gensym "low"
    and mid = Misc.gensym "mid"
    and hig = Misc.gensym "hig"
    in
	parse_stmt "{
	    var $low: int;
	    var $mid: int;
	    var $hig: int;
	    $low := $lb;
	    $hig := $ub;
	    while ($low < $hig) {
		$mid := ($low+$hig)/2;
		if ($i <= $idx[[$mid]]) {
		    $hig := $mid;
		} else {
		    $low := $mid+1;
		}
	    }
	    if ($i == $idx[[$low]]) {
		$thn;
	    } else {
		$els;
	    }
    }" [("$i",An_expr(i));
	("$lb",An_expr(lb));
	("$ub",An_expr(ub));
	("$low",An_string(low));
	("$mid",An_string(mid));
	("$hig",An_string(hig));
	("$idx",An_string(idx));
	("$thn",An_stmt(found_f (Expr_var(low))));
	("$els",An_stmt(not_found_f ()))]
;;

let interval_bounds0 n' args vars =
    let n = required an_expr_arg args n'
    and i = List'.match1 vars
    in
	parse_expr
	    "0 <= $i && $i <= $n-1"
	    [("$i",An_expr(i));("$n",An_expr(n))]
;;

let rectangular_bounds0 m' n' args vars =
    let m = required an_expr_arg args m'
    and n = required an_expr_arg args n'
    and (i,j) = List'.match2 vars
    in
	parse_expr
	    "0 <= $i && $i <= $m-1 && 0 <= $j && $j <= $n-1"
	    [("$i",An_expr(i));("$j",An_expr(j));
	     ("$m",An_expr(m));("$n",An_expr(n))]
;;

let rectangular_bounds1 m' n' args vars =
    let m = required an_expr_arg args m'
    and n = required an_expr_arg args n'
    and (i,j) = List'.match2 vars
    in
	parse_expr
	    "1 <= $i && $i <= $m && 1 <= $j && $j <= $n"
	    [("$i",An_expr(i));("$j",An_expr(j));
	     ("$m",An_expr(m));("$n",An_expr(n))]
;;

let triangular_bounds m n lower_p diag_p upper_p =
  let rectang_constrs =
    "0 <= $i && $i <= $m-1 && 0 <= $j && $j <= $n-1"
  and triang_constrs =
    match (lower_p,diag_p,upper_p) with
      (true,true,true) -> ""
    | (true,false,false) -> " && $i > $j"
    | (true,true,false) -> " && $i >= $j"
    | (false,true,true) -> " && $i <= $j"
    | (false,false,true) -> " && $i < $j"
    | _ -> failwith "Cannot construct such triangular bounds."
  in let constrs = rectang_constrs ^ triang_constrs
  in
  fun vars ->
    let (i,j) = List'.match2 vars
    in
    parse_expr constrs
      [("$i",An_expr(i));("$j",An_expr(j));
	("$m",An_expr(m));("$n",An_expr(n))]

(************************************************************************)
(* Printing access methods						*)
(************************************************************************)

open Printf;;

let rec output_access_method outch am =
  fprintf outch "{ am_name = %s; " am.am_name;
  fprintf outch "am_in_fields = %a; "
    (Io'.output_list output_string)am.am_in_fields;
  fprintf outch "am_out_field = %s; " am.am_out_field;
  fprintf outch "am_info = %a }" output_access_method_kind am.am_info

and output_access_method_kind outch amk =
  match amk with
    Am_fun(cost,afk) ->
      let cost_str =
	match cost with
	  O_1 -> "O_1"
	| O_logn -> " O_logn"
	| O_n -> "O_n"
      and afk_str =
	match afk with
	  Am_fun_search(_) -> "Am_fun_search(...)"
	| Am_fun_lookup(_) -> "Am_fun_lookup(...)"
      in
      fprintf outch "Am_fun(%s,%s)" cost_str afk_str
  | Am_rel(ark) ->
      let ark_str =
	match ark with
	  Am_rel_interval(_) -> "Am_rel_interval(...)"
	| Am_rel_general(_) -> "Am_rel_general(...)"
      in
      fprintf outch "Am_rel(%s)" ark_str

;;
