(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl

open Bmf
open Bmf_parser_aux
open Bb
open Bb_utils


(**************************************************************)
(*    B S R -- block sparse row format                        *)
(**************************************************************)

(*
  This is basically CRS where each entry is an lb-by-lb block.
  We provide two views of the matrix in BSR:

  - bsr_basic has index fields i and j
  
  - bsr also exposes block coordinates ib and jb

*)

type bsr_kind = 
    Basic              (* A(i,j,v) view *)
  | Block              (* A(ib,jb,i,j) view *)
  | Block_offset       (* A(ib,jb,ii,jj) view *)
      ;;

let on_kind kind a1 a2 a3 =
  match kind with
    Basic -> a1
  | Block -> a2
  | Block_offset -> a3
	;;

let bsr_maker kind skip_diag_p args = begin
  let n = required an_expr_arg args "n"
  and nb = required an_expr_arg args "nb"
  and lb = required an_expr_arg args "lb"
  and nzb = required an_expr_arg args "nzb"
  and bpntrb = required an_expr_var_arg args "bpntrb"
  and bpntre = required an_expr_var_arg args "bpntre"
  and bindx = required an_expr_var_arg args "bindx"
  and value_ptr = required an_expr_var_arg args "value"

  in let common_parse_args = [
    ("$bptrb",An_string(bpntrb));
    ("$bptre",An_string(bpntre));
    ("$bindx",An_string(bindx));
    ("$val",An_string(value_ptr));
    ("$n",An_expr(n));
    ("$nb",An_expr(nb));
    ("$lb",An_expr(lb)) ]

  in let enum_ib_m prereq body_f = (
    let _ = List'.match0 prereq
    in (body_f (Expr_int(1)) nb))
    
  and enum_jjb_m prereq body_f = (
    let ib = List'.match1 prereq
    in let args = ("$ib",An_expr(ib))::common_parse_args
    in (body_f
	  (parse_expr "$bptrb[[$ib]]" args)
	  (parse_expr "$bptre[[$ib]]-1" args)))

  and iter_jjb_no_diag_m prepreq = (
    let ib = List'.match1 prepreq
    in let jjb = Misc.gensym "jjb"
    in let jjb_id = Ident_valid(jjb,Type_scalar Type_int,
				Ident_var(Expr_void), An_list [])
    in let args = (
      [ ("$ib", An_expr(ib));
	("$jjb", An_string(jjb)) ] @ common_parse_args)
    in { decl_ids = [jjb_id];
	 init_st = (parse_stmt 
		      "{$jjb := $bptrb[[$ib]];
		        if (($jjb < $bptre[[$ib]]) && ($bindx[[$jjb]] == $ib)) {
			 $jjb := $jjb + 1;}}" 
		      args);
	 incr_st = (parse_stmt
		      "{$jjb := $jjb + 1;
		       if (($jjb < $bptre[[$ib]]) && ($bindx[[$jjb]] == $ib)) {
			 $jjb := $jjb + 1;}}"
		      args);
	 close_st = Stmt_nop;
	 valid_ex = (parse_expr "$jjb < $bptre[[$ib]]" args);
	 deref_ex = Expr_var(jjb) })
			   
  and lookup_jb_m prereq body_f = (
    let (ib,jjb) = List'.match2 prereq
    in (body_f
	  (parse_expr "$bindx[[$jjb]]"
	     (("$jjb",An_expr(jjb))::common_parse_args))))

  and enum_i_m prereq body_f = (
    let ib = List'.match1 prereq
    in let args = ("$ib",An_expr(ib))::common_parse_args
    in (body_f
	  (parse_expr "($ib - 1)* $lb + 1" args)
	  (parse_expr "$ib * $lb" args)))

  and enum_ii_m prereq body_f = (
    let (ib,jjb) = List'.match2 prereq
    in (body_f (Expr_int(1)) lb))
	  

  and lookup_i_m prereq body_f = (
    let (ib,ii) = List'.match2 prereq
    in (body_f
	  (parse_expr "$ii+($ib-1)*$lb"
	     (("$ii",An_expr(ii))::
	      ("$ib",An_expr(ib))::
	      common_parse_args))))

  (* j has exacy the same enumeration, given jb *)
  (* jj has exacy the same enumeration, given ib,jjb *)
  (* j has the same lookup, given jb,jj *)

  and lookup_v_m prereq body_f = (
    let (jjb,ii,jj) = List'.match3 prereq
    in let args = 
      [ ("$jjb",An_expr(jjb));
	("$jj",An_expr(jj));
	("$ii",An_expr(ii)) ] @ common_parse_args
    in (body_f
	  (parse_expr "$val[[$lb*$lb*($jjb-1)+($jj-1)*$lb+$ii]]"
	     args)))

  and bounds4 vars = (
    let (ib,jb,i,j) = List'.match4 vars
    in let args = 
      [ ("$i",An_expr(i));
	("$j",An_expr(j));
	("$ib",An_expr(ib));
	("$jb",An_expr(jb)) ] @ common_parse_args	
    in (parse_expr
	  "1<=$i && $i<=$n && 1<=$j && $j<=$n &&
	   1<=$ib && $ib<=$nb && 1<=$jb && $jb<=$nb"
	  args))

  and bounds4_offset vars = (
    let (ib,jb,ii,jj) = List'.match4 vars
    in let args = 
      [ ("$ii",An_expr(ii));
	("$jj",An_expr(jj));
	("$ib",An_expr(ib));
	("$jb",An_expr(jb)) ] @ common_parse_args	
    in (parse_expr
	  "1<=$ii && $ii<=$lb && 1<=$jj && $jj<=$lb &&
	   1<=$ib && $ib<=$nb && 1<=$jb && $jb<=$nb"
	  args))

  in let acc_mthd (name,in_fs,out_f,info) = {
    am_name = name; am_in_fields = in_fs;
    am_out_field = out_f; am_info = info
  }

  in let enum_jjb = (
    if skip_diag_p then
      ("enum_jjb",["ib"],"jjb",Am_rel(Am_rel_general(iter_jjb_no_diag_m)))
    else
      ("enum_jjb",["ib"],"jjb",Am_rel(Am_rel_interval(enum_jjb_m))))

  in let common_methods = 
    [ ("enum_ib",[],"ib",Am_rel(Am_rel_interval(enum_ib_m)));
      enum_jjb;
      ("lookup_jb_m",["ib";"jjb"],"jb",Am_fun(O_1,Am_fun_lookup(lookup_jb_m)));

      ("enum_ii",["ib";"jjb"],"ii",Am_rel(Am_rel_interval(enum_ii_m)));
      ("enum_jj",["ib";"jjb"],"jj",Am_rel(Am_rel_interval(enum_ii_m))); (* same function *)

      ("lookup_v",["jjb";"ii";"jj"],"v",Am_fun(O_1,Am_fun_lookup(lookup_v_m))); 
    ]
  and ij_lookups = 
    [ ("lookup_i",["ib";"ii"],"i",Am_fun(O_1,Am_fun_lookup(lookup_i_m)));
      ("lookup_j",["jb";"jj"],"j",Am_fun(O_1,Am_fun_lookup(lookup_i_m))); (* same function *)
    ] 

  in let acc_mthds = 
    (on_kind kind
       (common_methods @ ij_lookups)
       (common_methods @ ij_lookups)
       common_methods)

  and make_bounds = 
    (on_kind kind 
       (rectangular_bounds1 "n" "n" args)
       bounds4
       bounds4_offset)

  in { bb_index_fields = (on_kind kind
			    ["i";"j"]
			    ["ib";"jb";"i";"j"]
			    ["ib";"jb";"ii";"jj"]);
       bb_value_field = "v";
       bb_access_methods = List.map acc_mthd acc_mthds;
       bb_ordered_access = [];
       bb_bounds_maker = make_bounds }
      
end (* bsr_maker *)
    ;;
    

register "bsr_basic" (bsr_maker Basic false);;    
register "bsr_block" (bsr_maker Block false);;    
register "bsr_block_offset" (bsr_maker Block_offset false);;

register "bsr_basic_nd" (bsr_maker Basic true);;    
register "bsr_block_nd" (bsr_maker Block true);;    
register "bsr_block_offset_nd" (bsr_maker Block_offset true);;
    
