(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers

open Bmf;;
open Bmf_util;;
open Bmf_parser_aux;;
open Bb;;
open Bb_utils;;

(***********************************************************************
 *                 Black box for communication buffers
 ***********************************************************************)

let commbuffer_name arity vtype =
    let type_name = (
        match vtype with
        | (Type_scalar Type_int) -> "int"
        | (Type_scalar Type_real) -> "real"
        | (Type_scalar Type_void) -> "void"
        | _ -> failwith "commbuffer_name: invalid type")
    in 
       ("commbuffer_" ^ (string_of_int arity) ^ "_" ^ type_name)
;;

(*
  Access methods:
  [q] --> [ii]  (range enumeration)
  [ii] --> [q] (search)
  [ii] --> [i_k] (lookup)
*)

let commbuffer_maker_maker (q:string)
                           (ii:string) 
                           (i_list:(string list))
                           (v:string) 
                           (vtype:bmf_type) 
                           args = (

    let nproc = required an_expr_arg args "nproc"
    and handle = required an_expr_arg args "ptr"
    and arity = (List.length i_list)
    and i_vect = (Array.of_list i_list)

    in let enum_ii_m prereq body_f = (
       let q = List'.match1 prereq
       and s = Misc.gensym "start"
       and e = Misc.gensym "end"

       in let args = 
          [ ("$q", (An_expr q))
          ; ("$start", (An_string s))
          ; ("$end", (An_string e))
          ; ("$c", (An_expr handle))
          ; ("$body", (An_stmt (body_f (Expr_var s) (Expr_var e))))
          ]

       in (
          parse_stmt
              "{ var $start : int; var $end : int;
                 $start := call iptr4q_cbuf($c,$q);
                 $end := call iptr4q_cbuf($c,$q + 1);
                 $end := $end - 1;
                 $body; }"
              args))

    and enum_q_m prepreq body_f = (
        body_f
          (Expr_int(0))
          (parse_expr "$nproc - 1" [("$nproc", (An_expr nproc))])
    )

    and lookup_q_m prereq body_f = (
        let ii = List'.match1 prereq
        and q_name = Misc.gensym "q"
        in let q = Ident_valid(q_name,Type_scalar Type_int,
                               Ident_var(Expr_void), An_list [])
        in let args = 
           [ ("$q", (An_string q_name))
           ; ("$ii", (An_expr ii))
           ; ("$body", (An_stmt (body_f (Expr_var q_name))))
           ; ("$c", (An_expr handle))
           ]
        in (
           parse_stmt
              "{var $q : int;
                $q := call find_q4ii_cbuf($c, $ii);
                $body;}"
              args))

    and gen_lookup_i_m k prereq body_f = (
        let ii = List'.match1 prereq
        and i_name = Misc.gensym ("i_"^(string_of_int k))
        in let i = Ident_valid(i_name,Type_scalar Type_int,
                               Ident_var(Expr_void), An_list [])
        in let args =
           [ ("$i", (An_string i_name))
           ; ("$ii", (An_expr ii))
           ; ("$body", (An_stmt (body_f (Expr_var i_name))))
           ; ("$c", (An_expr handle))
           ; ("$k", (An_expr (Expr_int (k-1))))
           ]
        in (
           parse_stmt
              "{var $i : int;
                $i := call ind4ii_cbuf ($c, $ii, $k);
                $body;}"
              args))

    and gen_search_i_m k prereq found_f not_found_f = (
      failwith "Not implemented.";
      Stmt_nop)

    and lookup_v_m prereq body_f = (
        let ii = List'.match1 prereq
        and v_name = Misc.gensym "v"
        and lookup_func = (
            match vtype with
            | (Type_scalar Type_int) -> "int_val4ii_cbuf"
            | (Type_scalar Type_real) -> "real_val4ii_cbuf"
            | (Type_scalar Type_void) -> "void_val4ii_cbuf"
            | _ -> failwith "commbuffer: bad type")
        in let v = Ident_valid(v_name, vtype, Ident_var(Expr_void), An_list [])
        in let args =
           [ ("$v", (An_string v_name))
           ; ("$t", (An_type vtype))
           ; ("$ii", (An_expr ii))
           ; ("$body", (An_stmt (body_f (Expr_var v_name))))
           ; ("$c", (An_expr handle))
           ; ("$f", (An_string lookup_func))
           ]
        in (
           parse_stmt
              "{var $v : $t;
                $v := call $f ($c, $ii);
                $body;}"
              args))

    in let field_lookups = (
       let f lst j = (
           ( ("lookup_i"^(string_of_int j))
           , [ii]
           , i_vect.(j-1)
           , Am_fun(O_1, Am_fun_lookup(gen_lookup_i_m j))
           ) :: lst)
       in (Range.fold_left f [] 1 arity))

    in let searches = (
       List.map
          (fun j ->
	     ( ("search_i"^(string_of_int j))
	     , [i_vect.(j)]
	     , ii
	     , Am_fun(O_logn, Am_fun_search(gen_search_i_m j))
             ))
	  (try
	      (match String_map.find "searches" args with
		An_expr(Expr_int(j)) -> [j]
	      |	An_list(lst) ->
		  List.map
		    (fun an ->
		      match an with
			An_expr(Expr_int(j)) -> j
		      | _ -> failwith "Ill-formed \"search\" argument")
		    lst
	      |	_ -> failwith "Ill-formed \"search\" argument")
	   with Not_found -> [])
    )

    in let acc_mthds = (
      [ ("lookup_q", [ii], q, Am_fun(O_logn, Am_fun_lookup(lookup_q_m)))
      ; ("enum_ii", [q], ii, Am_rel(Am_rel_interval(enum_ii_m)))
      ; ("enum_q", [], q, Am_rel(Am_rel_interval(enum_q_m)))
      ; ("lookup_v", [ii], v, Am_fun(O_1, Am_fun_lookup(lookup_v_m)))
      ] @ field_lookups @ searches)

    in let acc_mthd (name,in_fs,out_f,info) = {
	am_name = name; am_in_fields = in_fs;
	am_out_field = out_f; am_info = info
    } 

    in
	List.map acc_mthd acc_mthds
)
;;           
           
let register_commbuffer arity vtype =
    let i = Misc.gensyms "i" arity
    and q = "q"
    and v = "v"
    and ii = "ii"
    in let make_acc_mthds =
	commbuffer_maker_maker q ii i v vtype
    in (
      register (commbuffer_name arity vtype)
	(fun args -> {
	    bb_index_fields = (q :: ii :: i);
	    bb_value_field = v;
	    bb_access_methods = make_acc_mthds args;
		  bb_ordered_access = [];
	    bb_bounds_maker = (fun lst -> (Expr_bool true))
	}))
;;

let register_lazy_commbuffer arity vtype =
    if ((not ((is_registered (commbuffer_name arity vtype))))) then
       register_commbuffer arity vtype
    else
       ()
;;

register_commbuffer 2 (Type_scalar Type_void);;
register_commbuffer 1 (Type_scalar Type_void);;
register_commbuffer 0 (Type_scalar Type_real);;
register_commbuffer 1 (Type_scalar Type_real);;

(***********************************************************************
 *                 Black box for combined comm-buffers
 ***********************************************************************)

(*
  Here is how it works

  Say we have two CBs: A and B.
  A has arity da and value type void.
  B has arity db and value type 
  one of the CBs must be of type void.

    

*)
