(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl

open Bmf;;
open Bmf_util;;
open Bmf_parser_aux;;
open Bb;;
open Bb_utils;;

(***********************************************************************
 *
 *                             Generalized block distribution BB
 *
 ***********************************************************************)

(*
  This BB takes three parameters:
  npart         -- the number of partitions
  ptr           -- pointer to the array of partition boundaries
                   
  The ptr array must have npart+1 elements.  To translate a global index
  x we need to find k such that npart[k] <= x < npart[k+1]; then the
  translation is: x' = x - npart[k].

  This means that if we translate the range n1..n2 into m partitions,
  then:
    ptr[0] = n1
    ptr[m] = n2+1

  and each partition k contains the range ptr[k]..(ptr[k+1]-1).


  Say q, x and x' are the partition number (starts with 0), global
  index and the translated index. Then we have the following access
  methods:

    x --> q       a search for a partition number and local index
    q, x --> x'   a lookup for the offset (x'= x - ptr[q])
    q --> {x}     a range enumeration

  There is also a methods to get a (void) value, but it shoule never
  be called
*)

let gen_binary_search4genblock ptr npart x found_f = (
    (* ptr is the name of the array
       npart is the expression for the number of partitions
       x is the expression for the global index
       found_f has the same meaning as in the search
       methods 

       This is somewhat different than the code in bb_utils
    *)
    let low = Misc.gensym "low"
    and mid = Misc.gensym "mid"
    and hig = Misc.gensym "hig"
    in
	parse_stmt "{
	    var $low: int;
	    var $mid: int;
	    var $hig: int;
	    $low := 0;
	    $hig := $npart;
            if ($x >= $ptr[[$hig]]) { call fail; }
            if ($x < $ptr[[$low]]) { call fail; }
	    while ($hig - $low > 1) {
		$mid := ($low+$hig)/2;
		if ($x >= $ptr[[$mid]]) {
		    $low := $mid;
		} else {
		    $hig := $mid;
		}
	    }
            $found;}"
       [("$x",An_expr(x));
	("$npart",An_expr(npart));
        ("$ptr", An_string(ptr));
	("$low",An_string(low));
	("$mid",An_string(mid));
	("$hig",An_string(hig));
        ("$found",An_stmt(found_f (Expr_var low)))]
)
;;


let genblock_maker args = (
    let npart = required an_expr_arg args "npart"
    and ptr = required an_expr_var_arg args "ptr"

    in let search_q_m prereq found_f not_found_f = (
        let x = List'.match1 prereq
        in (gen_binary_search4genblock ptr npart x found_f)
    )

    and lookup_xx_m prereq body_f = (
        let (q,x) = List'.match2 prereq
        in body_f
            (parse_expr
              "$x - $ptr[[$q]]"
              [ ("$q", (An_expr q))
              ; ("$x", (An_expr x))
              ; ("$ptr", (An_string ptr))
              ])
    )

    and enum_q_m prereq body_f = (
        let _ = List'.match0 prereq
        in body_f
           (parse_expr "0" [])
           (parse_expr "($npart-1)" [ ("$npart", (An_expr npart)) ])
    )

    and enum_x_top_m prereq body_f = (
        let _ = List'.match0 prereq
        in let parse_args =
           [ ("$npart", (An_expr npart))
           ; ("$ptr", (An_string ptr))
           ]
        in body_f
           (parse_expr "$ptr[[0]]" parse_args)
           (parse_expr "$ptr[[npart]]-1" parse_args)
    )

    and enum_xx_m prereq body_f = (
        let (q,x) = List'.match2 prereq
        in let parse_args =
           [ ("$ptr", (An_string ptr))
           ; ("$q", (An_expr q))
           ; ("$x", (An_expr x))
           ]
        in body_f
            (parse_expr "($x-$ptr[[$q]])" parse_args)
            (parse_expr "($x-$ptr[[$q]])" parse_args)
    )

    and enum_x_m prereq body_f = (
        let q = List'.match1 prereq
        in let parse_args = 
           [ ("$ptr", (An_string ptr))
           ; ("$q", (An_expr q))
           ]
        in body_f
            (parse_expr "$ptr[[$q]]" parse_args)
            (parse_expr "$ptr[[$q+1]]-1" parse_args)
    )
    
    (* this is a fake method, which should never be called
       because the value is of type void
       *)
    and lookup_v_m prereq body_f = (
        let (q,x,xx) = List'.match3 prereq
        in body_f (parse_expr "no_such_expr" [])
    )

    and search_v_m prereq found_f not_found_f = (
        let (q,x,xx) = List'.match3 prereq
        in (found_f (parse_expr "no_such_expr" []))
    )

    and enum_v_m prereq body_f = (
        let (q,x,xx) = List'.match3 prereq
        in body_f 
             (parse_expr "no_such_expr" [])
             (parse_expr "no_such_expr" [])
    )

    in let acc_mthds =
       [ ("enum_q", [], "q",
                    Am_rel(Am_rel_interval(enum_q_m)))
       ; ("enum_x_top", [], "x",
                        Am_rel(Am_rel_interval(enum_x_top_m)))
       ; ("search_q", ["x"], "q",
                      Am_fun(O_logn,Am_fun_search(search_q_m)))
       ; ("enum_x", ["q"], "x",
                    Am_rel(Am_rel_interval(enum_x_m)))
 
       ; ("enum_xx", ["q";"x"], "xx",
                    Am_rel(Am_rel_interval(enum_xx_m)))

       ; ("lookup_xx", ["q"; "x"], "xx",
                       Am_fun(O_1,Am_fun_lookup(lookup_xx_m)))
       ; ("lookup_v", ["q"; "x"; "xx"], "v",
                      Am_fun(O_1,Am_fun_lookup(lookup_v_m)))
       ; ("search_v", ["q"; "x"; "xx"], "v",
                      Am_fun(O_1,Am_fun_search(search_v_m)))
       ; ("enum_v", ["q"; "x"; "xx"], "v",
                    Am_rel(Am_rel_interval(enum_v_m)))
       ]

    and acc_mthd (name,in_fs,out_f,info) = 
        { am_name = name
        ; am_in_fields = in_fs
        ; am_out_field = out_f
        ; am_info = info
        }

    and make_fake_bounds args = (
        Expr_bool(true)
    )

    in 
       { bb_index_fields = ["q";"x";"xx"]
       ; bb_value_field = "v"
       ; bb_access_methods = List.map acc_mthd acc_mthds
	     ; bb_ordered_access = []
       ; bb_bounds_maker = make_fake_bounds
       }
)
;;

register "genblock" genblock_maker;;
    



