(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers

open Bmf;;
open Bmf_util;;
open Bmf_parser_aux;;
open Bb;;
open Bb_utils;;

(***********************************************************************
 *
 *                             Hash tables
 *
 ***********************************************************************)


(***********************************************************************
 *                          Some useful stuff
 ***********************************************************************)

(*
  str: << << (e1) (e2) >> << (e3) (e4) >> ... >>
  becomes
  [(e1,e2);(e3,e4);...]
*)
let an_list_of_expr_pairs_arg m str default_f =
    let f an = (
        match an with
        | (An_list [(An_expr e1); (An_expr e2)]) -> (e1,e2)
        | _ -> failwith "an_list_of_expr_pairs_arg: not a pair")
    in
      try (
          match (String_map.find str m) with
          | (An_list l) -> (List.map f l)
          | _ -> failwith "an_list_of_expr_pairs_arg: must be an_list")
      with Not_found -> default_f
;;

(***********************************************************************
 *********************************************************************** 
 ***********************************************************************
 *                       Hash table "maker maker"
 *
 * Each combinations of tuple size, key size and value type requires
 * its own black-box.  In the beginning we register black boxes with
 * names of the form: hashtable_%t_%k_%v 
 * where %t is the tuple size, %k is the key size, %v is the value
 * field type.
 *
 ***********************************************************************)

let hashtable_name ndim nkey vtype = (
    (* this returs a canonical name of a particular BB *)
    let type_name = (
        match vtype with
        | (Type_scalar Type_int) -> "int"
        | (Type_scalar Type_real) -> "real"
        | (Type_scalar Type_void) -> "void"
        | _ -> failwith "register_hash_table: bad value type")
    in "hashtable_" ^ (string_of_int ndim) ^ "_" ^
                    (string_of_int nkey)^"_"^type_name
)
;;
    
(*
  The arguments are:
  i_list -- list of names for the index fields
  iter_name -- name of the iterator field
  nkey -- how many of theseare to be used for hashing
  v -- name of the value field
  vtype -- type of the value field (real, int or void)

  args -- the usual maker keyword args
*)
let hashtable_maker_maker i_list iter_name nkey v vtype args =
    let ndims = List.length i_list
    and i_vec = Array.of_list i_list
    and ptr = required an_expr_arg args "ptr"
    
    in let ptr_name = (
       match ptr with
       | (Expr_var(s)) -> s
       | _ -> failwith "hashtable: invalid ptr name")

    in let (tuple_size, deref_fun_name) = (
        match vtype with
        | (Type_scalar Type_int) -> 
            (1 + ndims, "deref_hash_iter_as_int")
        | (Type_scalar Type_real) -> 
            (2 + ndims, "deref_hash_iter_as_real")
        | (Type_scalar Type_void) -> (ndims, "no_such_fn")
        | _ -> failwith "hashtable: Can not determine type size")

    in let hash_stream (l:(bmf_expr list)) = (
       let valid_p_name = Misc.gensym "valid_p"
       and iter_name = Misc.gensym "it"

       in let valid_p_id = Ident_valid ( valid_p_name,
                                         Type_scalar Type_bool,
                                         Ident_var(Expr_void), An_list [])
       and iter_id = Ident_valid ( iter_name,
                                   Type_scalar Type_int,
                                   Ident_var(Expr_void), An_list [])

       in let parse_args = 
          [ ("$ptr",(An_string ptr_name))
          ; ("$it",(An_string iter_name))
          ; ("$valid_p", (An_string valid_p_name))
          ]
       in
        { decl_ids = [valid_p_id; iter_id]
        ; init_st = parse_stmt 
                     "{$it := call open_hash_iter($ptr);
                       $valid_p := call valid_hash_iter($it);}"
                     parse_args
        ; incr_st = parse_stmt
                     "{call next_hash_iter($it);
                       $valid_p := call valid_hash_iter($it);}"
                     parse_args
        ; close_st = parse_stmt
                      "call close_hash_iter($it);"
                      parse_args
        ; valid_ex = Expr_var(valid_p_name)
        ; deref_ex = Expr_var(iter_name)
         })
    
    and gen_lookup_field_iter j prereq found_f = (
        (* generates lookup methods for the individual fields 
           in the tuple through the iterator *)
        let it = List'.match1 prereq
        and i_name = Misc.gensym "i"
        and j_expr = Expr_int(j)
        in let i = (
           Ident_valid( i_name, Type_scalar Type_int,
                        Ident_var(Expr_void),An_list []))
        in let parser_args = 
           [ ("$it", (An_expr it))
           ; ("$i", (An_string i_name))
           ; ("$j", (An_expr j_expr))
           ]
        in let assign_stmt = (
           parse_stmt 
              "$i := call deref_hash_iter_as_int ($it,$j-1);"
              parser_args)
        in 
           Stmt_let (
              (Frame [i]),
              Stmt_seq (assign_stmt, found_f (Expr_var(i_name)))))
        
    and lookup_value_iter prereq found_f = (
        let it = List'.match1 prereq
        and v_name = Misc.gensym "v"
        in let v = (
           Ident_valid ( v_name, vtype, Ident_var(Expr_void),
                         An_list []))
        in let parser_args = 
           [ ("$it", (An_expr it))
           ; ("$v", (An_string v_name))
           ; ("$dim", (An_expr (Expr_int(ndims))))
           ; ("$deref", (An_string deref_fun_name))
           ]
        in let assign_stmt = (
           parse_stmt 
              "$v := call $deref ($it,$dim);"
              parser_args)
        in 
           Stmt_let (
              (Frame [v]),
              Stmt_seq (assign_stmt, found_f (Expr_var(v_name)))))

    in let iter2field_lookups = (
       let f lst j = (
           ( ("lookup_i_"^(string_of_int j))
           , [iter_name]
           , i_vec.(j-1)
           , Am_fun(O_1,Am_fun_lookup(gen_lookup_field_iter j))
           ) :: lst)
       in Range.fold_left f [] 1 ndims)

    and value_lookup = (
        ( "lookup_v", [iter_name], v, 
                      Am_fun( O_1,
                              Am_fun_lookup(lookup_value_iter))))
    and enum_iter = 
        ("enum_iter", [], iter_name,
                     Am_rel(Am_rel_general(hash_stream)))

    and field2iter_lookup prereq found_f not_found_f = (
        (zassert ((List.length prereq) = ndims) 
               (Failure("field2iter_lookup: wrong number of dimensions")));

        let i_list = prereq
        in let m = (List.length i_list)

        and tuple = Misc.gensym "tuple"
        and valid_p_name = Misc.gensym "valid_p"
        and iter_name = Misc.gensym "it"

        in let tuple_assignments = (
           (* returns (j,stmt_list) pair *)

           let f (j,lst) i = (
               let args = 
                   [ ("$t", (An_string tuple))
                   ; ("$i", (An_expr i))
                   ; ("$j", (An_expr (Expr_int j)))
                   ]
               in let stmt = (
                   parse_stmt "$t[[$j-1]] := $i;" args)

               in ((j+1), stmt::lst))
           in 
              stmt_list2seq 
                ((fun (a,b) -> b) (List.fold_left f (1,[]) i_list)))

        in let parse_args1 = (
           [ ("$it", (An_string iter_name))
           ; ("$t", (An_string tuple))
           ; ("$m", (An_expr (Expr_int (m-1))))
           ; ("$ptr",(An_string ptr_name))
           ; ("$valid_p",(An_string valid_p_name))
           ])

        in let search = (
           parse_stmt 
              "{$it := call open_hash_iter($ptr);
                call search_hash_table($it, $t);}"
              parse_args1 )
        and check_valid = (
            parse_stmt "$valid_p := call valid_hash_iter($it);"
                       parse_args1 )
        in let parse_args2 = (parse_args1 @
           [ ("$assign", (An_stmt tuple_assignments))
           ; ("$search", (An_stmt search))
           ; ("$check_valid", (An_stmt check_valid))
           ; ("$found", (An_stmt (found_f (Expr_var iter_name))))
           ; ("$not_found", (An_stmt (not_found_f ())))
           ])

        in (
            (* $it should just become a ptr in the translation to C *)
            parse_stmt
              "{var $valid_p : bool;
                var $it : int;
                var $t : array(0:$m) of int;
                $assign;
                $search;
                $check_valid;
                if ($valid_p) { $found; } else { $not_found; }}"
              parse_args2
              )
        )

    in let  iter_lookup = (
        ( "lookup_it", List'.sublist i_list 0 nkey, iter_name, 
                      Am_fun( O_1,
                              Am_fun_search(field2iter_lookup))))

    in let acc_mthd (name,in_fs,out_f,info) = {
	am_name = name; am_in_fields = in_fs;
	am_out_field = out_f; am_info = info
    } 
    in let all_methods = 
       (enum_iter :: value_lookup :: iter_lookup :: iter2field_lookups)

    in List.map acc_mthd all_methods
;;

let register_hash_table ndim nkey vtype =
    let i = Misc.gensyms "i" ndim
    and v = Misc.gensym "v"
    in let make_acc_mthds =
	hashtable_maker_maker i "it" nkey v vtype
    in
      register (hashtable_name ndim nkey vtype)
	(fun args -> {
	    bb_index_fields = i;
	    bb_value_field = v;
	    bb_access_methods = make_acc_mthds args;
		  bb_ordered_access = [];
	    bb_bounds_maker = (fun lst -> Expr_bool(true))
	})
        
;;

let register_lazy_hash_table ndim nkey vtype =
    if (not ((is_registered (hashtable_name ndim nkey vtype)))) then
       register_hash_table ndim nkey vtype
    else ()
;;

register_hash_table 1 1 (Type_scalar Type_void);;
register_hash_table 3 2 (Type_scalar Type_void);;
register_hash_table 1 1 (Type_scalar Type_real);;
register_hash_table 2 2 (Type_scalar Type_real);;
