(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl

open Bmf;;
open Bmf_parser_aux;;
open Bb;;
open Bb_utils;;

(* This implementation of the sparse vector storage format is found in the  *)
(* source file [[sparsev.ml]]. In addition to the definition of             *)
(* [[sparsev_maker]], this file contains a call to [[bb__register]] in order *)
(* to register this black-box module with the sparse compiler.              *)

(* A sparse vector is a relation with three fields, $ii$, $i$, and $v$,*)
(* which hold the offset, dense index, and value of each non-zero entry,*)
(* respectively.  The core of the sparse vector black-box is the function,  *)
(* [[sparsev_maker]], which generates black-box objects for sparse vector   *)
(* variables. When invokes with a set of black-box arguments, this function *)
(* must,                                                                    *)
(*									    *)
(* + Parse these argument and extract the relevant information,         *)
(* + Create closures for each of the access methods, and                *)
(* + Create a record of all of the information that constitutes the     *)
(*   black-box object.                                                      *)

let sparsev_maker args =

    (* There are four arguments that must be provided in order to assemble*)
    (* the black-box.*)

    (* {[[n]]} -- The upper bound of index field, $i$, which will range*)
    (*   from $0$ to $[[n]]-1$.*)
    (* {[[nzs]]} -- The number of entries stored in the sparse vector. The*)
    (*   offset field, $ii$, will range in value from $0$ to $[[nzs]]-1$. *)
    (* {[[index]]} -- The name of the dense integer vector which holds the*)
    (*   dense index of each non-zero entry. The value of the $i$ field*)
    (*   associated with an offset, $ii$, is obtained by $[[index]][ii]$*)
    (* {[[value]]} -- The name of the dense real vector which holds the*)
    (*   value of each non-zero entry. The value of the $v$ field*)
    (*   associated with an offset, $ii$, is obtained by $[[value]][ii]$*)

    let n = required an_expr_arg args "n"
    and nzs = required an_expr_arg args "nzs"
    and index_ptr = required an_expr_var_arg args "index"
    and value_ptr = required an_expr_var_arg args "value"

    (**************************************************
    The access methods
    **************************************************)

    (* With these four arguments, it is possible to create closures that*)
    (* will generate code for each of the four access methods that are*)
    (* provided for a sparse vector.*)

    (* The access method, [[enum_ii]], will produce a stream of the values*)
    (* of $ii$. More precisely, it produces the lower and upper bounds, $0$*)
    (* and $[[nzs]]-1$, on the range of offsets. These bounds are passed to*)
    (* [[body_f]], the function provided by the compiler for generate the*)
    (* code that uses these bounds. *)

    (* The argument, [[prereq]], will be the list of arguments that*)
    (* appeared in the access method invokation. The [[List'.match0*)
    (* prereq]] expression is used to destruct this list. In this case, the*)
    (* invokation of this access method takes no arguments, so all that*)
    (* [[List'.match0]] does is check that [[prereq]] is a length zero*)
    (* list. However, in subsequent methods, [[List'.matchN]] will return*)
    (* the [[N]] arguments from the [[prereq]] list. *)
    in let enum_ii_m prereq body_f =
               let _ = List'.match0 prereq
               in
                   body_f (Expr_int(0)) (Expr_op(Op_plus,[nzs;Expr_int(-1)]))

    (* The access methods, [[lookup_i]] and [[lookup_v]], are used to*)
    (* dereference an $ii$ offset to obtain the associated $i$ and $v$*)
    (* values. [[body_f]] is the function provided by the compiler to*)
    (* generate the code that uses the results of the dereference. *)
    and lookup_i_m prereq body_f =
	let ii = List'.match1 prereq
	in
	    body_f (parse_expr "$indp[[$ii]]"
			[("$indp",An_string(index_ptr));
			 ("$ii",An_expr(ii))])
    and lookup_v_m prereq body_f =
	let ii = List'.match1 prereq
	in
	    body_f (parse_expr "$valp[[$ii]]"
			[("$valp",An_string(value_ptr));
			 ("$ii",An_expr(ii))])

    (* The access method, [[search_i]], uses a binary search to find the*)
    (* $ii$ offset associated with a particular value of $i$. The compiler*)
    (* supplied function, [[found_f]], is invoked to generate the code for*)
    (* when such an $ii$ is found, and the function, [[not_found_f]], is*)
    (* invoked to generate the code for when case when it is not found. *)
    and search_i_m prereq found_f not_found_f =
	let i = List'.match1 prereq
	in
	    gen_binary_search i (Expr_int(0)) nzs index_ptr
		found_f not_found_f
    in 
        (* With the closures for the access methods in place, the black-box*)
	(* object that will be returned by [[sparsev_maker]] can be*)
	(* constructed. It contains the schema of the relation, the arrays*)
	(* bounds of the vector, in terms of the dense index field, $i$,*)
	(* and the list of available access methods. *)
        {
            bb_index_fields = ["i"];
            bb_value_field = "v";
            bb_bounds_maker = (interval_bounds0 "n" args);
            bb_access_methods = 
                (* The list of available access methods includes, for each*)
		(* method, the name, input fields, output fields, and code*)
		(* generation closure. For the singleton  methods, an*)
		(* estimate of the cost of invoking the method is included*)
		(* as well. *)
                ([
                     { am_name="enum_ii";
		         am_in_fields=[]; am_out_field="ii";
                         am_info=Am_rel(Am_rel_interval(enum_ii_m)) };

                     { am_name="lookup_i";
			 am_in_fields=["ii"]; am_out_field="i";
                         am_info=Am_fun(O_1,Am_fun_lookup(lookup_i_m)) };

                     { am_name="lookup_v";
		         am_in_fields=["ii"]; am_out_field="v";
                         am_info=Am_fun(O_1,Am_fun_lookup(lookup_v_m)) };

                     { am_name="search_i";am_in_fields=["i"];
			 am_out_field="ii";
                         am_info=Am_fun(O_logn,Am_fun_search(search_i_m)) }
                 ]);
			      bb_ordered_access = [("enum_ii","lookup_i")]
        }
;;

(* Register the black-box with the compiler *)
register "sparsev" sparsev_maker;;
