(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl

open Bmf
open Bmf_parser_aux
open Bb
open Bb_utils

(**********************************************************)
(*      V B R -- variable block compressed row            *)
(**********************************************************)

type vbr_kind = 
    Basic
  | Block
  | Block_offset
      ;;

let on_kind kind a1 a2 a3 =
  match kind with
    Basic -> a1
  | Block -> a2
  | Block_offset -> a3
	;;

let vbr_maker kind args = begin
  let n = required an_expr_arg args "n"
  and mb = required an_expr_arg args "mb"
  and nb = required an_expr_arg args "nb"
  and value_ptr = required an_expr_var_arg args "value"
  and bpntrb = required an_expr_var_arg args "bpntrb"
  and bpntre = required an_expr_var_arg args "bpntre"
  and bindx = required an_expr_var_arg args "bindx"
  and indx = required an_expr_var_arg args "indx"
  and rpntr = required an_expr_var_arg args "rpntr"
  and cpntr = required an_expr_var_arg args "cpntr"

  in let common_parse_args = [
    ("$bptrb",An_string(bpntrb));
    ("$bptre",An_string(bpntre));
    ("$bindx",An_string(bindx));
    ("$indx",An_string(indx));
    ("$cpntr",An_string(cpntr));
    ("$rpntr",An_string(rpntr));
    ("$val",An_string(value_ptr));
    ("$n",An_expr(n));
    ("$nb",An_expr(nb));
    ("$mb",An_expr(mb)) ]

  in let enum_ib_m prereq body_f = (
    let _ = List'.match0 prereq
    in (body_f (Expr_int(1)) mb))

  and enum_jjb_m prereq body_f = (
    let ib = List'.match1 prereq
    in let args = ("$ib",An_expr(ib))::common_parse_args
    in (body_f
	  (parse_expr "$bptrb[[$ib]]" args)
	  (parse_expr "$bptre[[$ib]]-1" args)))

  and lookup_jb_m prereq body_f = (
    let (ib,jjb) = List'.match2 prereq
    in (body_f
	  (parse_expr "$bindx[[$jjb]]"
	     (("$jjb",An_expr(jjb))::common_parse_args))))

  and enum_ii_m prereq body_f = (
    let (ib,jjb) = List'.match2 prereq
    in (body_f (Expr_int(1)) 
	  (parse_expr "$rpntr[[$ib+1]]-$rpntr[[$ib]]"
	     (("$ib",An_expr(ib))::common_parse_args))))


  and enum_jj_m prereq body_f = (
    let (ib,jjb,jb) = List'.match3 prereq
    in (body_f (Expr_int(1)) 
	  (parse_expr "$cpntr[[$jb+1]]-$cpntr[[$jb]]"
	     (("$jb",An_expr(jb))::common_parse_args))))

  and lookup_i_m prereq body_f = (
    let (ib,ii) = List'.match2 prereq
    in (body_f
	  (parse_expr "$ii-1+$rpntr[[$ib]]"
	     (("$ii",An_expr(ii))::
	      ("$ib",An_expr(ib))::
	      common_parse_args))))

  and lookup_j_m prereq body_f = (
    let (jb,jj) = List'.match2 prereq
    in (body_f
	  (parse_expr "$jj-1+$cpntr[[$jb]]"
	     (("$jj",An_expr(jj))::
	      ("$jb",An_expr(jb))::
	      common_parse_args))))

  and lookup_v_m prereq body_f = (
    let (ib,jb,jjb,ii,jj) = List'.match5 prereq
    in let args = 
      [ ("$ii",An_expr(ii));
	("$jj",An_expr(jj));
	("$ib",An_expr(ib));
	("$jb",An_expr(jb));
	("$jjb",An_expr(jjb))
      ] @ common_parse_args
    in (body_f
	  (parse_expr
	     "$val[[$indx[[$jjb]]+$ii-1+($jj-1)*($rpntr[[$ib+1]]-$rpntr[[$ib]])]]"
	     args)))

  and bounds4 vars = (
    let (ib,jb,i,j) = List'.match4 vars
    in let args = 
      [ ("$i",An_expr(i));
	("$j",An_expr(j));
	("$ib",An_expr(ib));
	("$jb",An_expr(jb)) ] @ common_parse_args	
    in (parse_expr
	  "1<=$i && $i<=$n && 1<=$j && $j<=$n &&
	   1<=$ib && $ib<=$mb && 1<=$jb && $jb<=$nb"
	  args))

  and bounds4_offset vars = (
    let (ib,jb,ii,jj) = List'.match4 vars
    in let args = 
      [ ("$ii",An_expr(ii));
	("$jj",An_expr(jj));
	("$ib",An_expr(ib));
	("$jb",An_expr(jb)) ] @ common_parse_args	
    in (parse_expr
	  "1<=$ii && $ii<=$n && 1<=$jj && $jj<=$n &&
	   1<=$ib && $ib<=$mb && 1<=$jb && $jb<=$nb"
	  args))

  in let acc_mthd (name,in_fs,out_f,info) = {
    am_name = name; am_in_fields = in_fs;
    am_out_field = out_f; am_info = info
  }

  in let common_methods = 
    [ ("enum_ib",[],"ib",Am_rel(Am_rel_interval(enum_ib_m)));
      ("enum_jjb",["ib"],"jjb",Am_rel(Am_rel_interval(enum_jjb_m)));
      ("lookup_jb_m",["ib";"jjb"],"jb",Am_fun(O_1,Am_fun_lookup(lookup_jb_m)));

      ("enum_ii",["ib";"jjb"],"ii",Am_rel(Am_rel_interval(enum_ii_m)));
      ("enum_jj",["ib";"jjb";"jb"],"jj",Am_rel(Am_rel_interval(enum_jj_m)));

      ("lookup_v",["ib";"jb";"jjb";"ii";"jj"],
       "v",Am_fun(O_1,Am_fun_lookup(lookup_v_m))); 
    ]
  and ij_lookups = 
    [ ("lookup_i",["ib";"ii"],"i",Am_fun(O_1,Am_fun_lookup(lookup_i_m)));
      ("lookup_j",["jb";"jj"],"j",Am_fun(O_1,Am_fun_lookup(lookup_j_m)));
    ] 

  in let acc_mthds = 
    (on_kind kind
       (common_methods @ ij_lookups)
       (common_methods @ ij_lookups)
       common_methods)

  and make_bounds = 
    (on_kind kind 
       (rectangular_bounds1 "n" "n" args)
       bounds4
       bounds4_offset)

  in { bb_index_fields = (on_kind kind
			    ["i";"j"]
			    ["ib";"jb";"i";"j"]
			    ["ib";"jb";"ii";"jj"]);
       bb_value_field = "v";
       bb_access_methods = List.map acc_mthd acc_mthds;
       bb_ordered_access = [];
       bb_bounds_maker = make_bounds }
      
end (* vbr_maker *)
    	;;


register "vbr_basic" (vbr_maker Basic);;    
register "vbr_block" (vbr_maker Block);;    
register "vbr_block_offset" (vbr_maker Block_offset);;    


      




