(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Bsct_base

let usage str =
    if str <> "" then
	Printf.fprintf stderr "%s\n" str;
    Printf.fprintf stderr "%s\n"
	("Usage: bernoulli [-m4] [-m4cmd prog] [-m4opt opts]"
	    ^" [file.bmf] [-o filename.c]");
    Pervasives.exit 1
;;

(***************************************************************************)

let rec main () =

    (* Preprocessing *)
    let m4_p = ref false
    and m4cmd = ref Config.m4_command
    and m4opts = ref ""
    and default_m4opts = ""
    in let set_m4 () = m4_p := true
    and set_m4cmd str = m4cmd := str
    and add_m4opt str =
	if !m4opts = "" then
	    m4opts := str
	else
	    m4opts := (!m4opts)^" "^str

    (* Input file *)
    in let infile_p = ref false
    and infile_name = ref "-"
    in let anonfun str = 
	if !infile_p then usage "Multiple source files not allowed.";
	infile_p := true;
	infile_name := str

    (* Output file *)
    in let outfile_p = ref false
    and outfile_name = ref "-"
    and outfile_width = ref 72
    in let set_outfile_name str =
	if !outfile_p then usage "Multiple output files not allowed.";
	outfile_p := true;
	outfile_name := str
    and set_outfile_width w =
      outfile_width := w

    (* Parse the command line args *)
    in let dummy =
	Arg.parse
	    [
	     ("-o", Arg.String(set_outfile_name),
	      "Output file");
             ("-ow", Arg.Int(set_outfile_width),
	      "Width of output file (def=80, infty=-1)");
	     ("-m4", Arg.Unit(set_m4),
	      "Run the M4 preprocessor");
	     ("-m4cmd", Arg.String(set_m4cmd),
	      "Command for invoking M4");
	     ("-m4opt", Arg.String(add_m4opt),
	      "Options to M4");
	     ("--", Arg.String anonfun,
	      "Input source file")
	    ]
	    anonfun "err msg goes here"
    in

    (* M4 Preprocessing *)
    let prepout_name =
      if !m4_p then
	Bmf_parser_aux.do_m4_preproc !m4cmd (!m4opts^default_m4opts)
	  !infile_name
      else
	!infile_name
    in
    (* Read the source file *)
    let top_decls =
      if prepout_name = "-"  then
	Bmf_parser_aux.parse_stdin ()
      else
	Bmf_parser_aux.parse_file prepout_name
    in
    (* Remove the preprocessor output file *)
    if !infile_name <> prepout_name then
      Sys.remove prepout_name;

    (* sparsity analysis *)
    let top_decls = Sa.sparsity_analysis_procs top_decls in
    (* query scheduling *)
    let top_decls = Scheduler_bmf.schedule_procs false top_decls in
    (* dense join scheduling *)
    let top_decls =
      Joiner.schedule_procs Joiner.Sched_just_dense 1 top_decls
    in
    (* dense instantiation *)
    let top_decls =
      Instantiator.instantiate top_decls Instantiator.Inst_just_dense
    in
    (* dense optimizations and BLAS substitutions *)
    let top_decls = Dense.optimize_procs true top_decls in
    (* remaining join scheduling *)
    let top_decls =
      Joiner.schedule_procs Joiner.Sched_everything 1 top_decls
    in
    (* remaining instantiation *)
    let top_decls =
      Instantiator.instantiate top_decls
	    Instantiator.Inst_everything
    in

    (* expand intrinsics for generated C. *)
    let top_decls = List.map Genc.expand_intrinsics top_decls in
    (* "Normalize" the output. I.e, cleanup and simplify gensym'ed names. *)
    let top_decls =  Bmf_printer.normalize_globals top_decls in
    (* Write the output file *)
    if not !outfile_p && !infile_p then
      outfile_name :=
	 (if Filename.check_suffix !infile_name ".bmf" then
	   Filename.chop_suffix !infile_name ".bmf"
	 else
	   !outfile_name)
	 ^".c";
    if !outfile_name = "-" then
      Genc.to_stdout top_decls
    else
      Genc.to_file !outfile_name top_decls
;;

Printexc.catch main ();;

