(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers
open Bsct_base

(**************************************************************)
(*              Data types for BMF language                   *)
(**************************************************************)

type 
  annotation =
    An_type of bmf_type
  | An_stmt of bmf_stmt
  | An_expr of bmf_expr
  | An_string of string
  | An_list of annotation list
and
  annotations =
    An_map              of annotation String_map.t

and
  scalar_type =
    Type_int
  | Type_real
  | Type_bool
  | Type_void
  | Type_string

and 
  bmf_type = 
    Type_scalar         of scalar_type
  | Type_proc           of bmf_type list * bmf_type
  | Type_array		of array_type
  | Type_ptr		of bmf_type
  | Type_ann            of annotation
  | Type_unknown
and array_type =
  | Array_dense         of dense_lb_type   (* lower bounds, 0 or 1 *)
			 * bmf_expr list   (* upper bounds dimen *)
			 * dense_majorness (* majorness *)
			 * scalar_type     (* base type *)
  | Array_bb            of int * scalar_type * 
	                   string * (string * annotation) list
  | Array_view of 
      (*out_vars*) bmf_frame
	* (*in_vars*) bmf_frame
	* (*constraints*) bmf_expr
        * (*base_array*) array_type
and dense_lb_type = Dense_lb_0 | Dense_lb_1
and dense_majorness = Dense_row_major | Dense_col_major

and
  bmf_ident =
    Ident_valid of string * bmf_type * bmf_ident_stuff * annotation
  | Ident_null

and
  bmf_ident_stuff =
    (* mutable variable: name, type *)
    Ident_var		of bmf_expr

    (* non-mutable variable used in sets, loops *)
  | Ident_bind

    (* args, return val, stmt *)
  | Ident_proc          of bmf_frame * bmf_ident * bmf_stmt

and
  bmf_frame = 
     Frame              of bmf_ident list

and
  bmf_expr = 
    Expr_int            of int
  | Expr_real           of float
  | Expr_string         of string
  | Expr_bool           of bool
  | Expr_void
  | Expr_var            of string
  | Expr_op             of bmf_op * bmf_expr list
  | Expr_forall         of bmf_frame * bmf_expr
  | Expr_exist          of bmf_frame * bmf_expr

  (* accesses and lookup can happen using different functional
     depedencies. string is the name of the fucntional dependency
  *)
  | Expr_access         of bmf_access * bmf_expr * string * bmf_expr list
  | Expr_set            of bmf_frame * bmf_expr

and
  bmf_op =
    Op_plus
  | Op_times
  | Op_not
  | Op_neg
  | Op_div
  | Op_implies
  | Op_and
  | Op_or
  | Op_leq
  | Op_eq
  | Op_sqrt
  | Op_addr
  | Op_deref
  | Op_other of string
and
  bmf_access =
    Access_lookup
  | Access_weak
  | Access_strong
and
  bmf_stmt =
    Stmt_assign         of bmf_expr * bmf_expr
  | Stmt_seq            of bmf_stmt * bmf_stmt
  
  (* call lhs_expr := f_ident (expr_list) *)
  | Stmt_call           of bmf_expr      (* f_ident *)
                         * bmf_expr      (* lhs_expr *)
                         * bmf_expr list (* expr_list *)

  | Stmt_for            of bmf_ident (* index var *)
                         * bmf_expr  (* lb *)
                         * bmf_expr  (* ub *)
                         * bmf_expr  (* step *)
                         * bmf_stmt  (* body *)
  | Stmt_while          of bmf_expr * bmf_stmt
  | Stmt_if             of bmf_expr * bmf_stmt * bmf_stmt
  | Stmt_let            of bmf_frame     (* decls of the idents *)
			 * bmf_stmt      (* body *)
  | Stmt_do             of bmf_expr * bmf_stmt
  | Stmt_nop
  | Stmt_directive      of annotation * bmf_stmt
;;

(************************************************************************)

exception Invalid_bmf_s_expr of string

let bad_bmf_s_expr sx =
    raise (Invalid_bmf_s_expr (Sx.to_string sx))
let make_list f els = Sx.from_list (List.map f els)
let destr_list f sx = List.map f (Sx.to_list sx)
let sx_to_list sx =
  try Sx.to_list sx
  with Sx.Syntax_error -> bad_bmf_s_expr sx

let rec to_s_expr top_decls =
  Sx.Pair(Sx.Symbol("bmf"),
	  Sx.from_list (List.map bmf_ident_to_s_expr top_decls))

and from_s_expr sx =
    match sx with
      Sx.Pair(Sx.Symbol("bmf"),sx') ->
    	List.map bmf_ident_from_s_expr (Sx.to_list sx')
    | _ -> raise (Invalid_argument "Bmf.from_s_expr")

and annotation_to_s_expr ann =
  Sx.from_list
    (match ann with
      An_type(typ) -> [ Sx.Symbol("An_type") ; bmf_type_to_s_expr typ ]
    | An_stmt(stm) -> [ Sx.Symbol("An_stmt") ; bmf_stmt_to_s_expr stm ]
    | An_expr(exp) -> [ Sx.Symbol("An_expr") ; bmf_expr_to_s_expr exp ]
    | An_string(str) -> [ Sx.Symbol("An_string") ; Sx.String(str) ]
    | An_list(lst) -> [ Sx.Symbol("An_list") ;
			make_list annotation_to_s_expr lst ])

and annotation_from_s_expr sx =
  match sx_to_list sx with
    [ Sx.Symbol("An_type"); typ ] -> An_type(bmf_type_from_s_expr typ)
  | [ Sx.Symbol("An_stmt"); stm ] -> An_stmt(bmf_stmt_from_s_expr stm)
  | [ Sx.Symbol("An_expr"); exp ] -> An_expr(bmf_expr_from_s_expr exp)
  | [ Sx.Symbol("An_string"); Sx.String(str) ] -> An_string(str)
  | [ Sx.Symbol("An_list"); lst ] ->
      An_list(destr_list annotation_from_s_expr lst)
  | _ -> bad_bmf_s_expr sx

and bmf_ident_to_s_expr id =
  Sx.from_list
    (match id with
      Ident_null -> [ Sx.Symbol("Ident_null") ]
    | Ident_valid(name,typ,stuff,ann) ->
	[ Sx.Symbol("Ident_valid") ;
	  Sx.Symbol(name) ;
	  bmf_type_to_s_expr typ ;
	  bmf_ident_stuff_to_s_expr stuff ;
	  annotation_to_s_expr ann ])

and bmf_ident_from_s_expr sx =
  match sx_to_list sx with
    [ Sx.Symbol("Ident_null") ] -> Ident_null
  | [ Sx.Symbol("Ident_valid"); Sx.Symbol(name); typ; stuff; ann ] ->
	Ident_valid
	(name,
	 bmf_type_from_s_expr typ,
	 bmf_ident_stuff_from_s_expr stuff,
	 annotation_from_s_expr ann)
  | _ -> bad_bmf_s_expr sx
    

and bmf_ident_stuff_to_s_expr stuff =
  Sx.from_list
    (match stuff with
      Ident_var(init) -> [ Sx.Symbol("Ident_var") ;
			   bmf_expr_to_s_expr init ]
    | Ident_bind -> [ Sx.Symbol("Ident_bind") ]
    | Ident_proc(Frame(args),res,body) ->
	[ Sx.Symbol("Ident_proc") ;
	  make_list bmf_ident_to_s_expr args ;
	  bmf_ident_to_s_expr res ;
	  bmf_stmt_to_s_expr body ])

and bmf_ident_stuff_from_s_expr sx =
  match sx_to_list sx with
    [ Sx.Symbol("Ident_var"); init ] ->
      Ident_var(bmf_expr_from_s_expr init)
  | [ Sx.Symbol("Ident_bind") ] -> Ident_bind
  | [ Sx.Symbol("Ident_proc"); args; res; body ] ->
      Ident_proc
	(Frame(destr_list bmf_ident_from_s_expr args),
	 bmf_ident_from_s_expr res,
	 bmf_stmt_from_s_expr body)
  | _ -> bad_bmf_s_expr sx

and bmf_type_to_s_expr typ =
  Sx.from_list
    (match typ with
      Type_scalar(sc_typ) -> 
	[ Sx.Symbol("Type_scalar") ;
	  scalar_type_to_s_expr sc_typ ]
    | Type_proc(args_typ,res_typ) ->
	[ Sx.Symbol("Type_proc") ;
	  make_list bmf_type_to_s_expr args_typ ;
	  bmf_type_to_s_expr res_typ ]
    | Type_array(atyp) ->
	[ Sx.Symbol("Type_array") ; 
	  array_type_to_s_expr atyp ]
    | Type_ptr(typ') ->
	[ Sx.Symbol("Type_ptr") ;
	  bmf_type_to_s_expr typ' ]
    | Type_ann(ann) ->
	[ Sx.Symbol("Type_ann") ;
	  annotation_to_s_expr ann ]
    | Type_unknown -> [ Sx.Symbol("Type_unknown") ])

and bmf_type_from_s_expr sx =
  match sx_to_list sx with
    [ Sx.Symbol("Type_scalar"); sc_typ ] ->
      Type_scalar(scalar_type_from_s_expr sc_typ)
  | [ Sx.Symbol("Type_proc"); args_typ; res_typ ] ->
      Type_proc
	(destr_list bmf_type_from_s_expr args_typ,
	 bmf_type_from_s_expr res_typ)
  | [ Sx.Symbol("Type_array") ; atyp ] ->
      Type_array(array_type_from_s_expr atyp)
  | [ Sx.Symbol("Type_ann"); ann ] ->
	Type_ann(annotation_from_s_expr ann)
  | _ -> bad_bmf_s_expr sx

and array_type_to_s_expr atyp =
  Sx.from_list
    (match atyp with
    | Array_bb(dim,base_typ,bb_name,bb_args) ->
	[ Sx.Symbol("Array_bb") ;
	  Sx.Int(dim) ;
	  scalar_type_to_s_expr base_typ ;
	  Sx.Symbol(bb_name) ;
	  Sx.from_list
	    (List.map 
	       (fun (tag,val0) ->
		 Sx.Pair(Sx.Symbol(tag), annotation_to_s_expr val0))
	       bb_args)	]
    | Array_dense(lb, ubs, maj, base_typ) ->
	[ Sx.Symbol("Array_dense") ;
	  Sx.Int
	    (match lb with
	      Dense_lb_0 -> 0
	    | Dense_lb_1 -> 1) ;
	  make_list bmf_expr_to_s_expr ubs ;
	  Sx.Symbol
	    (match maj with
	      Dense_row_major -> "row"
	    | Dense_col_major -> "col") ;
	  scalar_type_to_s_expr base_typ ]
    | Array_view(Frame(out_frm),Frame(in_frm),constr,base_atyp) ->
	[ Sx.Symbol("Array_view") ;
	  make_list bmf_ident_to_s_expr out_frm ;
	  make_list bmf_ident_to_s_expr in_frm ;
	  bmf_expr_to_s_expr constr ;
	  array_type_to_s_expr base_atyp ])

and array_type_from_s_expr sx =
  match sx_to_list sx with
  | [ Sx.Symbol("Array_bb"); Sx.Int(dim); base_typ; 
      Sx.Symbol(bb_name); bb_args_sx ] ->
	let bb_args =
	  List.map
	    (fun xx ->
	      match xx with
		Sx.Pair(Sx.Symbol(tag),val0) ->
		  (tag,annotation_from_s_expr val0)
	      | _ -> bad_bmf_s_expr sx)
	    (sx_to_list bb_args_sx)
	in
      	Array_bb
	  (dim, scalar_type_from_s_expr base_typ, bb_name, bb_args)
  | [ Sx.Symbol("Array_dense"); Sx.Int(lb); ubs; Sx.Symbol(maj);
      base_typ ] ->
	let lb =
	  match lb with
	    0 -> Dense_lb_0
	  | 1 -> Dense_lb_1
	  | _ -> bad_bmf_s_expr sx
	and maj =
	  match maj with
	    "row" -> Dense_row_major
	  | "col" -> Dense_col_major
	  | _ -> bad_bmf_s_expr sx
	in
	Array_dense(lb, destr_list bmf_expr_from_s_expr ubs, 
		   maj, scalar_type_from_s_expr base_typ)
  | [ Sx.Symbol("Array_view") ; out_frm ; in_frm; constr; base_atyp ] ->
      Array_view(Frame(destr_list bmf_ident_from_s_expr out_frm),
		 Frame(destr_list bmf_ident_from_s_expr in_frm),
		 bmf_expr_from_s_expr constr,
	  array_type_from_s_expr base_atyp)
  | _ -> bad_bmf_s_expr sx

and scalar_type_to_s_expr sc_typ =
  Sx.Symbol
    (match sc_typ with
      Type_int -> "Type_int"
    | Type_real -> "Type_real"
    | Type_bool -> "Type_bool"
    | Type_void -> "Type_void"
    | Type_string -> "Type_string")

and scalar_type_from_s_expr sx =
  match sx with
    Sx.Symbol("Type_int") -> Type_int
  | Sx.Symbol("Type_real") -> Type_real
  | Sx.Symbol("Type_bool") -> Type_bool
  | Sx.Symbol("Type_void") -> Type_void
  | Sx.Symbol("Type_string") -> Type_string
  | _ -> bad_bmf_s_expr sx
    
and bmf_expr_to_s_expr e =
  Sx.from_list
    (match e with
      Expr_void -> []
    | Expr_int(x) -> [Sx.Int(x)]
    | Expr_real(x) -> [Sx.Float(x)]
    | Expr_string(x) -> [Sx.String(x)]
    | Expr_bool(x) -> [Sx.Bool(x)]
    | Expr_var(name) -> [Sx.Symbol(name)]
    | Expr_op(op,args) ->
      	[ Sx.Symbol("Expr_op") ;
	  bmf_op_to_s_expr op;
	  make_list bmf_expr_to_s_expr args ]
    | Expr_set(Frame(ids),pred) ->
	[ Sx.Symbol("Expr_set") ;
	  make_list bmf_ident_to_s_expr ids ;
	  bmf_expr_to_s_expr pred ]
    | Expr_forall(Frame(ids),pred) ->
	[ Sx.Symbol("Expr_forall") ;
	  make_list bmf_ident_to_s_expr ids ;
	  bmf_expr_to_s_expr pred ]
    | Expr_exist(Frame(ids),pred) ->
	[ Sx.Symbol("Expr_exist") ;
	  make_list bmf_ident_to_s_expr ids ;
	  bmf_expr_to_s_expr pred ]
    | Expr_access(acc,r,mthd_name,args) ->
	[ Sx.Symbol("Expr_access") ;
	  Sx.Symbol
	    (match acc with
	      Access_lookup -> "lookup"
	    | Access_weak -> "weak"
	    | Access_strong -> "strong") ;
	  bmf_expr_to_s_expr r ;
	  Sx.String(mthd_name) ;
	  make_list bmf_expr_to_s_expr args ])
    
and bmf_expr_from_s_expr sx =
  match sx_to_list sx with
    [] -> Expr_void
  | [Sx.Int(x)] -> Expr_int(x)
  | [Sx.Float(x)] -> Expr_real(x)
  | [Sx.String(x)] -> Expr_string(x)
  | [Sx.Bool(x)] -> Expr_bool(x)
  | [Sx.Symbol(name)] -> Expr_var(name)
  | [ Sx.Symbol("Expr_op"); op; args ] ->
      Expr_op(bmf_op_from_s_expr op,
	      destr_list bmf_expr_from_s_expr args)
  | [ Sx.Symbol("Expr_set"); ids; pred ] ->
	  Expr_set(Frame(destr_list bmf_ident_from_s_expr ids),
		   bmf_expr_from_s_expr pred)
      
  | [ Sx.Symbol("Expr_forall"); ids; pred ] ->
	Expr_forall(Frame( destr_list bmf_ident_from_s_expr ids),
		    bmf_expr_from_s_expr pred)
  | [ Sx.Symbol("Expr_exist"); ids; pred ] ->
	Expr_exist(Frame(destr_list bmf_ident_from_s_expr ids),
		     bmf_expr_from_s_expr pred)
  | [ Sx.Symbol("Expr_access");	Sx.Symbol(acc_sx); r;
      Sx.String(mthd_name); args ] ->
	let acc =
	  match acc_sx with
	    "lookup" -> Access_lookup
	  | "weak" -> Access_weak
	  | "strong" -> Access_strong
	  | _ -> bad_bmf_s_expr sx
	in
	Expr_access(acc, bmf_expr_from_s_expr r, mthd_name,
		    destr_list bmf_expr_from_s_expr args)
  | _ -> bad_bmf_s_expr sx
	  
and bmf_op_to_s_expr op =
  match op with
    Op_plus -> Sx.Symbol("+")
  | Op_times -> Sx.Symbol("*")
  | Op_not -> Sx.Symbol("not")
  | Op_neg -> Sx.Symbol("-")
  | Op_div -> Sx.Symbol("/")
  | Op_and -> Sx.Symbol("and")
  | Op_or -> Sx.Symbol("or")
  | Op_leq -> Sx.Symbol("<=")
  | Op_eq -> Sx.Symbol("=")
  | Op_sqrt -> Sx.Symbol("sqrt")
  | Op_implies -> Sx.Symbol("=>")
  | Op_addr -> Sx.Symbol("%addr")
  | Op_deref -> Sx.Symbol("%deref")
  | Op_other(name) -> Sx.from_list [ Sx.Symbol("op") ; Sx.String(name) ]

and bmf_op_from_s_expr sx =
  match sx with
    Sx.Symbol("+") -> Op_plus
  | Sx.Symbol("*") -> Op_times
  | Sx.Symbol("not") -> Op_not
  | Sx.Symbol("-") -> Op_neg
  | Sx.Symbol("/") -> Op_div
  | Sx.Symbol("and") -> Op_and
  | Sx.Symbol("or") -> Op_or
  | Sx.Symbol("<=") -> Op_leq
  | Sx.Symbol("=") -> Op_eq
  | Sx.Symbol("sqrt") -> Op_sqrt
  | Sx.Symbol("=>") -> Op_implies
  | Sx.Symbol("%addr") -> Op_addr
  | Sx.Symbol("%deref") -> Op_deref
  | Sx.Pair(Sx.Symbol("op"), Sx.Pair(Sx.String(name), Sx.Nil)) -> 
      Op_other(name)
  | _ -> bad_bmf_s_expr sx

and bmf_stmt_to_s_expr s = 
  Sx.from_list
    (match s with
      Stmt_nop -> [ Sx.Symbol("Stmt_nop") ]
    | Stmt_seq(first,second) ->
	[ Sx.Symbol("Stmt_seq") ;
	  bmf_stmt_to_s_expr first ;
	  bmf_stmt_to_s_expr second ]
    | Stmt_assign(lhs,rhs) ->
	[ Sx.Symbol("Stmt_assign") ;
	  bmf_expr_to_s_expr lhs ;
	  bmf_expr_to_s_expr rhs ]
    | Stmt_call(f,lhs,args) ->
	[ Sx.Symbol("Stmt_call") ;
	  bmf_expr_to_s_expr f ;
	  bmf_expr_to_s_expr lhs ;
	  make_list bmf_expr_to_s_expr args ]
    | Stmt_for(i,lb,ub,step,body) ->
	[ Sx.Symbol("Stmt_for") ;
	  bmf_ident_to_s_expr i ;
	  bmf_expr_to_s_expr lb ;
	  bmf_expr_to_s_expr ub ;
	  bmf_expr_to_s_expr step ;
	  bmf_stmt_to_s_expr body ]
    | Stmt_while(pred,body) ->
	[ Sx.Symbol("Stmt_while") ;
	  bmf_expr_to_s_expr pred ;
	  bmf_stmt_to_s_expr body ]
    | Stmt_if(pred,thn,els) ->
	[ Sx.Symbol("Stmt_if") ;
	  bmf_expr_to_s_expr pred ;
	  bmf_stmt_to_s_expr thn ;
	  bmf_stmt_to_s_expr els ]
    | Stmt_let(Frame(ids),body) -> 
	[ Sx.Symbol("Stmt_let") ;
	  make_list bmf_ident_to_s_expr ids ;
	  bmf_stmt_to_s_expr body ]
    | Stmt_do(pred,body) ->
	[ Sx.Symbol("Stmt_do") ;
	  bmf_expr_to_s_expr pred ;
	  bmf_stmt_to_s_expr body ]
    | Stmt_directive(ann,body) ->
	[ Sx.Symbol("Stmt_directive") ;
	  annotation_to_s_expr ann ;
	  bmf_stmt_to_s_expr body ])

and bmf_stmt_from_s_expr sx = 
  match sx_to_list sx with
    [ Sx.Symbol("Stmt_nop") ] -> Stmt_nop
  | [ Sx.Symbol("Stmt_seq"); first; second ] ->
      Stmt_seq(bmf_stmt_from_s_expr first,
	       bmf_stmt_from_s_expr second)
  | [ Sx.Symbol("Stmt_assign"); lhs; rhs ] ->
      Stmt_assign(bmf_expr_from_s_expr lhs,
		  bmf_expr_from_s_expr rhs)
  | [ Sx.Symbol("Stmt_call"); f; lhs; args ] ->
      Stmt_call(bmf_expr_from_s_expr f,
		bmf_expr_from_s_expr lhs,
		destr_list bmf_expr_from_s_expr args)
  | [ Sx.Symbol("Stmt_for"); i; lb; ub; step; body ] ->
      Stmt_for(bmf_ident_from_s_expr i,
	       bmf_expr_from_s_expr lb,
	       bmf_expr_from_s_expr ub,
	       bmf_expr_from_s_expr step,
	       bmf_stmt_from_s_expr body)
  | [ Sx.Symbol("Stmt_while"); pred; body ] ->
      Stmt_while(bmf_expr_from_s_expr pred,
		 bmf_stmt_from_s_expr body)
  | [ Sx.Symbol("Stmt_if"); pred; thn; els ] ->
      Stmt_if(bmf_expr_from_s_expr pred,
	      bmf_stmt_from_s_expr thn,
	      bmf_stmt_from_s_expr els)
  | [ Sx.Symbol("Stmt_let"); ids; body ] ->
      Stmt_let(Frame(destr_list bmf_ident_from_s_expr ids),
	       bmf_stmt_from_s_expr body)
  | [ Sx.Symbol("Stmt_do"); pred; body ] ->
      Stmt_do(bmf_expr_from_s_expr pred,
	      bmf_stmt_from_s_expr body)
  | [ Sx.Symbol("Stmt_directive"); ann; body ] ->
      Stmt_directive(annotation_from_s_expr ann,
		     bmf_stmt_from_s_expr body) 
  | _ -> bad_bmf_s_expr sx

(************************************************************************)

let intrinsic_name_p name =
  (String.length name > 0) & (String.get name 0 = '%')

exception Bmf_null_ident of unit;;
exception Bmf_undeclared_var of string;;
exception Bmf_undefined_bb of string;;
exception Bmf_abort of string;;
exception Bmf_expr_type_mismatch of bmf_expr * bmf_type;;

exception Bmf_must_be_var_expr of bmf_expr;;
exception Bmf_must_be_proc_id of bmf_ident;;
exception Bmf_must_be_proc_expr of bmf_expr;;
exception Bmf_must_be_index of bmf_ident;;
exception Bmf_must_be_set_expr of bmf_expr;;
exception Bmf_must_be_bb_ident of bmf_ident;;

exception Bmf_dont_know_how_to_type of bmf_ident;;

exception Bmf_not_unary;;
exception Bmf_not_binary;;
exception Bmf_not_if;;

exception Bmf_not_array_or_relation of bmf_ident;;

let annotation_2_cell_list an =
    (* convert an annotation into a list of
       (name, ann) cells
     *)
    let f ann =
        match ann with
          | (An_list [An_string name; a2]) -> 
              (name, a2)

          | (An_list ((An_string name)::rest)) ->
              (name, An_list rest)

          | _ -> failwith "Bad annotation cell format"

    in
      match an with
        | (An_list lst) -> List.map f lst
        | _ -> failwith "Annotation must be a list of cells"
;;

let cell_list_2_annotation (lst : (string * annotation) list) =
    let f (name,an) = (
        match an with
          | (An_list lst') -> An_list ((An_string name)::lst')
          | _ -> An_list [ An_string name; an ])

    in An_list (List.map f lst)
;;

let collect_annotations an =
    An_map (
      List.fold_left
        (fun m (name,ann) -> (String_map.add name ann m))
        String_map.empty
        (annotation_2_cell_list an))
;;

let has_annotation (An_map m) (s:string) =
    match (find_value (fun () -> String_map.find s m)) with
      | Unfound -> false
      | _ -> true
;;

let get_annotation (An_map m) (s:string) =
    match (find_value (fun () -> String_map.find s m)) with
      | Unfound -> failwith ("Annotation "^s^" not found")
      | (Found an) -> an
;;

(* todo : verify that variables are not redefined *)

let rec 
  bmf_frame2list =
    function Frame l -> l
and
  bmf_ident_name =
    function
      (Ident_valid(s,_,_,_)) -> s
    | Ident_null -> raise (Bmf_null_ident ())
and
  bmf_ident_type =
    function
      (Ident_valid(_,t,_,_)) -> t
    | Ident_null -> raise (Bmf_null_ident ())

and bmf_lookup_scope s scope =
  begin
    match scope with
      (id::rest_scope) -> if (s = bmf_ident_name id)
                            then id
                            else bmf_lookup_scope s rest_scope
    | [] -> Ident_null
  end

and bmf_lookup_scope_assert scope s =
    begin
      match (bmf_lookup_scope s scope) with
        | Ident_null -> raise (Bmf_null_ident ())
        | id -> id
    end
and
  bmf_get_type4var =
    fun
      scope s -> 
        (match bmf_lookup_scope s scope with
          (Ident_valid(_,t,_,_)) -> t
        | Ident_null -> raise (Bmf_null_ident ()))
and
  bmf_is_in_scope =
    fun 
      s scope -> (Ident_null <> bmf_lookup_scope s scope)
and
  bmf_order_in_scope =
    fun
      scope s -> bmf_order_in_scope_aux scope s 0
and
  bmf_order_in_scope_aux =
    fun
      p0 p1 p2 -> match (p0,p1,p2) with (((Ident_valid(name,_,_,_))::rest), s, n) -> 
        if (name = s) then n else bmf_order_in_scope_aux rest s (n+1)
    | ((Ident_null::rest), s, n) ->  raise (Bmf_null_ident ())
    | ([], s, n) -> raise (Bmf_undeclared_var s)
and
  bmf_verify_frame =
    fun
      p0 p1 -> match (p0,p1) with ((id::rest), scope) -> (bmf_verify_ident id scope) & (bmf_verify_frame rest (id::scope))
    | ([], scope) -> true

and bmf_verify_ident id scope =
  match id with
    Ident_valid(_,_,Ident_var(init),_) ->
      bmf_verify_expr scope init
  | Ident_valid(_,_,Ident_bind,_) -> true
  | Ident_valid(_,_,Ident_proc(args, ret, stmt),_) as proc -> 
      let scope' = (bmf_frame2list args)@(proc::scope) in
      let scope'' =
	if ret = Ident_null then
	  scope'
	else
	  ret::scope'
      in bmf_verify_stmt scope'' stmt
  | Ident_null -> raise (Bmf_null_ident ())

and
  bmf_verify_expr =
    fun 
      p0 p1 -> match (p0,p1) with (scope, (Expr_int(_))) -> true
    | (scope, (Expr_real(_))) -> true
    | (scope, (Expr_string(_))) -> true
    | (scope, (Expr_bool(_))) -> true
    | (scope, (Expr_void)) -> true
    | (scope, (Expr_var(s))) -> 
        if (bmf_is_in_scope s scope) then true
        else if intrinsic_name_p s then true
	else raise (Bmf_undeclared_var s)
    | (scope, (Expr_op(op, l))) -> List.for_all (bmf_verify_expr scope) l
    | (scope, (Expr_forall(fr,e))) -> bmf_verify_expr ((bmf_frame2list fr)@scope) e
    | (scope, (Expr_exist(fr,e))) -> bmf_verify_expr ((bmf_frame2list fr)@scope) e
    | (scope, (Expr_set(fr,e))) -> bmf_verify_expr ((bmf_frame2list fr)@scope) e

    (* todo : make this verifiers *)

    | (scope, (Expr_access(ac, e, fdep, l))) -> List.for_all (bmf_verify_expr scope) l

and bmf_verify_stmt scope stmt = 
  match stmt with
    Stmt_assign(e1,e2) -> 
      (bmf_verify_expr scope e1) & (bmf_verify_expr scope e2)
  | Stmt_seq(s1,s2) -> 
      (bmf_verify_stmt scope s1) & (bmf_verify_stmt scope s2)
  | Stmt_call(Expr_var(name),lhs,args) ->
      ((bmf_verify_expr scope lhs) & 
        (let proc_ident = bmf_lookup_scope name scope
        in
        match proc_ident with
	  _ when intrinsic_name_p name -> true
	| Ident_null -> raise (Bmf_undeclared_var name)
        | (Ident_valid(_,_,Ident_proc(_,_,_),_)) -> true
        | _ -> raise (Bmf_must_be_proc_id proc_ident)) &
       (List.for_all (bmf_verify_expr scope) args))
  | Stmt_call(proc,lhs,args) -> raise (Bmf_must_be_proc_expr proc)
  | Stmt_for((Ident_valid(_,_,Ident_bind,_) as id),lb,ub,step,stmt) ->
      (let scope2 = id :: scope
      in
      (bmf_verify_expr scope2 lb) &
      (bmf_verify_expr scope2 ub) &
      (bmf_verify_expr scope2 step) &
      (bmf_verify_stmt scope2 stmt))
  | Stmt_for(var,_,_,_,_) -> raise (Bmf_must_be_index var)
  | Stmt_while(e,s) -> 
      (bmf_verify_expr scope e) & 
      (bmf_verify_stmt scope s)
  | Stmt_if(e,s1,s2) ->
      (bmf_verify_expr scope e) & 
      (bmf_verify_stmt scope s1) &
      (bmf_verify_stmt scope s2)
  | Stmt_let(f,stmt) ->
      (bmf_verify_stmt ((bmf_frame2list f) @ scope) stmt)
  | Stmt_do(Expr_set(f,e),s) -> 
      (let scope2 = (bmf_frame2list f) @ scope
      in
      (bmf_verify_expr scope2 e) & (bmf_verify_stmt scope2 s))
  | Stmt_do(e,_) -> raise (Bmf_must_be_set_expr e)
    (* *todo* *)
  | Stmt_directive(_,s) -> (bmf_verify_stmt scope s)
  | Stmt_nop -> true

and
  bmf_expr_op_arg e =
    match e with
      Expr_op(_,[arg]) -> arg
    | _ -> raise Bmf_not_unary
and
  bmf_expr_op_arg1 e =
    match e with
      Expr_op(_,[arg1;arg2]) -> arg1
    | _ -> raise Bmf_not_binary
and
  bmf_expr_op_arg2 e =
    match e with
      Expr_op(_,[arg1;arg2]) -> arg2
    | _ -> raise Bmf_not_binary
and
  bmf_expr_op_args e =
    match e with
      Expr_op(_,args) -> args
    | _ -> raise Bmf_not_unary
and
  (* bmf_expr_args_by_op "+" (+ 1 (+ 2) (+ 3 (+ 4 5) 6)) -> [1;2;3;4;5;6] *)
  bmf_expr_args_by_op op e =
    match e with
      Expr_op(op',args) when op' = op ->
        List'.flat_map (bmf_expr_args_by_op op) args
    | _ -> [e]
and
  bmf_stmt_if_then st = 
    match st with
      Stmt_if(test,thn,els) -> thn
    | _ -> raise Bmf_not_if
and
  bmf_stmt_if_else st = 
    match st with
      Stmt_if(test,thn,els) -> els
    | _ -> raise Bmf_not_if
;;

(************************************************************************)
(* Expression construction						*)
(************************************************************************)

let make_nary_expr op identity annihil args =
    let args' = List'.flat_map (bmf_expr_args_by_op op) args
    in let args'' = List'.difference args' [identity]
    in
        match args'' with
	  [] -> identity
	| [arg] -> arg
	| _ when List.mem annihil args'' -> annihil
	| args'' -> Expr_op(op,args'')
;;

let rec make_neg e =
    match e with
      Expr_op(Op_neg,[e']) -> e'
    | Expr_op(Op_plus,args) -> Expr_op(Op_plus, List.map make_neg args)
    | Expr_int(x) -> Expr_int(-x)
    | Expr_real(x) -> Expr_real(-.x)
    | _ -> Expr_op(Op_neg,[e])
;;

let make_not e =
    match e with
      Expr_op(Op_not,[e']) -> e'
    | _ -> Expr_op(Op_not,[e])
;;

let make_plus = make_nary_expr Op_plus (Expr_int(0)) Expr_void
and make_times = make_nary_expr Op_times (Expr_int(1)) (Expr_int(0))
and make_and =
    make_nary_expr Op_and (Expr_bool(true)) (Expr_bool(false))
and make_or =
    make_nary_expr Op_or (Expr_bool(false)) (Expr_bool(true))
;;

(* other stuff *)

let rec
  bmf_temp_var_name =
    fun 
      (names:(string list)) pfix -> 
        if ((not ((List.mem pfix names))))
        then pfix
        else 
          let i = ref 0
          in (
            while (List.mem (pfix ^ "_" ^ (string_of_int !i)) names)
            do (incr i) done ;
            pfix ^ "_" ^ (string_of_int !i))
and
  bmf_create_temp_names =
    fun
      names (n:int) (pfix:string) ->
        if (n = 0) then []
        else 
          let name=(bmf_temp_var_name names pfix)
          in name :: (bmf_create_temp_names (name::names) (n-1) pfix)
;;

let
  ident2expr_var =
    fun (id:bmf_ident) -> Expr_var(bmf_ident_name id)
;;

            
let is_a_proc_decl decl =
    match decl with
      Ident_valid(_,_,Ident_proc(_,_,_),_) -> true
    | _ -> false
;;

(************************************************************************)
(*			  Command line processing			*)
(************************************************************************)

let register_switch name doc bmf_f =
  let sx_f sx args =
    let bmf = from_s_expr sx 
    in let (bmf',args') = bmf_f bmf args in
    let sx' =
      if bmf == bmf' then
 	sx
      else
 	to_s_expr bmf' 
    in (sx',args')
  in
  Bsct_toplevel.register_switch name "bmf" doc sx_f

let _ =
  register_switch "-V"
    "-V (bmf) -- Validate the current BMF program repr."
    (fun top_decls args ->
      let name = Bsct_toplevel.get_arg_name (List.hd args)
      and (options,param,args') = Bsct_toplevel.parse_arg false args
      in let _ = Bsct_toplevel.ignore_options options name
      in let _ =
	List.iter 
	  (fun id -> 
      	    assert (bmf_verify_ident id top_decls = true))
	  top_decls
      in (top_decls,args'))
