(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Bsct_base
open Ssisl
open Containers

type annotation =
  | An_type of bmf_type
  | An_stmt of bmf_stmt
  | An_expr of bmf_expr
  | An_string of string
  | An_list of annotation list
and annotations = An_map of annotation String_map.t
and scalar_type =
  | Type_int
  | Type_real
  | Type_bool
  | Type_void
  | Type_string
and bmf_type =
  | Type_scalar of scalar_type
  | Type_proc of bmf_type list * bmf_type
  | Type_array of array_type
  | Type_ptr of bmf_type
  | Type_ann of annotation
  | Type_unknown
and array_type =
  | Array_dense         of dense_lb_type   (* lower bounds, 0 or 1 *)
			 * bmf_expr list   (* upper bounds dimen *)
			 * dense_majorness (* majorness *)
			 * scalar_type     (* base type *)
  | Array_bb of int * scalar_type * 
	       string * (string * annotation) list
  (* A view on another array. Let A' be "base" array and "A" be the view
     array. A[out_vars] names the indices of A, and A'[in_vars] names the
     indices of A'. The linear equalities and inequalities, constraints,
     relate the two sets of indices. *)
  | Array_view of 
      (*out_vars*) bmf_frame
	* (*in_vars*) bmf_frame
	* (*constraints*) bmf_expr
        * (*base_array*) array_type
and dense_lb_type = Dense_lb_0 | Dense_lb_1
and dense_majorness = Dense_row_major | Dense_col_major
and bmf_ident =
  | Ident_valid of string * bmf_type * bmf_ident_stuff * annotation
  | Ident_null
and bmf_ident_stuff =
  | Ident_var of bmf_expr
  | Ident_bind
  | Ident_proc of bmf_frame * bmf_ident * bmf_stmt
and bmf_frame = Frame of bmf_ident list
and bmf_expr =
  | Expr_int of int
  | Expr_real of float
  | Expr_string of string
  | Expr_bool of bool
  | Expr_void
  | Expr_var of string
  | Expr_op of bmf_op * bmf_expr list
  | Expr_forall of bmf_frame * bmf_expr
  | Expr_exist of bmf_frame * bmf_expr
  | Expr_access of bmf_access * bmf_expr * string * bmf_expr list
  | Expr_set of bmf_frame * bmf_expr
and bmf_op =
  | Op_plus
  | Op_times
  | Op_not
  | Op_neg
  | Op_div
  | Op_implies
  | Op_and
  | Op_or
  | Op_leq
  | Op_eq
  | Op_sqrt
  | Op_addr
  | Op_deref
  | Op_other of string
and bmf_access = Access_lookup | Access_weak | Access_strong
and bmf_stmt =
  | Stmt_assign of bmf_expr * bmf_expr
  | Stmt_seq of bmf_stmt * bmf_stmt
  | Stmt_call of bmf_expr * bmf_expr * bmf_expr list
  | Stmt_for of bmf_ident * bmf_expr * bmf_expr * bmf_expr * bmf_stmt
  | Stmt_while of bmf_expr * bmf_stmt
  | Stmt_if of bmf_expr * bmf_stmt * bmf_stmt
  | Stmt_let of bmf_frame * bmf_stmt
  | Stmt_do of bmf_expr * bmf_stmt
  | Stmt_nop
  | Stmt_directive of annotation * bmf_stmt

exception Invalid_bmf_s_expr of string

val from_s_expr: Sx.t -> bmf_ident list
val to_s_expr: bmf_ident list -> Sx.t

val intrinsic_name_p : string -> bool
exception Bmf_null_ident of unit
exception Bmf_undeclared_var of string
exception Bmf_undefined_bb of string
exception Bmf_abort of string
exception Bmf_expr_type_mismatch of bmf_expr * bmf_type
exception Bmf_must_be_var_expr of bmf_expr
exception Bmf_must_be_proc_id of bmf_ident
exception Bmf_must_be_proc_expr of bmf_expr
exception Bmf_must_be_index of bmf_ident
exception Bmf_must_be_set_expr of bmf_expr
exception Bmf_must_be_bb_ident of bmf_ident
exception Bmf_dont_know_how_to_type of bmf_ident
exception Bmf_not_unary
exception Bmf_not_binary
exception Bmf_not_if
exception Bmf_not_array_or_relation of bmf_ident

val annotation_2_cell_list : annotation -> (string * annotation) list
val cell_list_2_annotation : (string * annotation) list -> annotation
val collect_annotations : annotation -> annotations
val has_annotation : annotations -> string -> bool
val get_annotation : annotations -> string -> annotation
val bmf_frame2list : bmf_frame -> bmf_ident list
val bmf_ident_name : bmf_ident -> string
val bmf_ident_type : bmf_ident -> bmf_type
val bmf_lookup_scope : string -> bmf_ident list -> bmf_ident
val bmf_lookup_scope_assert : bmf_ident list -> string -> bmf_ident
val bmf_get_type4var : bmf_ident list -> string -> bmf_type
val bmf_is_in_scope : string -> bmf_ident list -> bool
val bmf_order_in_scope : bmf_ident list -> string -> int
val bmf_order_in_scope_aux : bmf_ident list -> string -> int -> int
val bmf_verify_frame : bmf_ident list -> bmf_ident list -> bool
val bmf_verify_ident : bmf_ident -> bmf_ident list -> bool
val bmf_verify_expr : bmf_ident list -> bmf_expr -> bool
val bmf_verify_stmt : bmf_ident list -> bmf_stmt -> bool
val bmf_expr_op_arg : bmf_expr -> bmf_expr
val bmf_expr_op_arg1 : bmf_expr -> bmf_expr
val bmf_expr_op_arg2 : bmf_expr -> bmf_expr
val bmf_expr_op_args : bmf_expr -> bmf_expr list
val bmf_expr_args_by_op : bmf_op -> bmf_expr -> bmf_expr list
val bmf_stmt_if_then : bmf_stmt -> bmf_stmt
val bmf_stmt_if_else : bmf_stmt -> bmf_stmt
val make_nary_expr :
  bmf_op -> bmf_expr -> bmf_expr -> bmf_expr list -> bmf_expr
val make_neg : bmf_expr -> bmf_expr
val make_not : bmf_expr -> bmf_expr
val make_plus : bmf_expr list -> bmf_expr
val make_times : bmf_expr list -> bmf_expr
val make_and : bmf_expr list -> bmf_expr
val make_or : bmf_expr list -> bmf_expr
val bmf_temp_var_name : string list -> string -> string
val bmf_create_temp_names : string list -> int -> string -> string list
val ident2expr_var : bmf_ident -> bmf_expr
val is_a_proc_decl : bmf_ident -> bool

(************************************************************************)
(* Command-line switch processing.					*)
(************************************************************************)

val register_switch:
    ((*name:*)string) -> ((*doc:*)string) ->
      (bmf_ident list -> string list -> bmf_ident list *string list) ->
	unit
