(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl

open Bmf;;
open Bmf_util;;

type join_type = {
    join_t_var : string;
    join_bounds : join_bounds_type;
    join_terms : join_term_type list;
    join_pred : Wff.Simple.t;
    join_assertion : Wff.Simple.t;
    join_body : bmf_stmt
} 
and determined_type = {
    determined_term : join_term_type;
    determined_pred : Wff.Simple.t;
    determined_assertion : Wff.Simple.t;
    determined_body : bmf_stmt
}
and join_term_type = {
    jt_flag : string;
    jt_kind : join_term_kind;
    jt_constr : bmf_expr;
    jt_bounds : join_bounds_type
}
and join_term_kind =
|   Jtk_enum_simple of join_term_equality
|   Jtk_enum_indexing of join_term_equality
                * join_term_equality
|   Jtk_enum_searching of join_term_equality
                * join_term_equality
                * join_term_equality
|   Jtk_lookup  of join_term_equality

and join_term_equality =
    Jtk_eq of string	    
	    * string	    
	    * string	    
	    * bmf_expr list 
and join_bounds_type = (bmf_expr list) * (bmf_expr list)
;;

exception Not_a_valid_join;;
exception Not_a_valid_determined;;
exception Not_valid_bounds;;

exception Not_a_valid_jt;;

let make_decl name =
    Ident_valid(name,Type_scalar(Type_int),Ident_bind,An_list([]))
and make_bool_decl name =
    Ident_valid(name,Type_scalar(Type_bool),Ident_bind,An_list([]))
;;

(* Joins *)

let rec stmt_is_join s =
  match s with
    Stmt_do(Expr_set(Frame(_),Expr_op(Op_other("%join"),_)),_) -> 
      true
  |  _ -> false

and join2stmt
    ({ join_t_var = t_var; join_bounds = bounds;
       join_terms = terms;
       join_pred = pred; join_assertion = assertion;
       join_body = body } as join_type) =
  Stmt_do
    (Expr_set
       (Frame
	  ((make_decl t_var)
	   ::(List'.flat_map bound_in_term terms)),
	Expr_op(Op_other("%join"),[
		Expr_op(Op_other("%join_t_var"),[
		    	Expr_var(t_var)]);
		Expr_op(Op_other("%join_bounds"),[
		    	bounds2expr (Expr_var(t_var))
			  bounds]);
		Expr_op(Op_other("%join_terms"),
		    	Sort.list (<=) (List.map join_term2expr terms));
		Expr_op(Op_other("%join_pred"),[
		    	wff_to_bmf pred]);
		Expr_op(Op_other("%join_assertion"),[
		    	wff_to_bmf assertion])
	      ])),
     body)
    
and destruct_join_stmt s =
  match s with
    Stmt_do
      (Expr_set
	 (_,Expr_op
	    (Op_other("%join"),[
	     Expr_op(Op_other("%join_t_var"),[Expr_var(t_var)]);
	     Expr_op(Op_other("%join_bounds"),[raw_bounds]);
	     Expr_op(Op_other("%join_terms"),raw_terms);
	     Expr_op(Op_other("%join_pred"),[raw_pred]);
	     Expr_op(Op_other("%join_assertion"),[raw_assertion])
	   ])),
       body) ->
	 (t_var,raw_bounds,raw_terms,raw_pred,raw_assertion,body)
  | _ -> raise Not_a_valid_join

and stmt2join s =
    let (t_var,raw_bounds,raw_terms,raw_pred,raw_assertion,body) =
	destruct_join_stmt s
    in try
	{ join_t_var = t_var;
	  join_bounds =
		expr2bounds' (Expr_var(t_var)) raw_bounds
		    Not_a_valid_join;
	  join_terms = List.map expr2join_term raw_terms;
	  join_pred = bmf_to_wff raw_pred;
	  join_assertion = bmf_to_wff raw_assertion;
	  join_body = body }
    with (Not_a_valid_jt|Not_a_wff(_)) ->
	raise Not_a_valid_join

(* Determineds *)

and stmt_is_determined s =
  match s with
    Stmt_do(Expr_set(Frame(_),Expr_op(Op_other("%determined"),_)),_) -> 
      true
  | _ -> false

and determined2stmt
    ({ determined_term=term;
       determined_pred=pred; determined_assertion=assertion;
       determined_body=body} : determined_type) =
  Stmt_do
    (Expr_set
       (Frame(bound_in_term term),
	Expr_op(Op_other("%determined"),[
		Expr_op(Op_other("%determined_term"),[
			join_term2expr term]);
		Expr_op(Op_other("%determined_pred"),[
			wff_to_bmf pred]);
		Expr_op(Op_other("%determined_assertion"),[
			wff_to_bmf assertion])
	      ])),
     body)

and destruct_determined_stmt s =
  match s with
    Stmt_do
      (Expr_set
	 (_,Expr_op(Op_other("%determined"),[
		    Expr_op(Op_other("%determined_term"),[raw_term]);
		    Expr_op(Op_other("%determined_pred"),[raw_pred]);
		    Expr_op(Op_other("%determined_assertion"),[raw_assertion])
	      ])),
       body) ->
	 (raw_term,raw_pred,raw_assertion,body)
  | _ -> raise Not_a_valid_determined

and stmt2determined s =
    let (raw_term,raw_pred,raw_assertion,body) =
	destruct_determined_stmt s
    in try
	{ determined_term = expr2join_term raw_term;
	  determined_pred = bmf_to_wff raw_pred;
	  determined_assertion = bmf_to_wff raw_assertion;
	  determined_body = body }
    with (Not_a_valid_jt|Not_a_wff(_)) ->
	raise Not_a_valid_determined

(* Join terms *)

and expr_is_join_term e =
    match e with
      Expr_op(Op_other("%join_term"),_) -> true
    | _ -> false

and join_term2expr jt =
    Expr_op(Op_other("%join_term"),[
	Expr_var(jt.jt_flag);
	jt_kind2expr jt.jt_kind;
	jt.jt_constr;
	bounds2expr (Expr_var(join_var_of_kind jt.jt_kind))
	    jt.jt_bounds])

and expr2join_term e =
    let (flag,raw_kind,constr,raw_bounds) =
	match e with
	  Expr_op(Op_other("%join_term"),[
		Expr_var(flag);
		raw_kind;
		constr;
		raw_bounds]) ->
	    (flag,raw_kind,constr,raw_bounds)
	| _ -> raise Not_a_valid_jt
    in let kind = expr2jt_kind raw_kind Not_a_valid_jt
    in let bounds =
	expr2bounds' (Expr_var(join_var_of_kind kind))
	    raw_bounds Not_a_valid_jt
    in
	{ jt_flag = flag; jt_kind = kind;
	  jt_constr = constr; jt_bounds = bounds }

(* Bounds *)

and bounds2expr t_val (t_lbs,t_ubs) =
    let t_lbs' = List.map (fun t_lb -> Expr_op(Op_leq,[t_lb;t_val])) t_lbs
    and t_ubs' = List.map (fun t_ub -> Expr_op(Op_leq,[t_val;t_ub])) t_ubs
    in
	make_and (t_lbs'@t_ubs')

and expr2bounds t_val e =
    expr2bounds' t_val e Not_valid_bounds

and expr2bounds' t_val e exn =
    List.fold_left
	(fun (t_lbs,t_ubs) term ->
	    match term with
	      Expr_op(Op_leq,[t_lb;t_val'])
		    when t_val = t_val' ->
		(t_lb::t_lbs,t_ubs)
	    | Expr_op(Op_leq,[t_val';t_ub])
		    when t_val = t_val' ->
		(t_lbs,t_ub::t_ubs)
	    | Expr_bool(true) -> (t_lbs,t_ubs)
	    | _ -> raise exn)
	([],[]) (bmf_expr_args_by_op Op_and e)

(* Join term kinds *)

and jt_kind2expr jtk =
    match jtk with
      Jtk_enum_simple(enum_acc) ->
	Expr_op(Op_other("%jtk_enum_simple"),[
	    jt_equal2expr enum_acc])
    | Jtk_enum_indexing(enum_acc,lookup_acc) ->
	Expr_op(Op_other("%jtk_enum_indexing"),[
	    jt_equal2expr enum_acc;
	    jt_equal2expr lookup_acc])
    | Jtk_enum_searching(enum_acc,lookup_acc,search_acc) ->
	Expr_op(Op_other("%jtk_enum_searching"),[
	    jt_equal2expr enum_acc;
	    jt_equal2expr lookup_acc;
	    jt_equal2expr search_acc])
    | Jtk_lookup(lookup_acc) ->
	Expr_op(Op_other("%jtk_lookup"),[
	    jt_equal2expr lookup_acc])

and expr2jt_kind e exn =
    match e with
      Expr_op(Op_other("%jtk_enum_simple"),[enum_acc]) ->
	Jtk_enum_simple(expr2jt_equal enum_acc exn)
    | Expr_op(Op_other("%jtk_enum_indexing"),[enum_acc;lookup_acc]) ->
	Jtk_enum_indexing(expr2jt_equal enum_acc exn,
	    expr2jt_equal lookup_acc exn)
    | Expr_op(Op_other("%jtk_enum_searching"),[
	    enum_acc;lookup_acc;search_acc]) ->
	Jtk_enum_searching(expr2jt_equal enum_acc exn,
	    expr2jt_equal lookup_acc exn,
	    expr2jt_equal search_acc exn)
    | Expr_op(Op_other("%jtk_lookup"),[lookup_acc]) ->
	Jtk_lookup(expr2jt_equal lookup_acc exn)
    | _ -> raise exn

(* join_term_equality's *)

and jt_equal2expr (Jtk_eq(var,a,name,args)) =
    Expr_op(Op_eq,[
	Expr_var(var);
	Expr_access(Access_strong,Expr_var(a),name,args)])

and expr2jt_equal e exn =
    match e with
      Expr_op(Op_eq,[
	    Expr_var(var);
	    Expr_access(Access_strong,Expr_var(a),name,args)]) ->
	Jtk_eq(var,a,name,args)
    | _ -> raise exn

(* Utilities *)

and rel_of_kind kind =
    match kind with
      Jtk_enum_simple(m1) -> rel_of_eq m1
    | Jtk_enum_indexing(m1,m2) -> rel_of_eq m2
    | Jtk_enum_searching(m1,m2,m3) -> rel_of_eq m2
    | Jtk_lookup(m1) -> rel_of_eq m1
and join_var_of_kind kind =
    match kind with
      Jtk_enum_simple(m1) -> var_of_eq m1
    | Jtk_enum_indexing(m1,m2) -> var_of_eq m2
    | Jtk_enum_searching(m1,m2,m3) -> var_of_eq m2
    | Jtk_lookup(m1) -> var_of_eq m1

and enum_var_of_kind kind =
    match kind with
      Jtk_enum_simple(m1) -> var_of_eq m1
    | Jtk_enum_indexing(m1,m2) -> var_of_eq m1
    | Jtk_enum_searching(m1,m2,m3) -> var_of_eq m1
    | Jtk_lookup(m1) -> var_of_eq m1

and join_expr_of_kind kind =
    match kind with
      Jtk_enum_simple(m1) -> expr_of_eq m1
    | Jtk_enum_indexing(m1,m2) -> expr_of_eq m2
    | Jtk_enum_searching(m1,m2,m3) -> expr_of_eq m2
    | Jtk_lookup(m1) -> expr_of_eq m1

and enum_expr_of_kind kind =
    match kind with
      Jtk_enum_simple(m1) -> expr_of_eq m1
    | Jtk_enum_indexing(m1,m2) -> expr_of_eq m1
    | Jtk_enum_searching(m1,m2,m3) -> expr_of_eq m1
    | Jtk_lookup(m1) -> expr_of_eq m1

and vars_of_kind kind =
    match kind with
      Jtk_enum_simple(m1) -> [var_of_eq m1]
    | Jtk_enum_indexing(m1,m2) -> [var_of_eq m1; var_of_eq m2]
    | Jtk_enum_searching(m1,m2,m3) -> [var_of_eq m1; var_of_eq m2]
    | Jtk_lookup(m1) -> [var_of_eq m1]

and rel_of_eq (Jtk_eq(_,r,_,_)) = r
and var_of_eq (Jtk_eq(v,_,_,_)) = v

and expr_of_eq (Jtk_eq(_,r,s,a)) =
	Expr_access(Access_strong,Expr_var(r),s,a)

and bound_in_term term =
  let field_vars = List.map make_decl (vars_of_kind term.jt_kind)
  and flag_var = make_bool_decl term.jt_flag
  in flag_var::field_vars
;;
