(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Bmf;;

(* The data structure for describe join operations. The operations
   described all manner of joins, including inner and outer, equiv and
   affine. *)

type join_type = {
    join_t_var : string;
	(* The parametric variable upon which the join is being performed.
	   *)
    join_bounds : join_bounds_type;
	(* The range of the parametric variable in the results produces by
	   the join. These bounds have been computed by the scheduler using
	   FME. *)
    join_terms : join_term_type list;
	(* The list of terms that are to participate in the join. *)
    join_pred : Wff.Simple.t;
	(* The predicate that is to be evaluated by the join *)
    join_assertion : Wff.Simple.t;
	(* A predicate that holds, regardless of the value of the
	   join_pred. That is, the assertion predicate describe all
	   possible truth assignments of the hterms that are feasible, with
	   the query assertions as constraints. Can be used to describe
	   properties of the data structures, for instance. *)
    join_body : bmf_stmt
	(* The body of the join. *)
} 

(* The data structure for describing a determined operation. These occurs
   when a join term cannot be scheduled with its corrisponding term,
   perhaps because its fields are not yet ready. *)

and determined_type = {
    determined_term : join_term_type;
	(* The term *)
    determined_pred : Wff.Simple.t;
	(* The predicate *)
    determined_assertion : Wff.Simple.t;
	(* The assertion *)
    determined_body : bmf_stmt
	(* The body *)
}

and join_term_type = {
    jt_flag : string;
	(* The name used to refer to the term. 
	   This name can also be used in the body of the join as a success
	   flag for this term. That is, the jt_name will be a boolean
	   variable, which will be true in the body when the join term
	   successfully bound the enum_var and join_var's. *)
    jt_kind : join_term_kind;
	(* The methods that comprise the term. *)
    jt_constr : bmf_expr;
	(* value of the term's join var, in terms of join_t_var *)
    jt_bounds : join_bounds_type
	(* range of join_var to check. This is the result of projecting
	   join_bounds onto the term's join var *)
}
and join_term_kind =

|   Jtk_enum_simple of join_term_equality
			    (* A[["enum_s",...]]   = enum_var *)

|   Jtk_enum_indexing of join_term_equality
			    (* A[["enum_s",...]]   = enum_var *)
                * join_term_equality
			    (* A[["lookup_s",...]] = join_var *)

|   Jtk_enum_searching of join_term_equality
			    (* A[["enum_s",...]]   = enum_var *)
                * join_term_equality
			    (* A[["lookup_s",...]] = join_var *)
                * join_term_equality
			    (* A[["determined_s",...]] = enum_var *)

|   Jtk_lookup  of join_term_equality
			    (* A[["lookup_s",...]] = join_var *)

and join_term_equality =
    (* A[[name,args]] = var *)
    Jtk_eq of string	    (* var *)
	    * string	    (* A *)
	    * string	    (* name *)
	    * bmf_expr list (* args *)

and join_bounds_type = (bmf_expr list) * (bmf_expr list)
;;

exception Not_a_valid_join;;
exception Not_a_valid_determined;;
exception Not_valid_bounds;;

val stmt_is_join : bmf_stmt -> bool;;
val join2stmt : join_type -> bmf_stmt;;
val stmt2join : bmf_stmt -> join_type;;

val stmt_is_determined : bmf_stmt -> bool;;
val determined2stmt : determined_type -> bmf_stmt;;
val stmt2determined : bmf_stmt -> determined_type;;


val bounds2expr : bmf_expr -> (bmf_expr list * bmf_expr list)
    -> bmf_expr;;
val expr2bounds : bmf_expr -> bmf_expr ->
    (bmf_expr list * bmf_expr list);;

val rel_of_kind : join_term_kind -> string
val join_var_of_kind : join_term_kind -> string;;
val enum_var_of_kind : join_term_kind -> string;;
val join_expr_of_kind : join_term_kind -> bmf_expr;;
val enum_expr_of_kind : join_term_kind -> bmf_expr;;
