(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
(***********************************************************************
 *                             Affine forms
 ***********************************************************************)

open Ssisl
open Containers

open Alpha;;
open Bmf;;

type
  affine_form =
    Form                 of (int String_map.t) * int
  | Non_linear

and
  affine_term = 
    Term                 of string * int

and
  affine_map =
    Affine_map           of matrix * vector
;;

exception Non_linear_form of unit;;
exception Non_conjunct of bmf_expr;;
exception Bad_affine_map of ((bmf_ident list) * affine_map );;
exception Bad_affine_rel of ((bmf_ident list) * affine_map * (bmf_ident list) * affine_map);;
exception Zero_term of unit;;
exception Invalid_expr_set of bmf_expr;;
exception Equalities_not_allowed of bmf_expr;;
exception Incompatible_sets;;

(***********************************************************************
 *  Affine forms
 ***********************************************************************)

let affine_map_rows (Affine_map(a,b)) = 
    matrix_rows a
;;

let affine_map_cols (Affine_map(a,b)) = 
    matrix_cols a
;;

let new_affine_map m n =
    Affine_map (create_matrix m n, make_vector m)
;;

let compose_affine_map (Affine_map(a,b)) (Affine_map(c,d)) =
    (* a*(c*x + d) + b --> (a*c)*x + (a*d+b) *)
    Affine_map(mmm a c, saxpy 1 (mvm a d) 1 b)
;;

let is_zero_affine_map (Affine_map(a,b)) =
    ((is_zero_matrix a) & (is_zero_vector b))
;;

let is_ident_affine_map (Affine_map(a,b)) =
    ((is_ident_matrix a) & (is_zero_vector b))
;;

let rec
  new_affine_form =
    function  
      (Builder_const(c)) -> Form (String_map.empty, c)
    | (Builder_var(c,id)) ->
        (let f = new_affine_form (Builder_const(0))
      in
        match f with 
          Form(m,cc) -> Form( (String_map.add id c m), cc )
        | Non_linear -> raise (Non_linear_form ()))
;;

let get_const4affine_form a =
  begin
    match a with
      (Form(m,cc)) -> cc
    | Non_linear -> raise (Non_linear_form ())
  end
;;

let
  get_term_list =
    function 
      (Form(m,_)) ->
        (let rec
          result = ref []
        and
          use_term = fun (s : string) (x : int) -> 
                      result := Term(s, x)::!result
        in
          ( String_map.iter use_term m ; !result ))
    | Non_linear -> []
;;

let rec
  collect_vars =
    function 
      (Form(m,_) as f1) ->
        (let rec
          result = ref []
        and
          use_term = fun (s : string) (x : int) -> 
                      result := s::!result
        in
          ( String_map.iter use_term m ; ! result ))
    | Non_linear -> []

and
  affine_add =
    fun
      p0 p1 -> match (p0,p1) with ((Form(m1,c1) as f1), (Form(m2,c2) as f2)) ->
        affine_add_by_list c1 (get_term_list f1) f2
    | (_, _) -> Non_linear

and
  affine_saxpy =
    fun
      p0 p1 p2 p3 -> match (p0,p1,p2,p3) with ((a1 : int), (Form(m1,c1) as f1), (a2 : int), (Form(m2,c2) as f2)) ->
        affine_add (affine_scale a1 f1) (affine_scale a2 f2)
    | (_, _, _, _) -> Non_linear

and
  affine_scale =
    fun 
      p0 p1 -> match (p0,p1) with ((a:int), (Form(m,c) as f)) ->
        affine_scale_by_list a c (get_term_list f)
    | (_, _) -> Non_linear

and
  affine_mult =
    fun
      p0 p1 -> match (p0,p1) with ((Form(m1,c1) as f1), (Form(m2,c2) as f2)) ->
        if ((String_map.size m1) = 0)
        then affine_scale c1 f2
        else 
          if ((String_map.size m2) = 0)
          then affine_scale c2 f1
          else Non_linear
    | (_, _) -> Non_linear

and
  affine_coeff4var =
    fun 
      p0 p1 -> match (p0,p1) with (s, (Form(m,_))) ->
        (let xx = find_value (fun () -> String_map.find s m)
        in
          match xx with
            Found x -> x
          | Unfound -> 0)
    | (s, Non_linear) -> raise (Non_linear_form ())

and
  affine_scale_by_list =
    fun
      p0 p1 p2 -> match (p0,p1,p2) with (a, c, (Term(s,x)::rest)) ->
        (let f = affine_scale_by_list a c rest
        in
          match f with
            (Form(m,cc)) -> Form((String_map.add s (a*x) (String_map.remove s m)), cc)
          | Non_linear -> raise (Non_linear_form ()))
    | (a, c, []) -> new_affine_form (Builder_const (a*c))

and
  affine_add_by_list =
    fun
      p0 p1 p2 -> match (p0,p1,p2) with (c1, (Term(s,x)::rest), f2) ->
        (let f2_1 = (affine_add_by_list c1 rest f2) in
          let x2 = (x + (affine_coeff4var s f2_1))
          in
            match f2_1 with
              (Form(m2,c2)) -> 
                if (x2 = 0) 
                then Form((String_map.remove s m2), c2)
                else Form((String_map.add s x2 (String_map.remove s m2)), c2)
            | Non_linear -> Non_linear)

    | (c1, [], f2) ->
      (match f2 with
        (Form(m,c)) -> Form(m, c + c1)
       | Non_linear -> Non_linear)

and
  expr2affine =
    function
      (Expr_var(s) as e) -> new_affine_form (Builder_var(1,s))
    | (Expr_int(i)) -> new_affine_form (Builder_const(i))
    | (Expr_op(Op_plus,l)) ->
        (let la = List.map expr2affine l
        in List.fold_left affine_add (new_affine_form (Builder_const 0)) la)
    | (Expr_op(Op_times,l)) ->
        (let la = List.map expr2affine  l
        in List.fold_left affine_mult (new_affine_form (Builder_const 1)) la)
    | (Expr_op(Op_neg,_) as e) -> 
        affine_scale (-1) (expr2affine (bmf_expr_op_arg e))
    | (Expr_op(Op_div,l)) -> Non_linear
    | (_ as e) -> raise (Bmf_expr_type_mismatch (e, Type_scalar Type_int))

and
  affine2expr =
    function
      (Form(m,c) as fm) ->
        let f = fun (Term(s,i)) -> 
                  if (i = 0) then raise (Zero_term ())
                  else 
                    if (i = 1) then Expr_var(s)
                    else 
                      if (i = -1) then Expr_op(Op_neg,[Expr_var(s)])
                      else Expr_op(Op_times,[Expr_var(s) ; Expr_int(i)])
        in (
          let l = 
            if (c <> 0) then ((Expr_int(c)) :: (List.map f (get_term_list fm)))
                        else (List.map f (get_term_list fm))
          in
            if ((List.length l) = 0) then Expr_int(0)
            else 
              if ((List.length l) = 1) then (List.hd l)
              else Expr_op(Op_plus,l))

    | Non_linear -> raise (Non_linear_form ())
and
  is_linear_form =
    function
      (Form(_,c)) -> (c = 0)
    | Non_linear -> false
and
  term_list2vector =
    fun
      names lst (v : (int array ref)) ->
        try
          (List.iter
            (fun (Term(s,c)) -> (!v).(List'.position s names) <- c)
            lst)
        with
          Not_found -> failwith
                        "term_list2vector: Variable not in names"
and           
  (* WARNING: this only collects LINEAR part *)
  (* raises non_linear_form exception if not linear *)
  affine2vector =
    fun
      p0 p1 -> match (p0,p1) with (names, (Form(m,c) as f)) ->
        let n = (List.length names)
        in let v = Array.create n 0
          in (
            (term_list2vector names (get_term_list f) (ref v)) ;
            v)
    | (names, Non_linear) -> raise (Non_linear_form ())

and affine_list2affine_map names l =
    let m = List.length l
    and n = List.length names
    in
      if (m = 0) then
         Affine_map(create_matrix 0 n, make_vector 0)
      else
         let get_vec = fun a -> affine2vector names a
         and get_const = 
              (function (Form(_,c)) -> c 
                   | Non_linear -> raise (Non_linear_form ()))
         in Affine_map (
              caml_to_alpha_matrix (Array.of_list (List.map get_vec l)),
              caml_to_alpha_vector (Array.of_list (List.map get_const l)))

and verify_affine_map names (Affine_map(a,b) as m) =
    if (matrix_rows a <> vector_length b) then 
       failwith "Bad affine map"
    else 
       if (matrix_cols a > (List.length names)) then 
          failwith "Bad affine map"
       else m

and affine_map2affine_list names (Affine_map(_,_) as mp) =
    let (Affine_map(a,b)) = (verify_affine_map names mp)
    in let m = matrix_rows a
    and n = matrix_cols a
    and l = ref [] 
    and names_vec = Array.of_list names
    in (
      for i = 1 to m do
        let f = ref (new_affine_form (Builder_const (vector_ref b i)))
        in (
          for j = 1 to n do
            if (matrix_ref a i j <> 0) then
              f := (affine_add !f 
                      (new_affine_form
                        (Builder_var 
                          (matrix_ref a i j, names_vec.(j-1)))))
            else ()
          done ;
          l := !l @ [(!f)]
        )
      done ;
      (!l)
    )

and expr_list2affine_map names (l : (bmf_expr list)) =
    let f e = expr2affine e
    in affine_list2affine_map names (List.map f l)

and
  affine_relation2expr_list =
    fun
      (x1:(string list)) (a1:affine_map) (x2:(string list)) (a2:affine_map) ->
	let l1 = (affine_map2affine_list x1 a1)
        and l2 = (affine_map2affine_list x2 a2)
        in
	  if ((List.length l1) <> (List.length l2)) then
	    failwith "Bad affine relation"
          else
	    let combine_form = fun form1 form2 -> 
				  Expr_op(
                                    Op_eq,
                                    [ (affine2expr form1) 
                                    ; (affine2expr form2) 
                                    ])
            in
	      (List.map2 combine_form l1 l2)
;;

let
  affine_map2string =
    fun
      (Affine_map(a,b)) -> (matrix2string a) ^ " " ^ (vector2string b)
;;

let
  create_identity_affine_map =
    fun
      n -> Affine_map((create_identity n), (make_vector n))
;;

let extract_rows_affine_map (Affine_map(a,b)) i1 i2 =
    let n = matrix_cols a
    in
      Affine_map((extract_matrix a i1 i2 1 n),
		 (extract_vector b i1 i2))
;;

let rec extend_affine_maps a1 a2 =
  begin
    let n1 = (affine_map_cols a1)
    and n2 = (affine_map_cols a2)
    in
      if (n1 = n2) then (a1,a2)
      else
	if (n1 > n2) then
	  let (b2,b1) = (extend_affine_maps a2 a1)
          in (b1,b2)
        else
	  match a1 with
	    (Affine_map(a,b)) ->
	      (Affine_map((add_cols a (n2-n1)),b), a2)
  end  
;;

let stack_affine_maps (Affine_map(a1,b1)) (Affine_map(a2,b2)) =
  begin
    Affine_map( (stack_matrices a1 a2), (concat_vectors b1 b2) )
  end
;;

let compare_affine_maps (Affine_map(a1,b1)) (Affine_map(a2,b2)) =
    compare_pair compare_matrices compare_vectors
                 (a1,b1) (a2,b2)
;;

let equal_affine_map a1 a2 =
    (0 = (compare_affine_maps a1 a2))
;;

let destructure_linear_map (Affine_map(a,b)) =
    assert (is_zero_vector b) 
           (* (Failure("destructure_linear_map: B != 0")) *);
    a
;;

let destructure_affine_map (Affine_map(a,b)) =
    (destructure_matrix a, destructure_vector b)
;;

let make_linear_map a =
    (Affine_map(a, make_vector (matrix_rows a)))
;;
    
  
(************************************************************************)
(* Interface to Fourier-Motzkin						*)
(************************************************************************)

let expr_set2builder_set (orig_e: bmf_expr) = (
    
 let rec walk_set e =
	    match e with
	      Expr_set(Frame(frm),e') ->
		check_all_int_bindings frm;
		let (exists,constrs) = walk_exist e'
		in
		    Builder_set(List.map bmf_ident_name frm, exists, constrs)
	    | Expr_bool(false) -> Builder_no_solution
	    | _ -> raise (Invalid_expr_set orig_e)
	and walk_exist e =
	    match e with
	      Expr_exist(Frame(frm),e') ->
		check_all_int_bindings frm;
		(List.map bmf_ident_name frm,
		 walk_constraints e')
	    | _ -> ([],walk_constraints e)
	and check_all_int_bindings ids =
	    List.iter
		(fun id ->
		    match id with
		      Ident_valid(_,Type_scalar(Type_int),_,_) -> ()
		    | _ -> raise (Invalid_expr_set orig_e))
		ids
	and walk_constraints e =
	    List'.flat_map walk_constraint
		(bmf_expr_args_by_op Op_and e)
	and walk_constraint e =
	    match e with
	      Expr_op(Op_not,[e']) ->
		let (Builder_constraint(kind,left,right)) =
		    walk_simple_constraint e'
		in
		    (match kind with
		      Builder_eq ->
			[Builder_constraint(Builder_lt,left,right);
			 Builder_constraint(Builder_gt,left,right)]
		    | Builder_le ->
			[Builder_constraint(Builder_gt,left,right)]
		    | _ -> failwith "Shouldn't have happened.")
	    | Expr_bool(true) -> []
	    | Expr_bool(false) -> raise (Invalid_expr_set orig_e)
	    | _ -> [walk_simple_constraint e]
	and walk_simple_constraint e =
	    match e with
	      Expr_op(Op_eq,[e1;e2]) ->
		Builder_constraint(Builder_eq,
		    walk_terms e1, walk_terms e2)
	    | Expr_op(Op_leq,[e1;e2]) ->
		Builder_constraint(Builder_le,
		    walk_terms e1, walk_terms e2)
	    | _ -> raise (Invalid_expr_set orig_e)
	and walk_terms e =
	    match e with
	      Expr_op(Op_plus,args) -> List'.flat_map walk_terms args
	    | Expr_op(Op_neg,[e]) ->
		List.map (fun t ->
			match t with
			  Builder_const(c) ->
			    Builder_const(-1*c)
			| Builder_var(c,v) ->
			    Builder_var(-1*c, v))
		    (walk_terms e)
	    | _ -> [walk_term e]
	and walk_term e =
	    match e with
	      Expr_int(c) -> Builder_const(c)
	    | Expr_var(v) -> Builder_var(1,v)
	    | Expr_op(Op_times,[Expr_int(c);Expr_var(v)]) ->
		Builder_var(c,v)
	    | Expr_op(Op_times,[Expr_var(v);Expr_int(c)]) ->
		Builder_var(c,v)
	    | _ -> raise (Invalid_expr_set orig_e)
 in walk_set orig_e
	
: affine_set_builder);;

let builder_set2expr_set (orig_s: affine_set_builder) = (
    
 let rec walk_set s =
	    match s with
	      Builder_set(set_vars,exist_vars,constrs) ->
		let set_ids = List.map make_ident set_vars
		and exist_ids = List.map make_ident exist_vars
		and body = walk_constrs constrs
		in
		    if exist_ids = [] then
			Expr_set(Frame(set_ids),body)
		    else
			Expr_set(Frame(set_ids),
			    Expr_exist(Frame(exist_ids), body))
	    | Builder_no_solution -> Expr_bool(false)
	and make_ident name =
	    Ident_valid(name,Type_scalar(Type_int),Ident_bind,An_list[])
	and walk_constrs constrs =
	    match constrs with
	      [] -> Expr_bool(true)
	    | [x] -> walk_constr x
	    | _ -> Expr_op(Op_and,(List.map walk_constr constrs))
	and walk_constr (Builder_constraint(kind,left,right)) =
	    let left' = walk_terms left
	    and right' = walk_terms right
	    in
    		match kind with
		  Builder_eq -> Expr_op(Op_eq,[left';right'])
		| Builder_le -> Expr_op(Op_leq,[left';right'])
		| Builder_ge -> Expr_op(Op_leq,[right';left'])
		| Builder_lt ->
		    Expr_op(Op_not,[Expr_op(Op_leq,[right';left'])])
		| Builder_gt ->
		    Expr_op(Op_not,[Expr_op(Op_leq,[left';right'])])
	and walk_terms terms =
	    match terms with
	      [] -> failwith "Empty term list?"
	    | [x] -> walk_term x
	    | _ -> Expr_op(Op_plus,(List.map walk_term terms))
	and walk_term t =
	    match t with
	      Builder_const(c) -> Expr_int(c)
	    | Builder_var(1,v) -> Expr_var(v)
	    | Builder_var(c,v) ->
		Expr_op(Op_times,[Expr_int(c);Expr_var(v)])
 in walk_set orig_s
	
: bmf_expr);;

let intersect_sets s1 s2 =
    let (s1_bound,s1_existential,s1_pred) =
	match s1 with
	  Expr_set(Frame(s1_bound),
		Expr_exist(Frame(s1_existential),s1_pred)) ->
	    (s1_bound,s1_existential,s1_pred)
	| Expr_set(Frame(s1_bound),s1_pred) ->
	    (s1_bound,[],s1_pred)
	| _ -> raise (Invalid_expr_set s1)
    and (s2_bound,s2_existential,s2_pred) =
	match s2 with
	  Expr_set(Frame(s2_bound),
		Expr_exist(Frame(s2_existential),s2_pred)) ->
	    (s2_bound,s2_existential,s2_pred)
	| Expr_set(Frame(s2_bound),s2_pred) ->
	    (s2_bound,[],s2_pred)
	| _ -> raise (Invalid_expr_set s2)
    in let bound = 
	zassert (List.map bmf_ident_name s1_bound =
		    List.map bmf_ident_name s2_bound)
	    Incompatible_sets;
	s1_bound
    and existential = List'.union s1_existential s2_existential
    and pred =
	make_and ((bmf_expr_args_by_op Op_and s1_pred)@
		  (bmf_expr_args_by_op Op_and s2_pred))
    in
	if existential = [] then
	    Expr_set(Frame(bound),pred)
	else
	    Expr_set(Frame(bound),
		Expr_exist(Frame(existential),pred))
;;

let simplify_set s =
    builder_set2expr_set (simplify (expr_set2builder_set s))
;;

let gist_sets s1 s2 =
    builder_set2expr_set
	(gist (expr_set2builder_set s1)
	    (expr_set2builder_set s2))
;;

let rec fm_code_gen orig_e =
    let simplified_e = simplify_set orig_e
    (* Destruct the set expr. Collect all of the inequalities *)
    in let (vars,constrs,orig_constrs) =
        match simplified_e with
	  Expr_set(Frame(frm),pred) ->
	    let vars = List.map bmf_ident_name frm
	    and constrs = 
		List'.flat_map
		    (fun conj ->
			match conj with
			  Expr_bool(true) -> []
			| Expr_op(Op_leq,[e1;e2]) -> [(e1,e2)]
			| Expr_op(Op_eq,[e1;e2]) -> [(e1,e2);(e2,e1)]
			| _ -> raise (Invalid_expr_set orig_e))
		    (bmf_expr_args_by_op Op_and pred)
	    in
		(vars,constrs,pred)
	| _ -> 
	    (* *todo*. The system is inconsistent. What should be done? *)
	    assert false 
    in let do_one_step var results_for_vars' =
	(* Constraints of the form e1 <= e2 ->
	    affines of the form e1-e2 <= 0 *)
	let affines =
	    List.map (fun (e1,e2) ->
		    affine_add (expr2affine e1)
			(affine_scale (-1) (expr2affine e2)))
		constrs
	(* Partition the constrs in
	    pos_var <- coeff of `var' is > 0.
	    neg_var <- coeff of `var' is < 0. *)
	in let (pos_var,neg_var) =
	    List.fold_left
		(fun (pos_var,neg_var) aff ->
		    let coef = affine_coeff4var var aff
		    in
			if coef > 0 then
			    (aff::pos_var,neg_var)
			else if coef < 0 then
			    (pos_var,aff::neg_var)
			else
			    (pos_var,neg_var))
		([],[]) affines
	and remove_var c aff =
	    let coef = affine_coeff4var var aff
	    in
		assert (coef = c) (* (Failure("Wrong coef?!?")) *);
		affine2expr
		    (affine_scale (-1*c)
			(affine_add aff
			    (new_affine_form
				(Builder_var(-1*c,var)))))
	in let lbs = List.map (remove_var (-1)) neg_var
	and ubs = List.map (remove_var 1) pos_var
	in let rec remove_dups lst =
	    match lst with
	      [] -> []
	    | x::lst' ->
		let lst'' = remove_dups lst'
		in
		    if List.mem x lst' then
			lst''
		    else
			x::lst''
	in
	    results_for_vars'
		@[(remove_dups lbs,remove_dups ubs)]
    in
	match List.rev vars with
	  [] -> failwith "Can't code gen with no vars"
	| [var] -> do_one_step var []
	| var::srav' ->
	    let vars' = List.rev srav'
	    and make_ident name =
		Ident_valid(name,Type_scalar(Type_int),Ident_bind,An_list[])
	    in let results_for_vars' =
		fm_code_gen
		    (Expr_set(Frame(List.map make_ident vars'),
			Expr_exist(Frame[make_ident var],
			    orig_constrs)))
	    in
		do_one_step var results_for_vars'
;;

(*
let ax = new_affine_form (builder_var (1, "x"));;
let ay = new_affine_form (builder_var (1, "y"));;
let a1 = (affine_add (affine_add ax ay) (new_affine_form (builder_const 3)));;
let a2 = (affine_add (affine_add ax ay) (new_affine_form (builder_const (-7))));;
*)

