/************************************************************************
 *	Copyright (C) 1998. Bernoulli Project, Cornell University.	*
 ************************************************************************/

%{
open Alpha
open Bmf
open Bmf_linear

let rec order_ids_by_list ids lst =
    List.map (find_id ids) lst
and find_id ids str =
    match ids with
        (Ident_valid(str',_,_,_) as id)::ids' when str = str' -> id
    |	x::ids' -> find_id ids' str
    |	_ -> failwith "Formals and param decls don't match"
;;


let rec ident_list_to_type_list lst =
    List.map ident_to_type lst
and ident_to_type id =
    match id with
        Ident_valid(id,ty,_,_) -> ty
    |	_ -> failwith "What is id?"
;;

let clean_up_stmt_seq s =
    let rec find_stmts s =
	match s with
	  Stmt_seq(x,y) -> (find_stmts x) @ (find_stmts y)
	| Stmt_nop -> []
	| _ -> [s]
    and build_seq lst =
	match lst with
	  [] -> Stmt_nop
	| [s] -> s
	| s::lst' -> Stmt_seq(s,build_seq lst')
    in
	build_seq (find_stmts s)
;;

let lower_bound_is_one_p x =
    match x with
      Expr_int(0) -> false
    | Expr_int(1) -> true
    | _ -> failwith "Lb must be 0 or 1."
;;

%}

%token VOID
%token BOOL
%token REAL
%token INT
%token DOTDOT
%token DOT
%token ARRAY
%token PTR
%token OF
%token ON
%token COMMA
%token PARAMBEGIN
%token PARAMEND

%token PLUS
%token MINUS
%token STAR
%token DIVIDE
%token AND
%token OR
%token NOT
%token IMPLIES
%token FORALL
%token EXIST
%token TRUE0
%token FALSE0
%token POWER
%token SQRT0
%token ADDR

%token OPEN_PAREN
%token CLOSE_PAREN
%token OPEN_CURLY
%token CLOSE_CURLY

%token OPEN_SQUARE
%token CLOSE_SQUARE

%token OPEN_ANN
%token CLOSE_ANN

%token COLON
%token BAR
%token SEMI_COLON
%token ASSIGN

%token LEFT_ARROW

%token EQUALS
%token LE
%token LTHAN
%token GT
%token GE
%token NEQ

%token WHILE0
%token IF0
%token ELSE
%token FOR0
%token DO0
%token CALL0
%token OP0

%token EXTERN
%token PROCEDURE
%token SUCH
%token THAT
%token VAR

%token RELATION
%token TYPE_ANN

%token<string> STRINGVAL
%token<int> DECIMALINT
%token<float> FLOAT
%token<string> IDENT

%token EOF

%token<Bmf.bmf_type> IN_PLACE_TYPE
%token<Bmf.bmf_stmt> IN_PLACE_STMT
%token<Bmf.bmf_expr> IN_PLACE_EXPR
%token<string> IN_PLACE_STRING
%token<Bmf.bmf_stmt list> IN_PLACE_STMT_LIST
%token<Bmf.bmf_expr list> IN_PLACE_EXPR_LIST
%token<string list> IN_PLACE_STRING_LIST
%token IN_PLACE_EMPTY_LIST

%start int_matrix
%start int_vector
%start int_affine_map
%type <Alpha.matrix> int_matrix
%type <Alpha.vector> int_vector
%type <Bmf_linear.affine_map> int_affine_map

%start file
%type <Bmf.bmf_ident list> file

%start statement_entry
%type <Bmf.bmf_stmt> statement_entry

%start expression_entry
%type <Bmf.bmf_expr> expression_entry

%start vars_entry
%type <Bmf.bmf_ident list> vars_entry

%start annotation_entry
%type <Bmf.annotation> annotation_entry

%start ident_entry
%type <Bmf.bmf_ident> ident_entry

/* Prec 0 */
%nonassoc IMPLIES
/* Prec 1 */
%left OR
/* Prec 2 */
%left AND
/* Prec 3 */
%nonassoc EQUALS LE LTHAN GT GE NEQ
/* Prec 4 */
%left PLUS MINUS
/* Prec 5 */
%left STAR DIVIDE

%%

/************************************************************************/
/* Entry points								*/
/************************************************************************/

file : file_decls EOF { $1 };
statement_entry : statement EOF { $1 };
expression_entry : expression EOF { $1 };
annotation_entry : annotation EOF { $1 };
ident_entry : global_definition EOF { $1 }
vars_entry : zero_or_more_variable_declarations EOF { $1 };

/************************************************************************/
/* Declarations and definitions						*/
/************************************************************************/

file_decls
    : file_decls external_declaration { $1 @ [$2] }
    | file_decls global_definition { $1 @ [$2] }
    | { [] }
    ;

external_declaration
    : EXTERN procedure_declaration { $2 }
    ;

global_definition
    : variable_declaration { $1 }
    | procedure_definition { $1 }
    ;

variable_declaration
    : VAR ident COLON arbitrary_type optional_variable_init
	    optional_annotation_list SEMI_COLON
	{ Ident_valid($2, $4, Ident_var($5), $6) }
    ;

optional_variable_init
    : /* empty */ { Expr_void }
    | ASSIGN expression { $2 }
    ;

procedure_declaration
    : PROCEDURE ident optional_ident_list_in_parens
      procedure_params_declaration
	{ 
	    let frm = order_ids_by_list $4 $3
	    in
		Ident_valid($2,
		    Type_proc((ident_list_to_type_list frm),
			Type_scalar(Type_void)),
		    Ident_proc(Frame(frm), Ident_null, Stmt_nop),
		    An_list([]))
	}
    ;

optional_ident_list_in_parens
    : OPEN_PAREN ident_list CLOSE_PAREN { $2 }
    | /* empty */ { [] }
    ;

procedure_params_declaration
    : PARAMBEGIN zero_or_more_variable_declarations PARAMEND
	{ $2 }
    ;

procedure_definition
    : procedure_declaration block_statement
	{
	    match $1 with
		Ident_valid(name, ty,
		  Ident_proc(frm, result, Stmt_nop), ann) ->
		    Ident_valid(name, ty,
		      Ident_proc(frm, result, $2), ann)
	    | _ -> failwith "Didn't match expected proc ident?"
	}
    ;

/************************************************************************/
/* Types								*/
/************************************************************************/

arbitrary_type
    : scalar_type { Type_scalar($1) }
    | array_type { Type_array($1) }
    | PTR OF arbitrary_type { Type_ptr($3) }
    | TYPE_ANN annotation_list { Type_ann($2) }
    | IN_PLACE_TYPE { $1 }
    ;

scalar_type
    : BOOL { Type_bool }
    | REAL { Type_real }
    | INT { Type_int }
    | VOID { Type_void }
    ;

array_type
    /*
     * BB arrays.
     */
    : ARRAY INT POWER DECIMALINT OF scalar_type
	    OPEN_PAREN ident keyword_list CLOSE_PAREN
	{ Array_bb($4, $6, $8, $9) }
    /*
     * Dense arrays: There are two syntaxes for dense array types.
     */
    /* "array[...] of ..." -- Row major dense array type */
    | ARRAY OPEN_SQUARE dense_bounds CLOSE_SQUARE OF scalar_type
	{
	    let (lb,ubs) = $3
	    in Array_dense(lb,ubs,Dense_row_major,$6)
	}
    /* "array(...) of ..." -- Column major dense array type */
    | ARRAY OPEN_PAREN dense_bounds CLOSE_PAREN OF scalar_type
	{
	    let (lb,ubs) = $3
	    in Array_dense(lb,ubs,Dense_col_major,$6)
	}
    /*
     * View arrays.
     */
    | ARRAY OPEN_CURLY bracketted_frame LEFT_ARROW
       		bracketted_frame COLON expression CLOSE_CURLY
		OF array_type
	{
            Array_view($3,$5,$7,$10)
        }	
    ;

keyword_list
    : keyword_list ident COLON annotation { $1 @ [($2,$4)] }
    |  /*empty*/ { [] }


/* Bounds are of the form "lb:ub", where lb is either 0 or 1.
    Within a list of dense_bounds, the lbs must be all 0 or all 1. */
dense_bounds
    : dense_bound
	{ let (lb,ub) = $1 in (lb, [ub]) }
    | dense_bounds COMMA dense_bound
	{
	    let (lb,ubs) = $1
	    and (lb',ub') = $3
	    in
		if lb <> lb' then begin
		    failwith "Lbs must be consistent"
		end;
		(lb,ubs@[ub'])
	}
    ;

dense_bound : expression COLON expression
    {
	match $1 with
	  Expr_int(0) -> (Dense_lb_0,$3)
	| Expr_int(1) -> (Dense_lb_1,$3)
	| _ -> failwith "Lb must be 0 or 1."
    } ;

/************************************************************************/
/* Expressions								*/
/************************************************************************/

set_expression
    : OPEN_CURLY set_frame BAR expression CLOSE_CURLY 
	{ Expr_set($2, $4) }
    ;

set_frame
    : bracketted_frame { $1 }
    ;

expression
    : expression IMPLIES expression { Expr_op(Op_implies,[$1;$3]) }
    | expression OR expression { Expr_op(Op_or,[$1;$3]) }
    | expression AND expression { Expr_op(Op_and,[$1;$3]) }
    | expression EQUALS expression { Expr_op(Op_eq,[$1;$3]) }
    | expression LE expression { Expr_op(Op_leq,[$1;$3]) }
    | expression LTHAN expression { Expr_op(Op_not,[Expr_op(Op_leq,[$3;$1])]) }
    | expression GT expression { Expr_op(Op_not,[Expr_op(Op_leq,[$1;$3])]) }
    | expression GE expression { Expr_op(Op_leq,[$3;$1]) }
    | expression NEQ expression { Expr_op(Op_not, [Expr_op(Op_eq,[$1;$3])]) }
    | expression PLUS expression { Expr_op(Op_plus,[$1;$3]) }
    | expression MINUS expression { Expr_op(Op_plus,[$1;Expr_op(Op_neg,[$3])]) }
    | expression STAR expression { Expr_op(Op_times,[$1;$3]) }
    | expression DIVIDE expression { Expr_op(Op_div,[$1;$3]) }
    | unary_expression { $1 }

    ;

relation_lookup
    : ident OPEN_PAREN expression_list CLOSE_PAREN
	{
	    let (name,args) =
		match $3 with
		  Expr_string(name)::args -> (name, args)
		| args -> ("", args)
	    in
		Expr_access(Access_lookup, Expr_var($1), name, args)
	}
    ;

relation_access
    : ident OPEN_SQUARE expression_list CLOSE_SQUARE
	{
	    let (name,args) =
		match $3 with
		  Expr_string(name)::args -> (name, args)
		| args -> ("", args)
	    in
		Expr_access(Access_weak, Expr_var($1), name, args)
	}
    ;

relation_strong_access
    : ident OPEN_SQUARE OPEN_SQUARE 
      expression_list CLOSE_SQUARE CLOSE_SQUARE
	{
	    let (name,args) =
		match $4 with
		  Expr_string(name)::args -> (name, args)
		| args -> ("", args)
	    in
		Expr_access(Access_strong, Expr_var($1), name, args)
	}
    ;

quantified_expression
    : OPEN_PAREN quantifier bracketted_frame BAR expression
      CLOSE_PAREN
	{
	    if ($2) then
		Expr_forall($3, $5)
	    else
		Expr_exist($3, $5)
        }
    ;

quantifier
    : FORALL { true }
    | EXIST { false }
    ;

bracketted_frame
    : OPEN_SQUARE frame_ident_list CLOSE_SQUARE	{ Frame($2) }
    | OPEN_SQUARE CLOSE_SQUARE	{ Frame([]) }
    ;

frame_ident_list
    : frame_ident_list COMMA frame_ident { $1 @ $3 }
    | frame_ident { $1 }
    ;

frame_ident
    : ident 
	{
	    [Ident_valid($1,Type_scalar(Type_int),Ident_bind, An_list([]))]
	}
    | ident COLON scalar_type
	{
	    [Ident_valid($1,Type_scalar($3),Ident_bind, An_list([]))]
	}
    | VAR ident COLON scalar_type
	{
	    [Ident_valid($2,Type_scalar($4),Ident_var(Expr_void),
		An_list([]))]
	}
    | IN_PLACE_STRING_LIST
	{
	    List.map (fun str -> 
		    Ident_valid(str, Type_scalar(Type_int),
			Ident_bind, An_list([])))
		$1
	}
    | IN_PLACE_EMPTY_LIST { [] }
    ;

unary_expression
    : NOT unary_expression { Expr_op(Op_not,[$2]) }
    | MINUS unary_expression { Expr_op(Op_neg,[$2]) }
    | ADDR unary_expression { Expr_op(Op_addr,[$2]) }
    | STAR unary_expression { Expr_op(Op_deref,[$2]) }
    | primary_expression { $1 }
    ;

primary_expression
    : OPEN_PAREN expression CLOSE_PAREN { $2 }
    | quantified_expression { $1 }
    | relation_lookup { $1 }
    | relation_access { $1 }
    | relation_strong_access { $1 }
    | ident { Expr_var($1) }
    | constant_expression { $1 }
    | set_expression { $1 }
    | prefix_intrinsic_expr { $1 }
    | IN_PLACE_EXPR { $1 }
    ;

prefix_intrinsic_expr
    : SQRT0 OPEN_PAREN expression_list CLOSE_PAREN
	{ Expr_op(Op_sqrt,$3) }
    | OP0 OPEN_PAREN ident CLOSE_PAREN
	    OPEN_PAREN expression_list CLOSE_PAREN
	{ Expr_op(Op_other($3),$6) }
    ;

constant_expression
    : TRUE0 { Expr_bool(true) }
    | FALSE0 { Expr_bool(false) }
    | DECIMALINT { Expr_int($1) }
    | FLOAT { Expr_real($1) }
    | STRINGVAL { Expr_string($1) }
    ;

/************************************************************************/
/* Statements								*/
/************************************************************************/

statement
    : conditional_statement { $1 }
    | assignment_statement { $1 }
    | block_statement { $1 }
    | while_loop_statement { $1 }
    | do_loop_statement { $1 }
    | for_loop_statement { $1 }
    | call_statement { $1 }
    | directive_statement { $1 }
    | annotation_statement { $1 }
    ;

block_statement
    : OPEN_CURLY
      zero_or_more_variable_declarations
      zero_or_more_statements
      CLOSE_CURLY
	{
	    let body = clean_up_stmt_seq($3)
	    in
		if ( $2 = [] ) then
		    body
		else
		    Stmt_let(Frame($2), body)
	}

zero_or_more_variable_declarations
    : zero_or_more_variable_declarations variable_declaration
	{ $1 @ [ $2 ] }
    | { [] }
    ;

zero_or_more_statements
    : statement zero_or_more_statements { Stmt_seq($1, $2) }
    | { Stmt_nop }
    ;

annotation_statement
    : foo_annotation_stmt { $1 }
    | foo_annotation_stmt SEMI_COLON { $1 }

foo_annotation_stmt
    : IN_PLACE_STMT { $1 }
    | IN_PLACE_STMT_LIST {
	let rec recurs lst =
	    match lst with
	      [] -> Stmt_nop
	    | [s] -> s
	    | s::lst' -> Stmt_seq(s,recurs lst')
	in
	    recurs $1
      }
    | IN_PLACE_EMPTY_LIST { Stmt_nop }
    ;

assignment_statement
    : expression ASSIGN expression SEMI_COLON { Stmt_assign($1, $3) }
    ;

while_loop_statement
    : WHILE0 OPEN_PAREN expression CLOSE_PAREN block_statement
	{ Stmt_while($3, $5) }
    ;

conditional_statement
    : IF0 OPEN_PAREN expression CLOSE_PAREN block_statement
      optional_else_clause
	{ Stmt_if($3, $5, $6) }

optional_else_clause
    : ELSE conditional_statement { $2 }
    | ELSE block_statement { $2 }
    | /* empty */ { Stmt_nop }
    ;

for_loop_statement
    : FOR0 ident ASSIGN expression COLON expression optional_step
      block_statement
	{ Stmt_for(Ident_valid($2,Type_scalar(Type_int),Ident_bind,
				An_list([])),
	    $4, $6, $7, $8) }
    ;

optional_step
    : COLON expression { $2 }
    | /* empty */ { Expr_int(1) }
    ;

do_loop_statement
    : DO0 do_test block_statement { Stmt_do($2, $3) }
    ;

do_test
    : set_expression { $1 }
    | OPEN_PAREN expression CLOSE_PAREN { $2 }
    ;


/* *todo* results */
call_statement
    : call_result CALL0 ident optional_arg_list SEMI_COLON
	{
	    (* *todo* Is this the right order of Stmt_call? *)
	    Stmt_call(Expr_var($3), $1, $4)
	}
    ;

call_result
    : expression ASSIGN { $1 }
    | /*null*/ { Expr_void }
    ;

directive_statement
    : annotation_list statement { Stmt_directive($1,$2) }
    ;

optional_arg_list
    : OPEN_PAREN expression_list CLOSE_PAREN { $2 }
    | /* empty */ { [] }
    ;

/************************************************************************/
/* Annotations								*/
/************************************************************************/

annotation
    : annotation_list { $1 }
    | block_statement { An_stmt($1) }
    | OPEN_PAREN expression CLOSE_PAREN { An_expr($2) }
    | OPEN_SQUARE arbitrary_type CLOSE_SQUARE { An_type($2) }
    | string { An_string($1) }
    | ident COLON { An_string($1) }
    ;

optional_annotation_list
    : /* empty */ { An_list([]) }
    | annotation_list { $1 }
    ;

annotation_list
    : OPEN_ANN annotation_items CLOSE_ANN { An_list($2) }
    ;

annotation_items
    : annotation_items annotation { $1 @ [$2] }
    | { [] }
    ;
    

/************************************************************************/
/* Misc.								*/
/************************************************************************/

string
    : STRINGVAL { $1 }
    | IN_PLACE_STRING { $1 }
    ;

ident_list
    : ident_list COMMA ident_list_item { $1 @ $3 }
    | ident_list_item { $1 }
    ;

ident_list_item
    : ident { [$1] }
    | IN_PLACE_STRING_LIST { $1 }
    | IN_PLACE_EMPTY_LIST { [] }
    ;

ident
    : IDENT { $1 }
    | IN_PLACE_STRING { $1 }
    ;

expression_list
    : expression_list COMMA expression_list_item { $1 @ $3 }
    | expression_list_item { $1 }
    ;

expression_list_item
    : expression { [$1] }
    | IN_PLACE_EXPR_LIST { $1 }
    | IN_PLACE_STRING_LIST { List.map (fun x -> Expr_var(x)) $1 }
    | IN_PLACE_EMPTY_LIST { [] }
    ;

int_list
    :  int_list2 { $1 }
    | { [] }
    
int_list2
    : int_list2 COMMA DECIMALINT { $1 @ [$3] }
    | DECIMALINT { [$1] }
    ;

int_vector
    : OPEN_SQUARE int_list CLOSE_SQUARE
      { (caml_to_alpha_vector (Array.of_list $2)) }
    ;
      
int_vector_list
    : int_vector_list2 { $1 }
    | { [] }
    ;

int_vector_list2
    : int_vector_list2 COMMA int_vector { $1 @ [$3] }
    | int_vector { [$1] }
    ;

int_matrix
    : OPEN_SQUARE col_marker_list CLOSE_SQUARE 
      { (create_matrix 0 $2) }
    | OPEN_SQUARE int_vector_list CLOSE_SQUARE 
      { if ((List.length $2) = 0) then create_matrix 0 0
        else 
          (caml_to_alpha_matrix 
            ((Array.of_list (List.map alpha_vector_to_caml $2)))) }

col_marker_list
    : col_marker_list DOT { $1+1 }
    | { 0 }

int_affine_map
    : int_matrix int_vector { Affine_map($1, $2) }

%%
