(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers

exception Parser_Error of string;;
exception Parser_String_Error of string * string;;

let start_parser entry lexbuf (args : (string * Bmf.annotation) list) =
    let old_args = !Bmf_scanner.parser_args
    in
	try
	    Bmf_scanner.parser_args :=
		List.fold_left (fun m (key,val0) ->
			    String_map.add key val0 m)
		    String_map.empty args;
	    let result = entry Bmf_scanner.token lexbuf
	    in
		Bmf_scanner.parser_args := old_args;
		result
	with exn ->
	    Bmf_scanner.parser_args := old_args;
	    match exn with
		Parsing.Parse_error ->
		    raise (Parser_Error(
			"Syntax error around char "
			  ^ (string_of_int
				(Lexing.lexeme_start lexbuf))
			  ^ "."))
	    |   Bmf_scanner.Lexical_error i ->
		    raise (Parser_Error(
			"Lexical error around char "
			  ^ (string_of_int i)
			  ^ "."))
	    |   _ -> raise exn
;;

let rec parse_channel inchan =
    start_parser Bmf_parser.file
	(Lexing.from_channel inchan)
	[]

and parse_file str =
    Io'.call_with_open_input_file str parse_channel
	
and parse_stdin () =
    parse_channel stdin
;;    

let
  parse_matrix str =
    let lexbuf = Lexing.from_string str
    in
      try
	Bmf_parser.int_matrix Bmf_scanner.token lexbuf
      with exn ->
	    begin
		match exn with
		    Parsing.Parse_error ->
			raise (Parser_Error(
			    "Syntax error parsing matrix around char "
			      ^ (string_of_int
				    (Lexing.lexeme_start lexbuf))
			      ^ "."))
		|   Bmf_scanner.Lexical_error i ->
			raise (Parser_Error(
			    "Lexical error parsing matrix around char "
			      ^ (string_of_int i)
			      ^ "."))
		|   _ -> raise exn
	    end
;;

let
  parse_vector str =
    let lexbuf = Lexing.from_string str
    in
      try
	Bmf_parser.int_vector Bmf_scanner.token lexbuf
      with exn ->
	    begin
		match exn with
		    Parsing.Parse_error ->
			raise (Parser_Error(
			    "Syntax error parsing vector around char "
			      ^ (string_of_int
				    (Lexing.lexeme_start lexbuf))
			      ^ "."))
		|   Bmf_scanner.Lexical_error i ->
			raise (Parser_Error(
			    "Lexical error parsing vector around char "
			      ^ (string_of_int i)
			      ^ "."))
		|   _ -> raise exn
	    end
;;

let
  parse_affine_map str =
    let lexbuf = Lexing.from_string str
    in
      try
	Bmf_parser.int_affine_map Bmf_scanner.token lexbuf
      with exn ->
	    begin
		match exn with
		    Parsing.Parse_error ->
			raise (Parser_Error(
			    "Syntax error parsing affine_map around char "
			      ^ (string_of_int
				    (Lexing.lexeme_start lexbuf))
			      ^ "."))
		|   Bmf_scanner.Lexical_error i ->
			raise (Parser_Error(
			    "Lexical error parsing affine_map around char "
			      ^ (string_of_int i)
			      ^ "."))
		|   _ -> raise exn
	    end
;;

let parse_xxx parse_f str args =
  try
    start_parser parse_f
      (Lexing.from_string str)
      args
  with Parser_Error(msg) ->
    raise (Parser_String_Error(str,msg))

let parse_stmt str args = parse_xxx Bmf_parser.statement_entry str args
let parse_expr str args = parse_xxx Bmf_parser.expression_entry str args
let parse_annot str args = parse_xxx Bmf_parser.annotation_entry str args
let parse_ident str args = parse_xxx Bmf_parser.ident_entry str args
let parse_vars str args = parse_xxx Bmf_parser.vars_entry str args

(************************************************************************)

exception M4Failure

let do_m4_preproc m4cmd m4opts infile_name =
    let tmpdir =
	(try
	    Sys.getenv "TMPDIR"
	with Not_found ->
	    "/tmp")
    and tmpnam =
	"bmf-m4-"^(string_of_int (Unix.getpid ()))^".bmf"
    in let tmpfile = Filename.concat tmpdir tmpnam
    in let infile_name =
	if infile_name = "-" then
	    ""
	else
	    infile_name
    in let cmd =
	m4cmd ^ " " ^ m4opts ^ " " ^ infile_name ^ " >" ^ tmpfile
    in let stat =
	Sys.command cmd
    in
	if stat <> 0 then begin
	    Sys.remove tmpfile;
	    raise M4Failure
	end;
	tmpfile

(************************************************************************)

(* fixme: add switches for changing these values. *)
let m4cmd = ref Config.m4_command
and m4opts = ref ""
and default_m4opts = ""

let _ = 
  Bsct_toplevel.register_switch "-in_bmf" "*"
    ("-in_bmf:m FILE -- Read BMF from FILE.\n"^
     "	FILE = - reads from stdin.\n"^
     "	:m -- preprocess FILE using m4.")
    (fun _ args ->
      let name = Bsct_toplevel.get_arg_name (List.hd args)
      and (options,param,args') = Bsct_toplevel.parse_arg true args
      in let (m4_p,options) = Bsct_toplevel.extract_option options 'm'
      in let _ = Bsct_toplevel.ignore_options options name
      in let prepout_name =
    	if m4_p then do_m4_preproc !m4cmd (!m4opts^default_m4opts) param
    	else param
      in let top_decls =
    	if prepout_name = "-" then parse_stdin ()
    	else parse_file prepout_name
      in
      if param <> prepout_name then
    	Sys.remove prepout_name;
      (Bmf.to_s_expr top_decls,args'))
