(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers

open Bmf;;

let lowest_prec = 0
and infin_prec = 9999;;
let op_precedence op =
    match op with
    | Op_implies -> 0
    | Op_or -> 1
    | Op_and -> 2
    | (Op_leq|Op_eq) -> 3
    | Op_plus -> 4
    | (Op_times|Op_div) -> 5
    | (Op_not | Op_neg | Op_sqrt | Op_addr | Op_deref | Op_other _)  
      -> infin_prec

;;

let indent_delta = 3;;

(************************************************************************)

let paren_list fmt lst f =
    Format.pp_print_string fmt "(";
    Format.pp_open_hovbox fmt 0;
    Format.pp_print_cut fmt ();
    (match lst with
      x::lst' ->
	f x;
	List.iter
	    (fun x ->
		Format.pp_print_string fmt ",";
		Format.pp_print_cut fmt ();
		f x)
	    lst'
    | [] -> ());
    Format.pp_print_string fmt ")";
    Format.pp_close_box fmt ()
;;

let maybe_a_paren_list fmt lst f =
    match lst with
      x::lst' -> paren_list fmt lst f
    | _ -> ()
;;

(************************************************************************)
(* Entry points								*)
(************************************************************************)

let rec unparse_channel outchan width decls =
  let fmt = 
    Format.make_formatter (output outchan) (fun () -> flush outchan)
  in
  Format.pp_set_max_boxes fmt 999999999;
  Format.pp_set_margin fmt (if width = -1 then (999999999-1) else width);
  List.iter 
    (fun decl ->
      print_decl fmt decl;
      Format.pp_print_newline fmt ();
      Format.pp_print_newline fmt ())
    decls;
  Format.pp_print_flush fmt ()

and unparse_file str width decls =
    Io'.call_with_open_output_file str
	(fun outchan -> unparse_channel outchan width decls)
and unparse_stdout width decls =
    unparse_channel stdout width decls

and idents_printer fmt ids =
    List.iter (fun decl ->
		print_decl fmt decl;
		Format.pp_print_newline fmt ())
	ids;
    Format.pp_print_flush fmt ()

and ident_printer fmt id =
    print_decl fmt id

and stmt_printer fmt stmt =
    print_stmt fmt stmt

and expr_printer fmt expr =
    print_expr fmt expr

and type_printer fmt typ =
    print_type fmt typ

and annotation_printer fmt an =
    print_annot fmt an

(************************************************************************)
(* Declarations								*)
(************************************************************************)

and print_decl fmt decl =
    let (name, typ, stuff, ann) =
	match decl with
	  Ident_valid(name, typ, stuff,ann) -> (name, typ, stuff, ann)
	| _ -> failwith("What sort of decl?")
    in begin
	match stuff with
	  Ident_bind -> failwith "Unexpected ident_bind"
	| Ident_var(init) ->
            Format.pp_open_hovbox fmt indent_delta;
	    Format.pp_print_string fmt ("var "^name^": ");
	    print_type fmt typ;
	    if init <> Expr_void then begin
		Format.pp_print_space fmt ();
		Format.pp_print_string fmt ":=";
		Format.pp_print_space fmt ();
		print_expr fmt init
	    end;
	    if ann <> An_list([]) then begin
		Format.pp_print_space fmt ();
		print_annot_list fmt ann
	    end;
	    Format.pp_print_string fmt ";";
            Format.pp_close_box fmt ()
	| Ident_proc(Frame(frm), result, body) ->
	    assert (ann = An_list([]))
		(* (Failure("No annotations on procs, yet.")) *);
	    Format.pp_open_vbox fmt 0;
	    Format.pp_print_string fmt ("procedure "^name);
	    maybe_a_paren_list fmt frm
		(fun id -> Format.pp_print_string fmt (bmf_ident_name id));
            Format.pp_print_space fmt ();
	    Format.pp_open_vbox fmt indent_delta;
	    Format.pp_print_string fmt "parambegin";
	    List.iter (fun id -> Format.pp_print_space fmt (); print_decl fmt id) frm;
	    Format.pp_close_box fmt ();
	    Format.pp_print_space fmt ();
	    Format.pp_print_string fmt "paramend";
            Format.pp_print_space fmt ();
	    print_block fmt body;
	    Format.pp_close_box fmt ()
    end

(************************************************************************)
(* Statements								*)
(************************************************************************)

and print_block fmt stmt =
begin
    let rec find_block stmt =
	match stmt with
	  Stmt_seq(first, second) ->
	    (find_block first) @ (find_block second)
	| Stmt_nop -> []
	| _ -> [stmt]
    in let (decls,stmt) =
	match stmt with
	  Stmt_let(Frame(decls),stmt) -> (decls,stmt)
	| _ -> ([],stmt)
    in let stmts = find_block stmt
    in
	Format.pp_open_vbox fmt 0;
	Format.pp_open_vbox fmt indent_delta;
        Format.pp_print_string fmt "{";
	List.iter
	    (fun decl ->
		Format.pp_print_space fmt ();
		print_decl fmt decl)
	    decls;
	if decls <> [] && stmts <> [] then begin
            Format.pp_force_newline fmt ();
	end;
	List.iter
	    (fun stmt ->
		Format.pp_print_space fmt ();
		print_stmt fmt stmt)
	    stmts;
        Format.pp_close_box fmt ();
	Format.pp_print_space fmt ();
        Format.pp_print_string fmt "}";
        Format.pp_close_box fmt ()
end

and print_stmt fmt stmt =
begin
    match stmt with
      Stmt_assign(left, right) ->
	Format.pp_open_hovbox fmt (2*indent_delta);
	print_expr fmt left;
	Format.pp_print_string fmt " := ";
	print_expr fmt right;
	Format.pp_print_string fmt ";";
        Format.pp_close_box fmt ()
    (* *todo* Is this the right order of stmt_call? *)
    | Stmt_call(f,result,args) ->
	let f_name =
	    match f with
	      Expr_var(f_name) -> f_name
	    | _ -> failwith "Bad function expr"
	in
	    Format.pp_open_hovbox fmt (2*indent_delta);
	    if result <> Expr_void then begin
		print_expr fmt result;
		Format.pp_print_string fmt " := "
	    end;
	    Format.pp_print_string fmt ("call "^f_name);
	    maybe_a_paren_list fmt args
		(fun arg -> print_expr fmt arg);
	    Format.pp_print_string fmt ";";
	    Format.pp_close_box fmt ()
    | Stmt_for(id,lb,ub,step,body) ->
	Format.pp_open_vbox fmt indent_delta;
	Format.pp_open_hovbox fmt (2*indent_delta);
	Format.pp_print_string fmt ("for "^(bmf_ident_name id)^" := ");
	print_expr fmt lb;
	Format.pp_print_string fmt ":";
	print_expr fmt ub;
	(match step with
	  Expr_int(1) -> ()
	| _ ->
	    Format.pp_print_string fmt ":";
	    print_expr fmt step);
	Format.pp_close_box fmt ();
        Format.pp_print_space fmt ();
	print_block fmt body;
        Format.pp_close_box fmt ()
    | Stmt_while(test,body) ->
	Format.pp_open_vbox fmt indent_delta;
	Format.pp_open_hovbox fmt (2*indent_delta);
	Format.pp_print_string fmt "while (";
	print_expr fmt test;
	Format.pp_print_string fmt ")";
	Format.pp_close_box fmt ();
        Format.pp_print_space fmt ();
	print_block fmt body;
        Format.pp_close_box fmt ()
    | Stmt_if(test,thn,els) ->
	
 let rec loop els = begin
		match els with
		  Stmt_nop -> ()
		| Stmt_if(test',thn',els') ->
		    Format.pp_print_space fmt ();
		    Format.pp_open_vbox fmt indent_delta;
		    Format.pp_open_hovbox fmt (2*indent_delta);
		    Format.pp_print_string fmt "else if (";
		    print_expr fmt test';
		    Format.pp_print_string fmt ") ";
		    Format.pp_close_box fmt ();
		    Format.pp_print_space fmt ();
		    print_block fmt thn';
		    Format.pp_close_box fmt ();
		    loop els'
		| _ ->
		    Format.pp_print_space fmt ();
		    Format.pp_open_vbox fmt indent_delta;
		    Format.pp_print_string fmt "else";
                    Format.pp_print_space fmt ();
		    print_block fmt els;
                    Format.pp_close_box fmt ()
	    end
 in Format.pp_open_vbox fmt indent_delta;
	Format.pp_open_hovbox fmt (2*indent_delta);
	Format.pp_print_string fmt "if (";
	print_expr fmt test;
	Format.pp_print_string fmt ")";
	Format.pp_close_box fmt ();
        Format.pp_print_space fmt ();
	print_block fmt thn;
	Format.pp_close_box fmt ();
	loop els
	    
    | Stmt_do(test,body) ->
	Format.pp_open_vbox fmt indent_delta;
        Format.pp_open_hovbox fmt (2*indent_delta);
	Format.pp_print_string fmt "do ";
	(match test with
	  Expr_set(_,_) -> print_expr fmt test
	| _ ->
	    Format.pp_print_string fmt "(";
	    print_expr fmt test;
	    Format.pp_print_string fmt ")");
        Format.pp_close_box fmt ();
	Format.pp_print_space fmt ();
	print_block fmt body;
        Format.pp_close_box fmt ()
    | Stmt_directive(ann,s) ->
	Format.pp_open_vbox fmt 0;
	print_annot_list fmt ann;
        Format.pp_print_space fmt ();
	print_stmt fmt s;
	Format.pp_close_box fmt ()
    | (Stmt_seq(_,_)|Stmt_let(_,_)|Stmt_nop) ->
	print_block fmt stmt

end

(************************************************************************)
(* Expression								*)
(************************************************************************)

and print_expr fmt expr =
    print_expr' fmt lowest_prec expr

and print_expr' fmt prec expr =
let precedence expr =
    match expr with
      Expr_op(Op_not,[Expr_op(Op_leq,_)]) -> (op_precedence Op_leq)
    | Expr_op(Op_not,[Expr_op(Op_eq,_)]) -> (op_precedence Op_eq)
    | Expr_op(op,_) -> op_precedence op
    | _ -> infin_prec
in let prec' = precedence expr
in begin
    if (prec >= prec') then Format.pp_print_string fmt "(";
    (match expr with
      Expr_op(Op_not,[Expr_op(Op_leq,[arg1;arg2])]) ->
	print_args fmt ">" prec' [arg1;arg2]
    | Expr_op(Op_not,[Expr_op(Op_eq,[arg1;arg2])]) ->
	print_args fmt "!=" prec' [arg1;arg2]
    | Expr_op(op,args) ->
	(match op with
	  (Op_neg|Op_not) ->
	    assert ((List.length args) == 1)
		(* (Failure("Supposed to be unary.")) *);
	    Format.pp_print_string fmt (string_of_op op);
	    print_args fmt "*Notused*" prec' args
	| (Op_div|Op_eq|Op_leq) ->
	    assert ((List.length args) ==2)
		(* (Failure("Supposed to be binary.")) *);
	    print_args fmt (string_of_op op) prec' args
	| (Op_plus|Op_times|Op_and|Op_or) ->
	    print_args fmt (string_of_op op) prec'
		(bmf_expr_args_by_op op expr)
	| _ ->
	    Format.pp_open_hvbox fmt indent_delta;
	    Format.pp_print_string fmt (string_of_op op);
	    Format.pp_print_string fmt "(";
	    List.fold_left
		(fun first_p e' ->
		    if (not (first_p)) then Format.pp_print_string fmt ",";
		    Format.pp_print_cut fmt ();
		    print_expr fmt e';
		    false)
		true args;
	    Format.pp_print_string fmt ")";
	    Format.pp_close_box fmt ()
	)
    | Expr_int(n) -> Format.pp_print_int fmt n
    | Expr_real(n) -> Format.pp_print_float fmt n
    | Expr_string(s) -> Format.pp_print_string fmt ("\""^(String.escaped s)^"\"")
    | Expr_bool(b) -> Format.pp_print_string fmt (if b then "true" else "false")
    | Expr_void -> failwith "void not implemented."
    | Expr_var(str) -> Format.pp_print_string fmt str
    | Expr_forall(Frame(frm),pred) ->
	print_quant_expr fmt "forall" frm pred
    | Expr_exist(Frame(frm),pred) ->
	print_quant_expr fmt "exist" frm pred
    | Expr_access(ac,r,mthd,args) ->
	let (left,right) =
	    match ac with
	      Access_lookup -> ("(",")")
	    | Access_weak -> ("[","]")
	    | Access_strong -> ("[[","]]")
	and args' =
	    if mthd = "" then args else Expr_string(mthd)::args
	in
	    print_expr fmt r;
	    Format.pp_print_string fmt left;
            Format.pp_print_space fmt ();
	    (match args' with
	      x::args'' ->
		print_expr fmt x;
		List.iter (fun x ->
			Format.pp_print_string fmt ",";
                        Format.pp_print_cut fmt ();
			print_expr fmt x)
		    args''
	    | [] -> ());
            Format.pp_print_space fmt ();
	    Format.pp_print_string fmt right
    | Expr_set(Frame(frm),pred) ->
	print_quant_expr fmt "" frm pred
    );
    if (prec >= prec') then Format.pp_print_string fmt ")"
end

and string_of_op op =
    (match op with
      Op_plus -> "+"
    | Op_times -> "*"
    | Op_div -> "/"
    | Op_and -> "&&"
    | Op_or -> "||"
    | Op_eq -> "=="
    | Op_leq -> "<="
    | Op_not -> "not"
    | Op_neg -> "-"
    | Op_sqrt -> "sqrt"
    | Op_implies -> "=>"
    | Op_addr -> "&"
    | Op_deref -> "*"
    | Op_other(s) -> "%op("^s^")"
    )

and print_args fmt op prec args =
    Format.pp_open_hovbox fmt indent_delta;
    (
 let rec loop args first_p =
	    if ((not (first_p))) then begin
		Format.pp_print_space fmt ();
		Format.pp_print_string fmt op;
		Format.pp_print_space fmt ()
	    end;
	    match args with
		[x] ->
		    print_expr' fmt prec x
	    |	x::args' ->
		    print_expr' fmt prec x;
		    loop args' false
	    |	[] -> failwith "No args to op?"
 in loop args true
	);
    Format.pp_close_box fmt ()

and print_quant_expr fmt q frm pred =
    Format.pp_open_hovbox fmt indent_delta;
    if q <> "" then
      begin
        Format.pp_print_string fmt "(";
  	Format.pp_print_string fmt q;
	Format.pp_print_space fmt ()
      end
    else
        Format.pp_print_string fmt "{";
    print_bracketted_frame fmt frm;
    Format.pp_print_space fmt ();
    Format.pp_print_string fmt "|";
    Format.pp_print_space fmt ();
    print_expr fmt pred;
    if q <> "" then
	Format.pp_print_string fmt ")"
    else
	Format.pp_print_string fmt "}";
    Format.pp_close_box fmt ()

and print_bracketted_frame fmt frm =
  Format.pp_open_hovbox fmt indent_delta;
  Format.pp_print_string fmt "[";
  (match frm with
    x::frm' ->
      print_frame_id fmt x;
      List.iter
	(fun x ->
	  Format.pp_print_string fmt ",";
	  Format.pp_print_cut fmt ();
	  print_frame_id fmt x)
	frm'
  | [] -> ());
  Format.pp_print_string fmt "]";
  Format.pp_close_box fmt ()


and print_frame_id fmt id =
    match id with
      Ident_valid(name,Type_scalar(Type_int),Ident_bind,_) ->
	Format.pp_print_string fmt name
    | Ident_valid(name,Type_scalar(typ),Ident_bind,_) ->
	Format.pp_print_string fmt name;
	Format.pp_print_cut fmt ();
	Format.pp_print_string fmt ":";
	Format.pp_print_cut fmt ();
	print_scalar_type fmt typ
    | Ident_valid(name,Type_scalar(typ),Ident_var(init),_) ->
	assert (init = Expr_void)
	    (* (Failure("Init not allowed for quantified var.")) *);
	Format.pp_print_string fmt "var";
	Format.pp_print_space fmt ();
	Format.pp_print_string fmt name;
	Format.pp_print_cut fmt ();
	Format.pp_print_string fmt ":";
	Format.pp_print_cut fmt ();
	print_scalar_type fmt typ
    | _ -> failwith "what sort of quant id is this ?!?"

(************************************************************************)
(* Types								*)
(************************************************************************)

and print_type fmt typ =
    match typ with
      Type_scalar(x) -> print_scalar_type fmt x
    | Type_proc(args,result) -> failwith "Unimplemented type"
    | Type_array(atyp) -> print_array_type fmt atyp
    | Type_ptr(btyp) ->
	Format.pp_print_string fmt "pointer";
      	Format.pp_print_space fmt ();
      	Format.pp_print_string fmt "of";
	print_type fmt btyp
    | Type_ann(ann) ->
	Format.pp_print_string fmt "type_ann";
	Format.pp_print_space fmt ();
	print_annot_list fmt ann
    | Type_unknown -> failwith "Unknown type"

and print_array_type fmt atyp =
  Format.pp_print_string fmt "array";
  Format.pp_print_space fmt ();
  Format.pp_open_hovbox fmt 3;
  (match atyp with
    Array_bb(n,typ',bb_name,bb_args) ->
      Format.pp_print_string fmt "int^";
      Format.pp_print_int fmt n;
      Format.pp_print_space fmt ();
      Format.pp_print_string fmt "of";
      Format.pp_print_space fmt ();
      print_scalar_type fmt typ';
      Format.pp_print_space fmt ();
      Format.pp_print_string fmt "(";
      Format.pp_print_string fmt bb_name;
      List.iter
	(fun (tag,val0) ->
	  Format.pp_print_space fmt ();
	  Format.pp_print_string fmt tag;
	  Format.pp_print_string fmt ":";
	  Format.pp_print_space fmt ();
	  print_annot fmt val0)
	bb_args;
      Format.pp_print_string fmt ")"
  | Array_dense(lb,ubs,majorness,typ') ->
      let print_bounds fmt ub =
	Format.pp_print_string fmt
	  (if lb = Dense_lb_1 then "1" else "0");
	Format.pp_print_string fmt ":";
	print_expr fmt ub
      in begin
	Format.pp_print_string fmt
	  (if majorness = Dense_col_major then "(" else "[");
	print_bounds fmt (List.hd ubs);
	List.iter
	  (fun ub -> 
	    Format.pp_print_string fmt ","; print_bounds fmt ub)
	  (List.tl ubs);
	Format.pp_print_string fmt 
	  (if majorness = Dense_col_major then ")" else "]");
	Format.pp_print_string fmt " of ";
	print_scalar_type fmt typ'	
      end
  | Array_view(Frame(out_frm),Frame(in_frm),constr,base_atyp) ->
      Format.pp_print_string fmt "{";
      Format.pp_print_space fmt ();
      print_bracketted_frame fmt out_frm;
      Format.pp_print_space fmt ();
      Format.pp_print_string fmt "<-";
      Format.pp_print_space fmt ();
      print_bracketted_frame fmt in_frm;
      Format.pp_print_string fmt ":";
      Format.pp_print_space fmt ();
      print_expr fmt constr;
      Format.pp_print_space fmt ();
      Format.pp_print_string fmt "}";
      Format.pp_print_space fmt ();
      Format.pp_print_string fmt "of";
      Format.pp_print_space fmt ();
      print_array_type fmt base_atyp
	);
  Format.pp_close_box fmt ()

and print_scalar_type fmt typ =
    match typ with
      Type_int -> Format.pp_print_string fmt "int"
    | Type_bool -> Format.pp_print_string fmt "bool"
    | Type_real -> Format.pp_print_string fmt "real"
    | Type_void -> Format.pp_print_string fmt "void"
    | Type_string -> Format.pp_print_string fmt "string"

(************************************************************************)
(* Annotations								*)
(************************************************************************)
and print_annot fmt ann =
    match ann with
      An_type(ty) ->
	Format.pp_print_string fmt "[";
	print_type fmt ty;
	Format.pp_print_string fmt "]"
    | An_stmt(s) -> 
        Format.pp_print_string fmt "{";
        print_stmt fmt s;
        Format.pp_print_string fmt "}"
    | An_expr(e) ->
	Format.pp_print_string fmt "(";
	print_expr fmt e;
	Format.pp_print_string fmt ")"
    | An_string(str) ->
	Format.pp_print_string fmt ("\""^(String.escaped str)^"\"")
    | An_list(lst') -> print_annot_list fmt ann

and print_annot_list fmt ann =
    Format.pp_open_hovbox fmt 3;
    Format.pp_print_string fmt "<<";
    (match ann with
      An_list(lst) ->
	List.iter (fun x ->
		    Format.pp_print_space fmt ();
		    print_annot fmt x)
	    lst
    | _ -> failwith "The top-most annot must be a list.");
    Format.pp_print_string fmt " >>";
    Format.pp_close_box fmt ()

(************************************************************************)
;;

(************************************************************************)
(* Routines for normalizing identifiers					*)
(*									*)
(* These routines will traverse BMF AST's and will try to remove, or at *)
(* least normalize the tags the are generated for gensym'ed routines.	*)
(* The reason for having these routines is to remove the effects of 	*)
(* gensym ordering from the generated code.				*)
(************************************************************************)

let rec normalize_global id =
  match normalize_globals [id] with
    [id'] -> id'
  | _ -> assert false

and normalize_globals globals =

  let global_names =
    List.fold_left (fun s id -> String_set.add (bmf_ident_name id) s)
	String_set.empty globals

  (* used_names = the maximum tag int used for each base name. *)
  (* new_names = map from old name -> new name. *)

  and initial_used_names = String_map.empty
  and initial_new_names = String_map.empty

  in let rec find_new_name base_name max_tag =
    let new_name =
      if max_tag = 0 then
	base_name
      else
	(* This is supposed to be the same method as in Misc.gensym. *)
	base_name ^ "__" ^ (string_of_int (max_tag+1))
    in
    if String_set.mem new_name global_names then
      find_new_name base_name (max_tag+1)
    else
      (max_tag,new_name)

  in let ident_def (used_names,new_names) id =
    match id with
      Ident_valid(name,_,_,_) ->
    	let base_name = Misc.gensym_base name in
    	let init_max_tag =
	  try String_map.find base_name used_names
	  with Not_found -> 0
    	in let (max_tag,new_name) = find_new_name base_name init_max_tag
    	in let used_names' = 
	  String_map.add base_name (max_tag+1) used_names
    	and new_names' = 
	  String_map.add name new_name new_names
    	in (used_names',new_names')
    | Ident_null -> (used_names,new_names)

  in let rec main () =
    List.map walk_global_ident globals

  and walk_global_ident ident =
    match ident with
      Ident_valid(name,typ,stuff,ann) ->
	let stuff' = (walk_stuff initial_used_names initial_new_names
			stuff)
	in Ident_valid(name,typ,stuff',ann)
    | Ident_null -> ident

  and walk_ident used_names new_names ident =
    match ident with
      Ident_valid(name,typ,stuff,ann) ->
	let name' = String_map.find name new_names
	and typ' = walk_type used_names new_names typ
	and stuff' = walk_stuff used_names new_names stuff
	and ann' = walk_ann used_names new_names ann
	in Ident_valid(name',typ',stuff',ann')
    | Ident_null -> ident

  and walk_stuff used_names new_names stuff =
    match stuff with
      Ident_var(init) -> 
	Ident_var(walk_expr used_names new_names init)
    | Ident_bind -> stuff
    | Ident_proc(Frame(params),result,body) ->
	let (used_names',new_names')
	    = (List.fold_left ident_def 
		 (ident_def (used_names,new_names) result)
		 params)
	in let params' 
	    = List.map (walk_ident used_names' new_names') params
	and result' =
	  walk_ident used_names' new_names' result
	and body' =
	  walk_stmt used_names' new_names' body
	in Ident_proc(Frame(params'),result',body')
       
  and walk_stmt used_names new_names stmt =
    match stmt with
      Stmt_assign(lhs,rhs) ->
	let lhs' = walk_expr used_names new_names lhs
	and rhs' = walk_expr used_names new_names rhs
	in Stmt_assign(lhs',rhs')
    | Stmt_seq(first,second) ->
	let first' = walk_stmt used_names new_names first
	and second' = walk_stmt used_names new_names second
	in Stmt_seq(first',second')
    | Stmt_call(f,lhs,args) ->
	let f' = walk_expr used_names new_names f
	and lhs' = walk_expr used_names new_names lhs
	and args' = List.map (walk_expr used_names new_names) args
	in Stmt_call(f',lhs',args')
    | Stmt_for(id,lb,ub,step,body) ->
	let (used_names',new_names') =
	  ident_def (used_names,new_names) id
	in let id' = walk_ident used_names' new_names' id
	and lb' = walk_expr used_names new_names lb
	and ub' = walk_expr used_names new_names ub
	and step' = walk_expr used_names new_names step
	and body' = walk_stmt used_names' new_names' body
	in Stmt_for(id',lb',ub',step',body')
    | Stmt_while(test,body) ->
	let test' = walk_expr used_names new_names test
	and body' = walk_stmt used_names new_names body
	in Stmt_while(test',body')
    | Stmt_if(test,cons,altn) ->
	let test' = walk_expr used_names new_names test
	and cons' = walk_stmt used_names new_names cons
	and altn' = walk_stmt used_names new_names altn
	in Stmt_if(test',cons',altn')
    | Stmt_let(Frame(ids),body) ->
	let (used_names',new_names') =
	  List.fold_left ident_def (used_names,new_names) ids
	in let ids' = List.map (walk_ident used_names' new_names') ids
	and body' = walk_stmt used_names' new_names' body
	in Stmt_let(Frame(ids'),body')
    | Stmt_do(Expr_set(Frame(ids),pred),body) ->
	let (used_names',new_names') =
	  List.fold_left ident_def (used_names,new_names) ids
	in let ids' = List.map (walk_ident used_names' new_names') ids
	and pred' = walk_expr used_names' new_names' pred
	and body' = walk_stmt used_names' new_names' body
	in Stmt_do(Expr_set(Frame(ids'),pred'),body')
    | Stmt_do(_,_) ->
	(* Can't happen. Every Stmt_do should be used with an Expr_set. *)
	assert false
    | Stmt_nop -> stmt
    | Stmt_directive(ann,body) ->
	let ann' = walk_ann used_names new_names ann
	and body' = walk_stmt used_names new_names body
	in Stmt_directive(ann',body')

  and walk_type used_names new_names typ =
      match typ with
      	Type_proc(param_typs,result_typ) ->
	  let param_typs' =
	    List.map (walk_type used_names new_names) param_typs
	  and result_typ' =
	    walk_type used_names new_names result_typ
	  in Type_proc(param_typs',result_typ')
      |	Type_ann(ann) ->
	  let ann' = walk_ann used_names new_names ann
	  in Type_ann(ann')
      |	Type_array(atyp) -> 
	  Type_array(walk_array_type used_names new_names atyp)
      |	Type_ptr(btyp) ->
	  Type_ptr(walk_type used_names new_names btyp)
      |	(Type_scalar(_)|Type_unknown) -> 
	  typ

  and walk_array_type used_names new_names atyp =
    match atyp with
    | Array_dense(lb,ubs,maj,base_sc_typ) ->
	let ubs' = List.map (walk_expr used_names new_names) ubs
	in Array_dense(lb,ubs',maj,base_sc_typ)
    | Array_bb(dim,base_sc_type,bb_name,bb_args) ->
	let bb_args' =
	  List.map
	    (fun (tag,val0) ->
	      let val0' = walk_ann used_names new_names val0
	      in (tag,val0'))
	    bb_args
	in Array_bb(dim,base_sc_type,bb_name,bb_args')
    | Array_view(Frame(out_frm),Frame(in_frm),constr,base_atyp) ->
	let (used_names',new_names') =
	  List.fold_left ident_def (used_names,new_names) out_frm
	in let (used_names'',new_names'') =
	  List.fold_left ident_def (used_names',new_names') in_frm
	in 
	(* constr is scoped inside of out_frm and in_frm. 
	   base_atyp is not. *)
	let out_frm' = 
	  List.map (walk_ident used_names'' new_names'') out_frm
	and in_frm' = 
	  List.map (walk_ident used_names'' new_names'') in_frm
	and constr' = 
	  walk_expr used_names'' new_names'' constr
	and base_atyp' =
	  walk_array_type used_names new_names base_atyp
	in Array_view(Frame(out_frm'),Frame(in_frm'),constr',base_atyp')

  and walk_expr used_names new_names expr =
    match expr with
      (Expr_int(_)|Expr_real(_)|Expr_string(_)|Expr_bool(_)|
      Expr_void) -> expr
    | Expr_var(name) -> 
	let name' =
	  try String_map.find name new_names
	  with Not_found -> name
	in Expr_var(name')
    | Expr_op(op,args) ->
	let args' = List.map (walk_expr used_names new_names) args
	in Expr_op(op, args')
    | Expr_forall(Frame(ids),body) ->
	let (used_names',new_names') =
	  List.fold_left ident_def (used_names,new_names) ids
	in let ids' = List.map (walk_ident used_names' new_names') ids
	and body' = walk_expr used_names' new_names' body
	in Expr_forall(Frame(ids'),body')
    | Expr_exist(Frame(ids),body) ->
	let (used_names',new_names') =
	  List.fold_left ident_def (used_names,new_names) ids
	in let ids' = List.map (walk_ident used_names' new_names') ids
	and body' = walk_expr used_names' new_names' body
	in Expr_forall(Frame(ids'),body')
    | Expr_set(Frame(ids),body) ->
	let (used_names',new_names') =
	  List.fold_left ident_def (used_names,new_names) ids
	in let ids' = List.map (walk_ident used_names' new_names') ids
	and body' = walk_expr used_names' new_names' body
	in Expr_forall(Frame(ids'),body')
    | Expr_access(acc,rel,name,args) ->
	let rel' = walk_expr used_names new_names rel
	and args' = List.map (walk_expr used_names new_names) args
	in Expr_access(acc,rel',name,args')

  and walk_ann used_names new_names ann =
    match ann with
      An_type(typ) -> An_type(walk_type used_names new_names typ)
    | An_stmt(stmt) -> An_stmt(walk_stmt used_names new_names stmt)
    | An_expr(expr) -> An_expr(walk_expr used_names new_names expr)
    | An_string(_) -> ann
    | An_list(anns) -> 
	An_list(List.map (walk_ann used_names new_names) anns)

  in main ()

(************************************************************************)

let _ = 
  register_switch "-norm"
    "-norm (bmf) -- Normalize the current BMF program repr."
    (fun top_decls args ->
      let name = Bsct_toplevel.get_arg_name (List.hd args)
      and (options,param,args') = Bsct_toplevel.parse_arg false args
      in let _ = Bsct_toplevel.ignore_options options name
      in let top_decls' = normalize_globals top_decls
      in (top_decls',args'))

let _ =
  register_switch "-out"
    ("-out FILE (bmf) -- Output the current BMF program repr. to FILE\n"^
     "	FILE = - writes to stdout.")
    (fun top_decls args ->
      let name = Bsct_toplevel.get_arg_name (List.hd args)
      and (options,param,args') = Bsct_toplevel.parse_arg true args
      in let _ = Bsct_toplevel.ignore_options options name
      in let _ =
    	if param = "-" then
	  unparse_stdout 72 top_decls
    	else
	  unparse_file param 72 top_decls
      in (top_decls,args'))

