(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
(************************************************************************)
(* Preamble								*)
(************************************************************************)

open Ssisl
open Containers

open Bmf;;
open Bmf_util;;
open Bmf_array;;

module WS = Wff.Simple

type query = {
    query_vars: (string * query_var_kind) list;
    query_used: string list;
    query_bindings: query_binding list;
    query_bounds: bmf_expr;
    query_pred: WS.t;
    query_assertion: WS.t;
    query_body: bmf_stmt
}
and query_var_kind =
      Qv_loop
    | Qv_value of scalar_type
    | Qv_flag

and query_binding = {
    qb_rel: string;
    qb_acc: bmf_expr list;
    qb_value: string;
    qb_flag: string;
}
;;

exception Not_a_valid_query;;

(************************************************************************)
(* Normalize a query							*)
(*									*)
(* Replace accesses in the body with their corrisponding value and	*)
(* flag variables							*)
(************************************************************************)

module Rel_acc_element =
  struct
    type t = string * (bmf_expr list)
    let compare = Pervasives.compare
  end
module Rel_acc_map = Map(Rel_acc_element)

let rec normalize_query (q:query) = (
    let acc_to_qb_map =
        List.fold_left
	    (fun m qb ->
		(Rel_acc_map.add (qb.qb_rel,qb.qb_acc) qb m))
	    Rel_acc_map.empty q.query_bindings
    in let normalized_body =
	replace_accs_in_stmt acc_to_qb_map q.query_body
    in let normalized_bindings =
      Sort.list (<=)
	(List.map
	    (fun qb ->
	      { qb_rel = qb.qb_rel;
		qb_acc = qb.qb_acc;
		qb_value = qb.qb_value;
		qb_flag = qb.qb_flag
	      })
	    q.query_bindings)
    in 
    { query_vars = List'.remove_duplicates q.query_vars;
      query_bindings = normalized_bindings;
      query_used = List'.remove_duplicates q.query_used;
      query_bounds = q.query_bounds;
      query_pred = q.query_pred;
      query_assertion = q.query_assertion;
      query_body = normalized_body
    }
: query)

(* Replacing access exprs, A[[F(i)]], with value and flag vars, v_123 and
    b_407 *)

and replace_accs_in_stmt acc_to_qb_map s =
    let s0 = replace_accs_in_stmt acc_to_qb_map
    and e = replace_accs_in_expr acc_to_qb_map
    and i = replace_accs_in_ident acc_to_qb_map
    in
	match s with
	  Stmt_assign(lhs,rhs) -> Stmt_assign(e lhs, e rhs)
	| Stmt_seq(first,second) -> Stmt_seq(s0 first, s0 second)
	| Stmt_call(f,lhs,args) -> Stmt_call(f, e lhs, List.map e args)
	| Stmt_for(i0,lb,ub,step,body) ->
	    Stmt_for(i0, e lb, e ub, e step, s0 body)
	| Stmt_while(pred,body) -> Stmt_while(e pred, s0 body)
	| Stmt_if(pred,thn,els) -> Stmt_if(e pred, s0 thn, s0 els)
	| Stmt_let(Frame(decls),body) ->
	    Stmt_let(Frame(List.map i decls), s0 body)
	| Stmt_do(pred,body) -> Stmt_do(e pred, s0 body)
	| Stmt_nop -> Stmt_nop
	| Stmt_directive(ann,s') -> Stmt_directive(ann, s0 s')

and replace_accs_in_expr acc_to_qb_map e =
    let e0 = replace_accs_in_expr acc_to_qb_map
    and i = replace_accs_in_ident acc_to_qb_map
    in
	match e with
	  (Expr_int(_)|Expr_real(_)|Expr_string(_)|Expr_bool(_)|Expr_void
		|Expr_var(_)) ->
	    e
	| Expr_op(kind,args) -> Expr_op(kind, List.map e0 args)
	| Expr_forall(Frame(decls),pred) ->
	    Expr_forall(Frame(List.map i decls), e0 pred)
	| Expr_exist(Frame(decls),pred) ->
	    Expr_exist(Frame(List.map i decls), e0 pred)
	| Expr_access(acc,r,name,args) ->
	    let rname =
		match r with
		  Expr_var(rname) -> rname
		| _ -> failwith
			"normalize_query: indirection in expr_access"
	    and args' = List.map e0 args
	    in let qb =
		try
		    Rel_acc_map.find (rname,args') acc_to_qb_map
		with Not_found ->
		    failwith "normalize_query: acc without query binding"
	    in
		assert (name = "")
		    (* (Failure("normalize_query: named acc in query body.")) *);
		if acc = Access_lookup then
		    Expr_var(qb.qb_flag)
		else
		    Expr_var(qb.qb_value)
	| Expr_set(Frame(decls),pred) ->
	    Expr_set(Frame(List.map i decls), e0 pred)

and replace_accs_in_ident acc_to_qb_map i =
    let e = replace_accs_in_expr acc_to_qb_map
    in
	match i with
	  Ident_valid(name,typ,stuff,ann) ->
	    let stuff' =
		match stuff with
		  Ident_var(init) -> Ident_var(e init)
		| Ident_bind -> Ident_bind
		| Ident_proc(_,_,_) -> failwith "Can't happen"
	    in
		assert (ann = An_list([])) 
		    (* (Failure("replace_accs_in_ident: ident annots not implemented")) *);
		Ident_valid(name,typ,stuff',ann)
	| Ident_null -> Ident_null

;;

(************************************************************************)
(* Query bindings							*)
(************************************************************************)

let qb2expr qb =
    Expr_op(Op_other("%qb"),[
        Expr_var(qb.qb_rel);
	Expr_op(Op_other("%qb_acc"),qb.qb_acc);
	Expr_var(qb.qb_value);
	Expr_var(qb.qb_flag)])

and expr2qb e =
    let (rel,acc,val0,flag) =
	match e with
	  Expr_op(Op_other("%qb"),[
		Expr_var(rel);
		Expr_op(Op_other("%qb_acc"),acc);
		Expr_var(val0);
		Expr_var(flag)]) ->
	    (rel,acc,val0,flag)
	| _ -> raise Not_a_valid_query
    in {
	qb_rel = rel;
	qb_acc = acc;
	qb_value = val0;
	qb_flag = flag
    }
;;

(************************************************************************)
(* Query var decls							*)
(************************************************************************)

let query_var2ident (name,kind) =
    match kind with
      Qv_loop ->
	Ident_valid(name,Type_scalar(Type_int),Ident_bind,An_list([]))
    | Qv_value(sc_ty) ->
	Ident_valid(name,Type_scalar(sc_ty),Ident_var(Expr_void),
	    An_list([]))
    | Qv_flag ->
	Ident_valid(name,Type_scalar(Type_bool),Ident_bind,An_list([]))

and ident2query_var ident =
    match ident with
      Ident_valid(name,Type_scalar(Type_int),Ident_bind,An_list([])) ->
	(name,Qv_loop)
    | Ident_valid(name,Type_scalar(sc_ty),Ident_var(Expr_void),
	    An_list([])) ->
	(name,Qv_value(sc_ty))
    | Ident_valid(name,Type_scalar(Type_bool),Ident_bind,An_list([])) ->
	(name,Qv_flag)
    | _ -> raise Not_a_valid_query
;;

(************************************************************************)
(* Queries								*)
(************************************************************************)

let stmt_is_query s =
    match s with
      Stmt_do(
	    Expr_set(Frame(_),
		Expr_exist(Frame(_),
		    Expr_op(Op_other("%query"),_))),
	    _) -> true
    | _ -> false
;;

let query2stmt query =
    let query = normalize_query query
    in let (used,unused) =
	List.fold_left
	    (fun (used,unused) decl ->
		if List.mem (bmf_ident_name decl) query.query_used then
		    (decl::used,unused)
		else
		    (used,decl::unused))
	    ([],[]) (List.map query_var2ident query.query_vars)
    in
	Stmt_do(
	    Expr_set(Frame(used),
		Expr_exist(Frame(unused),
		    Expr_op(Op_other("%query"),[
			Expr_op(Op_other("%query_bindings"),
			    List.map qb2expr query.query_bindings);
			Expr_op(Op_other("%query_bounds"),[
			    query.query_bounds]);
			Expr_op(Op_other("%query_pred"),[
			    wff_to_bmf query.query_pred]);
			Expr_op(Op_other("%query_assertion"),[
			    wff_to_bmf query.query_assertion])
		    ]))),
	    query.query_body)

and stmt2query stmt =
    let (used,unused,raw_bindings,bounds,pred,assertion,body) =
	match stmt with
	  Stmt_do(
		Expr_set(Frame(used),
		    Expr_exist(Frame(unused),
			Expr_op(Op_other("%query"),[
			    Expr_op(Op_other("%query_bindings"),
				raw_bindings);
			    Expr_op(Op_other("%query_bounds"),[
				bounds]);
			    Expr_op(Op_other("%query_pred"),[
				pred]);
			    Expr_op(Op_other("%query_assertion"),[
				assertion])
			]))),
		body) ->
	    (used,unused,raw_bindings,bounds,pred,assertion,body)
	| _ -> raise Not_a_valid_query
    in
	normalize_query {
	    query_vars = List.map ident2query_var (used@unused);
	    query_bindings = List.map expr2qb raw_bindings;
	    query_used = List.map bmf_ident_name used;
            query_bounds = bounds;
	    query_pred = bmf_to_wff pred;
            query_assertion = bmf_to_wff assertion;
	    query_body = body
	}
;;

(***********************************************************************
 *                           Useful utilities
 ***********************************************************************)

let collect_loop_vars q =
    (* preserves the order of the variables *)

    (List.fold_left
      (fun p0 p1 -> match (p0,p1) with (l, (s,Qv_loop)) -> l@[s] | (l, _) -> l)
      []
      q.query_vars)
;;

let is_pred_strong_p query r =
  let b_r =
    let qb =
      List'.find_if (fun qb -> qb.qb_rel = r) query.query_bindings
    in qb.qb_flag
  in
  WS.tautology_p 
    (WS.Implies(query.query_pred,WS.Atom(b_r)))

let is_assert_strong_p query r =
  let b_r =
    let qb =
      List'.find_if (fun qb -> qb.qb_rel = r) query.query_bindings
    in qb.qb_flag
  in
  WS.tautology_p 
    (WS.Implies(query.query_assertion,WS.Atom(b_r)))
