(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
{

open Ssisl
open Containers

open Bmf_parser

exception Lexical_error of int;;
exception No_such_parser_arg of string;;
exception Parse_arg_is_invalid_list of string;;

(* String parsing swiped from CL71's scanner.mll *)

let initial_string_buffer = String.create 256;;
let string_buff = ref initial_string_buffer;;
let string_index = ref 0;;

let reset_string_buffer () =
  string_buff := initial_string_buffer;
  string_index := 0
;;

let store_string_char c =
  (if !string_index >= String.length (!string_buff) then
    let new_buff = String.create (String.length (!string_buff) * 2) in
      String.blit new_buff 0 (!string_buff) 0 (String.length (!string_buff));
      string_buff := new_buff;
      ());
  String.set (!string_buff) (!string_index) c;
  incr string_index
;;

let get_stored_string () =
  let s = String.sub (!string_buff) 0 (!string_index) in
    string_buff := initial_string_buffer;
    s
;;

let keywords = (String_hashtbl.create 101 : token String_hashtbl.t);;

String_hashtbl.add keywords "array" ARRAY;;
String_hashtbl.add keywords "bool" BOOL;;
String_hashtbl.add keywords "void" VOID;;
String_hashtbl.add keywords "call" CALL0;;
String_hashtbl.add keywords "do" DO0;;
String_hashtbl.add keywords "else" ELSE;;
String_hashtbl.add keywords "exist" EXIST;;
String_hashtbl.add keywords "extern" EXTERN;;
String_hashtbl.add keywords "false" FALSE0;;
String_hashtbl.add keywords "for" FOR0;;
String_hashtbl.add keywords "forall" FORALL;;
String_hashtbl.add keywords "if" IF0;;
String_hashtbl.add keywords "int" INT;;
String_hashtbl.add keywords "of" OF;;
String_hashtbl.add keywords "on" ON;;
String_hashtbl.add keywords "parambegin" PARAMBEGIN;;
String_hashtbl.add keywords "paramend" PARAMEND;;
String_hashtbl.add keywords "procedure" PROCEDURE;;
String_hashtbl.add keywords "pointer" PTR;;
String_hashtbl.add keywords "real" REAL;;
String_hashtbl.add keywords "relation" RELATION;;
String_hashtbl.add keywords "sqrt" SQRT0;;
String_hashtbl.add keywords "true" TRUE0;;
String_hashtbl.add keywords "type_ann" TYPE_ANN;;
String_hashtbl.add keywords "var" VAR;;
String_hashtbl.add keywords "while" WHILE0;;


let parser_args =
    ref (String_map.empty : Bmf.annotation String_map.t);;

let is_an_stmt_list lst =
    List.for_all
	(fun an' ->
	    match an' with
	      Bmf.An_stmt(_) -> true
	    | _ -> false)
	lst
;;

let an_stmt_list_to_stmt_list lst =
    List.map
	(fun an' ->
	    match an' with
	      Bmf.An_stmt(s) -> s
	    | _ -> failwith "Not a an_stmt")
	lst
;;

let is_an_expr_list lst =
    List.for_all
	(fun an' ->
	    match an' with
	      Bmf.An_expr(_) -> true
	    | _ -> false)
	lst
;;

let an_expr_list_to_expr_list lst =
    List.map
	(fun an' ->
	    match an' with
	      Bmf.An_expr(e) -> e
	    | _ -> failwith "Not a an_expr")
	lst
;;

let is_an_string_list lst =
    List.for_all
	(fun an' ->
	    match an' with
	      Bmf.An_string(_) -> true
	    | _ -> false)
	lst
;;

let an_string_list_to_string_list lst =
    List.map
	(fun an' ->
	    match an' with
	      Bmf.An_string(str) -> str
	    | _ -> failwith "Not a an_string")
	lst
;;

}

rule token = parse
      ".." { DOTDOT }
    | "." { DOT }
    | "," { COMMA }
    | "+" { PLUS }
    | "-" { MINUS }
    | "*" { STAR }
    | "/" { DIVIDE }
    | "&" { ADDR }
    | "&&" { AND }
    | "||" { OR }
    | "!" { NOT }
    | "^" { POWER }
    | "(" { OPEN_PAREN }
    | ")" { CLOSE_PAREN }
    | "{" { OPEN_CURLY }
    | "}" { CLOSE_CURLY }
    | "[" { OPEN_SQUARE }
    | "]" { CLOSE_SQUARE }
    | ":" { COLON }
    | "|" { BAR }
    | ";" { SEMI_COLON }
    | ":=" { ASSIGN }
    | "<-" { LEFT_ARROW }
    | "==" { EQUALS }
    | "<=" { LE }
    | "<" { LTHAN }
    | ">" { GT }
    | ">=" { GE }
    | "!=" { NEQ }
    | "<<" { OPEN_ANN }
    | ">>" { CLOSE_ANN }
    | "=>" { IMPLIES }
    | "$" ['a'-'z''A'-'Z''0'-'9''_''%']+
	{
	    let str = (Lexing.lexeme lexbuf)
	    in let an =
		try
		    String_map.find str !parser_args
		with Not_found ->
		    raise (No_such_parser_arg(str))
	    in
		match an with
		  Bmf.An_type(ty) -> IN_PLACE_TYPE(ty)
		| Bmf.An_stmt(st) -> IN_PLACE_STMT(st)
		| Bmf.An_expr(ex) -> IN_PLACE_EXPR(ex)
		| Bmf.An_string(str) -> IN_PLACE_STRING(str)
		| Bmf.An_list([]) -> IN_PLACE_EMPTY_LIST
		| Bmf.An_list(lst) ->
		    if is_an_stmt_list lst then
			IN_PLACE_STMT_LIST(an_stmt_list_to_stmt_list lst)
		    else if is_an_expr_list lst then
			IN_PLACE_EXPR_LIST(an_expr_list_to_expr_list lst)
		    else if is_an_string_list lst then
			IN_PLACE_STRING_LIST(an_string_list_to_string_list lst)
		    else
			raise (Parse_arg_is_invalid_list(str))
	}
		    
    | (['1'-'9']['0'-'9']*|'0')
	{ DECIMALINT(int_of_string(Lexing.lexeme lexbuf)) }
    | (['0'-'9']+'.'['0'-'9']*|['0'-'9']*'.'['0'-'9']+)
      (['e''E'] (['-''+'])? ['0'-'9']+)?
	{ FLOAT(float_of_string(Lexing.lexeme lexbuf)) }
    | ['0'-'9']+ ['e''E'] (['-''+'])? ['0'-'9']+
	{ FLOAT(float_of_string(Lexing.lexeme lexbuf)) }
    | "%op" { OP0 }
    | ['a'-'z''A'-'Z''_''%'] ['a'-'z''A'-'Z''0'-'9''_''%']*
	{
	    let str = Lexing.lexeme lexbuf
	    in
		try
		    String_hashtbl.find keywords str
		with Not_found ->
		    IDENT(str)
	}
    | '"'
	{
	    reset_string_buffer();
	    string lexbuf;
	    STRINGVAL(get_stored_string()) 
	}
    | [' ' '\t' '\n']+ { token lexbuf }
    | '#' [^ '\n']* ('\n'|eof) { token lexbuf }
    | eof { EOF }
    | _ { raise (Lexical_error(Lexing.lexeme_start lexbuf)) }

and string = parse
      '"' { () }
    | '\\' ['\\' '"' 't' 'b' 'r' 'n']
	{
	    let c = (Lexing.lexeme_char lexbuf 1)
	    in let c' =
	        match c with
		    't' -> '\t'
		|   'b' -> '\b'
		|   'r' -> '\r'
		|   'n' -> '\n'
		|   _ -> c
	    in
		store_string_char(c');
		string lexbuf
	}
    | '\\' ['0'-'9'] ['0'-'9'] ['0'-'9']
	{
	    store_string_char(Char.chr
				(64*(Char.code(Lexing.lexeme_char lexbuf 1))
				+ 8*(Char.code(Lexing.lexeme_char lexbuf 2))
				+   (Char.code(Lexing.lexeme_char lexbuf 3))));
	    string lexbuf
	}
    | '\\' { raise (Lexical_error(Lexing.lexeme_start lexbuf)) }
    | eof { raise (Lexical_error(Lexing.lexeme_start lexbuf)) }
    | _
	{
	    store_string_char(Lexing.lexeme_char lexbuf 0);
	    string lexbuf
	}

