(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
(***********************************************************************
 *                Useful stuff for BMF
 ***********************************************************************)

open Bsct_base
open Ssisl
open Containers

open Bmf;;
open Bmf_linear;;

(***********************************************************************
 *                             Annotations
 ***********************************************************************)

let an_get_string an =
    begin
      match an with
        | (An_string s) -> s
        | _ -> failwith "an_get_string: Not a string"
    end
;;

let an_get_list an =
    begin
      match an with
        | (An_list l) -> l
        | _ -> failwith "an_get_list: Not a list"
    end
;;

let an_get_string_list an =
    (List.map an_get_string (an_get_list an))
;;

let string_list_2_an_list lst =
    An_list (List.map (fun x -> An_string x) lst)
;;

(***********************************************************************
 *                              Idents
 ***********************************************************************)

let collect_ann_4_ident id =
    match id with
      | (Ident_valid(_,_,_,ann)) -> (collect_annotations ann)
      | Ident_null -> failwith "collect_ann_4_ident: null ident"
;;

let is_valid_ident id = 
    match id with
      | (Ident_valid(_)) -> true
      | Ident_null -> false
;;

let
  is_array =
    function
      (Ident_valid(_,(Type_array(_)),_,_)) -> true
    | _ -> false
;;

let array_elt_type id =
  begin
    match id with
      (Ident_valid(_,Type_array(atyp),_,_)) ->
       	Bmf_array.array_base_type atyp
    | _ -> failwith "array_elt_type: not an array"
  end
;;

(*
  << "name1" "name2" .... >> --> [ ident_valid(name1), .... ]
*)
let an_list2index_list an =
    begin
      match an with
        | (An_list al) ->
          let an2index an =
                Ident_valid
                  ( (an_get_string an)
                  , Type_scalar Type_int
                  , Ident_bind
                  ,(An_list []))
          in (List.map an2index al)

        | a -> failwith "Not an annotation list"
    end
;;

let
  an_list2expr_list =
    function
      (An_list al) ->
        (List.map
          (function (An_expr e) -> e | _ -> failwith "an_expr expected")
          al)
    | _ -> failwith "Expecting an an_list"
;;

let
  expr_list2an_list =
    fun
      l -> An_list (List.map (fun x -> An_expr x) l)
;;

let
  var_list2an_list =
    fun
      l -> An_list (List.map (fun x -> An_string x) (List.map bmf_ident_name l))
;;

(***********************************************************************
 *                            Statements
 ***********************************************************************)

let rec
  compact_seq_stmt =
    function
      (Stmt_seq(s1,s2)) ->
        (match ((compact_seq_stmt s1),(compact_seq_stmt s2)) with
          (Stmt_nop,Stmt_nop) -> Stmt_nop
        | (Stmt_nop,s) -> s
        | (s,Stmt_nop) -> s
        | (s,t) -> Stmt_seq(s,t))

    | s -> s
;;

let 
  stmt_list2seq =
    function
      [] -> Stmt_nop
    | [s] -> s
    | (stmt::rest) -> (List.fold_left (fun s t -> Stmt_seq(s,t)) stmt rest)
;;


let rec seq2stmt_list s =
  match s with
    Stmt_seq(a,b) -> (seq2stmt_list a)@(seq2stmt_list b)
  | Stmt_nop -> []
  | s -> [s]
;;


exception Root_stmt

module Stmt_element =
  struct
    type t = bmf_stmt
    let equal = (==)
    let hash = Containers.hash
  end

module Stmt_hashtbl = Hashtbl(Stmt_element)

let parent_of_stmt (tree: bmf_stmt) = (
  let htbl = Stmt_hashtbl.create 101 in 
  recurs tree
    (fun recurs s ->
      match s with
      	Stmt_seq(l,r) ->
          Stmt_hashtbl.add htbl l s; recurs l;
          Stmt_hashtbl.add htbl r s; recurs r
      | Stmt_for(_,_,_,_,b) ->
          Stmt_hashtbl.add htbl b s; recurs b
      | Stmt_while(_,b) ->
          Stmt_hashtbl.add htbl b s; recurs b
      | Stmt_if(_,t,e) ->
          Stmt_hashtbl.add htbl t s; recurs t;
          Stmt_hashtbl.add htbl e s; recurs e
      | Stmt_let(_,b) ->
          Stmt_hashtbl.add htbl b s; recurs b
      | Stmt_do(_,b) ->
          Stmt_hashtbl.add htbl b s; recurs b
      (* Stmts with no children *)
      | (Stmt_assign(_,_) | Stmt_call(_,_,_) | Stmt_nop) ->
          ()
      | Stmt_directive(_,s') ->
	  Stmt_hashtbl.add htbl s' s; recurs s');
  (fun s ->
    if s == tree then
      raise Root_stmt
    else
      Stmt_hashtbl.find htbl s)
: bmf_stmt -> bmf_stmt)

(***********************************************************************
 *                           Expressions
 ***********************************************************************)

let collect_var_names lst =
    begin
      let rec collect l = 
          match l with
            | ((Expr_access(_,r,_,l))::rest) -> (collect (r::l@rest))
            | ((Expr_op(_,l))::rest) -> (collect (l@rest))
            | ((Expr_exist(_,p))::rest) -> 
                failwith "collect_var_names: too complicated"
            | ((Expr_forall(_,p))::rest) ->
                failwith "collect_var_names: too complicated"
            | ((Expr_var(s))::rest) -> s::(collect rest)
            | (_::rest) -> (collect rest)
            | [] -> []
      in (List'.remove_duplicates (collect lst))
    end
;;

let collect_vars scope l =
  begin
    (List.map (bmf_lookup_scope_assert scope) (collect_var_names l))
  end
;;

let rec collect_op_args op lst =
    let rec walk e =
        match e with
          Expr_op(op',lst') when op = op' -> collect_op_args op lst'
        | _ -> [e]
    in
        List'.flat_map walk lst
;;

let rec uses_var x e =
    match e with
      | (Expr_access(_,r,_,l)) -> (List.exists (uses_var x) (r::l))
      | (Expr_op(_,l)) -> (List.exists (uses_var x) l)
      | (Expr_var(s)) -> (s = x)
      | (Expr_forall(_)) -> failwith 
                            "uses_var: expression too complicated"
      | (Expr_exist(_)) -> failwith
                           "uses_var: expression too complicated"
      | _ -> false
;;

(***********************************************************************
 *                           Expressions: comparators
 ***********************************************************************)

(*
  Here is how expression comparators work:
  they take 
  - comparator for access expressions (comp_acc)
  - scope
  - expr1 and expr2
  and do the obviious thing, except that comp_acc is called for accesses.
  The important point here is that these comparators do not recurse into 
  the access arguments!!!
  The return value is -1, 0, 1 -- so that the comparators can be use din sets and
  maps.
  comp_acc must have type:  scope -> expr -> expr -> int

  IMPORTANT: it is assumed that access functions are affine functions of
             some variables
*)

exception Expr_diff of int;;

let rec compare_expr_list comp_acc lst1 lst2 =
    try
      (List.iter2
        (fun e1 e2 -> let c = (compare_expr comp_acc e1 e2)
                      in if (c <> 0) then raise (Expr_diff c)
                                     else ())
        lst1 lst2); 
      0
    with
      | (Invalid_argument "do_list2") -> ((List.length lst1) - (List.length lst2))
      | (Expr_diff c) -> c

and compare_expr comp_acc e1 e2 =
    match (e1, e2) with
      | (Expr_access(op1,r1,f1,l1), Expr_access(op2,r2,f2,l2)) ->
          (comp_acc e1 e2)
      | (Expr_var(s1),Expr_var(s2)) -> (Pervasives.compare s1 s2)
      | (Expr_op(op1,l1), Expr_op(op2,l2)) ->
          if (op1 <> op2) then (Pervasives.compare op1 op2)
          else (compare_expr_list comp_acc l1 l2)
      | _ -> failwith "compare_expr: Expression too complicated"
;;

let compare_access just_access_func e1 e2 =
    match (e1,e2) with 
      | (Expr_access(op1,r1,f1,l1),Expr_access(op2,r2,f2,l2)) ->
          let same = (just_access_func or (op1 = op2)) &&
                     (r1 = r2) &&
                     (f1 = f2)
          in 
             if same then
                let x1 = (collect_var_names l1)
                and x2 = (collect_var_names l2)
                in let x = x1 @ x2
                in let acc1 = (expr_list2affine_map x l1)
                and acc2 = (expr_list2affine_map x l2)
                in (compare_affine_maps acc1 acc2)
             else
              (Pervasives.compare (op1,r1,f1) (op2,r2,f2))
      | _ -> failwith "compare_access: must have expr_access"
;;

let compare_expr_just_access =
    compare_expr (compare_access true) 
;;

let compare_expr_strict =
    compare_expr (compare_access false)
;;

let rec is_subexpr comp e1 e2 =
    if (0 = (comp e1 e2)) then true
    else
      match e2 with
        | (Expr_access(op,r,f,l)) -> (List.exists (is_subexpr comp e1) (r::l))
        | (Expr_op(op,l)) -> (List.exists (is_subexpr comp e1) l)
        | (Expr_var(s)) -> false (* must have compared to 0 *)
        | _ -> failwith "is_subexpr: Expression too complicated"
;;

let collect_accesses l =
    let rec collect l = 
        match l with
          | ((Expr_access(_,_,_,_) as a)::rest) -> a :: (collect rest)
          | ((Expr_op(_,l))::rest) -> (collect (l@rest))
          | ((Expr_exist(_,p))::rest) -> 
              failwith "collect_accesses: Expression too complicated"
          | ((Expr_forall(_,p))::rest) ->
              failwith "collect_accesses: Expression too complicated"
          | (_::rest) -> (collect rest)
          | [] -> []
        in
          (List'.long_remove_duplicates
	     (fun x y -> (compare_expr (compare_access true)) x y = 0)
	     identity (collect l))
;;

        

(***********************************************************************
 *                      Expressions: renaming
 ***********************************************************************)

let rename_in_expr names1 names2 e =
    if ((List.length names1) <> (List.length names2)) then
      failwith "rename_in_expr: Must have name lists of same length"
    else
      let ren = (List.fold_left2 
                    (fun m x y -> (String_map.add x y m))
                    String_map.empty 
                    names1
                    names2)
      in let rec rename = 
        (function
          (Expr_access(op,r,fdep,l)) ->
            Expr_access(op,(rename r),fdep,(List.map rename l))
    
        | (Expr_op(op,l)) ->
            Expr_op(op,(List.map rename l))
    
        | (Expr_var(s) as e) ->
            (match (find_value (fun () -> String_map.find s ren)) with
              (Found s2) -> Expr_var(s2)
            | Unfound -> e)

        | (Expr_forall(_,p)) -> failwith "rename_in_expr: too complicated"
        | (Expr_exist(_,p)) -> failwith "rename_in_expr: too complicated"
        | (Expr_set(_,p)) -> failwith "rename_in_expr: too complicated"
        | e -> e)
    
      in (rename e)
;;

let verify_scope scope msg =
  begin
    (List.map
      (function (Ident_valid(_,_,_,_) as i) -> i | Ident_null -> failwith ("verify_scope: " ^ msg))
      scope)
  end
;;

(***********************************************************************
 *                   Expressions: simplification
 ***********************************************************************)


(* only works for conjuncts !! *)
(* x is the set of "protected" variables into which the other
   variables should be renamed, if possible
*)
(*
let eliminate_simple_equalities scope x pred =
  begin
    let     ind_vars = (collect_vars scope [pred])
    in let  ind_var_names = (map bmf_ident_name ind_vars)
    and     xnames = (map bmf_ident_name x)
    in let  u = (unionfind__init compare_strings (xnames @ ind_var_names))
    in let  subst_name a = 
      begin
        if (mem a xnames) then a
        else (unionfind__find u a)
      end
    in let  rename_var v = 
      begin
        match v with
          | (ident_valid(name,t,stuff,a)) -> 
              ident_valid((subst_name name),t,stuff,a)
          | ident_null -> failwith "eliminate_simple_equalities: ident_null"
      end
    and     compare a b = 
      begin
        match ((mem a xnames),(mem b xnames)) with
          | (true,true) -> (compare_strings a b)
          | (false,false) -> (compare_strings a b)
          | (true,false) -> (-1)
          | (false,true) -> 1
      end
    in let  {EQ=a; LEQ=b; NL=c} = (pred2conjunct pred)
    in let process_equality_affine_form l a =
      begin
        if (0 = (get_const4affine_form a)) then
          match (get_term_list a) with
            | [term(xx,1) ; term(yy,(-1))] -> ( (unionfind__union u xx yy) ; l )
            | [term(xx,(-1)) ; term(yy,1)] -> ( (unionfind__union u xx yy) ; l )
            | _ -> a::l
        else
          a::l
      end
    in let  a2 = (it_list process_equality_affine_form [] a)
        (* a2 has the equalities removed *)

    in let () = (unionfind__set_least_representatives u compare)
        (* this makes sure that vars are renamed into protected vars *)

    in let collect_protected_eqs l xx =
      begin
        let yy = (unionfind__find u xx)
        in 
          if (xx <> yy) then
            (affine_add 
              (new_affine_form (builder_var(1,xx)))
              (new_affine_form (builder_var((-1),yy)))) :: l
          else
            l
      end
    in let a_prot = (it_list collect_protected_eqs [] xnames)
        (* a_prot now has all equalities involving protected vars *)

    in let  ind_vars_new = (map rename_var ind_vars)
    in (rename_in_expr ind_vars ind_vars_new (conjunct2pred scope {EQ=(a2@a_prot);LEQ=b;NL=c}))
  end
;;
*)

(*
let reduce_exist_vars scope e =
  begin
    let e2 =
    ( (print_string "Before reduce\n")
    ; (bmf_printer__expr_printer e)
    ; (print_string "\n")
    ; (flush stdout)
    ; (match e with
      | (expr_set(frame x, expr_exist(frame y, pred))) ->
          let     pred2 = eliminate_simple_equalities (y@x@scope) (x@scope) pred
          in let  z = (collect_vars (y@x@scope) [pred2])
          in let  new_y = (list__subtract z (x@scope))
          in (expr_set(frame x,expr_exist(frame new_y, pred2)))

      | (expr_set(_,_)) -> e
      | _ -> failwith "reduce_exist_vars: Only expr_set allowed")
    )
    in
    ( (print_string "After reduce\n")
    ; (bmf_printer__expr_printer e2)
    ; (print_string "\n")
    ; e2
    )
  end
;;
*)

(*
let reduce_exist_vars scope e =
  begin
    match e with
      | (expr_set(frame x, expr_exist(frame y, pred))) ->
          let     pred2 = eliminate_simple_equalities (y@x@scope) (x@scope) pred
          in let  z = (collect_vars (y@x@scope) [pred2])
          in let  new_y = (list__subtract z (x@scope))
          in (expr_set(frame x,expr_exist(frame new_y, pred2)))
    
      | (expr_set(_,_)) -> e
      | _ -> failwith "reduce_exist_vars: Only expr_set allowed"
  end
;;
*)

(***********************************************************************
 *			   Expresions <-> WFF's
 ***********************************************************************)

module WS = Wff.Simple

let rec wff_to_bmf (e: WS.t) = ((
    match e with
      WS.Conjunct(e1,e2) ->
	Expr_op(Op_and,[wff_to_bmf e1; wff_to_bmf e2])
    | WS.Disjunct(e1,e2) ->
	Expr_op(Op_or,[wff_to_bmf e1; wff_to_bmf e2])
    | WS.Implies(e1,e2)  ->
	Expr_op(Op_implies,[wff_to_bmf e1; wff_to_bmf e2])
    | WS.Negation(e) -> Expr_op(Op_not,[wff_to_bmf e])
    | WS.Equality(e1,e2) ->
	Expr_op(Op_eq,[wff_to_bmf e1; wff_to_bmf e2])
    | WS.Atom(s) -> Expr_var(s)
    | WS.Constant(b) -> Expr_bool(b)
) : bmf_expr);;

exception Not_a_wff of bmf_expr;;

let rec bmf_to_wff (e: bmf_expr) = ((
    match e with
      Expr_op(Op_and,[e]) -> bmf_to_wff e
    | Expr_op(Op_and,e1::es) ->
	WS.Conjunct(bmf_to_wff e1,
	    bmf_to_wff (Expr_op(Op_and,es)))
    | Expr_op(Op_or,[e]) -> bmf_to_wff e
    | Expr_op(Op_or,e1::es) ->
	WS.Disjunct(bmf_to_wff e1,
	    bmf_to_wff (Expr_op(Op_or,es)))
    | Expr_op(Op_not,[e]) ->
	WS.Negation(bmf_to_wff e)
    | Expr_op(Op_eq,[e1;e2]) ->
	WS.Equality(bmf_to_wff e1, bmf_to_wff e2)
    | Expr_op(Op_implies,[e1;e2]) ->
	WS.Implies(bmf_to_wff e1, bmf_to_wff e2)
    | Expr_var(s) -> WS.Atom(s)
    | Expr_bool(b) -> WS.Constant(b)
    | _ -> raise (Not_a_wff(e))
) : WS.t);;

(***********************************************************************
 *                       Statements: uniquely rename
 ***********************************************************************)

(************************************************************************

Rename identifiers within a file or a single procedure such that the same
name is declared only once within each procedure, regardless of scope.

A variable that does not have a local declaration is assumed to be global
and is not renamed.

Identifers that appear within stmts, exprs, or types within annotations are
renamed. Annotation strings are not modified.

Also, detect multiply declarations for the same name in the same frame.

************************************************************************)

(* Entry points *)
let rec uniquely_rename_within_file decls =
  List.map uniquely_rename_within_proc decls
    
and uniquely_rename_within_proc decl =
  let m = String_map.empty in
  u_r_ident m m decl
      
(***************************************************************************)

(* Functions to add names of the map *)
and u_r_add_idents m ids =
  (* Check for duplic decls *)
  let _ =
    List.fold_left
      (fun s id ->
	match id with
	  Ident_valid(name,typ,stuff,ann) ->
	    let name = bmf_ident_name id in
	    if String_set.mem name s then 
	      failwith ("Duplicate decl: "^name);
	    String_set.add name s
	| Ident_null -> s)
      String_set.empty ids
  in
  (* Add the names and renames to m, the renaming map. *)
  List.fold_left
    (fun m id ->
      match id with
	Ident_valid(name,typ,stuff,ann) ->
	  let name = bmf_ident_name id in
	  String_map.add name (Misc.gensym name) m
      |	Ident_null -> m)
    m ids
    
(* Maps an old name to its new name. *)
and u_r_name m name =
  try
    String_map.find name m
  with Not_found ->
    (* Must be a global. Don't rename. *)
    name
      
(***************************************************************************)
      
(* Routines that do the renaming *)
and u_r_stmt m (st : bmf_stmt) = (
  let e = u_r_expr
  and es = u_r_exprs
  and s = u_r_stmt
  and is = u_r_idents
  and i = u_r_ident
  and a = u_r_annot
  in
  match st with
    Stmt_assign(lhs,rhs) -> Stmt_assign(e m lhs,e m rhs)
  | Stmt_seq(x,y) -> Stmt_seq(s m x,s m y)
  | Stmt_call(f,r,args) -> Stmt_call(e m f,e m r,es m args)
  | Stmt_for(i0,l,u,step,b) ->
      let m' = u_r_add_idents m [i0] in
      Stmt_for(i m m' i0,e m l,e m u,e m step,s m' b)
  | Stmt_while(p,b) -> Stmt_while(e m p,s m b)
  | Stmt_if(p,t,e0) -> Stmt_if(e m p,s m t,s m e0)
  | Stmt_let(Frame(ids),b) -> 
      let m' = u_r_add_idents m ids in
      Stmt_let(Frame(is m m' ids),s m' b)
  | Stmt_do(Expr_set(Frame(ids),p),b) ->
      let m' = u_r_add_idents m ids in
      Stmt_do(Expr_set(Frame(is m m' ids),e m' p),s m' b)
  | Stmt_do(p,b) -> assert false
  | Stmt_nop -> st
  | Stmt_directive(ann,body) -> Stmt_directive(a m ann,s m body)
: bmf_stmt)

(***************************************************************************)

and u_r_exprs m (es : bmf_expr list) = List.map (u_r_expr m) es

and u_r_expr m (ex : bmf_expr) = (
  let e = u_r_expr
  and es = u_r_exprs
  and i = u_r_ident
  and is = u_r_idents
  and a = u_r_annot
  and n = u_r_name
  in
  match ex with
    (Expr_int(_)|Expr_real(_)|Expr_string(_)|Expr_bool(_)|Expr_void)
    -> ex
  | Expr_var(v) -> Expr_var(n m v)
  | Expr_op(op,args) -> Expr_op(op,es m args)
  | Expr_forall(Frame(ids),p) ->
      let m' = u_r_add_idents m ids in
      Expr_forall(Frame(is m m' ids),e m' p)
  | Expr_exist(Frame(ids),p) ->
      let m' = u_r_add_idents m ids in
      Expr_exist(Frame(is m m' ids),e m' p)
  | Expr_access(ac,r,mthd,args) -> 
      Expr_access(ac,e m r,mthd,es m args)
  | Expr_set(Frame(ids),p) ->
      let m' = u_r_add_idents m ids in
      Expr_set(Frame(is m m' ids), e m' p)
: bmf_expr)
    
(***************************************************************************)

and u_r_types m (ts : bmf_type list) = List.map (u_r_type m) ts

and u_r_type m (ty : bmf_type) = (
  let t = u_r_type
  and ts = u_r_types
  and a = u_r_annot
  in
  match ty with
    (Type_scalar(_)|Type_unknown) -> ty
  | Type_array(atyp) -> Type_array(u_r_array_type m atyp)
  | Type_proc(args,res) -> Type_proc(ts m args,t m res)
  | Type_ptr(btyp) -> Type_ptr(t m btyp)
  | Type_ann(an) -> Type_ann(a m an)
: bmf_type)

and u_r_array_type m (atyp : array_type) = (
  let e = u_r_expr
  and es = u_r_exprs
  and t = u_r_type
  and is = u_r_idents
  and a = u_r_annot
  in
  match atyp with
    Array_dense(lb,ubs,maj,base_sc) -> Array_dense(lb,es m ubs,maj,base_sc)
  | Array_bb(dim,base_sc,bb_name,bb_args) ->
      Array_bb(dim,base_sc,bb_name,
	       List.map (fun (tag,val0) -> (tag,a m val0))
		 bb_args)
  | Array_view(Frame(out_frm),Frame(in_frm),constr,base_atyp) ->
      let m' = u_r_add_idents m (out_frm@in_frm) in
      Array_view(Frame(is m m' out_frm),Frame(is m m' in_frm),
		 e m' constr,u_r_array_type m base_atyp)

: array_type)

(***************************************************************************)

and u_r_idents m m' is = List.map (u_r_ident m m') is

(* m is the scope that surrounds id (we'll call this the outer scope).
   m' is the scope that contains the new name of id (we'll call this the
   inner scope). *)

and u_r_ident m m' (id : bmf_ident) = (
  let t = u_r_type
  and e = u_r_expr
  and is = u_r_idents
  and i = u_r_ident
  and s = u_r_stmt
  in
  match id with
  | Ident_valid(name,typ,stuff,annot) ->
      (* m' has the rename for name. *)
      let name' = u_r_name m' name
      (* Annotations are scoped in the outer scope (<--- ??? ) *)
      and annot' = u_r_annot m annot
      in let (typ',stuff') =
	(match stuff with
	  Ident_proc(Frame(args),res,body) ->
	    (* The Ident_proc and typ are scoped _inside_ the parameter 
	       declarations. *) 
	    let m'' = u_r_add_idents m' (res::args) in
	    (t m'' typ,Ident_proc(Frame(is m'' m'' args),
				  i m'' m'' res, s m'' body))
	| Ident_var(init) ->
	    (* var types and init expres are in the _outer_ scope. *)
	    (t m typ, Ident_var(e m init))
	| Ident_bind ->
	    (* binding types are in the _outer_ scope. *)
	    (t m typ, Ident_bind)
	  )
      in 
      Ident_valid(name',typ',stuff',annot')
  | Ident_null -> id
: bmf_ident)

(***************************************************************************)

and u_r_annots m ans = List.map (u_r_annot m) ans

and u_r_annot m (an : annotation) = (
  match an with
    An_type(ty)-> An_type(u_r_type m ty)
  | An_stmt(st) -> An_stmt(u_r_stmt m st)
  | An_expr(ex) -> An_expr(u_r_expr m ex)
  | An_list(ans) -> An_list(u_r_annots m ans)
  (* Don't rename strings! *)
  | An_string(_) -> an
: annotation)

;;

let _ =
  Sx.define_prim1 (Scheme.interaction_environment ())
    "uniquely_rename"
    (fun sx -> 
      let bmf = Bmf.from_s_expr sx in
      let bmf' = uniquely_rename_within_file bmf in
      let sx' = Bmf.to_s_expr bmf' in
      sx')
    ;;

(************************************************************************)
(* Expression simplification						*)
(************************************************************************)

let rec bmf_expr_simplify e =
    let e0 = bmf_expr_simplify
    in let rec make_nary_expr op identity f args =
	let args' = List'.flat_map (bmf_expr_args_by_op op) args
	in let args'' = List'.remove identity  args'
	in
	    match f args'' with
	      [] -> identity
	    | [arg] -> arg
	    | args''' -> Expr_op(op,args''')
    and make_neg e =
	match e with
	  Expr_op(Op_neg,[e']) -> e'
	| Expr_op(Op_plus,args) -> Expr_op(Op_plus, List.map make_neg args)
	| Expr_int(x) -> Expr_int(-x)
	| Expr_real(x) -> Expr_real(-.x)
	| _ -> Expr_op(Op_neg,[e])
    and make_not e =
	match e with
	  Expr_op(Op_not,[e']) -> e'
	| _ -> Expr_op(Op_not,[e])
    and make_plus e =
	make_nary_expr Op_plus (Expr_int(0))
	    (fun lst ->
		let (i,r,lst') =
		    List.fold_right
			(fun e (i,r,lst') ->
			    match e with
			      Expr_int(i') -> (i+i',r,lst')
			    | Expr_real(r') -> (i,r+.r',lst')
			    | e -> (i,r,e::lst'))
			lst (0,0.0,[])
		in let lst'' =
		    match (i,r) with
		      (0,0.0) -> lst'
		    | (i,0.0) -> (Expr_int(i))::lst'
		    | (i,r) -> (Expr_real((float i)+.r))::lst'
		in
		    lst'')
	    e
    and make_times e =
	make_nary_expr Op_times (Expr_int(1))
	    (fun lst ->
		let (i,r,lst') =
		    List.fold_right
			(fun e (i,r,lst') ->
			    match e with
			      Expr_int(i') -> (i*i',r,lst')
			    | Expr_real(r') -> (i,r*.r',lst')
			    | e -> (i,r,e::lst'))
			lst (1,1.0,[])
		in let lst'' =
		    match (i,r) with
		      ((0,_)|(_,0.0)) -> [Expr_int(0)]
		    | (1,1.0) -> lst'
		    | (i,1.0) -> (Expr_int(i))::lst'
		    | (i,r) -> (Expr_real((float i)*.r))::lst'
		in
		    lst'')
	    e
    and make_and e =
	make_nary_expr Op_and (Expr_bool(true))
	    (fun lst ->
		let (b,lst') =
		    List.fold_right
			(fun e (b,lst') ->
			    match e with
			      Expr_bool(b') -> (b'&&b,lst')
			    | e -> (b,e::lst'))
			lst (true,[])
		in let lst'' =
		    if b then
			lst'
		    else
			[Expr_bool(false)]
		in
		    lst'')
	    e
    and make_or e =
	make_nary_expr Op_or (Expr_bool(false))
	    (fun lst ->
		let (b,lst') =
		    List.fold_right
			(fun e (b,lst') ->
			    match e with
			      Expr_bool(b') -> (b'&&b,lst')
			    | e -> (b,e::lst'))
			lst (false,[])
		in let lst'' =
		    if (not (b)) then
			lst'
		    else
			[Expr_bool(true)]
		in
		    lst'')
	    e
    in
	match e with
	  (Expr_int(_)|Expr_real(_)|Expr_string(_)|Expr_bool(_)|
		Expr_void|Expr_var(_)) ->
	    e
	| Expr_forall(frm,pred) -> Expr_forall(frm, e0 pred)
	| Expr_exist(frm,pred) -> Expr_exist(frm, e0 pred)
	| Expr_set(frm,pred) -> Expr_set(frm, e0 pred)
	| Expr_access(acc,r,name,args) ->
	    Expr_access(acc, e0 r, name, List.map e0 args)
	| Expr_op(op,args) ->
	    let args' = List.map e0 args
	    in let sort_by_op op e =
		match e with
		  Expr_op(op',args) when op = op' ->
		    Expr_op(op,Sort.list ( (<)) args)
		| _ -> e
	    in
		match op with
		  Op_neg ->
		    (match args' with
		      [e] -> make_neg e
		    | _ -> failwith "bmf_expr_simplfy -- Isn't unary")
		| Op_plus -> sort_by_op Op_plus (make_plus args')
		| Op_times -> sort_by_op Op_times (make_times args')
		| Op_not ->
		    (match args' with
		      [e] -> make_not e
		    | _ -> failwith "bmf_expr_simplfy -- Isn't unary")
		| Op_and -> sort_by_op Op_and (make_and args')
		| Op_or -> sort_by_op Op_or (make_or args')
		| _ -> Expr_op(op,args')
;;

(************************************************************************)
(* Finding free vars in various forms.					*)
(************************************************************************)

let empty_s = String_set.empty
let singleton_s name = String_set.add name empty_s
let reduce_s f lst =
  List.fold_left 
    (fun result x -> String_set.union result (f x))
    empty_s lst
    
let rec remove_scoped ids s =
  match ids with
    [] -> s
  | Ident_null::ids' -> remove_scoped ids' s
  | Ident_valid(name,_,_,_)::ids' -> 
      remove_scoped ids' (String_set.remove name s)

(* Compute the free variables in various ast types.
   This should probably be in bmf.ml... *)
let rec stmt_fvs st =
  match st with
    Stmt_assign(lhs, rhs) -> 
      String_set.union (expr_fvs lhs) (expr_fvs rhs)
  | Stmt_seq(first,second) -> 
      String_set.union (stmt_fvs first) (stmt_fvs second)
  | Stmt_call(f,res,args) -> reduce_s expr_fvs (f::res::args)
  | Stmt_for(id,lb,ub,step,body) ->
      remove_scoped [id]
	(String_set.union (reduce_s expr_fvs [lb;ub;step])
	   (String_set.union (ident_fvs id) (stmt_fvs body)))
  | Stmt_while(pred,body) -> 
      String_set.union (expr_fvs pred) (stmt_fvs body)
  | Stmt_if(test,thn,els) ->
      String_set.union (expr_fvs test) 
	(String_set.union (stmt_fvs thn) (stmt_fvs els))
  | Stmt_let(Frame(ids),body) ->
      remove_scoped ids
	(String_set.union (idents_fvs ids) (stmt_fvs body))
  | Stmt_do(Expr_set(Frame(ids),pred),body) ->
      remove_scoped ids
	(String_set.union (idents_fvs ids)
 	   (String_set.union (expr_fvs pred) (stmt_fvs body)))
  | Stmt_do _ -> assert false
  | Stmt_nop -> empty_s
  | Stmt_directive(ann,body) -> 
      String_set.union (ann_fvs ann) (stmt_fvs body)
and expr_fvs e =
  match e with
  | (Expr_int _ | Expr_real _ | Expr_string _ | Expr_bool _ | 
    Expr_void ) -> empty_s
  | Expr_var(name) -> singleton_s name 
  | Expr_op(_,args) -> reduce_s expr_fvs args
  | Expr_forall(Frame(ids),pred) -> 
      remove_scoped ids
	(String_set.union (idents_fvs ids) (expr_fvs pred))
  | Expr_exist(Frame(ids),pred) ->
      remove_scoped ids
	(String_set.union (idents_fvs ids) (expr_fvs pred))
  | Expr_set(Frame(ids),pred) ->
      remove_scoped ids
	(String_set.union (idents_fvs ids) (expr_fvs pred))
  | Expr_access(_,a,_,args) -> 
      String_set.union (expr_fvs a) (reduce_s expr_fvs args)
and type_fvs (ty: bmf_type) =
  match ty with
    (Type_scalar _ | Type_unknown)-> empty_s
  | Type_proc(args_tys,res_ty) -> reduce_s type_fvs (res_ty::args_tys)
  | Type_array(atyp) -> atype_fvs atyp
  | Type_ptr(btyp) -> type_fvs btyp
  | Type_ann(ann) -> ann_fvs ann
and atype_fvs (atyp: array_type) =
  match atyp with
    Array_dense(_,ubs,_,_) -> reduce_s expr_fvs ubs
  | Array_bb(_,_,_,args) -> 
      reduce_s ann_fvs (List.map snd args)
  | Array_view(Frame(out_ids),Frame(in_ids),constrs,atyp') ->
      String_set.union (reduce_s ident_fvs (out_ids@in_ids))
      (String_set.union (expr_fvs constrs) (atype_fvs atyp'))
and ann_fvs (an: annotation) =
  match an with
    An_type(typ) -> type_fvs typ
  | An_stmt(s) -> stmt_fvs s
  | An_expr(e) -> expr_fvs e
  | An_string _ -> empty_s
  | An_list(ans) -> reduce_s ann_fvs ans
and idents_fvs (ids: bmf_ident list) =
  reduce_s ident_fvs ids
and ident_fvs (id: bmf_ident) =
  match id with
    Ident_null -> empty_s
  | Ident_valid(name,typ,stuff,an) ->
      let results_stuff =
	(match stuff with
	  Ident_var(init) -> expr_fvs init
	| Ident_bind -> empty_s
	| Ident_proc(Frame(arg_ids),res_id,body) ->
	    remove_scoped (res_id::arg_ids)
	      (String_set.union (reduce_s ident_fvs (res_id::arg_ids)) 
		 (stmt_fvs body)))
      in
      String_set.union (type_fvs typ) 
	(String_set.union results_stuff (ann_fvs an))
;;

(* *****************************************************************
   Place annotations for substitutions.

   pairs - list of (x,y) substitution pairs.
   body - stmt in which the substitutions are to be done.
   ***************************************************************** *)

let place_substitutions pairs body =
  let ids = 
    List.map
      (fun (x,y) ->
	Ident_valid(x,Type_scalar(Type_int),Ident_var(y),
		    An_list([An_string("%subst")])))
      pairs
  in
  Stmt_let(Frame(ids),body)
;;

(* *****************************************************************
   Perform substitutions.

   Looks for let-bindings of the form,
     { var x : real := y << "%subst" >>; body }
   and replaces them with
     body[x |-> y]
   ***************************************************************** *)

let rec perform_substitutions_in_proc id =
  match id with
    Ident_valid(name,typ,Ident_proc(args,res,body),ann) ->
      let body' = perform_substitutions_in_stmt body
      in Ident_valid(name,typ,Ident_proc(args,res,body'),ann)
  | _ -> id

and perform_substitutions_in_stmt st =
  stmt_subst String_map.empty st

and stmt_subst m st =
  let e = expr_subst m
  and es = exprs_subst m
  and s = stmt_subst m
  and i = ident_subst m
  and a = ann_subst m
  in
  match st with
    Stmt_assign(lhs,rhs) -> Stmt_assign(e lhs,e rhs)
  | Stmt_seq(s1,s2) -> Stmt_seq(s s1,s s2)
  | Stmt_call(f,r,args) -> Stmt_call(e f,e r,es args)
  | Stmt_for(id,lb,ub,step,body) ->
      Stmt_for(i id,e lb,e ub,e step,s body)
  | Stmt_while(pred,body) -> Stmt_while(e pred,s body)
  | Stmt_if(test,thn,els) -> Stmt_if(e test,s thn,s els)
  | Stmt_do(pred,body) -> Stmt_do(e pred,s body)
  | Stmt_nop -> st
  | Stmt_directive(ann,body) -> Stmt_directive(a ann,s body)
  | Stmt_let(Frame(ids),body) ->
      let ids' = idents_subst m ids in
      let (m',ids'') =
      	List.fold_right
	  (fun id (m',ids'') ->
	    match id with
	      Ident_valid(x,_,Ident_var(y),
			  An_list([An_string("%subst")])) 
	      ->
		(String_map.add x y m', ids'')
	    | _ -> (m',id::ids''))
	  ids' (m,[])
      in 
      let body' = stmt_subst m' body in
      if ids'' = [] then
	body'
      else
	Stmt_let(Frame(ids''),body')

and exprs_subst m exs = List.map (expr_subst m) exs
and expr_subst m ex = bmf_expr_simplify (expr_subst' m ex)

and expr_subst' m ex =
  let e = expr_subst' m
  and es = List.map (expr_subst' m)
  and is = idents_subst m
  in
  match ex with
    (Expr_int _ | Expr_real _ | Expr_string _ | Expr_bool _ | Expr_void) ->
      ex
  | Expr_var(x) ->
    (try
	let y = String_map.find x m in
	e y
      with Not_found ->
	ex)
  | Expr_op(op,args) -> Expr_op(op,es args)
  | Expr_forall(Frame(ids),pred) -> Expr_forall(Frame(is ids),e pred)
  | Expr_exist(Frame(ids),pred) -> Expr_exist(Frame(is ids),e pred)
  | Expr_set(Frame(ids),pred) -> Expr_set(Frame(is ids),e pred)
  | Expr_access(acc,a,name,args) -> Expr_access(acc,e a,name,es args)

and idents_subst m ids = List.map (ident_subst m) ids
and ident_subst m id =
  match id with
    Ident_valid(name,typ,stuff,ann) ->
      Ident_valid(name,type_subst m typ,stuff_subst m stuff,
		  ann_subst m ann)
  | Ident_null -> id
and stuff_subst m stuff =
  match stuff with
    Ident_proc _ -> assert false (* no nested procedures. *)
  | Ident_var(init) -> Ident_var(expr_subst m init)
  | Ident_bind -> Ident_bind

and type_subst m ty =
  match ty with
    (Type_scalar _ |Type_unknown) -> ty
  | Type_proc _ -> assert false (* no nested procedures. *)
  | Type_array(atyp) -> Type_array(atype_subst m atyp)
  | Type_ptr(btyp) -> Type_ptr(type_subst m btyp)
  | Type_ann(ann) -> Type_ann(ann_subst m ann)

and atype_subst m atyp =
  match atyp with
    Array_dense(lb,ubs,maj,sc) -> 
      Array_dense(lb,exprs_subst m ubs,maj,sc)
  | Array_bb(dim,base,name,args) ->
      Array_bb(dim,base,name,
	       List.map (fun (tag,val0) -> (tag,ann_subst m val0))
		 args)
  | Array_view(Frame(out_ids),Frame(in_ids),constr,base_atyp) ->
      Array_view(Frame(idents_subst m out_ids),
		 Frame(idents_subst m in_ids),
		 expr_subst m constr,
		 atype_subst m base_atyp)

and ann_subst m ann =
  match ann with
    An_type(ty) -> An_type(type_subst m ty)
  | An_stmt(st) -> An_stmt(stmt_subst m st)
  | An_expr(ex) -> An_expr(expr_subst m ex)
  | An_string _ -> ann
  | An_list(lst) -> An_list(List.map (ann_subst m) lst)


let _ =
  Sx.define_prim1 (Scheme.interaction_environment ())
    "perform_substitutions"
    (fun sx -> 
      let bmf = Bmf.from_s_expr sx in
      let bmf' = List.map perform_substitutions_in_proc bmf in
      let sx' = Bmf.to_s_expr bmf' in
      sx')
