(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
(* Case-sensitive Scheme *)
module Scheme = Sim.Make(Sim.Case_sensitive)

(************************************************************************)
(* S-expr's								*)
(************************************************************************)

module type Sx_type =
  sig
    type t =
       	Nil
      |	Bool of bool
      | Char of char
      | Float of float
      | Int of int
      | String of string
      | Symbol of string
      | Pair of t * t
      | Vector of t array
      | Other of Scheme.t
    exception Syntax_error
    val from_scheme : Scheme.t -> t
    val to_scheme : t -> Scheme.t
    val from_list : t list -> t
    val to_list : t -> t list
    val list_p : t -> bool
    val untag_symbol : t -> string
    val untag_symbol_list : t -> string list
    val tag_symbol : string -> t
    val tag_symbol_list : string list -> t
    val read : unit -> t
    val write : t -> unit 
    val pretty_print : t -> unit
    val to_string : t -> string
    val to_string_pretty : t -> string
    val define_prim0 : Scheme.env_type -> string -> (unit -> t) -> unit
    val define_prim1 : Scheme.env_type -> string -> (t -> t) -> unit
    val define_prim2 : Scheme.env_type -> string -> (t -> t -> t) -> unit
    val define_prim3 : Scheme.env_type -> string -> (t -> t -> t -> t) -> unit
    val define_prim4 : 
	Scheme.env_type -> string -> (t -> t -> t -> t -> t) -> unit
    val define_prim5 : 
	Scheme.env_type -> string -> (t -> t -> t -> t -> t -> t) -> unit
    val define_primn : Scheme.env_type -> string -> (t -> t) -> unit
  end

module Sx =
  struct
    type t =
	Nil
      |	Bool of bool
      | Char of char
      | Float of float
      | Int of int
      | String of string
      | Symbol of string
      | Pair of t * t
      | Vector of t array
      | Other of Scheme.t

    exception Syntax_error

    let rec from_scheme x =
      if x = Scheme.nil then
	Nil
      else if Scheme.bool_p x then
	Bool(Scheme.untag_bool x)
      else if Scheme.char_p x then
	Char(Scheme.untag_char x)
      else if Scheme.float_p x then
	Float(Scheme.untag_float x)
      else if Scheme.int_p x then
	Int(Scheme.untag_int x)
      else if Scheme.string_p x then
	String(Scheme.untag_string x)
      else if Scheme.symbol_p x then
	Symbol(Scheme.symbol_to_string x)
      else if Scheme.pair_p x then
	Pair(from_scheme (Scheme.car x), from_scheme (Scheme.cdr x))
      else if Scheme.vector_p x then
	Vector(Array.map from_scheme (Scheme.untag_vector x))
      else
	Other(x)

    let rec to_scheme x =
      match x with
	Nil -> Scheme.nil
      |	Bool(b) -> Scheme.tag_bool b
      |	Char(c) -> Scheme.tag_char c
      |	Float(f) -> Scheme.tag_float f
      |	Int(i) -> Scheme.tag_int i
      |	String(s) -> Scheme.tag_string s
      |	Symbol(s) -> Scheme.intern s
      |	Pair(x,y) -> Scheme.cons (to_scheme x) (to_scheme y)
      |	Vector(v) -> Scheme.tag_vector (Array.map to_scheme v)
      |	Other(x) -> x

    let rec from_list lst =
      match lst with
	[] -> Nil
      |	x::lst' -> Pair(x, from_list lst')

    let rec to_list x =
      match x with
	Nil -> []
      |	Pair(car,cdr) -> car::(to_list cdr)
      |	_ -> raise Syntax_error

    let rec list_p x =
      match x with
	Nil -> true
      |	Pair(car,cdr) -> list_p cdr
      |	_ -> false

    let untag_symbol e =
      match e with
    	Symbol(x) -> x
      | _ -> raise Syntax_error

    let untag_symbol_list x = (List.map untag_symbol (to_list x))

    let tag_symbol str = Symbol(str)

    let tag_symbol_list lst = (from_list (List.map tag_symbol lst))

    let read () = 
      let x = Scheme.read () in
      if Scheme.eq_p x Scheme.eof then
	raise End_of_file
      else 
	from_scheme x
    let write x = Scheme.write (to_scheme x)
    let pretty_print x = Scheme.pretty_print (to_scheme x)

    let to_string x =
      let (port,getter) = Scheme.make_output_port_to_string () in
      Scheme.with_output_to_port port
	(fun () -> Scheme.write (to_scheme x));
      getter ()

    let to_string_pretty x =
      let (port,getter) = Scheme.make_output_port_to_string () in
      Scheme.with_output_to_port port
	(fun () -> Scheme.pretty_print (to_scheme x));
      getter ()

    let define_prim0 env name f =
      Scheme.define_prim0 env name
	(fun () ->
	  let res_sx = f () in
	  let res_scm = to_scheme res_sx in
	  res_scm)
	  
    let define_prim1 env name f =
      Scheme.define_prim1 env name
	(fun a0_scm ->
	  let a0_sx = from_scheme a0_scm in
	  let res_sx = f a0_sx in
	  let res_scm = to_scheme res_sx in
	  res_scm)
	  
    let define_prim2 env name f =
      Scheme.define_prim2 env name
	(fun a0_scm a1_scm ->
	  let a0_sx = from_scheme a0_scm
	  and a1_sx = from_scheme a1_scm in
	  let res_sx = f a0_sx a1_sx in
	  let res_scm = to_scheme res_sx in
	  res_scm)
	  
    let define_prim3 env name f =
      Scheme.define_prim3 env name
	(fun a0_scm a1_scm a2_scm ->
	  let a0_sx = from_scheme a0_scm
	  and a1_sx = from_scheme a1_scm
	  and a2_sx = from_scheme a2_scm in
	  let res_sx = f a0_sx a1_sx a2_sx in
	  let res_scm = to_scheme res_sx in
	  res_scm)
	  
    let define_prim4 env name f =
      Scheme.define_prim4 env name
	(fun a0_scm a1_scm a2_scm a3_scm ->
	  let a0_sx = from_scheme a0_scm
	  and a1_sx = from_scheme a1_scm
	  and a2_sx = from_scheme a2_scm
	  and a3_sx = from_scheme a3_scm in
	  let res_sx = f a0_sx a1_sx a2_sx a3_sx in
	  let res_scm = to_scheme res_sx in
	  res_scm)

    let define_prim5 env name f =
      Scheme.define_prim5 env name
	(fun a0_scm a1_scm a2_scm a3_scm a4_scm ->
	  let a0_sx = from_scheme a0_scm
	  and a1_sx = from_scheme a1_scm
	  and a2_sx = from_scheme a2_scm
	  and a3_sx = from_scheme a3_scm
	  and a4_sx = from_scheme a4_scm in
	  let res_sx = f a0_sx a1_sx a2_sx a3_sx a4_sx in
	  let res_scm = to_scheme res_sx in
	  res_scm)
	  
    let define_primn env name f =
      Scheme.define_primn env name
	(fun args_scm ->
	  let args_sx = from_scheme args_scm in
	  let res_sx = f args_sx in
	  let res_scm = to_scheme res_sx in
	  res_scm)
	  
  end


(************************************************************************)
(* Atoms								*)
(************************************************************************)

module type AtomType =
  sig
    type t
    val compare: t -> t -> int
    exception Invalid_s_expr of Sx.t
    val from_sexpr: Sx.t -> t
    val to_sexpr: t -> Sx.t
  end

module SimpleAtom =
  struct
    type t = string
    let compare = Pervasives.compare
    exception Invalid_s_expr of Sx.t
    let from_sexpr x =
      match x with
	Sx.Symbol(str) -> str
      |	_ -> raise (Invalid_s_expr(x))
    let to_sexpr x = Sx.Symbol(x)
  end
