(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers
open Bsct_base

(************************************************************************)

let get_arg_name str =
  if String.length str = 0 || String.get str 0 <> '-' then
    failwith ("Expected command arg before \""^str^"\"");
  let colon_indx = 
    try (String.index_from str 1 ':')
    with Not_found -> String.length str
  in let name = String.sub str 1 (colon_indx-1)
  in name

(* Returns (options,param,args') *)
and parse_arg param_p args =
  assert (args <> []);
  let arg = List.hd args
  in let options =
    try
      let colon_indx = String.index_from arg 1 ':'
      and len = (String.length arg)
      in
      String.sub arg (colon_indx+1) (len-colon_indx-1)
    with Not_found -> ""
  and (param,args') =
    if param_p then
      begin
	if List.tl args = [] then
	  failwith ("Missing parameter after \""^arg^"\"");
	((List.hd (List.tl args)), (List.tl (List.tl args)))
      end
    else
      ("", List.tl args)
  in
  (options,param,args')
  
(* Returns (bool,options') *)
and extract_option options chr =
  try
    let indx = String.index options chr
    and len = String.length options
    in let before_options = String.sub options 0 indx
    and after_options = String.sub options (indx+1) (len-indx-1)
    in
    (true,(before_options^after_options))
  with Not_found ->
    (false,options)

and ignore_options options name =
  if options <> "" then
    Printf.fprintf stderr "Ignoring options \"%s\" to -%s\n" 
      options name;
  ()


(************************************************************************)
(* Mechanism for registering command line flags *)

(* hashtbl from "switch_name:tag" -> (func,doc) *)

let arg_funcs = String_hashtbl.create 101

let register_switch str tag doc f =
  (* Remove the leading - *)
  let str =
    if String.length str > 0 && String.get str 0 = '-' then
      String.sub str 1 ((String.length str)-1)
    else
      str
  in 
  if str = "" then failwith "command with with name \"\"??";
  let key = str ^ ":" ^ tag in
  try
    String_hashtbl.find arg_funcs key;
    failwith ("duplicate command line switch -"^str^" for tag "^tag)
  with Not_found ->
    String_hashtbl.add arg_funcs key (f,doc)

let rec find_switch str tag =
  try
    let (f,_) = String_hashtbl.find arg_funcs (str^":"^tag) in f
  with Not_found -> try
    let (f,_) = String_hashtbl.find arg_funcs (str^":*") in f
  with Not_found -> 
    Printf.fprintf stderr "No command line switch -%s for tag %s\n" 
      str tag;
    print_usage ();
    exit 1

and print_usage () =
  let docs = 
    String_hashtbl.fold (fun _ (_,doc) lst -> doc::lst) 
      arg_funcs []
  in let docs = Sort.list (<=) docs in
  Printf.fprintf stderr "Valid switches:\n";
  List.iter (fun doc -> Printf.fprintf stderr "%s\n" doc)
    docs

(************************************************************************)

let rec main () =
  process_args
    (* skip to the first switch *)
    (recurs (Array.to_list Sys.argv)
       (fun recurs args ->
	 if args = [] then
	   []
	 else if String.get (List.hd args) 0 = '-' then
	   args
	 else
	   recurs (List.tl args)))

and process_args args =
  recurs (Sx.Nil,args)
    (fun recurs (sx,args) ->
      if args = [] then
    	()
      else
    	let name = get_arg_name (List.hd args) in
	let tag =
	  match sx with
	    Sx.Nil -> ""
	  | Sx.Pair(Sx.Symbol(tag),_) -> tag
	  | _ -> assert false
	in let func = find_switch name tag in 
	let (sx',args') = func sx args in
	(* Sanity check on new sx. *)
	(match sx with
	  (Sx.Nil|Sx.Pair(Sx.Symbol(_),_)) -> ()
	| _ -> failwith "Bsct_toplevel.process_args: Invalid s-expr.");
	(* Process the next switch. *)
    	recurs (sx',args'))

(************************************************************************)
(*			       Help message				*)
(************************************************************************)

let _ = 
  register_switch "-help" "*"
    "-help -- Print this message."
    (fun sx args ->
      let name = get_arg_name (List.hd args)
      and (options,param,args') = parse_arg false args
      in let _ = ignore_options options name
      in
      print_usage ();
      (sx,args'))

(************************************************************************)
(*			      Scheme funcall				*)
(************************************************************************)

let _ =
  register_switch "-f" "*"
    "-f FNAME -- Call Scheme function FNAME with current program repr."
    (fun sx args ->
      let name = get_arg_name (List.hd args)
      and (options,param,args') = parse_arg true args
      in let _ = ignore_options options name
      in let f = !(Scheme.lookup_var (Scheme.interaction_environment ()) param)
      and arg = Sx.to_scheme sx
      in let result = Scheme.funcall1 f arg
      in let sx' = Sx.from_scheme result
      in (sx',args'))
