(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers
open Bsct_base

(************************************************************************)
(*		       Routines for running the BSCT			*)
(************************************************************************)

(************************************************************************)
(*			       Entry points				*)
(************************************************************************)

(* main.
   Processes command line args. Invoke as 
   Bsct_toplevel.main ();; *)
val main: unit -> unit
(* Invoke the BSCT by passing it a command line to process. *)
val process_args: string list -> unit

(************************************************************************)
(*	Registering routines to be invoked by command line switches	*)
(************************************************************************)

(* Register a function, f, to be invoked when the command line switch
   "-name" is encountered. The routine will only be invoked if the car of
   the s-expr of the current program representation is the symbol "tag".
   The "doc" string is used when printing a usage message.

   E.g., 
   register_switch "-foo" "xtag" doc xyzzy;;
   will register the routine "xyzzy" to be invoked when the command line
   switch "-foo" is encountered, and the current program representation is
   "xtag". The string "doc" is used when printing a usage message. 

   Use tag "*" to register routines that will be invoked for any
   representation. Such routines are used only if representation specific
   routines cannot be found (ie, tag "bmf" matches before tag "*").

   The function, "f", takes the s-expr of the current program
   representation and the current list of command line arguments, and
   returns an s-expr for the new program representation and the subsequent
   list of command line arguments.

   Please observe the following conventions when formatting "doc" strings:

   ("-foo:xyz param (xtag) -- Peform operation foo on the current XTAG"^
   "                               program rep."^
   "       param can be explained here."^
   "       :x is explained here."^
   "       :y is explained here."^
   "       :z is explained here.")
   *)

val register_switch: 
    ((*name:*) string) -> ((*tag:*) string) -> ((*doc:*) string) ->
      ((*f:*) (Sx.t -> string list -> (Sx.t * string list))) -> 
	unit

(************************************************************************)
(*	Utilities for processing arguments in registered routines.	*)
(************************************************************************)

(* Command line switches take the form
   ... -name:options param ... *)

(* extract "name" from "-name:options". *)
val get_arg_name: string -> string

(* Takes the current list of "args" and "param_p" which indicates whether
   the switch expects a parameter. Returns a "options", "param", and a list
   of subsequent arguments, "args'" *)
val parse_arg: 
    ((*param_p:*)bool) ->
      ((*args:*) string list) ->
	((*options:*)string * (*param:*)string * (*args':*)string list)

(* Extract a letter option from a list of options. I.e, finds the
   character, "letter", in the string, "options". Returns "present_p", a
   flag indicating whether "letter" was found, and "options'", the options
   with "letter" removed. *)
val extract_option: 
    ((*options:*)string) -> 
      ((*letter:*)char) -> 
	((*present_p:*)bool * (*options':*)string)

(* Prints a warning that the "options" to switch "-name" were ignored. *)
val ignore_options: (*options:*) string -> (*name*) string -> unit

