(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers

open Bmf;;
open Bmf_util;;

exception No_matching_blas;;

(************************************************************************)

let rec find_blas_in_procs ids =
    Genc.add_header_func blas_header_fun;
    List.map find_blas_in_proc ids

and find_blas_in_proc id =
    match id with
      Ident_valid(name,typ,Ident_proc(Frame(args),res,body),ann) ->
	let parent_of = parent_of_stmt body
	and dense_arrays =
	    add_to_dense_arrays (no_dense_arrays ()) (res::args)
	in let body' =
	    find_blas_in_stmt parent_of dense_arrays body
	in
	    Ident_valid(name,typ,Ident_proc(Frame(args),res,body'),ann)
    | _ -> id

and find_blas_in_stmt parent_of dense_arrays st =
    let st' = (
	try 
	    find_mvm dense_arrays st
	with No_matching_blas -> try
	    find_bzero dense_arrays st
	with No_matching_blas -> try
	    find_gather dense_arrays st
	with No_matching_blas -> try
	    find_scatter_add dense_arrays st
	with No_matching_blas ->
	    find_blas_in_stmt' parent_of dense_arrays st
    ) in
        let parent =
	  try parent_of st
	  with Root_stmt -> Stmt_nop
	in
	match parent with
	  Stmt_seq(_,_) -> st'
	| _ ->
	    try
		find_bs_mvm dense_arrays st'
	    with No_matching_blas ->
		st'

and find_blas_in_stmt' parent_of dense_arrays st =
    let s = find_blas_in_stmt parent_of dense_arrays
    in
	match st with
	  (Stmt_assign(_,_)|Stmt_call(_,_,_)|Stmt_nop) -> st
	| Stmt_seq(first,second) ->	Stmt_seq(s first,s second)
	| Stmt_for(id,lb,ub,step,body) -> Stmt_for(id,lb,ub,step,s body)
	| Stmt_while(pred,body) -> Stmt_while(pred,s body)
	| Stmt_if(pred,thn,els) -> Stmt_if(pred,s thn,s els)
	| Stmt_let(Frame(ids),body) ->
	    let dense_arrays' = add_to_dense_arrays dense_arrays ids
	    in
		Stmt_let(Frame(ids),
		    find_blas_in_stmt parent_of dense_arrays' body)
	| Stmt_do(pred,body) -> Stmt_do(pred,s body)
	| Stmt_directive(ann,body) -> Stmt_directive(ann,s body)

(************************************************************************)

and find_bs_mvm dense_arrays st =
    let rec recurs lst =
	match lst with
	| (Stmt_call(Expr_var("%dzero"),Expr_void,[
			(* Y := *)Expr_var(y_name);
			(* LB := *)lb_e;
			(* M := *)m]))
	  ::(Stmt_call(Expr_var("%dcolmvm"),Expr_void,[
		(* A := *)Expr_var(a_name);
		(* X := *)Expr_var(x_name);
		(* Y := *)Expr_var(y_name');
		(* LB := *)lb_e';
		(* M := *)m';
		(* N := *)n;
		(* T? := *)Expr_bool(false);
		(* L? := *)Expr_bool(true);
		(* D? := *)Expr_bool(true);
		(* U? := *)Expr_bool(true)]))
	  ::(Stmt_call(Expr_var("%dscatteradd"),Expr_void,[
			(* Y := *)Expr_var(y_name'');
			(* BIG_Y := *)Expr_var(big_y_name);
			(* IND := *)Expr_var(ind_name);
			(* LB := *)lb_e'';
			(* M := *)m'']))
	  ::lst' when (
		    let b1 = compare_printed_exprs m m'
		    and b2 = compare_printed_exprs m m''
		    and b3 = y_name = y_name'
		    and b4 = y_name = y_name''
		    and b5 = compare_printed_exprs lb_e lb_e'
		    and b6 = compare_printed_exprs lb_e lb_e''
		    in
			b1 && b2 && b3 && b4 && b5 && b6
		    ) ->
	    (Stmt_call(Expr_var("%dbscolmvm"),Expr_void,[
		    (* M := *)m;
		    (* N := *)n;
		    (* LB := *)lb_e;
		    (* A := *)Expr_var(a_name);
		    (* X := *)Expr_var(x_name);
		    (* Y := *)Expr_var(y_name);
		    (* BIG_Y := *)Expr_var(big_y_name);
		    (* IND := *)Expr_var(ind_name)])
	    )::(recurs lst')
	| (Stmt_call(Expr_var("%dgather"),Expr_void,[
			(* X := *)Expr_var(x_name);
			(* BIG_X := *)Expr_var(big_x_name);
			(* IND := *)Expr_var(ind_name);
			(* LB := *)lb_e;
			(* M := *)m]))
	  ::(Stmt_call(Expr_var("%dcolmvm"),Expr_void,[
		(* A := *)Expr_var(a_name);
		(* X := *)Expr_var(x_name');
		(* Y := *)Expr_var(y_name);
		(* LB := *)lb_e';
		(* M := *)m';
		(* N := *)n;
		(* T? := *)Expr_bool(true);
		(* L? := *)Expr_bool(true);
		(* D? := *)Expr_bool(true);
		(* U? := *)Expr_bool(true)]))
	  ::lst' when (
		    let b1 = compare_printed_exprs m m'
		    and b2 = compare_printed_exprs lb_e lb_e'
		    and b3 = x_name = x_name'
		    in
			b1 && b2 && b3
		    ) ->
	    (Stmt_call(Expr_var("%dbscolmtvm"),Expr_void,[
		    (* M := *)m;
		    (* N := *)n;
		    (* LB := *)lb_e;
		    (* A := *)Expr_var(a_name);
		    (* X := *)Expr_var(x_name);
		    (* Y := *)Expr_var(y_name);
		    (* BIG_X := *)Expr_var(big_x_name);
		    (* IND := *)Expr_var(ind_name)])
	    )::(recurs lst')
	| st::lst' -> st::(recurs lst')
	| [] -> []
    in
	stmt_list2seq (recurs (seq2stmt_list st))

(************************************************************************)

and find_mvm dense_arrays st =
    let (i,i_lb,i_ub,st') = destruct_for_loop st
    in let (j,j_lb,j_ub,st'') = destruct_for_loop st'
    in let (y_acc,a_acc,x_acc) =
	match (ignore_directives st'') with
	  Stmt_assign(y_acc_1,Expr_op(Op_plus,[y_acc_2;
		    Expr_op(Op_times,[a_acc;x_acc])]))
		when y_acc_1 = y_acc_2
		    && is_a_1d_acc dense_arrays y_acc_1
		    && is_a_2d_acc dense_arrays a_acc
		    && is_a_1d_acc dense_arrays x_acc ->
	    (y_acc_1,a_acc,x_acc)
	| Stmt_assign(y_acc_1,Expr_op(Op_plus,[
		    Expr_op(Op_times,[a_acc;x_acc]);
		    y_acc_2]))
		when y_acc_1 = y_acc_2
		    && is_a_1d_acc dense_arrays y_acc_1
		    && is_a_2d_acc dense_arrays a_acc
		    && is_a_1d_acc dense_arrays x_acc ->
	    (y_acc_1,a_acc,x_acc)
	| Stmt_assign(y_acc_1,Expr_op(Op_plus,[y_acc_2;
		    Expr_op(Op_times,[x_acc;a_acc])]))
		when y_acc_1 = y_acc_2
		    && is_a_1d_acc dense_arrays y_acc_1
		    && is_a_2d_acc dense_arrays a_acc
		    && is_a_1d_acc dense_arrays x_acc ->
	    (y_acc_1,a_acc,x_acc)
	| Stmt_assign(y_acc_1,Expr_op(Op_plus,[
		    Expr_op(Op_times,[x_acc;a_acc]);
		    y_acc_2]))
		when y_acc_1 = y_acc_2
		    && is_a_1d_acc dense_arrays y_acc_1
		    && is_a_2d_acc dense_arrays a_acc
		    && is_a_1d_acc dense_arrays x_acc ->
	    (y_acc_1,a_acc,x_acc)
	| _ -> raise No_matching_blas
    in let (y_name,y_index,a_name,a_row_index,a_col_index,x_name,x_index) =
	match (y_acc,a_acc,x_acc) with
	  (Expr_access(_,Expr_var(y_name),_,[Expr_var(y_index)]),
	   Expr_access(_,Expr_var(a_name),_,[
		Expr_var(a_row_index);Expr_var(a_col_index)]),
	   Expr_access(_,Expr_var(x_name),_,[Expr_var(x_index)])) ->
	    (y_name,y_index,a_name,a_row_index,a_col_index,x_name,x_index)
	| _ -> raise No_matching_blas
    in let (lb,maj,a_row_ub,a_col_ub) =
	try
	    match (find_a_dense_array dense_arrays y_name,
		   find_a_dense_array dense_arrays a_name,
		   find_a_dense_array dense_arrays x_name) with
	      ((y_lb,[_],_,Type_real),
		    (a_lb,[a_row_ub;a_col_ub],a_maj,Type_real),
		    (x_lb,[_],_,Type_real))
			when y_lb = a_lb && a_lb = x_lb ->
		(y_lb,a_maj,a_row_ub,a_col_ub)
	    | _ -> raise No_matching_blas
	with Not_found ->
	    raise No_matching_blas
    in let (lb_e,lb_i) =
	match lb with
	  Dense_lb_0 -> (Expr_int(0),0)
	| Dense_lb_1 -> (Expr_int(1),1)
    in let (i,i_ub,j,j_ub,lower_p,diag_p,upper_p) =
	(* The following patterns ASSUME that bmf_expr_simplify has been
	    called on the bounds. Simplification ensures that constant
	    terms appear before variable terms, for instance. Simplication
	    is currently done in dense.ml *)
	match (i_lb,i_ub,j_lb,j_ub) with
	  (* Rectangular bounds do not have to be square; the rest do *)
	| (Expr_int(i_lb_i),i_ub,Expr_int(j_lb_i),j_ub)
		when i_lb_i = lb_i && j_lb_i = lb_i &&
		    (not (String_set.mem i (expr_fvs j_ub))) ->
	    if i = a_row_index && j = a_col_index then begin
		zassert (i_ub = a_row_ub) No_matching_blas;
		zassert (j_ub = a_col_ub) No_matching_blas;
		(i,i_ub,j,j_ub,true,true,true)
	    end else if j = a_row_index && i = a_col_index then begin
		zassert (j_ub = a_row_ub) No_matching_blas;
		zassert (i_ub = a_col_ub) No_matching_blas;
		(j,j_ub,i,i_ub,true,true,true)
	    end else
		raise No_matching_blas
	  (* lower triangular, including diagonal *)
	| (Expr_int(i_lb_i),i_ub,Expr_int(j_lb_i),Expr_var(i'))
		    when i_lb_i = lb_i && j_lb_i = lb_i && i' = i ->
	    zassert (a_row_ub = a_col_ub) No_matching_blas;
	    if i = a_row_index && j = a_col_index then begin
		zassert (i_ub = a_row_ub) No_matching_blas;
		(i,i_ub,j,j_ub,true,true,false)
	    end else if j = a_row_index && i = a_col_index then begin
		zassert (i_ub = a_col_ub) No_matching_blas;
		(j,j_ub,i,i_ub,false,true,true)
	    end else
		raise No_matching_blas
	  (* Strictly lower triangular *)
	| (Expr_int(i_lb_i),i_ub,Expr_int(j_lb_i),
		Expr_op(Op_plus,[Expr_int(-1);Expr_var(i')]))
		    when i_lb_i = lb_i+1 && j_lb_i = lb_i && i' = i ->
	    zassert (a_row_ub = a_col_ub) No_matching_blas;
	    if i = a_row_index && j = a_col_index then begin
		zassert (i_ub = a_row_ub) No_matching_blas;
		(i,i_ub,j,j_ub,true,false,false)
	    end else if j = a_row_index && i = a_col_index then begin
		zassert (i_ub = a_col_ub) No_matching_blas;
		(j,j_ub,i,i_ub,false,false,true)
	    end else
		raise No_matching_blas
	  (* upper triangular, including diagonal *)
	| (Expr_int(i_lb_i),i_ub,Expr_var(i'),j_ub)
		when i_lb_i = lb_i && i' = i && i_ub = j_ub ->
	    zassert (a_row_ub = a_col_ub) No_matching_blas;
	    if i = a_row_index && j = a_col_index then begin
		zassert (j_ub = a_row_ub) No_matching_blas;
		(i,i_ub,j,j_ub,false,true,true)
	    end else if j = a_row_index && i = a_col_index then begin
		zassert (j_ub = a_col_ub) No_matching_blas;
		(j,j_ub,i,i_ub,true,true,false)
	    end else
		raise No_matching_blas
	  (* Strictly upper triangular *)
	| (Expr_int(i_lb_i),i_ub,
		Expr_op(Op_plus,[Expr_int(1);Expr_var(i')]),j_ub)
		    when i_lb_i = lb_i && i' = i &&
			i_ub = (bmf_expr_simplify
				(Expr_op(Op_plus,[j_ub;Expr_int(-1)]))) ->
	    zassert (a_row_ub = a_col_ub) No_matching_blas;
	    if i = a_row_index && j = a_col_index then begin
		zassert (i_ub = a_row_ub) No_matching_blas;
		(i,i_ub,j,j_ub,false,false,true)
	    end else if j = a_row_index && i = a_col_index then begin
		zassert (i_ub = a_col_ub) No_matching_blas;
		(j,j_ub,i,i_ub,true,false,false)
	    end else
		raise No_matching_blas
	| _ -> raise No_matching_blas
    in let transpose_p =
	if i = y_index && j = x_index then
	    false
	else if j = y_index && i = x_index then
	    true
	else
	    raise No_matching_blas
    (* If A is row-maj, swap everything so that it looks col-maj *)
    in let (transpose_p,lower_p,diag_p,upper_p,a_row_ub,a_col_ub) =
	if maj = Dense_row_major then
	    ((not (transpose_p)),upper_p,diag_p,lower_p,a_col_ub,a_row_ub)
	else
	    (transpose_p,lower_p,diag_p,upper_p,a_row_ub,a_col_ub)
    in let m = make_dimen lb a_row_ub
    and n = make_dimen lb a_col_ub
    in
	Stmt_call(Expr_var("%dcolmvm"),Expr_void,[
	    (* A := *)Expr_var(a_name);
	    (* X := *)Expr_var(x_name);
	    (* Y := *)Expr_var(y_name);
	    (* LB := *)lb_e;
	    (* M := *)m;
	    (* N := *)n;
	    (* T? := *)Expr_bool(transpose_p);
	    (* L? := *)Expr_bool(lower_p);
	    (* D? := *)Expr_bool(diag_p);
	    (* U? := *)Expr_bool(upper_p)
	])

(************************************************************************)

(*
    for i := i_lb:i_ub
	x_name[i] := 0.0;
*)

and find_bzero dense_arrays st =
    let (i,i_lb,i_ub,st') = destruct_for_loop st
    in let (x_name,zero) =
	match (ignore_directives st') with
	  Stmt_assign(Expr_access(Access_strong,Expr_var(x_name),"",[
				    Expr_var(i')]),
			zero) when i = i' ->
	    (x_name,zero)
	| _ -> raise No_matching_blas
    in let _ =
	zassert (zero = Expr_real(0.0) || zero = Expr_int(0))
	    No_matching_blas
    in let (x_lb,x_sc_ty) =
	try
	    match find_a_dense_array dense_arrays x_name with
	      (x_lb,[_],_,x_sc_ty) -> (x_lb,x_sc_ty)
	    | _ -> raise No_matching_blas
	with Not_found ->
	    raise No_matching_blas
    in let _ =
	match (zero,x_sc_ty) with
	  (Expr_int(0),Type_int) -> ()
	| (Expr_real(0.0),Type_real) -> ()
	| _ -> raise No_matching_blas
    in let lb_e =
	match (i_lb,x_lb) with
	  (Expr_int(0),Dense_lb_0) -> Expr_int(0)
	| (Expr_int(1),Dense_lb_1) -> Expr_int(1)
	| _ -> raise No_matching_blas
    in let x_dimen = make_dimen x_lb i_ub
    in let make_call f_name =
	Stmt_call(Expr_var(f_name),Expr_void,[
	    (* X := *)Expr_var(x_name);
	    (* LB := *)lb_e;
	    (* N := *)x_dimen
	])
    in
	match x_sc_ty with
	  Type_int -> make_call "%zzero"
	| Type_real -> make_call "%dzero"
	| _ -> assert false (* Can't happen *)

(************************************************************************)

(*
    for i := i_lb:i_ub
	x_name[i] := y_name[ind_name[i]];
*)

and find_gather dense_arrays st =
    let (i,i_lb,i_ub,st') = destruct_for_loop st
    in let (x_name,y_name,ind_name) =
	match (ignore_directives st') with
	  Stmt_assign(
		    Expr_access(Access_strong,Expr_var(x_name),"",[
			Expr_var(i')]),
		    Expr_access(Access_strong,Expr_var(y_name),"",[
			Expr_access(Access_strong,Expr_var(ind_name),"",[
			    Expr_var(i'')])]))
		when i = i' && i = i'' ->
	    (x_name,y_name,ind_name)
	| _ -> raise No_matching_blas
    in let (x_lb,x_sc_ty,ind_lb) =
	try
	    match (find_a_dense_array dense_arrays x_name,
		   find_a_dense_array dense_arrays ind_name) with
	      ((x_lb,[_],_,x_sc_ty),(ind_lb,[_],_,Type_int)) ->
		(x_lb,x_sc_ty,ind_lb)
	    | _ -> raise No_matching_blas
	with Not_found ->
	    raise No_matching_blas
    in let lb_e =
	match (i_lb,x_lb,ind_lb) with
	  (Expr_int(0),Dense_lb_0,Dense_lb_0) -> Expr_int(0)
	| (Expr_int(1),Dense_lb_1,Dense_lb_1) -> Expr_int(1)
	| _ -> raise No_matching_blas
    in let x_dimen = make_dimen x_lb i_ub
    in let f_name =
	match x_sc_ty with
	  Type_real -> "%dgather"
	| Type_int -> "%zgather"
	| _ -> raise No_matching_blas
    in
	Stmt_call(Expr_var(f_name),Expr_void,[
	    (* X := *)Expr_var(x_name);
	    (* Y := *)Expr_var(y_name);
	    (* IND := *)Expr_var(ind_name);
	    (* LB := *)lb_e;
	    (* N := *)x_dimen
	])

(************************************************************************)

(*
    for i := i_lb:i_ub
	y_name[ind_name[i]] := y_name[ind_name[i]] + x_name[i];
*)

and find_scatter_add dense_arrays st =
    let (i,i_lb,i_ub,st') = destruct_for_loop st
    in let (x_name,y_name,ind_name) =
	match (ignore_directives st') with
	  Stmt_assign(
		    Expr_access(Access_strong,Expr_var(y_name),"",[
			Expr_access(Access_strong,Expr_var(ind_name),"",[
			    Expr_var(i')])]),
		    Expr_op(Op_plus,[
			Expr_access(Access_strong,Expr_var(y_name'),"",[
			    Expr_access(Access_strong,Expr_var(ind_name'),"",[
				Expr_var(i'')])]);
			Expr_access(Access_strong,Expr_var(x_name),"",[
			    Expr_var(i''')])
			]))
		when i = i' && i = i'' && i = i'''
		    && y_name = y_name' && ind_name = ind_name' ->
	    (x_name,y_name,ind_name)
	| Stmt_assign(
		    Expr_access(Access_strong,Expr_var(y_name),"",[
			Expr_access(Access_strong,Expr_var(ind_name),"",[
			    Expr_var(i')])]),
		    Expr_op(Op_plus,[
			Expr_access(Access_strong,Expr_var(x_name),"",[
			    Expr_var(i'')]);
			Expr_access(Access_strong,Expr_var(y_name'),"",[
			    Expr_access(Access_strong,Expr_var(ind_name'),"",[
				Expr_var(i''')])])
			]))
		when i = i' && i = i'' && i = i'''
		    && y_name = y_name' && ind_name = ind_name' ->
	    (x_name,y_name,ind_name)
	| _ -> raise No_matching_blas
    in let (x_lb,x_sc_ty,ind_lb) =
	try
	    match (find_a_dense_array dense_arrays x_name,
		   find_a_dense_array dense_arrays ind_name) with
	      ((x_lb,[_],_,x_sc_ty),(ind_lb,[_],_,Type_int)) ->
		(x_lb,x_sc_ty,ind_lb)
	    | _ -> raise No_matching_blas
	with Not_found ->
	    raise No_matching_blas
    in let lb_e =
	match (i_lb,x_lb,ind_lb) with
	  (Expr_int(0),Dense_lb_0,Dense_lb_0) -> Expr_int(0)
	| (Expr_int(1),Dense_lb_1,Dense_lb_1) -> Expr_int(1)
	| _ -> raise No_matching_blas
    in let x_dimen = make_dimen x_lb i_ub
    in let f_name =
	match x_sc_ty with
	  Type_real -> "%dscatteradd"
	| Type_int -> "%zscatteradd"
	| _ -> raise No_matching_blas
    in
	Stmt_call(Expr_var(f_name),Expr_void,[
	    (* X := *)Expr_var(x_name);
	    (* Y := *)Expr_var(y_name);
	    (* IND := *)Expr_var(ind_name);
	    (* LB := *)lb_e;
	    (* N := *)x_dimen
	])

(************************************************************************)

and ignore_directives st =
    match st with
      Stmt_directive(an,st') -> ignore_directives st'
    | _ -> st

and destruct_for_loop st =
    match (ignore_directives st) with
      Stmt_for(Ident_valid(i,Type_scalar(Type_int),_,An_list([])),
		i_lb,i_ub,Expr_int(1),
	    st') ->
	(i,i_lb,i_ub,st')
    | Stmt_for(Ident_valid(i,Type_scalar(Type_int),_,An_list([])),
		i_ub,i_lb,Expr_int(-1),
	    st') ->
	(i,i_lb,i_ub,st')
    | _ -> raise No_matching_blas

and is_a_1d_acc dense_arrays ex =
    match ex with
      Expr_access(Access_strong,Expr_var(name),"",[_]) ->
	(try
	    find_a_dense_array dense_arrays name;
	    true
	with Not_found ->
	    false)
    | _ -> false

and is_a_2d_acc dense_arrays ex =
    match ex with
      Expr_access(Access_strong,Expr_var(name),"",[_;_]) ->
	(try
	    find_a_dense_array dense_arrays name;
	    true
	with Not_found ->
	    false)
    | _ -> false

and make_dimen lb ub =
    match lb with
      Dense_lb_0 -> bmf_expr_simplify (Expr_op(Op_plus,[ub;Expr_int(1)]))
    | Dense_lb_1 -> ub

(************************************************************************)

and no_dense_arrays () = String_map.empty

and add_to_dense_arrays dense_arrays ids =
    if ids = [] then
	dense_arrays
    else
	add_to_dense_arrays
	    (match (List.hd ids) with
	      Ident_valid(name,Type_array(Array_dense(lb,ubs,maj,sc_ty)),_,_) ->
		String_map.add name (lb,ubs,maj,sc_ty) dense_arrays
	    | _ -> dense_arrays)
	    (List.tl ids)

and find_a_dense_array dense_arrays name =
    String_map.find name dense_arrays

(************************************************************************)

and compare_printed_exprs e1 e2 =
    match (e1,e2) with
      (Expr_int(n1),Expr_int(n2)) -> n1 = n2
    | (Expr_real(n1),Expr_real(n2)) -> n1 = n2
    | (Expr_string(s1),Expr_string(s2)) -> s1 = s2
    | (Expr_bool(b1),Expr_bool(b2)) -> b1 = b2
    | (Expr_void,Expr_void) -> true
    | (Expr_var(v1),Expr_var(v2)) -> v1 = v2
    | (Expr_op(op1,args1),Expr_op(op2,args2)) ->
	op1 = op2 &&
	    compare_printed_expr_lists
		(bmf_expr_args_by_op op1 e1)
		(bmf_expr_args_by_op op2 e2)
    | (Expr_forall(frm1,pred1),Expr_forall(frm2,pred2)) ->
	frm1 = frm2 && compare_printed_exprs pred1 pred2
    | (Expr_exist(frm1,pred1),Expr_forall(frm2,pred2)) ->
	frm1 = frm2 && compare_printed_exprs pred1 pred2
    | (Expr_set(frm1,pred1),Expr_forall(frm2,pred2)) ->
	frm1 = frm2 && compare_printed_exprs pred1 pred2
    | (Expr_access(acc1,r1,name1,args1),Expr_access(acc2,r2,name2,args2))
	    ->
	acc1 = acc2 && compare_printed_exprs r1 r2 && name1 = name2 &&
	    compare_printed_expr_lists args1 args2
    | (_,_) -> false

and compare_printed_expr_lists lst1 lst2 =
    match (lst1,lst2) with
      ([],[]) -> true
    | (e1::lst1',e2::lst2') ->
	compare_printed_exprs e1 e2 &&
	    compare_printed_expr_lists lst1' lst2'
    | (_,_) -> false

(************************************************************************)

and blas_header_fun outchan =
    ()

and dbscolmvm_transformer st =
    match st with
      Stmt_call(Expr_var("%dbscolmvm"),Expr_void,[
	    (* M := *)m;
	    (* N := *)n;
	    (* LB := *)lb_e;
	    (* A := *)Expr_var(a_name);
	    (* X := *)Expr_var(x_name);
	    (* Y := *)Expr_var(y_name);
	    (* BIG_Y := *)Expr_var(big_y_name);
	    (* IND := *)Expr_var(ind_name)]) ->
	Genc.Ir_stmt(Bmf_parser_aux.parse_stmt
	    (
		match lb_e with
		  Expr_int(0) ->
		    "{
			var $mm : int := $m;
			var $nn : int := $n;
			if ($mm == 0 || $nn == 0) {
			} else if ($nn > DGEMV_UNROLL_LVL) {
			    call %dzero($y,$lb,$mm);
			    call %dcolmvm($A,$x,$y,$lb,$mm,$nn,false,
				    true,true,true);
			    call %dscatteradd($y,$big_y,$ind,$lb,$mm);
			} else {
			    call MY_DGEMV_N_1111($mm,$nn,$A,$mm,$x,$big_y,
				    $ind);
			}
		    }"
		| Expr_int(1) ->
		    "{
			var $mm : int := $m;
			var $nn : int := $n;
			if ($mm > 0 && $nn > 0) {
			    call %dzero($y,$lb,$m);
			    call %dcolmvm($A,$x,$y,$lb,$m,$n,false,
				    true,true,true);
			    call %dscatteradd($y,$big_y,$ind,$lb,$m);
			}
		    }"
		| _ -> assert false (* Can't happen *)
	    )
	    [
		("$mm",An_string(Misc.gensym "m"));
		("$nn",An_string(Misc.gensym "n"));
		("$m",An_expr(m));
		("$n",An_expr(n));
		("$lb",An_expr(lb_e));
		("$A",An_string(a_name));
		("$x",An_string(x_name));
		("$y",An_string(y_name));
		("$big_y",An_string(big_y_name));
		("$ind",An_string(ind_name))
	    ])
    | _ -> assert false (* Can't happen *)

and dbscolmtvm_transformer st =
    match st with
      Stmt_call(Expr_var("%dbscolmtvm"),Expr_void,[
	    (* M := *)m;
	    (* N := *)n;
	    (* LB := *)lb_e;
	    (* A := *)Expr_var(a_name);
	    (* X := *)Expr_var(x_name);
	    (* Y := *)Expr_var(y_name);
	    (* BIG_X := *)Expr_var(big_x_name);
	    (* IND := *)Expr_var(ind_name)]) ->
	Genc.Ir_stmt(Bmf_parser_aux.parse_stmt
	    (
		match lb_e with
		  Expr_int(0) ->
		    "{
			var $mm : int := $m;
			var $nn : int := $n;
			if ($mm == 0 || $nn == 0) {
			} else if ($nn > DGEMV_UNROLL_LVL) {
			    call %dgather($x,$big_x,$ind,$lb,$mm);
			    call %dcolmvm($A,$x,$y,$lb,$mm,$nn,true,
				    true,true,true);
			} else {
			    call MY_DGEMV_Y_1111($mm,$nn,$A,$mm,$big_x,
				    $ind,$y);
			}
		    }"
		| Expr_int(1) ->
		    "{
			var $mm : int := $m;
			var $nn : int := $n;
			if ($mm > 0 && $nn > 0) {
			    call %dgather($x,$big_x,$ind,$lb,$n);
			    call %dcolmvm($A,$x,$y,$lb,$m,$n,true,
				    true,true,true);
			}
		    }"
		| _ -> assert false (* Can't happen *)
	    )
	    [
		("$mm",An_string(Misc.gensym "m"));
		("$nn",An_string(Misc.gensym "n"));
		("$m",An_expr(m));
		("$n",An_expr(n));
		("$lb",An_expr(lb_e));
		("$A",An_string(a_name));
		("$x",An_string(x_name));
		("$y",An_string(y_name));
		("$big_x",An_string(big_x_name));
		("$ind",An_string(ind_name))
	    ])
    | _ -> assert false (* Can't happen *)

and dcolmvm_transformer st =
    match st with
      Stmt_call(Expr_var("%dcolmvm"),Expr_void,[
	    (* A := *)Expr_var(a_name);
	    (* X := *)Expr_var(x_name);
	    (* Y := *)Expr_var(y_name);
	    (* LB := *)lb_e;
	    (* M := *)m;
	    (* N := *)n;
	    (* T? := *)Expr_bool(transpose_p);
	    (* L? := *)Expr_bool(lower_p);
	    (* D? := *)Expr_bool(diag_p);
	    (* U? := *)Expr_bool(upper_p)]) ->
	let new_st =
	    if lower_p && diag_p && upper_p then
		make_gemv_call a_name x_name y_name lb_e m n transpose_p
	    else
		make_dtrmv_call a_name x_name y_name lb_e m n transpose_p
		    lower_p diag_p upper_p
	in
	    Genc.Ir_stmt(new_st)
    | _ -> assert false (* Can't happen *)

(* DGEMV call *)
and make_gemv_call a_name x_name y_name lb_e m n transpose_p =
    let args = [
	("$tr",An_expr(Expr_string(
	    if transpose_p then "t" else "n")));
	("$mm",An_string(Misc.gensym "m"));
	("$nn",An_string(Misc.gensym "n"));
	("$m",An_expr(m));
	("$n",An_expr(n));
	("$lb",An_expr(lb_e));
	("$A",An_string(a_name));
	("$x",An_string(x_name));
	("$y",An_string(y_name))
    ] in
	Bmf_parser_aux.parse_stmt
	    "{
		var $mm : int := $m;
		var $nn : int := $n;
		call DGEMV($tr,&$mm,&$nn,
		    &pvs_one_r,$A,&$mm,$x,
		    &pvs_one_i,&pvs_one_r,$y,
		    &pvs_one_i);
	    }"
	    args

(* DTRMV call *)
and make_dtrmv_call a_name x_name y_name lb_e m n transpose_p
	lower_p diag_p upper_p =
    let mm = Misc.gensym "m"
    and nn = Misc.gensym "n"
    in let args = [
	("$f_name",An_string(
	    match (transpose_p,upper_p) with
	      (false,false) -> "MY_DTRMV_N_L"
	    | (false,true) ->  "MY_DTRMV_N_U"
	    | (true,false) ->  "MY_DTRMV_T_L"
	    | (true,true) ->   "MY_DTRMV_T_U"));
	("$lb",An_expr(lb_e));
	(* The real dimension of A *)
	("$m",An_expr(m));
	("$mm",An_string(mm));
	(* The reduced dimension of A when (not diag_p) *)
	("$n",An_expr(
	    if (not (diag_p)) then
		Expr_op(Op_plus,[m;Expr_int(-1)])
	    else
		m));
	("$nn",An_string(nn));
	(* A, displaced appropriately *)
	("$A",An_expr(
	    if diag_p then
		Expr_var(a_name)
	    else if lower_p then
		Expr_op(Op_plus,[Expr_var(a_name);Expr_int(1)])
	    else (* if upper_p *)
		Expr_op(Op_plus,[Expr_var(a_name);Expr_var(nn)])));
	(* x, displaced appropriately *)
	("$x",An_expr(
	    if (not (diag_p)) &&
		((lower_p && transpose_p) ||
		 (upper_p && (not (transpose_p)))) then
		Expr_op(Op_plus,[Expr_var(x_name);Expr_int(1)])
	    else
		Expr_var(x_name)));
	(* y, displaced appropriately *)
	("$y",An_expr(
	    if (not (diag_p)) &&
		((lower_p && (not (transpose_p))) ||
		 (upper_p && transpose_p)) then
		Expr_op(Op_plus,[Expr_var(y_name);Expr_int(1)])
	    else
		Expr_var(y_name)))
    ]
    in
	(match (lower_p,diag_p,upper_p) with
	| ((false,false,true)|(true,true,false)
	    |(false,true,true)|(true,false,false)) -> ()
	| (_,_,_) -> assert false (* Can't happen *));
	Bmf_parser_aux.parse_stmt
	    "{
		# Matrix is square
		var $mm : int := $m;
		{ var $nn : int := $n;
		call $f_name($nn,$A,$mm,$x,$y);
	    }}"
	    args

and zero_transformer st =
    match st with
      Stmt_call(Expr_var(f_name),Expr_void,[
	    (* X := *)Expr_var(x_name);
	    (* LB := *)lb_e;
	    (* N := *)x_dimen])
	    when f_name = "%dzero" || f_name = "%zzero" ->
	Genc.Ir_stmt(Bmf_parser_aux.parse_stmt
	    "for $i := $lb:($n+$lb-1) {
		$x[[$i]] := $zero;
	    }"
	    [
		("$i",An_string(Misc.gensym "i"));
		("$x",An_string(x_name));
		("$lb",An_expr(lb_e));
		("$n",An_expr(x_dimen));
		("$zero",An_expr(
		    match f_name with
		      "%dzero" -> Expr_real(0.0)
		    | "%zzero" -> Expr_int(0)
		    | _ -> assert false (* Can't happen *)))
	    ])
    | _ -> assert false (* Can't happen *)

and gather_transformer st =
    match st with
      Stmt_call(Expr_var(f_name),Expr_void,[
	    (* X := *)Expr_var(x_name);
	    (* Y := *)Expr_var(y_name);
	    (* IND := *)Expr_var(ind_name);
	    (* LB := *)lb_e;
	    (* N := *)x_dimen])
	    when f_name = "%dgather" || f_name = "%zgather" ->
	Genc.Ir_stmt(Bmf_parser_aux.parse_stmt
	    "for $i := $lb:($n+$lb-1) {
		$x[[$i]] := $y[[ $ind[[ $i ]] ]];
	    }"
	    [
		("$i",An_string(Misc.gensym "i"));
		("$x",An_string(x_name));
		("$y",An_string(y_name));
		("$ind",An_string(ind_name));
		("$lb",An_expr(lb_e));
		("$n",An_expr(x_dimen))
	    ])
    | _ -> assert false (* Can't happen *)

and scatter_add_transformer st =
    match st with
      Stmt_call(Expr_var(f_name),Expr_void,[
	    (* X := *)Expr_var(x_name);
	    (* Y := *)Expr_var(y_name);
	    (* IND := *)Expr_var(ind_name);
	    (* LB := *)lb_e;
	    (* N := *)x_dimen])
	    when f_name = "%dscatteradd" || f_name = "%zscatteradd" ->
	Genc.Ir_stmt(Bmf_parser_aux.parse_stmt
	    "for $i := $lb:($n+$lb-1) {
		$y[[ $ind[[ $i ]] ]] := $y[[ $ind[[ $i ]] ]] + $x[[$i]];
	    }"
	    [
		("$i",An_string(Misc.gensym "i"));
		("$x",An_string(x_name));
		("$y",An_string(y_name));
		("$ind",An_string(ind_name));
		("$lb",An_expr(lb_e));
		("$n",An_expr(x_dimen))
	    ])
    | _ -> assert false (* Can't happen *)

(************************************************************************)

;;

Genc.add_intrinsic_handler "%dbscolmvm" dbscolmvm_transformer;;
Genc.add_intrinsic_handler "%dbscolmtvm" dbscolmtvm_transformer;;
Genc.add_intrinsic_handler "%dcolmvm" dcolmvm_transformer;;
Genc.add_intrinsic_handler "%dzero" zero_transformer;;
Genc.add_intrinsic_handler "%zzero" zero_transformer;;
Genc.add_intrinsic_handler "%dgather" gather_transformer;;
Genc.add_intrinsic_handler "%zgather" gather_transformer;;
Genc.add_intrinsic_handler "%dscatteradd" scatter_add_transformer;;
Genc.add_intrinsic_handler "%zscatteradd" scatter_add_transformer;;
