(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers

open Bmf;;
open Bmf_util;;
open Bmf_join;;

exception Not_part_of_for_nest;;

let rec do_fm id =
    match id with
      Ident_valid(name,ty,Ident_proc(Frame(args),res,body),an) ->
	let body' = find_for_nests (Expr_bool(true)) body
	in
	    Ident_valid(name,ty,Ident_proc(Frame(args),res,body'),an)
    | _ -> id

and find_for_nests symb_bounds st =
    let s = find_for_nests symb_bounds
    in
	match st with
	(* bo-query *)
	| Stmt_directive(
		An_list([
		    An_string("bo-query");An_expr(symb_bounds')])
		    as annot,
		body) ->
	    assert (symb_bounds = Expr_bool(true))
		(* (Failure("Can't happen")) *);
	    Stmt_directive(annot,
		find_for_nests symb_bounds' body)
	(* eo-query *)
	| Stmt_directive(An_list([An_string("eo-query")]) as annot,
		body) ->
	    Stmt_directive(annot,
		find_for_nests (Expr_bool(true)) body)
	(* Beginning of a loop nest *)
	| Stmt_directive(An_list([An_string("do-any")]) as annot,body) ->
	    let (stuff,body') = determine_for_nest symb_bounds st
	    in
		if body' <> st then
		    process_for_nest stuff (s body')
		else
		    Stmt_directive(annot,s body)
	| Stmt_for(id,lb,ub,step,body) ->
	    let (stuff,body') = determine_for_nest symb_bounds st
	    in
		if body' <> st then
		    process_for_nest stuff (s body')
		else
		    Stmt_for(id,lb,ub,step,s body)
	(* Other kinds of non-nest statements *)
	| (Stmt_assign(_,_)|Stmt_call(_,_,_)|Stmt_nop) -> st
	| Stmt_seq(first,second) -> Stmt_seq(s first,s second)
	| Stmt_while(pred,body) -> Stmt_while(pred,s body)
	| Stmt_if(pred,thn,els) -> Stmt_if(pred,s thn,s els)
	| Stmt_let(frm,body) ->
	    if stmt_is_determined st then
		let { determined_term = term; 
		      determined_pred = pred;
		      determined_assertion = assertion;
		      determined_body = body } =
		    stmt2determined st
		in
		determined2stmt 
		  { determined_term = term; 
		    determined_pred = pred;
		    determined_assertion = assertion;
		    determined_body = s body }
	    else
		Stmt_let(frm,s body)
	| Stmt_do(pred,body) ->
	    if stmt_is_determined st then
		let { determined_term = term;
		      determined_pred = pred;
		      determined_assertion = assertion;
		      determined_body = body } =
		    stmt2determined st
		in
		determined2stmt
		  { determined_term = term; 
		    determined_pred = pred;
		    determined_assertion = assertion;
		    determined_body = s body }
	    else if stmt_is_join st then
		let { join_t_var = t_var; 
		      join_bounds = bounds;
		      join_terms = terms; 
		      join_pred = pred;
		      join_assertion = assertion;
		      join_body = body } =
			stmt2join st
		in
		join2stmt
		  { join_t_var = t_var; 
		    join_bounds = bounds;
		    join_terms = terms; 
		    join_pred = pred;
		    join_assertion = assertion;
		    join_body = s body }
	    else
		Stmt_do(pred,s body)
	| Stmt_directive(annot,body) -> Stmt_directive(annot,s body)

and process_for_nest (bounds,raw_free_vars,loop_vars,do_any_p,steps) body =
    let _ =
	assert (loop_vars <> []) (* (Failure("Shouldn't happen")) *)
    in let free_vars = List'.difference raw_free_vars loop_vars
    in let r =
	Expr_set(Frame(List.map make_id (free_vars@loop_vars)),bounds)
    in let all_bounds = Bmf_linear.fm_code_gen r
    in let loop_bounds =
	List.fold_left (fun lst _ -> (List.tl lst)) all_bounds free_vars
    in 
	reconstruct_for_nest loop_vars loop_bounds do_any_p
	    steps body

and determine_for_nest symb_bounds st =
    let not_for_nest_results st =
	((symb_bounds,
	  String_set.to_list (expr_fvs symb_bounds),
	  [],[],[]),st)
    in try (
	(* Match the do-any directive *)
	let (is_do_any,st') =
	    match st with
	      Stmt_directive(An_list([An_string("do-any")]),st') ->
		(true,st')
	    | _ -> (false,st)
	(* Match the loop. Abort if no loop *)
	in let (id,lb,ub,c,st'') =
	    match st' with
	      Stmt_for(id,lb,ub,Expr_int(1),st'') -> (id,lb,ub,1,st'')
	    | Stmt_for(id,ub,lb,Expr_int(-1),st'') -> (id,lb,ub,-1,st'')
	    | _ -> raise Not_part_of_for_nest
	in let var = bmf_ident_name id
	in let lbs1 = bmf_expr_args_by_op (Op_other("max")) lb
	and ubs1 = bmf_expr_args_by_op (Op_other("min")) ub
	in let _ =
	    if (not ((List.for_all affine_expr_p lbs1 &&
		    List.for_all affine_expr_p ubs1))) then begin
		raise Not_part_of_for_nest
	    end
	(* Match the additional bounds if stmt *)
	in let (pred2,st''') =
	    match st'' with
	      Stmt_if(pred,st''',Stmt_nop)
		    when (String_set.mem var (expr_fvs pred)) &&
			(affine_constr_expr_p pred) ->
		(pred,st''')
	    | _ -> (Expr_bool(true),st'')
	in let ((bounds,free_vars,loop_vars,do_any_p,steps),body) =
	    determine_for_nest symb_bounds st'''
	in let var_bound =
	    Expr_op(Op_and,[
		bounds2expr (Expr_var(var)) (lbs1,ubs1);
		pred2])
	in
	    ((make_and [var_bound;bounds],
	      List'.union 
		(String_set.to_list (expr_fvs var_bound))
		free_vars,
	      var::loop_vars,
	      is_do_any::do_any_p,
	      c::steps),
	     body)
    ) with Not_part_of_for_nest ->
	not_for_nest_results st

and reconstruct_for_nest vars bounds do_any_p steps body =
    match (vars,bounds,do_any_p,steps) with
      ([],[],[],[]) -> body
    | (var::vars',(lbs,ubs)::bounds',is_do_any::do_any_p',
	    step::steps') ->
	let body' = 
	    reconstruct_for_nest vars' bounds' do_any_p' steps' body
	in let id = make_id var
	in let (lb,ub) = (make_max lbs,make_min ubs)
	in let body'' =
	    if step > 0 then
		Stmt_for(id,lb,ub,Expr_int(step),body')
	    else
		Stmt_for(id,ub,lb,Expr_int(step),body')
	in let body''' =
	    if is_do_any then
		Stmt_directive(An_list([An_string("do-any")]),body'')
	    else
		body''
	in
	    body'''
    | _ -> assert false (* Can't happen *)

and affine_constr_expr_p e =
    match e with
      Expr_op(op,[lhs;rhs]) when op = Op_eq || op = Op_leq ->
	affine_expr_p lhs && affine_expr_p rhs
    | _ -> false

and affine_expr_p e =
    List.for_all affine_term_expr_p (bmf_expr_args_by_op Op_plus e)

and affine_term_expr_p e =
    match e with
      (Expr_int(_)|Expr_var(_)) -> true
    | Expr_op(Op_times,[Expr_int(_);Expr_var(_)]) -> true
    | Expr_op(Op_times,[Expr_var(_);Expr_int(_)]) -> true
    | _ -> false

and make_id name =
    Ident_valid(name,Type_scalar(Type_int),Ident_bind,An_list([]))

and make_max lbs =
    match lbs with
      [] -> assert false (* Can't happen *)
    | [lb] -> lb
    | lbs -> Expr_op(Op_other("max"),lbs)

and make_min ubs =
    match ubs with
      [] -> assert false (* Can't happen *)
    | [ub] -> ub
    | ubs -> Expr_op(Op_other("min"),ubs)

;;
