(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers

open Bmf;;
open Bmf_util;;
open Dense_misc;;

exception Not_an_update_assign;;
exception Gather_attempt_failure;;

(************************************************************************)
(* Entry								*)
(************************************************************************)

let empty_s = Int_set.empty

type gathers_stuff = {
    g_defs_of_name : bmf_stmt -> string -> Int_set.t;
    g_parent_of : bmf_stmt -> bmf_stmt option
};;

type ref_type =
    Ref_invalid
|   Ref_rhs
|   Ref_update of bmf_op
;;

type unprocd_ref_type = {
    ur_name : string;
    ur_ref : bmf_expr;
    ur_fvs : String_set.t;
    ur_fvs_defs : Int_set.t;
    ur_type : ref_type
};;

let rec do_gathers aggressive_p id =
    match id with
      Ident_valid(name,ty,Ident_proc(Frame(args),res,body),an) ->
	let (ord,in_h,_) = 
	  Hybrid_rds.compute_reaching_defs body
	and aliases_m = 
	  Hybrid_aliases.compute_aliases_by_decls id
	in let in_defs_at_stmt_of_name =
	    Hybrid_rds.defs_at_stmt_by_name ord in_h aliases_m
	and parent_of =
	    let raw_parent_of = parent_of_stmt body
	    in
		(fun st ->
		    try Some(raw_parent_of st)
		    with Root_stmt -> None)
	in let stuff = {
	    g_defs_of_name = in_defs_at_stmt_of_name;
	    g_parent_of = parent_of
	} in let scope = add_to_scope [] (res::args)
	in let (placements,urs,body') =
	    do_gathers_in_stmt aggressive_p stuff scope body
	in let (placements',body'') =
	    place_at_placement placements body body'
	in let id' =
	    Ident_valid(name,ty,Ident_proc(Frame(args),res,body''),an)
	in
	    assert (placements' = [])
		(* (Failure("Why are there pending placements?")) *);
	    id'
    | _ -> id

(************************************************************************)
(* First layer of stmt walker.						*)
(************************************************************************)

and do_gathers_in_stmt aggressive_p stuff scope st =
    let (placements,urs,st') =
	do_gathers_in_stmt' aggressive_p stuff scope st
    in let (placements',st'') =
	place_at_placement placements st st'
    in let urs' =
	remove_variant_refs stuff st urs
    in
	(placements',urs',st'')

(* Check for stuff that needs to needs to be placed at this point. *)

and place_at_placement placements st st' =
    List.fold_right
	(fun ((placement_st,placement_f) as placement)
		(placements',st'') ->
	    if placement_st == st then
		let new_st'' = placement_f st''
		in
		    (placements',new_st'')
	    else
		(placement::placements',st''))
	placements ([],st')

(* If an unprocessed reference is not invariant with respect to the current
    statement, then toss it: it's not safe to hoist it any higher *)

and remove_variant_refs stuff st urs =
    List.fold_right
	(fun ur urs' ->
	    let fvs_defs' = all_defs_of_fvs stuff st ur.ur_fvs
	    in
		if not (Int_set.equal fvs_defs' ur.ur_fvs_defs) then
		    urs'
		else
		    ur::urs')
	urs []

(************************************************************************)
(* Second layer of stmt walker.						*)
(************************************************************************)

and do_gathers_in_stmt' aggressive_p stuff scope st =
    match st with
    (* Interesting stuff happens here. *)
      Stmt_assign(lhs, rhs) ->
	do_gathers_in_stmt_assign stuff scope st lhs rhs
    | Stmt_call(f,res,args) ->
	do_gathers_in_stmt_call stuff scope st res args
    | Stmt_for(id,lb,ub,step,body) ->
	do_gathers_in_stmt_for aggressive_p stuff scope st id lb ub step body
    (* The rest of the stmt are straight forward. *)
    | Stmt_nop -> ([],[],st)
    | Stmt_directive(an,body) ->
	let (p,urs,body') =
	    do_gathers_in_stmt aggressive_p stuff scope body
	in
	    (p,urs,Stmt_directive(an,body'))
    | Stmt_seq(first,second) ->
	let (first_p,first_urs,first') =
	    do_gathers_in_stmt aggressive_p stuff scope first
	and (second_p,second_urs,second') =
	    do_gathers_in_stmt aggressive_p stuff scope second
	in
	    (first_p@second_p,
	     combine_unprocd_refs stuff first_urs second_urs,
	     Stmt_seq(first',second'))
    | Stmt_if(pred,thn,els) ->
	let pred_urs = rvalue_unprocd_refs stuff scope st pred
	and (thn_p,thn_urs,thn') =
	    do_gathers_in_stmt aggressive_p stuff scope thn
	and (els_p,els_urs,els') =
	    do_gathers_in_stmt aggressive_p stuff scope els
	in
	    (thn_p@els_p,
	     combine_unprocd_refs stuff pred_urs
		(combine_unprocd_refs stuff thn_urs els_urs),
	     Stmt_if(pred,thn',els'))
    | Stmt_while(pred,body) ->
	let pred_urs = rvalue_unprocd_refs stuff scope st pred
	and (body_p,body_urs,body') =
	    do_gathers_in_stmt aggressive_p stuff scope body
	in
	    (body_p,
	     combine_unprocd_refs stuff pred_urs body_urs,
	     Stmt_while(pred,body'))
    | Stmt_do(Expr_set(Frame(ids),pred),body) ->
	let pred_urs = rvalue_unprocd_refs stuff scope st pred
	and (body_p,body_urs,body') =
	    do_gathers_in_stmt aggressive_p stuff scope body
	in
	    (body_p,
	     combine_unprocd_refs stuff pred_urs body_urs,
	     Stmt_do(Expr_set(Frame(ids),pred),body'))
    | Stmt_do(_,_) -> failwith "Ill-formed stmt_do"
    | Stmt_let(Frame(ids),body) ->
	let scope' = add_to_scope scope ids
	in let ids_urs =
	    List.fold_right List'.union (List.map (find_urs_in_ident stuff scope st) ids) []
	and (body_p,body_urs,body') =
	    do_gathers_in_stmt aggressive_p stuff scope' body
	in
	    (body_p,
	     combine_unprocd_refs stuff ids_urs body_urs,
	     Stmt_let(Frame(ids),body'))

and do_gathers_in_stmt_assign stuff scope st lhs rhs =
    let urs =
	try (
	    let (op,lhs,args) = destruct_update_assign st
	    in
		combine_unprocd_refs stuff
		    (lvalue_unprocd_refs stuff scope st lhs
			(Ref_update(op)))
		    (rvalues_unprocd_refs stuff scope st args)
	) with Not_an_update_assign ->
	    combine_unprocd_refs stuff
		(lvalue_unprocd_refs stuff scope st lhs Ref_invalid)
		(rvalue_unprocd_refs stuff scope st rhs)
    in
	([],urs,st)

and do_gathers_in_stmt_call stuff scope st res args =
    let urs =
	List.fold_right
	    (fun arg urs ->
		combine_unprocd_refs stuff
		    (lvalue_unprocd_refs stuff scope st arg Ref_invalid)
		    urs)
	    (res::args) []
    in
	([],urs,st)

and do_gathers_in_stmt_for aggressive_p stuff scope st id lb ub step body =
    let id_urs = find_urs_in_ident stuff scope st id
    and lb_urs = rvalue_unprocd_refs stuff scope st lb
    and ub_urs = rvalue_unprocd_refs stuff scope st ub
    and step_urs = rvalue_unprocd_refs stuff scope st step
    and (body_p,body_urs,body') =
	do_gathers_in_stmt aggressive_p stuff scope body
    in let (body_p',body_urs',body'') =
	List.fold_right (attempt_to_gather aggressive_p stuff scope st id lb ub step)
	    body_urs (body_p,[],body')
    in
	(body_p',
	 combine_unprocd_refs stuff (id_urs@lb_urs@ub_urs@step_urs) body_urs',
	 Stmt_for(id,lb,ub,step,body''))

and attempt_to_gather aggressive_p stuff scope st id lb ub step ur
    (p',urs',body') =
let name = (bmf_ident_name id)
in
    if (not ((String_set.mem name ur.ur_fvs))) then
	(p',ur::urs',body')
    else begin
	(* fake_ref is ur.ur_ref with all occurances of name removed. *)
	let fake_ref = Dense_walk.expr_of_walk_kind
	    (replace_expr (Dense_walk.Walk_expr(ur.ur_ref))
			    (Expr_var(name)) Expr_void)
	(* fake_expr is used to figure out where the gather code can 
	    be placed. It simply is a bogus expr that contains all of
	    the exprs that we need to worry about. *)
	in let fake_expr =
	    Expr_op(Op_plus,[lb;ub;step;fake_ref])
	in let place_st =
	    compute_invariant_placement stuff st fake_expr
	in let (v_name,v_acc_name,v_index) =
	    match ur.ur_ref with
	      Expr_access(Access_strong,Expr_var(v_name),v_acc_name,
		    [v_index]) ->
		(v_name,v_acc_name,v_index)
	    | _ -> failwith "Can't happen"
	in let (v_maj,v_sc_ty) = List.assoc v_name scope
	in (try (
	    (* Try to do a real gather. *)
	    (* Test for profitability *)
	    (let rec iter x =
		match x with
		(* Not profitable. *)
		  None -> raise Gather_attempt_failure
		| Some(st') when st' = place_st ->
		    raise Gather_attempt_failure
		(* Profitable. *)
		| Some(Stmt_for(_,_,_,_,_)|Stmt_while(_,_)|Stmt_do(_,_)) ->
		    ()
		(* Keep going *)
		| Some(st') -> iter (stuff.g_parent_of st')
	    in
		if ((not (aggressive_p))) then begin
		    (* Don't count the current for loop *)
		    iter (stuff.g_parent_of st);
		end);
	    (* Test whether or not the index expression is ``complex'' *)
	    (let complex_p =
		Dense_walk.it_bmf
		    (fun ctx form ->
			match form with
			  Dense_walk.Walk_expr(Expr_access(_,_,_,_)) ->
			    Dense_walk.Iba_return(true)
			| _ ->
			    Dense_walk.Iba_visit([Dense_walk.WALK_EXPRS],ctx,
				(fun lst -> Dense_walk.Iba_return(List.mem true lst))))
		    () (Dense_walk.Walk_expr(v_index))
	    in
		zassert complex_p Gather_attempt_failure);
	    (* Do the gather *)
	    let v_name' = Misc.gensym ur.ur_name
	    and v_lb =
		match lb with
		  Expr_int(0) -> Dense_lb_0
		| Expr_int(1) -> Dense_lb_1
		| _ -> raise Gather_attempt_failure
	    in let old_ref = ur.ur_ref
	    and new_ref = Expr_access(Access_strong,Expr_var(v_name'),
			    "",[Expr_var(name)])
	    and new_decl =
		Ident_valid(v_name',
			Type_array(Array_dense(v_lb,[ub],v_maj,v_sc_ty)),
			Ident_var(Expr_void),An_list([]))
	    and replace_name name' ex =
		Dense_walk.expr_of_walk_kind
		    (replace_expr (Dense_walk.Walk_expr(ex))
			(Expr_var(name)) (Expr_var(name')))
	    in let place_rhs_gather place_st' =
		let name' = Misc.gensym name
		in let new_ref' = replace_name name' new_ref
		and old_ref' = replace_name name' old_ref
		in
		    Stmt_let(Frame([new_decl]),
			Stmt_seq(
			    Stmt_for(Ident_valid(name',
					Type_scalar(Type_int),
					Ident_bind,An_list([])),
				lb,ub,step,
				Stmt_assign(new_ref',old_ref')),
			    place_st'))
	    and place_update_gather op place_st' =
		let name' = Misc.gensym name
		and name'' = Misc.gensym name
		and idem = idempotent_element op v_sc_ty
		in let new_ref' = replace_name name' new_ref
		and new_ref'' = replace_name name'' new_ref
		and old_ref'' = replace_name name'' old_ref
		in
		    Stmt_let(Frame([new_decl]),
			Stmt_seq(Stmt_seq(
			    Stmt_for(Ident_valid(name',
					Type_scalar(Type_int),
					Ident_bind,An_list([])),
				lb,ub,step,
				Stmt_assign(new_ref',idem)),
			    place_st'),
			    Stmt_for(Ident_valid(name'',
					Type_scalar(Type_int),
					Ident_bind,An_list([])),
				lb,ub,step,
				Stmt_assign(old_ref'',
				    Expr_op(op,[old_ref'';new_ref''])))
			    ))
	    and body'' =
		Dense_walk.stmt_of_walk_kind
		    (replace_expr (Dense_walk.Walk_stmt(body')) old_ref new_ref)
	    in
		((place_st,
		    (match ur.ur_type with
		      Ref_rhs -> place_rhs_gather
		    | Ref_update(op) -> place_update_gather op
		    | Ref_invalid -> failwith "Can't happen")
		 )::p',urs',body'')
	) with Gather_attempt_failure -> (
	    (* Nothing worked. Drop the ur and move on... *)
	    (p',urs',body')
	))
    end

(************************************************************************)
(* Misc.								*)
(************************************************************************)

and add_to_scope scope ids =
    List.fold_right
	(fun id scope ->
	    match id with
	      Ident_valid(name,Type_array(Array_dense(_,[_],maj,sc_ty)),
		    Ident_var(_),_) ->
		(name,(maj,sc_ty))::scope
	    | _ -> scope)
	ids scope
and replace_expr form old_ex new_ex =
    Dense_walk.map_bmf
	(fun ctx form ->
	    match form with
	      Dense_walk.Walk_expr(ex) when ex = old_ex ->
		Dense_walk.Mba_return(Dense_walk.Walk_expr(new_ex))
	    | _ ->
		Dense_walk.Mba_visit([Dense_walk.WALK_ALL],ctx,
				   (fun x -> Dense_walk.Mba_return x)))
	() form

and lvalue_unprocd_refs stuff scope st ex typ =
    match ex with
      Expr_access(_,_,_,_) -> make_unprocd_ref stuff scope st ex typ
    | _ -> rvalue_unprocd_refs stuff scope st ex

and rvalues_unprocd_refs stuff scope st exs =
    List.fold_right
	(fun ex urs ->
	    List'.union (rvalue_unprocd_refs stuff scope st ex) urs)
	exs []

and rvalue_unprocd_refs stuff scope st ex =
    let e = rvalue_unprocd_refs stuff scope st
    in
	match ex with
	  (Expr_int(_)|Expr_real(_)|Expr_string(_)|Expr_bool(_)
		|Expr_void|Expr_var(_)) ->
	    []
	| Expr_access(_,Expr_var(name),_,_) ->
	    make_unprocd_ref stuff scope st ex Ref_rhs
	| Expr_access(_,_,_,_) ->
	    failwith "Ill-formed expr_access"
	| Expr_op(Op_addr,args) ->
	    (match args with
	      [Expr_var(name)] ->
		[make_invalid_unprocd_ref stuff name]
	    | [Expr_access(_,Expr_var(name),_,args)] ->
		List.fold_right List'.union (List.map e args)
		    [make_invalid_unprocd_ref stuff name]
	    | _ -> failwith "Ill-formed op_other(addr)")
	| Expr_op(op,args) -> List.fold_right List'.union (List.map e args) []
	| Expr_forall(_,pred) -> e pred
	| Expr_exist(_,pred) -> e pred
	| Expr_set(_,pred) -> e pred

and find_urs_in_ident stuff scope st id =
    match id with
      Ident_valid(_,_,id_stuff,_) ->
	(match id_stuff with
	  Ident_bind -> []
	| Ident_var(e) -> rvalue_unprocd_refs stuff scope st e
	| Ident_proc(_,_,_) -> failwith "Can't happen")
    | Ident_null -> []

and make_invalid_unprocd_ref stuff name =
    { ur_name = name;
      ur_ref = Expr_void;
      ur_fvs = String_set.empty;
      ur_fvs_defs = empty_s;
      ur_type = Ref_invalid
    }

and make_unprocd_ref stuff scope st ex typ =
    try (
	let (acc,name) =
	    match ex with
	      Expr_access(acc,Expr_var(name),_,_) -> (acc,name)
	    | _ -> failwith "Shouldn't happen"
	in let (_,sc_ty) = List.assoc name scope
	in
	    if typ = Ref_invalid || acc <> Access_strong then
		[make_invalid_unprocd_ref stuff name]
	    else begin
		(match typ with
		  Ref_update(op) ->
		    idempotent_element op sc_ty; ()
		| _ -> ());
		let fvs = free_vars_in_expr ex
		in let fvs_defs = all_defs_of_fvs stuff st fvs
		in [{
		    ur_name = name;
		    ur_ref = ex;
		    ur_fvs = fvs;
		    ur_fvs_defs = fvs_defs;
		    ur_type = typ
		}]
	    end
    ) with Not_found ->
	[]

and combine_unprocd_refs stuff urs1 urs2 =
    let name_to_urs_f ur m =
	let urs =
	    try
		String_map.find ur.ur_name m
	    with Not_found ->
		[]
	in
	    String_map.add ur.ur_name (ur::urs) m
    in let name_to_urs_m =
	List.fold_right name_to_urs_f urs1
	    (List.fold_right name_to_urs_f urs2
		String_map.empty)
    in let partition_urs urs =
	List.fold_right
	    (fun ur (invalids,updates,rhs) ->
		let add x lst =
		    if List.mem x lst then lst else x::lst
		in
		    match ur.ur_type with
		      Ref_invalid -> (add ur invalids,updates,rhs)
		    | Ref_update(_) -> (invalids,add ur updates,rhs)
		    | Ref_rhs -> (invalids,updates,add ur rhs))
	    urs ([],[],[])
    in
	String_map.fold
	    (fun name urs lst ->
		match partition_urs urs with
		  ([],[],rhs) -> rhs@lst
		| ([],[update],[]) -> update::lst
		| _ -> 
		    (make_invalid_unprocd_ref stuff name)
		    ::lst)
	    name_to_urs_m []

and destruct_update_assign st =
    let (lhs,rhs) =
	match st with
	  Stmt_assign(lhs,rhs) -> (lhs,rhs)
	| _ -> raise Not_an_update_assign
    in let op =
	match rhs with
	  Expr_op(op,_) -> op
	| _ -> raise Not_an_update_assign
    in let (found_eq_lhs,not_eq_lhs) =
	List.fold_right
	    (fun ex (found_eq_lhs,not_eq_lhs) ->
		(* Only catch the first ex = lhs *)
		if ex = lhs && (not (found_eq_lhs)) then
		    (true,not_eq_lhs)
		else
		    (found_eq_lhs,not_eq_lhs@[ex]))
	    (bmf_expr_args_by_op op rhs)
	    (false,[])
    in
	if found_eq_lhs then
	    (op,lhs,not_eq_lhs)
	else
	    raise Not_an_update_assign

and idempotent_element op sc_ty =
    match op with
      Op_plus when sc_ty = Type_int -> Expr_int(0)
    | Op_plus when sc_ty = Type_real -> Expr_real(0.0)
    | Op_times when sc_ty = Type_int -> Expr_int(1)
    | Op_times when sc_ty = Type_real -> Expr_real(1.0)
    | Op_and when sc_ty = Type_bool -> Expr_bool(true)
    | Op_or when sc_ty = Type_bool -> Expr_bool(false)
    | _ -> raise Not_found

(* Compute the highest parent of st at which ex is _still_ invariant. *)
and compute_invariant_placement stuff st ex =
    let fvs = free_vars_in_expr ex
    in let fvs_defs = all_defs_of_fvs stuff st fvs
    in
    recurs st
      (fun recurs st' ->
      	let parent_o = stuff.g_parent_of st' in
      	match parent_o with
	  None -> st'
      	|	Some(parent_st) ->
	    let fvs_defs' = all_defs_of_fvs stuff parent_st fvs in
	    if not (Int_set.equal fvs_defs fvs_defs') then
	      st'
	    else
	      recurs parent_st)

and all_defs_of_fvs stuff st fvs =
    let all_defs =
	String_set.fold
	    (fun fv defs ->
	      Int_set.union
		defs
		(try
		  stuff.g_defs_of_name st fv
		with Not_found ->
		  empty_s))
	    fvs empty_s
    in
	all_defs
;;
