(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers

open Bmf;;
open Dense_misc;;

(************************************************************************)
(*			  Hoisting invariant code			*)
(************************************************************************)

(************************************************************************)
(* Hoist let bindings							*)
(*									*)
(* Place stmt_let identifiers as ``high'' in the procedure as possible. *)
(* It gets interesting because we have to handle nested bindings        *)
(* correctly. E.g., { var x := 1; { var y := x; ... } }			*)
(*									*)
(* N.B. This implementation assumes that renaming has been performed    *)
(*									*)
(* *todo* BUG -- this implementation does not account for assignments to*)
(* init expressions!							*)
(************************************************************************)

let empty_s = Int_set.empty

type h_l_b_stuff = {
    hlb_defs_of_name : bmf_stmt -> string -> Int_set.t;
    hlb_ord : Hybrid_rds.Def_ordinator.t
};;

(* Entry point for procedures *)

let rec hoist_let_bindings (id: bmf_ident) =
    match id with
      Ident_valid(name,typ,Ident_proc(args,res,body),an) ->
	let (ord,in_ht,_) = 
	  Hybrid_rds.compute_reaching_defs body
	and aliases_m = 
	  Hybrid_aliases.compute_aliases_by_decls id
	in let in_defs_at_stmt_of_name =
	    Hybrid_rds.defs_at_stmt_by_name ord in_ht aliases_m
	in let stuff = {
	    hlb_defs_of_name = in_defs_at_stmt_of_name;
	    hlb_ord = ord
	} in let results =
	    h_l_b_walk stuff body body
	in let body' = place_all_let_bindings results
	in
	    Ident_valid(name,typ,Ident_proc(args,res,body'),an)
    | _ -> id

(* Returns "binding pair" list * new stmt, where "binding pair" is a
     bmf_ident * (all free idents in the bmf_ident) *)

(* This proc looks at the defs in the parent and the defs in the bindings
    and figures out if a binding needs to be placed b/c it is no longer
    invariant. *)

and h_l_b_walk stuff (parent_st: bmf_stmt) (st: bmf_stmt) = (
    let (st_bps,st') =
	h_l_b_walk' stuff parent_st st
    in let (ids_to_place,remaining_bps) =
	List.fold_right
	    (fun ((id,fvs,fvs_defs) as bp)
		    (ids_to_place,remaining_bps) ->
		let fvs_defs' = expr_defs_of_fvs stuff parent_st fvs
		in
		    if not (Int_set.equal fvs_defs' fvs_defs) then
			(id::ids_to_place,remaining_bps)
		    else
			(ids_to_place,bp::remaining_bps))
	    st_bps ([],[])
    in
	if ids_to_place = [] then
	    (st_bps,st')
	else
	    let (body_bps,body) =
		place_let_bindings false (List.map bmf_ident_name ids_to_place)
		    remaining_bps st'
	    in
		(body_bps,Stmt_let(Frame(ids_to_place),body))
: (bmf_ident * String_set.t * Int_set.t) list * bmf_stmt)

(* This proc actually traverses the statements and collects and returns the
    bindings. *)

and h_l_b_walk' stuff (parent_st: bmf_stmt) (st: bmf_stmt) = (
    let s = h_l_b_walk stuff st
    in
	match st with
	  (Stmt_assign(_,_)|Stmt_call(_,_,_)|Stmt_nop) ->
	    ([],st)
	| Stmt_seq(first,second) ->
	    let (first_bps,first') = s first
	    and (second_bps,second') = s second
	    in
		(first_bps@second_bps, Stmt_seq(first',second'))
	| Stmt_directive(an,body) ->
	    let (body_bps,body') = s body
	    in
		(body_bps,Stmt_directive(an,body'))
	(* Collect bindings from let *)
	| Stmt_let(Frame(ids),body) ->
	    let (body_bps,body') = s body
	    in let ids_bps =
		List.map (fun id ->
			let fvs = free_vars_in_ident id
			in let fvs_defs = expr_defs_of_fvs stuff st fvs
			in
			    (id,fvs,fvs_defs))
		    ids
	    in
		(ids_bps@body_bps,body')
	(* Flush bindings on either side of conditionals *)
	| Stmt_if(pred,thn,els) ->
	    ([],Stmt_if(pred,
		    (place_all_let_bindings (s thn)),
		    (place_all_let_bindings (s els))))
	(* We do _not_ assume that while loops are executed at all. *)
	| Stmt_while(pred,body) ->
	    ([],Stmt_while(pred,
		    (place_all_let_bindings (s body))))
	(* We assume that for loops are executed at least once. *)
	| Stmt_for(id,lb,ub,step,body) ->
	    let (body_bps,body') = s body
	    in let (body_bps',body'') =
		place_let_bindings false [bmf_ident_name id] body_bps
		    body'
	    in
		(body_bps', Stmt_for(id,lb,ub,step,body''))
	(* We assume that do loops are executed at least once. *)
	| Stmt_do(Expr_set(Frame(ids),pred),body) ->
	    let (body_bps,body') = s body
	    in let (body_bps',body'') =
		place_let_bindings false (List.map bmf_ident_name ids)
		    body_bps body'
	    in
		(body_bps', Stmt_do(Expr_set(Frame(ids),pred),body''))
	| Stmt_do(_,_) -> failwith "Ill-formed stmt_do"
: (bmf_ident * String_set.t * Int_set.t) list * bmf_stmt)

and place_all_let_bindings (binding_pairs,body) =
    let collected_fvs =
	List.fold_left
	    (fun vars (_,fvs,_) -> String_set.union vars fvs)
	    String_set.empty binding_pairs
    in let fvs_not_in_b_ps =
	remove_idents (List.map (fun (id,_,_) -> id) binding_pairs)
	    collected_fvs
    in let (binding_pairs',body') =
	place_let_bindings true (String_set.elements fvs_not_in_b_ps)
	    binding_pairs body
    in
	assert (binding_pairs' = []) (* (Failure "Can't happen") *);
	body'

and place_let_bindings place_no_fvs_p names binding_pairs body =
    let (ids_w_names,bps_wo_names) =
	List.fold_left
	    (fun (ids_w_names,bps_wo_names) ((id,fvs,_) as bp) ->
		if (place_no_fvs_p && String_set.is_empty fvs) ||
			List.exists (fun name -> String_set.mem name fvs) names
		then
		    (id::ids_w_names,bps_wo_names)
		else
		    (ids_w_names,bp::bps_wo_names))
	    ([],[]) binding_pairs
    in
	if ids_w_names <> [] then
	    let (binding_pairs',body') =
		place_let_bindings false (List.map bmf_ident_name ids_w_names)
		    bps_wo_names body
	    in
		(binding_pairs',Stmt_let(Frame(ids_w_names),body'))
	else
	    (binding_pairs,body)

and expr_defs_of_fvs stuff st fvs =
    let all_defs =
	String_set.fold
	    (fun fv defs ->
	      Int_set.union defs
		(stuff.hlb_defs_of_name st fv))
	    fvs empty_s
    in
	remove_id_defs stuff all_defs

and remove_id_defs stuff defs =
  Int_set.fold
    (fun def defs' ->
      match Hybrid_rds.Def_ordinator.of_int stuff.hlb_ord def with
	Hybrid_rds.Def_write _ -> Int_set.add def defs'
      | Hybrid_rds.Def_entry _ -> defs')
    defs empty_s
;;
