(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers

open Bmf;;
(* open Bmf_walk;; *)

(************************************************************************)
(*	     Misc. transformations for the dense optimizations		*)
(************************************************************************)

let remove_idents ids vars =
    List.fold_left
	(fun vars id ->
	    match id with
	      Ident_valid(name,_,_,_) -> String_set.remove name vars
	    | Ident_null -> vars)
	vars ids
;;

(************************************************************************)
(* Collecting free variables that occur in program trees.		*)
(* These routines are obsolete. Use the Bmf.xxx_fvs routines instead.	*)
(************************************************************************)

let reduce_s f lst =
  List.fold_left 
    (fun result x -> String_set.union result (f x))
    String_set.empty lst

let free_vars_in_ident id = (Bmf_util.ident_fvs id)
let free_vars_in_idents ids = (Bmf_util.idents_fvs ids)
let free_vars_in_stmt st = (Bmf_util.stmt_fvs st)
let free_vars_in_stmts sts = reduce_s free_vars_in_stmt sts
let free_vars_in_expr ex = (Bmf_util.expr_fvs ex)
let free_vars_in_exprs exs = reduce_s free_vars_in_expr exs
let free_vars_in_type ty = (Bmf_util.type_fvs ty)
let free_vars_in_types tys = reduce_s free_vars_in_type tys
let free_vars_in_ann an = (Bmf_util.ann_fvs an)
let free_vars_in_anns ans = reduce_s free_vars_in_ann ans
  

(************************************************************************)
(* Remove unused let-declarations					*)
(*									*)
(* Do a bottom-up traversal of the tree. Remove let-decl's for		*)
(* identifiers that are not referenced					*)
(************************************************************************)

let empty_s = String_set.empty

let rec remove_let_decls_in_proc (id: bmf_ident) = (
  match id with
    Ident_valid(name,typ,Ident_proc(args,res,body),an) ->
      Ident_valid(name,typ,Ident_proc
		    (args,res,
		     remove_let_decls_in_stmt body),an)
  | _ -> id
: bmf_ident)

and remove_let_decls_in_stmt (st: bmf_stmt) = (
  let (st',_) = remove_let_decls_in_stmt' st
  in st'
: bmf_stmt)

(* Takes a stmt and returns a (transformed stmt * free vars in transformed
   stmt) pair. *)
and remove_let_decls_in_stmt' (st: bmf_stmt) = (
  let union_s = String_set.union
  and diff_s = String_set.diff 
  in
  let r = remove_let_decls_in_stmt'
  and f_a = Bmf_util.ann_fvs
  and f_e = Bmf_util.expr_fvs
  and f_i = Bmf_util.ident_fvs
  in
  match st with
    Stmt_nop -> (st, empty_s)
  | Stmt_assign(lhs,rhs) -> (st, union_s (f_e lhs) (f_e rhs))
  | Stmt_seq(first,second) ->
      let (first',first_f) = r first
      and (second',second_f) = r second
      in (Stmt_seq(first',second'), union_s first_f second_f)
  | Stmt_call(f,res,args) -> 
      (st, union_s (f_e f) 
	 (union_s (f_e res)
	    (List.fold_left (fun s arg -> union_s s (f_e arg))
	       empty_s args)))
  | Stmt_while(pred,body) ->
      let (body',body_f) = r body in
      (Stmt_while(pred,body'),
       union_s (f_e pred) body_f)
  | Stmt_if(test,thn,els) ->
      let (thn',thn_f) = r thn and (els',els_f) = r els in
      (Stmt_if(test,thn',els'),
       union_s (f_e test) (union_s thn_f els_f))
  | Stmt_directive(ann,body) ->
      let (body',body_f) = r body in
      (Stmt_directive(ann,body'),
       union_s (f_a ann) body_f)
  | Stmt_for(id,lb,ub,step,body) ->
      let (body',body_f) = r body in
      let body_f' = String_set.remove (bmf_ident_name id) body_f in
      (Stmt_for(id,lb,ub,step,body'),
       union_s (f_i id) 
	 (union_s (f_e lb)
	    (union_s (f_e ub)
	       (union_s (f_e step) body_f'))))
  | Stmt_do(Expr_set(Frame(ids),pred),body) ->
      let (body',body_f) = r body
      and pred_f = f_e pred
      and bound = 
	List.fold_left (fun s id -> String_set.add (bmf_ident_name id) s)
	  empty_s ids
      and ids_f = 
	List.fold_left  (fun s id -> union_s s (f_i id))
	  empty_s ids
      in
      (Stmt_do(Expr_set(Frame(ids),pred),body'),
       union_s ids_f (diff_s (union_s body_f pred_f) bound))
  | Stmt_do _ -> assert false
  | Stmt_let(Frame(ids),body) ->
      let (body',body_f) = r body
      and bound = 
	List.fold_left (fun s id -> String_set.add (bmf_ident_name id) s)
	  empty_s ids
      in let unused_bound = diff_s bound body_f in
      let bound' = diff_s bound unused_bound in
      let ids' = 
	List'.remove_if_not
	  (fun id -> String_set.mem (bmf_ident_name id) bound')
	  ids
      in let ids_f =
	List.fold_left  (fun s id -> union_s s (f_i id))
	  empty_s ids'
      in
      (Stmt_let(Frame(ids'),body'),
       union_s ids_f (diff_s body_f bound'))

: bmf_stmt * String_set.t)
