(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers

open Bmf;;
open Bmf_util;;
open Bb_decl;;
open Bmf_join;;
open Dense_misc;;

let empty_s = String_set.empty
let singleton_s x = String_set.add x empty_s

let rec do_reindex id =
    match id with
      Ident_valid(name,ty,Ident_proc(Frame(args),res,body),an) ->
	let body' = Dense_walk.deep_copy_stmt body
	in let denses_s =
	    add_denses
		(add_denses empty_s [res])
		args
	in let dtrmds =
	    Dense_walk.it_bmf
		(fun denses_s form ->
		    let st = Dense_walk.stmt_of_walk_kind form
		    in let st_lst =
			if is_reindexable_determined_p denses_s st then
			    [st]
			else
			    []
		    and denses_s' =
			match st with
			  Stmt_let(Frame(ids),_) ->
			    add_denses denses_s ids
			| _ -> denses_s
		    in
			Dense_walk.Iba_visit([Dense_walk.WALK_STMTS],denses_s',
			    (fun lsts ->
				Dense_walk.Iba_return(List.fold_left (fun x y -> x@y)
						st_lst lsts))))
		denses_s (Dense_walk.Walk_stmt(body'))
	in let body'' =
	    List.fold_left
		(fun body'' dtrmd ->
		    let decl'' =
			Ident_valid(name,ty,
			    Ident_proc(Frame(args),res,body''),an)
		    in
			attempt_reindex (res::args) dtrmd body''
			    (Hybrid_aliases.compute_aliases_by_decls decl''))
		body' dtrmds
	in
	    Ident_valid(name,ty,Ident_proc(Frame(args),res,body''),an)
    | _ -> id

and is_reindexable_determined_p denses_s st =
    try (
	let { determined_term = {
		jt_kind = kind;
		jt_constr = constr
		(* todo: *)
		(* jt_strength = strength *) };
	      determined_body = body } = stmt2determined st
	in begin
	    (* todo: *)
	(*
	    match (kind,strength,constr) with
	      ((jtk_enum_indexing(_,jtk_eq(_,_,_,[_]))
		    |jtk_enum_searching(_,jtk_eq(_,_,_,[_]),_)),
	       jts_no_test,
	       expr_access(access_strong,expr_var(A),"",_))
		    when (set__mem A denses_s) ->
		true
	    | _ -> false
	*)
	false
	end
    ) with Not_a_valid_determined ->
	false

and add_denses denses_s ids =
  match ids with
    [] -> denses_s
  | id::ids' ->
      add_denses
	(match id with
	  Ident_valid(_,Type_array(atyp),_,_) ->
	    (match atyp with
	      Array_dense _ -> String_set.add (bmf_ident_name id) denses_s 
	    | Array_bb _ -> denses_s
	    | Array_view _ -> 
		(* Views should be removed by the scheduler. *)
		assert false
		  )
	| _ -> denses_s)
	ids'

and attempt_reindex param_decls dtrmd_st proc_body aliases_m =
    let parent_of = parent_of_stmt proc_body
    and { determined_term = { jt_kind = kind; jt_constr = constr };
	  determined_body = dtrmd_body } as dtrmd
	= stmt2determined dtrmd_st
    in let (a_name,a_ref) =
	match constr with
	  Expr_access(_,(Expr_var (a_name) as a_ref),_,_) ->
	    (a_name,a_ref)
	| _ -> failwith "Can't happen"
    and x_name =
	match kind with
	  Jtk_enum_indexing(_,Jtk_eq(_,x_name,_,_)) -> x_name
	| Jtk_enum_searching(_,Jtk_eq(_,x_name,_,_),_) -> x_name
	| _ -> failwith "Can't happen"
    in let a_aliases =
	(try
	    String_map.find a_name aliases_m
	with Not_found ->
	    singleton_s a_name)
    in let a_decl =
	find_decl param_decls parent_of a_name dtrmd_st
    in let another_A_ref_p =
	(* Does A_name, or any alias of A_name, appear anywhere else
	    in the determined? *)
	Dense_walk.it_bmf
	    (fun ctx form ->
		match form with
		  Dense_walk.Walk_expr(Expr_var(a_name') as var_ref)
			when 
                  String_set.mem a_name' a_aliases ->
		    Dense_walk.Iba_return(var_ref != a_ref)
		| _ ->
		    Dense_walk.Iba_visit([Dense_walk.WALK_ALL],ctx,
			(fun vals -> Dense_walk.Iba_return(List.mem true vals))))
	    () (Dense_walk.Walk_stmt(dtrmd_st))
    in let safe_p = (not (another_A_ref_p))
    in let (profitable_p,placement) =
	if safe_p then
	    determine_placement parent_of a_name a_aliases
		x_name dtrmd_st
	else
	    (false,Stmt_nop) (* just savin' a few cycles *)
    in
	if (not ((safe_p && profitable_p))) then
	    proc_body
	else
	    gen_reindex dtrmd_st proc_body a_decl
		placement dtrmd

and find_decl param_decls parent_of name st =
    let rec do_match st =
	match st with
	  Stmt_let(Frame(ids),_) ->
	    (try
		find_in_ids ids
	    with Not_found ->
		do_next st)
	| _ ->
	    do_next st
    and do_next st =
	(try
	    do_match (parent_of st)
	with Root_stmt ->
	    find_in_ids param_decls)
    and find_in_ids ids =
	match ids with
	  [] -> raise Not_found
	| (Ident_valid(name',_,_,_) as id)::_
		when name = name' ->
	    id
	| _::ids' -> find_in_ids ids'
    in
	(try
	    do_next st
	with Not_found ->
	    failwith "Failed to find decl.")

and determine_placement parent_of a_name a_aliases
	x_name dtrmd_st =
    let rec contains_A_alias fvs =
	(not (String_set.is_empty (String_set.inter a_aliases fvs)))
    and loop st seen_query_top_p seen_loop_after_query_p =
	let profitable_p = seen_query_top_p && seen_loop_after_query_p
	in
	    (try
		let st' = parent_of st
		in
		    loop' st st' profitable_p seen_query_top_p 
			seen_loop_after_query_p
	    with Root_stmt ->
		(profitable_p,st))
    and loop' st st' profitable_p seen_query_top_p
	    seen_loop_after_query_p =
	match st' with
	  Stmt_seq(first,second) ->
	    if first == st then
		if contains_A_alias (stmt_fvs second) then
		    (profitable_p,st)
		else
		    loop st' seen_query_top_p seen_loop_after_query_p
	    else if second == st then
		if contains_A_alias (stmt_fvs first) then
		    (profitable_p,st)
		else
		    loop st' seen_query_top_p seen_loop_after_query_p
	    else
		failwith "Can't happen"
	| Stmt_for(id,lb,ub,step,_) ->
	    if contains_A_alias
		(List.fold_left String_set.union empty_s
		    [ident_fvs id;
		     expr_fvs lb;
		     expr_fvs ub])
	    then
		(profitable_p,st)
	    else
		loop st' seen_query_top_p seen_query_top_p
	| Stmt_while(pred,_) ->
	    if contains_A_alias (expr_fvs pred) then
		(profitable_p,st)
	    else
		loop st' seen_query_top_p seen_query_top_p
	| Stmt_do(pred,_) ->
	    if contains_A_alias (expr_fvs pred) then
		(profitable_p,st)
	    else
		loop st' seen_query_top_p seen_query_top_p
	(* *todo* is this right? *)
	| Stmt_if(_,_,_) ->	(profitable_p,st)
	| Stmt_let(Frame(ids),_) ->
	    let names = List.map bmf_ident_name ids
	    in
		if List.mem a_name names || List.mem x_name names ||
		    contains_A_alias (idents_fvs ids)
		then
		    (profitable_p,st)
		else 
		    loop st' seen_query_top_p seen_loop_after_query_p
	| Stmt_directive(an,_) ->
	    if contains_A_alias (ann_fvs an) then
		(profitable_p,st)
	    else
		let seen_query_top_p' =
		    match an with
		      _ when seen_query_top_p -> true
		    | An_list((An_string("bo-query"))::_) -> true
		    | _ -> false
		in		
		    loop st' seen_query_top_p'
			seen_loop_after_query_p
	| _ -> failwith "Can't happen"
    in
	 loop dtrmd_st false false

(* todo: *)
(*
and gen_reindex dtrmd_st proc_body A_decl placement dtrmd =
    let { determined_term = {
	    jt_kind = kind;
	    jt_strength = strength;
	    jt_enum_var = enum_var;
	    jt_join_var = join_var;
	    jt_constr = constr;
	    jt_bounds = bounds } as dtrmd_term;
	  determined_body = dtrmd_body } as dtrmd
	= stmt2determined dtrmd_st
    in let (A_ref,A_name,A_mthd,args) =
	match constr with
	  expr_access(_,(expr_var(A_name) as A_ref),A_mthd,args) ->
	    (A_ref,A_name,A_mthd,args)
	| _ -> failwith "Can't happen"
    in let (lb,ubs) =
	match A_decl with
	  ident_valid(_,type_dense(lb,ubs,_,_),_,_) -> (lb,ubs)
	| _ -> failwith "Can't happen"
    in let gen_loop body_f =
	let rec recurs args ubs =
	    match ubs with
	      [] -> body_f args
	    | ub::ubs' ->
		let i = gensym "i"
		in
		    stmt_for(ident_valid(i,type_scalar(type_int),
				ident_bind,an_list([])),
			expr_int(if lb = dense_lb_0 then 0 else 1),
			ub,expr_int(1),
			recurs (args@[expr_var(i)]) ubs')
	in
	    recurs [] ubs
    in let (x_name,x_mthd) =
	match kind with
	  jtk_enum_indexing(_,jtk_eq(_,x_name,x_mthd,_)) ->
	    (x_name,x_mthd)
	| jtk_enum_searching(_,jtk_eq(_,x_name,x_mthd,_),_) ->
	    (x_name,x_mthd)
	| _ -> failwith "Can't happen"
    in let enum_e = constr
    and join_e =
	expr_access(access_strong,expr_var(x_name),
	    x_mthd,[constr])
    in let reindex_loop =
	gen_loop
	    (fun args ->
		determined2stmt {
		    determined_term = {
			jt_kind = kind;
			jt_strength = strength;
			jt_enum_var = enum_var;
			jt_join_var = join_var;
			jt_constr =
			    expr_access(access_strong,A_ref,A_mthd,args);
			jt_bounds = bounds;
			};
		    determined_body =
			stmt_assign(
			    expr_access(access_strong,A_ref,A_mthd,args),
			    expr_var(enum_var)) })
    and unreindex_loop =
	gen_loop
	    (fun args ->
		stmt_assign(
		    expr_access(access_strong,A_ref,A_mthd,args),
		    expr_access(access_strong,expr_var(x_name),x_mthd,[
			expr_access(access_strong,A_ref,A_mthd,args)])))
    in
	stmt_of_walk_kind(
	    Dense_walk.map_bmf
		(fun ctx form ->
		    match form with
		      walk_stmt(st) when st == dtrmd_st ->
			mba_recurs(ctx,walk_stmt(dtrmd_body),mba_return)
		    | walk_stmt(st) when st == placement ->
			mba_visit([WALK_ALL],ctx,
			    (fun form ->
				mba_return(
				    walk_stmt(
					stmt_seq(reindex_loop,
					    stmt_seq(stmt_of_walk_kind(form),
						unreindex_loop))))))
		    | walk_expr(expr_var(v)) when v = enum_var ->
			mba_recurs(ctx,walk_expr(enum_e),mba_return)
		    | walk_expr(expr_var(v)) when v = join_var ->
			mba_recurs(ctx,walk_expr(join_e),mba_return)
		    | _ -> mba_visit([WALK_ALL],ctx,
			    (fun form -> mba_return(form))))
		() (walk_stmt(proc_body)))

*)

and gen_reindex dtrmd_st proc_body a_decl placement dtrmd =
    Stmt_nop
;;
