(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers

open Bmf;;
open Bmf_util;;
open Dense_misc;;

exception Not_an_update_assign;;
exception Reptr_attempt_failure;;

(* do_reptrs performs the following transformations:

	for i = 0:n-1 {
	    ... y[m + i] ...
	}
	--->
	var y' : array (0:n-1) of real = &y[m];
	for i = 0:n-1 {
	    ... y'[i] ...
	}

   This is a useful precursor to BLAS pattern matching.
   *)

(************************************************************************)
(* Entry								*)
(************************************************************************)

let empty_s = Int_set.empty

module Stmt_element =
  struct
    type t = bmf_stmt
    let equal = (==)
    let hash = Containers.hash
  end
module Stmt_hashtbl = Hashtbl(Stmt_element)

type reptrs_stuff = {
    rpt_defs_of_name : bmf_stmt -> string -> Int_set.t;
    rpt_parent_of : bmf_stmt -> bmf_stmt option;
    rpt_placements : placement_type list Stmt_hashtbl.t
} and placement_type = {
    p_old_expr : bmf_expr;
    p_new_expr : bmf_expr;
    p_ident: bmf_ident;
};;

let rec do_reptrs id =
    match id with
      Ident_valid(name,ty,Ident_proc(Frame(args),res,body),an) ->
	let (ord,in_ht,_) = 
	  Hybrid_rds.compute_reaching_defs body
	and aliases_m = 
	  Hybrid_aliases.compute_aliases_by_decls id
	in let in_defs_at_stmt_of_name =
	    Hybrid_rds.defs_at_stmt_by_name ord in_ht aliases_m
	and parent_of =
	    let raw_parent_of = parent_of_stmt body
	    in
		(fun st ->
		    try Some(raw_parent_of st)
		    with Root_stmt -> None)
	in let stuff = {
	  rpt_defs_of_name = in_defs_at_stmt_of_name;
	  rpt_parent_of = parent_of;
	  rpt_placements = Stmt_hashtbl.create 101
	} in let scope = add_to_scope [] (res::args)
	in let body' =
	    do_reptrs_in_stmt stuff scope body
	in
	    Ident_valid(name,ty,Ident_proc(Frame(args),res,body'),an)
    | _ -> id

and do_reptrs_in_stmt stuff scope st =
    let st' = (
	let s = do_reptrs_in_stmt stuff scope
	and e = do_reptrs_in_expr stuff scope st
	in
	    match st with
	      Stmt_assign(lhs,rhs) -> Stmt_assign(e lhs,e rhs)
	    | Stmt_seq(first,second) -> Stmt_seq(s first,s second)
	    | Stmt_call(f,res,args) -> Stmt_call(f,e res,List.map e args)
	    | Stmt_for(id,lb,ub,step,body) ->
		Stmt_for(id,e lb,e ub,e step,s body)
	    | Stmt_while(pred,body) -> Stmt_while(e pred,s body)
	    | Stmt_if(pred,thn,els) -> Stmt_if(e pred,s thn,s els)
	    | Stmt_let(Frame(ids),body) ->
		let scope' = add_to_scope scope ids
		in let body' = do_reptrs_in_stmt stuff scope' body
		in
		    Stmt_let(Frame(ids),body')
	    | Stmt_do(pred,body) -> Stmt_do(e pred,s body)
	    | Stmt_nop -> st
	    | Stmt_directive(ann,body) -> Stmt_directive(ann,s body)
    ) in
	do_placements stuff st st'

and do_reptrs_in_expr stuff scope st ex =
    let ex' = (
	let e = do_reptrs_in_expr stuff scope st
	in
	    match ex with
	      (Expr_int(_)|Expr_real(_)|Expr_string(_)|Expr_bool(_)
		    |Expr_void|Expr_var(_)) ->
		ex
	    | Expr_op(op,args) -> Expr_op(op,List.map e args)
	    | Expr_forall(frm,pred) -> Expr_forall(frm,e pred)
	    | Expr_exist(frm,pred) -> Expr_exist(frm,e pred)
	    | Expr_set(frm,pred) -> Expr_set(frm,e pred)
	    | Expr_access(acc,r,name,args) ->
		Expr_access(acc,e r,name,List.map e args)
    ) in let ex'' =
	subst_by_placements stuff st ex'
    in
	try
	    attempt_to_reptr_expr stuff scope st ex''
	with Reptr_attempt_failure ->
	    ex''

(************************************************************************)

and do_placements stuff st st' =
    do_placements_loop (find_placements stuff st) st'

and do_placements_loop ps st' =
    if ps = [] then
	st'
    else begin
	let names =
	    List.fold_right
		(fun p names ->
		    String_set.add (bmf_ident_name p.p_ident) names)
		ps String_set.empty
	in let (ready_ps,not_ready_ps) =
	    List.fold_right
		(fun p (ready_ps,not_ready_ps) ->
		    let fvs_in_p = free_vars_in_ident p.p_ident
		    in
			if String_set.is_empty (String_set.inter names fvs_in_p) then
			    (p::ready_ps,not_ready_ps)
			else
			    (ready_ps,p::not_ready_ps))
		ps ([],[])
	in let _ =
	    assert (ready_ps <> []) (* (Failure("No ready placements?!?")) *)
	in let ids =
	    List.fold_right (fun p ids -> p.p_ident::ids)
		ready_ps []
	in
	    Stmt_let(Frame(ids),
		do_placements_loop not_ready_ps st')
    end

and find_placements stuff st =
    try
	Stmt_hashtbl.find stuff.rpt_placements st
    with Not_found ->
	[]

and add_placements stuff st p =
    let ps = find_placements stuff st
    in
	Stmt_hashtbl.remove stuff.rpt_placements st;
        Stmt_hashtbl.add stuff.rpt_placements st (p::ps)

and subst_by_placements stuff st ex =
  match ex with
    (Expr_int(_)|Expr_real(_)|Expr_string(_)|Expr_bool(_) | 
    Expr_void|Expr_var(_)) -> ex
  | _ -> 
      recurs (find_placements stuff st)
    	(fun recurs lst ->
	  if lst = [] then
	    (match stuff.rpt_parent_of st with
	      None -> ex
	    | Some(parent_st) -> 
	      	subst_by_placements stuff parent_st ex)
	  else if ex = (List.hd lst).p_old_expr then
	    (List.hd lst).p_new_expr
	  else
	    recurs (List.tl lst))

(*and subst_by_placements stuff x ex =*)
(*    let rec iter lst =*)
(*	if lst = [] then*)
(*	  (match x with*)
(*	    None -> ex*)
(*	  | Some(st) -> *)
(*	      subst_by_placements stuff (stuff.rpt_parent_of st) ex)*)
(*	else if ex = (List.hd lst).p_old_expr then*)
(*	    (List.hd lst).p_new_expr*)
(*	else*)
(*	    iter (List.tl lst)*)
(*    in*)
(*	match ex with*)
(*	  (Expr_int(_)|Expr_real(_)|Expr_string(_)|Expr_bool(_)*)
(*		|Expr_void|Expr_var(_)) -> ex*)
(*	| _ -> iter (find_placements stuff st)*)

and names_at_placement stuff st =
    List.fold_right (fun p s -> String_set.add (bmf_ident_name p.p_ident) s)
	(find_placements stuff st)
	String_set.empty

(************************************************************************)

and attempt_to_reptr_expr stuff scope st ex =
    let (name,arg) =
	match ex with
	  Expr_access(Access_strong,Expr_var(name),"",[arg]) ->
	    (name,arg)
	| _ -> raise Reptr_attempt_failure
    in let (v_lb,v_ub,maj,sc_ty) =
	try
	    find_in_scope scope name
	with Not_found ->
	    raise Reptr_attempt_failure
    in let v_lb_i =
	match v_lb with
	  Dense_lb_0 -> 0
	| Dense_lb_1 -> 1
    in let (invar_st,_) = find_hoisting_target stuff st arg
    in let (id,lp_lb,lp_ub) =
	match stuff.rpt_parent_of invar_st with
	  Some(Stmt_for(id,lp_lb,lp_ub,Expr_int(1),_)) -> (id,lp_lb,lp_ub)
	| Some(Stmt_for(id,lp_ub,lp_lb,Expr_int(-1),_)) -> (id,lp_lb,lp_ub)
	| _ -> raise Reptr_attempt_failure
    in let i = bmf_ident_name id
    in let (operands,complex_p,found_i_p) =
	List.fold_right
	    (fun operand (operands,complex_p,found_i_p) ->
		match operand with
		  Expr_var(i') when i = i' ->
		    zassert ((not (found_i_p))) Reptr_attempt_failure;
		    (operands,complex_p,true)
		| (Expr_int(_)|Expr_real(_)|Expr_string(_)|Expr_bool(_)) ->
		    (operand::operands,complex_p,found_i_p)
		| _ ->
		    zassert ((not ((String_set.mem i (free_vars_in_expr operand)))))
			Reptr_attempt_failure;
		    (operand::operands,true,found_i_p))
	    (bmf_expr_args_by_op Op_plus arg) ([],false,false)
    in let _ = zassert complex_p Reptr_attempt_failure
    in let _ = zassert found_i_p Reptr_attempt_failure
    in let arg' = Expr_op(Op_plus,operands)
    in let (invar_st',profitable_p) = find_hoisting_target stuff st arg'
    in let _ = zassert profitable_p Reptr_attempt_failure
    in let name' = Misc.gensym name
    in let new_expr =
	Expr_access(Access_strong,Expr_var(name'),"",[Expr_var(i)])
    in let v_ub' = bmf_expr_simplify
	(make_plus [v_ub;make_neg lp_lb;Expr_int(v_lb_i)])
    in let offset = bmf_expr_simplify
	(make_plus [arg';Expr_int(v_lb_i)])
    in let ident =
	Ident_valid(name',
	    Type_array(Array_dense(v_lb,[v_ub'],maj,sc_ty)),
	    Ident_var(Expr_op(Op_addr,[
		Expr_access(Access_strong,Expr_var(name),"",
		    [offset])])),
	    An_list([]))
    in
	add_placements stuff invar_st' {
	    p_old_expr = ex;
	    p_new_expr = new_expr;
	    p_ident = ident
	};
	new_expr	

(************************************************************************)

and find_hoisting_target stuff st ex =
  let fvs = free_vars_in_expr ex
  in let fvs_defs = all_defs_of_fvs stuff st fvs
  in
  recurs (st,false)
    (fun recurs (st,profitable_p) ->
      match stuff.rpt_parent_of st with
	None -> (st,profitable_p)
      | Some(st') ->
	  if not (String_set.is_empty 
		    (String_set.inter fvs
		       (names_at_placement stuff st))) then
	    (st,profitable_p)
	  else if not (Int_set.equal (all_defs_of_fvs stuff st' fvs)
			 fvs_defs) then
	    (st,profitable_p)
	  else
	    (match st with
	      (Stmt_for(_,_,_,_,_)|Stmt_while(_,_)|Stmt_do(_,_)) ->
		recurs (st',true)
	    | _ -> recurs (st',profitable_p)))

and add_to_scope scope ids =
    List.fold_right
	(fun id scope ->
	    match id with
	      Ident_valid(name,Type_array(Array_dense(lb,[ub],maj,sc_ty)),
		    Ident_var(_),_) ->
		(name,(lb,ub,maj,sc_ty))::scope
	    | _ -> scope)
	ids scope

and find_in_scope scope name =
    List.assoc name scope

and expr_of_dense_lb lb =
    match lb with
      Dense_lb_0 -> Expr_int(0)
    | Dense_lb_1 -> Expr_int(1)

and all_defs_of_fvs stuff st fvs =
    let all_defs =
	String_set.fold
	    (fun fv defs ->
	      Int_set.union
		defs
		(try
		  stuff.rpt_defs_of_name st fv
		with Not_found ->
		  empty_s))
	    fvs empty_s
    in
	all_defs

(************************************************************************)

;;
