(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl

open Bmf;;

type walk_kind =
    Walk_stmt of bmf_stmt
|   Walk_expr of bmf_expr
|   Walk_type of bmf_type
|   Walk_ident of bmf_ident
|   Walk_ann of annotation
;;

type walk_what =
    WALK_ALL
|   WALK_STMTS
|   WALK_EXPRS
|   WALK_TYPES
|   WALK_IDENTS
|   WALK_ANNS
;;

type ('a,'b) transform_bmf_action =
    Tba_return of 'b * walk_kind 
|   Tba_recurs of 'a
		* walk_kind
	        * ('b -> walk_kind -> ('a,'b) transform_bmf_action)
|   Tba_visit of walk_what list
	        * 'a
		* ('b list -> walk_kind -> ('a,'b) transform_bmf_action)
;;

type 'a map_bmf_action =
    Mba_return of walk_kind
|   Mba_recurs of 'a
		* walk_kind
	        * (walk_kind -> 'a map_bmf_action)
|   Mba_visit of walk_what list
	        * 'a
		* (walk_kind -> 'a map_bmf_action)
;;

type ('a,'b) it_bmf_action =
    Iba_return of 'b
|   Iba_recurs of 'a
		* walk_kind
	        * ('b -> ('a,'b) it_bmf_action)
|   Iba_visit of walk_what list
	        * 'a
		* ('b list -> ('a,'b) it_bmf_action)
;;


(************************************************************************)
(* A General Program Transformer					*)
(************************************************************************)

let stmt_of_walk_kind wk =
    match wk with
      Walk_stmt(st) -> st
    | _ -> failwith "Not a walk_stmt."
and expr_of_walk_kind wk =
    match wk with
      Walk_expr(st) -> st
    | _ -> failwith "Not a walk_expr."
and type_of_walk_kind wk =
    match wk with
      Walk_type(st) -> st
    | _ -> failwith "Not a walk_type."
and ident_of_walk_kind wk =
    match wk with
      Walk_ident(st) -> st
    | _ -> failwith "Not a walk_ident."
and ann_of_walk_kind wk =
    match wk with
      Walk_ann(st) -> st
    | _ -> failwith "Not a walk_ann."
;;

let rec transform_bmf (f: 'a -> walk_kind -> ('a,'b) transform_bmf_action)
	(ctx: 'a) (form : walk_kind) = (

    
 let rec next_action a =
	match a with
	  Tba_return((x0,x1)) -> (x0,x1)
	| Tba_recurs(ctx',form',contin) ->
	    let (lst,form) = (transform_bmf f ctx' form')
	    in
		next_action (contin lst form)
	| Tba_visit(whats,ctx',contin) ->
	    let (lst,form) = (do_visit whats ctx' form)
	    in
		next_action (contin lst form)
    and do_visit whats ctx form =
	let do_stmts_p = (List.mem WALK_ALL whats || List.mem WALK_STMTS whats)
	and do_exprs_p = (List.mem WALK_ALL whats || List.mem WALK_EXPRS whats)
	and do_types_p = (List.mem WALK_ALL whats || List.mem WALK_TYPES whats)
	and do_idents_p = (List.mem WALK_ALL whats || List.mem WALK_IDENTS whats)
	and do_anns_p = (List.mem WALK_ALL whats || List.mem WALK_ANNS whats)
	and nop x = ([],x)
	in let s =
	    if do_stmts_p then
		(fun st ->
		    let (lst,form') = transform_bmf f ctx (Walk_stmt(st))
		    in
			([lst],stmt_of_walk_kind form'))
	    else
		nop
	and e =
	    if do_exprs_p then
		(fun es ->
		    let (lst,form') = transform_bmf f ctx (Walk_expr(es))
		    in
			([lst],expr_of_walk_kind form'))
	    else
		nop
	and t =
	    if do_types_p then
		(fun ty ->
		    let (lst,form') = transform_bmf f ctx (Walk_type(ty))
		    in
			([lst],type_of_walk_kind form'))
	    else
		nop
	and i =
	    if do_idents_p then
		(fun id ->
		    let (lst,form') = transform_bmf f ctx (Walk_ident(id))
		    in
			([lst],ident_of_walk_kind form'))
	    else
		nop
	and a =
	    if do_anns_p then
		(fun an ->
		    let (lst,form') = transform_bmf f ctx (Walk_ann(an))
		    in
			([lst],ann_of_walk_kind form'))
	    else
		nop
	in let rec map_preserve f lst =
	    match lst with
	      [] -> ([],[])
	    | x::rest -> 
		let (x_lst,x') = f x
		and (rest_lst,rest') = map_preserve f rest
		in
		    (x_lst@rest_lst,
		     (if x == x' && rest == rest' then
			lst
		      else
			x'::rest'))
	in let es = if do_exprs_p then (map_preserve e) else nop
	and ts = if do_types_p then (map_preserve t) else nop
	and is = if do_idents_p then (map_preserve i) else nop
	and as0 = if do_anns_p then (map_preserve a) else nop
	in let f0 =
	    if do_idents_p then
		(fun (Frame(ids)) ->
		    let (lst,ids') = is ids
		    in
			(lst,Frame(ids')))
	    else
		nop
	in
	    match form with
	       Walk_stmt(st) -> (
		match st with
		  Stmt_assign(lhs,rhs) ->
		    let (lhs_lst,lhs') = e lhs
		    and (rhs_lst,rhs') = e rhs
		    in
			(lhs_lst@rhs_lst,
			 (if lhs == lhs' && rhs == rhs' then
			    form
			  else
			    Walk_stmt(Stmt_assign(lhs',rhs'))))
		| Stmt_seq(first,second) ->
		    let (first_lst,first') = s first
		    and (second_lst,second') = s second
		    in
			(first_lst@second_lst,
			 (if first == first' && second == second' then
			    form
			  else
			    Walk_stmt(Stmt_seq(first',second'))))
		| Stmt_call(f,res,args) ->
		    let (f_lst,f') = e f
		    and (res_lst,res') = e res
		    and (args_lst,args') = es args
		    in	
			(f_lst@res_lst@args_lst,
			 (if f == f' && res == res' && args == args' then
			    form
			  else
			    Walk_stmt(Stmt_call(f',res',args'))))
		| Stmt_for(id,lb,ub,step,body) ->
		    let (id_lst,id') = i id
		    and (lb_lst,lb') = e lb
		    and (ub_lst,ub') = e ub
		    and (step_lst,step') = e step
		    and (body_lst,body') = s body
		    in
			(id_lst@lb_lst@ub_lst@step_lst@body_lst,
			 (if id == id' && lb == lb' && ub == ub' &&
				step == step' && body == body' then
			    form
			  else
			    Walk_stmt(Stmt_for(id',lb',ub',step',body'))))
		| Stmt_while(pred,body) ->
		    let (pred_lst,pred') = e pred
		    and (body_lst,body') = s body
		    in
			(pred_lst@body_lst,
			 (if pred == pred' && body == body' then
			    form
			  else
			    Walk_stmt(Stmt_while(pred',body'))))
		| Stmt_if(pred,thn,els) ->
		    let (pred_lst,pred') = e pred
		    and (thn_lst,thn') = s thn
		    and (els_lst,els') = s els
		    in
			(pred_lst@thn_lst@els_lst,
			 (if pred == pred' && thn == thn' && 
				els == els' then
			    form
			  else
			    Walk_stmt(Stmt_if(pred',thn',els'))))
		| Stmt_let(frm,body) ->
		    let (frm_lst,frm') = f0 frm
		    and (body_lst,body') = s body
		    in
			(frm_lst@body_lst,
			 (if frm == frm' && body == body' then
			    form
			  else
			    Walk_stmt(Stmt_let(frm',body'))))
		| Stmt_do(pred,body) ->
		    let (pred_lst,pred') = e pred
		    and (body_lst,body') = s body
		    in
			(pred_lst@body_lst,
			 (if pred == pred' && body == body' then
			    form
			  else
			    Walk_stmt(Stmt_do(pred',body'))))
		| Stmt_nop -> ([],form)
		| Stmt_directive(an,body) ->
		    let (an_lst,an') = a an
		    and (body_lst,body') = s body
		    in
			(an_lst@body_lst,
			 (if an == an' && body == body' then
			    form
			  else
			    Walk_stmt(Stmt_directive(an',body'))))
	    )| Walk_expr(ex) -> (
		match ex with
		  (Expr_int(_)|Expr_real(_)|Expr_string(_)|Expr_bool(_)
			|Expr_void|Expr_var(_))  ->
		    ([],form)
		| Expr_op(op,args) ->
		    let (args_lst,args') = es args
		    in
			(args_lst,
			 (if args == args' then
			    form
			  else
			    Walk_expr(Expr_op(op,args'))))
		| Expr_forall(frm,pred) ->
		    let (frm_lst,frm') = f0 frm
		    and (pred_lst,pred') = e pred
		    in
			(frm_lst@pred_lst,
			 (if frm == frm' && pred == pred' then
			    form
			  else
			    Walk_expr(Expr_forall(frm',pred'))))
		| Expr_exist(frm,pred) ->
		    let (frm_lst,frm') = f0 frm
		    and (pred_lst,pred') = e pred
		    in
			(frm_lst@pred_lst,
			 (if frm == frm' && pred == pred' then
			    form
			  else
			    Walk_expr(Expr_exist(frm',pred'))))
		| Expr_access(acc,r,name,args) -> 
		    let (r_lst,r') = e r
		    and (args_lst,args') = es args
		    in
			(r_lst@args_lst,
			 (if r == r' && args == args' then
			    form
			  else
			    Walk_expr(Expr_access(acc,r',name,args'))))
		| Expr_set(frm,pred) ->
		    let (frm_lst,frm') = f0 frm
		    and (pred_lst,pred') = e pred
		    in
			(frm_lst@pred_lst,
			 (if frm == frm' && pred == pred' then
			    form
			  else
			    Walk_expr(Expr_set(frm',pred'))))
	    )| Walk_type(ty) -> (
		match ty with
		  (Type_scalar(_)|Type_unknown) ->
		    ([], form)
		| Type_proc(args_tys,res_ty) ->
		    let (args_tys_lst,args_tys') = ts args_tys
		    and (res_ty_lst,res_ty') = t res_ty
		    in
			(args_tys_lst@res_ty_lst,
			 (if args_tys == args_tys' &&
				res_ty == res_ty' then
			    form
			  else
			    Walk_type(Type_proc(args_tys',res_ty'))))
		| Type_ann(an) ->
		    let (an_lst,an') = a an
		    in
			(an_lst,
			 (if an == an' then
			    form
			  else
			    Walk_type(Type_ann(an'))))
		| Type_ptr(base_typ) ->
		    let (base_typ_lst,base_typ') = t base_typ in
		    (base_typ_lst,
		     (if base_typ = base_typ' then
		       form
		     else
		       Walk_type(Type_ptr(base_typ'))))
		| Type_array(atyp) ->
		    let rec recurs atyp =
		      match atyp with
		 	Array_dense(lb,ubs,maj,base_ty) ->
			  let (ubs_lst,ubs') = es ubs
			  in
			  (ubs_lst,
			   (if ubs == ubs' then
			     atyp
			   else
			    Array_dense(lb,ubs',maj,base_ty)))
		      | Array_view(out_frm,in_frm,constr,base_atyp) ->
			  let (out_frm_lst,out_frm') = f0 out_frm
			  and (in_frm_lst,in_frm') = f0 in_frm
			  and (constr_lst,constr') = e constr
			  and (base_atyp_lst,base_atyp') = recurs base_atyp
			  in
			  (out_frm_lst@in_frm_lst@constr_lst@base_atyp_lst,
			   (if out_frm == out_frm' && in_frm == in_frm' &&
			     constr == constr' && base_atyp == base_atyp' then
			     atyp
			   else
			     Array_view(out_frm',in_frm',constr',base_atyp')))
		      | Array_bb(dim,base_ty,bb_name,bb_args) ->
			  let rec recurs bb_args =
			    match bb_args with
			      [] -> ([],bb_args)
			    | ((tag,val0) as bb_arg)::rest_bb_args ->
			  	let (rest_bb_args_lst,rest_bb_args') =
				  recurs rest_bb_args
			  	and (val0_lst,val0') = a val0
			  	in let bb_arg' =
				  if val0 == val0' then
				    bb_arg
				  else
				    (tag,val0')
			  	in let bb_args' =
				  if bb_arg == bb_arg' &&
				    rest_bb_args == rest_bb_args' then
				    bb_args
				  else
				    bb_arg'::rest_bb_args'
			  	in
			  	(val0_lst@rest_bb_args_lst, bb_args')
			  in let (bb_args_lst,bb_args') =
		            recurs bb_args
			  in
		          (bb_args_lst,
			   (if bb_args == bb_args' then
			     atyp
			   else
			     Array_bb(dim,base_ty,bb_name,
				      bb_args')))
		    in let (atyp_lst,atyp') = recurs atyp
		    in
		    (atyp_lst,
		     (if atyp == atyp' then
		       form
		     else
		       Walk_type(Type_array(atyp'))))
	    )| Walk_ident(id) -> (
		match id with
		  Ident_valid(name,ty,stuff,an) ->
		    let (stuff_lst,stuff') = (
			match stuff with
			  Ident_var(init) ->
			    let (init_lst,init') = e init
			    in
				(init_lst,
				 (if init == init' then
				    stuff
				  else
				    Ident_var(init')))
			| Ident_bind -> ([],stuff)
			| Ident_proc(args,res,body) ->
			    let (args_lst,args') = f0 args
			    and (res_lst,res') = i res
			    and (body_lst,body') = s body
			    in
				(args_lst@res_lst@body_lst,
				 (if args == args' && res == res' &&
					body == body' then
				    stuff
				  else
				    Ident_proc(args',res',body')))
		    ) in let (ty_lst,ty') = t ty
		    and (an_lst,an') = a an
		    in
			(ty_lst@stuff_lst@an_lst,
			 (if ty == ty' && stuff == stuff' && an == an' then
			    form
			  else
			    Walk_ident(Ident_valid(name,ty',stuff',an'))))
		| Ident_null -> ([],form)
	    )| Walk_ann(an) -> (
		match an with
		  An_type(ty) ->
		    let (ty_lst,ty') = t ty
		    in
			(ty_lst,
			 (if ty == ty' then
			    form
			  else
			    Walk_ann(An_type(ty'))))
		| An_stmt(st) ->
		    let (st_lst,st') = s st
		    in
			(st_lst,
			 (if st == st' then
			    form
			  else
			    Walk_ann(An_stmt(st'))))
		| An_expr(ex) ->
		    let (ex_lst,ex') = e ex
		    in
			(ex_lst,
			 (if ex == ex' then
			    form
			  else
			    Walk_ann(An_expr(ex'))))
		| An_string(s0) -> ([],form)
		| An_list(ans) ->
		    let (ans_lst,ans') = as0 ans
		    in
			(ans_lst,
			 (if ans == ans' then
			    form
			  else
			    Walk_ann(An_list(ans'))))
	    )
 in next_action (f ctx form)

    
: 'b * walk_kind)
;;

(************************************************************************)

let rec map_bmf (f: 'a -> walk_kind -> 'a map_bmf_action)
	(ctx: 'a) (form : walk_kind) = (
    let rec do_action a =
	match a with
	  Mba_return(form') -> Tba_return((),form')
	| Mba_recurs(ctx',form',contin) ->
	    Tba_recurs(ctx',form',
		(fun _ form' -> do_action (contin form')))
	| Mba_visit(whats,ctx',contin) ->
	    Tba_visit(whats,ctx',
		(fun _ form' -> do_action (contin form')))
    in let (_,form') =
	transform_bmf
	    (fun ctx form -> do_action (f ctx form))
	    ctx form
    in
	form'
: walk_kind);;

(************************************************************************)

let rec it_bmf (f: 'a -> walk_kind -> ('a,'b) it_bmf_action)
	(ctx: 'a) (form : walk_kind) = (
    let rec do_action form a =
	match a with
	  Iba_return(val0) -> Tba_return(val0,form)
	| Iba_recurs(ctx',form',contin) ->
	    Tba_recurs(ctx',form',
		(fun vals form' -> do_action form' (contin vals)))
	| Iba_visit(whats,ctx',contin) ->
	    Tba_visit(whats,ctx',
		(fun vals form' -> do_action form' (contin vals)))
    in let (val0,_) =
	transform_bmf (fun ctx form -> do_action form (f ctx form))
	    ctx form
    in
	val0
: 'b);;

(************************************************************************)

let transform_bmf_ident f ctx id =
    let (res,form) = transform_bmf f ctx (Walk_ident(id))
    in
	(res,ident_of_walk_kind(form))
;;

let transform_bmf_stmt f ctx id =
    let (res,form) = transform_bmf f ctx (Walk_stmt(id))
    in
	(res,stmt_of_walk_kind(form))
;;

let transform_bmf_expr f ctx id =
    let (res,form) = transform_bmf f ctx (Walk_expr(id))
    in
	(res,expr_of_walk_kind(form))
;;

let transform_bmf_type f ctx id =
    let (res,form) = transform_bmf f ctx (Walk_type(id))
    in
	(res,type_of_walk_kind(form))
;;

let transform_annotation f ctx id =
    let (res,form) = transform_bmf f ctx (Walk_ann(id))
    in
	(res,ann_of_walk_kind(form))
;;

(************************************************************************)

let transform_bmf_idents f ctx op init ids =
    let rec recurs ids =
	match ids with
	  [] -> (init,[])
	| id::rest_ids ->
	    let (id_result,id') = transform_bmf_ident f ctx id
	    and (rest_ids_result,rest_ids') = recurs rest_ids
	    in
		(op id_result rest_ids_result,
		 if id == id' && rest_ids == rest_ids' then
		    ids
		 else
		    id'::rest_ids')
    in
	recurs ids
;;

let transform_bmf_stmts f ctx op init ids =
    let rec recurs ids =
	match ids with
	  [] -> (init,[])
	| id::rest_ids ->
	    let (id_result,id') = transform_bmf_stmt f ctx id
	    and (rest_ids_result,rest_ids') = recurs rest_ids
	    in
		(op id_result rest_ids_result,
		 if id == id' && rest_ids == rest_ids' then
		    ids
		 else
		    id'::rest_ids')
    in
	recurs ids
;;

let transform_bmf_exprs f ctx op init ids =
    let rec recurs ids =
	match ids with
	  [] -> (init,[])
	| id::rest_ids ->
	    let (id_result,id') = transform_bmf_expr f ctx id
	    and (rest_ids_result,rest_ids') = recurs rest_ids
	    in
		(op id_result rest_ids_result,
		 if id == id' && rest_ids == rest_ids' then
		    ids
		 else
		    id'::rest_ids')
    in
	recurs ids
;;

let transform_bmf_types f ctx op init ids =
    let rec recurs ids =
	match ids with
	  [] -> (init,[])
	| id::rest_ids ->
	    let (id_result,id') = transform_bmf_type f ctx id
	    and (rest_ids_result,rest_ids') = recurs rest_ids
	    in
		(op id_result rest_ids_result,
		 if id == id' && rest_ids == rest_ids' then
		    ids
		 else
		    id'::rest_ids')
    in
	recurs ids
;;

let transform_annotations f ctx op init ids =
    let rec recurs ids =
	match ids with
	  [] -> (init,[])
	| id::rest_ids ->
	    let (id_result,id') = transform_annotation f ctx id
	    and (rest_ids_result,rest_ids') = recurs rest_ids
	    in
		(op id_result rest_ids_result,
		 if id == id' && rest_ids == rest_ids' then
		    ids
		 else
		    id'::rest_ids')
    in
	recurs ids
;;

(************************************************************************)
(* Deep copying								*)
(************************************************************************)

let rec deep_copy_ident id =
    let i = deep_copy_ident
    and f = deep_copy_frame
    and s = deep_copy_stmt
    and e = deep_copy_expr
    and a = deep_copy_ann
    and t = deep_copy_type
    in
	match id with
	  Ident_valid(name,ty,stuff,an) ->
	    let stuff' =
		match stuff with
		  Ident_var(init) -> Ident_var(e init)
		| Ident_bind -> Ident_bind
		| Ident_proc(args,res,body) ->
		    Ident_proc(f args,i res,s body)
	    in
		Ident_valid(name,t ty,stuff',a an)
	| Ident_null -> Ident_null

and deep_copy_frame (Frame(ids)) = Frame(List.map deep_copy_ident ids)

and deep_copy_stmt st =
    let i = deep_copy_ident
    and f = deep_copy_frame
    and s = deep_copy_stmt
    and e = deep_copy_expr
    and a = deep_copy_ann
    in let es = List.map e
    in
	match st with
	  Stmt_assign(lhs,rhs) -> Stmt_assign(e lhs,e rhs)
	| Stmt_seq(first,second) -> Stmt_seq(s first,s second)
	| Stmt_call(f0,res,args) -> Stmt_call(e f0,e res,es args)
	| Stmt_for(id,lb,ub,step,body) ->
	    Stmt_for(i id,e lb,e ub,e step,s body)
	| Stmt_while(pred,body) -> Stmt_while(e pred,s body)
	| Stmt_if(pred,thn,els) -> Stmt_if(e pred,s thn,s els)
	| Stmt_let(frm,body) -> Stmt_let(f frm,s body)
	| Stmt_do(pred,body) -> Stmt_do(e pred,s body)
	| Stmt_nop -> Stmt_nop
	| Stmt_directive(an,body) -> Stmt_directive(a an,s body)

and deep_copy_expr ex =
    let f = deep_copy_frame
    and e = deep_copy_expr
    in let es = List.map e
    in
	match ex with
	  Expr_int(x) -> Expr_int(x)
	| Expr_real(x) -> Expr_real(x)
	| Expr_string(x) -> Expr_string(x)
	| Expr_bool(x) -> Expr_bool(x)
	| Expr_void -> Expr_void
	| Expr_var(x) -> Expr_var(x)
	| Expr_op(op,args) -> Expr_op(op,es args)
	| Expr_forall(frm,pred) -> Expr_forall(f frm,e pred)
	| Expr_exist(frm,pred) -> Expr_exist(f frm,e pred)
	| Expr_access(acc,r,name,args) -> Expr_access(acc,e r,name,es args)
	| Expr_set(frm,pred) -> Expr_set(f frm,e pred)

and deep_copy_type ty =
    let e = deep_copy_expr
    and a = deep_copy_ann
    and t = deep_copy_type
    in let es = List.map e
    and ts = List.map t
    in
	match ty with
	  Type_scalar(sc_ty) -> Type_scalar(sc_ty)
	| Type_proc(args_tys,res_ty) -> Type_proc(ts args_tys,t res_ty)
	| Type_array(atyp) -> Type_array(deep_copy_array_type atyp)
	| Type_ptr(btyp) -> Type_ptr(t btyp)
	| Type_ann(an) -> Type_ann(a an)
	| Type_unknown -> Type_unknown

and deep_copy_array_type atyp =
    let f = deep_copy_frame
    and a = deep_copy_ann
    and e = deep_copy_expr
    and at = deep_copy_array_type
    in let es = List.map e
    in
    match atyp with
      Array_bb(d,base_ty,bb_name,bb_args) -> 
	let bb_args' =
	  List.map (fun (tag,val0) -> (tag, a val0)) bb_args
	in Array_bb(d,base_ty,bb_name,bb_args')
    | Array_dense(lb,ubs,maj,base_ty) ->
	Array_dense(lb,es ubs,maj,base_ty)
    | Array_view(out_frm,in_frm,constr,base_atyp) ->
     	Array_view(f out_frm,f in_frm,e constr, at base_atyp) 

and deep_copy_ann an =
    let s = deep_copy_stmt
    and e = deep_copy_expr
    and a = deep_copy_ann
    and t = deep_copy_type
    in let as0 = List.map a
    in
	match an with
	  An_type(ty) -> An_type(t ty)
	| An_stmt(st) -> An_stmt(s st)
	| An_expr(ex) -> An_expr(e ex)
	| An_string(s0) -> An_string(s0)
	| An_list(ans) -> An_list(as0 ans)
;;


