(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Bsct_base

(**************************************************************)
(*              Expression manipulation                       *)
(**************************************************************)

(*
module type Ordered_type_with_printer =
  sig
    type t
    val compare : t -> t -> int
    val to_string : t -> string
  end
*)

module type E =
  sig
    module ATOM : AtomType
    type atom = ATOM.t

    exception Invalid_s_expr of Sx.t

    type cnst =
      | Int of int
      | Real of float
      | String of string
      | Bool of bool
      | Void

    type op =
	Plus
      | Times
      | Neg
      | Div
      | Sqrt
      | Not
      | And
      | Or
      | Leq
      | Eq
      |	Lt
      |	Gt
      |	Geq
      |	Apply

    type expr = 
	Atom of atom 
      | Const of cnst 
      | Op of op * expr list

    val op_arg : expr -> expr
    val op_arg1 : expr -> expr
    val op_arg2 : expr -> expr
    val op_args : expr -> expr list
    val args_by_op : op -> expr -> expr list
    val make_nary_expr : op -> expr -> expr -> expr list -> expr
    val make_neg : expr -> expr
    val make_not : expr -> expr
    val make_plus : expr list -> expr
    val make_minus : expr -> expr -> expr
    val make_times : expr list -> expr
    val make_and : expr list -> expr
    val make_or : expr list -> expr
    val make_binary : op -> expr -> expr -> expr
    val collect_atoms : expr -> atom list
    val rename : atom list -> atom list -> expr -> expr

    val to_sexpr : expr -> Sx.t
    val to_string : expr -> string

    module Type :
        sig
          type t =
            | Atom
            | Num
            | Int
            | Real
            | String
            | Bool
            | Void
            | Top
	  val type_lub : t -> t -> t
	  val type_lub_list : t list -> t
	  val apply_op : op -> t list -> t
        end
    val expr_type : expr -> Type.t
  end


module Make(A : AtomType) =
  struct

    module ATOM = A
    type atom = A.t

    exception Invalid_s_expr of Sx.t


(*
    let atom_compare = A.compare
    let atom_to_string = A.to_string *)

    type cnst =
	Int of int
      | Real of float
      | String of string
      | Bool of bool
      | Void

    type op =
	Plus
      | Times
      | Neg
      | Div
      | Sqrt
      | Not
      | And
      | Or
      | Leq
      | Eq
      |	Lt
      |	Gt
      |	Geq
      |	Apply

    type expr = 
	Atom of atom
      | Const of cnst
      | Op of op * (expr list)


    let op_arg e =
      match e with
	Op(_,[arg]) -> arg
      | _ -> failwith "op_arg: Must be Op with a single argument"

    let op_arg1 e =
      match e with
	Op(_,[arg1;arg2]) -> arg1
      | _ -> failwith "op_arg1: Must be Op with exactly two args"

    let op_arg2 e =
      match e with
	Op(_,[arg1;arg2]) -> arg2
      | _ -> failwith "op_arg2: Must be Op with exactly two args"

    let op_args e = 
      match e with
	Op(_,args) -> args
      | _ -> failwith "op_args: Must be Op"

      (* args_by_op "+" (+ 1 (+ 2) (+ 3 (+ 4 5) 6)) -> [1;2;3;4;5;6] *)
    let rec args_by_op op e =
      match e with
	Op(op',args) when (op' = op) ->
          List'.flat_map (args_by_op op) args
      | _ -> [e]


    let make_nary_expr op identity annihil args =
      let args' = List'.flat_map (args_by_op op) args
      in let args'' = List'.difference args' [identity]
      in
      match args'' with
	[] -> identity
      | [arg] -> arg
      | _ when List.mem annihil args'' -> annihil
      | args'' -> Op(op,args'')
	    
    let rec make_neg e =
      match e with
	Op(Neg,[e']) -> e'
      | Op(Plus,args) -> Op(Plus, List.map make_neg args)
      | Const(Int(x)) -> Const(Int(-x))
      | Const(Real(x)) -> Const(Real(-.x))
      | _ -> Op(Neg,[e])

    let make_not e =
      match e with
	Op(Not,[e']) -> e'
      | _ -> Op(Not,[e])

    let make_plus = 
      make_nary_expr Plus (Const(Int(0))) (Const(Void))

    let make_minus e1 e2 =
      make_plus [e1; Op(Neg,[e2])]

    let make_times = 
      make_nary_expr Times (Const(Int(1))) (Const(Int(0)))

    let make_and =
      make_nary_expr And (Const(Bool(true))) (Const(Bool(false))) 

    let make_or =
      make_nary_expr Or (Const(Bool(false))) (Const(Bool(true)))

    let make_binary op e1 e2 =
      Op(op,[e1;e2])

    let rec collect_atoms e =
      match e with
	Atom(a) -> [a]
      | Op(_,l') -> List'.flat_map collect_atoms l'
      | _ -> []

(**************
    module Op_comparator =
      struct
	type t = op
	let compare = Pervasives.compare
      end

    module Op_map_helper = Map'.Make(Op_comparator)
    module Op_map = Op_map_helper.S

    let op_name_map = Op_map_helper.absorb_list Op_map.empty 
	[ (Plus, "Plus") ;
	  (Neg, "Neg") ;
	  (Times, "Times") ;
    	  (Div, "Div") ;
    	  (Sqrt, "Sqrt") ;
    	  (Not, "Not") ;
    	  (And, "And") ;
    	  (Or, "Or") ;
    	  (Leq, "Leq") ;
    	  (Geq, "Geq") ;
    	  (Lt, "Lt") ;
    	  (Gt, "Gt") ;
    	  (Eq, "Eq") ;
	  (Apply, "Apply") ]

    let rec to_string e =
      let rec cat_list (l:(string list)) = 
	match l with
	  [] -> ""
	| [s] -> s
	| s::t::rest -> s^","^(cat_list (t::rest)) in
      match e with
	Op(op,l) -> (Op_map.find op op_name_map)^"("^
	  (cat_list (List.map to_string l))^")"
      |	Atom(a) -> ATOM.to_string a
      |	Const(Int(i)) -> string_of_int i
      |	Const(Real(g)) -> string_of_float g
      |	Const(String(s)) -> "\""^s^"\""
      |	Const(Bool(b)) -> string_of_bool b
      |	Const(Void) -> "#void"

***************)

    let rename from_names to_names e =
      let l2 = List.combine from_names to_names in
      let rec rename_aux e' =
	match e' with
	  Atom(a) -> 
	    (try Atom(List.assoc a l2)
	    with Not_found -> Atom(a))
	| Op(op,l) -> Op(op, List.map rename_aux l)
	| _ -> e'
      in
      (rename_aux e)


    (********************************************************
     *		       Conversion to sexpr
     *********************************************************)

    module From_expr =
      struct
    	module Expr_ops =
	  struct 
	    type t = op
	    let compare = Pervasives.compare
	  end
	    
    	module Op_map_helper = Map'.Make(Expr_ops)
    	module Op_map = Op_map_helper.S
	    
    	let op_name_map = 
	  Op_map_helper.absorb_list Op_map.empty 
	    [ (Plus,  "+") ;
	      (Neg,   "-") ;
	      (Times, "*") ;
    	      (Div,   "/") ;
    	      (Not,   "!") ;
    	      (Sqrt,  "sqrt") ;
    	      (And,   "&&") ;
    	      (Or,    "||") ;
    	      (Leq,   "<=") ;
    	      (Eq,    "=") ;
    	      (Geq,   ">=") ;
    	      (Lt,    "<") ;
    	      (Gt,    ">") ]
      end

    let rec to_sexpr ex =
      let find_op op = 
      	try (Sx.Symbol (From_expr.Op_map.find op From_expr.op_name_map))
      	with Not_found -> failwith "to_s_expr: operator not found" in
      match ex with
	(Op(Apply, l)) -> (Sx.from_list (List.map to_sexpr l))
      |	(Op(op, l)) -> (Sx.from_list ((find_op op)::(List.map to_sexpr l)))
      |	(Atom s) -> ATOM.to_sexpr s
      |	(Const (Void)) -> Sx.Nil
      |	(Const (Real f)) -> Sx.Float f
      |	(Const (Int i)) -> Sx.Int i
      |	(Const (Bool b)) -> Sx.Bool b
      |	(Const (String s)) -> Sx.String s

    let to_string ex = (Sx.to_string (to_sexpr ex))

    module Type =
      struct
	type t = 
	    Atom | Num | Int | Real 
	  | String | Bool | Void | Top

	let type_lub t1 t2 =
	  match (t1,t2) with
	    (Atom,_) -> t2
	  | (_,Atom) -> t1
	  | ((Top,_)|(_,Top)) -> Top
	  | ((Num,Int)|(Int,Num)) -> Int
	  | ((Num,Real)|(Real,Num)) -> Real
	  | ((Int,Real)|(Real,Int)) -> Real
	  | _ -> if (t1 = t2) then t1 else Top

	let type_lub_list l =
	  (List.fold_left type_lub Atom l)
	    
	let apply_op op l =
	    (* NOTE: the type of "no inputs" is Atom *)
	  let t = type_lub_list l in
	  match (op,t) with
	    ((Plus|Times|Neg|Div),(Int|Real|Num)) -> t
	  | ((Plus|Times|Neg|Div),Atom) -> Num
	  | (Sqrt,(Int|Real|Num|Atom)) -> Real
	  | ((Plus|Times|Neg|Div|Sqrt),_) -> Top
	  | ((Not|And|Or),(Bool|Atom)) -> Bool
	  | ((Leq|Eq|Lt|Gt),(Bool|Void)) -> Top
	  | ((Leq|Eq|Lt|Gt),_) -> Bool
	  | (Apply,_) -> Top
	  | _ -> Top

      end
	
    let rec expr_type e =
      match e with
	Atom(_) -> Type.Atom
      | Const(Int(_)) ->  Type.Int
      | Const(Real(_)) -> Type.Real
      | Const(String(_)) -> Type.String
      | Const(Void) -> Type.Void
      | Const(Bool(_)) -> Type.Bool
      | Op(op,l) -> Type.apply_op op (List.map expr_type l)
	    
  end


