(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Bsct_base

module E = Expr.Make(SimpleAtom)

module Input_for_lambda =
  struct
    module EG = 
      struct
    	module Ex = E
    	let gensym = (fun () -> (Misc.gensym "t"))
    	let gensyms = (Misc.gensyms "t")
      end

    module Key = SimpleAtom
  end

module L = Lambda_expr.Make(Input_for_lambda)
module L_affine = Lambda_affine.Make(L)
module A = L_affine.A

module Convert =
  struct


    (***************************************************)
    (*  Conversion from s_expr                         *)
    (***************************************************)
    exception Invalid_s_expr of Sx.t

    module To_expr =
      struct

	module Op_map_helper = Map'.Make(SimpleAtom)
	module Op_map = Op_map_helper.S
	    
	type args_upper_bound = Num of int | Inf
	and op_spec = Op_spec of E.op * int * args_upper_bound
	    (* (op,minargs,maxargs,trans) *)

	let op_name_map = 
	  Op_map_helper.absorb_list Op_map.empty 
	    [ ("+",    Op_spec(E.Plus ,1,Inf)   ) ;
	      ("-",    Op_spec(E.Neg  ,1,Num 2) ) ;
	      ("*",    Op_spec(E.Times,1,Inf)   ) ;
    	      ("/",    Op_spec(E.Div  ,2,Num 2) ) ;
    	      ("!",    Op_spec(E.Not  ,1,Num 1) ) ;
    	      ("sqrt", Op_spec(E.Sqrt ,1,Num 1) ) ;
    	      ("&&",   Op_spec(E.And  ,2,Inf)   ) ;
	      ("and",  Op_spec(E.And  ,2,Inf)   ) ;
    	      ("||",   Op_spec(E.Or   ,2,Inf)   ) ;
    	      ("<=",   Op_spec(E.Leq  ,2,Num 2) ) ;
    	      ("=",    Op_spec(E.Eq   ,2,Num 2) ) ;
    	      (">=",   Op_spec(E.Geq  ,2,Num 2) ) ;
    	      ("<",    Op_spec(E.Lt   ,2,Num 2) ) ;
    	      (">",    Op_spec(E.Gt   ,2,Num 2) ) ]

	let is_within_arg_bounds min_arg max_arg n =
	  match max_arg with
	    Inf -> (min_arg <= n)
	  | Num m -> (min_arg <= n && n <= m)

      end
	
    let rec from_sexpr s_expr =
      if (Sx.list_p s_expr) then 
	begin
	  match (Sx.to_list s_expr) with
	    [] -> E.Const(E.Void)
	  | (Sx.Symbol sym)::l' -> 
	      let l_expr = (List.map from_sexpr l') in
	      let arg_len = List.length l_expr in
	      begin
	      	try
	      	  let To_expr.Op_spec(op,min_arg,max_arg) = 
		    (To_expr.Op_map.find sym To_expr.op_name_map) in
		  if (To_expr.is_within_arg_bounds min_arg max_arg arg_len) then
		    if (op = E.Neg) then
		      if (arg_len = 1) then E.Op(E.Neg, l_expr)
		      else E.Op(E.Plus, 
				[List.hd l_expr;
				  E.Op(E.Neg, [List.hd (List.tl l_expr)])])
		    else
		      E.Op(op, l_expr)
		  else
		    raise (Invalid_s_expr s_expr)
	      	with Not_found ->
	      	  (E.Op(E.Apply, 
		      	(E.Const(E.String sym))::l_expr))
	  	| _ -> raise (Invalid_s_expr s_expr)
	      end
	  | _ -> raise (Invalid_s_expr s_expr)
      	end
      else begin
	match s_expr with
	  Sx.Nil -> failwith "from_s_expr: Nil is a list!!"
	| (Sx.Symbol sym) -> E.Atom sym
	| (Sx.Pair _) -> raise (Invalid_s_expr s_expr)
	| (Sx.Vector _) -> raise (Invalid_s_expr s_expr)
	| (Sx.Bool b) -> E.Const (E.Bool b)
	| (Sx.Char c) -> 
	    let s = String.create 1 in 
	    String.set s 0 c; (E.Const (E.String s))
	| (Sx.String s) -> (E.Const (E.String s))
	| (Sx.Int i) -> (E.Const (E.Int i))
	| (Sx.Float f) -> (E.Const (E.Real f))
	| (Sx.Other _) -> raise (Invalid_s_expr s_expr)
      end


    (*********************************************************)
    (* simply inherit the converter from the Expr module     *)
    (*********************************************************)
    let to_sexpr = E.to_sexpr
	    
      
  end


(***********************************************************
 *	   Lambda expressions instantiated for string atoms
 ***********************************************************)

module L_convert =
  struct
    let a_sym = Sx.Symbol "A" 
    and b_sym = Sx.Symbol "B" 
    and sub_sym = Sx.Symbol "Sub"

    let parse_desc sexpr =
      match (Sx.to_list sexpr) with
	[sym; Sx.Symbol(key1); Sx.Symbol(key2)] when sym = a_sym -> L.A(key1,key2)
      |	[sym; Sx.Symbol(key1); Sx.Symbol(key2)] when sym = b_sym -> L.B(key1,key2)
      |	[sym; Sx.Symbol(key1); Sx.Symbol(key2)] 
	  when sym = sub_sym -> L.Sub(key1,key2)
      |	_ -> failwith ("L_convert.parse_desc: invalid s_expr: "^
		       (Sx.to_string sexpr))

    let unparse_desc desc =
      match desc with
	L.A(key1,key2) ->
	  (Sx.from_list [ a_sym; Sx.Symbol(key1); Sx.Symbol(key2) ])
      |	L.B(key1,key2) ->
	  (Sx.from_list [ b_sym; Sx.Symbol(key1); Sx.Symbol(key2) ])
      |	L.Sub(key1,key2) ->
	  (Sx.from_list [ a_sym; Sx.Symbol(key1); Sx.Symbol(key2) ])

    (*********************************************************)
    let parse sexpr = begin
      (* the format is:
	 (key1 a1) (key2 a2) ... (keyn an) expr
	 where ai are lists of variables 
	 each ai can be empty
	 *)
      
      let rec parse' accumulator lst = (
	match lst with
	  (* this branch accumulates (key,list) pairs *)
	  (Sx.Pair(Sx.Symbol(k),var_lst_expr))::form'::rest ->
	    let new_binding = 
	      (k, (List.map Sx.untag_symbol (Sx.to_list var_lst_expr))) in
	    (parse' (new_binding::accumulator) (form'::rest))

	| (* this branch catches the predicate expression at the end *)
	  [form] -> 
	    let the_predicate = (
	      try (Convert.from_sexpr form)
	      with 
	    	Convert.Invalid_s_expr(_) ->
		  raise (SimpleAtom.Invalid_s_expr sexpr)) in
	    (L.make { L.bindings=accumulator ;
		      L.predicate=the_predicate })

	| [] -> failwith("L_convert.parse: expr is too short")
	| _ -> failwith ("L_convert.parse: error")) in

      (parse' [] (Sx.to_list sexpr))
    end (* parse sexpr *)

    (*********************************************************)
    let unparse lbda = begin
      let { L.bindings = binds; 
	    L.predicate = the_predicate} =
	(L.destructure lbda) in

      let unparse_binding (k,lst) =
	(Sx.from_list 
	   (List.map 
	      (fun str -> Sx.Symbol(str)) 
	      (k::lst))) in

      (Sx.from_list
	 (List.fold_left
	    (fun accumulator bndg -> 
	      (unparse_binding bndg)::accumulator)
	    [Convert.to_sexpr the_predicate]
	    binds))
    end (* unparse lbda *)
  end
