(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Printf;;
open Bsct_base
open Ssisl
open Containers
open Bmf;;

(*
Bug : The local array allocation stuff assumes that there is no recursion.
*)

(************************************************************************)
(* Intrinsics								*)
(************************************************************************)

(* Hooks for other passes *)

type intrinsic_result =
    Ir_string of string
|   Ir_stmt of bmf_stmt
;;

let header_funcs = ref [];;
let add_header_func (f:out_channel -> unit) =
    header_funcs := f::(!header_funcs);;

let intrinsic_handlers = ref [];;
let add_intrinsic_handler (name:string)
	(f:bmf_stmt -> intrinsic_result) =
    intrinsic_handlers := (name,f)::(!intrinsic_handlers);;

(* Some useful intrinsics. *)

add_intrinsic_handler "%halt"
    (fun st ->
      match st with
    	Stmt_call(Expr_var(_),Expr_void,[]) ->
	  Ir_stmt
	    (Stmt_call(Expr_var("assert"),Expr_void,[Expr_bool(false)]));
      | _ -> assert false (* Can't happen *))
    ;;

(* Expand the intrinsics in a procedure. *)

let rec expand_intrinsics id =
  match id with
    Ident_valid(name,typ,Ident_proc(Frame(args),res,body),ann) ->
      let body' = expand_intrinsics_stmt body in
      Ident_valid(name,typ,Ident_proc(Frame(args),res,body'),ann)
  | _ -> id

and expand_intrinsics_stmt st =
  let r = expand_intrinsics_stmt in
  match st with
    Stmt_call(Expr_var(f_name),result,args) ->
      (try 
	let ir = (List.assoc f_name !intrinsic_handlers) st in
	match ir with
	  Ir_stmt(st') -> r st'
	| Ir_string _ -> failwith "Not yet implemented."
      with Not_found -> st)
  | (Stmt_assign _ | Stmt_call _ | Stmt_nop) -> st
  | Stmt_seq(s1,s2) -> Stmt_seq(r s1,r s2)
  | Stmt_for(id,lb,ub,step,body) -> Stmt_for(id,lb,ub,step,r body)
  | Stmt_while(pred,body) -> Stmt_while(pred,r body)
  | Stmt_if(test,thn,els) -> Stmt_if(test,r thn,r els)
  | Stmt_let(frm,body) -> Stmt_let(frm,r body)
  | Stmt_do(pred,body) -> Stmt_do(pred,r body)
  | Stmt_directive(an,body) -> Stmt_directive(an,r body)

let _ =
  Sx.define_prim1 (Scheme.interaction_environment ())
    "expand_intrinsics"
    (fun sx -> 
      let bmf = Bmf.from_s_expr sx in
      let bmf' = List.map expand_intrinsics bmf in
      let sx' = Bmf.to_s_expr bmf' in
      sx')

(************************************************************************)
(* Indentation								*)
(************************************************************************)

let max_padding = 100;;

let indent =
    let v = Array.create max_padding ""
    in
	for i = 0 to max_padding-1 do
	    Array.set v i (String.make i ' ')
	done;
	(fun i ->
	    if i = -1 then
		" "
	    else
		Array.get v i)
;;

let newline i =
    if i = -1 then
	""
    else
	"\n"
;;

let indent_incr i =
    if i = -1 then
	-1
    else
	i+4
;;

(************************************************************************)
(* Utilities								*)
(************************************************************************)

let iter2 f lst1 lst2 =
    assert ((List.length lst1) = (List.length lst2)) (* (Failure("Foo!")) *);
    List.iter2 f lst1 lst2
;;

let real_type = ref "double";;

let lowest_prec = 0
and infin_prec = 9999;;
let op_precedence op =
    match op with
      Op_or -> 1
    | Op_and -> 2
    | (Op_leq|Op_eq) -> 3
    | Op_plus -> 4
    | (Op_times|Op_div) -> 5
    | _ -> infin_prec
;;

let rec ann_to_list ann =
    match ann with
      An_list(lst) -> lst
    | _ -> failwith "ann_to_list: not a list"

and is_ref_param lst =
    match lst with
      An_string("value")::lst -> false
    | An_string("ref")::lst -> true
    | x::lst -> is_ref_param lst
    | [] -> false
;;

let paren_list outchan lst f =
    fprintf outchan "(";
    (match lst with
      x::lst' ->
	f x;
	List.iter (fun x ->
		fprintf outchan ",";
		f x)
	    lst'
    | [] -> ());
    fprintf outchan ")"
;;

(************************************************************************)
(* Codegen Environment							*)
(************************************************************************)

let rec cgi_search str lst default_f =
    match lst with
      (str',info)::lst' ->
	if str' = str then
	    info
	else
	    cgi_search str lst' default_f
    | [] -> default_f ()
;;

let cgi_rename str env =
    let (renames,dense_info) = env
    in
	cgi_search str renames (fun () -> str)
;;

let cgi_dense_info str env =
    let (renames,dense_info) = env
    in
	cgi_search str dense_info (fun () -> failwith "No dense info")
;;

(************************************************************************)
(* Entry points								*)
(************************************************************************)

let rec to_channel outchan decls =
    print_header outchan;
    List.iter (fun decl ->
		print_top_decl outchan 0 decl true)
	decls;
    fprintf outchan "\n";
    List.iter (fun decl ->
		print_top_def outchan decl)
	decls;
    flush outchan
and to_file str decls =
    Io'.call_with_open_output_file str
	(fun outchan -> to_channel outchan decls)
and to_stdout decls =
    to_channel stdout decls

and print_header outchan =
    fprintf outchan "
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>

#ifndef BMF_INTRINSICS_DEFINED
#define BMF_INTRINSICS_DEFINED
#ifdef __cplusplus
static inline bool IMPLIES(bool x,bool y) {
    return (bool)(!x || y); }
static inline int MIN(int x,int y) {
    if (x < y) return x; else return y; }
static inline int MAX(int x,int y) {
    if (x > y) return x; else return y; }
static inline double MIN(double x,double y) {
    if (x < y) return x; else return y; }
static inline double MAX(double x,double y) {
    if (x > y) return x; else return y; }
#else
#define IMPLIES(x,y) (!(x) || (y))
#define MIN(x,y) ((x) < (y) ? (x) : (y))
#define MAX(x,y) ((x) > (y) ? (x) : (y))
#endif
#endif

#ifndef BMF_INLINE
#define BMF_INLINE
#endif

";
    List.iter (fun f -> f outchan; fprintf outchan "\n")
	!header_funcs

(************************************************************************)
(* Forward declarations							*)
(************************************************************************)

and print_top_decl outchan i decl extern_p =
    let (name,typ,stuff,ann) =
	match decl with
	  Ident_valid(name,typ,stuff,ann) -> (name,typ,stuff,ann)
	| Ident_null -> failwith "ident_null"
    in
	fprintf outchan "%s" (indent i);
	(match stuff with
	  Ident_var(_) ->
	    if extern_p then fprintf outchan "extern ";
	    c_type outchan typ name
	| Ident_proc(Frame(args),result,_) ->
	    print_proc_header outchan name result args
	| Ident_bind -> failwith "Extern ident_bind??");
	fprintf outchan ";%s" (newline i)
	
and print_internal_decl outchan i decl env =
    let (name,typ,stuff,ann) =
	match decl with
	  Ident_valid(name,typ,stuff,ann) -> (name,typ,stuff,ann)
	| Ident_null -> failwith "ident_null"
    in let init =
	match stuff with
	  Ident_var(init) -> init
	| Ident_bind -> failwith "How, exactly, do ident_bind's arise?"
	| Ident_proc(_,_,_) -> failwith "Internal proc decl?!?"
    in
	match typ with
	  (Type_scalar _ | Type_ptr _) ->
	    fprintf outchan "%s" (indent i);
	    c_type outchan typ name;
	    fprintf outchan " = 0;%s" (newline i)
	| Type_array(atyp) ->
	    (match atyp with
	    | Array_dense(_,_,_,sc_ty) ->
	    	print_internal_dense_decl outchan i name sc_ty init env
	    | (Array_bb _ | Array_view _)->
	    	failwith "Shouldn't happen. Did you run instantiation?")
	| (Type_proc(_,_)|Type_ann(_)|Type_unknown) ->
	    failwith "Shouldn't happen. Did you run instantiation?"

and print_internal_init outchan i decl env =
    let (name,typ,stuff,ann) =
	match decl with
	  Ident_valid(name,typ,stuff,ann) -> (name,typ,stuff,ann)
	| Ident_null -> failwith "ident_null"
    in let init =
	match stuff with
	  Ident_var(init) -> init
	| Ident_bind -> failwith "How, exactly, do ident_bind's arise?"
	| Ident_proc(_,_,_) -> failwith "Internal proc decl?!?"
    in
	match typ with
	  Type_scalar(_) ->
	    if init <> Expr_void then begin
		fprintf outchan "%s" (indent i);
		fprintf outchan "%s = " name;
		print_expr outchan init env;
		fprintf outchan ";%s" (newline i)
	    end
	| Type_array(atyp) ->
	    (match atyp with
	    | Array_dense(_,_,_,sc_ty) ->
	    	print_internal_dense_init outchan i name sc_ty init env
	    | (Array_bb _ | Array_view _) ->
	    	failwith "Shouldn't happen. Did you run instantiation?")
	| (Type_ptr(_)|Type_proc(_,_)|Type_ann(_)|Type_unknown) ->
	    failwith "Shouldn't happen. Did you run instantiation?"

and compute_dense_bound_expr name env =
    let (_,_,dimens) = cgi_dense_info name env
    in let snemid = List.rev dimens
    in
	List.fold_left
	    (fun result dimen ->
		match (dimen,result) with
		  (Expr_int(x),Expr_int(y)) -> Expr_int(x*y)
		| (dimen,Expr_op(Op_times,dimens')) ->
		    Expr_op(Op_times,dimen::dimens')
		| (dimen,result) ->
		    Expr_op(Op_times,[dimen;result]))
	    (List.hd snemid) (List.tl snemid)

and print_internal_dense_decl outchan i name sc_ty init env =
    let sc_ty_str =
	match sc_ty with
	  Type_int -> "int"
	| Type_real -> "double"
	| (Type_bool|Type_void|Type_string) -> failwith "Unimplemented"
    in
	match (init,compute_dense_bound_expr name env) with
	  (Expr_void,Expr_int(c)) when c <= 4096 ->
	    fprintf outchan "%s%s %s[%d];%s"
		(indent i) sc_ty_str name c (newline i)
	| (Expr_void,bound_expr) ->
	    fprintf outchan "%sstatic %s *%s = 0;%s"
		(indent i) sc_ty_str name (newline i);
	    fprintf outchan "%sstatic int %s = 0;%s"
		(indent i) (name^"__len") (newline i)
	| (_,_) ->
	    fprintf outchan "%s%s *%s;%s" (indent i) sc_ty_str name
		(newline i)

and print_internal_dense_init outchan i name sc_ty init env =
    let sc_ty_str =
	match sc_ty with
	  Type_int -> "int"
	| Type_real -> "double"
	| (Type_bool|Type_void|Type_string) -> failwith "Unimplemented"
    in
	match (init,compute_dense_bound_expr name env) with
	  (Expr_void,Expr_int(c)) when c <= 4096 -> ()
	| (Expr_void,bound_expr) ->
	    let i' = indent_incr i
	    in let i'' = indent_incr i'
	    in let new_len = (name^"__new_len")
	    in let name_len = (name^"__len")
	    in
		fprintf outchan "%s{%s" (indent i) (newline i);
		fprintf outchan "%sint %s = " (indent i') new_len;
		print_expr outchan bound_expr env;
		fprintf outchan ";%s" (newline i');
		fprintf outchan "%sif (%s < %s && %s) {%s"
		    (indent i') name_len new_len name (newline i');
		fprintf outchan "%sfree(%s);%s"
		    (indent i'') name (newline i'');
		fprintf outchan "%s%s = 0;%s"
		    (indent i'') name (newline i'');
		fprintf outchan "%s%s = 0;%s"
		    (indent i'') name_len (newline i'');
		fprintf outchan "%s}%s" (indent i') (newline i');
		fprintf outchan "%sif (%s == 0) {%s"
		    (indent i') name (newline i');
		fprintf outchan "%s%s = 2*%s+1;%s"
		    (indent i'') name_len new_len (newline i'');
		fprintf outchan "%s%s = (%s *)calloc(%s,sizeof(%s));%s"
		    (indent i'') name sc_ty_str name_len sc_ty_str
		    (newline i'');
		fprintf outchan "%sassert(%s);%s"
		    (indent i'') name (newline i'');
		fprintf outchan "%s}%s" (indent i') (newline i');
		fprintf outchan "%s}%s" (indent i) (newline i)
	| (_,_) ->
	    fprintf outchan "%s%s =" (indent i) name;
	    print_expr outchan init env;
	    fprintf outchan ";%s"(newline i)

and print_proc_header outchan name result args =
    let out_arg arg =
	match arg with
	  Ident_valid(name,typ,stuff,ann) ->
	    let name' =
		if is_ref_param (ann_to_list ann) then
		    "*" ^ name
		else
		    name
	    in
		c_type outchan typ name'
	| Ident_null -> failwith "ident_null??"
    in
	fprintf outchan "BMF_INLINE ";
	(match result with
	  Ident_valid(result_name,result_typ,_,_) ->
	    c_type outchan result_typ ""
	| Ident_null -> fprintf outchan "void");
	fprintf outchan " %s(" name;
	(match args with
	  arg::args' ->
	    out_arg arg;
	    List.iter (fun arg -> 
			fprintf outchan ", ";
			out_arg arg)
		 args'
	| [] -> ());
	fprintf outchan ")"

(************************************************************************)
(* Types								*)
(************************************************************************)

and c_type outchan typ name =
    match typ with
      Type_scalar(x) ->
	let name = if name <> "" then " "^name else ""
	in
	    (match x with
	      Type_int -> fprintf outchan "int%s" name
	    | Type_real -> fprintf outchan "%s%s" !real_type name
	    | Type_bool -> fprintf outchan "char%s" name
	    | Type_void -> fprintf outchan "void%s" name
	    | Type_string -> fprintf outchan "char *%s" name)
    | (Type_ptr(Type_array _)|Type_array _) ->
	let atyp =
	  match typ with
	    Type_ptr(Type_array(atyp)) -> atyp
	  | Type_array(atyp) -> atyp
	  | _ -> assert false
	in
	(match atyp with
    	| Array_dense(_,_,_,x) -> 
	    c_type outchan (Type_scalar(x)) ("*"^name)
    	| (Array_bb _ | Array_view _) -> failwith "Can't happen")
    | Type_ptr(btyp) ->
	c_type outchan btyp ("*"^name)
    | Type_proc(_,_) -> failwith "Not implemented."
    | Type_ann(_) -> failwith "Not implemented."
    | Type_unknown -> failwith "Unknown type??"

(************************************************************************)
(* Definitions								*)
(************************************************************************)

and print_top_def outchan decl =
    let (name,typ,stuff,ann) =
	match decl with
	  Ident_valid(name,typ,stuff,ann) -> (name,typ,stuff,ann)
	| Ident_null -> failwith "ident_null"
    in
	match stuff with
	  Ident_var(init) ->
	    c_type outchan typ name;
	    if init != Expr_void then begin
		fprintf outchan " = ";
		print_expr outchan init ([],[])
	    end;
	    fprintf outchan ";"
	| Ident_bind -> failwith "Top-level ident_bind?"
	| Ident_proc(Frame(args),result,body) ->
	    print_proc_def outchan name args result body

and process_dense_decls old_dense_list decls =
    let dense_decls =
	List'.remove_if_not
	     (fun id ->
		match id with
		  Ident_valid(name,typ,stuff,ann) ->
		    (match typ with
		      Type_array(atyp) ->
		      	(match atyp with
			  Array_dense(_,_,_,_) -> true
			| (Array_bb _ | Array_view _) -> assert false)
		    | _ -> false)
		| Ident_null -> assert false (*Huh?*))
	    decls
    in let dense_lst =
	List.map (fun id ->
		match id with
		  Ident_valid(name,typ,stuff,ann) ->
		    (match typ with
		      Type_array(Array_dense(lb,ubs,majorness,base)) ->
			let dimens =
			    List.map
				(fun ub ->
				    if lb = Dense_lb_1 then
					ub
				    else
					Expr_op(Op_plus,[ub;Expr_int(1)]))
				ubs
			in let holders = 
			  List.map
			    (fun i -> name^"__ub"^(string_of_int i))
			    (List'.from_to 0 (List.length dimens))
			in let indexer =
			    dense_indexer lb majorness holders
			in
			    assert
				((List.length ubs) = (List.length dimens))
				(* (Failure("array ann doesn't match type")) *);
			    (name, (indexer, holders, dimens))
		    | _ -> assert false (*Huh?*))
		| Ident_null -> assert false (*Huh?*))
	    dense_decls
in
    dense_lst

and print_proc_def outchan name args result body =
    let i = (indent_incr 0)
    and ref_args =
	List'.remove_if_not (fun id ->
		match id with
		  Ident_valid(name,typ,stuff,ann) ->
		    is_ref_param (ann_to_list ann)
		| Ident_null -> false)
	    args
    in let rename_lst =
	List.map (fun id ->
		    match id with
		      Ident_valid(name,_,_,_) -> (name, name^"__")
		    | Ident_null -> assert false (*Huh?*))
	    ref_args
    in let dense_lst = process_dense_decls [] args
    in let env = (rename_lst,dense_lst)
    in let copy_in_copy_out_vars =
	List.map2
	    (fun rename_info typ ->
		let (name,name') = rename_info
		in
		    ((Ident_valid(name',typ,Ident_var(
			Expr_op(Op_deref,[Expr_var(name)])),
			An_list([]))),
		    ([],[])))
	    rename_lst (List.map bmf_ident_type ref_args)
    in let dimen_holder_vars =
	List'.flat_map
	    (fun (name,(indexer,holders,dimens)) ->
		List.map2
		    (fun holder dimen ->
			((Ident_valid(holder,Type_scalar(Type_int),
			    Ident_var(dimen),An_list([]))),
			(rename_lst,[])))
		    holders dimens)
	    dense_lst
    in
	(* Header *)
	print_proc_header outchan name result args;
	fprintf outchan "\n{\n";
	(* print some declarations and initializations *)
	let do_decl (id,env) = print_internal_decl outchan i id env
	and do_init (id,env) = print_internal_init outchan i id env
	in begin
	    List.iter do_decl copy_in_copy_out_vars;
	    List.iter do_decl dimen_holder_vars;
	    if result != Ident_null then
		print_internal_decl outchan i result env;
	    List.iter do_init copy_in_copy_out_vars;
	    List.iter do_init dimen_holder_vars
	end;
	(* print the body *)
	fprintf outchan "%s" (indent i);
	print_block outchan i body env;
	(* copy-out the ref param vars *)
	List.iter (function (name,name') ->
		    fprintf outchan "%s*%s = %s;\n" (indent i) name name')
	    rename_lst;
	(* return the result var *)
	if result != Ident_null then
	    fprintf outchan "%sreturn %s;\n"
		(indent i) (bmf_ident_name result);
	fprintf outchan "}\n\n"

(************************************************************************)
(* Statements								*)
(************************************************************************)

and print_block outchan i stmt ((rename_lst,dense_lst) as env) =
begin
    let rec find_block stmt =
	match stmt with
	  Stmt_seq(first, second) ->
	    (find_block first) @ (find_block second)
	| Stmt_nop -> []
	| _ -> [stmt]
    in let (decls,stmt) =
	match stmt with
	  Stmt_let(Frame(decls),stmt) -> (decls,stmt)
	| _ -> ([],stmt)
    in let stmts = find_block stmt
    in let new_dense_lst = process_dense_decls dense_lst decls
    in let new_env = (rename_lst,new_dense_lst@dense_lst)
    in let dimen_holder_vars =
	List'.flat_map
	    (fun (name,(indexer,holders,dimens)) ->
		List.map2
		    (fun holder dimen ->
			(Ident_valid(holder,Type_scalar(Type_int),
			    Ident_var(dimen),An_list([]))))
		    holders dimens)
	    new_dense_lst
    in let i' = indent_incr i
    in
	fprintf outchan "{%s" (newline i);
	if dimen_holder_vars <> [] || decls <> [] then begin
	    let do_decl decl = print_internal_decl outchan i' decl new_env
	    and do_init decl = print_internal_init outchan i' decl new_env
	    in
		List.iter do_decl dimen_holder_vars;
		List.iter do_decl decls;
		List.iter do_init dimen_holder_vars;
		List.iter do_init decls;
		fprintf outchan "%s" (newline i)
	end;
	List.iter (fun stmt -> print_stmt outchan i' stmt new_env)
	    stmts;
	fprintf outchan "%s}" (indent i);
	fprintf outchan "%s" (newline i)
end

and print_stmt outchan i stmt env =
begin
    match stmt with
      Stmt_assign(lhs,rhs) ->
	fprintf outchan "%s" (indent i);
	print_expr outchan lhs env;
	fprintf outchan " = ";
	print_expr outchan rhs env;
	fprintf outchan ";%s" (newline i)
    | Stmt_call(Expr_var(f_name),result,args) ->
	assert ((cgi_rename f_name env) == f_name)
	  (* (Failure("cgi_info for a function name?")) *);
	fprintf outchan "%s" (indent i);
	if result <> Expr_void then begin
	  print_expr outchan result env;
	  fprintf outchan " = "
	end;
	fprintf outchan "%s" f_name;
	paren_list outchan args
	  (fun arg -> print_expr outchan arg env);
	fprintf outchan ";%s" (newline i)
    | Stmt_call(_,_,_) -> failwith "Bad function expr"
    | Stmt_for(id,lb,ub,step,body) ->
	let name = (bmf_ident_name id)
	and step_c =
	    match step with
	      Expr_int(step_c) -> step_c
	    | _ -> failwith
		"genc doesn't deal with FOR's with non-const steps"
	in
	    fprintf outchan "%s{ int %s; for (%s=" (indent i) name name;
	    print_expr outchan lb env;
	    (if step_c > 0 then
		fprintf outchan ";%s<=" name
	    else if step_c < 0 then
		fprintf outchan ";%s>=" name
	    else
		failwith "FOR with step of 0?!?");
	    print_expr outchan ub env;
	    fprintf outchan ";%s+=" name;
	    print_expr outchan step env;
	    fprintf outchan ") ";
	    print_block outchan i body env;
	    fprintf outchan " }%s" (newline i)
    | Stmt_while(test,body) ->
	fprintf outchan "%swhile (" (indent i);
	print_expr outchan test env;
	fprintf outchan ") ";
	print_block outchan i body env;
	fprintf outchan "%s" (newline i)
    | Stmt_if(test,thn,els) ->
	
 let rec loop els = begin
		match els with
		  Stmt_nop ->
		    fprintf outchan "%s" (newline i)
		| Stmt_if(test',thn',els') ->
		    fprintf outchan " else if (";
		    print_expr outchan test' env;
		    fprintf outchan ") ";
		    print_block outchan i thn' env;
		    loop els'
		| _ ->
		    fprintf outchan " else ";
		    print_block outchan i els env;
		    fprintf outchan "%s" (newline i)
	    end
 in fprintf outchan "%sif (" (indent i);
	print_expr outchan test env;
	fprintf outchan ") ";
	print_block outchan i thn env;
	loop els
	    
    | (Stmt_seq(_,_)|Stmt_let(_,_)|Stmt_nop) ->
	fprintf outchan "%s" (indent i);
	print_block outchan i stmt env;
	fprintf outchan "%s" (newline i)
    | Stmt_do(test,body) ->
	failwith "Unexpected do stmt in genc"
    | Stmt_directive(ann,s') ->
	failwith "Unexpected stmt_directive in genc"
end

(************************************************************************)
(* Expressions								*)
(************************************************************************)

and print_expr outchan expr env =
    print_expr' outchan lowest_prec expr env 

and print_expr' outchan prec expr env =
let precedence expr =
    match expr with
      Expr_op(op,_) -> op_precedence op
    | _ -> infin_prec
in let prec' = precedence expr
in let rec nary_to_binary outchan f args env =
    match args with
      [] -> failwith "Can't happen"
    | [e] -> print_expr outchan e env
    | e::args' ->
	fprintf outchan "%s(" f;
	print_expr outchan e env;
	fprintf outchan ",";
	nary_to_binary outchan f args' env;
	fprintf outchan ")"
in begin
    if prec >= prec' then fprintf outchan "(";
    (match expr with
      Expr_op(Op_not,[Expr_op(Op_leq,[arg1;arg2])]) ->
	print_args outchan ">" prec' [arg1;arg2] env
    | Expr_op(Op_not,[Expr_op(Op_eq,[arg1;arg2])]) ->
	print_args outchan "<>" prec' [arg1;arg2] env
    | Expr_op(op,args) ->
	(match op with
	  (Op_neg|Op_not|Op_deref|Op_addr) ->
	    assert ((List.length args) == 1)
		(* (Failure("Supposed to be unary.")) *);
	    fprintf outchan "%s" (string_of_op op);
	    print_args outchan "*Notused*" prec' args env
	| (Op_div|Op_eq|Op_leq) -> 
	    assert ((List.length args) ==2)
		(* (Failure("Supposed to be binary.")) *);
	    print_args outchan (string_of_op op) prec' args env
	| (Op_plus|Op_times|Op_and|Op_or) ->
	    print_args outchan (string_of_op op) prec'
		(bmf_expr_args_by_op op expr) env
	| Op_implies ->
	    fprintf outchan "(";
	    nary_to_binary outchan "IMPLIES" args env;
	    fprintf outchan ")"
	| Op_other("min") ->
	    fprintf outchan "(";
	    nary_to_binary outchan "MIN" args env;
	    fprintf outchan ")"
	| Op_other("max") ->
	    fprintf outchan "(";
	    nary_to_binary outchan "MAX" args env;
	    fprintf outchan ")"
	| _ ->
	    fprintf outchan "%s" (string_of_op op);
	    paren_list outchan args
		(fun e -> print_expr outchan e env)
	)
    | Expr_int(n) -> fprintf outchan "%d" n
    | Expr_real(n) -> fprintf outchan "%e" n
    | Expr_string(s) -> fprintf outchan "\"%s\"" (String.escaped s)
    | Expr_bool(b) -> fprintf outchan "%s" (if b then "1" else "0")
    | Expr_var(str) ->
        fprintf outchan "%s" (cgi_rename str env)
    | Expr_void -> failwith "Unexpected expr_void in genc"
    | Expr_forall(_,_) -> failwith "Unexpected expr_forall in genc"
    | Expr_exist(_,_) -> failwith "Unexpected expr_exit in genc"
    | Expr_access(Access_strong,Expr_var(r),mthd,args) ->
	let r' = (cgi_rename r env)
	and (indexer,dimens,_) = (cgi_dense_info r env)
	in
	    assert (mthd = "") (* (Failure("Can't gen C for named access methods.")) *);
	    fprintf outchan "%s" r';
	    fprintf outchan "[";
	    print_expr outchan (indexer args) env;
	    fprintf outchan "]"
    | Expr_access(_,_,_,_) -> failwith "Unexpected expr_access in genc"
    | Expr_set(_,_) -> failwith "Unexpected expr_set in genc"
    );
    if prec >= prec' then fprintf outchan ")"
end

and string_of_op op =
    (match op with
      Op_plus -> "+"
    | Op_times -> "*"
    | Op_div -> "/"
    | Op_and -> "&&"
    | Op_or -> "||"
    | Op_eq -> "=="
    | Op_leq -> "<="
    | Op_not -> "!"
    | Op_neg -> "-"
    | Op_sqrt -> "sqrt"
    | Op_deref -> "*"
    | Op_addr -> "&"
    | Op_implies -> assert false
    | Op_other(s) -> failwith "Unimplemented op_other"
    )

and print_args outchan op prec args env =
    match args with
	[x] -> print_expr' outchan prec x env
    |	x::args' ->
	    print_expr' outchan prec x env;
	    fprintf outchan "%s" op;
	    print_args outchan op prec args' env
    |	[] -> failwith "No args to op?"

(************************************************************************)
(* Dense array indexing							*)
(************************************************************************)

and dense_indexer lb majorness dimens args =
    match majorness with
      Dense_row_major -> dense_row_major_indexer lb dimens args
    | Dense_col_major -> dense_row_major_indexer lb (List.rev dimens) (List.rev args)

and dense_row_major_indexer lb dimens args =
    assert ((List.length dimens) = (List.length args))
	(* (Failure("Num of index args != num of dimens")) *);
    let gen_index arg =
	if lb = Dense_lb_1 then
	    Expr_op(Op_plus,[arg;Expr_int(-1)])
	else
	    arg
    in let rec recurs dimens args =
    	match (dimens,args) with
	  ([],[]) -> ([],[])
	| (dimen::dimens',arg::args') ->
	    let (strides,offsets) = recurs dimens' args'
	    in let dimen' = Expr_var(dimen)
	    and index' = make_times ((gen_index arg)::strides)
	    in
		(dimen'::strides,index'::offsets)
	| (_,_) -> failwith "This can't happen"
    in let (_,offsets) = recurs dimens args
    in
	make_plus offsets

(************************************************************************)

let _ =
  register_switch "-intrins_c"
    ("-intrins_c (bmf) -- Expand intrinsics into C.")
    (fun top_decls args ->
      let name = Bsct_toplevel.get_arg_name (List.hd args)
      and (options,param,args') = Bsct_toplevel.parse_arg false args
      in let _ = Bsct_toplevel.ignore_options options name
      in let top_decls' = List.map expand_intrinsics top_decls
      in (top_decls',args'))

let _ =
  register_switch "-genc"
    ("-genc FILE (bmf) -- Generate C to FILE\n"^
     "	FILE = - writes to stdout.")
    (fun top_decls args ->
      let name = Bsct_toplevel.get_arg_name (List.hd args)
      and (options,param,args') = Bsct_toplevel.parse_arg true args
      in let _ = Bsct_toplevel.ignore_options options name
      in let _ =
    	if param = "-" then
	  to_stdout top_decls
    	else
	  to_file param top_decls
      in (top_decls,args'))

