(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers

open Bmf;;
open Bb_decl;;

(************************************************************************)
(* Collecting free variables that occur in program trees.		*)
(* These routines are obsolete. Use the Bmf.xxx_fvs routines instead.	*)
(************************************************************************)

(* fixme. Obsolete. Use the routines in Bmf_util instead. *)

let reduce_s f lst =
  List.fold_left 
    (fun result x -> String_set.union result (f x))
    String_set.empty lst

let free_vars_in_ident id = (Bmf_util.ident_fvs id)
let free_vars_in_idents ids = (Bmf_util.idents_fvs ids)
let free_vars_in_stmt st = (Bmf_util.stmt_fvs st)
let free_vars_in_stmts sts = reduce_s free_vars_in_stmt sts
let free_vars_in_expr ex = (Bmf_util.expr_fvs ex)
let free_vars_in_exprs exs = reduce_s free_vars_in_expr exs
let free_vars_in_type ty = (Bmf_util.type_fvs ty)
let free_vars_in_types tys = reduce_s free_vars_in_type tys
let free_vars_in_ann an = (Bmf_util.ann_fvs an)
let free_vars_in_anns ans = reduce_s free_vars_in_ann ans
  
(************************************************************************)

type aliases_map = String_set.t String_map.t;;
let empty_s = String_set.empty
let singleton_s x = String_set.add x empty_s

let expr_fvs e =
  let lst = String_set.elements (free_vars_in_expr e) in
  String_set.of_list lst

let map_incr name val0 m =
  String_map.update m name empty_s (fun s -> String_set.add val0 s)

(************************************************************************)
(* Compute aliasing between variables, based on their declarations.	*)
(*									*)
(* _Very_ conservative.							*)
(*									*)
(* In BMF, we assume no aliasing between parameters and globals ( la   *)
(* Fortran), yet aliasing can still arise in several circumstances:     *)
(*									*)
(* 1. Arguments to bb rel's are used as the underlying storage for those*)
(* relations. Thus, the arguments and the bb rel's are aliased.		*)
(*									*)
(* 2. The op(addr)(...) form can be used to initialize a dense array var*)
(* with a pointer into another dense array.  Thus, the two arrays are   *)
(* aliased.								*)
(*									*)
(* 3. A relation name is aliased with each of its field names, but the	*)
(* field names are independent from one another.			*)
(*									*)
(* This routine computes and returns a map from each name, A, to the    *)
(* set of names that are possibly invalidated by writes to A.		*)
(*									*)
(* For ease of implementation, this routine assumes that no two		*)
(* declarations are for the same name. Perform renaming before using	*)
(* this routine.							*)
(************************************************************************)

(* *****************************************************************
   Entry point.
   ***************************************************************** *)

let rec compute_aliases_by_decls id = (
    (* First, collect all of the decls in form. *)
    let decls = walk_id id
    (* Second, check for the same name occuring in more than one decl. *)
    in let names_s =
      List.fold_left
	(fun names_s decl ->
	  let name = bmf_ident_name decl in
	  if String_set.mem name names_s then
	    failwith 
	      "hybrid_aliases: Same name in multiple decls. Rename first!";
	  String_set.add name names_s)
	empty_s decls
    (* Third, parse the decls to construct the Aliases, ArgsOf, and ArgOf
	relations. *)
    in let (aliases_m, argsOf_m, argOf_m) =
        let initial_m =
	    List.fold_right 
	        (fun name m -> String_map.add name empty_s m)
	        (String_set.to_list names_s) String_map.empty
	in
	    List.fold_left process_decl
	        (initial_m,initial_m,initial_m)
		decls
    (* Finally, for each name, compute its invalidate set. *)
    in
	List.fold_right
	    (fun name m ->
		String_map.add
		   name
		   (invalidate_names aliases_m argsOf_m argOf_m
			name empty_s)
		    m)
	    (String_set.to_list names_s) String_map.empty
: aliases_map)

(* ***************************************************************** 
   Routines to walk the AST's and find all var decls.
   ***************************************************************** *)

and walk_ids ids = List'.flat_map walk_id ids

and walk_id id =
  match id with
    Ident_valid(_,_,stuff,_) ->
      (match stuff with
	Ident_proc(Frame(args),res,body) ->
	  id::(walk_ids (res::args))@(walk_st body)
      |	_ -> [id])
  | Ident_null -> []

and walk_st st =
  match st with
    (Stmt_assign _ | Stmt_call _ | Stmt_nop) -> []
  | Stmt_seq(s1,s2) -> (walk_st s1)@(walk_st s2)
  | Stmt_for(id,_,_,_,body) -> (walk_id id)@(walk_st body)
  | Stmt_while(_,body) -> walk_st body
  | Stmt_if(_,thn,els) -> (walk_st thn)@(walk_st els)
  | Stmt_let(Frame(ids),body) -> (walk_ids ids)@(walk_st body)
  | Stmt_do(Expr_set(Frame(ids),_),body) -> 
      (walk_ids ids)@(walk_st body)
  | Stmt_do _ -> assert false
  | Stmt_directive(_,body) -> walk_st body


(* *****************************************************************
   Dispatch routine for processing declarations.
   ***************************************************************** *)

and process_decl (aliases_m, argsOf_m, argOf_m) decl =
    match decl with
      Ident_valid(_,Type_array(atyp),_,_) ->
	(match atyp with
	  Array_bb(_,_,_,_) ->
	    let (argsOf_m', argOf_m') =
	      process_bb_decl argsOf_m argOf_m decl
	    in (aliases_m, argsOf_m', argOf_m')
    	| Array_dense(_,_,_,_) ->
	    let aliases_m' =
	      process_dense_decl aliases_m decl
	    in (aliases_m', argsOf_m, argOf_m)
	| Array_view _ ->
	    (* Views should be removed by the scheduler. *)
	    assert false 
	      )
    | _ -> (aliases_m, argsOf_m, argOf_m)

(* *****************************************************************
   Given a bb decl,

       var A : array int^?? (bb_name arg1: (x) arg2: (y) ...) of ??;

   Add all (A,x), (A,y), ... pairs to ArgsOf_m and all (x,A), (y,A), ...
   pairs to ArgOf_m. The idea is that a write to A may invalidate x,y,...,
   and a write to x may invalidate A, but none of y,...

   Note, if an argument is not a simple var or array access expr, then we
   assume that it can't be used as an lvalue by the bb. Is this a
   reasonable assumption?
   ***************************************************************** *)

and process_bb_decl argsOf_m argOf_m decl =
    let name = bmf_ident_name decl
    and (_,keywords) = extract_bb_directive decl
    in let args_s =
	List.fold_left
	    (fun args_s (_,ann) ->
		match ann with
		  An_expr(Expr_var(arg)) ->
		    String_set.add arg args_s
		| An_expr(Expr_access(_,Expr_var(arg),_,_)) ->
		    String_set.add arg args_s
		| _ -> args_s)
	    empty_s keywords
    in
	List.fold_right
	    (fun r (argsOf_m',argOf_m') ->
		(map_incr name r argsOf_m',
		 map_incr r name argOf_m'))
	    (String_set.to_list args_s) (argsOf_m,argOf_m)

(* *****************************************************************
   Dense array decls can be initialized with a "pointer".

       var inode ... := op(addr)(A[[...]]);

   In this case, inode and A are aliases of one another.
   ***************************************************************** *)

and process_dense_decl aliases_m decl =
    let (name,aliases_s) =
	match decl with
	  Ident_valid(name,_,
		Ident_var(Expr_op(Op_addr,[
		    Expr_access(_,Expr_var(r),_,_)])),_) ->
	    (name, singleton_s r)
	| Ident_valid(name,_,Ident_var(Expr_void),_) ->
	    (name, empty_s)
	| _ -> assert false (* Can't happen *)
    in
	List.fold_right
	    (fun r aliases_m' ->
		map_incr name r (map_incr r name aliases_m'))
	    (String_set.to_list aliases_s) aliases_m

(* *****************************************************************

   Compute the set of each invalidate names for each name.

   All bb relations in which the name appears as an argument are
   invalidated. Then, all bb relations in which these names appear are also
   invalidated. Etc. for all "ancestor" names.

   Also, all names that appear as bb arguments in name's decl are
   invalidated. Etc. for all "descendents".
   
   Finally, recursively invalidate any name to that is aliases to any names
   that we have invalidated so far.

   Here are a couple of example.

	var A : array int^?? (bb_name arg1: (x) arg2: (y) ...) of ??;
	var M : array int^?? (bb_name' arg: (A) arg: (B) ...) of ??;
	var A' : array int^?? (bb_name arg1: (x') arg2: (y') ...) of ??
	    := op(addr)(A[[...]])

   Writing to A invalidates x, y, M, A', x' and y', but not B.
   Writing to x invalidates A, M, A', x' and y' but not y.

   Clear as mud?

   *todo*: The invalidation that occurs because of aliasing between bb
   decl's is odd, and may be inconceived, but it is conservative. Since, we
   don't actually allowing aliasing withween bb decl's at the moment,
   perhaps this is something to revisit at a future time...

***************************************************************** *)

and invalidate_names aliases_m argsOf_m argOf_m name invalid_s =
    if String_set.mem name invalid_s then
	invalid_s
    else begin
	let rec ancestors_of name =
	    String_set.fold
		(fun name' result ->
		    String_set.union (ancestors_of name') result)
		(String_map.find name argOf_m) (singleton_s name)
	and descendents_of name =
	    String_set.fold
		(fun name' result ->
		    String_set.union (descendents_of name')  result)
		(String_map.find name argsOf_m) (singleton_s name)
	in let relations =
	    String_set.union (ancestors_of name) (descendents_of name)
	in let aliases =
	    String_set.fold
		(fun name' results ->
		  String_set.union (String_map.find name' aliases_m) 
		    results)
		relations empty_s
	in
	    String_set.fold
		(invalidate_names aliases_m argsOf_m argOf_m)
		aliases (String_set.union invalid_s relations)
    end
;;

