(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers
open Bmf

module type PARAMS =
  sig
    type edge_val_t
    type context_t
    val bottom: context_t -> edge_val_t
    val equal: context_t -> edge_val_t -> edge_val_t -> bool
    val at_assign: context_t -> bmf_stmt -> edge_val_t -> edge_val_t
    val at_call: context_t -> bmf_stmt -> edge_val_t -> edge_val_t
    val at_let_enter: context_t -> bmf_stmt -> edge_val_t -> edge_val_t
    val at_let_exit: context_t -> bmf_stmt -> edge_val_t -> edge_val_t
    val at_directive_enter: context_t -> bmf_stmt -> edge_val_t -> edge_val_t
    val at_directive_exit: context_t -> bmf_stmt -> edge_val_t -> edge_val_t
    val at_if_enter: context_t -> bmf_stmt -> edge_val_t -> 
      (edge_val_t * edge_val_t)
    val at_if_exit: context_t -> bmf_stmt -> edge_val_t -> edge_val_t -> 
      edge_val_t
    val at_for: context_t -> bmf_stmt -> edge_val_t -> edge_val_t -> 
      (edge_val_t * edge_val_t)
    val at_while: context_t -> bmf_stmt -> edge_val_t -> edge_val_t -> 
      (edge_val_t * edge_val_t)
    val at_do: context_t -> bmf_stmt -> edge_val_t -> edge_val_t -> 
      (edge_val_t * edge_val_t)
  end


(************************************************************************)

module Stmt_element =
  struct
    type t = bmf_stmt
    let equal = (==)
    let hash = Containers.hash
  end
module type RESULT = (HASHTBL with type key = bmf_stmt)
module Result = Hashtbl(Stmt_element)
let make_result () = Result.create 101

(************************************************************************)

module Solver(P: PARAMS) =
  struct
    type edge_val_t = P.edge_val_t
    type context_t = P.context_t

    (*****************************************************************)

    let rec init_hts (ctx: context_t) in_h out_h s =
      let r = init_hts ctx in_h out_h in
      Result.replace in_h s (P.bottom ctx);
      Result.replace out_h s (P.bottom ctx);
      match s with
	Stmt_seq(first,second) -> r first; r second
      | Stmt_for(_,_,_,_,body) -> r body
      | Stmt_while(_,body) -> r body
      | Stmt_if(_,thn,els) -> r thn; r els
      | Stmt_let(_,body) -> r body
      | Stmt_do(_,body) -> r body
      | Stmt_directive(_,body) -> r body
      | ( Stmt_assign _ | Stmt_call _ | Stmt_nop ) -> ()

    (*****************************************************************)

    let forward_solve (ctx: context_t) (s: bmf_stmt) (in_val : 'a) =
      let in_h = make_result ()
      and out_h = make_result ()
      in
      (* This is where execution starts *)
      let rec main () = (
	(* Initialize the hashtables. *)
	init_hts ctx in_h out_h s;
	(* Propogate the values. *)
	walk s in_val;
	(* Return the hashtables as the results *)
	(in_h,out_h)
      : ('a Result.t * 'a Result.t))

      (* walk and walk': traverse the AST and do the analysis. *)

      (* Updates the in and out edges of s and calls walk' to do the real
         work. *)
      and walk s in_val =
	Result.replace in_h s in_val;
	let result = walk' s in_val in
	Result.replace out_h s result;
	result

      (* Dispatch on s and compute the new output value for s. *)
      and walk' s in_val =
	match s with
	  Stmt_assign _ -> P.at_assign ctx s in_val
	| Stmt_call _ -> P.at_call ctx s in_val
	| Stmt_nop -> in_val
	| Stmt_seq(first,second) -> walk second (walk first in_val)
	| Stmt_if(_,thn,els) ->
	    let (thn_in_val,els_in_val) = P.at_if_enter ctx s in_val
	    in let thn_out_val = walk thn thn_in_val
	    and els_out_val = walk els els_in_val
	    in let out = P.at_if_exit ctx s thn_out_val els_out_val
	    in out
	(* Looping constructs. *)
	| ( Stmt_for _  | Stmt_while _ | Stmt_do _ ) ->
	    let (body, at_loop) =
	      match s with
		Stmt_for(_,_,_,_,body) -> (body, P.at_for)
	      |	Stmt_while(_,body) -> (body, P.at_while)
	      |	Stmt_do(_,body) -> (body, P.at_do)
	      |	_ -> assert false
	    in 
	    recurs (Result.find out_h body)
	      (fun recurs back_val ->
	      	let (out_val, body_in_val) = at_loop ctx s in_val back_val
	      	in let body_out_val = walk body body_in_val
	      	in
	      	if P.equal ctx back_val body_out_val then
		  out_val
	      	else
		  recurs body_out_val)
	(* "Scoping" constructs. No looping. *)
	| ( Stmt_let _ | Stmt_directive _ ) ->
	    let (body, at_enter, at_exit) =
	      match s with
		Stmt_let(_,body) ->
		  (body, P.at_let_enter, P.at_let_exit)
	      |	Stmt_directive(_,body) ->
		  (body, P.at_directive_enter, P.at_directive_exit)
	      |	_ -> assert false
	    in let body_in_val = at_enter ctx s in_val
	    in let body_out_val = walk body body_in_val
	    in let out = at_exit ctx s body_out_val
	    in out

      in main()

    (*****************************************************************)

    let backward_solve (ctx: context_t) (s: bmf_stmt) (out_val : 'a) =
      let in_h = make_result ()
      and out_h = make_result ()
      in
      (* This is where execution starts *)
      let rec main () = (
	(* Initialize the hashtables. *)
	init_hts ctx in_h out_h s;
	(* Propogate the values. *)
	walk s out_val;
	(* Return the hashtables as the results *)
	(in_h,out_h)
      : ('a Result.t * 'a Result.t))

      (* walk and walk': traverse the AST and do the analysis. *)

      (* Updates the in and out edges of s and calls walk' to do the real
         work. *)
      and walk s out_val =
	Result.replace out_h s out_val;
	let result = walk' s out_val in
	Result.replace in_h s result;
	result

      (* Dispatch on s and compute the new input value for s. *)
      and walk' s out_val =
	match s with
	  Stmt_assign _ -> P.at_assign ctx s out_val
	| Stmt_call _ -> P.at_call ctx s out_val
	| Stmt_nop -> out_val
	| Stmt_seq(first,second) -> walk first (walk second out_val)
	| Stmt_if(_,thn,els) ->
	    let (thn_out_val,els_out_val) = P.at_if_enter ctx s out_val
	    in let thn_in_val = walk thn thn_out_val
	    and els_in_val = walk els els_out_val
	    in let in_val = P.at_if_exit ctx s thn_in_val els_in_val
	    in in_val
	(* Looping constructs. *)
	| ( Stmt_for _  | Stmt_while _ | Stmt_do _ ) ->
	    let (body, at_loop) =
	      match s with
		Stmt_for(_,_,_,_,body) -> (body, P.at_for)
	      |	Stmt_while(_,body) -> (body, P.at_while)
	      |	Stmt_do(_,body) -> (body, P.at_do)
	      |	_ -> assert false
	    in
	    recurs (Result.find in_h body)
	      (fun recurs back_val ->
	      	let (in_val, body_out_val) = at_loop ctx s out_val back_val
	      	in let body_in_val = walk body body_out_val
	      	in
	      	if P.equal ctx back_val body_in_val then
		  in_val
	      	else
		  recurs body_in_val)
	(* "Scoping" constructs. No looping. *)
	| ( Stmt_let _ | Stmt_directive _ ) ->
	    let (body, at_enter, at_exit) =
	      match s with
		Stmt_let(_,body) ->
		  (body, P.at_let_enter, P.at_let_exit)
	      |	Stmt_directive(_,body) ->
		  (body, P.at_directive_enter, P.at_directive_exit)
	      |	_ -> assert false
	    in let body_out_val = at_enter ctx s out_val
	    in let body_in_val = walk body body_out_val
	    in let in_val = at_exit ctx s body_in_val
	    in in_val

      in main()
  end

(************************************************************************)

module Forward_gen_kill_params =
  struct
    type edge_val_t = Int_set.t
    type context_t = 
    	(*gens_ht:*) Int_set.t Result.t * (*kills_ht:*) Int_set.t Result.t

    let bottom ctx = Int_set.empty
    let equal ctx x y = Int_set.equal x y
    (* Simple statements *)
    let classic_df (gens_ht,kills_ht) s in_s =
      Int_set.union (Result.find gens_ht s)
      	(Int_set.diff in_s (Result.find kills_ht s))
    let at_assign = classic_df
    let at_call = classic_df
    let at_directive_enter (gens_ht,kills_ht) s in_s = in_s
    let at_directive_exit (gens_ht,kills_ht) s in_s = in_s
    (* Ifs. No gens or kills *)
    let at_if_enter (gens_ht,kills_ht) s in_s = 
      ((*thn_in*)in_s,(*els_in*)in_s)
    let at_if_exit (gens_ht,kills_ht) s thn_out els_out = 
      Int_set.union thn_out els_out
    (* Let's gens_ht is the Def_entry's for the scoped ids.
       kills_ht removes all defs when exitted the scope. *)
    let at_let_enter (gens_ht,kills_ht) s in_s =
      Int_set.union in_s
	(Result.find gens_ht s)
    let at_let_exit (gens_ht,kills_ht) s body_out =
      Int_set.diff body_out 
	(Result.find kills_ht s)
    (* loops. *)
    let loop_df (gens_ht,kills_ht) s in_s body_out =
      let body_in =
      	Int_set.union
	  (Int_set.union in_s
	     (Result.find gens_ht s))
	  body_out
      in let out_s =
      	Int_set.diff body_in
	  (Result.find kills_ht s)
      in
      (out_s, body_in)
    let at_for = loop_df
    let at_do = loop_df
    let at_while = loop_df
  end

