(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Bmf
open Containers

module type PARAMS =
  sig
    (* type of the values propagated on the edges. *)
    type edge_val_t
    (* type of the problem context that is passed to other parameter
       functions. *)
    type context_t

    (* The value used to initialize all edges. *)
    val bottom: context_t -> edge_val_t
    (* equality predicate for edge values. *)
    val equal: context_t -> edge_val_t -> edge_val_t -> bool

    (* The following functions are used to propogate values through bmf
       nodes.

       The sense in which values are propagates is determined by the class
       of problem being solve. If, for instance, a forward problem is being
       solved, then the #at_assign method will be expected to take the in
       edge value and compute the out edge value. If a backward problem is
       being solved, then this message will be expected to take the out
       edge value and compute the in edge value.
       *)

    (* Simple statements.
       Forward: bmf_stmt -> in edge value -> out edge value
       Backward: bmf_stmt -> out edge value -> in edge value *)
    val at_assign: context_t -> bmf_stmt -> edge_val_t -> edge_val_t
    val at_call: context_t -> bmf_stmt -> edge_val_t -> edge_val_t
    (* Scoping statements.
       Forward:
       #at_xxx_enter: bmf_stmt -> in edge value -> body in edge value
       #at_xxx_exit: bmf_stmt -> body out edge value -> out edge value
       Backward:
       #at_xxx_enter: bmf_stmt -> out edge value -> body out edge value
       #at_xxx_exit: bmf_stmt -> body in edge value -> in edge value *)
    val at_let_enter: context_t -> bmf_stmt -> edge_val_t -> edge_val_t
    val at_let_exit: context_t -> bmf_stmt -> edge_val_t -> edge_val_t
    val at_directive_enter: context_t -> bmf_stmt -> edge_val_t -> edge_val_t
    val at_directive_exit: context_t -> bmf_stmt -> edge_val_t -> edge_val_t
    (* If stmt.
       Similar to the scoping statements, except that two bodies are
       involved, the "then" body and the "else" body. *)
    val at_if_enter: context_t -> bmf_stmt -> edge_val_t -> 
      (edge_val_t * edge_val_t)
    val at_if_exit: context_t -> bmf_stmt -> edge_val_t -> edge_val_t -> 
      edge_val_t
    (* Looping constructs.
       Forward: bmf_stmt -> in edge value -> body out edge value ->
		    (out edge value, body in edge value)
       Backward: bmf_stmt -> out edge value -> body in edge value ->
		     (out in value, body out edge value) *)
    val at_for: context_t -> bmf_stmt -> edge_val_t -> edge_val_t -> 
      (edge_val_t * edge_val_t)
    val at_while: context_t -> bmf_stmt -> edge_val_t -> edge_val_t -> 
      (edge_val_t * edge_val_t)
    val at_do: context_t -> bmf_stmt -> edge_val_t -> edge_val_t -> 
      (edge_val_t * edge_val_t)
  end


module type RESULT = (HASHTBL with type key = bmf_stmt)
module Result : RESULT

module Solver(P: PARAMS) : 
  sig
    type edge_val_t = P.edge_val_t
    type context_t = P.context_t

    (* Takes a context, a stmt, and the initial input.
       Returns two hashtables, (in_ht,out_ht), indexed by stmts.
	 Result.find in_ht s -> value on in edge of stmt s
	 Result.find out_ht s -> value on out edge of stmt s
       *)
    val forward_solve: 
	context_t -> bmf_stmt -> edge_val_t ->
	    edge_val_t Result.t * edge_val_t Result.t
    val backward_solve: 
	context_t -> bmf_stmt -> edge_val_t ->
	    edge_val_t Result.t * edge_val_t Result.t
  end


module Forward_gen_kill_params :
    (PARAMS with type edge_val_t = Int_set.t
	     and type context_t =
		 (*gens_ht:*) Int_set.t Result.t * 
                (*kills_ht:*) Int_set.t Result.t)
