(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers

open Bsct_base
open Bmf
open Bmf_join
open Bb

module WS = Wff.Simple

(* *****************************************************************

   Iteration Based Reindexing (IBR).

        ...
        while do
	    ...
            for ...
                do (Aii,Ai,Bii,Bi) := join ...

        |
        v

        ...
        count = 0
        for ...
            do (Aii,Ai,Bii,Bi) := join ...
                count++
        int tAii[count], tBii[count];
        count = 0
        for ...
            do (Aii,Ai,Bii,Bi) := join ...
                tBii[count] = Bii
                count++
        while
            count = 0;
            for ...
                do (Aii,Ai) := enum A ...
                Bii = tBii[count]; Bi = ...
                count++


   Basic algorithm
   * Compute aliases
   * Compute reaching defs
   * walk proc body using placement/substr machinery
     + Find a join
     + Compute the query "slice" = top-of-query -> ... -> join
     + Find free vars of slice
     + Find invariant pt of free vars. highest pt at which in defs don't
       change.
     + Establish profitability of hoisting slice to invariant pt
     + Code gen
     + Recurse on new join body w/subs
     + Return new body and placements

   Unfortunately, we don't have any easy update the AST or the reaching def 
   information in place. This makes it very difficult to handle nested
   joins with a single pass over the AST.

   Instead, we choose to reindex only the outermost joins that are
   reachable by traversing the AST top-down. Once these joins have been
   reindexed and the new AST has been generated, we repeat the entire
   process. We do this until no more reindexable joins are found.

   todo: 
   o try to partly schedule join before reindexing.
   ***************************************************************** *)

exception Cant_reindex

type ibr_stuff = {
    ibr_defs_of_name : bmf_stmt -> string -> Int_set.t;
    ibr_parent_of : bmf_stmt -> bmf_stmt option
};;

let expr_fvs = Bmf_util.expr_fvs
and stmt_fvs = Bmf_util.stmt_fvs
and ident_fvs = Bmf_util.ident_fvs
and idents_fvs = Bmf_util.idents_fvs
and ann_fvs = Bmf_util.ann_fvs

let empty_s = String_set.empty
let singleton_s x = String_set.add x empty_s
let reduce_s f lst =
  List.fold_left (fun s x -> String_set.union s (f x)) empty_s lst

(* *****************************************************************
   Entry point. 
   ***************************************************************** *)

let rec do_proc id =
  let id = Bmf_util.uniquely_rename_within_proc id in
  match id with
   Ident_valid(name,ty,Ident_proc(Frame(args),res,body),an) ->
     (* Reaching defs, group by aliases. *)
     let in_defs_at_stmt_of_name =
       let (ord,in_h,_) = Hybrid_rds.compute_reaching_defs body
       and aliases_m = 
	 Hybrid_aliases.compute_aliases_by_decls id
       in let in_defs_at_stmt_of_name =
	 Hybrid_rds.defs_at_stmt_by_name ord in_h aliases_m
       in in_defs_at_stmt_of_name
     in 
     (* parent_of function. *)
     let parent_of =
       let raw_parent_of = Bmf_util.parent_of_stmt body
       in
       (fun st ->
	 try Some(raw_parent_of st)
	 with Bmf_util.Root_stmt -> None)
     in
     (* Global state. *)
     let stuff = {
       ibr_defs_of_name = in_defs_at_stmt_of_name;
       ibr_parent_of = parent_of
     } in
     (* Walk and do the reindexing *)
     let (reindexed_p,placements,body') =
       outer_walk stuff body
     in let body'' =
       Bmf_util.perform_substitutions_in_stmt body'
     in let id' =
       Ident_valid(name,ty,Ident_proc(Frame(args),res,body''),an)
     in
     (* Recurs and look for more reindexing opportunities if we found some
     	this go around. Otherwise, exit. *)
     if reindexed_p then do_proc id' else id'
  (* We only do reindexing in procedures. Duh. *)
  | _ -> id

(* *****************************************************************
   The recursive tree walk. Returns (reindexed_p,placements,st').

   st' = the new statement.
   reindexed_p = was any reindexing done in st'?
   placements = placement information for code being hoisted out of st.

   placement information is a list of (st1,xform) pairs. st1 is a statement
   in the original AST. xform is a function that takes, st1', the new
   version of and returns st1'', a further transformed version of st1.
   ***************************************************************** *)
	
and outer_walk stuff st =
  (* Destruct and recurs. *)
  let r = outer_walk stuff in  
  let (reindexed_p,placements,st') =
    match st with
      _ when stmt_is_join st -> process_join stuff st
    | _ when stmt_is_determined st -> process_determined stuff st
    | (Stmt_assign _ | Stmt_call _ | Stmt_nop) -> (false,[],st)
    | Stmt_seq(st1,st2) ->
      	let (reindexed1_p,placement1,st1') = r st1
      	and (reindexed2_p,placement2,st2') = r st2 in
      	(reindexed1_p || reindexed2_p,
	 placement1 @ placement2,
	 Stmt_seq(st1',st2'))
    | Stmt_for(id,lb,ub,step,body) ->
	let (reindexed_p,placement,body') = r body in
	(reindexed_p,placement,Stmt_for(id,lb,ub,step,body'))
    | Stmt_while(pred,body) ->
	let (reindexed_p,placement,body') = r body in
	(reindexed_p,placement,Stmt_while(pred,body'))
    | Stmt_if(test,thn,els) ->
	let (reindexed_thn_p,placement_thn,thn') = r thn
	and (reindexed_els_p,placement_els,els') = r els in
      	(reindexed_thn_p || reindexed_els_p,
	 placement_thn @ placement_els,
	 Stmt_if(test,thn',els'))
    | Stmt_let(frm,body) ->
	let (reindexed_p,placement,body') = r body in
	(reindexed_p,placement,Stmt_let(frm,body'))
    | Stmt_do(pred,body) ->
	let (reindexed_p,placement,body') = r body in
	(reindexed_p,placement,Stmt_do(pred,body'))
    | Stmt_directive(ans,body) ->
	let (reindexed_p,placement,body') = r body in
	(reindexed_p,placement,Stmt_directive(ans,body'))
  in
  (* Do any placements that apply to the current statement. *)
  let (placements',st'') =
    List.fold_left
      (fun (placements',st'') ((targ_st,xform_f) as placement) ->
	if targ_st = st then
	  (placements',xform_f st'')
	else
	  (placement::placements',st''))
      ([],st') placements
  in
  (* Done *)
  (reindexed_p, placements', st'')

(* *****************************************************************
   Joins
   ***************************************************************** *)

and process_join 
    ({ ibr_defs_of_name = defs_of_name; ibr_parent_of = parent_of }
       as stuff) join_st =
  let ({ join_t_var = t_var; join_bounds = bounds;
	 join_terms = terms; join_pred = pred;
	 join_assertion = assertion; join_body = body } as join) =
    stmt2join join_st
  in
  try
    let (slice,slice_fvs) = construct_slice stuff join_st in
    let invar_st = find_invariant_pt stuff slice slice_fvs in
    (* If we get this far, it's profitable, so preceed with the reindexing. *)
    let (new_join_terms,target_terms) =
      determine_terms_for_reindexing stuff invar_st join
    in 
    if target_terms = [] then raise Cant_reindex;
    let (placements,new_join) =
      reindex_join join slice invar_st
	new_join_terms target_terms
    in (true,placements,join2stmt new_join)
  with Cant_reindex ->
    let (reindexed_p,placements,body') = outer_walk stuff body in
    (reindexed_p, placements,
     join2stmt 
       { join_t_var = t_var; join_bounds = bounds;
	 join_terms = terms; join_pred = pred;
	 join_assertion = assertion; join_body = body' })
      
(* fixme: make this general *)
and determine_terms_for_reindexing 
    ({ ibr_parent_of = parent_of } as stuff) invar_st join  =
  let bbm =
    (* go from the root down to the invar_st, accumulating array decls into 
       bbm. *)
    recurs invar_st
      (fun recurs st ->
	let bbm' =
	  match (parent_of st) with
	    Some(st') -> recurs st'
	  | None -> Bb_decl.empty_black_box_map
	in
	match st with
	  Stmt_let(Frame(ids),_) -> Bb_decl.add_array_decls ids bbm'
	| _ -> bbm')
  in
  let (t1,t2) =
    (match join.join_terms with
      [t1;t2] -> (t1,t2)
    | _ -> raise Cant_reindex)
  in
  let r1 = rel_of_kind t1.jt_kind
  and r2 = rel_of_kind t2.jt_kind in
  let bb1 = String_map.find r1 bbm
  and bb2 = String_map.find r2 bbm in
  let n1 = List.length bb1.bb_index_fields
  and n2 = List.length bb2.bb_index_fields in
  if n1 > n2 then
    ([t1],[t2])
  else if n1 < n2 then
    ([t2],[t1])
  else
    raise Cant_reindex

(* fixme *)
(* returns ([inspector_placement,init_placement] as placements,new_join)
   where

   inspector_placement -- adds inspector:
     (invar_st,inspector_f)
   init_placement -- adds "count := 0" before slice:
     (top of slice, init_f)
   *)

and reindex_join join slice invar_st new_join_terms 
    target_terms = (
  let { join_t_var = t_var; join_bounds = bounds;
      	join_terms = terms; join_pred = pred;
      	join_assertion = assertion; join_body = body } = join
  in
  (* Variable names used in the generated code. *)
  let num_count_v = Misc.gensym "num_count"
  and count_v = Misc.gensym "count"
  in
  (* Loop nest to count the number of iterations. *)
  let count_body =
    gen_with_slice slice
      (Bmf_parser_aux.parse_stmt
	 "$num_count := $num_count + 1;"
	 [("$num_count",An_string(num_count_v))])
  in
  (* (vars,code) pairs.
     vars is list of temps used to store results of target terms. 
     code is assignment stmt used to initialize the vars. *)
  let reindexing_data =
    List.map
      (fun target_term ->
	let join_var = join_var_of_kind (target_term.jt_kind)
	and enum_var = enum_var_of_kind (target_term.jt_kind) in
  	let join_tmp = Misc.gensym ("i"^join_var)
	and enum_tmp = Misc.gensym ("i"^enum_var) in
	let join_repl = Expr_access(Access_strong,Expr_var(join_tmp),
				    "",[Expr_var(count_v)])
	and enum_repl = Expr_access(Access_strong,Expr_var(enum_tmp),
				    "",[Expr_var(count_v)])
	in
	let parse_args =
	  [("$join_tmp",An_string(join_tmp));
	    ("$join_var",An_string(join_var));
	    ("$enum_tmp",An_string(enum_tmp));
	    ("$enum_var",An_string(enum_var));
	    ("$count",An_string(count_v))]
	in
	if join_var = enum_var then
	  ([join_tmp],
	   [(join_var,join_repl)],
	   Bmf_parser_aux.parse_stmt 
	     "$join_tmp[[$count]] := $join_var;" 
	     parse_args)
	else
	  ([join_tmp;enum_tmp],
	   [(join_var,join_repl);(enum_var,enum_repl)],
	   Bmf_parser_aux.parse_stmt
	     ("{ $join_tmp[[$count]] := $join_var;"^
	      "  $enum_tmp[[$count]] := $enum_var; }")
	     parse_args))
      target_terms
  in 
  (* tmp variable declarations. *)
  let tmp_var_decls =
    List'.flat_map
      (fun (vars,_,_) ->
	List.map
	  (fun var ->
	    Bmf_parser_aux.parse_ident
	      "var $var: array (0:$num_count-1) of int;"
	      [("$var",An_string(var));
		("$num_count",An_string(num_count_v))])
	  vars)
      reindexing_data
  in
  (* Loop nest to initialize the temps. *)
  let init_body =
    gen_with_slice slice
      (Bmf_parser_aux.parse_stmt
	 "{ $inits; $count := $count + 1; }"
	 [("$count",An_string(count_v));
	   ("$inits",
	    An_list(List.map (fun (_,_,init) -> An_stmt(init)) 
		      reindexing_data))])
  in
  (* Inspector code. *)
  let inspector_target_st = invar_st
  and inspector_f invar_st' =
    Bmf_parser_aux.parse_stmt
      ("{"^
       "  var $num_count: int := 0; "^
       "  var $count: int := 0; "^
       "  $count_body;"^
       "  $xxx;"^
       "}")
      [("$num_count",An_string(num_count_v));
	("$count",An_string(count_v));
	("$count_body",An_stmt(count_body));
	("$xxx",An_stmt(Stmt_let(Frame(tmp_var_decls),
				 Stmt_seq(init_body,invar_st'))))]
  (* Code placed just outside of executor join. *)
  and init_target_st = List.hd slice
  and init_f top_st =
    Bmf_parser_aux.parse_stmt
      ("{ $count := 0; $top_st; }")
      [("$count",An_string(count_v));
	("$top_st",An_stmt(top_st))]
  (* The reindexed join. *)
  and new_join =
    let new_body =
      Bmf_util.place_substitutions
	(List'.flat_map (fun (_,substs,_) -> substs) reindexing_data)
	body
    and wff_bindings =
      List.map (fun term -> (term.jt_flag,WS.Constant(true)))
	target_terms
    in
    (* fixme. this is totally bogus. *)
    let new_pred = 
      WS.simplify (WS.subst wff_bindings join.join_pred)
    and new_assertion = 
      WS.simplify (WS.subst wff_bindings join.join_assertion)
    in
    { join_t_var = join.join_t_var;
      join_bounds = join.join_bounds;
      join_pred = new_pred;
      join_assertion = new_assertion;
      join_terms = new_join_terms;
      join_body = new_body }
  in
  ([(inspector_target_st,inspector_f);
     (init_target_st,init_f)], 
   new_join)
)

(* *****************************************************************
   Determineds
   ***************************************************************** *)

and process_determined 
    ({ ibr_defs_of_name = defs_of_name; ibr_parent_of = parent_of }
       as stuff) determined_st =
  let ({ determined_term = term; determined_pred = pred;
	 determined_assertion = assertion; determined_body = body }
	 as determined) = stmt2determined determined_st
  in
  try
    let (slice,slice_fvs) = construct_slice stuff determined_st in
    let invar_st = find_invariant_pt stuff slice slice_fvs in
    (* If we get this far, it's profitable, so preceed with the reindexing. *)
    let (inspector_placement,new_executor) =
      reindex_determined slice invar_st body
    in (true,inspector_placement,new_executor)
  with Cant_reindex ->
    let (reindexed_p,placements,body') = outer_walk stuff body in
    (reindexed_p, placements,
     determined2stmt
       { determined_term = term; determined_pred = pred;
	 determined_assertion = assertion; determined_body = body' })

and reindex_determined slice invar_st body =
  (* Unimplemented. *)
  raise Cant_reindex

(* *****************************************************************
   Construct the slice and slice-related info. 
   ***************************************************************** *)

and construct_slice stuff targ_st = (
  let { ibr_defs_of_name = defs_of_name; ibr_parent_of = parent_of } =
    stuff
  and body =
    if stmt_is_join targ_st then
      (stmt2join targ_st).join_body
    else
      (stmt2determined targ_st).determined_body
  in 

  (* *****************************************************************
     The query slice. 

     This is all statements from the top of the partially join-selected (a
     verb?) query down to, and including, targ_st.

     The slice is a list of the statements. The first statement is the
     outermost. The last is targ_st.
     ***************************************************************** *)
  let slice =
    recurs (targ_st,[])
      (fun recurs (st,result) ->
	match st with
	  Stmt_directive(An_list([An_string("bo-query"); _]),
			       	_) -> st::result
	| _ ->
	    let st' =
	      (match parent_of st with
	    	None -> assert false (* No bo-query? how? *)
	      | Some(st') -> st')
	    in
	    recurs (st',st::result))
  in

  (* *****************************************************************
     Check the slice for validity. There are only certain kinds of
     statements that we want to be present if we are going to do
     reindexing.
     ***************************************************************** *)
  List.iter
    (fun st ->
      match st with
      (* These statements are ok. *)
	_ when stmt_is_join st -> ()
      |	_ when stmt_is_determined st -> ()
      |	(Stmt_seq _ | Stmt_for _ | Stmt_while _ | Stmt_let _|
	Stmt_directive _ ) -> ()
      (* Anything else is not. *)
      |	_ -> raise Cant_reindex)
    slice;

  (* *****************************************************************
     Collect the free vars in the slice.

     We have to be careful not to collect the free vars of the statements
     that the slice encloses; these vars do no affect whether or not the
     slice can be hoisted. Only the variables that appear within the slice
     should be recorded.
     ***************************************************************** *)
  let (_,fvs) =
    List.fold_right
      (fun st (child_st,fvs) ->
	let new_fvs =
	  match st with
	    _ when stmt_is_join st ->
	      let { join_t_var = t_var; join_bounds = bounds;
		    join_terms = terms; join_pred = pred;
		    join_assertion = assertion; join_body = body } = 
	      	stmt2join st
	      in
	      assert (child_st == body);
	      let bound_by_join = 
		String_set.add t_var (bound_by_terms terms)
	      and fvs' =
	      	String_set.union fvs 	       	  (String_set.union (bounds_fvs bounds) 
		     (String_set.union (terms_fvs terms)
		     (String_set.union (wff_fvs pred) (wff_fvs assertion))))
	      in String_set.diff fvs' bound_by_join
	  | _ when stmt_is_determined st ->
	      let { determined_term = term; determined_pred = pred;
		    determined_assertion = assertion; 
		    determined_body = body } = 
	      	stmt2determined st
	      in
	      assert (child_st == body);
	      let bound_by_determined = bound_by_term term
	      and fvs' =
	      	String_set.union fvs 	       	  (String_set.union
	      	(term_fvs term) 		     (String_set.union (wff_fvs pred) (wff_fvs assertion)))
	      in String_set.diff fvs' bound_by_determined
	  | Stmt_seq(s1,s2) ->
	      if child_st = s1 then
	      	String_set.union fvs (stmt_fvs s2)
	      else
	      	String_set.union fvs (stmt_fvs s1)
	  | Stmt_for(id,lb,ub,step,body) ->
	      assert (child_st == body);
	      String_set.union (String_set.remove (bmf_ident_name id) fvs)
	      		      	(String_set.union (ident_fvs id)
	      		      	(String_set.union (expr_fvs lb)
	      		      	(String_set.union (expr_fvs ub) ((expr_fvs step)))))
	  | Stmt_while(test,body) ->
	      assert (child_st == body);
	      String_set.union fvs (expr_fvs test)
	  | Stmt_let(Frame(ids),body) ->
	      assert (child_st == body);
	      let ids_name_set =
	      	String_set.of_list (List.map bmf_ident_name ids)
	      in
	      String_set.union (String_set.diff fvs ids_name_set) (idents_fvs ids)
	  | Stmt_directive(ans,body) ->
	      assert (child_st == body);
	      String_set.union fvs (ann_fvs ans)
	  | _ -> assert false
	in
	(st,new_fvs))
      slice (body,empty_s)
  in
  (slice,fvs)
)

(* *****************************************************************
  Find the point to which the query slice can be hoisted. Compute
  profitability along the way.

  Reindexing is profitable if the slice can be hoisted outside of a loop 
  of any kind.

  Abort if hoisting isn't profitable.
  ***************************************************************** *)

and find_invariant_pt stuff slice slice_fvs = (
  let { ibr_defs_of_name = defs_of_name; ibr_parent_of = parent_of } =
    stuff
  in
  let top_st = List.hd slice in
  let get_defs st =
    String_set.fold
      (fun fv defs -> Int_set.union defs (defs_of_name st fv))
      slice_fvs Int_set.empty
  in let slice_defs = get_defs top_st in
  let (profitable_p,invar_st) =
    recurs (false,top_st)
      (fun recurs (profitable_p,st) ->
	match parent_of st with
	  None -> (profitable_p,st)
	| Some(st') ->
	    let defs' = get_defs st' in
	    if not (Int_set.equal slice_defs defs') then
	      (profitable_p,st)
	    else
	      match st' with
		  (* Yes! It's profitable! *)
		(Stmt_for _ | Stmt_do _ | Stmt_while _) ->
		  recurs (true,st')
	      | _ -> recurs (profitable_p,st'))
      
  in
  if not profitable_p then raise Cant_reindex;
  invar_st
)

(* *****************************************************************
   Misc routines.
   ***************************************************************** *)

(* Replace the innermost st in a slice with another. *)
and replace_innermost_of_slice slice st' =
    match slice with
      [] -> assert false
    | [st] -> [st']
    | st::slice' -> st::(replace_innermost_of_slice slice' st')

(* Generating code from slices. Returns an AST constructed from SLICE with
   the body being BODY. *)
and gen_with_slice slice body =
  match slice with
    [] -> body
  | st::slice' when stmt_is_join st ->
      let { join_t_var = t_var; join_bounds = bounds;
	    join_terms = terms; join_pred = pred;
	    join_assertion = assertion } = stmt2join st in
      join2stmt
	{ join_t_var = t_var; join_bounds = bounds;
	  join_terms = terms; join_pred = pred;
	  join_assertion = assertion;
 	  join_body = gen_with_slice slice' body }
  | st::slice' when stmt_is_determined st ->
      let { determined_term = term; determined_pred = pred;
	    determined_assertion = assertion } = stmt2determined st in
      determined2stmt
	{ determined_term = term; determined_pred = pred;
	  determined_assertion = assertion;
 	  determined_body = gen_with_slice slice' body }
  | Stmt_seq(s1,s2)::slice' ->
      assert (slice' <> []);
      let s' = List.hd slice' in
      if s1 = s' then
	Stmt_seq(gen_with_slice slice' body, s2)
      else
	Stmt_seq(s1, gen_with_slice slice' body)
  | Stmt_for(id,lb,ub,step,_)::slice' ->
      Stmt_for(id,lb,ub,step, gen_with_slice slice' body)
  | Stmt_while(test,_)::slice' ->
      Stmt_while(test, gen_with_slice slice' body)
  | Stmt_let(frm,_)::slice' ->
      Stmt_let(frm, gen_with_slice slice' body)
  | Stmt_directive(ann,_)::slice' ->
      Stmt_directive(ann, gen_with_slice slice' body)
  | _ -> assert false  

(* Free and bound vars of various things *)

and bounds_fvs (lbs,ubs) =
  String_set.union (reduce_s expr_fvs lbs) (reduce_s expr_fvs ubs)

and bound_by_terms terms =
  List.fold_left 
    (fun s term -> String_set.union s (bound_by_term term))
    empty_s terms
and bound_by_term term =
  let { jt_kind = kind } = term in
  String_set.union (singleton_s (join_var_of_kind kind))     (singleton_s (enum_var_of_kind kind))

and terms_fvs terms =
  List.fold_left (fun s term -> String_set.union s (term_fvs term))
    empty_s terms
and term_fvs term =
  let { jt_kind = kind; jt_constr = constr; jt_bounds = bounds } = term in
  let rec jtk_fvs jtk =
    match jtk with
      Jtk_enum_simple(enum) -> (jte_fvs enum)
    | Jtk_enum_indexing(enum,lookup) -> 
	String_set.union (jte_fvs enum) (jte_fvs lookup)
    | Jtk_enum_searching(enum,lookup,search) ->
	String_set.union (jte_fvs enum) (String_set.union (jte_fvs lookup) (jte_fvs search))
    | Jtk_lookup(lookup) -> (jte_fvs lookup)
  and jte_fvs jte =
    let Jtk_eq(v,a,name,args) = jte in
    String_set.union (singleton_s a) (reduce_s expr_fvs args)
  in
  String_set.union (jtk_fvs kind)     (String_set.union (expr_fvs constr) (bounds_fvs bounds))

and wff_fvs w =
  match w with
    Wff.Simple.Conjunct(a,b) -> String_set.union (wff_fvs a) (wff_fvs b)
  | Wff.Simple.Disjunct(a,b) -> String_set.union (wff_fvs a) (wff_fvs b)
  | Wff.Simple.Implies(a,b) -> String_set.union (wff_fvs a) (wff_fvs b)
  | Wff.Simple.Equality(a,b) -> String_set.union (wff_fvs a) (wff_fvs b)
  | Wff.Simple.Negation(a) -> wff_fvs a
  | Wff.Simple.Atom(v) -> singleton_s v
  | Wff.Simple.Constant(_) -> empty_s

(* *****************************************************************
   Scheme entry points.
   ***************************************************************** *)

let _ =
  Sx.define_prim1 (Scheme.interaction_environment ())
    "dense_ibr"
    (fun sx -> 
      let bmf = Bmf.from_s_expr sx in
      let bmf' = List.map do_proc bmf in
      let sx' = Bmf.to_s_expr bmf' in
      sx')
