(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
(* todo:
   expr printing
   sort names when printing sets.
   *)

open Printf

open Ssisl
open Containers
open Bsct_base
open Bmf

open Hybrid_dataflow

(************************************************************************)
(* All of the types							*)
(************************************************************************)

type def_t =
    Def_entry of string
  | Def_write of bmf_stmt * bmf_expr
	
let defs_equal def1 def2 =
  match (def1,def2) with
    (Def_entry(name1),Def_entry(name2)) -> name1 = name2
  | (Def_write(s1,e1),Def_write(s2,e2)) -> s1 == s2 && e1 == e2
  | _ -> false

module Def_element =
  struct
    type t = def_t
    let equal = defs_equal
  end

module Def_ordinator = Ordinator(Def_element)

type result_t = Int_set.t Result.t

(************************************************************************)

let empty_s = Int_set.empty
let list_to_set ord lst = 
  List.fold_left
    (fun s d -> Int_set.add (Def_ordinator.to_int ord d) s)
    empty_s lst
let singleton_s ord d = list_to_set ord [d]

let name_of d =
  match d with
    Def_entry(name) -> name
  | Def_write(_,Expr_var(name)) -> name
  | Def_write(_,Expr_access(_,Expr_var(name),_,_)) -> name
  | _ -> assert false

(************************************************************************)
(* Computing the reaching defs						*)
(************************************************************************)

module Df = Solver(Forward_gen_kill_params)

let rec compute_reaching_defs stmt =
  let (ord, in_ht, out_ht, gens_ht, kills_ht) =
    compute_reaching_defs' stmt
  in (ord,in_ht, out_ht)

and compute_reaching_defs' stmt =
  let ord = Def_ordinator.create () in
  (* 1. Find all names that appear in the stmt tree *)
  let all_names = find_all_names stmt
  in let (gens_ht,kills_ht) = 
    ((Result.create 101 : result_t),
     (Result.create 101 : result_t))
  (* 2. Compute the set of defs generated at each stmt *)
  in let _ = compute_gens ord gens_ht stmt
  (* 3. Compute the set of free vars and then the set of entry defs.*)
  and entry_defs = compute_entry_defs ord stmt
  (* 4. Compute the set of definitions killed at each stmt *)
  in let _ = compute_kills ord gens_ht entry_defs kills_ht
  (* 5. Solve the analysis problem. *)
  in let (in_ht, out_ht) = 
    Df.forward_solve (gens_ht,kills_ht) stmt entry_defs
  in (ord, in_ht, out_ht, gens_ht, kills_ht)

(* 1. Find all names that appear in the stmt tree *)
and find_all_names stmt =

  let empty = String_set.empty in
  let singleton name = String_set.add name empty in
  let reduce f lst =
    List.fold_left 
      (fun result x -> String_set.union result (f x))
      empty lst
  in

  let rec w_stmt (s: bmf_stmt) =
    match s with
      Stmt_assign(lhs,rhs) -> String_set.union (w_expr lhs) (w_expr rhs)
    | Stmt_seq(first,second) -> String_set.union (w_stmt first) (w_stmt second)
    | Stmt_call(f,res,args) -> reduce w_expr (f::res::args)
    | Stmt_for(id,lb,ub,step,body) ->
	String_set.union (w_id id) 	  (String_set.union (reduce w_expr [lb;ub;step]) (w_stmt body))
    | Stmt_while(test,body) -> String_set.union (w_expr test) (w_stmt body)
    | Stmt_if(test,thn,els) -> 
	String_set.union (w_expr test) (reduce w_stmt [thn;els]) 
    | Stmt_let(Frame(ids),body) ->
	String_set.union (reduce w_id ids) (w_stmt body)
    | Stmt_do(pred,body) -> String_set.union (w_expr pred) (w_stmt body)
    | Stmt_nop -> empty
    | Stmt_directive(ann,body) -> String_set.union (w_ann ann) (w_stmt body)
  and w_expr (e: bmf_expr) =
    match e with
    | (Expr_int _ | Expr_real _ | Expr_string _ | Expr_bool _ | 
      Expr_void ) -> empty
    | Expr_var(name) -> singleton name 
    | Expr_op(_,args) -> reduce w_expr args
    | Expr_forall(Frame(ids),pred) -> 
	String_set.union (reduce w_id ids) (w_expr pred)
    | Expr_exist(Frame(ids),pred) ->
	String_set.union (reduce w_id ids) (w_expr pred)
    | Expr_set(Frame(ids),pred) ->
	String_set.union (reduce w_id ids) (w_expr pred)
    | Expr_access(_,a,_,args) -> String_set.union (w_expr a) (reduce w_expr args)
  and w_type (ty: bmf_type) =
    match ty with
      (Type_scalar _ | Type_unknown)-> empty
    | Type_proc(args_tys,res_ty) -> reduce w_type (res_ty::args_tys)
    | Type_array(atyp) -> w_atype atyp
    | Type_ptr(btyp) -> w_type btyp
    | Type_ann(ann) -> w_ann ann
  and w_atype (atyp: array_type) =
    match atyp with
      Array_dense(_,ubs,_,_) -> reduce w_expr ubs
    | Array_bb(_,_,_,args) -> 
	reduce w_ann (List.map snd args)
    | Array_view(Frame(out_ids),Frame(in_ids),constrs,atyp') ->
	String_set.union (reduce w_id (out_ids@in_ids))
	(String_set.union (w_expr constrs) (w_atype atyp'))
  and w_ann (an: annotation) =
    match an with
      An_type(typ) -> w_type typ
    | An_stmt(s) -> w_stmt s
    | An_expr(e) -> w_expr e
    | An_string _ -> empty
    | An_list(ans) -> reduce w_ann ans
  and w_id (id: bmf_ident) =
    match id with
      Ident_null -> empty
    | Ident_valid(name,typ,stuff,an) ->
	let results_stuff =
	  (match stuff with
	    Ident_var(init) -> w_expr init
	  | Ident_bind -> empty
	  | Ident_proc(Frame(arg_ids),res_id,body) ->
	      String_set.union (reduce w_id (res_id::arg_ids)) (w_stmt body))
	in
	String_set.union (singleton name)
 	  (String_set.union (w_type typ) 
	     (String_set.union results_stuff (w_ann an)))
    
  in w_stmt stmt

(* 2. Compute the set of defs generated at each stmt *)
and compute_gens ord gens_ht stmt =
    let gens =
      match stmt with
      	Stmt_assign(lhs,rhs) -> 
	  singleton_s ord (Def_write(stmt,lhs))
      | Stmt_call(f,res,args) -> 
	  (* We're assuming call by references, so everything is a
	     potential def. *)
	  list_to_set ord
	    (List.map 
	       (fun ex -> Def_write(stmt,ex))
	       (List'.remove_if (fun ex -> ex = Expr_void)
		  (res::args)))
      | Stmt_for(id,lb,ub,step,body) -> 
	  singleton_s ord (Def_entry(bmf_ident_name id))
      | Stmt_let(Frame(ids),body) ->
	  list_to_set ord
	    (List.map (fun id -> Def_entry(bmf_ident_name id)) ids)
      | Stmt_do(Expr_set(Frame(ids),pred),body) -> 
	  list_to_set ord
	    (List.map (fun id -> Def_entry(bmf_ident_name id)) ids)
      |	Stmt_do _ -> assert false
      | ( Stmt_seq _ | Stmt_while _ | Stmt_if _ | Stmt_nop | 
	Stmt_directive _ ) -> empty_s
    in 
    (* Double-check the validity of the def that we just created. *)
    let _ =
      Int_set.iter
	(fun i -> assert (name_of (Def_ordinator.of_int ord i) <> ""))
	gens
    in
    (* Set the gens for the current stmt. *)
    let _ = Result.add gens_ht stmt gens in
    (* Now, recurs. *)
    let recurs = compute_gens ord gens_ht in
    match stmt with
      	Stmt_assign(lhs,rhs) -> ()
      | Stmt_seq(first,second) -> recurs first; recurs second
      | Stmt_call(f,res,args) -> ()
      | Stmt_for(id,lb,ub,step,body) -> recurs body
      | Stmt_while(test,body) -> recurs body
      | Stmt_if(test,thn,els) -> recurs thn; recurs els
      | Stmt_let(Frame(ids),body) -> recurs body
      | Stmt_do(pred,body) ->  recurs body
      | Stmt_nop -> ()
      | Stmt_directive(ann,body) -> recurs body


(* 3. Compute the set of free vars and then the set of entry defs.*)
and compute_entry_defs ord stmt =
  let fvs_of_stmt = Bmf_util.stmt_fvs stmt in
  let entry_defs =
    Int_set.of_list
      (List.map (fun name -> Def_ordinator.to_int ord (Def_entry(name)))
	 (String_set.to_list fvs_of_stmt))
  in entry_defs

(* 4. Compute the set of definitions killed at each stmt *)
and compute_kills ord gens_ht entry_defs kills_ht =
  (* Group defs by name *)
  let dt = String_hashtbl.create 101 in
  let _ =
    (* Insert all of these defs into dt by their name. *)
    for i = 0 to (Def_ordinator.size ord)-1 do
      let def = Def_ordinator.of_int ord i in
      String_hashtbl.add dt (name_of def) i
    done
  in
  let i_defs_by_name = String_hashtbl.create 101 in
  let _ =
    (* Collect all (name,def) pairs into a single (name, {defs}) pair, and
       insert it into i_defs_by_name *)
    List.iter
      (fun name ->
	let i_defs_s = 
	  Int_set.of_list (String_hashtbl.find_all dt name) 
	in
	String_hashtbl.add i_defs_by_name name i_defs_s)
      (String_hashtbl.keys dt)
  in
  (* Function to determine whether a def is a "killing" def. *)
  let killing_i_def_p i_def =
    let def = Def_ordinator.of_int ord i_def in
    match def with
      (* Entry points are killing points of all defs for a name since the
	 values do not persist outside of the id's scope. *)
      Def_entry _ -> true
      (* A write to a var name (as opposed to an array ref, for instance),
	 is assumed to overwrite the variable entirely. This is certainly
	 true for scalars. Is that enough? *)
    | Def_write(_,Expr_var _) -> true
    | Def_write _ -> false
  in
  (* Assign kill sets to each stmt *)
  (*   for each (stmt,gened_defs) pair, *)
  Result.iter
    (fun st gened_i_defs ->
      (* The killing defs of st *)
      let killing_i_defs_lst = 
	List'.remove_if_not killing_i_def_p 
	  (Int_set.to_list gened_i_defs)
      in
      (* All defs killed by the killing defs *)
      let killed_i_defs_s =
	List.fold_left
	  (fun killed_i_defs_s killing_i_def ->
	    let killing_def = Def_ordinator.of_int ord killing_i_def in
	    Int_set.union killed_i_defs_s 
	      (String_hashtbl.find i_defs_by_name (name_of killing_def)))
	  empty_s killing_i_defs_lst
      in
      (* Insert into kills_ht *)
      Result.add kills_ht st killed_i_defs_s)
    gens_ht

(************************************************************************)
(* Group defs by name and aliases					*)
(************************************************************************)

(* Group defs in ht by (st,name) pairs *)
module St_ex_element =
  struct
    type t = bmf_stmt * string
    let equal (st1,name1) (st2,name2) =
      st1 == st2 && name1 = name2
    let hash = Containers.hash
  end
module St_ex_hashtbl = Hashtbl(St_ex_element)

let defs_at_stmt_by_name ord ht aliases = 
  let stmt_name_ht = St_ex_hashtbl.create 101 in
  Result.iter
    (fun st i_defs ->
      Int_set.iter
	(fun i_def ->
	  let def = Def_ordinator.of_int ord i_def in
	  St_ex_hashtbl.add stmt_name_ht (st,name_of def) i_def)
	i_defs)
    ht;
  (* function that maps st and name to the defs at st for name and all of
     its aliases. *)
  let make_result st name =
    let name_and_aliases = String_map.find name aliases in
    String_set.fold
      (fun name s -> 
	let i_defs_of_name_lst =
	  try St_ex_hashtbl.find_all stmt_name_ht (st,name)
	  with Not_found -> []
	in Int_set.absorb s i_defs_of_name_lst)
      name_and_aliases
      Int_set.empty
  in 
  (* Return a function that memoizes the results. *)
  let cache_ht = St_ex_hashtbl.create 101 in
  (fun st name ->
    let key = (st,name) in
    try St_ex_hashtbl.find cache_ht key
    with Not_found ->
      let result = make_result st name in
      St_ex_hashtbl.add cache_ht key result;
      result)


(************************************************************************)
(* Routines for testing.						*)
(************************************************************************)

module Stmt_element =
  struct
    type t = bmf_stmt
    let equal = (==)
  end
module Stmt_ordinator = Ordinator(Stmt_element)

let rec test_proc id =
  match Bmf_printer.normalize_global id with
  | Ident_valid(name,typ,Ident_proc(Frame(arg_ids),res_id,body),ann) ->
      let (def_ord,in_ht,out_ht,gens_ht,kills_ht) = 
	compute_reaching_defs' body in
      let (stmt_ord,body') = number_stmts body in
      let id' = Ident_valid(name,typ,
			    Ident_proc(Frame(arg_ids),res_id,body'),ann) in
      printf "\n*** Results for %s ***\n" name;
      print_results def_ord stmt_ord in_ht out_ht gens_ht kills_ht;
      printf "\nThe code:\n";
      Bmf_printer.unparse_stdout 80 [id']
  | _ -> ()

and number_stmts body =
  let ord = Stmt_ordinator.create () in
  let body' =
    recurs body
      (fun recurs st ->
	let number f =
	  let i = Stmt_ordinator.to_int ord st in
	  Stmt_directive
	    (An_list([An_string("#"^(string_of_int i))]),
	     f ())
	in
	match st with
	  Stmt_nop -> st
	| (Stmt_assign _ | Stmt_call _) -> number (fun () -> st)
	| Stmt_seq(first,second) ->
	    let first' = recurs first in
	    let second' = recurs second in
	    Stmt_seq(first', second')
	| Stmt_for(id,lb,ub,step,body) ->
	    number (fun () -> Stmt_for(id,lb,ub,step,recurs body))
	| Stmt_while(test,body) -> 
	    number (fun () -> Stmt_while(test,recurs body))
	| Stmt_if(test,thn,els) ->
	    number
	      (fun () -> 
	      	let thn' = recurs thn in
	      	let els' = recurs els in
	      	Stmt_if(test,thn',els'))
	| Stmt_let(ids,body) -> 
	    number (fun () -> Stmt_let(ids,recurs body))
	| Stmt_do(pred,body) ->
	    number (fun () -> Stmt_do(pred,recurs body))
	| Stmt_directive(ann,body) -> Stmt_directive(ann,recurs body)
	      )
  in (ord, body')

and print_results def_ord stmt_ord in_ht out_ht gens_ht kills_ht =

  (* Print the defs *)
  printf "\nDefs:\n";
  let ex_printer outch ex =
    let fmt =
      Format.make_formatter (output outch) (fun () -> flush outch)
    in
    Format.pp_set_margin fmt (999999999-1);
    Bmf_printer.expr_printer fmt ex;
    Format.pp_print_flush fmt ()
  in
  for j = 0 to (Def_ordinator.size def_ord)-1 do
    match Def_ordinator.of_int def_ord j with
      Def_entry(name) -> printf "%3d: Def_entry(\"%s\")\n" j name
    | Def_write(st,ex) ->
	printf "%3d: Def_write(%d,%a)\n" j 
	  (Stmt_ordinator.to_int stmt_ord st) ex_printer ex
  done;

  (* Print the in and out sets. *)
  printf "\nReaching defs:\n";
  for i = 0 to (Stmt_ordinator.size stmt_ord)-1 do
    let st = Stmt_ordinator.of_int stmt_ord i in
    printf "\nin[%d] =" i;
    print_defs_set def_ord (Result.find in_ht st);
    printf "\ngen[%d] =" i;
    print_defs_set def_ord (Result.find gens_ht st);
    printf "\nkill[%d] =" i;
    print_defs_set def_ord (Result.find kills_ht st);
    printf "\nout[%d] =" i;
    print_defs_set def_ord (Result.find out_ht st);
    printf "\n"
  done

and print_defs_set def_ord s =
  let defs_by_name = String_hashtbl.create 101
  in
  (* Group the defs by name *)
  Int_set.iter
    (fun i_def ->
      let def = Def_ordinator.of_int def_ord i_def in
      String_hashtbl.add defs_by_name (name_of def) i_def)
    s;
  (* for each name... *)
  List.iter
    (fun name ->
      printf " %s:" name;
      (* for each def ... *)
      List.iter
	(fun def_num -> printf " %d" def_num)
	(String_hashtbl.find_all defs_by_name name))
    (String_hashtbl.keys defs_by_name)


