(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Containers

open Bmf;;
open Bb;;
open Bb_decl;;
open Bmf_parser_aux;;
open Bmf_join;;

type inst_what_type = Inst_just_dense | Inst_everything;;

(************************************************************************
Overview

This code implements Black Box instantiation.

There are six primary constructs that have to be handled:
o [[bmf_stmt]]
o [[bmf_expr]]
o Global [[bmf_frame]]
o Parameter [[bmf_frame]] and result [[bmf_ident]]
o Let [[bmf_frame]]
o Quantified and unquantified set [[bmf_frame]].

Access expressions may legally occur within [[bmf_expr]]'s that appear
within, [[bmf_stmt]]'s, other [[bmf_expr]]'s, and [[ident_var]]
initialization expressions and handle declaration arguments within let
[[bmf_frame]]'s. Of course, access expressions may appear within the bodies
of top level [[ident_proc]]'s, but they may not appear in parameter or set
frames.

The tricky thing about instantiating access expressions within
[[bmf_expr]]'s is that it may require inserting statements around the
statement that contains the access expression. For instance, the
[[stmt_assign]]
\begin{quote}\begin{verbatim}
x := ... A[[ "search", i ]] ...;
\end{verbatim}\end{quote}
may result in the code,
\begin{quote}\begin{verbatim}
... code for binary search ...
if (found) {
    x := ... Aval[[ii]] ...;
}
\end{verbatim}\end{quote}
This is reflected in the fact that the functional access methods in the
black boxes have the types,
\begin{quote}\begin{verbatim}
type am_lookup_type ==
    bmf_expr list              ( * args to acc expr * )
    -> ( bmf_expr -> bmf_stmt) ( * found contin. * )
    -> bmf_stmt;;              ( * result * )

type am_search_type ==
    bmf_expr list              ( * args to acc expr * )
    -> ( bmf_expr -> bmf_stmt) ( * found contin. * )
    -> ( unit -> bmf_stmt )    ( * not found contin. * )
    -> bmf_stmt;;              ( * result * )
\end{verbatim}\end{quote}
A continuation argument is used to place the transformed access expression
back into the code, and the result type of the access methods is
[[bmf_stmt]].

As a result of this continuation-passing style (CPS) in the Black Box
protocal, portions of the instantiator are continuation passing as
well\footnote{
  Actually, none of this is exactly CPS. It's really a hybrid
  of CPS and more conventional call/return.
}. In particular,
[[process_expr]], [[process_let_frame]], and [[process_ann]] are in CPS.
************************************************************************)


(* Intrinsic names *)

let handle_init =  "%handle_init";;
let handle_incr =  "%handle_incr";;
let handle_close = "%handle_close";;
let handle_valid = "%handle_valid";;
let handle_deref = "%handle_deref";;

module Expr_element =
  struct
    type t = bmf_expr
    let compare = Pervasives.compare
  end
module Expr_map = Map(Expr_element)

(* Types *)

(* There are two binding environments that are maintained as the program is *)
(* traversed.                                                               *)

(* The [[id_env]] records declarations of Black Box relation identifiers,   *)
(* handle identifiers, and primitive array identifiers.                     *)
type id_env_type = id_entry_type String_map.t
     and id_entry_type =
         Id_relvar of black_box
     |   Id_handle of am_handler_record
     |   Id_prim_array of black_box

(* The [[acc_env]] records mappings from access expressions to their        *)
(* replacement expressions after instantiation. It is assumed that both the *)
(* key and value expressions stored in [[acc_env]] have already been        *)
(* instantiated.                                                            *)
and acc_env_type = bmf_expr Expr_map.t
;;

(* Searches for an access method. If not found, throw an appropriate
   exception. *)

let get_method name bbi =
  try find_method name bbi
  with Not_found -> assert false (* no such access method. *)
    
let empty_id_env = String_map.empty;;
let rec add_to_id_env name id id_env =
    String_map.add name id id_env
and find_in_id_env name id_env =
    String_map.find name id_env
and id_of_rel_var_expr r id_env =
    let r_name =
        match r with
          Expr_var(r_name) -> r_name
        | _ -> assert false (* Expected a expr_var *)

    in
        try
            find_in_id_env r_name id_env
        with
            Not_found ->
                assert false (* No decl found for target of access. *)
;;

let empty_acc_env = Expr_map.empty;;
let rec add_to_acc_env orig replc acc_env =
    Expr_map.add orig replc acc_env
and find_in_acc_env orig acc_env =
    Expr_map.find orig acc_env
;;

let make_max lbs =
        match lbs with
          [] -> assert false (* Can't happen *)
        | [lb] -> lb
        | lbs -> Expr_op(Op_other("max"),lbs)
and make_min ubs =
        match ubs with
          [] -> assert false (* Can't happen *)
        | [ub] -> ub
        | ubs -> Expr_op(Op_other("min"),ubs)
;;

let rec instantiate global_decls inst_what =
    let (Frame(global_decls')) =
	process_global_frame inst_what (Frame(global_decls))
    in
	global_decls'


(* The core routines *)

(* The routine [[process_exprs]] is used to process lists of expressions.   *)
and process_exprs inst_what id_env acc_env es k =
    match es with
      [] -> k acc_env []
    | e::remaining_es ->
	process_expr inst_what id_env acc_env e
	    (fun acc_env' e' ->
		process_exprs inst_what id_env acc_env' remaining_es
		    (fun acc_env' remaining_es' ->
			k acc_env' (e'::remaining_es')))

(* The function for processing [[bmf_expr]]'s does a depth-first traversal*)
(* of the expression tree in a CPS manner.*)
and process_expr inst_what (id_env: id_env_type) (acc_env: acc_env_type)
	(e: bmf_expr) (k: (acc_env_type -> bmf_expr -> bmf_stmt)) = ((
    (* In the following, each of the kinds of expressions*)
    (* that require instantiator are discussed. *)
    match e with
    (* Traverse [[R]] and [[args]]. *)
    | Expr_access(acc,r,name,args) ->
	process_expr inst_what id_env acc_env r
	    (fun acc_env' r' ->
		process_exprs inst_what id_env acc_env' args 
		    (fun acc_env'' args' ->
			instantiate_access_expr inst_what id_env
			    acc r' name args' acc_env'' k))
    (* Set inclusion test *)
    | Expr_op(Op_other("in"),lst) ->
	  let (val0,r,name,args,bounds_expr) =
	      match lst with
		[val0;Expr_access(Access_strong,r,name,args)] ->
		  (val0,r,name,args,Expr_bool(true))
	      | [val0;Expr_access(Access_strong,r,name,args);bounds_expr] ->
		  (val0,r,name,args,bounds_expr)
	      | _ -> assert false (* Can't happen 5 *)
	  in
	      process_expr inst_what id_env acc_env val0
		  (fun acc_env' val' ->
		      process_expr inst_what id_env acc_env' r
			  (fun acc_env'' r' ->
			      process_exprs inst_what id_env acc_env'' args
				  (fun acc_env''' args' ->
				      process_expr inst_what id_env
					    acc_env''' bounds_expr
					  (fun acc_env'''' bounds_expr' ->
					      instantiate_inclusion_check
						inst_what id_env
						val' r' name args' bounds_expr'
						acc_env'''' k))))
    (* Handle validity and deferencing *)
    | Expr_op(Op_other(f),args) when f = handle_valid ->
	  let h = deref_handle_args args id_env
	  in
	      k acc_env h.valid_ex
    | Expr_op(Op_other(f),args) when f = handle_deref ->
	  let h = deref_handle_args args id_env
	  in
	      k acc_env h.deref_ex
    | (Expr_int(_)|Expr_real(_)|Expr_string(_)|Expr_bool(_)
	    |Expr_void|Expr_var(_)) ->
	k acc_env e
    | Expr_op(op,args) ->
	process_exprs inst_what id_env acc_env args
	    (fun acc_env' args' ->
		k acc_env' (Expr_op(op,args')))
    | Expr_forall(frm,pred) ->
	process_expr inst_what
	    (process_set_frame inst_what id_env acc_env frm)
	    acc_env pred
	    (fun acc_env' pred' ->
		k acc_env' (Expr_forall(frm,pred')))
    | Expr_exist(frm,pred) ->
	process_expr inst_what
	    (process_set_frame inst_what id_env acc_env frm)
	    acc_env pred
	    (fun acc_env' pred' ->
		k acc_env' (Expr_exist(frm,pred')))
    | Expr_set(frm,pred) ->
	process_expr inst_what
	    (process_set_frame inst_what id_env acc_env frm)
	    acc_env pred
	    (fun acc_env' pred' ->
		k acc_env' (Expr_set(frm,pred')))
) :  bmf_stmt)

(* We examine [[R']] and determine whether this is an*)
(* access to a  primitive array or a relation. Primitive array*)
(* accesses are simply generated. Relation access methods have*)
(* to instantiated. *)
and instantiate_access_expr inst_what id_env acc r' name args' acc_env'' k =
    let e' = Expr_access(acc,r',name,args')
    and bbi = get_bbi_from_rel_expr r' id_env
    in
	(match acc with
	  Access_strong ->
	      instantiate_strong_access inst_what bbi
		  acc_env'' e' k
	| Access_lookup 
	      when inst_what = Inst_everything ->
	    let b = Misc.gensym("b")
	    in let b_v = Expr_var(b)
	    in let new_id =
		Ident_valid(b,Type_scalar(Type_bool),
		    Ident_var(Expr_void),An_list([]))
	    and new_first =
		instantiate_lookup_access inst_what bbi acc_env'' e'
		    (fun _ result -> 
			Stmt_assign(b_v,result))
		    (fun _ result ->
			Stmt_assign(b_v,result))
	    and new_second =
		k (add_to_acc_env e' b_v acc_env'') b_v
	    in
		Stmt_let(Frame([new_id]),
		    Stmt_seq(new_first,new_second))
	| Access_lookup -> k acc_env'' e'
	| Access_weak -> 
	    failwith "Instantiator: Access_weak is not implemented.")

(* Instantiating set inclusion check *)
and instantiate_inclusion_check inst_what id_env
	val' r' name args' bounds_expr'
	acc_env k =
    let bbi = (get_bbi_from_rel_expr r' id_env)
    and check_e = (Expr_op(Op_other("in"),[
		      val';
		      Expr_access(Access_strong,r',name,args');
		      bounds_expr']))
    in try
	let replc = find_in_acc_env check_e acc_env
	in
	    assert (replc<>Expr_void)
		(* (Failure("How did that happen?!?")) *);
	    k acc_env replc
    with Not_found ->
	let (lbs,ubs) = expr2bounds val' bounds_expr'
	in let { am_info = mthd } = get_method name bbi
	in
	    (* We generate different inclusion tests depending upon the *)
	    (* kind of access method *)
	    match mthd with
	    (* The access method is for an interval. Just generate an*)
	    (* interval test. *)
	      Am_rel(Am_rel_interval(range_f)) ->
		range_f args'
		    (fun lb ub ->
			let lb' = make_max (lb::lbs)
			and ub' = make_min (ub::ubs)
			in let test =
			    make_and [
				Expr_op(Op_leq,[lb';val']);
				Expr_op(Op_leq,[val';ub'])]
			in
			    k (add_to_acc_env check_e test acc_env) test)
	    (* Don't instantiate stream relation methods during the dense*)
	    (* run. *)
	    | Am_rel(Am_rel_general(_))
		    when inst_what = Inst_just_dense ->
		k acc_env check_e
	    (* The access method .... *)
	    | Am_rel(Am_rel_general(_)) ->
		let b = Misc.gensym("b")
		in let b_v = Expr_var(b)
		in let search_code =
		    (parse_stmt "{
			    var $h : type_ann <<\"handle\" ($acc)>>;
			    $b := false;
			    call $handle_init($h);
			    while (%op($handle_valid)($h) && !$b) {
				$b := (%op($handle_deref)($h) == $val);
				if (!$b) {
				    call $handle_incr($h);
				}
			    }
			    call $handle_close($h);
			}" [("$h",An_string(Misc.gensym "h"));
			    ("$b",An_string(b));
			    ("$val",An_expr(val'));
			    ("$acc",An_expr(
				Expr_access(Access_strong,r',name,args')));
			    ("$handle_init",An_string(handle_init));
			    ("$handle_incr",An_string(handle_incr));
			    ("$handle_close",An_string(handle_close));
			    ("$handle_valid",An_string(handle_valid));
			    ("$handle_deref",An_string(handle_deref))
			    ])
		in
		    Stmt_let(Frame([
				Ident_valid(b,Type_scalar(Type_bool),
				    Ident_var(Expr_void),An_list([]))]),
			Stmt_seq(
			    (process_stmt inst_what id_env acc_env
				search_code),
			    (k (add_to_acc_env check_e b_v acc_env)
				b_v)))
	    | _ -> assert false (* Can't happen 7 *)

(* Instantiating strong relation accesses *)
and instantiate_strong_access inst_what bbi acc_env strong_e k =
    try
	let replc = find_in_acc_env strong_e acc_env
	in
	    assert (replc<>Expr_void)
		(* (Failure("expr_void -- bogus strong access?")) *);
	    k acc_env replc
    with Not_found ->
	let (r,name,args) =
	    match strong_e with
	      Expr_access(Access_strong,r,name,args) -> (r,name,args)
	    | _ -> assert false (* Can't happen 1 *)
	in let { am_info = mthd } = get_method name bbi
	in let found_f replc =
	    k (add_to_acc_env strong_e replc acc_env) replc
	and not_found_f () =
	    Stmt_call(Expr_var("%fail"),Expr_void,[])
	in
	    match mthd with
	      Am_fun(cost,kind)
		    when cost = O_1 || inst_what = Inst_everything ->
		(match kind with
		  Am_fun_lookup(lookup_f) ->
		    lookup_f args found_f
		| Am_fun_search(search_f) ->
		    search_f args found_f not_found_f)
	    | Am_fun(_,_) -> k acc_env strong_e
	    | _ -> assert false (* Can't happen 2 *)

(* Instantiating lookup relation accesses *)
and instantiate_lookup_access inst_what bbi acc_env lookup_e found_k
	not_found_k =
    assert (inst_what = Inst_everything)
	(* (Failure("Too late for just dense.")) *);
    try
	let replc = find_in_acc_env lookup_e acc_env
	in
	    match replc with
	      Expr_void ->
		assert false (* expr_void -- bogus lookup access? *)
	    | Expr_bool(true) -> found_k acc_env replc
	    | Expr_bool(false) -> not_found_k acc_env replc
	    | _ ->
		Stmt_if(replc,
		    found_k acc_env replc,
		    not_found_k acc_env replc)
    with Not_found ->
	let (r,name,args) =
	    match lookup_e with
	      Expr_access(Access_lookup,r,name,args) -> (r,name,args)
	    | _ -> assert false (* Can't happen 3 *)
	in let { am_info = mthd } = get_method name bbi
	in let strong_e = Expr_access(Access_strong,r,name,args)
	in let found_f replc =
	    let acc_env' =
		add_to_acc_env lookup_e (Expr_bool(true))
		    (add_to_acc_env strong_e replc acc_env)
	    in
		found_k acc_env' replc
	and not_found_f () =
	    let acc_env' =
		add_to_acc_env lookup_e (Expr_bool(false))
		    (add_to_acc_env strong_e Expr_void acc_env)
	    in
		not_found_k acc_env' Expr_void
	in
	    match mthd with
	      Am_fun(_,Am_fun_lookup(lookup_f)) ->
		lookup_f args found_f
	    | Am_fun(_,Am_fun_search(search_f)) ->
		search_f args found_f not_found_f
	    | _ -> assert false (* Can't happen 4 *)

and process_term inst_what id_env acc_env term k =
    let { jt_kind = kind; jt_flag = flag;
	  jt_constr = constr; jt_bounds = (lbs,ubs) } =
	term
    in let k' acc_env' kind' =
	let rec continuation_chain () =
	    process_expr inst_what id_env acc_env' constr
		(fun acc_env'' constr' ->
		    process_exprs inst_what id_env acc_env'' lbs
			(fun acc_env''' lbs' ->
			    process_exprs inst_what id_env acc_env''' ubs
				(fun acc_env'''' ubs' ->
				    build_result constr' lbs' ubs'
					acc_env'''')))
	and build_result constr' lbs' ubs' acc_env'''' =
	    k acc_env''''
		{ jt_kind = kind';
		  jt_flag = flag;
		  jt_constr = constr';
		  jt_bounds = (lbs',ubs') }
	in
	    continuation_chain ()
    in let process_jtk_eq acc_env (Jtk_eq(var,a,name,args)) k =
	process_exprs inst_what id_env acc_env args
	    (fun acc_env' args' ->
		k acc_env' (Jtk_eq(var,a,name,args')))
    in begin
	match kind with
	  Jtk_enum_simple(enum_eq) ->
	    process_jtk_eq acc_env enum_eq
		(fun acc_env' enum_eq' ->
		    k' acc_env' (Jtk_enum_simple(enum_eq)))
	| Jtk_enum_indexing(enum_eq,lookup_eq) ->
	    process_jtk_eq  acc_env enum_eq
		(fun acc_env' enum_eq' ->
		    process_jtk_eq  acc_env' lookup_eq
			(fun acc_env'' lookup_eq' ->
			    k' acc_env''
				(Jtk_enum_indexing(enum_eq',lookup_eq'))))
	| Jtk_enum_searching(enum_eq,lookup_eq,search_eq) ->
	    process_jtk_eq  acc_env enum_eq
		(fun acc_env' enum_eq' ->
		    process_jtk_eq  acc_env' lookup_eq
			(fun acc_env'' lookup_eq' ->
			    process_jtk_eq  acc_env'' search_eq
				(fun acc_env''' search_eq' ->
				    k' acc_env'''
					(Jtk_enum_searching(enum_eq',lookup_eq',
					    search_eq')))))
	| Jtk_lookup(lookup_eq) ->
	    process_jtk_eq acc_env lookup_eq
		(fun acc_env' lookup_eq' ->
		    k' acc_env' (Jtk_lookup(lookup_eq)))
    end

and process_terms inst_what id_env acc_env ts k =
    match ts with
      [] -> k acc_env []
    | t::remaining_ts ->
	process_term inst_what id_env acc_env t
	    (fun acc_env' t' ->
		process_terms inst_what id_env acc_env' remaining_ts
		    (fun acc_env' remaining_ts' ->
			k acc_env' (t'::remaining_ts')))

(* Processing Annotations *)
and process_ann inst_what (id_env: id_env_type) (acc_env: acc_env_type)
	(ann: annotation) (k: (acc_env_type -> annotation -> bmf_stmt)) = ((
    match ann with
      An_type(typ) -> (* todo *) k acc_env ann
    | An_stmt(s) ->
	k acc_env (An_stmt(process_stmt inst_what id_env acc_env s))
    | An_expr(e) ->
	process_expr inst_what id_env acc_env e
	    (fun acc_env' e' ->
		k acc_env' (An_expr(e')))
    | An_string(str) -> k acc_env ann
    | An_list([]) -> k acc_env ann
    | An_list(x::remaining_lst) ->
	process_ann inst_what id_env acc_env x
	    (fun acc_env' x' ->
		process_ann inst_what id_env acc_env' (An_list(remaining_lst))
		    (fun acc_env'' t ->
			let remaining_lst' =
			    match t with
			      An_list(remaining_lst') -> remaining_lst'
			    | _ -> assert false (* Can't happen 9 *)
			in
			    k acc_env'' (An_list(x'::remaining_lst'))))
) : bmf_stmt)

(* Processing [[bmf_stmt]]'s *)

(* Most of this is pretty straight forward ``CPS''. The only cases that we  *)
(* need to treat specially are [[stmt_do]]'s over relational access methods, *)
(* [[stmt_if]]'s with lookup access methods, and functions for handle       *)
(* manipulation.                                                            *)
(* The remaining routines are not in CPS\footnote{                          *)
(*     It might be possible to use CPS to pass [[acc_env]]'s between        *)
(*     statements (e.g., [[stmt_seq]]), but it's not clear to me when it is *)
(*     safe to do so, so I haven't implemented this.                        *)
(* }.                                                                       *)


and process_stmt inst_what (id_env: id_env_type) (acc_env: acc_env_type)
	(s: bmf_stmt) = ((
    match s with
      (* Relational access enumeration *)
      Stmt_do(Expr_set(Frame([
		      Ident_valid(index,Type_scalar(sc_ty),Ident_bind,
			  An_list([]))]),
		  Expr_op(Op_other("in"),lst)),
	      body) ->
	  let (r,name,args,bounds_expr) =
	      match lst with
		[Expr_var(index'); Expr_access(Access_strong,r,name,args)]
		      when index = index' ->
		  (r,name,args,Expr_bool(true))
	      | [Expr_var(index'); Expr_access(Access_strong,r,name,args);
		      bounds_expr] when index = index' ->


		  (r,name,args,bounds_expr)
	      | _ -> assert false (* Can't happen 17 *)
	  in let rec continuation_chains () =
	      process_expr inst_what id_env acc_env r
		  (fun acc_env' r' ->
		      process_exprs inst_what id_env acc_env' args
			  (fun acc_env'' args' ->
			      process_expr inst_what id_env acc_env''
				    bounds_expr
				  (fun acc_env''' bounds_expr' ->
				      build_result r' args' bounds_expr'
					acc_env''')))
	  and build_result r' args' bounds_expr' acc_env''' =
	      let body' =
		  process_stmt inst_what id_env acc_env''' body
	      in
	      (* Instantiate relational access enum *)
	      let bbi = get_bbi_from_rel_expr r' id_env
	      in let { am_info = mthd } = get_method name bbi
	      in let body'' =
		  if bounds_expr' = Expr_bool(true) then
		      body'
		  else
		      Stmt_if(bounds_expr',body',Stmt_nop)
	      in
		  match mthd with
		    Am_rel(Am_rel_interval(range_f)) ->
		      range_f args'
			  (fun lb ub ->
			      process_stmt inst_what id_env acc_env'''
				  (parse_stmt "{
				      <<\"do-any\">>
				      for $i := $lb:$ub { $body; }
				  }" [("$i",An_string(index));
				      ("$lb",An_expr(lb));
				      ("$ub",An_expr(ub));
				      ("$body",An_stmt(body''))]))
		  | Am_rel(Am_rel_general(_))
			  when inst_what = Inst_everything ->
		      process_stmt inst_what id_env acc_env'''
			  (parse_stmt "{
			      var $h : type_ann <<\"handle\" ($acc)>>;
			      var $i : int;
			      call $handle_init($h);
			      while (%op($handle_valid)($h)) {
				  $i := %op($handle_deref)($h);
				  $body;
				  call $handle_incr($h);
			      }
			      call $handle_close($h);
			  }" [("$h",An_string(Misc.gensym "h"));
			      ("$acc",An_expr(
				  Expr_access(Access_strong,r',name,args')));
			      ("$i",An_string(index));
			      ("$body",An_stmt(body''));
			      ("$handle_init",An_string(handle_init));
			      ("$handle_incr",An_string(handle_incr));
			      ("$handle_close",An_string(handle_close));
			      ("$handle_valid",An_string(handle_valid));
			      ("$handle_deref",An_string(handle_deref))
			      ])
		  | Am_rel(Am_rel_general(_)) ->
		      Stmt_do(Expr_set(Frame([
			      Ident_valid(index,Type_scalar(sc_ty),Ident_bind,
				  An_list([]))]),
			      Expr_op(Op_other("in"),[
				  Expr_var(index);
				  Expr_access(Access_strong,r',name,args');
				  bounds_expr'])),
			  body')
		  | _ -> assert false (* Can't happen 18 *)
	  in
	      continuation_chains ()

    (* Lookup Access Conditionals *)
    | Stmt_if(Expr_access(Access_lookup,r,name,args),thn,els) ->
	let rec continuation_chain () =
	    process_expr inst_what id_env acc_env r
		(fun acc_env' r' ->
		    process_exprs inst_what id_env acc_env' args
			(fun acc_env'' args' ->
			    build_result r' args' acc_env''))
	and build_result r' args' acc_env'' =
 	    if inst_what = Inst_just_dense then
		Stmt_if(Expr_access(Access_lookup,r',name,args'),
		    process_stmt inst_what id_env acc_env'' thn,
		    process_stmt inst_what id_env acc_env'' els)
	    else
		(* Instantiate lookup access conditionals *)
		(let e' = Expr_access(Access_lookup,r',name,args')
		 in let bbi = get_bbi_from_rel_expr r' id_env
		 in
		     instantiate_lookup_access inst_what bbi acc_env'' e'
			 (fun acc_env''' _ ->
			    process_stmt inst_what id_env acc_env''' thn)
			 (fun acc_env''' _ ->
			    process_stmt inst_what id_env acc_env''' els))
	in
	    continuation_chain ()
    | Stmt_call(Expr_var(f), res, args) when f = handle_init ->
	  assert (res = Expr_void)
	      (* (Failure("Can't happen 19")) *);
	  let h = deref_handle_args args id_env
	  in
	      h.init_st
    | Stmt_call(Expr_var(f), res, args) when f = handle_incr ->
	  assert (res = Expr_void)
	      (* (Failure("Can't happen 20")) *);
	  let h = deref_handle_args args id_env
	  in
	      h.incr_st
    | Stmt_call(Expr_var(f), res, args) when f = handle_close ->
	  assert (res = Expr_void)
	      (* (Failure("Can't happen 21")) *);
	  let h = deref_handle_args args id_env
	  in
	      h.close_st
    (* Unscheduled Joins *)
    | _ when stmt_is_join s ->
	let { join_t_var = t_var; 
	      join_bounds = (lbs,ubs);
	      join_terms = terms; 
	      join_pred = pred;
	      join_assertion = assertion;
	      join_body = body } =
	    stmt2join s
	in let rec continuation_chain () =
	    process_exprs inst_what id_env acc_env lbs
		(fun acc_env' lbs' ->
		    process_exprs inst_what id_env acc_env' ubs
			(fun acc_env'' ubs' ->
			    process_terms inst_what id_env acc_env'' terms
				(fun acc_env''' terms' ->
				    build_result lbs' ubs' terms' acc_env''')))
	and build_result lbs' ubs' terms' acc_env''' =
	    let body' =
		process_stmt inst_what id_env acc_env''' body
	    in let join' = 
	      {	join_t_var = t_var;
		join_bounds = (lbs',ubs');
		join_terms = terms';
	      	join_pred = pred;
	      	join_assertion = assertion;
		join_body = body'
	      } 
	    in
	    join2stmt join'
	in 
	    continuation_chain ()
    (* Unscheduled Determined *)
    | _ when stmt_is_determined s ->
	  let { determined_term = term; 
		determined_pred = pred;
		determined_assertion = assertion;
		determined_body = body } =
	      stmt2determined s
	  in
	      process_term inst_what id_env acc_env term
		  (fun acc_env' term' ->
		      let body' =
			  process_stmt inst_what id_env acc_env' body
		      in let determined' = 
			{ determined_term = term';
			  determined_pred = pred;
			  determined_assertion = assertion;
			  determined_body = body'
		      } in
			  determined2stmt determined')
    | Stmt_assign(lhs,rhs) ->
	process_expr inst_what id_env acc_env lhs
	    (fun acc_env' lhs' ->
		process_expr inst_what id_env acc_env' rhs
		    (fun acc_env'' rhs' ->
			Stmt_assign(lhs',rhs')))

    | Stmt_seq(first,second) ->
	Stmt_seq(process_stmt inst_what id_env acc_env first,
	    process_stmt inst_what id_env acc_env second)
    | Stmt_call(f,lhs,args) ->
	process_expr inst_what id_env acc_env f
	    (fun acc_env' f' ->
		process_expr inst_what id_env acc_env' lhs
		    (fun acc_env'' lhs' ->
			process_exprs inst_what id_env acc_env'' args
			    (fun acc_env''' args' ->
				Stmt_call(f',lhs',args'))))
    | Stmt_for(i,lb,ub,step,body) ->
	process_expr inst_what id_env acc_env lb
	    (fun acc_env' lb' ->
		process_expr inst_what id_env acc_env' ub
		    (fun acc_env'' ub' ->
			process_expr inst_what id_env acc_env'' step
			    (fun acc_env''' step' ->
				Stmt_for(i,lb',ub',step',
				    process_stmt inst_what id_env acc_env'''
					body))))
    | Stmt_while(pred,body) ->
	process_expr inst_what id_env acc_env pred
	    (fun acc_env' pred' ->
		Stmt_while(pred', process_stmt inst_what id_env acc_env' body))
    | Stmt_if(pred,thn,els) ->
	process_expr inst_what id_env acc_env pred
	    (fun acc_env' pred' ->
		Stmt_if(pred',
		    process_stmt inst_what id_env acc_env' thn,
		    process_stmt inst_what id_env acc_env' els))
    | Stmt_let(frm,body) ->
	process_let_frame inst_what id_env acc_env frm
	    (fun id_env' acc_env' frm' ->
		Stmt_let(frm',
		    process_stmt inst_what id_env' acc_env' body))
    | Stmt_do(Expr_set(frm,pred),body) ->
	process_expr inst_what (process_set_frame inst_what id_env acc_env frm) acc_env pred
	    (fun acc_env' pred' ->
		Stmt_do(Expr_set(frm,pred'),
		    process_stmt inst_what id_env acc_env' body))
    | Stmt_do(_,_) -> assert false (* Shouldn't happen 6 *)
    | Stmt_nop -> Stmt_nop
    | Stmt_directive(ann,body) ->
	let body' = process_stmt inst_what id_env acc_env body
	in
	    if inst_what = Inst_just_dense then
		Stmt_directive(ann,body')
	    else
		(* Strip the directive *)
		body'
) : bmf_stmt)

(* \section{Processing Frames}                                              *)

(* \subsection{Processing Let [[bmf_frame]]'s}                              *)

and process_let_frame inst_what (orig_id_env: id_env_type)
	(orig_acc_env: acc_env_type) (Frame(decls))
	(k: (id_env_type -> acc_env_type -> bmf_frame -> bmf_stmt)) = 
  ((let rec recurs id_env_f  acc_env processed_decls unprocessed_decls =
    match unprocessed_decls with
      [] -> k (id_env_f orig_id_env) acc_env (Frame(processed_decls))
    | decl::unprocessed_decls' ->
	let k' id_env_f' acc_env' decls' =
	  recurs (fun ie -> id_env_f (id_env_f' ie)) acc_env'
	    (processed_decls@decls') unprocessed_decls'
	in
	match decl with
	(* BB array decls *)
	  Ident_valid(name,(Type_array(Array_bb(_,_,_,_)) as typ),
		      Ident_var(init),ann) 
	  ->
	    if (init <> Expr_void) then
	      failwith "Instantiator: BB decls cannot be inited.";
	    process_ann inst_what orig_id_env acc_env ann
	      (fun acc_env' ann' ->
		let decl' =
		  Ident_valid(name,typ,Ident_var(Expr_void),ann')
		in let bbi = black_box_of_bb_decl decl'
		in let id_env_f ie =
		  add_to_id_env name (Id_relvar(bbi)) ie
		in
		if inst_what = Inst_just_dense then
		  k' id_env_f acc_env' [decl']
		else
		  k' id_env_f acc_env' [])
	(* Dense array decls *)
	| Ident_valid(name,Type_array(Array_dense(lb,ubs,maj,sc_ty)),
		      Ident_var(init),ann) 
	  ->
	    process_exprs inst_what orig_id_env acc_env ubs
	      (fun acc_env' ubs' ->
		process_expr inst_what orig_id_env acc_env' init
		  (fun acc_env'' init' ->
		    process_ann inst_what orig_id_env acc_env'' ann
		      (fun acc_env''' ann' ->
			let decl' =
			  Ident_valid
			    (name,
			     Type_array(Array_dense(lb,ubs',maj,sc_ty)),
			     Ident_var(init'),ann')
			in let bbi = black_box_of_dense_decl(decl')
			in let id_env_f ie =
			  add_to_id_env name
			    (Id_prim_array(bbi)) ie
			in
			k' id_env_f acc_env''' [decl'])))
	(* Handle decls *)
	| Ident_valid
	    (h,
	     Type_ann(An_list
			([
			 An_string("handle");
			 An_expr(Expr_access
				   (Access_strong,r,name,args))])),
	     Ident_var(Expr_void),An_list([]))
	  ->
	    process_expr inst_what orig_id_env acc_env r
	      (fun acc_env' r' ->
		process_exprs inst_what orig_id_env acc_env' args
		  (fun acc_env'' args' ->
		    let bbi =
		      get_bbi_from_rel_expr r'
			orig_id_env
		    in let { 
		      am_info = mthd; 
		      am_out_field = out_field 
		    } = get_method name bbi
		    in let stream_f =
		      match mthd with
			Am_rel(Am_rel_interval(range_f)) ->
			  interval_stream_f range_f
			    r' out_field
		      | Am_rel(Am_rel_general(stream_f)) ->
			  stream_f
		      | _ -> assert false (* Can't happen 12 *)
		    in let decl' =
		      Ident_valid
			(h,
			 Type_ann(An_list
				    ([
				     An_string("handle");
				     An_expr(Expr_access(
					     Access_strong,r',
					     name,args'))])),
			 Ident_var(Expr_void),An_list([]))
		    in let r0 = stream_f args'
		    in let id_env_f ie =
		      add_to_id_env h (Id_handle(r0)) ie
		    in
		    if inst_what = Inst_just_dense then
		      k' id_env_f acc_env [decl']
		    else
		      k' id_env_f acc_env'' r0.decl_ids))
        (* Scalar decls *)
	| Ident_valid(name,Type_scalar(sc_ty),Ident_var(init),ann) ->
	    process_expr inst_what orig_id_env acc_env init
	      (fun acc_env' init' ->
		process_ann inst_what orig_id_env acc_env' ann
		  (fun acc_env'' ann' ->
		    let decl' =
		      Ident_valid(name,Type_scalar(sc_ty),
				  Ident_var(init'),ann')
		    in
		    k' identity acc_env'' [decl']))
	| Ident_valid(name,Type_scalar(sc_ty),Ident_bind,ann) ->
	    process_ann inst_what orig_id_env acc_env ann
	      (fun acc_env' ann' ->
		let decl' =
		  Ident_valid(name,Type_scalar(sc_ty),Ident_bind,ann')
		in
		k' identity acc_env [decl'])
	| _ -> assert false (* What kind of simple decl is this? *)
  in
  recurs identity orig_acc_env [] decls
) : bmf_stmt)

(* \subsection{Processing Global [[bmf_frame]]'s}                           *)
and process_global_frame inst_what (Frame(decls)) = 
  ((
   let rec recurs id_env acc_env unprocessed_decls =
     assert (acc_env == empty_acc_env)
	    (* (Failure("Why isn't acc_env empty?")) *);
     match unprocessed_decls with
       [] -> []
     | decl::unprocessed_decls' ->
	 let k' id_env' acc_env' decls' =
	   decls'@(recurs id_env' acc_env' unprocessed_decls')
	 in
	 match decl with
	   Ident_valid(name,(Type_array(Array_bb(_,_,_,_)) as typ),
		       Ident_var(Expr_void),ann) 
	   ->
	     let ann' = dont_process_ann inst_what id_env acc_env ann
	     in let decl' =
	       Ident_valid(name,typ,Ident_var(Expr_void),ann')
	     in let bbi = black_box_of_bb_decl decl'
	     in let id_env' =
	       add_to_id_env name (Id_relvar(bbi)) id_env
	     in
	     if inst_what = Inst_just_dense then
	       k' id_env' acc_env [decl']
	     else
	       k' id_env' acc_env []
	 | Ident_valid(name,Type_array(Array_dense(lb,ubs,maj,sc_ty)),
		       Ident_var(init),ann) 
	   ->
	     let ubs' = dont_process_exprs inst_what id_env empty_acc_env ubs
	     and init' = dont_process_expr inst_what id_env empty_acc_env init
	     and ann' = dont_process_ann inst_what id_env empty_acc_env ann
	     in let decl' =
	       Ident_valid(name,Type_array(Array_dense(lb,ubs',maj,sc_ty)),
			   Ident_var(init'),ann')
	     in let bbi = black_box_of_dense_decl decl'
	     in let id_env' =
	       add_to_id_env name (Id_prim_array(bbi)) id_env
	     in
	     k' id_env' empty_acc_env [decl']
	 | Ident_valid(name,(Type_proc(_,_) as typ),
		       Ident_proc(args_frame,res_decl,body),
		       ann)
	   ->
	     let id_env' = process_param_frame inst_what id_env empty_acc_env
		 res_decl args_frame
	     in let body' = process_stmt inst_what id_env' empty_acc_env body
	     in let ann' = dont_process_ann inst_what id_env empty_acc_env ann
	     in let decl' =
	       Ident_valid(name,typ,Ident_proc(args_frame,res_decl,body'),ann')
	     in
	     k' id_env acc_env [decl']
	 | Ident_valid(name,Type_scalar(sc_ty),Ident_var(init),ann) ->
	     let init' = dont_process_expr inst_what id_env empty_acc_env init
	     and ann' = dont_process_ann inst_what id_env empty_acc_env ann
	     in let decl' =
	       Ident_valid(name,Type_scalar(sc_ty),Ident_var(init'),ann')
	     in
	     k' id_env empty_acc_env [decl']
	 | Ident_valid(name,Type_scalar(sc_ty),Ident_bind,ann) ->
	     let ann' = dont_process_ann inst_what id_env empty_acc_env ann
	     in let decl' =
	       Ident_valid(name,Type_scalar(sc_ty),Ident_bind,ann')
	     in
	     k' id_env acc_env [decl']
	 | _ -> assert false (* What kind of simple decl is this? *)
   in
   Frame(recurs empty_id_env empty_acc_env decls)
) : bmf_frame)

and process_param_frame inst_what (orig_id_env: id_env_type)
	(orig_acc_env: acc_env_type) (res_decl: bmf_ident)
	(Frame(args_decls)) = 
  ((
   let rec recurs id_env_f acc_env processed_decls unprocessed_decls =
     match unprocessed_decls with
       [] ->
	 assert (orig_acc_env = acc_env)
		(* (Failure("process_param_frame -- "^
		   "Why did the acc_env change?")) *);
	 (id_env_f,processed_decls)
     | decl::unprocessed_decls' ->
	 process_decl id_env_f processed_decls decl unprocessed_decls'
   and process_decl id_env_f processed_decls decl unprocessed_decls' =
     let k' id_env_f' acc_env' decls' =
       recurs (fun ie -> id_env_f (id_env_f' ie)) acc_env'
	 (processed_decls@decls') unprocessed_decls'
     in
     match decl with
       Ident_valid(name,Type_array(Array_dense(lb,ubs,maj,sc_ty)),
		   Ident_var(init),ann) 
       ->
	 (let ubs' = dont_process_exprs inst_what orig_id_env orig_acc_env ubs
	 and init' = dont_process_expr inst_what orig_id_env orig_acc_env init
	 and ann' = dont_process_ann inst_what orig_id_env orig_acc_env ann
	 in let decl' =
	   Ident_valid(name,Type_array(Array_dense(lb,ubs',maj,sc_ty)),
		       Ident_var(init'),ann)
	 in let bbi = black_box_of_dense_decl decl'
	 in let id_env_f ie =
	   add_to_id_env name (Id_prim_array(bbi)) ie
	 in
	 k' id_env_f orig_acc_env [decl'])
     | Ident_valid(name,Type_scalar(sc_ty),Ident_var(init),ann) ->
	 let init' = dont_process_expr inst_what orig_id_env orig_acc_env init
	 and ann' = dont_process_ann inst_what orig_id_env orig_acc_env ann
	 in let decl' =
	   Ident_valid(name,Type_scalar(sc_ty),Ident_var(init'),ann')
	 in
	 k' identity orig_acc_env [decl']
     | Ident_valid(name,Type_scalar(sc_ty),Ident_bind,ann) ->
	 let ann' = dont_process_ann inst_what orig_id_env orig_acc_env ann
	 in let decl's =
	   Ident_valid(name,Type_scalar(sc_ty),Ident_bind,ann')
	 in
	 k' identity orig_acc_env [decl]
     | _ -> assert false (* What kind of simple decl is this? *)
   in let (res_id_env_f,res_decls') =
     if res_decl <> Ident_null then
       recurs identity orig_acc_env [] [res_decl]
     else
       (identity,[res_decl])
   in let (args_id_env_f,args_decls') =
     recurs res_id_env_f orig_acc_env [] args_decls
   in
   (* *todo* We really should check this.
   assert ([res_decl] = res_decls' && args_decls = args_decls')
   (* (Failure("process_param_frame -- Why did the decls change?")) *);
   *)
   args_id_env_f orig_id_env
) : id_env_type)

(* \subsection{Processing Set and Param [[bmf_frame]]'s}                    *)
and process_set_frame inst_what (orig_id_env: id_env_type)
	(orig_acc_env: acc_env_type) (Frame(decls)) = 
  ((
   let rec recurs id_env_f acc_env processed_decls unprocessed_decls =
     match unprocessed_decls with
       [] ->
	 assert (orig_acc_env = acc_env)
	   (* (Failure("process_param_frame -- "^
	      "Why did the acc_env change?")) *);
	 (id_env_f,processed_decls)
     | decl::unprocessed_decls' ->
	 process_decl id_env_f processed_decls decl unprocessed_decls'
   and process_decl id_env_f processed_decls decl unprocessed_decls' =
     let k' id_env_f' acc_env' decls' =
       recurs (fun ie -> id_env_f (id_env_f' ie)) acc_env'
	 (processed_decls@decls') unprocessed_decls'
     in
     match decl with
       Ident_valid(name,Type_scalar(sc_ty),Ident_var(init),ann) ->
	 let init' = dont_process_expr inst_what orig_id_env orig_acc_env init
	 and ann' = dont_process_ann inst_what orig_id_env orig_acc_env ann
	 in let decl' =
	   Ident_valid(name,Type_scalar(sc_ty),Ident_var(init'),ann')
	 in
	 k' identity orig_acc_env [decl']
     | Ident_valid(name,Type_scalar(sc_ty),Ident_bind,ann) ->
	 let ann' = dont_process_ann inst_what orig_id_env orig_acc_env ann
	 in let decl's =
	   Ident_valid(name,Type_scalar(sc_ty),Ident_bind,ann')
	 in
	 k' identity orig_acc_env [decl]
     | _ -> assert false (* What kind of simple decl is this? *)
   in let (id_env_f',decls') =
     recurs identity orig_acc_env [] decls
   in
   (* *todo* We really should check this.
   assert (decls = decls')
   (* (Failure("process_set_frame -- Why did the decls change?")) *);
   *)
   id_env_f' orig_id_env
) : id_env_type)

(************************************************************************
``Don't process'' versions
************************************************************************)

(* A hack. A {\em really} bad hack.                                         *)
and dont_process_ann inst_what id_env acc_env ann =
    let result = ref (An_list([]))
    and set_p = ref false
    in let s =
	process_ann inst_what id_env acc_env ann
	    (fun acc_env' ann' ->
		assert (acc_env == acc_env')
		    (* (Failure("Instantiated obj is wrong context")) *);
		result := ann';
		set_p := true;
		Stmt_nop)
    in
	assert (s = Stmt_nop)
	    (* (Failure("Can't happen 22")) *);
	assert (!set_p)
	    (* (Failure("Can't happen 23")) *);
	!result

(* Another {\em really} bad hack.                                           *)
and dont_process_expr inst_what id_env acc_env e =
    let result = ref Expr_void
    and set_p = ref false
    in let s =
	process_expr inst_what id_env acc_env e
	    (fun acc_env' e' ->
		assert (acc_env == acc_env')
		    (* (Failure("Instantiated obj is wrong context")) *);
		result := e';
		set_p := true;
		Stmt_nop)
    in
	assert (s = Stmt_nop)
	    (* (Failure("Can't happen 24")) *);
	assert (!set_p)
	    (* (Failure("Can't happen 25")) *);
	!result

(* This one isn't so bad.                                                   *)
and dont_process_exprs inst_what id_env acc_env es =
    List.map (dont_process_expr inst_what id_env acc_env) es

and deref_handle_args args id_env =
    let h =
	match args with
	  [h] -> h
	| _ -> assert false (* Can't happen 8 *)
    in
	match (id_of_rel_var_expr h id_env) with
	  (Id_prim_array(_)|Id_relvar(_)) ->
	    assert false (* Shouldn't happen 3 *)
	| Id_handle(r) -> r

and get_bbi_from_rel_expr r id_env =
    match (id_of_rel_var_expr r id_env) with
      Id_handle(_) ->
	  assert false (* Shouldn't happen 1 *)
    | Id_prim_array(bbi) -> bbi
    | Id_relvar(bbi) -> bbi

and interval_stream_f range_f r out_field args =
    let r_name =
        match r with
          Expr_var(r_name) -> r_name
        | _ -> assert false (* Expected a expr_var *)
    in let index = Misc.gensym (r_name ^ out_field)
    in let ub = Misc.gensym ("ub_" ^index)
    in let e_index = Expr_var(index)
    and e_ub = Expr_var(ub)
    and make_decl name =
	Ident_valid(name,Type_scalar(Type_int),
	    Ident_var(Expr_void),An_list([]))
    in
	{ decl_ids = [make_decl index; make_decl ub];
	  init_st=
	    range_f args
		(fun e_lb_val e_ub_val ->
		    Stmt_seq(Stmt_assign(e_index,e_lb_val),
			     Stmt_assign(e_ub,e_ub_val)));
	  incr_st=
	    Stmt_assign(e_index,
		make_plus [e_index; Expr_int(1)]);
	  close_st= Stmt_nop;
	  valid_ex= Expr_op(Op_leq,[e_index;e_ub]);
	  deref_ex= e_index
	  }

(************************************************************************)
(*			  Command line processing			*)
(************************************************************************)

let _ =
  register_switch "-inst"
    "-inst (bmf) -- Instantiation."
    (fun top_decls args ->
      let name = Bsct_toplevel.get_arg_name (List.hd args)
      and (options,param,args') = Bsct_toplevel.parse_arg false args
      in let (just_dense_p,options) = Bsct_toplevel.extract_option options 'd'
      in let _ = Bsct_toplevel.ignore_options options name
      in let top_decls' =
	instantiate
	  top_decls
	  (if just_dense_p then Inst_just_dense 
	  else Inst_everything)
      in (top_decls',args'))
