(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Bsct_base
open Ssisl
open Containers

open Bmf
open Bmf_join
open Bmf_parser_aux
open Bb
open Bb_decl

type sched_what_type = Sched_just_dense | Sched_everything;;
(* 0 = search and nested join. 1 = 0 + scatter/gather & two-finger *)
type sched_level_type = int;;

type scatter_gather_type = {
    (* Temporary array for scatter *)
    sg_id : bmf_ident;
    (* Initialization code for scatter *)
    sg_init : bmf_stmt;
    (* Scatter code *)
    sg_scatter : bmf_stmt;
    (* Gather code *)
    sg_gather : bmf_stmt;
    (* Tested body of join *)
    sg_body : bmf_stmt
  }	

exception Cant_use_2_finger
exception Cant_use_scatter_gather

(************************************************************************
Entry points and determined/join finder

Code to find determineds and joins, and invoke the appropriate scheduler.
************************************************************************)

let rec schedule_procs sched_what sched_level (decls: bmf_ident list) = (
  List.rev (List.map (schedule_proc sched_what sched_level)
	      (List.rev decls))
    : bmf_ident list)

and schedule_proc sched_what sched_level (decl: bmf_ident) = (
  match decl with
    Ident_valid(name,typ,Ident_proc(Frame(args),res,body),ann) ->
      let body' = 
	walk_stmt sched_what sched_level (add_array_decls (res::args)
				String_map.empty)
	  body
      in 
      let body'' = Bmf_util.perform_substitutions_in_stmt body' in
      Ident_valid(name,typ,Ident_proc(Frame(args),res,body''),ann)
  | _ -> decl
	: bmf_ident)

and walk_stmt sched_what sched_level bbm s = (
  let r = walk_stmt sched_what sched_level bbm
  in
  match s with
    _ when stmt_is_determined s ->
      schedule_determined sched_what sched_level bbm
	(stmt2determined s)
  | _ when stmt_is_join s ->
      schedule_join sched_what sched_level bbm (stmt2join s)
  | (Stmt_assign _ | Stmt_call _ | Stmt_nop ) -> s
  | Stmt_seq(first,second) -> Stmt_seq(r first, r second)
  | Stmt_for(i,lb,ub,step,body) -> Stmt_for(i, lb, ub, step, r body)
  | Stmt_while(pred,body) -> Stmt_while(pred, r body)
  | Stmt_if(pred,thn,els) -> Stmt_if(pred, r thn, r els)
  | Stmt_let(Frame(decls),body) ->
      let bbm' = add_array_decls decls bbm in
      Stmt_let(Frame(decls), walk_stmt sched_what sched_level bbm' body)
  | Stmt_do(pred,body) -> Stmt_do(pred, r body)
  | Stmt_directive(ann,s') -> Stmt_directive(ann, r s')
: bmf_stmt)

(* *****************************************************************
   Join selection heuristics
   ***************************************************************** *)

and schedule_determined sched_what sched_level (bbm: black_box_map)
    (determined: determined_type) = (
  let { determined_term = term;
	determined_pred = pred;
	determined_assertion = assertion;
	determined_body = body } = determined
  in let { jt_kind = kind;
	   jt_flag = term_flag;
	   jt_constr = constr;
	   jt_bounds = bounds } = term
  in let enum_var = enum_var_of_kind kind
  and join_var = join_var_of_kind kind
  in let dont_schedule_determined () =
    let body' = walk_stmt sched_what sched_level bbm body
    in let determined' = { determined_term = { 
			   jt_kind = kind;
			   jt_flag = term_flag;
			   jt_constr = constr;
			   jt_bounds = bounds };
			   determined_pred = pred;
			   determined_assertion = assertion;
			   determined_body = body'}
    in
    determined2stmt determined'
  in
  match kind with
    Jtk_enum_simple(Jtk_eq(_,rel,enum_s,enum_args)) ->
      let bbi = String_map.find rel bbm
      in let { am_info = am } = find_method enum_s bbi	
      in let final_body =
	begin
	  match am with
	  | Am_rel(Am_rel_interval(_)) ->
	    (* We don't generate inclusion tests for
	       [[Am_rel_interval]]'s because we {\em assume} that
	       the $t$ bounds check generated by  [[scheduler.nw]]
	       is sufficient. *)
	    (* *todo*: This reasoning is a bit spurious. Rethink
	       and reimplement. *)
	      Bmf_util.place_substitutions 
		[(enum_var,constr);(term_flag,Expr_bool(true))] 
		body
	  | _ when sched_what = Sched_just_dense ->
	      dont_schedule_determined ()
	  | Am_rel(Am_rel_general(_)) ->
	      Bmf_util.place_substitutions [(enum_var,constr)]
		(parse_stmt
		   "if (%op(in)($enum_var, $R[[$enum_s,$enum_args]])) {
		      $body;
		    }"
		   [("$R",An_string(rel));
		     ("$enum_var",An_string(enum_var));
		     ("$enum_s",An_expr(Expr_string(enum_s)));
		     ("$enum_args",An_list(List.map (fun x -> An_expr(x)) 
					     enum_args));
		     ("$body",
		      An_stmt
			(Bmf_util.place_substitutions
			   [(term_flag,Expr_bool(true))]
			   body))
		   ])
	  | Am_fun(_,_) -> failwith "Can't happen"
      	end
      in walk_stmt sched_what sched_level bbm final_body
	
  (* Nest and test. *)
  | Jtk_enum_indexing(Jtk_eq(_,rel,enum_s,enum_args),
		      Jtk_eq(_,_,lookup_s,lookup_args))
      when sched_what = Sched_everything ->
	let final_body =
	  Bmf_util.place_substitutions 
	    [(join_var,constr)]
	    (parse_stmt
	      "do {[$enum_var]| %op(in)($enum_var,$R[[$enum_s,$enum_args]])} {
		 if ($join_var == $R[[$lookup_s,$lookup_args]]) {
		   $body;
		 }
	       }"
	      [("$enum_var",An_string(enum_var));
		("$join_var",An_string(join_var));
		("$R",An_string(rel));
		("$enum_s",An_expr(Expr_string(enum_s)));
		("$enum_args",An_list(List.map (fun x -> An_expr(x)) 
					enum_args));
		("$lookup_s",An_expr(Expr_string(lookup_s)));
		("$lookup_args",An_list(List.map (fun x -> An_expr(x)) 
					  lookup_args));
		("$body",
		 An_stmt
		   (Bmf_util.place_substitutions
		      [(term_flag,Expr_bool(true))]
		      body))])
	in
	walk_stmt sched_what sched_level bbm final_body
	  
  (* Generate a ``determined'' access. *)
  | Jtk_enum_searching(Jtk_eq(_,_,enum_s,enum_args),
		       Jtk_eq(_,_,lookup_s,lookup_args),
		       Jtk_eq(_,rel,determined_s,determined_args))
      when sched_what = Sched_everything ->
	let parse_args =
	  [("$R",An_string(rel));
	    ("$determined_s",An_expr(Expr_string(determined_s)));
	    ("$determined_args",An_list(List.map (fun x -> An_expr(x)) 
					  determined_args))]
	in let final_body =
	  Bmf_util.place_substitutions
	    [(join_var,constr)]
	    (Bmf_util.place_substitutions
	       [(enum_var,
		 (parse_expr "$R[[$determined_s,$determined_args]]"
		    parse_args))]
	       (parse_stmt
		  "if ($R($determined_s,$determined_args)) {
	             $body;
	           }"
		  (parse_args
		   @[("$body",
		      An_stmt
		     	(Bmf_util.place_substitutions
		  	   [(term_flag,Expr_bool(true))]
		  	   body))])))
	in
	walk_stmt sched_what sched_level bbm final_body
					
  (* Schedule a [[Jtk_lookup]] determined *)
  | Jtk_lookup(Jtk_eq(_,rel,lookup_s, lookup_args)) ->
      let parse_args =
	[("$R",An_string(rel));
	  ("$lookup_s",An_expr(Expr_string(lookup_s)));
	  ("$lookup_args",An_list(List.map (fun x -> An_expr(x)) 
				    lookup_args))]
      in let final_body =
	Bmf_util.place_substitutions
	  [(join_var,constr)]
	  (Bmf_util.place_substitutions
	     [(enum_var,
	       (parse_expr "$R[[$lookup_s,$lookup_args]]"
		  parse_args))]
	     (parse_stmt
		"if ($R($lookup_s,$lookup_args)) {
	           $body;
	         }"
		(parse_args
		   @[("$body",
		      An_stmt
		     	(Bmf_util.place_substitutions
		  	   [(term_flag,Expr_bool(true))]
		  	   body))])))
      in
      walk_stmt sched_what sched_level bbm final_body
  | _ -> dont_schedule_determined ()
	: bmf_stmt)

(*--------------------------------------------------------------------------
  In order to schedule a join, we first divide its terms into three       
  categories,                                                             
  \begin{description}                                                     
  \item[ [[lookup_terms]] ] There are [[Jtk_lookup]] join terms.            
  \item[ [[trivial_terms]] ] These a [[Jtk_enum_simple]] terms in which 
  the access
  method is a [[Am_rel_interval]]. In other words, these join terms are  
  simple loops from $lb$ to $ub$.                                        
  \item[ [[non_trivial_terms]] ] Those terms that don't fit into the above 
  categories.                                                            
  \end{description}                                                        
  ------------------------------------------------------------------------*)
and schedule_join sched_what sched_level (bbm: black_box_map)
    (join: join_type) = (
  let { join_t_var = t_var; join_terms = terms;
	join_pred = pred; join_assertion = assertion;
	join_body = body; join_bounds = t_bounds } = join
  in 

  (* Seperate trivial, non-trivial, and lookup join terms *)
  let (non_trivial_terms,trivial_terms,lookup_terms) =
    split_join_terms bbm terms
  in
  match (non_trivial_terms,trivial_terms,lookup_terms) with
    ([],[],[]) -> failwith "Can't happen"
	
  (* This case arises when a joinable lookup is scheduled by the*)
  (* query scheduler.*)
  | (_,_,lookup_term::lookup_terms') ->
      walk_stmt sched_what sched_level bbm
      	(gen_join_nested_loop join
      	   lookup_term (non_trivial_terms @ trivial_terms) body)
	
  (* A list of trivial terms are a list of simple loops. Try to*)
  (* scheduler the ``tightest'' loop outermost. Schedule the rest of*)
  (* the loops as determineds. *)
  | ([],_,[]) ->
      let ordered_terms =
	Sort.list
	  (fun {jt_bounds = (lbs1,ubs1)}
	      {jt_bounds = (lbs2,ubs2)} ->
		let num_lbs1 = List.length lbs1
		and num_ubs1 = List.length ubs1
		and num_lbs2 = List.length lbs2
		and num_ubs2 = List.length ubs2
		in let max1 = max num_lbs1 num_ubs1
		and max2 = max num_lbs2 num_ubs2
		and min1 = min num_lbs1 num_ubs1
		and min2 = min num_lbs2 num_ubs2
		in
		if max1 <> max2 then
		  max1 < max2
		else if min1 <> min2 then
		  min1 < min2
		else
		  false)
	  trivial_terms
      in
      walk_stmt sched_what sched_level bbm
      	(gen_join_nested_loop join
	   (List.hd ordered_terms) (List.tl ordered_terms) body)

  (* If there is a single non-trivial term and any number of trivial*)
  (* terms, then go ahead and schedule the join, even if we are only*)
  (* doing dense joins. After all, it doesn't matter what*)
  (* optimizations are done; this join is   still going to be*)
  (* scheduled in this manner.  *)
  | ([non_trivial_term],_,[]) ->
      walk_stmt sched_what sched_level bbm
	(gen_join_nested_loop join 
	   non_trivial_term trivial_terms body)

	(* If there is more than one non-trivial term, and we are only*)
	(* scheduling dense joins, then don't schedule the join completely.*)
	(* Instead, simply schedule the trivial terms, and leave the*)
	(* non-trivial terms unscheduled. *)
  | (_,_,[]) when sched_what = Sched_just_dense ->
      let body' = insert_determineds join trivial_terms body
      in let body'' = walk_stmt sched_what sched_level bbm body'
      in let join' = update_join join non_trivial_terms body'' in
      join2stmt join'
  (* Generate a non-trivial join. *)
  | (non_trivial_terms,trivial_terms,[]) -> 
      let body' = insert_determineds join trivial_terms body in
      let body'' = 
      	match non_trivial_terms with
	  [] -> assert false (* handled above. *)
      	| term::[] -> 
	    gen_join_nested_loop join term [] body
      	| hd_term::tl_terms -> 
	    (try
	      if sched_level = 0 then raise Cant_use_2_finger;
	      attempt_join_2_finger sched_what sched_level bbm join 
	      	non_trivial_terms body
	    with Cant_use_2_finger -> try
	      if sched_level = 0 then raise Cant_use_scatter_gather;
	      attempt_join_scatter sched_what sched_level bbm join 
	      	non_trivial_terms body
	    with Cant_use_scatter_gather ->
	      gen_join_nested_loop join hd_term tl_terms body)
      in
      walk_stmt sched_what sched_level bbm body''

: bmf_stmt)

(* *****************************************************************
   Try a two-finger join.
   ***************************************************************** *)

and attempt_join_2_finger sched_what sched_level bbm join terms body =
  let get_info term = 
    match term.jt_kind with
      Jtk_enum_indexing(Jtk_eq(_,rel,enum_s,_),
			Jtk_eq(_,_,lookup_s,_)) ->
			  (rel,enum_s,lookup_s)
    |	Jtk_enum_searching(Jtk_eq(_,rel,enum_s,_),
			     Jtk_eq(_,_,lookup_s,_),_) ->
			       (rel,enum_s,lookup_s)
    |	_ -> failwith "Can't happen"
  in
  let test_term term =
    match term.jt_kind with
      Jtk_enum_indexing(_,_) | Jtk_enum_searching(_,_,_) -> (
	let (rel,enum_s,lookup_s) = get_info term
	in
	let bbi = String_map.find rel bbm
	in
	let ord = bbi.bb_ordered_access
	and enum_m = find_method enum_s bbi
	in
	List.mem (enum_s,lookup_s) ord)
    | _ -> false
  in
  let rec
      find_pair terms rest pair =
    if (List.length pair) = 2 then
      (pair,terms@rest)
    else if (List.length pair) > 2 then
      failwith "Can't happen"
    else
      match terms with
	head::tail ->
	  if (test_term head) then
	    find_pair tail rest (head::pair)
	  else
	    find_pair tail (head::rest) pair
      |	[] -> ([],[])
  in
  if (List.length terms) < 2 then raise Cant_use_2_finger;
  match (find_pair terms [] []) with
    (t1::t2::[],rest) -> 
      walk_stmt sched_what sched_level bbm
	(gen_join_2_finger t1 t2 
	   (insert_determineds join rest body))
  | _ -> raise Cant_use_2_finger

(* *****************************************************************
   Try a scatter/gather join.
   ***************************************************************** *)

and attempt_join_scatter sched_what sched_level bbm join terms body =
  let test_term term =
    match term.jt_kind with
      Jtk_enum_indexing(_,_) | Jtk_enum_searching(_,_,_) ->
	true
    |	_ -> false
  in
  let rec
      find_term terms rest =
    match terms with
      head::tail ->
	if (test_term head) then 
	  head::(rest@tail)
	else
	  find_term tail (head::rest)
    |	[] -> []
  in
  if((List.length terms) < 2) then raise Cant_use_scatter_gather;
  match (find_term terms []) with
    (scatter_term::other_terms) ->
      let sg = 
	gen_join_scatter bbm join 
	  scatter_term body
      in let join' = update_join join other_terms sg.sg_body
      in let final_body =
	Stmt_let
	  (Frame([sg.sg_id]),
	   parse_stmt
	     ("{"^
	      "  $init;"^
	      "  $scatter;"^
	      "  $body;"^
	      "  $gather;"^
	      "}")
	     [("$init",An_stmt(sg.sg_init));
	       ("$scatter",An_stmt(sg.sg_scatter));
	       ("$body",An_stmt(join2stmt join'));
	       ("$gather",An_stmt(sg.sg_gather))])
      in
      walk_stmt sched_what sched_level bbm final_body 
  | [] -> raise Cant_use_scatter_gather

(* *****************************************************************
   Seperate trivial, non-trivial, and lookup join terms 
   ***************************************************************** *)

and split_join_terms bbm terms =
  let (non_trivial_terms,trivial_terms,lookup_terms) =
    List.fold_left
      (fun (non_trivial_terms,trivial_terms,lookup_terms) term ->
      	match term with
	  { jt_kind = Jtk_enum_simple(Jtk_eq(_,rel,enum_s,enum_args)) } ->
	    let bbi = String_map.find rel bbm
	    in let { am_info = am } = find_method enum_s bbi
	    in begin
	      match am with
	      	Am_rel(Am_rel_interval(_)) ->
		  (non_trivial_terms,term::trivial_terms,
		   lookup_terms)
	      | _ -> (term::non_trivial_terms,trivial_terms,
		      lookup_terms)
	    end
      	| { jt_kind = Jtk_lookup(_) } ->
	    (non_trivial_terms,trivial_terms,term::lookup_terms)
      	| _ -> (term::non_trivial_terms,trivial_terms,lookup_terms))
      ([],[],[]) terms
  in
  (* Put into a "canonical" order. This will help to normalized the 
     output. *)
  (Sort.list (<=) non_trivial_terms,
   Sort.list (<=) trivial_terms,
   Sort.list (<=) lookup_terms)		

(* *****************************************************************
   Generate a nested loop join 
   ***************************************************************** *)

(* loop_term - the term that is to be used for the loop.
   search_terms - the terms that are to be used for the tests and searches.
                  these terms are nested in the same order in which they
                  appear in this list. *)

and gen_join_nested_loop join
    loop_term search_terms body =
  gen_join_term_loop join loop_term
    (insert_determineds join search_terms body)

(* Generate the outer loop of a join nest. *)
and gen_join_term_loop join term body =
  (* Destruct the term to be used for the loop *)
  let { jt_kind = kind;
	jt_flag = term_flag;
	jt_constr = constr;
	jt_bounds = bounds } = term
  in let enum_var = enum_var_of_kind kind
  and join_var = join_var_of_kind kind
  and t_var = join.join_t_var
      
  (* Construct the body w/ substitutions *)
  in let t_val = 
    let (c,rest) = solve_for constr t_var
    in
    match c with
      1 -> make_plus (Expr_var(join_var)::
		      (List.map make_neg rest))
    | -1 -> make_plus (make_neg (Expr_var(join_var))::rest)
    | _ -> failwith "Non-unit coef for t"
  in let final_body =
    match kind with
      Jtk_enum_simple(Jtk_eq(_,rel,enum_s,enum_args)) ->
	let bounds_expr = bounds2expr (Expr_var(enum_var)) bounds in
	(* Construct the loop *)
	parse_stmt
      	  ("do {[$enum_var]|"^
	   "      %op(in)($enum_var,$R[[$enum_s,$enum_args]],$bounds)} {"^
	   "  $body;"^
	   "}")
      	  [("$enum_var",An_string(enum_var));
	    ("$R",An_string(rel));
	    ("$enum_s",An_expr(Expr_string(enum_s)));
	    ("$enum_args",An_list(List.map (fun x -> An_expr(x)) enum_args));
	    ("$bounds",An_expr(bounds_expr));
	    ("$body",
	     An_stmt
	       (Bmf_util.place_substitutions 
		  [(t_var,t_val);(term_flag,Expr_bool(true))]
		  body))]	  
    | (Jtk_enum_indexing _ | Jtk_enum_searching _) ->
	let (Jtk_eq(_,rel,enum_s,enum_args),
	     Jtk_eq(_,_,lookup_s,lookup_args)) =
	  match kind with
	    Jtk_enum_indexing(enum_eq,lookup_eq) -> (enum_eq,lookup_eq)
	  | Jtk_enum_searching(enum_eq,lookup_eq,_) -> (enum_eq,lookup_eq)
	  | _ -> assert false
	in
	let bounds_expr = bounds2expr (Expr_var(join_var)) bounds in
	(* Construct the loop *)
	parse_stmt
	  ("do {[$enum_var] |"^
	   "      %op(in)($enum_var,$R[[$enum_s,$enum_args]])} {"^
	   "  $body;"^
	   "}")
      	  [("$enum_var",An_string(enum_var));
	    ("$R",An_string(rel));
	    ("$enum_s",An_expr(Expr_string(enum_s)));
	    ("$enum_args",An_list(List.map (fun x -> An_expr(x)) enum_args));
	    ("$body",An_stmt(
	     Bmf_util.place_substitutions 
	       [(join_var,
		 Expr_access(Access_strong,Expr_var(rel),lookup_s,
			     lookup_args))]
	       (Bmf_util.place_substitutions [(t_var,t_val)]
		  (let body' =
		    Bmf_util.place_substitutions
		      [(term_flag,Expr_bool(true))]
		      body
		  in
		  if bounds_expr = Expr_bool(true) then
		    body'
		  else
		    (parse_stmt "if ($bounds) { $body; }")
      		      [("$bounds",An_expr(bounds_expr));
		       	("$body",An_stmt(body'))]))))]
    | Jtk_lookup(Jtk_eq(_,rel,lookup_s,lookup_args)) ->
	let bounds_expr = bounds2expr (Expr_var(join_var)) bounds in
	(* Construct the loop *)
	Bmf_util.place_substitutions 
	  [(join_var,
	    Expr_access(Access_strong,Expr_var(rel),lookup_s,lookup_args))] 
	  (Bmf_util.place_substitutions [(t_var,t_val)]	
      	     (let body' =
	       Bmf_util.place_substitutions
		 [(term_flag,Expr_bool(true))]
		 body
	     in
	     if bounds_expr = Expr_bool(true) then
	       body'
	     else
	       (parse_stmt "if ($bounds) { $body; }")
      	     	 [("$bounds",An_expr(bounds_expr));
		   ("$body",An_stmt(body'))]))
  in
  final_body
      
			
(* *****************************************************************
   2 finger joins
   ***************************************************************** *)

and gen_join_2_finger term1 term2 body =
  (* enum term1 normally, and finger with term2 *)
  let ev1 = enum_var_of_kind term1.jt_kind
  and ev2 = enum_var_of_kind term2.jt_kind
  and ee1 = enum_expr_of_kind term1.jt_kind
  and ee2 = enum_expr_of_kind term2.jt_kind
  and jv1 = join_var_of_kind term1.jt_kind
  and jv2 = join_var_of_kind term2.jt_kind
  and je1 = join_expr_of_kind term1.jt_kind
  and je2 = join_expr_of_kind term2.jt_kind
  in
  let h1 = String.concat "" ["h"; ev1]
  and h2 = String.concat "" ["h"; ev2]
  in 
  let final_body =
    parse_stmt 
      ("{"^
       "  var $h1 : type_ann <<\"handle\" ($ee1)>>;"^
       "  var $h2 : type_ann <<\"handle\" ($ee2)>>;"^
       "  $body;"^
       "}")
      [ ("$h1",An_string(h1));
	("$h2",An_string(h2));
	("$ee1",An_expr(ee1));
	("$ee2",An_expr(ee2));
	("$body",An_stmt(
	 Bmf_util.place_substitutions
	   [(ev1,(Expr_op(Op_other("%handle_deref"),[Expr_var(h1)])));
      	     (ev2,(Expr_op(Op_other("%handle_deref"),[Expr_var(h2)])))]
	   (Bmf_util.place_substitutions [(jv1,je1); (jv2,je2)]
	      (parse_stmt 
		 ("{"^
		  "  call %handle_init($h1);"^
		  "  call %handle_init($h2);"^
		  "  "^
		  "  while (%op(%handle_valid)($h1) && "^
		  "         %op(%handle_valid)($h2)) {"^
		  "    if ($jv1 < $jv2) {"^
		  "      call %handle_incr($h1);"^
		  "    }"^
		  "    else if ($jv1 > $jv2) {"^
		  "      call %handle_incr($h2);"^
		  "    }"^
		  "    else {"^
		  "      $body;"^
		  "      call %handle_incr($h1);"^
		  "      call %handle_incr($h2);"^
		  "    }"^
		  "  }"^
		  "  call %handle_close($h1);"^
		  "  call %handle_close($h2);"^
		  "}")
		 [ ("$h1",An_string(h1));
		   ("$h2",An_string(h2));
		   ("$jv1",An_string(jv1));
		   ("$jv2",An_string(jv2));
		   ("$body",
		    An_stmt
		      (Bmf_util.place_substitutions
			 [(term1.jt_flag,Expr_bool(true));
			   (term2.jt_flag,Expr_bool(true))]
			 body))]))
	   ))]
  in final_body

(* *****************************************************************
   Scatter/gather joins.
   ***************************************************************** *)

and gen_join_scatter bbm join term body = 
  let (enum_var,relation,lookup_s,lookup_args) =
    match term.jt_kind with
      Jtk_enum_indexing(Jtk_eq(var,rel,_,_),
			Jtk_eq(_,_,s,args)) -> (var,rel,s,args)
    |	Jtk_enum_searching(Jtk_eq(var,rel,_,_),
			     Jtk_eq(_,_,s,args),_) -> (var,rel,s,args)
    |	_ -> failwith "Can't happen"
  in
  let make_array_bounds relation =
    (* Should be using FM!!! *)
    let rec expr2bounds expr var =
      match expr with
	Expr_op(Op_and,elist) ->
	  let blist = List.map (fun x -> expr2bounds x var) elist
	  and concat_bounds (l1,u1) (l2,u2) = (l1@l2,u1@u2)
	  in
	  List.fold_left concat_bounds ([],[]) blist
      |	Expr_op(Op_leq,l::g::[]) ->
	  if(l = Expr_var(var)) then
	    ([],[g])
	  else if (g = Expr_var(var)) then
	    ([l],[])
	  else
	    ([],[])
      |	_ -> failwith "Replace with Fourier Motzkin!!!"
    in		
    let bbi = String_map.find relation bbm in
    let { bb_index_fields = index_fields;
	  bb_bounds_maker = bounds_maker
	} = bbi 
    and {am_out_field = field} = find_method lookup_s bbi
    in 
    let b_expr = bounds_maker 
	(List.map (fun x -> Expr_var(x)) index_fields)
    in 
    let (lbs,ubs) = expr2bounds b_expr field in
    (make_max lbs, make_min ubs)
  in
  let array = String.concat "" ["ta"; enum_var]
  and (lb,ub) = make_array_bounds relation
  in
  { sg_id = Ident_valid(array, 
			Type_array(Array_dense(Dense_lb_0,[ub],
					       Dense_row_major,Type_int)),
			Ident_var(Expr_void),
			An_list([]));

    sg_init = parse_stmt 
      "for $var := $lb : $ub { $ta[[$var]] := -1; }"
      [ ("$var",An_string(enum_var));
	("$ta",An_string(array));
	("$lb",An_expr(lb));
	("$ub",An_expr(ub))];

    sg_scatter = 
    (gen_join_nested_loop join 
       term []
       (parse_stmt 
	  "$ta[[$R[[$lookup_s,$lookup_args]]]] := $var;"
	  [ ("$var",An_string(enum_var));
	    ("$ta",An_string(array));
	    ("$R",An_string(relation));
	    ("$lookup_s",An_expr(Expr_string(lookup_s)));
	    ("$lookup_args",An_list(List.map (fun x -> An_expr(x)) 
				      lookup_args))]));

    sg_gather = Stmt_nop;

    sg_body = 
    (let { jt_kind = kind;
	   jt_flag = term_flag;
	   jt_constr = constr;
	   jt_bounds = bounds } = term
    in
    let enum_var = enum_var_of_kind kind
    and join_var = join_var_of_kind kind
    in 
    Bmf_util.place_substitutions [(join_var,constr)]
    (Bmf_util.place_substitutions 
       [(enum_var,
	 Expr_access(Access_strong,Expr_var(array),"",
		     [Expr_var(join_var)]))]
      (parse_stmt
	 ("{ if ($evar >= 0) { $body; } }")
	 [ ("$evar",An_string(enum_var));
	   ("$body",
	    An_stmt
	      (Bmf_util.place_substitutions
		 [(term_flag,Expr_bool(true))]
		 body))])))

  }

(* *****************************************************************
   Insert code for a list of unscheduled determineds. 
   ***************************************************************** *)

and insert_determineds join terms body =
  List.fold_right (insert_determined join)
    terms body

and insert_determined join term body =
  determined2stmt 
    { determined_term = term; 
      (* VSM: Does this make sense??? *)
      determined_pred = compute_term_pred join term join.join_pred;
      determined_assertion = Wff.Simple.Constant(true);
      determined_body = body }

(* *****************************************************************
   Make a new join from an old join by updating the list of terms and the
   body. 
   ***************************************************************** *)

and update_join old_join new_terms new_body =
  List.iter
    (fun new_term ->
      assert (List.mem new_term old_join.join_terms))
    new_terms;
  (* fixme: need to project the old jt_flag's, etc. from everything. *)
  { join_t_var = old_join.join_t_var;
    join_bounds = old_join.join_bounds;
    join_pred = old_join.join_pred;
    join_assertion = old_join.join_assertion;
    join_terms = new_terms;
    join_body = new_body }

(* *****************************************************************
   Utilities
   ***************************************************************** *)

and compute_term_pred join term p =
  let { jt_flag = term_flag } = term in
  match p with
    Wff.Simple.Constant(true) -> p
  | Wff.Simple.Atom(term_flag') -> 
      if term_flag = term_flag' 
      then p 
      else Wff.Simple.Constant(true)
  | Wff.Simple.Conjunct(p1,p2) -> (
      let p1' = compute_term_pred join term p1 in
      match p1' with
	Wff.Simple.Constant(true) -> compute_term_pred join term p2
      |	_ -> p1')
  | _ -> failwith "Invalid Wff"
					 
and solve_for e var =
  let (c,terms) =
    List.fold_left
      (fun (c,terms) term ->
	match term with
	  Expr_op(Op_times,[Expr_int(c');Expr_var(var')]) ->
	    if var = var' then (c+c',terms) else (c,term::terms)
	| Expr_op(Op_times,[Expr_var(var');Expr_int(c')]) ->
	    if var = var' then (c+c',terms) else (c,term::terms)
	| Expr_op(Op_neg,[Expr_var(var')]) ->
	    if var = var' then (c-1,terms) else (c,term::terms)
	| Expr_var(var') ->
	    if var = var' then (c+1,terms) else (c,term::terms)
	| Expr_int(_) -> (c,term::terms)
	| _ -> failwith "Invalid term")
      (0,[]) (bmf_expr_args_by_op Op_plus e)
  in
  (c,terms)

(* These are duplicated in other files but not exported anywhere. *)
and make_max lbs =
  match lbs with
    [] -> failwith "Can't happen"
  | [lb] -> lb
  | lbs -> Expr_op(Op_other("max"),lbs)
				
and make_min ubs =
  match ubs with
    [] -> failwith "Can't happen"
  | [ub] -> ub
  | ubs -> Expr_op(Op_other("min"),ubs)
				
(************************************************************************)
(*			  Command line processing			*)
(************************************************************************)

let _ =
  register_switch "-jse"
    "-jse (bmf) -- Join implementation selection."
    (fun top_decls args ->
      let name = Bsct_toplevel.get_arg_name (List.hd args)
      and (options,param,args') = Bsct_toplevel.parse_arg false args
      in let (just_dense_p,options) = Bsct_toplevel.extract_option options 'd'
      in let (level_0_p,options) = Bsct_toplevel.extract_option options '0'
      in let (level_1_p,options) = Bsct_toplevel.extract_option options '1'
      in let _ = Bsct_toplevel.ignore_options options name
      in let sched_what =
    	if just_dense_p then Sched_just_dense 
    	else Sched_everything
      in let sched_level = 
    	if not level_0_p then 1 else 0
      in let top_decls' = 
	schedule_procs sched_what sched_level top_decls
      in (top_decls',args'))
