(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Bsct_base
open Printf

module type L_affine =
  sig
    module L : Lambda_expr.L
    type atom = L.atom
    type expr = L.expr
    type key = L.key

    module A : 
	Affine.A with type atom = atom
    	and type expr = expr

    type analysis_result =
	{ param_t : key;
	  param_u : key;
	  key_input : key;
	  key_output : key;
	  summary_lbda : L.t;
	  desc : analysis_result_desc }

    and analysis_result_desc =
      |	Overconstrained
      |	Universal of L.t          (* Lin Lout Lu . out= f(in,u) *)
      |	Universal_unique of L.t   (* Lin Lout . out = f(in) *)
      |	Bound of L.t * L.t        (* Lin Lt . in=f(t) ; Lout Lt Lu . out=g(t,u) *)
      |	Bound_unique of L.t * L.t (* Lin Lt . in=f(t) ; Lout Lt . out=g(t) *)

    val analyze : L.t -> (*out*) key -> (*in*) key -> 
      (*t*) key -> (*u*) key -> analysis_result

    val analyze1 : L.t -> (*out*) key -> (*t*) key -> analysis_result

  end

module Make = 
  functor (Lbda :  Lambda_expr.L) ->
    struct
      module L = Lbda
      type atom = L.atom
      type expr = L.expr
      type key = L.key

      module Key = L.Key

      module A = Affine.Make(Lbda.Ex)
      module E = L.Ex

      type analysis_result =
	  { param_t : key;
	    param_u : key;
	    key_input : key;
	    key_output : key;
	    summary_lbda : L.t;
	    desc : analysis_result_desc }
	    
      and analysis_result_desc =
      	| Overconstrained
      	| Universal of L.t          (* Lin Lout Lt . out= f(in,t) *)
      	| Universal_unique of L.t   (* Lin Lout . out = f(in) *)
      	| Bound of L.t * L.t        (* Lin Lt . in=f(t) ; Lout Lt Lu . out=g(t,u) *)
      	| Bound_unique of L.t * L.t (* Lin Lt . in=f(t) ; Lout Lt . out=g(t) *)


      (*********************************************************)
      let normalize_predicate p =
	(* a==b & c==d & e==f  ---> [a-b; c-d; e-f] *)
	(* have to watch out for "true" and "false" *)
	(* true becomes 0==0, false becomes 1==0 *)
	let zero = E.Const(E.Int 0)
	and one = E.Const(E.Int 1) in 
	let process_expr e = (
	  match e with
	    E.Const(E.Bool(true)) -> zero (* == 0 *)
	  | E.Const(E.Bool(false)) -> one (* == 0 *)
	  | _ -> 
	      (match (A.normalize_eq e) with
	      | E.Op(E.Eq,[e';E.Const(E.Int(0))]) -> e' (* == 0 *)
	      |	_ -> failwith "strip_eq: not normalized")) in
	(List.map process_expr (E.args_by_op E.And p))
	
      (*********************************************************)
      let make_assignment_predicate x lst =
	(zassert ((List.length lst) = (List.length x))
	   (Invalid_argument "make_assignment_predicate"));
	let make_1_assignment (x,e) = 
	  (E.make_binary E.Eq (E.Atom(x)) e) in
	(E.make_and
	   (List.map make_1_assignment 
	      (List.combine x lst)))
	   
      (*********************************************************)
      let k2s k = (Sx.to_string (Key.to_sexpr k))

      let analyze_worker key_out key_in key_t key_u 
	  all_vars x y constraints = begin
	match (A.analyze_relation L.gensyms all_vars x y constraints)
	with
	  A.Overconstrained -> Overconstrained
	| A.Bound_x(t,u,f,g) -> 
	    (* we have that x =f(t) and y=g(t,u)
	       make two separate lambdas *)
	    let lbda_x = 
	      (L.make
		 { L.bindings = [(key_in,x); (key_t,t)];
		   L.predicate = (make_assignment_predicate x f) })
	    and lbda_y =
	      (L.make
		 { L.bindings = 
		   if (u = []) then 
		     [(key_out,y); (key_t,t)] 
		   else 
		     [(key_out,y); (key_t,t); (key_u,u)];
		   L.predicate = (make_assignment_predicate y g) }) in
	    if (u = []) then Bound_unique(lbda_x, lbda_y)
	    else Bound(lbda_x, lbda_y)
	| A.Universal_x(u,f) ->
	    (* we havethat y = f(x,u) *)
	    let lbda_y = 
	      (L.make
		 { L.bindings = 
		   if (u = []) then 
		     [(key_in,x); (key_out,y)] 
		   else 
		     [(key_in,x); (key_out,y); (key_u,u)];
		   L.predicate = (make_assignment_predicate y f) }) in
	    if (u = []) then Universal_unique(lbda_y)
	    else Universal(lbda_y)
      end (* analyze_worker *)

      let summarize descriptor keyi keyj key_t key_u = begin
	match descriptor with
	  Overconstrained -> (L.make { L.bindings=[];
				       L.predicate=E.Const(E.Bool(false)); })
	| Universal(lbda) -> (L.rename_key lbda key_u key_t)
	| Universal_unique(lbda) -> (L.add_key lbda key_t 0)
	| Bound_unique(lbda1, lbda2) ->
	    (* lbda1 is of the form Lkeyj Lt aj=f(t) *)
	    (* lbda2 is of the form Lkeyi Lt ai=g(t) *)
	    (L.apply lbda1 lbda2
	       [ L.A(keyj,keyj);
		 L.A(key_t, key_t);
		 L.B(keyi,keyi);
		 L.Sub(key_t,key_t) ])

	| Bound(lbda1, lbda2) ->
	    (* lbda1 is of the form Lkeyj Lt aj=f(t) *)
	    (* lbda2 is of the form Lkeyi Lt Lu ai=g(t,u) *)
	    let lbda' = (L.apply lbda1 lbda2
			   [ L.A(keyj,keyj);
			     L.A(key_t, key_t);
			     L.B(keyi,keyi);
			     L.B(key_u,key_u);
			     L.Sub(key_t,key_t) ]) in
	    (L.concat_vars lbda' [key_t; key_u] key_t)	    
      end (* summarize *)
	
      let analyze lbda key_out key_in key_t key_u = begin
	if L.debug_p() then begin
	  print_string "==============================\n";
	  print_string "L.analyze trace\n";
	  (printf "lbda=%s to=%s from=%s\n"
	     (Sx.to_string (L.to_sexpr lbda)) 
	     (k2s key_out) (k2s key_in));
	end
	else ();

	let constraints = (normalize_predicate 
			     (L.get_predicate lbda))
	and x = (L.get_var_list lbda key_in)
	and y = (L.get_var_list lbda key_out)
	and all_vars = (L.get_all_vars lbda) in
	let descriptor = (analyze_worker 
			    key_out key_in key_t key_u
			    all_vars x y
			    constraints) in

	{ param_t = key_t; param_u = key_u;
	  key_output = key_out; key_input = key_in;
	  summary_lbda = (summarize descriptor key_out key_in key_t key_u);
	  desc = descriptor }
      end (* analyze *)

      let analyze1 lbda key_in key_t = begin
	if L.debug_p() then begin
	  print_string "==============================\n";
	  print_string "L.analyze1 trace\n";
	  (printf "lbda=%s from=%s\n"
	     (Sx.to_string (L.to_sexpr lbda)) 
	     (k2s key_in));
	end
	else ();

	let constraints = (normalize_predicate 
			     (L.get_predicate lbda))
	and x = (L.get_var_list lbda key_in)
	and all_vars = (L.get_all_vars lbda) in
	let descriptor = (analyze_worker 
			    key_in key_in key_t key_t
			    all_vars x []
			    constraints) in
	{ param_t = key_t; param_u = key_t;
	  key_output = key_in; key_input = key_in;
	  summary_lbda = (summarize descriptor key_in key_in key_t key_t);
	  desc = descriptor }
      end (* analyze1 *)

    end (* module Make ... *)
	
