(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Bsct_base

module type Expr_with_generator_type =
  sig
    module Ex : Expr.E
    val gensym : unit -> Ex.atom
    val gensyms : int -> Ex.atom list
  end

module type L = 
    sig
      module Ex : Expr.E
      module Key : AtomType
      type atom = Ex.atom
      type expr = Ex.expr
      type key = Key.t
      type t
      type builder = { bindings : (key * (atom list)) list;
		       predicate : expr }

      type descriptor = 
	  A of key * key 
	| B of key * key
	| Sub of key * key

      val debug_p : unit -> bool
      val debug_on : unit -> unit
      val debug_off : unit -> unit

      val gensym : unit -> atom
      val gensyms : int -> atom list

      val verify : t -> string -> t
      val verify_p : builder -> bool

      val make : builder -> t
      val destructure : t -> builder
      val rename_unique : t -> t

      val n_var_lists : t -> int
      val var_list_size : t -> key -> int
      val get_var_list : t -> key -> atom list
      val get_all_vars : t -> atom list
      val get_all_keys : t -> key list
      val check_keys_p : t -> key list -> bool

      val get_predicate : t -> expr
      val set_predicate : t -> expr -> t

      val concat_vars : t -> key list -> key -> t
      val rename_key : t -> key -> key -> t
      val add_key : t -> key -> int -> t
      val add_key_names : t -> key -> atom list -> t
      val split_vars : t -> key -> int -> key -> key -> t
      val apply : t -> t -> descriptor list -> t
      val key_blit : t -> key list -> (key * (atom list)) list -> t

      val to_sexpr : t -> Sx.t
    end

module type L_input_type =
  sig
    module EG : Expr_with_generator_type
    module Key : AtomType
  end

module Make (IN : L_input_type) =
  struct
    module EG = IN.EG
    module Key = IN.Key

    module Ex = EG.Ex       

    type atom = Ex.atom
    type expr = Ex.expr
    type key = Key.t

    module KeyMap' = Map'.Make(Key)
    module KeyMap = KeyMap'.S

    type t = { vars : (atom list) KeyMap.t ;
	          (* vars is a map from key type to atom list *)
	       pred : expr }

    type builder = { bindings : (key * (atom list)) list;
		     predicate : expr }
	  
    type descriptor = 
	A of key * key 
      | B of key * key 
      | Sub of key * key

    let debug = ref false
    let debug_p () = !debug
    and debug_on () = debug := true
    and debug_off () = debug := false

    let gensym = EG.gensym
    let gensyms = EG.gensyms

    module ATOM = Ex.ATOM (* just for convenience *)
    module S = Map.Make(ATOM)
    module U = Map'.Make(ATOM)
    module VarSet = Set.Make(ATOM)
	  


    (*********************************************************)
    let to_sexpr {vars=v_map; pred=p} =
      let process_one (k,lst) = 
	(Sx.Pair((Key.to_sexpr k),
		 (Sx.from_list (List.map ATOM.to_sexpr lst)))) in
      (Sx.from_list
	 ((List.map process_one
	    (KeyMap'.list_of_bindings v_map)) 
	  @ [ (Ex.to_sexpr p) ]))

    (****************************************************************)
    (*take a list of variables lists and an expression
      and verify that they are suitable to construct a
      lambda expression

      lst is the list of (key,var_list) pairs
      the_predicate is the predicate

      Here are the rules:
      - variable lists are all disjoint
      - keys are unique
      - all variables used in the_predicate 
          must be defined in lsts
      *)
    let verify_p {bindings=lst; predicate=the_predicate} =
      let all_vars = 
	(List'.flat_map (fun (k,l) -> l) lst) in
      let all_keys =
	(List.map (fun (k,l) -> k) lst) in
      if ((List'.uniqp all_vars) && (List'.uniqp all_keys)) then begin
	let used_vars = (Ex.collect_atoms the_predicate) in
	(List'.subsetp used_vars all_vars)
      end
      else false

    let verify_ready lbda msg =
      let ok = (verify_p 
		  { bindings=(KeyMap'.list_of_bindings lbda.vars);
		    predicate=lbda.pred }) in
      if ok then lbda
      else begin
	let str = (Sx.to_string (to_sexpr lbda)) in
	raise (Failure (msg^" : "^str))
      end

    let verify = verify_ready

    (******************************************************************)
    let make {bindings=lst; predicate=p} = 
      let v_map = (KeyMap'.absorb_list 
		     KeyMap.empty
		     lst) in
      let lbda = { vars=v_map; pred=p } in
      (verify_ready lbda "Lambda_expr.make ")

    let destructure {vars=v_map; pred=p} =
      { bindings=(KeyMap'.list_of_bindings v_map); 
	predicate=p }

    let rename_unique {vars=v_map; pred=the_predicate} =
      (* gives the variables unique names *)
      let rename_a_list lst = (gensyms (List.length lst)) in
      let bind_lst = (KeyMap'.list_of_bindings v_map) in
      let bind_lst' = 
	(List.map (fun (k,l) -> (k,(rename_a_list l))) bind_lst) in
      let var_lst = (List'.flat_map Pervasives.snd bind_lst)
      and var_lst' = (List'.flat_map Pervasives.snd bind_lst') in
      let new_predicate = 
	(Ex.rename var_lst var_lst' the_predicate) in
      (make {bindings=bind_lst'; predicate=new_predicate} )

    let n_var_lists {vars=v_map} = (KeyMap'.map_size v_map)

    let fail_not_found_key a k msg = (
      let a_str = (Sx.to_string (to_sexpr a))
      and k_str = (Sx.to_string (Key.to_sexpr k)) in
      failwith (msg ^ ": key not found: " ^ k_str ^ " in: " ^ a_str))

    let var_list_size a k =
      try (List.length (KeyMap.find k a.vars))
      with
	Not_found -> 
	  (fail_not_found_key a k "Lambda_expr.var_list_size")

    let get_var_list a k =
      try (KeyMap.find k a.vars)
      with
	Not_found -> 
	  (fail_not_found_key a k "Lambda_expr.get_var_list")

    let get_all_vars a =
      (List'.flat_map Pervasives.snd
	 (KeyMap'.list_of_bindings a.vars))

    let get_all_keys a =
      (List.map Pervasives.fst
	 (KeyMap'.list_of_bindings a.vars))

    let check_keys_p a lst = 
      let lst' = (get_all_keys a) in
      (List'.samep lst lst')

    let get_predicate a = a.pred
    let set_predicate a p =
      (verify_ready
	 { pred = p; vars = a.vars }
	 "Lambda_expr.set_predicate")

    let key_to_string k = (Sx.to_string (Key.to_sexpr k))
    let key_lst_to_string lst =
      (String.concat " " (List.map key_to_string lst))

    let atom_to_string v = (Sx.to_string (ATOM.to_sexpr v))
    let atom_lst_to_string lst =
      (String.concat " " (List.map atom_to_string lst))

    let make_msg msg k = (msg ^ " : " ^ (key_to_string k))

    let verify_key a k msg =
      try ((KeyMap.find k a.vars); ())
      with
	Not_found -> failwith(make_msg msg k)

    let verify_not_key a k ok_lst msg =
      try
	((KeyMap.find k a.vars);
	 if (List.mem k ok_lst) then ()
	 else (failwith(make_msg msg k)))
      with
	Not_found -> ()

    (*********************************************************)
    let concat_vars a key_src_lst key_dest =
      (List.iter 
	 (fun k -> verify_key a k "Lambda_expr.concat_vars: src")
	 key_src_lst);
      (verify_not_key a key_dest key_src_lst
	"Lambda_expr.concat_vars: key_dest");

      let new_lst = 
	(List'.flat_map (get_var_list a) key_src_lst) in
      let b = 
      	{ vars = (KeyMap.add key_dest new_lst
		    (KeyMap'.remove_list a.vars key_src_lst));
	  pred = a.pred } in
      (verify_ready b "Lambda_expr.concat_vars")

    let rename_key a from_key to_key =
      (verify_key a from_key "Lambda_expr.concat: src");
      (verify_not_key a to_key [from_key] "Lambda_expr.concat: src");
      (concat_vars a [from_key] to_key)

    let add_key_names a k lst =
      (verify_not_key a k [] "add_key");
      { vars = (KeyMap.add k lst a.vars);
	pred = a.pred }

    let add_key a k n =
      (add_key_names a k (gensyms n))

    let split_vars a key_from j key1 key2 = 
	(* split the list with key=key_from at the j-th variable:
	   (x0 x1 ... xj ... x(m-1)) -->
	   (x0 ... x(j-1))  (xj ... x(m-1))
	   and place the first list at key1, second -- at key2
	   *)

      (zassert (key1 <> key2) (Failure "Lambda_expr.split: same keys"));
      verify_key a key_from "Lambda_expr.split: key_from";
      verify_not_key a key1 [key_from] "Lambda_expr.split: key1";
      verify_not_key a key2 [key_from] "Lambda_expr.split: key2";

      let lst = (get_var_list a key_from) in
      let m = (List.length lst) in
      zassert (0<=j && j<=m) (Invalid_argument("Lamda_expr.split_vars"));
      let arr = (Array.of_list lst) in
      let lst1 = (Array.to_list (Array.sub arr 0 j))
      and lst2 = (Array.to_list (Array.sub arr j (m-j))) in
      let b = { vars = (KeyMap'.absorb_list
			  (KeyMap.remove key_from a.vars)
			  [ (key1,lst1) ; (key2,lst2) ]);
		pred = a.pred } in
      (verify_ready b "Lamda_expr.split_vars ")
 
    (*********************************************************)
    let apply a b the_desc_lst = 
      let rec place_directly intermediate lbda key_from key_to =
	(verify_not_key intermediate key_to [] 
	   "Lamda_expr.apply:place_directly key_to");
	(verify_key lbda key_from 
	   "Lamda_expr.apply:place_directly key_from");
	{ vars = (KeyMap.add key_to
		    (get_var_list lbda key_from)
		    intermediate.vars) ;
	  pred = intermediate.pred }

      and substitute intermediate key_from_b key_to_in_result =
	(verify_key b key_from_b "Lamda_expr.apply:substitute key_from_b");
	(verify_key intermediate key_to_in_result 
	   "Lamda_expr.apply:substitute key_to_in_result");
	(zassert ((var_list_size b key_from_b) = 
		  (var_list_size intermediate key_to_in_result))
	   (Failure("Lamda_expr.apply:substitute sizes")));
	{ vars = intermediate.vars ;
	  pred = (Ex.rename 
		    (get_var_list b key_from_b)
		    (get_var_list intermediate key_to_in_result)
		    intermediate.pred) }

      and driver intermediate desc = (
	match desc with
	  A(key_from, key_to) -> place_directly intermediate a key_from key_to
	| B(key_from, key_to) -> place_directly intermediate b key_from key_to
	| Sub(key_from_b,key_to_in_result) ->
	    substitute intermediate key_from_b key_to_in_result)

      in
      let init = { vars=KeyMap.empty; 
		   pred = (Ex.make_and [a.pred ; b.pred]) } in
      (verify_ready
	 (List.fold_left driver init the_desc_lst)
	 "Lamda_expr.apply ")

    (*********************************************************)
    let key_blit a lst blst =
      let new_keys = (List.map Pervasives.fst blst)
      and vars = (List'.flat_map (get_var_list a) lst)
      and new_vars = (List'.flat_map Pervasives.snd blst) in
      (zassert (List'.samep vars new_vars)
	 (Failure("Lambda_expr.key_blit not the same vars: "^
		  (atom_lst_to_string vars) ^ "," ^
		  (atom_lst_to_string new_vars))));
      (verify_ready
	 { vars=(KeyMap'.absorb_list 
		   (KeyMap'.remove_list a.vars lst)
		   blst);
	   pred = a.pred }
	 "Lambda_expr.key_blit")
      
  end
	

