(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
(****************************************************************

   Lambda module provides predicates with scoped variables
   In general a lambda affine expression is of the form:
     La1.La2...Lan P(a1,...an), n >= 0
   --> where each ai is a LIST of variables. <-- 
   Basically, we group the variables for convenience.

   When the expression is built the variables are uniquelly 
   renamed to prevent capture in subsequent operations.

   The main operation is an combination, which is defined as follows:

   Suppose we have two lambda expressions:
     A = La1.La2...Lan P(a1,...,an)
     B = Lb1.Lb2...Lbm Q(b1,...,bm)
   In order to combine A and B, we need to specify how each ai and bj
   appears in the result. This is done using a list of data structures
   called a descriptor. This data structure is an array 

   Let D[1..r] be a list of r descriptors, and that each descriptor has
   the type:
     type descriptor = A of int | B of int | Subst of int * int

   Each D[i] tells how the i-th list of variables in the result of the
   combination of A and B should be constructed:
   - D[i] == A(k) then the i-th group is taken from A
   - D[i] == B(k) ................................. B
   - D[i] == Subst(j,k) then the variables bj are replaced by ak in Q.
             It is assumed that D[i'] == A(k) for some i'

   The sanity rule is that at the end all the variables should be scoped.
   The variables are uniquelly renamed after the combination.

   Notice that combination can be used to reorder the variable lists 
   (by supplying an empty second argument)

***********************************************************************)

open Bsct_base
open Ssisl

(* To construct a module of lambda expressions all we need a module
   for the expressions + a symbol (atom) generator *)

module type Expr_with_generator_type =
  sig
    module Ex : Expr.E
    val gensym : unit -> Ex.atom
    val gensyms : int -> Ex.atom list
  end

module type L = 
    sig
      module Ex : Expr.E
      module Key : AtomType
      type atom = Ex.atom
      type expr = Ex.expr
      type key = Key.t
      type t
      type builder = { bindings : (key * (atom list)) list;
		       predicate : expr }

      type descriptor = 
	  A of key * key 
	| B of key * key
	| Sub of key * key

      val debug_p : unit -> bool
      val debug_on : unit -> unit
      val debug_off : unit -> unit

      val gensym : unit -> atom
      val gensyms : int -> atom list

      val verify : t -> (* failure msg *) string -> t
      val verify_p : builder -> bool

      val make : builder -> t
      val destructure : t -> builder
      val rename_unique : t -> t

      val n_var_lists : t -> int
      val var_list_size : t -> key -> int
      val get_var_list : t -> key -> atom list
      val get_all_vars : t -> atom list
      val get_all_keys : t -> key list
      val check_keys_p : t -> key list -> bool

      val get_predicate : t -> expr
      val set_predicate : t -> expr -> t

      val concat_vars : t -> key list -> key -> t
      val rename_key : t -> key -> key -> t
      val add_key : t -> key -> int -> t
      val add_key_names : t -> key -> atom list -> t
      val split_vars : t -> key -> int -> key -> key -> t
      (* This is the combination operation. *)
      val apply : t -> t -> descriptor list -> t

      (* reorganize the variables in lst_in as specified by the
	 list of bindings blst *)
      val key_blit : t -> (* lst *) key list 
	 -> (* blst *) (key * (atom list)) list -> t

      val to_sexpr : t -> Sx.t
    end

module type L_input_type =
  sig
    module EG : Expr_with_generator_type
    module Key : AtomType
  end

module Make (IN : L_input_type) :
    L with module Ex = IN.EG.Ex and module Key = IN.Key

