(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
(*************************************************************)
(*Operations on lambda expressions in the scheme interpreter *)
(*************************************************************)
open Printf
open Ssisl
open Bsct_base
open Expr_string


(*********************************************************)
(*         Helper functions                              *)
(*********************************************************)
let untag_key form = 
  (Scheme.symbol_to_string form)


(*********************************************************)
(*********************************************************)
let lbda_to_scheme lbda = (Sx.to_scheme (L_convert.unparse lbda))
and scheme_to_lbda form = (L_convert.parse (Sx.from_scheme form))

let verify form =
  (lbda_to_scheme (scheme_to_lbda form))

let apply form_desc form_a form_b =
  let desc_lst = 
    (List.map L_convert.parse_desc
       (Sx.to_list 
          (Sx.from_scheme form_desc)))
  and a = (scheme_to_lbda form_a)
  and b = (scheme_to_lbda form_b) in
  let c = L.apply a b desc_lst in
  (lbda_to_scheme c)

let analyze form_x form_y form_t form_u form_a =
  let key_x = (untag_key form_x)
  and key_y = (untag_key form_y)
  and key_t = (untag_key form_t)
  and key_u = (untag_key form_u)
  and a =  (scheme_to_lbda form_a) in
  match ((L_affine.analyze a key_y key_x key_t key_u).L_affine.desc) with
  | L_affine.Overconstrained -> Scheme.nil
  | L_affine.Bound(lbda_x, lbda_y) ->
      (Scheme.cons
         (lbda_to_scheme lbda_x)
         (lbda_to_scheme lbda_y))
  | L_affine.Bound_unique(lbda_x, lbda_y) ->
      (Scheme.cons
         (lbda_to_scheme lbda_x)
         (lbda_to_scheme lbda_y))
  | L_affine.Universal(lbda_y) ->
      (lbda_to_scheme lbda_y)
  | L_affine.Universal_unique(lbda_y) ->
      (lbda_to_scheme lbda_y)

let summary form_x form_y form_t form_u form_a =
  let key_x = (untag_key form_x)
  and key_y = (untag_key form_y)
  and key_t = (untag_key form_t)
  and key_u = (untag_key form_u)
  and a =  (scheme_to_lbda form_a) in
  let sum_lbda = 
    (L_affine.analyze a key_y key_x key_t key_u).L_affine.summary_lbda in
  (lbda_to_scheme sum_lbda)
    
let rename form_a =
  (lbda_to_scheme
     (L.rename_unique
        (scheme_to_lbda form_a)))

let concat_vars form_src form_to form_a =
  let keys_from = (List.map untag_key 
		     (Scheme.to_list form_src))
  and key_to = (untag_key form_to) 
  and lbda = (scheme_to_lbda form_a) in
  (lbda_to_scheme (L.concat_vars lbda keys_from key_to))

let split_vars form_src form_j form_dst1 form_dst2  form_a =
  let key_from = (untag_key form_src)
  and j = (Scheme.untag_int form_j)
  and key_to1 = (untag_key form_dst1)
  and key_to2 = (untag_key form_dst2)
  and lbda =  (scheme_to_lbda form_a) in
  (lbda_to_scheme 
     (L.split_vars lbda key_from j key_to1 key_to2))

(*********************************************************)
(*   Definitions of Scheme primitives                    *)
(*********************************************************)
let attach env name_prefix = 
  (Scheme.define_prim1 env (name_prefix ^ "verify")   verify);
  (Scheme.define_prim3 env (name_prefix ^ "apply")    apply);
  (Scheme.define_prim5 env (name_prefix ^ "analyze")  analyze);
  (Scheme.define_prim5 env (name_prefix ^ "summary")  summary);
  (Scheme.define_prim3 env (name_prefix ^ "concat")   concat_vars);
  (Scheme.define_prim5 env (name_prefix ^ "split")    split_vars);
  (Scheme.define_prim1 env (name_prefix ^ "rename")   rename);
  ()
  



