(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Bsct_base

open Expr_string

module NameMap' = Map'.Make(SimpleAtom)
module NameMap = NameMap'.S

(*********************************************************)
type t = {
    bmap : binding_map;
    iter : string list;
    out_vars : string list;
    eqs : lambda_expression;
    ineqs : lambda_expression;
    pred : lambda_expression
  }

and binding = {
    relation : string;
    index_vars: string list;
    value_var : string;
    flag_var : string }

and binding_map = binding NameMap.t (* string-to-binding map *)

and lambda_expression = Expr_string.L.t

and  bound_var_type =
    Index of string
  | Value of string
  | Flag of string 
  | Relation of string
    
and var_type = 
    Loop | Bound of bound_var_type

(*********************************************************)

let get_inverse_binding_list re = begin
  let list4binding (name,bind) = (
    (List.map (fun s -> (s,Bound(Index(name)))) bind.index_vars) @
    [ (bind.value_var, Bound(Value(name)));
      (bind.flag_var, Bound(Flag(name)));
      (bind.relation, Bound(Relation(name))) ]) in

  ((List.map (fun s -> (s,Loop)) re.iter) @
   (List'.flat_map list4binding (NameMap'.list_of_bindings re.bmap))
     )

end (* get_inverse_binding_list *)

let get_value_vars re =
  let is_value_var desc = (
    match desc with
      (_,Bound(Value(_))) -> true
    | _ -> false) in
  (List.map Pervasives.fst
     (List'.remove_if_not is_value_var 
	(get_inverse_binding_list re)))

let get_index_vars re =
  let is_index_var desc = (
    match desc with
      (_,Bound(Index(_))) -> true
    | (_,Loop) -> true
    | _ -> false) in
  (List.map Pervasives.fst
     (List'.remove_if_not is_index_var 
	(get_inverse_binding_list re)))

let get_flag_vars re =
  let is_flag_var desc = (
    match desc with
      (_,Bound(Flag(_))) -> true
    | _ -> false) in
  (List.map Pervasives.fst
     (List'.remove_if_not is_flag_var 
	(get_inverse_binding_list re)))

let get_relation_names re =
  let is_rel_name desc = (
    match desc with
      (_,Bound(Relation(_))) -> true
    | _ -> false) in
  (List.map Pervasives.fst
     (List'.remove_if_not is_rel_name
	(get_inverse_binding_list re)))

let get_binding_names re =
  (List.map Pervasives.fst
     (NameMap'.list_of_bindings re.bmap))

let get_binding_name4var_type bvt = (
  match bvt with
    Relation(s) -> s
  | Value(s) -> s
  | Flag(s) -> s
  | Index(s) -> s)


(*********************************************************)

let verify re msg_str = begin
  let msg s = (msg_str ^ " : " ^ s) in
  let all_vars = (List.map Pervasives.fst
		    (get_inverse_binding_list re)) in
  (zassert (List'.uniqp all_vars) 
     (Failure(msg "not unique vars")));
  (L.verify re.eqs (msg "eqs"));
  (L.verify re.ineqs (msg "ineqs"));
  (L.verify re.pred (msg "pred"));

  (zassert ((L.var_list_size re.eqs "symb") =
	    (L.var_list_size re.ineqs "symb"))
     (Failure(msg "eq/ineq symbolics")));

  (* keys in the equality and inequality 
     lambdas must be the same as the names of 
     the bindings + "symb" + "iter" *)
  let keys_must = "symb" :: "iter" :: (get_binding_names re) in
  (zassert (L.check_keys_p re.eqs keys_must) 
     (Failure(msg "eq keys")));
  (zassert (L.check_keys_p re.ineqs keys_must) 
     (Failure(msg "ineq keys")));

  (* the dimensions of array indices must agree
     with the sizes of corresponding variable
     lists in eq and ineq *)

  (NameMap.iter
     (fun name bind ->
       (zassert ((List.length bind.index_vars) =
		 (L.var_list_size re.eqs name))
	  (Failure(msg ("eq mismatch in "^name))));
       (zassert ((List.length bind.index_vars) =
		 (L.var_list_size re.ineqs name))
	  (Failure(msg ("ineq mismatch in "^name)))))
     re.bmap);

  (* the number of loop vars must agree with the "iter" list
      in eqs and ineqs *)

  (zassert ((List.length re.iter) = 
	    (L.var_list_size re.eqs "iter"))
     (Failure(msg "eq mismatch in iter")));
  (zassert ((List.length re.iter) = 
	    (L.var_list_size re.ineqs "iter"))
     (Failure(msg "ineq mismatch in iter")));
       

  (* now, check the out_vars : can be anything *)
  (zassert (List'.subsetp re.out_vars all_vars)
     (Failure(msg "out_vars")));

  re

  (* NOTE: notice that the lambdas do not have to carry the
     same variable names as the bindings -- just the 
     dimensions  have to agree *)

end (* verify *)

(*********************************************************)
let make 
      iter_vars out bindings 
      eq_lbda ineq_lbda pred_lbda = begin
      { iter = iter_vars;
	out_vars = out;
	bmap = (NameMap'.absorb_list NameMap.empty bindings);
	eqs = eq_lbda;
	ineqs = ineq_lbda;
	pred = pred_lbda }
end (* make *)

(*********************************************************)

let get_all_bindings re = (NameMap'.list_of_bindings re.bmap);;
  
let get_binding_names re = (List.map Pervasives.fst
			      (get_all_bindings re));;

let get_binding re name =
  try (NameMap.find name re.bmap)
  with
    Not_found -> 
      (failwith ("Rel_expr.get_binding: not found "^name))

let has_binding_p re name = 
  try (NameMap.find name re.bmap; true)
  with Not_found -> false

let set_binding re name bind =
  (zassert (has_binding_p re name)
     (Failure ("Rel_expr.set_binding: not found"^name)));
  { bmap = (NameMap'.replace name bind re.bmap) ;
    out_vars = re.out_vars;
    iter = re.iter ;
    eqs = re.eqs;
    ineqs = re.ineqs;
    pred = re.pred }

let remove_binding re name =
  { bmap = (NameMap.remove name re.bmap);
    out_vars = re.out_vars;
    iter = re.iter ;
    eqs = re.eqs;
    ineqs = re.ineqs;
    pred = re.pred }

let rename_binding re name name' =
  (zassert (has_binding_p re name)
     (Failure ("Rel_expr.rename_binding: not found"^name)));
  (zassert (not (has_binding_p re name'))
     (Failure ("Rel_expr.rename_binding: already exists"^name')));
  let bind = (get_binding re name) in
  { bmap = (NameMap.add name' bind
	      (NameMap.remove name re.bmap));
    out_vars = re.out_vars;
    iter = re.iter ;
    eqs = re.eqs;
    ineqs = re.ineqs;
    pred = re.pred }
  
  
(*********************************************************)

let get_eqs re = re.eqs
let set_eqs re lbda =
  { bmap = re.bmap ;
    out_vars = re.out_vars;
    iter = re.iter ;
    eqs = lbda;
    ineqs = re.ineqs;
    pred = re.pred }

let get_ineqs re = re.ineqs
let set_ineqs re lbda =
  { bmap = re.bmap ;
    out_vars = re.out_vars;
    iter = re.iter ;
    eqs = re.eqs;
    ineqs = lbda;
    pred = re.pred }

let get_pred re = re.pred
let set_pred re lbda =
  { bmap = re.bmap ;
    out_vars = re.out_vars;
    iter = re.iter ;
    eqs = re.eqs;
    ineqs = re.ineqs;
    pred = lbda }

let get_iter re = re.iter
let set_iter re lst =
  { bmap = re.bmap ;
    out_vars = re.out_vars;
    iter = lst ;
    eqs = re.eqs;
    ineqs = re.ineqs;
    pred = re.pred }

let get_out_vars re = re.out_vars
let set_out_vars re lst =
  { bmap = re.bmap ;
    out_vars = lst;
    iter = re.iter;
    eqs = re.eqs;
    ineqs = re.ineqs;
    pred = re.pred }

(*********************************************************)
let get_array_access_fn re access_name = begin
  let eqs = (get_eqs re) in
  let eqs' = (L.concat_vars eqs ["symb";"iter"] "symb_iter") in
  let { L_affine.desc = descriptor; } = 
    (L_affine.analyze eqs' access_name "symb_iter" "$t" "$u") in
  match descriptor with
    L_affine.Universal_unique(lbda) ->
      (L.key_blit lbda ["symb_iter"]
	 [ ("symb", (L.get_var_list eqs "symb"));
	   ("iter", (L.get_var_list eqs "iter")) ])
  | _ -> failwith "Rel_expr.get_array_access_fn: not unique"
end (* get_array_access_fn *)


(*********************************************************)
let from_sexpr sexpr = begin
  let parse_binding b_s_expr = (
    match (Sx.to_list b_s_expr) with
      [n;r;a;v;f] ->
	((Sx.untag_symbol n),
	 { relation=(Sx.untag_symbol r);
	   index_vars=(Sx.untag_symbol_list a);
	   value_var=(Sx.untag_symbol v);
	   flag_var=(Sx.untag_symbol f) 
	 }) (* this is a (name,binding) pair *)
    | _ -> (failwith("Rel_expr.parse_binding : "^
		    (Sx.to_string b_s_expr)))) in

  let parse_binding_map sexpr = (
      (List.map parse_binding (Sx.to_list sexpr))) in

  match (Sx.to_list sexpr) with
    [s_i;s_out;s_b;s_eq;s_ineq;s_pred] ->
      (verify (make
		 (Sx.untag_symbol_list s_i)
		 (Sx.untag_symbol_list s_out)
		 (parse_binding_map s_b)
		 (L_convert.parse s_eq)
		 (L_convert.parse s_ineq)
		 (L_convert.parse s_pred))
	 ("Rel_expr: parse "^(Sx.to_string sexpr)))

  | _ -> failwith("Rel_expr: parse "^(Sx.to_string sexpr))
  
end (* from_sexpr *)

let to_sexpr re = begin
  let unparse_binding (name,bind) =
	(Sx.from_list
	   [ (Sx.tag_symbol name);
	     (Sx.tag_symbol bind.relation);
	     (Sx.tag_symbol_list bind.index_vars);
	     (Sx.tag_symbol bind.value_var);
	     (Sx.tag_symbol bind.flag_var) ]) in
  
  (Sx.from_list
     [ (Sx.tag_symbol_list re.iter);
       (Sx.tag_symbol_list re.out_vars);
       (Sx.from_list
	  (List.map unparse_binding
	     (NameMap'.list_of_bindings re.bmap)));
       (L_convert.unparse re.eqs);
       (L_convert.unparse re.ineqs);
       (L_convert.unparse re.pred) ])

end (* to_sexpr *)
	  

(*********************************************************)

let rel_to_scheme re = (Sx.to_scheme (to_sexpr re))
and scheme_to_rel form = (from_sexpr (Sx.from_scheme form))

let verify_scheme form = 
  (rel_to_scheme 
     (verify (scheme_to_rel form) "verify_scheme"))

let get_array_access_fn_scheme form_name form_re =
  let access_name = Scheme.symbol_to_string form_name
  and re = scheme_to_rel form_re in
  let lbda = (get_array_access_fn re access_name) in
  (Sx.to_scheme
     (L.to_sexpr lbda))

let attach env prefix =
  (Scheme.define_prim1 env (prefix ^ "verify") verify_scheme);
  (Scheme.define_prim2 env (prefix ^ "access-fn") 
     get_array_access_fn_scheme);
  ()
