/************************************************************************
 *	Copyright (C) 1998. Bernoulli Project, Cornell University.	*
 ************************************************************************/
/*
 * Communication buffers
 */

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>

#include "rt-memory.h"
#include "commbuffer.h"

commbuffer
make_cbuf (int nproc, int nind, int arity, cbuf_val_type t)
{
    commbuffer c = rt_alloc0(commbuffer_t);

    nproc_cbuf(c) = nproc;
    nind_cbuf(c) = nind;
    arity_cbuf(c) = arity;
    val_type_cbuf(c) = t;

    iptr_cbuf(c) = rt_nalloc0(int, nproc + 1);
    ind_cbuf(c) = rt_nalloc0(int, nind * arity);
    
    switch (t) {
    case CBUF_INT:
	vals_cbuf(c) = rt_nalloc0(int, nind);
	break;
    case CBUF_REAL:
	vals_cbuf(c) = (int *) rt_nalloc0(double, nind);
	break;
    case CBUF_VOID:
	vals_cbuf(c) = 0;
	break;
    }

    return c;
}

void
counts_to_iptr (int nproc, int counts[], int iptr[])
{
    int q, save, sum;

    /*
     * now do the prefix sum on iptr:
     * iptr[q] for q=0..nprocs gets sum_{j=0}^{q-1} iptr[j]
     */
    sum = 0;
    for (q = 0; q <= nproc; q++) {
	save = iptr[q];
	counts[q] = sum;
	sum += save;
    }
}

commbuffer
make_cbuf_from_iptr (int nproc, int *iptr, int arity, cbuf_val_type t)
{
    int nind = iptr[nproc];
    commbuffer c = make_cbuf (nproc, nind, arity, t);
    int i;

    for (i = 0; i <= nproc; i++) {
	iptr4q_cbuf(c,i) = iptr[i];
    }

    return c;
}

commbuffer
pack_cbuf (int nproc, int arity, cbuf_val_type t, hash_table h)
{
    int nind = nelem_hash_table(h);
    commbuffer c;
    int q, ii, j;
    int *htuple;
    int *iptr, *next_ptr;
    hash_iter it;

    /*
     * find how many elements go to each proc
     */
    iptr = rt_nalloc0 (int, nproc+1);
    for (it = open_hash_iter(h); valid_hash_iter(it);
	 next_hash_iter (it)) {
	q = deref_hash_iter (it)[0];
	iptr[q] ++;
    }
    close_hash_iter(it);

    counts_to_iptr (nproc, iptr, iptr);
    assert (iptr[nproc] == nind);

    c = make_cbuf_from_iptr (nproc, iptr, arity, t);
    rt_nfree(int, nproc+1, iptr);
    iptr = iptr_cbuf (c);

    /*
     * now go through the table and fill the buffer
     */
    next_ptr = rt_nalloc0(int, nproc+1);
    for (q = 0; q <= nproc; q++) {
	next_ptr[q] = iptr[q];
    }
    for (it = open_hash_iter(h); valid_hash_iter(it);
	 next_hash_iter(it)) {
	htuple = deref_hash_iter(it);
	q = htuple[0];
	htuple++; /* skip the proc number */

	assert (next_ptr[q] < next_ptr[q+1]);
	
	ii = next_ptr[q];
	next_ptr[q]++;

	for (j = 0; j < arity; j++, htuple++) {
	    ind_tuple_elt_cbuf(c,ii,j) = *htuple;
	}

	/* now htuple points to the value field */
	switch (t) {
	case CBUF_REAL:
	    val4ii_cbuf (double, c, ii) = *((double *) htuple);
	    break;
	case CBUF_INT:
	    val4ii_cbuf (int, c, ii) = *htuple;
	    break;
	case CBUF_VOID:
	    break;
	default:
	    assert (0);
	}
    }
    close_hash_iter(it);
    rt_nfree (int, nproc+1, next_ptr);

    return c;
}

void
free_cbuf (commbuffer c)
{
    int nproc = nproc_cbuf(c);
    int arity = arity_cbuf(c);
    int nind = nind_cbuf(c);
    cbuf_val_type t = val_type_cbuf(c);

    rt_nfree (int, nproc+1, iptr_cbuf(c));
    rt_nfree (int, arity * nind, ind_cbuf(c));
    
    switch(t){
    case CBUF_INT: 
	rt_nfree(int, nind, vals_cbuf(c));
	break;
    case CBUF_REAL: 
	rt_nfree(double, nind, vals_cbuf(c));
	break;
    case CBUF_VOID: 
	break;
    default:
	assert (0);
    }

    rt_free (commbuffer_t, c);
}

int
count_non_empty_procs (commbuffer c)
{
    int q;
    int nproc = nproc_cbuf (c);
    int cnt = 0;

    for (q = 0; q < nproc; q++) {
	if (iptr_cbuf(c)[q+1] > iptr_cbuf(c)[q]) {
	    cnt++;
	}
    }
   
    return cnt;
}

int
find_q4ii_cbuf (commbuffer c, int ii)
{
    int *iptr = iptr_cbuf(c);
    int nproc = nproc_cbuf(c);
    int q1, q, q2;
    
    /* want to find **the largest** q such that iptr[q] <= ii <
       iptr[q+1] */

    q1 = 0;
    q2 = nproc;
    if (ii < 0) {
	fprintf (stderr, "find_q4ii_cbuf: ii must be >= 0\n");
	abort ();
    }
    if (ii >= iptr[nproc]) {
	fprintf (stderr, "find_q4ii_cbuf: ii must be < iptr[nproc]\n");
	abort ();
    }
    while (q2 - q1 > 1) {
	q = q1 + (q2 - q1)/2;
	if (ii < iptr[q]) {
	    q2 = q;
	}
	else {
	    q1 = q;
	}
    }

    return q1;
}

int
find_ii4tuple_cbuf (commbuffer c, int q, int *t)
{
    int arity = arity_cbuf(c);
    int ii;

    for (ii = iptr4q_cbuf(c,q); ii < iptr4q_cbuf(c,q+1); ii++){
	int *t2 = ind_tuple_cbuf(c,ii);
	int k, same;
	
	for (k = 0, same = 1; (k < arity); k++) {
	    same = same && (t[k] == t2[k]);
	}
	
	if (same) {
	    return ii;
	}
    }

    return -1;
}

void 
verify_cbuf(int me, commbuffer c)
{
    int q;
    int ii;
    int arity = arity_cbuf(c);

    for (q = 0; q < nproc_cbuf(c); q++) {
	for (ii = iptr4q_cbuf(c,q); ii < iptr4q_cbuf(c,q+1); ii++) {
	    int *t = ind_tuple_cbuf(c,ii);
	    int ii1;

	    assert (q == find_q4ii_cbuf (c, ii));
	    assert (ii == find_ii4tuple_cbuf (c, q, t));

	    /*
	    ii1 = find_ii4tuple_cbuf (c, q, t);
	    if (ii != ii1) {
		fprintf (stderr, "cbuf failure on %d\n", me);
		abort ();
		
	    }*/
	}
    }
}

commbuffer
proj_cbuf (commbuffer c, int new_arity, int keep_vals)
{
    commbuffer c2 = 
	make_cbuf(
	    nproc_cbuf(c),
	    nind_cbuf(c),
	    new_arity,
	    keep_vals ? val_type_cbuf(c) : CBUF_VOID);
    int q;
    
    assert (new_arity <= arity_cbuf(c));
    /*
     * copy the iptr, but DO NOT copy the tuples
     */
    for (q = 0; q <= nproc_cbuf(c); q++) {
	iptr_cbuf(c2)[q] = iptr_cbuf(c)[q];
    }

    return c2;
}

void
copy_common_prefix_fields_cbuf (commbuffer cto, commbuffer cfrom)
{
    int ii;
    int j;

    assert (nind_cbuf(cto) == nind_cbuf(cfrom));
    assert (arity_cbuf(cto) <= arity_cbuf(cfrom));

    for (ii = 0; ii < nind_cbuf(cto); ii++) {
	for (j = 0; j < arity_cbuf(cto); j++) {
	    ind_tuple_elt_cbuf(cto,ii,j) = 
		ind_tuple_elt_cbuf(cfrom,ii,j);
	}
    }
}

void
copy_values_cbuf (commbuffer cto, commbuffer cfrom)
{
    int ii;
    assert (nind_cbuf(cto) == nind_cbuf(cfrom));
    assert (val_type_cbuf(cto) == val_type_cbuf(cfrom));

    switch (val_type_cbuf(cto)) {
    case CBUF_INT:
	for (ii = 0; ii < nind_cbuf(cto); ii++) {
	    val4ii_cbuf(int,cto,ii) = val4ii_cbuf(int,cfrom,ii);
	}
	break;
    case CBUF_REAL:
	for (ii = 0; ii < nind_cbuf(cto); ii++) {
	    val4ii_cbuf(double,cto,ii) = val4ii_cbuf(double,cfrom,ii);
	}
	break;
    case CBUF_VOID:
	break;
    default:
	assert (0);
    }
}
