/************************************************************************
 *	Copyright (C) 1998. Bernoulli Project, Cornell University.	*
 ************************************************************************/
/*
 * Hash tables 
 */

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>

#include "rt-memory.h"
#include "hash.h"

hash_cell
make_hash_cell (hash_cell_arena *pa, int tuple_sz,
		int *tuple, hash_cell next)
{
    hash_cell_arena a;
    hash_cell c;
    int cell_size = hash_cell_size(tuple_sz);
    int i;
    int *t;

    a = *pa;
    if (0 == nleft_hc_arena(a)) {
	a = rt_alloc0(hash_cell_arena_t);
	next_hc_arena(a) = *pa;
	*pa = a;
	
	nleft_hc_arena(a) = HASH_CELL_ARENA_SIZE /cell_size;
	avail_hc_arena(a) = data_hc_area(*pa);
    }
    
    c = (hash_cell) avail_hc_arena(a);
    avail_hc_arena(a) += cell_size;
    nleft_hc_arena(a) --;
    next_hash_cell(c) = next;

    t = tuple_hash_cell(c);
    for (i = 0; i < tuple_sz; i++) {
	t[i] = tuple[i];
    }

    return c;
}

hash_table
make_hash_table (int m, int ksz, int tsz)
{
    hash_table h = rt_alloc0(hash_table_t);

    size_hash_table(h) = m;
    ksz_hash_table(h) = ksz;
    tsz_hash_table(h) = tsz;
    nelem_hash_table(h) = 0;
    table_hash_table(h) = rt_nalloc0(hash_cell, m+1);
    /* the 0th cell is used for a sentinel */

    /* init the arena */
    arena_hash_table(h) = rt_alloc0(hash_cell_arena_t);
    nleft_hc_arena(arena_hash_table(h)) = 
	HASH_CELL_ARENA_SIZE / hash_cell_size (tsz);
    avail_hc_arena(arena_hash_table(h)) = data_hc_area(arena_hash_table(h));
    
    return h;
}
    
void
free_hash_table(hash_table h)
{
    hash_cell_arena a, next;

    rt_nfree(hash_cell, 1+size_hash_table(h), table_hash_table(h));
    
    for (a = arena_hash_table(h); a; a = next) {
	next = next_hc_arena(a);
	rt_free (hash_cell_arena_t, a);
    }

    rt_free (hash_table_t, h);
}

void
search_hash_table (hash_iter it, int *tuple)
{
    hash_cell c;
    hash_table h = table_hash_iter(it);
    int ksz = ksz_hash_table(h);
    int i;
    int hf = 0;
    int *t;
    int same;

    assert (h == table_hash_iter(it));

    for (i = 0; i < ksz; i++) {
	hf = hf ^ (tuple[i]);
    }
    hf = (hf > 0) ? hf : -hf;
    hf = 1+ hf % (size_hash_table(h));

    assert ((hf > 0) && (size_hash_table(h) >= hf));

    pos_hash_iter (it) = hf;

    for (c = table_hash_table(h)[hf]; c; c = next_hash_cell(c)) {
	t = tuple_hash_cell(c);
	for (i = 0, same = 1; i < ksz; i++) {
	    same = same && (t[i] == tuple[i]);
	}
	if (same) {
	    cell_hash_iter (it) = c;
	    return;
	}
    }

    cell_hash_iter(it) = 0;
}

void
insert_hash_table (hash_iter it, int *tuple)
{
    hash_table h = table_hash_iter(it);
    int j = pos_hash_iter(it);
    int ksz = ksz_hash_table(h);
    int tsz = tsz_hash_table(h);
    int *t;
    int i;

    if (! valid_hash_iter(it)) {
	table_hash_table(h)[j] = 
	    make_hash_cell (& arena_hash_table(h),
			    tsz, tuple, table_hash_table(h)[j]);
	cell_hash_iter(it) = table_hash_table(h)[j];
	nelem_hash_table(h) ++;
    }
    else {
	t = deref_hash_iter(it);

#ifndef NDEBUG
	{
	    /* check if the key part is the same */
	    int same = 1;
	    for (i = 0; i < ksz; i++) {
		same = same && (t[i] == tuple[i]);
	    }
	    assert (same);

	}
#endif

	/* copy the data part */
	for (i = ksz; i < tsz; i++) {
	    t[i] = tuple[i];
	}
    }
}


/***********************************************************************
 *			   Iterator interface
 ***********************************************************************/


static void
advance_iter_to_valid (hash_iter it)
{
    while (!cell_hash_iter(it) && pos_hash_iter(it)) {
	-- pos_hash_iter(it);
	cell_hash_iter(it) = 
	    table_hash_table(table_hash_iter(it))[pos_hash_iter(it)];
    }
    /* if we get to the 0th position, the cell is 0 automatically */
}

hash_iter 	
open_hash_iter (hash_table h)
{
    hash_iter it = rt_alloc0(hash_iter_t);
    int m = size_hash_table(h);

    table_hash_iter(it) = h;
    pos_hash_iter(it) = m;
    cell_hash_iter(it) = table_hash_table(h)[m];

    advance_iter_to_valid(it);

    return it;
}

void
close_hash_iter (hash_iter it)
{
    rt_free (hash_iter_t, it);
}

int
valid_hash_iter (hash_iter it)
{
    return (0 != cell_hash_iter(it));
}

void
next_hash_iter (hash_iter it)
{
    cell_hash_iter(it) = next_hash_cell(cell_hash_iter(it));
    advance_iter_to_valid(it);
}

int *
deref_hash_iter(hash_iter it)
{
    if (0 != cell_hash_iter(it)) {
	return tuple_hash_cell(cell_hash_iter(it));
    }
    else {
	fprintf (stderr, "ERROR: Can not dereference hash table iter\n");
	abort ();
	return 0; /* so that gcc does not complain */
    }
}

int
deref_hash_iter_as_int (hash_iter it, int offset)
{
   int *p = deref_hash_iter (it);
   return p[offset];
}

double
deref_hash_iter_as_real (hash_iter it, int offset)
{
   int *p = deref_hash_iter (it) + offset;
   return *((double *)p);
}
