/************************************************************************
 *	Copyright (C) 1998. Bernoulli Project, Cornell University.	*
 ************************************************************************/
/* 
 * Hash tables 
 */

#ifndef _hash_h
#define _hash_h

/***********************************************************************
 * These hash tables use  tuple of integers as a key. The data field
 * can be anything of fixed length.
 *
 * As a relation, a hash table has the following structure:
 *    key --> iter  (can search by key)
 *    [] --> iter (enumeration)
 *    iter --> key, data (cheap lookup)
 *
 ***********************************************************************/


/***********************************************************************
 *                            Hash table cells
 ***********************************************************************/

/* forward declaration */
typedef struct hash_cell_arena_s *hash_cell_arena;

typedef struct hash_cell_s hash_cell_t;
typedef struct hash_cell_s *hash_cell;

struct hash_cell_s {
    hash_cell   next;
    int         tuple[1];
};

#define         hash_cell_size(tuple_sz) \
(sizeof(hash_cell_t) + ((tuple_sz)-1)*sizeof(int))

#define         next_hash_cell(c)       ((c)->next)

/* ksz == size of the key (in ints)
   dsz == size of the data field (in ints)
   */

#define         tuple_hash_cell(c)              ((c)->tuple)

hash_cell       make_hash_cell (hash_cell_arena *pa, int tuple_sz, 
                                int *tuple, hash_cell next);

/***********************************************************************
 *                 Arenas to keep hash table cells
 ***********************************************************************/

#define         HASH_CELL_ARENA_SIZE            (2 << 14)

typedef struct hash_cell_arena_s hash_cell_arena_t;

struct hash_cell_arena_s {
    hash_cell_arena next;
    int         nleft;
    char        *avail;
    char        data[HASH_CELL_ARENA_SIZE];
};

#define         ncells_hc_arena(tsz)    (HASH_CELL_ARENA_SIZE / hash_cell_size(tsz))
#define         next_hc_arena(a)        ((a)->next)
#define         nleft_hc_arena(a)       ((a)->nleft)
#define         avail_hc_arena(a)       ((a)->avail)
#define         data_hc_area(a)         ((a)->data)

/* manipulation of arenas is "inlined" in make_hash_cell */

/***********************************************************************
 *                           Hash tables
 ***********************************************************************/

typedef struct hash_table_s hash_table_t;
typedef struct hash_table_s *hash_table;

typedef struct hash_iter_s *hash_iter; /* forward decl. */

struct hash_table_s {
    int         m;      /* size of the table */
    int         n;      /* number of elements */
    int         ksz;    /* size of the keys */ 
    int         tsz;    /* size of the whole tuples */
    hash_cell   *t;     /* pointers to cells */
    hash_cell_arena a;  /* where cells are kept */
};

#define         size_hash_table(h)      ((h)->m)
#define         nelem_hash_table(h)     ((h)->n)
#define         table_hash_table(h)     ((h)->t)
#define         ksz_hash_table(h)       ((h)->ksz)
#define         tsz_hash_table(h)       ((h)->tsz)
#define         arena_hash_table(h)     ((h)->a)

hash_table      make_hash_table (int m, int ksz, int tsz);
void            free_hash_table (hash_table h);

void            search_hash_table (hash_iter it, int *tuple);
void            insert_hash_table (hash_iter it, int *tuple);

/***********************************************************************
 *                         Iter interface
 ***********************************************************************/

typedef struct hash_iter_s hash_iter_t;

struct hash_iter_s {
    hash_table  h;
    hash_cell   c;
    int         j;
};

#define         table_hash_iter(s)      ((s)->h)
#define         cell_hash_iter(s)       ((s)->c)
#define         pos_hash_iter(s)        ((s)->j)

hash_iter       open_hash_iter (hash_table h);
void            close_hash_iter (hash_iter it);
int             valid_hash_iter (hash_iter it);
void            next_hash_iter (hash_iter it);

int *           deref_hash_iter (hash_iter it);
                /* only the non-key part should be modified */

int             deref_hash_iter_as_int (hash_iter it, int offset);
double          deref_hash_iter_as_real (hash_iter it, int offset);

#endif

