/************************************************************************
 *	Copyright (C) 1998. Bernoulli Project, Cornell University.	*
 ************************************************************************/
/*
  Testing run-time system
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/time.h>

#include "hash.h"
#include "commbuffer.h"
#include "rt-memory.h"
#include "rt-comm.h"

extern double drand48();

static int
int_rand (int N)
{
    return (int)( ((double)N) * drand48() );
}

#define MAXN (1000000)

static void
random_unique (int *A, int N, int n)
{
    static char flags[MAXN];
    int i;
    int x;

    assert (N <= MAXN);

    for (i = 0; i < n; i++) {
	do { /* not very efficient, but who cares! */
	    x = int_rand(N);
	} while (flags[x]);
	flags[x] = 1;
	A[i] = x;
    }

    for (i = 0; i < n; i++) { /* clean up */
	flags[A[i]] = 0;
    }
    
}

static double
timer ()
{
    struct timeval tv;
    gettimeofday (&tv, NULL);

    return ((double)(tv.tv_sec % 1000)) + 
        (((double)(tv.tv_usec)) / 1.0e6);
}

static int
comp_int (const void *x, const void *y)
{
    return (*((int*)x)) - (*((int*)y));
}

static int
uniq (int to[], int from[], int n)
{    
    int i, j;
    if (0 == n) return 0;

    qsort ((char *) from, n, sizeof(int), comp_int);
    to[0] = from[0];

    for (i = 0, j = 1; j < n; j++) {
        if (from[j] != to[i]) {
            to[++i] = from[j];
        }
    }

    /* i+1 has the number of unique elements */
    return i+1;
}

int
test_hash_dot (int argc, char **argv)
{
    union _tuple {
	struct _e {
	    int i;
	    double v;
	} e;
	int t[sizeof(struct _e)/sizeof(int)];
    };
    
    typedef union _tuple tuple_t;


    int n = atoi(argv[1]);
    int B = atoi(argv[2]);
    int P = n / B;
    hash_table hx = make_hash_table (1023, 1, 3);
    hash_table hy = make_hash_table (1023, 1, 3);
    hash_iter it, itx, ity;
    int i, k;
    double sum;
    int *Qptr = (int *) calloc(P+1, sizeof(int));

    extern void dot_hash_bmf (int, hash_table, hash_table, double *);
    void genb_test(int, hash_table, hash_table, int *,
		   int, double *, int);
    
    assert (n % B == 0);

    for (i = 0, k = 1; i <= P; i++, k+=B){
	Qptr[i] = k;
    }

    /* tuples (i, 123.0) for  1 <= i <= 3n/4 go into the X table */
    /* tuples (i, 123.0) for  n/4 <= i <= n go into the Y table */
    it = open_hash_iter (hx);
    for (i = 1; i <= (3*n)/4; i++) {
	static tuple_t t;
	t.e.i = i;
	t.e.v = 123.0;
	search_hash_table (it, t.t);
	insert_hash_table (it, t.t);
    }
    close_hash_iter(it);
    it = open_hash_iter (hy);
    for (i = n/4; i <= n; i++) {
	static tuple_t t;
	t.e.i = i;
	t.e.v = 123.0;
	search_hash_table (it, t.t);
	insert_hash_table (it, t.t);
    }
    close_hash_iter(it);

    /* dot_hash_bmf (n, hx, hy, &sum); */
    genb_test (n, hx, hy, Qptr, P, &sum, B);
    printf ("After BMF: %lf\n", sum/(123.0 * 123.0));

    itx = open_hash_iter (hx);
    ity = open_hash_iter (hy);
    sum = 0.0;
    while (valid_hash_iter(itx)) {
	static tuple_t *pt;
	static tuple_t t;
	static double vx, vy;

	pt = (tuple_t *) deref_hash_iter (itx);
	t.e.i = pt->e.i;
	/* vx = pt->e.v; */
	vx = deref_hash_iter_as_real (itx, 1);
	search_hash_table (ity, t.t);
	if (valid_hash_iter(ity)) {
	    pt = (tuple_t *) deref_hash_iter (itx);
	    /* vy = pt->e.v; */
	    vy = deref_hash_iter_as_real (ity, 1);
	    sum += vx * vy;
	}

	next_hash_iter (itx);
    }
    close_hash_iter(itx);
    close_hash_iter(ity);

    free_hash_table (hx);
    free_hash_table (hy);

    printf ("n/4 = %d, sum = %lf\n", n/4, sum/(123.0 * 123.0));
    
    return 0;
}

int
test_hash_table (int argc, char **argv)
{
    /*
     * we feed 20*n integers in the range 0..n-1 to the table
     */
    int n = atoi(argv[1]);
    int N, Nuniq;

    /* table hold 3-tuples with the first field being the key */
    hash_table h = make_hash_table (3*n, 1, 3);

    int *arr, *arr2, *arr3;
    int i;
    int x;
    double t;
    hash_iter it;
    int data[3];
    int *ptuple;

    N = 20*n;
    arr = (int *) calloc (N, sizeof(int));
    arr2 = (int *) calloc (N, sizeof(int));
    arr3 = (int *) calloc (N, sizeof(int));
    for (i = 0; i < N; i++) {
        x = int_rand (n);
        arr[i] = x;
    }

    it = open_hash_iter(h);
    t = timer();
    for (i = 0; i < N; i++) {
        data[0] = x = arr[i];
        data[1] = 10 * x;
        data[2] = 100 * x;
        search_hash_table(it, data);
        insert_hash_table(it, data);
    }
    t = timer() - t;
    close_hash_iter(it);

    printf ("Time: %f\n", t);

    it = open_hash_iter(h);
    for (i = 0; i < N; i++) {
        data[0] = arr[i];
        search_hash_table(it, data);
        assert (valid_hash_iter(it));
    }
    close_hash_iter (it);

    for (i = 0, it=open_hash_iter(h); valid_hash_iter(it); 
         next_hash_iter(it), i++) {
        ptuple = deref_hash_iter (it);

        arr3[i] = x = ptuple[0];
        assert (ptuple[1] == 10 * x);
        assert (ptuple[2] == 100 * x);
    }

    close_hash_iter(it);

    Nuniq = uniq (arr2, arr, N);
    assert (Nuniq == i);
    qsort ((char *) arr3, i, sizeof(int), comp_int);

    for (i = 0; i < Nuniq; i++) {
        assert (arr3[i] == arr2[i]);
    }

    free (arr);
    free (arr2);
    free (arr3);
    free_hash_table (h);

    return 0;
}

int
test_commbuffer (int argc, char **argv)
{
    /* we will insert tuples 
       (M* < 10 is a random (for each q) number):
       q,i,j,k
       with i ranging from 10*q to 10*q + Mi
       j ranging from 100*q to 100*q + Mj
       k ranging from 1000*q to 1000*q + Mk

       the invariant for each tuple is (q,i,j,k):

       q = i/10 = j/100 = k/1000
       */

    int Q = atoi (argv[1]);
    hash_table h = make_hash_table (Q, 4, 4);
    hash_iter it;
    commbuffer c;
    int q, i, j, k, Mi, Mj, Mk, ii, n, nind;
    int tuple[4];
    int *p, *iptr;

    it = open_hash_iter (h);

    for (q = 0; q < Q; q++) {
        Mi = int_rand (10);
        Mj = int_rand (10);
        Mk = int_rand (10);
        for (i = 10*q; i <= 10*q + Mi; i++) {
            for (j = 100*q; j <= 100*q + Mj; j++) {
                for (k = 1000*q; k <= 1000*q + Mk; k++) { 
                    tuple[0] = q;
                    tuple[1] = i;
                    tuple[2] = j;
                    tuple[3] = k;
                    search_hash_table (it, tuple);
                    insert_hash_table (it, tuple);
                }
            }
        }
        
    }

    close_hash_iter (it);

    for (it = open_hash_iter(h); valid_hash_iter(it);
         next_hash_iter(it)) {
        p = deref_hash_iter (it);
        q = p[0];
        i = p[1];
        j = p[2];
        k = p[3];

        assert (q == i / 10);
        assert (q == j / 100);
        assert (q == k / 1000);
    }
    close_hash_iter (it);
    
    c = pack_cbuf (Q, 3, CBUF_VOID, h); /* arity only counts i,j,k */
    iptr = iptr_cbuf (c);
    for (q = 0; q < Q; q++) {
        for (ii = iptr[q]; ii < iptr[q+1]; ii++) {
            i = ind_tuple_elt_cbuf(c,ii,0);
            j = ind_tuple_elt_cbuf(c,ii,1);
            k = ind_tuple_elt_cbuf(c,ii,2);
            
            assert (q == i / 10);
            assert (q == j / 100);
            assert (q == k / 1000);
        }
    }

    nind = nind_cbuf (c);
    for (n = 10*Q; n--; ) {
        ii = int_rand (nind);
        q = find_q4ii_cbuf (c, ii);
        i = ind_tuple_elt_cbuf(c,ii,0);
        j = ind_tuple_elt_cbuf(c,ii,1);
        k = ind_tuple_elt_cbuf(c,ii,2);
            
        assert (q == i / 10);
        assert (q == j / 100);
        assert (q == k / 1000);
    }
    
    free_hash_table (h);
    free_cbuf (c);

    return 0;
}

int
test_comm (int argc, char **argv)
{
    int q, ii, nproc, me, n, i;
    int M = atoi (argv[1]);
    commbuffer c;
    commbuffer d;
    hash_table h;
    hash_iter it;
    int tuple[3];
    FILE *fp;
    char out_c_name[20], out_d_name[20];
    int *A = (int *) calloc (M, sizeof(int));

    MPI_Init(&argc,&argv);
    MPI_Comm_size(MPI_COMM_WORLD,&nproc);
    MPI_Comm_rank(MPI_COMM_WORLD,&me);
    srand48 (me);

    h = make_hash_table (2 * nproc, 3, 3);
    it = open_hash_iter (h);

    printf ("%d generating random\n", me);

    for (q = 0; q < nproc; q++) {
	n = 1 + int_rand(M);
        tuple[0] = q;
	random_unique (A, M, n);
	for (i = 0; i < n; i++) {
            tuple[1] = A[i];
            tuple[2] = 10000 - tuple[0] - tuple[1];

            /*
            printf ("inserting %d %d %d %d\n",
                    me, q, tuple[1], tuple[2]);
                    */

            search_hash_table (it, tuple);
            insert_hash_table (it, tuple);
        }

    }

    printf ("%d done random\n", me);

    close_hash_iter (it);

    c = pack_cbuf (nproc, 2, CBUF_VOID, h);

    printf ("%d starting exchange\n", me);

    d = exchange_cbuf (me, nproc, c);

    printf ("%d done exchange\n", me);

    verify_cbuf(me, c);
    verify_cbuf(me, d);

    printf ("%d done verifying\n", me);

    sprintf (out_c_name, "out_c.%d", me);
    fp = fopen (out_c_name, "w");
    assert (fp);

    for (q = 0; q < nproc; q++) {
        for (ii = iptr_cbuf(c)[q]; ii < iptr_cbuf(c)[q+1]; ii++) {
            fprintf (fp, "%d %d %d %d\n",
                     me, q, 
                     ind_tuple_elt_cbuf(c,ii,0),
                     ind_tuple_elt_cbuf(c,ii,1));
        }
    }
    fclose (fp);

    printf ("%d done writing c\n", me);

    sprintf (out_d_name, "out_d.%d", me);
    fp = fopen (out_d_name, "w");
    assert (fp);

    for (q = 0; q < nproc; q++) {
        for (ii = iptr_cbuf(d)[q]; ii < iptr_cbuf(d)[q+1]; ii++) {
            fprintf (fp, "%d %d %d %d\n",
                     q, me,
                     ind_tuple_elt_cbuf(d,ii,0),
                     ind_tuple_elt_cbuf(d,ii,1));
        }
    }
    fclose (fp);

    printf ("%d done writing d\n", me);

    free_hash_table (h);
    free_cbuf (c);
    free_cbuf (d);
    free (A);

    return 0;
}

int
main (int argc, char **argv)
{
    return test_comm (argc, argv);
}






