/************************************************************************
 *	Copyright (C) 1998. Bernoulli Project, Cornell University.	*
 ************************************************************************/
/*
 * Communication routines
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "rt-memory.h"
#include "rt-comm.h"

static int exchange_tag = 0;

static commbuffer 
exchange_prepare_cbuf (int me, int nproc, commbuffer c)
{
    int *bits = rt_nalloc0(int, nproc);
    int *counts = rt_nalloc0(int, nproc);
    int *send_counts = rt_nalloc0(int, nproc);
    int *recv_counts = rt_nalloc0(int, nproc+1);
    int q;
    MPI_Request not_used;
    int rc;
    int m;
    int x;
    MPI_Status status;
    commbuffer d;

    assert (nproc == nproc_cbuf (c));

    exchange_tag++;

    for (q = 0; q < nproc; q++) {
	send_counts[q] = (iptr_cbuf(c)[q+1] - iptr_cbuf(c)[q]);
	bits[q] = (send_counts[q] > 0) ? 1 : 0;
    }

    reduce_int_sum (bits, counts, nproc);

    /* send counts[q] to q */
    for (q = 0; q < nproc; q++) {
	if (bits[q]) {
	    rc = MPI_Isend (&send_counts[q], 1, MPI_INT, q,
			    exchange_tag, 
			    MPI_COMM_WORLD,
			    &not_used);
	    assert (rc == MPI_SUCCESS);
	}
    }

    m = counts[me];
    while (m--) {
	rc = MPI_Recv (&x, 1, MPI_INT, MPI_ANY_SOURCE, 
		       exchange_tag,
		       MPI_COMM_WORLD,
		       &status);
	assert (rc == MPI_SUCCESS);
	recv_counts[ status.MPI_SOURCE ] = x;
    }

    counts_to_iptr (nproc, recv_counts, recv_counts);
    d = make_cbuf_from_iptr (nproc, recv_counts, 
			     arity_cbuf(c), val_type_cbuf(c));

    rt_nfree (int, nproc, bits);
    rt_nfree (int, nproc, counts);
    rt_nfree (int, nproc, send_counts);
    rt_nfree (int, nproc+1, recv_counts);
    
    return d;
}


commbuffer 
exchange_cbuf (int me, int nproc, commbuffer c)
{
    commbuffer d = exchange_prepare_cbuf (me, nproc, c);

    send_cbuf (me, nproc, c, exchange_tag);
    recv_cbuf (me, nproc, d, exchange_tag);
    
    return d;
}

void 
send_cbuf (int me, int nproc, commbuffer c, int tag)
{
    int q, ii, m;
    MPI_Request not_used;
    int rc;
    int tag_ind = make_user_ind_tag(tag);
    int tag_val = make_user_val_tag(tag);

    assert (nproc == nproc_cbuf (c));

    /* printf ("%d starting send\n", me); */

    /* send the indices */
    for (q = 0; q < nproc; q++) {
	ii = iptr_cbuf(c)[q];
	m = iptr_cbuf(c)[q+1] - ii;
	if (0 < m) {
	    rc = MPI_Isend (ind_tuple_cbuf(c,ii),
			    m * arity_cbuf(c), MPI_INT,
			    q,
			    tag_ind,
			    MPI_COMM_WORLD,
			    &not_used);
	    assert (rc == MPI_SUCCESS);
	}
    }

    /* send the values */
    if (val_type_cbuf(c) != CBUF_VOID) {
	for (q = 0; q < nproc; q++) {
	    ii = iptr_cbuf(c)[q];
	    m = iptr_cbuf(c)[q+1] - ii;
	    if (0 < m) {
		if (val_type_cbuf(c) == CBUF_REAL) {
		    rc = MPI_Isend (&typed_vals_cbuf(double,c)[ii],
				    m, 
				    MPI_DOUBLE,
				    tag_val,
				    q,
				    MPI_COMM_WORLD,
				    &not_used);
		    assert (rc == MPI_SUCCESS);
		}
		else if (val_type_cbuf(c) == CBUF_INT) {
		    rc = MPI_Isend (&typed_vals_cbuf(double,c)[ii],
				    m,
				    MPI_INT,
				    tag_val,
				    q,
				    MPI_COMM_WORLD,
				    &not_used);
		    assert (rc == MPI_SUCCESS);
		}
		else {
		    fprintf (stderr, "Invalid value type\n");
		    abort ();
		}
	    }
	}
    }

    /* printf ("%d done starting send\n", me); */

}

void 
recv_cbuf (int me, int nproc, commbuffer c, int tag)
{
    int q, i, ii, m, rc;
    int cnt = count_non_empty_procs (c);
    int cnt_comm;
    int tag_ind = make_user_ind_tag(tag);
    int tag_val = make_user_val_tag(tag);
    
    MPI_Request *rqs;
    MPI_Status *stats;
    
    assert (nproc == nproc_cbuf (c));

    cnt_comm = (val_type_cbuf(c) != CBUF_VOID) ? 2*cnt : cnt;
    rqs = rt_nalloc0 (MPI_Request, cnt_comm);
    stats = rt_nalloc0 (MPI_Status, cnt_comm);

    /* post recvs for the indices */
    i = 0;
    for (q = 0; q < nproc; q++) {
	ii = iptr_cbuf(c)[q];
	m = iptr_cbuf(c)[q+1] - ii;
	if (0 < m) {
	    rc = MPI_Irecv (ind_tuple_cbuf(c,ii),
			    m * arity_cbuf(c), 
			    MPI_INT,
			    q,
			    tag_ind,
			    MPI_COMM_WORLD,
			    &rqs[i]);
	    assert (rc == MPI_SUCCESS);
	    i++;
	}
    }

    /* post the receives for the values */
    if (val_type_cbuf(c) != CBUF_VOID) {
	for (q = 0; q < nproc; q++) {
	    ii = iptr_cbuf(c)[q];
	    m = iptr_cbuf(c)[q+1] - ii;
	    if (0 < m) {
		if (val_type_cbuf(c) == CBUF_REAL) {
		    rc = MPI_Irecv (&typed_vals_cbuf(double,c)[ii],
				    m,
				    MPI_DOUBLE,
				    tag_val,
				    q,
				    MPI_COMM_WORLD,
				    &rqs[i]);
		    assert (rc == MPI_SUCCESS);
		    i++;
		}
		else if (val_type_cbuf(c) == CBUF_INT) {
		    rc = MPI_Irecv (&typed_vals_cbuf(double,c)[ii],
				    m,
				    MPI_INT,
				    tag_val,
				    q,
				    MPI_COMM_WORLD,
				    &rqs[i]);
		    assert (rc == MPI_SUCCESS);
		    i++;
		}
		else {
		    fprintf (stderr, "Invalid value type\n");
		    abort ();
		}
	    }
	}
    }

    assert (i == cnt_comm);

    /* printf ("%d waiting for recv\n", me); */

    /* not wait for all receives to complete */
    rc = MPI_Waitall (cnt_comm, rqs, stats);
    assert (rc == MPI_SUCCESS);

    /* printf ("%d done waiting for recv\n", me); */

#ifndef NDEBUG
    /* check if all messages arrived OK */
    i = 0;
    for (q = 0; q < nproc; q++) {
	int m2;

	ii = iptr_cbuf(c)[q];
	m = iptr_cbuf(c)[q+1] - ii;
	if (0 < m) {
	    assert (q == stats[i].MPI_SOURCE);
	    assert (tag_ind == stats[i].MPI_TAG);
	    MPI_Get_count(&stats[i], MPI_INT, &m2);
	    assert (m * arity_cbuf(c) == m2);
	    i++;
	}
    }

    /* post the receives for the values */
    if (val_type_cbuf(c) != CBUF_VOID) {
	int m2;
	for (q = 0; q < nproc; q++) {
	    ii = iptr_cbuf(c)[q];
	    m = iptr_cbuf(c)[q+1] - ii;
	    if (0 < m) {
		if (val_type_cbuf(c) == CBUF_REAL) {
		    assert (q == stats[i].MPI_SOURCE);
		    assert (tag_val == stats[i].MPI_TAG);
		    MPI_Get_count(&stats[i], MPI_DOUBLE, &m2);
		    assert (m == m2);
		    i++;
		}
		else if (val_type_cbuf(c) == CBUF_INT) {
		    assert (q == stats[i].MPI_SOURCE);
		    assert (tag_val == stats[i].MPI_TAG);
		    MPI_Get_count(&stats[i], MPI_INT, &m2);
		    assert (m == m2);
		    i++;
		}
		else {
		    fprintf (stderr, "Invalid value type\n");
		    abort ();
		}
	    }
	}
    }
#endif
    
    rt_nfree(MPI_Request, cnt_comm, rqs);
    rt_nfree(MPI_Status, cnt_comm, stats);
}


