(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
(************************************************************************)
(* Preamble								*)
(************************************************************************)

open Ssisl
open Containers

open Bmf;;
open Bmf_util;;
open Bmf_query;;

module WS = Wff.Simple

(************************************************************************)
(* Misc.								*)
(************************************************************************)

let true_e = Expr_bool(true);;
let false_e = Expr_bool(false);;

module Aref_element =
  struct
    type t = string * bmf_expr list
    let equal = (=)
    let hash = Containers.hash
  end

module Aref_hashtbl = Hashtbl(Aref_element)

(************************************************************************)
(* Entry points								*)
(************************************************************************)

let rec sparsity_analysis_procs (decls : bmf_ident list) = (
  List.rev (List.map sparsity_analysis_proc (List.rev decls))
    : bmf_ident list)

and sparsity_analysis_proc (decl : bmf_ident) = (
  match decl with
    Ident_valid(name,typ,Ident_proc(Frame(args),res,body),ann) ->
      walk_outer_type typ;
      walk_outer_ann ann;
      let array_info =
	add_arrays_to_map (walk_outer_ident res)
	  (add_arrays_to_map (List'.flat_map walk_outer_ident args)
	     String_map.empty)
      in	
      Ident_valid(name,typ,
		  Ident_proc(Frame(args),res,
			     walk_outer_stmt array_info 
			       (Expr_bool(true)) body),
		  ann)
  | _ -> decl
	: bmf_ident)

(************************************************************************)
(* The ``Outer'' walkers.						*)
(************************************************************************)

(* walk_outer_stmt looks for do loops *)

and walk_outer_stmt 
    (array_info: (bool * scalar_type) String_map.t)
    (assertion: bmf_expr) (s: bmf_stmt) =
  ((let s0 = walk_outer_stmt array_info assertion
  and e = walk_outer_expr array_info
  in let es = List.map e
  in
  (match s with
    Stmt_assign(lhs,rhs) -> Stmt_assign(e lhs, e rhs)
  | Stmt_seq(first,second) -> Stmt_seq(s0 first,s0 second)
  | Stmt_call(f,res,args) -> Stmt_call(e f, e res, es args)
  | Stmt_for(i,lb,ub,step,body) ->
      assert (walk_outer_ident i ==  [])
          (* sparse directive in for index var?!? *);
      Stmt_for(i, e lb, e ub, e step, s0 body)
  | Stmt_while(pred,body) -> Stmt_while(e pred,s0 body)
  | Stmt_if(pred,thn,els) -> Stmt_if(e pred,s0 thn,s0 els)
  | Stmt_let(Frame(ids),body) ->
      let new_array_info =
	add_arrays_to_map (List'.flat_map walk_outer_ident ids)
	  array_info
      in
      Stmt_let(Frame(ids), 
	       walk_outer_stmt new_array_info assertion body)
  | Stmt_do(Expr_set(Frame(ids),pred),body) ->
      do_stmt2querry array_info assertion s
  | Stmt_do(_,_) -> failwith "stmt_do without expr_set"
  | Stmt_nop -> s
  | Stmt_directive(ann,s2) -> 
      try
      	let cell_list = annotation_2_cell_list ann in
	let assertion' = 
	  match List.assoc "query_assertion" cell_list with
	    An_expr(x) -> x
	  | _ -> raise Not_found
	in let cell_list' = 
	  
	    (List'.remove_if 
	       (fun (tag,val0) -> tag = "query_assertion") 
	       cell_list)
	in let s2' = walk_outer_stmt array_info assertion' s2 in
	if cell_list' = [] then
	  s2'
	else
	  Stmt_directive(cell_list_2_annotation cell_list',s2')
      with Not_found ->
	Stmt_directive(ann, s0 s2))
    ) : bmf_stmt)

(* walk_outer_expr:
   o looks for sparse array accesses within stmts which aren''t do loops
     and signals an error. 
   o Makes all weak refs to dense arrays strong.
   *)

and walk_outer_expr (array_info: (bool * scalar_type) String_map.t) 
    (ex: bmf_expr) = (
  let e = walk_outer_expr array_info
  in let es = List.map e
  in
  match ex with
    (Expr_int(_)|Expr_real(_)|Expr_string(_)|Expr_bool(_)|
    Expr_void|Expr_var(_)) -> ex
  | Expr_op(op,args) -> Expr_op(op, es args)
  | Expr_forall(Frame(ids),pred) -> 
      assert (List'.flat_map walk_outer_ident ids == [])
		(* sparse array directives inside forall idents?!? *);
      Expr_forall(Frame(ids), e pred)
  | Expr_exist(Frame(ids),pred) ->
      assert (List'.flat_map walk_outer_ident ids == [])
		(* sparse array directives inside exists idents?!? *);
      Expr_exist(Frame(ids), e pred)
  | Expr_set(Frame(ids),pred) ->
      assert (List'.flat_map walk_outer_ident ids == [])
		(* sparse array directives inside set idents?!? *);
      Expr_set(Frame(ids), e pred)
  | Expr_access(acc,Expr_var(a),nam,args) ->
      let (sparse_p,_) = String_map.find a array_info 
      in let _ = 
      	assert ((not (sparse_p)))
	  (* expr_access outside of an stmt_do! *);
      and acc' = 
	if not sparse_p && acc = Access_weak then Access_strong else acc
      in 
      Expr_access(acc',Expr_var(a),nam, es args)
  | Expr_access(_,_,_,_) ->
      failwith "Ill-formed expr_access"
: bmf_expr)

(* walk_outer_ident looks for ``sparse'' directives on ident decls. *)

and walk_outer_ident (id: bmf_ident) = (
  match id with
    Ident_valid(name,Type_array(atyp),Ident_var(_),ann) ->
      (* *ARRAY* this needs to be fixed *)
      let rec recurs atyp =
	match atyp with
	  Array_bb(dimens,base_typ,_,_) -> [name,true,base_typ]
	| Array_dense(_,_,_,base_typ) -> [name,false,base_typ]
	| Array_view(_,_,_,base_atyp) -> recurs base_atyp
      in recurs atyp
  | Ident_valid(name,typ,stuff,ann) ->
      walk_outer_type typ;
      walk_outer_ann ann;
      (match stuff with
	(Ident_var(_)|Ident_bind) -> ()
      | Ident_proc(_,_,_) ->
	  failwith "walk_outer_ident: nested proc?!?");
      []
    | Ident_null -> []
: (string * bool * scalar_type) list)

and add_arrays_to_map (lst : (string * bool * scalar_type) list)
	(m : (bool * scalar_type) String_map.t) =
    List.fold_left
	(fun m (name,sparse_p,base_ty) ->
	    String_map.add name (sparse_p,base_ty) m)
	m lst

(* walk_outer_type traverses type and dispatches to the other walkers. *)

and walk_outer_type (ty: bmf_type) = (
    match ty with
      (Type_scalar(_)|Type_unknown) ->
	()
    | Type_proc(args_ty,res_ty) ->
	List.iter walk_outer_type args_ty;
	walk_outer_type res_ty
    | Type_ann(ann) -> walk_outer_ann ann
    | Type_ptr(btyp) -> walk_outer_type btyp
    | Type_array(atyp) -> 
	let rec recurs atyp =
	  match atyp with
	    Array_bb(dim,base_sc,bb_name,bb_args) ->
	      List.iter
		(fun (tag,val0) -> walk_outer_ann val0) bb_args
	  | Array_dense(lb,ubs,maj,base_typ) -> ()
	  | Array_view(out_frm,in_frm,constr,base_atyp) ->
	      recurs base_atyp
	in
	recurs atyp
: unit)

(* walk_outer_ann traverses annotations and dispatches to the other
    walkers. *)

and walk_outer_ann (ann: annotation) = (
    match ann with
      An_type(ty) -> walk_outer_type ty
    | An_stmt(_) -> failwith "walk_outer_ann: an_stmt encountered"
    | An_expr(e) -> () (* walk_outer_expr e *)
    | An_string(_) -> ()
    | An_list(lst) -> List.iter walk_outer_ann lst
: unit)

(************************************************************************)
(* Process a single do stmt.						*)
(************************************************************************)

and do_stmt2querry (array_info: (bool * scalar_type) String_map.t) 
    (assertion : bmf_expr) (st : bmf_stmt) = (
    let (ids,loop_pred,body) =
      match st with
	Stmt_do(Expr_set(Frame(ids),pred),body) -> (ids,pred,body)
      |	_ -> assert false
    in
    assert (List'.flat_map walk_outer_ident ids == [])
	(* sparse array directives inside set idents?!? *);
    (* refs_htbl maps from refs, A[F(i,j)], to the flag and value idents,
	("b_A","v_A") *)
    let refs_htbl = Aref_hashtbl.create 101
    (* The loop pred and the new body, after sparsity analysis. *)
    in let (body_pred,new_body) =
        guard_of_stmt array_info refs_htbl body
    (* The strong references, dense or assertions. *)
    and strong_refs = asserts_of_stmt array_info refs_htbl body
    (* "Gist" the sparsity guard from the guards of the body. *)
    in let new_body = perform_gists (bmf_to_wff body_pred) new_body
    (* The names of the index variables *)
    in let index_names =
	List.map
	    (fun id ->
		match id with
		  Ident_valid(name,Type_scalar(Type_int),Ident_bind,_) ->
		    name
		| _ -> failwith "Unexpected set id decl.")
	    ids
    (* The index variables *)
    in let index_vars =	List.map (fun name -> (name,Qv_loop)) index_names
    (* The names of the value and flag bindings introduced *)
    in let (val_names,flag_names) =
	Aref_hashtbl.fold
	    (fun (a,f) (v_A,b_A) (val_names,flag_names) ->
		(v_A::val_names, b_A::flag_names))
	    refs_htbl ([],[])
    (* Declarations of the value and flag bindings *)
    and (value_vars,flag_vars) =
	Aref_hashtbl.fold
	    (fun (a,f) (v_A,b_A) (value_vars,flag_vars) ->
		let (_,base_typ) = String_map.find a array_info
		in let v_A_var = (v_A,Qv_value(base_typ))
		and b_A_var = (b_A,Qv_flag)
		in
		    if base_typ = Type_bool then (
			assert (v_A = b_A) (* v_A != b_A ?!? *);
			(v_A_var::value_vars,flag_vars)
		    ) else (
			(v_A_var::value_vars,b_A_var::flag_vars)
		    ))
	    refs_htbl ([],[])
    (* (A,F,v_A,b_A,P_A) bindings *)
    in let bindings =
	Aref_hashtbl.fold
	    (fun (a,f) (v_A,b_A) bindings ->
		let (sparse_p,_) = String_map.find a array_info
		in let _ = () (* fixme: Compute strength here? *)
		in let binding =
		    {qb_rel=a; qb_acc=f; qb_value=v_A; qb_flag=b_A}
		in
		    binding::bindings)
	    refs_htbl []
    in let new_assertion = massage_assertion refs_htbl assertion
    in let all_vars = index_vars @ value_vars @ flag_vars
    in let q = { 
	query_vars = all_vars;
	query_bindings = bindings;
	query_used = List.map fst all_vars;
        query_bounds = loop_pred;
	query_pred = bmf_to_wff body_pred;
        query_assertion = bmf_to_wff new_assertion;
	query_body = new_body
    }
    in let result = query2stmt q
    in
	stmt2query result;
	result

: bmf_stmt)

(* Convert the assertion that appeared in the query_assertion directive.
   This involves replacing all references, A[i,j], in the assertion with
   the appropriate flag, b_A. *)

and massage_assertion refs_htbl ex =
  let e = massage_assertion refs_htbl in
  let es = List.map e in
  match ex with
    Expr_access(_,Expr_var(a),"",f) ->
      (try
	let (_,b_a) = Aref_hashtbl.find refs_htbl (a,f) in
	Expr_var(b_a)
      with Not_found ->
	failwith "Ill-formed query_assertion directive.")
  | Expr_bool(_) -> ex
  | Expr_op(op,exs) -> Expr_op(op, es exs)
  | _ ->
      failwith "Ill-formed query_assertion directive."

(************************************************************************)
(* Walking the body computing the sparsity guards			*)
(************************************************************************)

and guard_of_stmt (array_info: (bool * scalar_type) String_map.t)
	(refs_htbl: (string * string) Aref_hashtbl.t) 
	(s: bmf_stmt) = (
    match s with
      Stmt_assign(lhs,rhs) -> 
	guard_of_assign array_info refs_htbl lhs rhs
    | Stmt_nop -> (Expr_bool(false),Stmt_nop)
    | Stmt_seq(first,second) ->
        let (pred_first,new_first) = 
	    guard_of_stmt array_info refs_htbl first
	and (pred_second,new_second) = 
	    guard_of_stmt array_info refs_htbl second
	in
	    (make_or [pred_first;pred_second],
	     Stmt_seq(new_first,new_second))
    | Stmt_let(frm,body) ->
	let (pred,new_body) = 
	    guard_of_stmt array_info refs_htbl body
	in
	    (pred, Stmt_let(frm,new_body))
    | Stmt_if(be,body1,body2) ->
	guard_of_if array_info refs_htbl be body1 body2
    | (Stmt_call(_,_,_)|Stmt_for(_,_,_,_,_)|Stmt_while(_)|
      Stmt_do(_,_)|Stmt_directive(_,_)) ->
	failwith "guard_of_stmt: can't deal with stmt"
: bmf_expr * bmf_stmt)

(************************************************************************)
(* Process an if stmt in a do loop					*)
(************************************************************************)

and guard_of_if (array_info: (bool * scalar_type) String_map.t)
    (refs_htbl: (string * string) Aref_hashtbl.t) 
    (be: bmf_expr) (body1: bmf_stmt) (body2: bmf_stmt) = (
  let (pred1,new_body1) = 
    guard_of_stmt array_info refs_htbl body1
  and (pred2,new_body2) = 
    guard_of_stmt array_info refs_htbl body2
  and (be_t,be_f) = 
    guards_of_bool array_info refs_htbl be
  in
  let new_pred1 = make_and[be_t;pred1]
  and new_pred2 = make_and[be_f;pred2]
  in
  let final_body1 = 
    match new_pred1 with
      Expr_bool(false) -> Stmt_nop
    |	_ -> Stmt_if(new_pred1,bmf_stmt_if_then new_body1,Stmt_nop)
  and final_body2 = 
    match new_pred2 with
      Expr_bool(false) -> Stmt_nop
    |	_ -> Stmt_if(new_pred2,bmf_stmt_if_then new_body2,Stmt_nop)
  in
  (make_or[new_pred1;new_pred2],
   Stmt_if(rewrite_expr refs_htbl be,
	   final_body1,
	   final_body2))
 : bmf_expr * bmf_stmt)

(************************************************************************)
(* Process a single assignment stmt in a do loop			*)
(************************************************************************)

and guard_of_assign (array_info: (bool * scalar_type) String_map.t)
	(refs_htbl: (string * string) Aref_hashtbl.t) 
	(lhs: bmf_expr) (rhs: bmf_expr) = (

    let rec init_expr e =
	match e with
	  Expr_op(_,args) -> List.iter init_expr args
	| Expr_access(acc,Expr_var(a),"",f) 
		when acc = Access_strong || acc = Access_weak ->
	    (try
		Aref_hashtbl.find refs_htbl (a,f);
		()
	    with Not_found ->
		Aref_hashtbl.add refs_htbl (a,f)
		    (Misc.gensym ("v_"^a),Misc.gensym ("b_"^a)))
	| (Expr_int(_)|Expr_real(_)|Expr_var(_)) -> ()
	| _ -> failwith "init_expr: unexpected expr."

    in let rec ne e =
	match e with
	  _ when e = lhs -> false_e
	| Expr_op(Op_plus,args) ->
	    (match args with
	      [] -> failwith "Shouldn't happen"
	    | [arg] -> ne arg
	    | _ ->
		let rec loop prefx lst =
		    match lst with
		      [] -> []
		    | arg::suffx ->
			(make_or
			    ((List.map (nz array_info refs_htbl) prefx) @ [ne arg] @ (List.map (nz array_info refs_htbl) suffx)))
			::(loop (prefx@[arg]) suffx)
		in
		    make_and (loop [] args))
	| Expr_op(Op_neg,_) -> true_e
	| _ -> true_e

    in let update_op lhs rhs =
	match rhs with
	  Expr_op(op,args) ->
	    let all_args = bmf_expr_args_by_op op rhs
	    in
		if List.mem lhs all_args then
		    (op,Expr_op(op,List'.difference all_args [lhs]))
		else
		    (* Not a valid update operator. *)
		    (Op_eq,rhs)
	  (* Not a valid update operator. *)
	| _ -> (Op_eq,rhs)
	
    in
	init_expr lhs;
	init_expr rhs;	
	let (op,rhs_for_nz) = update_op lhs rhs
	in let pred =
	    if op = Op_plus then
		nz array_info refs_htbl rhs_for_nz
	    else
		make_and [ne rhs; make_or [nz array_info refs_htbl lhs; nz array_info refs_htbl rhs]]
	in
	    (pred,
	     Stmt_if(pred,
		Stmt_assign(rewrite_expr refs_htbl lhs, rewrite_expr refs_htbl rhs),
		Stmt_nop))

: bmf_expr * bmf_stmt)

(************************************************************************)
(* Process a boolean expression						*)
(************************************************************************)

and guards_of_bool (array_info: (bool * scalar_type) String_map.t)
    (refs_htbl: (string * string) Aref_hashtbl.t) 
    (be: bmf_expr) = (
  match be with
    Expr_bool(x) -> (Expr_bool(x),Expr_bool(not x))
  | Expr_op(Op_not,_) ->
      let (t,f) = guards_of_bool array_info refs_htbl (bmf_expr_op_arg be)
      in (f,t)
  | Expr_op(Op_and,_) ->
      let (t1,f1) = guards_of_bool array_info refs_htbl (bmf_expr_op_arg1 be)
      and (t2,f2) = guards_of_bool array_info refs_htbl (bmf_expr_op_arg2 be)
      in 
          (make_and[t1;t2],make_or[f1;f2])  
  | Expr_op(Op_or,_) ->
      let (t1,f1) = guards_of_bool array_info refs_htbl (bmf_expr_op_arg1 be)
      and (t2,f2) = guards_of_bool array_info refs_htbl (bmf_expr_op_arg2 be)
      in 
          (make_or[t1;t2],make_and[f1;f2])  
  | Expr_op(Op_leq,_) -> 
      let nz1 = nz array_info refs_htbl (bmf_expr_op_arg1 be)
      and nz2 = nz array_info refs_htbl (bmf_expr_op_arg2 be)
      in 
          (true_e,make_or[nz1;nz2])
  | Expr_op(Op_eq,_) -> 
      let nz1 = nz array_info refs_htbl (bmf_expr_op_arg1 be)
      and nz2 = nz array_info refs_htbl (bmf_expr_op_arg2 be)
      in 
          (true_e,make_or[nz1;nz2])
  | Expr_var(_) -> (true_e,true_e)
  | _ -> failwith "guards_of_bool: not a boolean expr.")

(************************************************************************)
(* Check if an arithmetic expression can be non-zero			*)
(************************************************************************)

and nz (array_info: (bool * scalar_type) String_map.t)
    (refs_htbl: (string * string) Aref_hashtbl.t) 
    (e: bmf_expr) = (
  match e with
    Expr_op(Op_plus,args) -> make_or (List.map (nz array_info refs_htbl) args)
  | Expr_op(Op_times,args) -> make_and (List.map (nz array_info refs_htbl) args)
  | Expr_op(Op_div,_) -> nz array_info refs_htbl (bmf_expr_op_arg1 e)
  | Expr_op(Op_neg,_) -> nz array_info refs_htbl (bmf_expr_op_arg e)
  | Expr_op(Op_sqrt,_) -> nz array_info refs_htbl (bmf_expr_op_arg e)
  | Expr_access(_,Expr_var(a),_,f) ->
      (try
	let (sparse_p,_) = String_map.find a array_info
	and (_,b_A) = Aref_hashtbl.find refs_htbl (a,f)
	in
	if sparse_p then
	  Expr_var(b_A)
	else
	  true_e
      with Not_found ->
	failwith "nz: No array_info ?!?")
  | Expr_int(x) -> Expr_bool(x<>0)
  | Expr_real(x) -> Expr_bool(x<>0.0)
  | Expr_var(_) -> true_e
  | _ -> failwith "nz: unexpected expr.")


(************************************************************************)
(* Replace array refs with either value holders or "presence" holders.  *)
(************************************************************************)

and rewrite_expr refs_htbl e =
  let zero_p x =
    match x with
      (Expr_int(0) | Expr_real(0.0)) -> true
    |	 _ -> false
  in
  match e with
    Expr_op(Op_eq,[Expr_access(_,Expr_var(a),_,f);arg2]) 
      when zero_p arg2 ->
	let (_,b_A) = Aref_hashtbl.find refs_htbl (a,f)
	in Expr_op(Op_not,[Expr_var(b_A)])
  | Expr_op(Op_eq,[arg1;Expr_access(_,Expr_var(a),_,f)]) 
      when zero_p arg1 ->
	let (_,b_A) = Aref_hashtbl.find refs_htbl (a,f)
	in Expr_op(Op_not,[Expr_var(b_A)])
  | Expr_op(op,args) -> 
      Expr_op(op,List.map (rewrite_expr refs_htbl) args)
  | Expr_access(_,Expr_var(a),_,f) ->
      let (v_A,_) = Aref_hashtbl.find refs_htbl (a,f)
      in
      Expr_var(v_A)
  | (Expr_int(_)|Expr_real(_)|Expr_var(_)) -> e
  | _ -> failwith "rewrite_expr: unexpected expr."

(************************************************************************)
(* Gist the sparsity guard out of the guards in a stmt tree		*)
(************************************************************************)

and perform_gists body_pred_wff stmt =
  match stmt with
    Stmt_if(pred,thn,els) ->
      (try
      	let pred_wff = bmf_to_wff pred
	in let pred_wff' = Wff.Simple.gist pred_wff body_pred_wff
	and thn' = 
	  perform_gists (Wff.Simple.Conjunct(body_pred_wff,pred_wff)) 
	    thn
	and els' =
	  perform_gists (Wff.Simple.Conjunct
			   (body_pred_wff,
			    Wff.Simple.Negation(pred_wff)))
	    els
	in let pred' = wff_to_bmf pred_wff'
	in
	(match pred' with
	  Expr_bool(true) -> thn'
	| Expr_bool(false) -> els'
	| _ -> Stmt_if(pred',thn',els'))
      with Not_a_wff(_) ->
	Stmt_if(pred, 
		perform_gists body_pred_wff thn,
		perform_gists body_pred_wff els))
  | _ -> stmt

(************************************************************************)
(* Computing the strength assertions for a statement tree.		*)
(************************************************************************)

and asserts_of_stmt (array_info: (bool * scalar_type) String_map.t)
	(refs_htbl: (string * string) Aref_hashtbl.t) 
	(s: bmf_stmt) = ((

    let e = asserts_of_expr array_info refs_htbl
    and s0 = asserts_of_stmt array_info refs_htbl
    in
	match s with
	  Stmt_assign(lhs,rhs) ->
	    String_set.union (e lhs) (e rhs)
	| Stmt_nop -> String_set.empty
	| Stmt_seq(first,second) ->
	    String_set.union (s0 first) (s0 second)
	| Stmt_let(frm,body) -> s0 body
	| Stmt_if(test,_,_) -> e test
	| (Stmt_call(_,_,_)|Stmt_for(_,_,_,_,_)|Stmt_while(_)|
		Stmt_do(_,_)|Stmt_directive(_,_)) ->
	    failwith "asserts_of_stmt: can't deal with stmt"
	    
) : (* strong refs *) String_set.t)

and asserts_of_expr (array_info: (bool * scalar_type) String_map.t)
	(refs_htbl: (string * string) Aref_hashtbl.t) 
	(e: bmf_expr) = ((

    let e0 = asserts_of_expr array_info refs_htbl
    in let es es = List.fold_left String_set.union String_set.empty (List.map e0 es)
    in
	match e with
	  Expr_access(kind,Expr_var(a),"",f) ->
	    let (sparse_p,_) = String_map.find a array_info
	    and (_,b_A) = Aref_hashtbl.find refs_htbl (a,f)
	    in
		if ((not (sparse_p))) || (kind = Access_strong)  then
		    (* Refs to dense arrays and assertion references are
		       strong. *)
		    String_set.add b_A String_set.empty
		else
		    String_set.empty
	| Expr_op(_,args) -> es args
	| (Expr_int(_)|Expr_real(_)|Expr_string(_)|Expr_bool(_)|
		Expr_void|Expr_var(_)) ->
	    String_set.empty
	| (Expr_access(_,_,_,_)|Expr_forall(_,_)|Expr_exist(_,_)|
		Expr_set(_,_)) ->
	    failwith "asserts_of_expr: unimplemented expr kind."

) : (* strong refs *) String_set.t)

(************************************************************************)
(*			  Command line processing			*)
(************************************************************************)

let _ =
  register_switch "-qf"
    "-qf (bmf) -- Query formulation."
    (fun top_decls args ->
      let name = Bsct_toplevel.get_arg_name (List.hd args)
      and (options,param,args') = Bsct_toplevel.parse_arg false args
      in let _ = Bsct_toplevel.ignore_options options name
      in let top_decls' = sparsity_analysis_procs top_decls
      in (top_decls',args'))

(************************************************************************)
(* End of File								*)
(************************************************************************)
