(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Printf;;

open Alpha;;

open Ssisl;;
open Containers

open Bmf;;
open Bmf_util;;
open Bmf_query;;
open Bmf_join;;
open Bmf_linear;;
open Bmf_parser_aux;;
open Bb;;
open Bb_decl;;

open Scheduler_types;;
open Scheduler_hier;;
open Scheduler_linear;;
open Scheduler_stages;;

module WS = Wff.Simple

exception No_solutions;;

exception No_avail_tests;;
exception No_determined_enums;;
exception No_avail_joins;;
exception No_avail_enums;;
exception No_determined_lookups;;
exception No_joinable_lookups;;
exception No_unjoinable_lookups;;

let pending_joinable_field_p gs ss f =
  (joinable_field_p gs f) & not (determined_field_p gs ss f)

(************************************************************************)
(* make_jm_of_field constructs a map from a joinable field name to the
   ready hterm that can be used to access it.*)
(* todo: jm_of_field is broken by design. In a sense, what it does is
   determine the hterm that will be used for each field. This is
   _wrong_. The heuristic should determine this. *)
(************************************************************************)

type join_mthd_type =
| Jm_no_such		(* No ready hterm *)
| Jm_loop of string	(* Name of the loop variable *)
| Jm_term
    of hterm_kind	(* methods for joining the field *)
;;

let make_jm_of_field gs	ss = 
  (* Construct a map from loop fields -> Jm_loop(_) values. *)
  let loop_fields_m =
    List.fold_left
      (fun m loop_name ->
	let loop_field = Field_loop(loop_name) in
	if pending_joinable_field_p gs ss loop_field then
	  Fn_map.add loop_field (Jm_loop(loop_name)) m
	else
	  m)
      Fn_map.empty
      (collect_loop_vars (global_state_query gs))
  in 
  (* Compare two hterms and return the "better" of the two.
     Totally heuristic. *)
  let better_hterm hterm1 hterm2 =
    match (hterm1,hterm2) with
      (Hterm_lookup _,Hterm_lookup _) ->
	assert false (* todo: not implemented. *)
    | (Hterm_lookup _, _) -> hterm1
    | (_, Hterm_lookup _) -> hterm2
    | (Hterm_enum_searching _,Hterm_enum_searching _) ->
	assert false (* todo: not implemented. *)
    | (Hterm_enum_searching _, _) -> hterm1
    | (_, Hterm_enum_searching _) -> hterm2
    | (Hterm_enum_indexing _,Hterm_enum_indexing _) ->
	assert false (* todo: not implemented. *)
    | (Hterm_enum_indexing _, _) -> hterm1
    | (_, Hterm_enum_indexing _) -> hterm2
    | (Hterm_enum_simple _,Hterm_enum_simple _) ->
	(* todo: not implemented. *)
	if hterm1 < hterm2 then
	  hterm1
	else
	  hterm2
  in
  (* Add to this map, the name/hterm pair of the join field of each ready
     hterm. *)
  let field_to_jm_m =
    List.fold_left
      (fun m hterm ->
	let f = primary_field_of_hterm hterm in
	if pending_joinable_field_p gs ss f then
	  try
	    let v = Fn_map.find f m in
	    match v with
	      Jm_term(hterm2) ->
		Fn_map.add f (Jm_term(better_hterm hterm hterm2)) m
	    | _ -> assert false; (* Is this possible? *)
	  with Not_found ->
	    Fn_map.add f (Jm_term(hterm)) m
	else
	  m)
      loop_fields_m (step_state_ready_hterms ss)
  (* The result is a function that looks for a field name within this map,
     and returns Jm_no_such, if it is not found. *)
  in
  (fun field ->
    assert (pending_joinable_field_p gs ss field)
      (* (Failure("Not a pending joinable field.")) *);
    try
      Fn_map.find field field_to_jm_m
    with
      Not_found -> Jm_no_such)

(************************************************************************)
(* Return the equiv class for a pending join field. *)
(************************************************************************)

let find_equiv_class gs ss join_field =
  assert (pending_joinable_field_p gs ss join_field)
      (* (Failure("Not a pending joinable field.")) *);
  let ls = step_state_ls ss
  in let equiv_classes = ls.ls_equiv_classes
  in let join_index = ls.ls_index_of_field join_field
  in let rec recurs equiv_classes =
    match equiv_classes with
      [] -> assert false
	  (* Can't happen. Pending join fields must have equiv classes. *)
    | equiv_class::_ when List.mem join_index equiv_class ->
	equiv_class
    | _::equiv_classes' ->
	recurs equiv_classes'
  in
  recurs equiv_classes

let equiv_class_to_hterms jm_of_field ss equiv_class =
    let ls = step_state_ls ss
    in
	List'.flat_map
	    (fun k ->
		let field = ls.ls_index_to_field_v.(k-1)
		in let jm = jm_of_field field
		in
		    match jm with
		      Jm_term(hterm) -> [hterm]
		    | _ -> [])
	    equiv_class
;;


let call_with_backtracing verbose_p name fn gs ss =
  Tracing.wrap
    (fun () -> fn gs ss)
    name
    (fun outch -> 
      if verbose_p then
	fprintf outch "ss = %a" output_step_state ss
      else
	fprintf outch "...")
    (fun outch result -> fprintf outch "...")

(************************************************************************
The back-tracking mechanism. Apply `f' to each element of `lst', in
term. Return the result of the first application that succeeds without
throwing No_solutions. If no elements of `lst' succeed, then
throw `exn'.
************************************************************************)

let try_until_success exn lst f =
    let rec iter lst =
	match lst with
	  [] -> raise exn
	| x::lst' ->
	    (try
	      Tracing.pass
		(fun () -> f x)
		(fun ex -> ex = No_solutions)
	    with No_solutions ->
	        if !verbose then
		  begin
	            fprintf stderr "*** Retry! ***\n";
	            flush stderr
		  end;
		iter lst')
    in
	iter lst
;;

(************************************************************************
Schedule a single query.
************************************************************************)
let rec schedule_query (bbm: black_box_map) (q: query)
    (verbose_p: bool) = (
  call_with_dynamic_binding verbose verbose_p
    (fun () -> 
      Tracing.top (!verbose)
    	(fun () ->
	  schedule_query' bbm q))
: hlp_type)

and schedule_query' (bbm: black_box_map) (q: query) = (
    let gs = make_global_state bbm q
    in let initial_ss = make_step_state gs
    in let stages = preprocess gs initial_ss
    in (gs,stages)
: hlp_type)

(************************************************************************
Preprocessing

An fields whose access function is invariant with respect to the present
query needs to be scheduled first. There two cases.

The first occurs when a field is accessed by a constant expression. In this
case, the corrisponding row in $H$ will be all zeros. These rows must be
permuted to the very top of $H$ and all of these fields marked as
``determined''.

The second case occurs when a field is accessed by an invariant variable
(e.g., a symbolic). In this case, a set of rows will exist in $H$
corrisponding to this ``join''. These rows need to be permuted and marked
as determined also.

Note: we do \underline{not} have to generate anything for the plan at this
point. By simply marking these fields as determined, we will cause the
appropriate searches to be generated by the core heuristic.
************************************************************************)

and preprocess (gs: global_state_type) (ss: step_state_type) = (
  Tracing.wrap
    (fun () -> 
      let (ss',stage) = make_constant_stage gs ss
      in Stage_constant(stage,(schedule_invariants gs ss')))
    "preprocess"
    (fun outch -> fprintf outch "ss = %a" output_step_state ss)
    (fun outch result -> fprintf outch "...")
: stage_type)

(* Schedule every join term that contains an invariant field. *)
(* todo: doesn't generate a range test! *)
and schedule_invariants (gs: global_state_type) (ss: step_state_type) =
  
  let ls = step_state_ls ss in
  (* true iff field at row i is invariant *)
  let rec invar_i_p i =
    match ls.ls_index_to_field_v.(i-1) with
      Field_inv(_) -> true
    | _ -> false
  in
  
  let rec iter equiv_classes =
    match equiv_classes with
      equiv_class::_ when List.exists invar_i_p equiv_class ->
	let (ss',stage) = make_invariant_stage gs ss equiv_class
	in Stage_invariant(stage,(schedule_invariants gs ss'))
    | _::equiv_classes' -> iter equiv_classes'
    |  [] -> next_step gs ss
  in
  iter ls.ls_equiv_classes

(************************************************************************
Enums
************************************************************************)

and next_step (gs: global_state_type) (ss: step_state_type) =
  call_with_backtracing true "next_step" next_step' gs ss

and next_step' (gs: global_state_type) (ss: step_state_type) = (

    (* If everything is bound, then we are done. *)
    if all_bounded_p (step_state_bound ss)
	    (global_state_all_fields gs) then
	Stage_body(make_body_stage gs ss)

    (* If lookup terms are ready, then schedule one *)
    else if List.exists
		(fun hterm ->
		    match hterm with
		      Hterm_lookup(_) -> true
		    | _ -> false)
		(step_state_ready_hterms ss) then
	lookup_heuristic gs ss
    else
	enum_heuristic gs ss

: stage_type)


(************************************************************************
Scheduling Enum terms
************************************************************************)

and enum_heuristic (gs: global_state_type)
	(ss: step_state_type) = ((

    (**************************************************
    The heuristic for scheduling joins is as follows:
    **************************************************)

    (* Schedule a test. A test is generated when the value of a loop index
      is already determined, and it is only necessary to see that index falls
      within the range of indices of the loop. *)
    try
      Tracing.pass
	(fun () -> 
	  call_with_backtracing false "try_test_enums" 
	    try_test_enums gs ss)
	(fun ex -> ex = No_avail_tests)
    with No_avail_tests ->

    (* Schedule a determined enum. *)
    try
      Tracing.pass
	(fun () ->
	  call_with_backtracing false "try_determined_enums"
 	    try_determined_enums gs ss)
	(fun ex -> ex = No_determined_enums)
    with No_determined_enums ->

    (* Schedule a non-trivial join. Favor the join with the most number of
       relations. *)
    try
      Tracing.pass
	(fun () ->
	  call_with_backtracing false "try_join_enums"
 	    try_join_enums gs ss)
	(fun ex -> ex = No_avail_joins)
    with No_avail_joins ->

    (* Schedule an enumeration. Favor enumerations of array indices and
       offsets over vector indices and offsets. *)
    try
      Tracing.pass
	(fun () ->
	  call_with_backtracing false "try_trivial_enums"
 	    try_trivial_enums gs ss)
	(fun ex -> ex = No_avail_enums)
    with No_avail_enums ->
    (* Deadlock. No progress can be made. This case shouldn't every happen.
       *)
    raise No_solutions
    )
: stage_type)


(************************************************************************
Looking for a ready hterm that is a simple enum, and whose join field
is determined.
************************************************************************)

and try_test_enums (gs: global_state_type) (ss: step_state_type) =
    let determined_loop_hterms =
      List'.remove_if_not
	(fun hterm ->
	  match hterm with
	    Hterm_enum_simple(enum_m)
	      when determined_field_p gs ss enum_m.nam_out_field ->
		true
	  | _ -> false)
	(step_state_ready_hterms ss)
    in let do_hterm hterm =
      let (ss',stage) =
	make_determined_stage gs ss hterm
      in
      Stage_determined(stage,(next_step gs ss'))
    in let stages =
      try_until_success No_avail_tests determined_loop_hterms
	do_hterm
    in
    stages

(************************************************************************
Look for a ready enum hterm whose outfield is determined.
************************************************************************)

and try_determined_enums (gs: global_state_type) (ss: step_state_type) =
    let determined_hterms =
      List'.remove_if_not
	(fun hterm ->
	  match hterm with
	    (Hterm_enum_simple(_)|Hterm_enum_indexing(_,_)|
	    Hterm_enum_searching(_,_,_))
	      when determined_field_p gs ss (primary_field_of_hterm hterm) ->
		true
	  | _ -> false)
	(step_state_ready_hterms ss)
    in let do_hterm hterm =
      let (ss',stage) =
      	make_determined_stage gs ss hterm
      in
      Stage_determined(stage,(next_step gs ss'))
    in let stages =
      try_until_success No_determined_enums determined_hterms
	do_hterm
    in
    stages

(************************************************************************
Look for a non-trivial join to schedule. When more than one non-trivial
join is available, heuristically select one.
************************************************************************)

and try_join_enums (gs: global_state_type) (ss: step_state_type) =
    let ls = step_state_ls ss
    and jm_of_field = make_jm_of_field gs ss
    (* list of (equiv_class). *)
    in let equiv_classes = ls.ls_equiv_classes
    (* list of (equiv_class,hterms). `hterms' is the ready terms for the
       fields in the equiv_class. *)
    in let equiv_classes' =
	List.map
	    (fun equiv_class ->
	      let hterms =
		equiv_class_to_hterms jm_of_field ss equiv_class
	      in
	      (equiv_class,hterms))
	    equiv_classes
    in let non_triv_equiv_classes =
	List'.remove_if_not (fun (equiv_class,hterms) -> List.length hterms >= 2)
	    equiv_classes'
    (* list of (equiv_class,hterms,num_pred). `num_pred' is number of
       hterms whose relation appears as a conjunct of the query
       predicate. *)
    in let non_triv_equiv_classes' =
	List.map (fun (equiv_class,hterms) ->
		(equiv_class,hterms,count_pred_hterms gs hterms))
	    non_triv_equiv_classes
    (* list of (equiv_class,hterms,num_pred,num_assert). `num_assert' is
       the number of hterms whose relation will appear as a conjunct of the
       join predicate and which represents a strong reference to a sparse
       relations. *)
    in let non_triv_equiv_classes'' =
	List.map (fun (equiv_class,hterms,num_pred) ->
		(equiv_class,hterms,num_pred,
		    count_assert_hterms gs hterms))
	    non_triv_equiv_classes'
    (* Sorting the equiv classes. *)
    in let sorted_equiv_classes =
	Sort.list
	    (fun (_,hterms1,num_pred1,num_assert1)
		    (_,hterms2,num_pred2,num_assert2) ->
		(* This is the heuristic taken from the old scheduler. It
		   could probably use some work... *)

		(* 1. Prefer joins w/more hterms *)
		let num_hterms1 = List.length hterms1
		and num_hterms2 = List.length hterms2
		in
		    if num_hterms1 > num_hterms2 then
			true
		    else if num_hterms1 < num_hterms2 then
			false
		    else

		(* 2. Prefer joins w/more predicates being satisfied. *)
		if num_pred1 > num_pred2 then
		    true
		else if num_pred1 < num_pred2 then
		    false
		else

		(* 3. Prefer joins w/fewer assertion referenced being
		      tested by the join. The idea is to defer the joins in
		      the hopes that they will eventually no require
		      testing. This occurs in BlockSolve between A and
		      X. *)
		if num_assert1 < num_assert2 then
		    true
		else if num_assert1 > num_assert2 then
		    false
		else

		(* 4. Prefer joins on larger relations. *)
		let rec degree_of_hterms hterms =
		    List.fold_left
		        (fun sum hterm -> sum + (degree_of_hterm hterm))
		        0 hterms
		and degree_of_hterm hterm =
		    let r' = rel_of_hterm hterm
		    in let r = (global_state_qb_of_rel gs r').qb_rel
		    and bbm = global_state_bbm gs
		    in
			(List.length
			   ((String_map.find r bbm).bb_index_fields))
		in
		let degree1 = degree_of_hterms hterms1
		and degree2 = degree_of_hterms hterms2
		in
		    if degree1 > degree2 then
		        true
		    else if degree1 < degree2 then
		        false
		    else
		  
		(* the two joins are "equal". Pick one at random. *)
		let sorted_hterms1 = Sort.list (<=) hterms1
		and sorted_hterms2 = Sort.list (<=) hterms2
		in
		sorted_hterms1 < sorted_hterms2)
	    non_triv_equiv_classes''
    in let do_join (equiv_class,hterms,_,_) =
      let (ss',stage) = make_join_stage gs ss equiv_class hterms 
      in Stage_join(stage,(next_step gs ss'))
    in let stages =
      try_until_success No_avail_joins sorted_equiv_classes
	do_join
    in
    stages

(* Count the number of hterms whose relations appear as conjunctive terms
   of the query's predicate. *)

and count_pred_hterms gs hterms = 
    let flag_to_rel_m =
	List.fold_left
	    (fun m r' ->
		let qb = global_state_qb_of_rel gs r'
		in
		    String_map.add qb.qb_flag r' m)
	    String_map.empty
	    (global_state_all_rels gs)
    in let conjunct_rels =
	List.fold_left
	    (fun s e ->
		match e with
		  WS.Atom(flag) ->
		    (try
			let r' = String_map.find flag flag_to_rel_m
			in
			    String_set.add r' s
		    with Not_found ->
			s)
		| _ -> s)
	    String_set.empty
	    (WS.to_conjuncts ((global_state_query gs).query_pred))
    in let conjunct_hterms =
	List'.remove_if_not
	    (fun hterm ->
		String_set.mem (rel_of_hterm hterm) conjunct_rels)
	    hterms
    in
	List.length conjunct_hterms

(* Count the number of hterms whose relation will appear as a conjunct of the
   join predicate and which represents a strong reference to a sparse
   relations. *)

and count_assert_hterms gs hterms = 
  List.length
    (List'.remove_if_not
       (fun hterm ->
	 let r' = rel_of_hterm hterm
	 in let r = (global_state_qb_of_rel gs r').qb_rel
	 in is_assert_strong_p (global_state_query gs) r)
       hterms)

(************************************************************************
Find a simple enum hterm for the relation with the maximum dimension.
************************************************************************)

and try_trivial_enums (gs: global_state_type) (ss: step_state_type) =
    let mthd_to_dim hterm =
	let r' = rel_of_hterm hterm
	in let r = (global_state_qb_of_rel gs r').qb_rel
	and bbm = global_state_bbm gs
	in
	    List.length ((String_map.find r bbm).bb_index_fields)
    in let ready_enums =
      List'.remove_if_not
	(fun hterm ->
	  let field = primary_field_of_hterm hterm in
	  if determined_field_p gs ss field then
	    false
	  else
	    match hterm with
	      Hterm_enum_simple(_) -> true
	    | Hterm_enum_indexing(_,_) -> true
	    | Hterm_enum_searching(_,_,_) -> true
	    | _ -> false)
	(step_state_ready_hterms ss)
    in let ready_enums' =
	Sort.list
	    (fun h1 h2 -> mthd_to_dim h1 >= mthd_to_dim h2)
	    ready_enums
    in let do_ready_hterm hterm =
	let field = primary_field_of_hterm hterm in
	if joinable_field_p gs field then
	  let equiv_class =
	    find_equiv_class gs ss field
	  and jm_of_field = make_jm_of_field gs ss
	  in let hterms =
	    equiv_class_to_hterms jm_of_field ss equiv_class
	  in let (ss',stage) =
	    make_join_stage gs ss equiv_class hterms
	  in
	  Stage_join(stage,(next_step gs ss'))
	else
	  let (ss',stage) =
	    make_unjoinable_stage gs ss hterm
	  in
	  Stage_unjoinable(stage,(next_step gs ss'))
    in let stages =
      try_until_success No_avail_enums ready_enums' do_ready_hterm
    in
    stages

(************************************************************************
Lookups
************************************************************************)
and lookup_heuristic (gs: global_state_type)
        (ss: step_state_type) = (

    (* Schedule a lookup on a determined field. Doing the test first may
       prevent other, unnecessary work from being done. *)
    try
      Tracing.pass
	(fun () ->
	  call_with_backtracing false "try_determined_lookups"
 	    try_determined_lookups gs ss)
	(fun ex -> ex = No_determined_lookups)
    with No_determined_lookups ->

    (* Schedule an undetermined join field. *)
    try
      Tracing.pass
	(fun () ->
	  call_with_backtracing false "try_joinable_lookups"
 	    try_joinable_lookups gs ss)
	(fun ex -> ex = No_joinable_lookups)
    with No_joinable_lookups ->

    (* Schedule an unjoinable field. *)
    try
      Tracing.pass
	(fun () ->
	  call_with_backtracing false "try_unjoinable_lookups"
 	    try_unjoinable_lookups gs ss)
	(fun ex -> ex = No_unjoinable_lookups)
    with No_unjoinable_lookups ->

    (* Deadlock. No progress can be made. This case shouldn't every
       happen. *)
    raise No_solutions

: stage_type)

(************************************************************************
************************************************************************)

and try_determined_lookups (gs: global_state_type) (ss: step_state_type) =
    let determined_lookup_hterms =
	List'.remove_if_not
	    (fun hterm ->
		match hterm with
		  Hterm_lookup(lookup_m)
			when determined_field_p gs ss
			    (lookup_m.nam_out_field) ->
		    true
		| _ -> false)
	    (step_state_ready_hterms ss)
    in let do_hterm hterm =
      let (ss',stage) = make_determined_stage gs ss hterm
      in Stage_determined(stage,(next_step gs ss'))
    in let stages =
      try_until_success No_determined_lookups
	determined_lookup_hterms do_hterm
    in
    stages

(************************************************************************
************************************************************************)

and try_joinable_lookups (gs: global_state_type) (ss: step_state_type) =
    let joinable_lookup_hterms =
	List'.remove_if_not
	    (fun hterm ->
		match hterm with
		  Hterm_lookup(lookup_m)
			when pending_joinable_field_p gs ss
				lookup_m.nam_out_field ->
		    true
		| _ -> false)
	    (step_state_ready_hterms ss)
    in let do_hterm hterm =
      let join_field = primary_field_of_hterm hterm
      in let equiv_class =
	find_equiv_class gs ss join_field
      and jm_of_field = make_jm_of_field gs ss
      in let hterms =
	equiv_class_to_hterms jm_of_field ss equiv_class    
      in let (ss',stage) = make_join_stage gs ss equiv_class hterms
      in Stage_join(stage,(next_step gs ss'))
    in let stages =
      try_until_success No_joinable_lookups joinable_lookup_hterms
	do_hterm
    in
    stages

(************************************************************************
************************************************************************)

and try_unjoinable_lookups (gs: global_state_type) (ss: step_state_type) =
    let unjoinable_lookup_hterms =
	List'.remove_if_not
	    (fun hterm ->
		match hterm with
		  Hterm_lookup(_) -> true
		| _ -> false)
	    (step_state_ready_hterms ss)
    in let do_hterm hterm =
      let (ss',stage) = make_unjoinable_stage gs ss hterm
      in Stage_unjoinable(stage,(next_step gs ss'))
    in let stages =
      try_until_success No_unjoinable_lookups unjoinable_lookup_hterms
	do_hterm
    in
    stages

;;

