(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Printf;;
open Ssisl
open Containers

open Bb;;
open Bmf_query;;
open Bmf_join;;
open Scheduler_types;;

type hterm_kind =
| Hterm_enum_simple
    of named_access_method    (* the enun_m for the join var *)
| Hterm_enum_indexing
    of named_access_method    (* the enum_m *)
     * named_access_method    (* the lookup_m for the join var *)
| Hterm_enum_searching
    of named_access_method    (* the enun_m *)
     * named_access_method    (* the lookup_m for the join var *)
     * named_access_method    (* the search_m on the join var *)
| Hterm_lookup
    of named_access_method    (* the lookup_m for the join var *)
;;

let output_hterm_kind outch hterm =
  match hterm with
    Hterm_lookup(m1) ->
      fprintf outch "Hterm_lookup(%a)"
 	output_named_access_method m1
  | Hterm_enum_simple(m1) ->
      fprintf outch "Hterm_enum_simple(%a)"
 	output_named_access_method m1
  | Hterm_enum_indexing(m1,m2) ->
      fprintf outch "Hterm_enum_indexing(%a,%a)"
 	output_named_access_method m1
 	output_named_access_method m2
  | Hterm_enum_searching(m1,m2,m3) ->
      fprintf outch "Hterm_enum_searching(%a,%a,%a)"
 	output_named_access_method m1
 	output_named_access_method m2
 	output_named_access_method m3
;;

let mthds_of_hterm hterm =
    match hterm with
      Hterm_lookup(m1) -> [m1]
    | Hterm_enum_simple(m1) -> [m1]
    | Hterm_enum_indexing(m1,m2) -> [m1;m2]
    | Hterm_enum_searching(m1,m2,m3) -> [m1;m2;m3]
;;

let primary_field_of_hterm hterm =
    match hterm with
      Hterm_lookup(m1) -> m1.nam_out_field
    | Hterm_enum_simple(m1) -> m1.nam_out_field
    | Hterm_enum_indexing(m1,m2) -> m2.nam_out_field
    | Hterm_enum_searching(m1,m2,m3) -> m2.nam_out_field
;;

let rel_of_hterm hterm =
    let m = List.hd (mthds_of_hterm hterm)
    in
	m.nam_rel
;;

let bound_by_hterm hterm =
    let ms = mthds_of_hterm hterm
    in let all_fields = List.map (fun m -> m.nam_out_field) ms
    in let bound_fields = List'.remove_duplicates all_fields
    in
	bound_of_list bound_fields
;;

let rec compute_ready_hterms bbm rel_to_qb_m bound =
    String_map.fold
	(fun r' qb lst ->
	    let bbi = String_map.find qb.qb_rel bbm
	    in let ready_hterms =
		compute_ready_hterms_for_relation r' bbi bound
	    in
		ready_hterms@lst)
	rel_to_qb_m []

and compute_ready_hterms_for_relation r' bbi bound =

    (**************************************************
    Values used when computing the ready hterms
    **************************************************)
    (* The fields of R' that are bound *)
    let bound_for_R' =
	List.fold_left
	    (fun lst name ->
		match name with
		  Field_rel(r'',field) when r' = r'' -> field::lst
		| _ -> lst)
	    [] (list_of_bound bound)
    in

    (**************************************************
    Functions used when computing the ready hterms
    **************************************************)

    (* A method is ready if all of its in fields are bound, and its out
       field isn't *)
    let ready_method_p am =
	(List'.difference am.am_in_fields bound_for_R' = []) &&
	    (not ((List.mem am.am_out_field bound_for_R')))
    in

    (* A method is almost ready w.r.t the field, f, it is not currently
       ready, but it will be if f is bound. *)
    let almost_ready_wrt_p f am =
	(List'.difference am.am_in_fields bound_for_R' = [f]) &&
	    (not ((List.mem am.am_out_field bound_for_R')))
    in

    (* Returns true iff mthd is a singleton method. Otherwise, it is a
       stream method *)
    let singleton_method_p mthd =
	match mthd.am_info with
	  Am_fun(_) -> true
	| Am_rel(_) -> false
    in
    

    (* A field is a join field if it is an index field. *)
    let join_field_p f =
	List.mem f bbi.bb_index_fields
    in

    (* A routine to put an access method together *)
    let nam am =
	make_named_access_method r' am
    in

    (**************************************************
    Computing the ready hterms
    **************************************************)

    (* Find all of the ready methods *)
    let all_ready_methods =
	List'.remove_if_not ready_method_p bbi.bb_access_methods
    in

    (* Partition these into lookup (singleton) and enum (stream)
       methods *)
    let (lookup_mthds,enum_mthds) =
	(List'.remove_if_not singleton_method_p all_ready_methods,
	 List'.remove_if singleton_method_p all_ready_methods)
    in

    (* Construct the lookup hterms *)
    let lookup_hterms =
	List.map (fun x -> Hterm_lookup x) (List.map nam lookup_mthds)
    in

    (* Find all of the enum methods whose out field is a join field. These
       will be used to construct the simple enum hterms. *)
    let (joinable_mthds,unjoinable_mthds) =
      let pred mthd = join_field_p mthd.am_out_field
      in
      (List'.remove_if_not pred enum_mthds,
       List'.remove_if pred enum_mthds)
    in

    (* Construct the simple enum hterms *)
    let enum_simple_hterms =
	List.map (fun x -> Hterm_enum_simple x) (List.map nam joinable_mthds)
    in

    (* For each enum mthd whose out field is unjoinable, find all mthds,
       (a) that are almost ready wrt to the out field of the enum mthd,
       (b) that are singleton methods, and
       (c) whose out field is a join field. 
       Each enum mthd is pairs with a list of such methods. *)
    let indexing_mthds_pairs =
	List.map
	    (fun enum_mthd ->
		let indexing_mthds =
		    List'.remove_if_not
			(fun mthd ->
			    (singleton_method_p mthd)
			    && (almost_ready_wrt_p enum_mthd.am_out_field 
				    mthd)
			    && (join_field_p mthd.am_out_field))
			bbi.bb_access_methods
		in
		    (enum_mthd,indexing_mthds))
	    unjoinable_mthds
    in

    (* Enum methods with no corrisponding indexing methods become simple
       enum hterms. *)
    let more_enum_simple_hterms =
	List.map (fun (enum_mthd,_) -> Hterm_enum_simple(nam enum_mthd))
	    (List'.remove_if_not (fun (_,indexing_mthds) -> indexing_mthds = [])
		indexing_mthds_pairs)
    in

    (* Such methods are combined with their corrisponding enum methods and
       made into pairs. The result is a list of these pairs. *)
    let indexing_pairs =
	List'.flat_map
	    (fun (enum_mthd,indexing_mthds) ->
		List.map (fun indexing_mthd -> (enum_mthd,indexing_mthd))
		    indexing_mthds)
	    indexing_mthds_pairs
    in

    (* Turn each (enum_mthd,indexing_mthd) pair into a triple
       (enum_mthd,indexing_mthd,search_mthds), where search_mthds are a
       list of relevant search mthds for the pair. A method is a relevant
       search method when,
       (a) It's a single method,
       (b) It's almost ready wrt the outfield of the index method, and
       (c) Its out field is the same as the enum method. *)
    let searching_triples =
	List.map
	    (fun (enum_mthd,indexing_mthd) ->
		let search_mthds =
		    List'.remove_if_not
			(fun mthd ->
			    (singleton_method_p mthd)
			    && (almost_ready_wrt_p
				    indexing_mthd.am_out_field
				    mthd)
			    && (mthd.am_out_field =
				    enum_mthd.am_out_field))
			bbi.bb_access_methods
		in
		    (enum_mthd,indexing_mthd,search_mthds))
	    indexing_pairs
    in

    (* Construct the list of indexing hterms and searching hterms from these
       triples. *)
    let (enum_indexing_hterms,enum_searching_hterms) =
	List.fold_left
	    (fun (indexing_hterms,searching_hterms)   
		    (enum_mthd,indexing_mthd,search_mthds) ->
		if search_mthds = [] then
		    let new_hterm =
			Hterm_enum_indexing(nam enum_mthd,nam indexing_mthd)
		    in
			(new_hterm::indexing_hterms,searching_hterms)
		else
		    let new_hterms =
			List.map (fun search_mthd ->
				Hterm_enum_searching(
				    nam enum_mthd, nam indexing_mthd,
				    nam search_mthd))
			    search_mthds
		    in
			(indexing_hterms,new_hterms@searching_hterms))
	    ([],[]) searching_triples
    in

    (* The results *)
    lookup_hterms @ enum_simple_hterms @ more_enum_simple_hterms @
	enum_indexing_hterms @ enum_searching_hterms
    
;;

(************************************************************************
Generate a join_term_kind from an hterm_type
************************************************************************)
let gen_join_term_kind hterm f =
    match hterm with
      Hterm_enum_simple(enum_m) ->
	Jtk_enum_simple(f enum_m)
    | Hterm_enum_indexing(enum_m,lookup_m) ->
	Jtk_enum_indexing(f enum_m, f lookup_m)
    | Hterm_enum_searching(enum_m,lookup_m,search_m) ->
	Jtk_enum_searching(f enum_m, f lookup_m, f search_m)
    | Hterm_lookup(lookup_m) ->
	Jtk_lookup(f lookup_m)
;;
