(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Printf;;

open Ssisl
open Containers

open Alpha;;
open Bmf;;
open Scheduler_types;;

type linear_state_type = {
    ls_H : matrix;             
    ls_x0 : vector;            
    ls_index_to_field_v: field_name array;
    ls_index_of_field : field_name -> int;
    ls_i : int; 
    ls_j : int; 
    ls_initial_i : int; 
    ls_initial_j : int;
    ls_old_is : int list;
    ls_equiv_classes : int list list
}
;;

let output_int outch i = fprintf outch "%d" i;;

let output_linear_state_type outch ls =
  let rec gen_old_equiv_classes (i: int) (lst: int list)
      (old_is: int list) (results: int list list) = (
    if old_is = [] then
      List.rev results
    else if i = List.hd old_is then
      gen_old_equiv_classes i [] (List.tl old_is)
	((List.rev lst)::results)
    else
      gen_old_equiv_classes (i+1) (i::lst) old_is results
	: int list list)

  in let output_row i =
    fprintf outch " %a(%d): [" 
      output_field_name (ls.ls_index_to_field_v.(i-1)) i;
    for j = 1 to matrix_cols ls.ls_H do
      if j = ls.ls_initial_j then
	fprintf outch " |"
      else if j = ls.ls_j then
	fprintf outch " ||";
      fprintf outch " %d" (matrix_ref ls.ls_H i j)
    done;
    fprintf outch "] = [ %d ]" (vector_ref ls.ls_x0 i)
    ;
  in
  let output_equiv_class lst =
    fprintf outch " {";
    if lst <> [] then
      begin
    	output_row (List.hd lst);
	List.iter 
	  (fun i -> 
	    fprintf outch ", ";
	    output_row i)
	  (List.tl lst)
      end;
    fprintf outch " }"
  in
  fprintf outch "[";
  List.iter output_equiv_class
    (gen_old_equiv_classes 1 []
       (ls.ls_old_is@[ls.ls_i]) []);
  fprintf outch " ||";
  List.iter output_equiv_class ls.ls_equiv_classes;
  fprintf outch " ]"

;;

type linear_advance_kinds =
        Linear_initial 
    |   Linear_subseq
;;

let make_index_of_field index_to_field_v =
    let field_to_index_m =
	Range.fold_left
	    (fun m i -> Fn_map.add index_to_field_v.(i-1) i m)
	    Fn_map.empty 1 (Array.length index_to_field_v)
    in
	(fun name ->
	    Fn_map.find name field_to_index_m)
;;

let rec compute_equivs h starting_row starting_col =
    let rs = matrix_rows h
    and cs = matrix_cols h
    in let rs' = rs - starting_row + 1
    and cs' = cs - starting_col + 1
    in let h' =
	if rs' = 0 || cs' = 0 then
	    create_matrix rs' cs'
	else
	    extract_matrix h starting_row rs starting_col cs
    (* Scale a row by its gcd *)
    in let scale_row_by_gcd i =
	let c = Range.fold_left
	    (fun c j ->
		let x = matrix_ref h' i j
		in
		    if x = 0 then
			c
		    else if c = -1 then
			x
		    else
			gcd x c)
	    (-1) 1 cs'
	in
	    for j = 1 to cs' do
		matrix_set h' i j ((matrix_ref h' i j) / c)
	    done
    (* Test for row eq *)
    and rows_eq i1 i2 =
	let rec loop j =
	    if j > cs' then
		true
	    else if matrix_ref h' i1 j <> matrix_ref h' i2 j then
		false
	    else
		loop (j+1)
	in
	    loop 1
    in let _ =
	zassert (cs' >= 0) (Invalid_argument("compute_equivs"));
	(* Scale each row by its gcd *)
	for i = 1 to rs' do
	    scale_row_by_gcd i
	done
    (* Group the rows into equiv classes *)
    (*  classes is a list of refs of lists of row indices *)
    (*  each list of row indices is an equiv class. *)
    in let classes =
	Range.fold_left
	    (fun classes i ->
		let rec loop lst =
		    match lst with
		      [] ->
			(ref [i])::classes
		    (* If row i is equal to the
			representative row in equiv class ``x'',
			then put row i into that equiv class. *)
		    | x::lst' when rows_eq i (List.hd !x) ->
			x := i::!x;
			classes
		    | _::lst' ->
			loop lst'
		in
		    loop classes)
	    [] 1 rs'
    in
	List.map (fun x ->
		List.map (fun y ->
			y + starting_row - 1)
		    !x)
	    classes
;;

let make_linear_state h x0 index_to_field_v i j initial_i initial_j
	old_is =
    { ls_H = h;
      ls_x0 = x0;
      ls_index_to_field_v = index_to_field_v;
      ls_index_of_field = make_index_of_field index_to_field_v;
      ls_i = i;
      ls_j = j;
      ls_initial_i = initial_i;
      ls_initial_j = initial_j;
      ls_old_is = old_is;
      ls_equiv_classes = (compute_equivs h i j)
    }
;;

let make_initial_linear_state initial_H initial_x0
	initial_index_to_field_v =
    make_linear_state initial_H initial_x0 initial_index_to_field_v
	1 1 1 1 []
;;

let amend_linear_state ls h' x0' index_to_field_v' i' j' initial_p=
    let (initial_i',initial_j') =
	if initial_p then
	    (i',j')
	else
	    (ls.ls_initial_i,ls.ls_initial_j)
    in
	make_linear_state h' x0' index_to_field_v'
	    i' j' initial_i' initial_j' (ls.ls_old_is@[ls.ls_i])
;;

(************************************************************************
Use unimodular column transformations to zero $H[i,j+1:n]$.
************************************************************************)
let one_step_toward_lower_echelon h i j = (
    let rs = matrix_rows h
    and cs = matrix_cols h
    in
	for j' = cs downto j+1 do
	    let a = matrix_ref h i j
	    and b = matrix_ref h i j'
	    in
		if a <> 0 && b <> 0 then (
		    let (d,u,v) = egcd a b
		    in
			for i' = i to rs do
			    let x = matrix_ref h i' j
			    and y = matrix_ref h i' j'
			    in
				matrix_set h i' j (u*x + v*y);
				matrix_set h i' j' (-b/d*x + a/d*y)
			done
		) else if a = 0 && b <> 0 then (
		    for i' = i to rs do
			let x = matrix_ref h i' j
			and y = matrix_ref h i' j'
			in
			    matrix_set h i' j y;
			    matrix_set h i' j' x
		    done
		)
	done;
	(* Is the row entirely zero? *)
	matrix_ref h i j = 0
);;

let permute_rows (h:matrix) i j (x0:vector)
                (index_to_field_v: 'a array) rows =
    let swap v1 set1 v2 set2 = set1 v2; set2 v1
    and rs = matrix_rows h
    and cs = matrix_cols h
    and rows' = Sort.list (fun x y -> x <= y) rows
    in let new_i =
	List.fold_left
	    (fun i i' ->
		assert (i<=i') (* (Failure("permute_rows")) *);
		if i <> i' then begin
		    for j = 1 to cs do
			swap (matrix_ref h i j) (matrix_set h i j)
			    (matrix_ref h i' j) (matrix_set h i' j)
		    done;
		    swap (vector_ref x0 i) (vector_set x0 i)
			(vector_ref x0 i') (vector_set x0 i');
		    swap (index_to_field_v.(i-1))
			    (Array.set index_to_field_v (i-1))
			 (index_to_field_v.(i'-1))
			    (Array.set index_to_field_v (i'-1))
		end;
		i + 1)
	    i rows'
    in
	new_i
;;

let advance_linear_state kind
	({ ls_H = h; ls_x0 = x0; ls_index_to_field_v = index_to_field_v;
	    ls_i=i; ls_j=j } as ls)
	equiv_class =
    let h' = copy_matrix h
    and x0' = copy_vector x0
    and index_to_field_v' = Array.copy index_to_field_v
    in let i' =
	permute_rows h' i j x0' index_to_field_v' equiv_class
    in let zero_rows_p =
	one_step_toward_lower_echelon h' i j
    in let (j',initial_p) =
	match kind with
	  Linear_initial ->
	    let j' = if zero_rows_p then j else j+1
	    in
		(j',true)
	| Linear_subseq ->
	    zassert ((not (zero_rows_p)))
		(Failure("one_step_toward_lower_echelon: Zero row in H?!?"));
	    (j+1,false)
    in
	amend_linear_state ls h' x0' index_to_field_v' i' j' initial_p
;;


(************************************************************************
Generate value of
$g_i=h^{(i)}_1*t_1 + h^{(i)}_2*t_2 + \cdots + h^{(i)}_n*t_n +
\tilde{x}^{(i)}$

[[(column_term j)]] returns the [[bmf_expr]] to be used for each $t_j$.
************************************************************************)
let expr_of_row t_var_name_v ls i =
    let {ls_H = h; ls_x0 = x0} = ls
    in let x0_term =
	match vector_ref x0 i with
	  0 -> []
	| c -> [Expr_int(c)]
    in let h_terms =
	Range.fold_left
	    (fun h_terms j ->
		let h_var = Expr_var(t_var_name_v.(j-1))
		in
		    match (matrix_ref h i j) with
		      0 -> h_terms
		    | 1 -> h_terms@[h_var]
		    | -1 -> h_terms@[make_neg h_var]
		    | c -> h_terms@[make_times [Expr_int(c);h_var]])
	    x0_term 1 (matrix_cols h)
    in
	make_plus h_terms
;;
