(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Printf;;

open Alpha;;

open Ssisl;;
open Containers

open Bmf;;
open Bmf_util;;
open Bmf_query;;
open Bmf_join;;
open Bmf_linear;;
open Bmf_parser_aux;;
open Bb;;
open Bb_decl;;

open Scheduler_types;;
open Scheduler_hier;;
open Scheduler_linear;;

module WS = Wff.Simple

module Hterm_element =
  struct
    type t = hterm_kind
    let compare = Pervasives.compare
  end
module Hterm_map = Map(Hterm_element)

(************************************************************************
Global state
************************************************************************)

type global_state_type = {
    gs_bbm : black_box_map;
    gs_q : query;
    gs_rel_to_qb_m: query_binding String_map.t;
    gs_all_fields: bound_type;
    gs_join_fields: Fn_set.t;
    gs_find_acc_mthd:
	(*R*) string -> (*name*) string -> named_access_method;
    gs_all_acc_mthds: named_access_method list;
    gs_initial_ls: linear_state_type;
    gs_t_var_name_v: string array
};;

let global_state_all_fields gs =
    gs.gs_all_fields
and global_state_all_rels gs =
    String_map.fold (fun rel _ lst -> rel::lst) gs.gs_rel_to_qb_m []
and global_state_qb_of_rel gs rel =
    String_map.find rel gs.gs_rel_to_qb_m
and global_state_bbm gs =
    gs.gs_bbm
and global_state_query gs =
    gs.gs_q
and global_state_t_var_name_v gs =
    gs.gs_t_var_name_v
;;

(************************************************************************
Naming the relation instances within the query

We need to use the relation name in some places, but the relation name is
not enough to determine which reference is being referenced. This is
because you might have two references to the same relation with different
access functions, e.g., $A[i,j]$ and $A[i+1,j+1]$.

We resolve the problem, by creating artifial relation names, and mapping
them back to their respective query bindings, which are unique.
************************************************************************)
let make_rel_to_qb_m q =
    List.fold_left
	(fun m qb ->
	    let r = qb.qb_rel
	    in let r' =
		(* Only create a new name if an entry in the map already
		    exists. *)
		try
		    String_map.find r m;
		    Misc.gensym r
		with Not_found ->
		    r
	    in
		String_map.add r' qb m)
    String_map.empty q.query_bindings
;;

(************************************************************************
************************************************************************)

let make_var_of_field () =
  let field_to_var_ht = Fn_hashtbl.create 101
  in let var_of_field field =
    try
      Fn_hashtbl.find field_to_var_ht field
    with Not_found -> (
      let var =
	match field with
	  Field_loop(i) -> i
	| Field_inv(n) -> n
	| Field_rel(r,f) -> Misc.gensym (r^f)
      in
      Fn_hashtbl.add field_to_var_ht field var;
      var)
  in var_of_field


(************************************************************************
Constructing the linear system

\label{sec:constr-linear}

We construct a system $Ax=b$ that contains the linear constraints of the
query. We then ``solve'' this system for the new parametric system,
$Ht+x_0 = x$, where $H$ is the null space of $A$ and $t$ is a set of
parametric variables.
************************************************************************)

let construct_linear_system bbm q rel_to_qb_m =
    let var_of_field = make_var_of_field ()
    in let (initial_scope,add_to_scope,free_vars_in) =
        (* "Scopes" are sets of names. *)
	let initial_scope = String_set.empty
	and add_to_scope field scope = 
	  String_set.add (var_of_field field) scope
	(* free_vars_in will find the free vars in e within the scope. IE,
	   the free vars of e minus the names in the scope *)
	and free_vars_in scope e =
	  String_set.elements
	    (String_set.diff (expr_fvs e) scope)
	in
	    (initial_scope,add_to_scope,free_vars_in)
    in let notice_field (fields,scope) field =
	(field::fields,add_to_scope field scope)
    (**************************************************
    Utility function adding all free variables to [[scope]] and [[fields]].
    **************************************************)
    in let add_free_vars_in (fields,scope) e  =
	       List.fold_left
		   (fun (fields,scope) var ->
		       notice_field (fields,scope) (Field_inv(var)))
		   (fields,scope)
		   (free_vars_in scope e)
    in let scope = initial_scope
    in let (loop_fields,scope) =
	 (**************************************************
	 Extract all of the loop fields from the query bindings.
	 **************************************************)
	 (List.fold_left
	     (fun (fields,scope) var ->
		 notice_field (fields,scope) (Field_loop(var)))
	     ([],scope) (collect_loop_vars q))
    in let (all_array_fields,scope) =
	 (**************************************************
	 Extract all of the necessary fields from the arrays
	 **************************************************)
	let rec extract_stuff_from_qb r' qb (all_array_fields,scope) =
	    let r = qb.qb_rel
	    in let bbi = 
	      try
	      	String_map.find r bbm
	      with Not_found ->
		failwith ("Unbound array name: " ^ r)
	    in let names = all_fields_in_black_box bbi
	    in let fields = List.map (fun name -> Field_rel(r',name)) names
	    in
		List.fold_left notice_field (all_array_fields,scope) fields
	in
	    String_map.fold extract_stuff_from_qb rel_to_qb_m ([],scope)
    in let (array_join_fields,array_constraints,scope) =
	 (**************************************************
	 Extract all of the necessary fields and constraints from the access
	 functions.
	 **************************************************)
	let rec extract_stuff_from_qb r' qb
		(array_fields,array_constraints,scope) =
	    let r = qb.qb_rel
	    in let bbi = String_map.find r bbm
	    in
		assert ((List.length(bbi.bb_index_fields)) =
			(List.length(qb.qb_acc)))
		    (* (Failure("Number of exprs in access does match bb")) *);
		List.fold_left2 (extract_stuff_from_acc r')
		    (array_fields,array_constraints,scope)
		    bbi.bb_index_fields
		    qb.qb_acc
	and extract_stuff_from_acc r' (array_fields,array_constraints,scope)
		name acc =
	    let (array_fields,scope) =
		add_free_vars_in (array_fields,scope) acc
	    in let field = Field_rel(r',name)
	    in let var = var_of_field field
	    in let (array_fields,scope) =
		notice_field (array_fields,scope) field
	    in let constraint0 = make_plus [acc; (make_neg (Expr_var(var)))]
	    in
		(array_fields, constraint0::array_constraints, scope)
	in
	    String_map.fold extract_stuff_from_qb rel_to_qb_m ([],[],scope)
    in let (pred_join_fields,pred_constraints,scope) =
	 (**************************************************
	 Extract all of the necessary fields and constraints from the
	 query predicate.
	 **************************************************)
	let equalities =
	    List'.remove_if_not
		(fun constr ->
		    match constr with
		      Expr_op(Op_eq,_) -> true
		    | _ -> false)
		(bmf_expr_args_by_op Op_and (q.query_bounds))
	in let (pred_fields,scope) =
	    List.fold_left add_free_vars_in ([],scope) equalities
	and pred_constraints =
	    List.map
		(fun constr ->
		    match constr with
		      Expr_op(Op_eq,[lhs;rhs]) ->
			make_plus [lhs; make_neg rhs]
		    | _ -> assert false (* Can't happen *))
		equalities
	in
	    (pred_fields,pred_constraints,scope)
    in let all_fields = bound_of_list(loop_fields @ all_array_fields)
    and join_fields = loop_fields @ array_join_fields @ pred_join_fields
    and constraints = array_constraints @ pred_constraints
    in let join_vars = List.map var_of_field join_fields
    in let index_to_field_v = Array.of_list join_fields
    in let (Affine_map(a,b)) = expr_list2affine_map join_vars constraints
    in let (h,x0) = solve_linear a b
    in
	(h,x0,all_fields,join_fields,index_to_field_v)
;;

module Str_str_element =
  struct
    type t = string * string
    let compare = Pervasives.compare
  end
module Str_str_map = Map(Str_str_element)

let make_global_state bbm q =
    let rel_to_qb_m = make_rel_to_qb_m q
    in let var_of_field = make_var_of_field ()
    in let (initial_H,initial_x0,all_fields,join_fields,
		initial_index_to_field_v) =
      construct_linear_system bbm q rel_to_qb_m
    in let initial_ls =
	make_initial_linear_state initial_H initial_x0
	    initial_index_to_field_v
    in let (find_acc_mthd,all_acc_mthds) =
	(* Name the access methods in the query. *)
	let all_acc_mthds =
	    String_map.fold
		(fun r' qb lst ->
		    let r = qb.qb_rel
		    in let bbi = String_map.find r bbm
		    in
			List.fold_left
			    (fun lst am ->
				(make_named_access_method r am)::lst)
			    lst bbi.bb_access_methods)
		rel_to_qb_m []
	in let rname_to_nam_map =
	    List.fold_left
		(fun rname_to_nam_map mthd ->
		    let r = mthd.nam_rel
		    and name = mthd.nam_unnamed.am_name
		    in
			Str_str_map.add (r,name) mthd rname_to_nam_map)
		Str_str_map.empty all_acc_mthds
	in let find_acc_mthd r name =
	    Str_str_map.find (r,name) rname_to_nam_map
	in
	    (find_acc_mthd,all_acc_mthds)
    in let t_var_name_v =
	Misc.build_vector
	    (fun j -> Misc.gensym ("t"^(string_of_int (j+1))))
	    (matrix_cols initial_H)
    in {
        gs_bbm = bbm;
        gs_q = q;
	gs_rel_to_qb_m = rel_to_qb_m;
	gs_all_fields = all_fields;
	gs_join_fields = 
          List.fold_right Fn_set.add join_fields Fn_set.empty;
	gs_find_acc_mthd = find_acc_mthd;
	gs_all_acc_mthds = all_acc_mthds;
	gs_initial_ls = initial_ls;
	gs_t_var_name_v = t_var_name_v
    }
;;

let joinable_field_p
	({gs_join_fields = join_fields}
	    as gs: global_state_type) field =
    Fn_set.mem field join_fields
;;


(************************************************************************
Bindings
************************************************************************)
type bindings_type = bmf_expr Fn_map.t
;;

(************************************************************************
Keeping try of the values of fields

As joins, loops, tests, etc., are scheduled, the values of various fields
become bound and others become determined. When a field is bound, its value
has been computed by some enclosing code and can be used. When a field is
determined, it has not been computed yet, but its value is already known.
This can happen when some terms of a join cannot be scheduled with the
join, because they are not ready.

The variable [[bound]], which is part of the state passed to
[[next_stage]], keeps track of the set of bound variables. The variable
[[bindings]] keeps track of the values of fields that have been either
bound or determined.

The code in this subsection implements the abstraction to [[bindings]].
[[bindings]] is a non-destrucive map whose key is a field, and whose value
is a BMF expression.
************************************************************************)
let rec add_binding bindings field expr =
    Fn_map.add field expr bindings
and find_binding bindings field =
    Fn_map.find field bindings

and has_binding_p bindings field =
    try
	find_binding bindings field;
	true
    with Not_found ->
	false

(************************************************************************
The routine [[add_linear_binding]] adds a binding constructed from the
appropriate row of the system, $H$.
************************************************************************)
and add_linear_binding (gs: global_state_type)
	bindings ls field =
    let index = ls.ls_index_of_field field
    in let val0 = expr_of_row gs.gs_t_var_name_v ls index
    in let val' = bmf_expr_simplify val0
    in
	add_binding bindings field val'

(************************************************************************
The routine [[add_var_binding]] adds a binding constructed from a variable
name.
************************************************************************)
and add_var_binding bindings field name =
    add_binding bindings field (Expr_var(name))

(************************************************************************
The routine [[add_self_binding]] add the variable binding obtained by
generating the variable name from the field name.
************************************************************************)
and add_self_binding var_of_field bindings field =
    add_var_binding bindings field (var_of_field field)

(************************************************************************
The routine [[add_lookup_binding]] adds an access method invocation
expression to use as the binding for a field.
************************************************************************)
and add_lookup_binding
	({gs_rel_to_qb_m = rel_to_qb_m}
	    as gs: global_state_type)
	bindings lookup_m =
    let r' = lookup_m.nam_rel
    and lookup_s = lookup_m.nam_unnamed.am_name
    and in_fields = lookup_m.nam_in_fields
    and out_field = lookup_m.nam_out_field
    in let val0 =
	parse_expr "$R[[$lookup_s,$args]]"
	    [("$R",An_string((String_map.find r' rel_to_qb_m).qb_rel));
	     ("$lookup_s",An_expr(Expr_string(lookup_s)));
	     ("$args",An_list(
		List.map (fun in_field ->
			An_expr(find_binding bindings in_field))
		    in_fields))]
    in
	add_binding bindings out_field val0
;;

(************************************************************************
Step state
************************************************************************)

type step_state_type = {
    ss_bound: bound_type;
    ss_ls : linear_state_type;
    ss_ready_hterms : hterm_kind list
};;

let step_state_bound ss =
    ss.ss_bound;;
let step_state_ls ss =
    ss.ss_ls;;
let step_state_ready_hterms ss =
    ss.ss_ready_hterms;;

let make_initial_step_state gs initial_ls =
    let ready_hterms =
	compute_ready_hterms gs.gs_bbm gs.gs_rel_to_qb_m initial_bound
    in {
	ss_bound = initial_bound;
	ss_ls = initial_ls;
	ss_ready_hterms = ready_hterms
    }
;;

let make_step_state
	({gs_initial_ls = initial_ls} as gs) =
    make_initial_step_state gs initial_ls

;;

let update_ss_bound gs bound' ss =
    let ready_hterms =
	compute_ready_hterms gs.gs_bbm gs.gs_rel_to_qb_m bound'
    in {
	ss_bound = bound';
	ss_ls = ss.ss_ls;
	ss_ready_hterms = ready_hterms
    }
and update_ss_ls ls' ss =
    { ss_bound = ss.ss_bound;
      ss_ls = ls';
      ss_ready_hterms = ss.ss_ready_hterms }
;;

(* A field is determined if it is a joinable field and its value is
   determined w.r.t. the linear system. 

   The function returns false when either the field is not joinable or is
   not determined w.r.t. the linear system. Don't change this behavior. *)

let determined_field_p (gs: global_state_type) ss field =
    try
	let ls = ss.ss_ls
	in let i = ls.ls_i
	and i' = ls.ls_index_of_field field
	in
	    (* i is the index of the first _unprocessed_ row of H *)
	    i' < i
    with Not_found ->
	false
;;

let output_step_state outch ss =
  fprintf outch "{ ss_bound = %a; " output_bound_type ss.ss_bound;
  fprintf outch "ss_ls = %a; " output_linear_state_type ss.ss_ls;
  fprintf outch "ss_ready_hterms = %a }"
    (Io'.output_list output_hterm_kind) ss.ss_ready_hterms
;;

(************************************************************************
************************************************************************)

type re_type = {
    re_vars : string list;
    re_fields :  field_name list;
    re_equates : bmf_expr;
    re_bounds : bmf_expr
}
and range_info_type = {
    ri_symb_bounds : bmf_expr;
    ri_of_col : int -> join_bounds_type;
    ri_of_row : int -> join_bounds_type
}
;;

let range_info_of_symbolics ri =
    ri.ri_symb_bounds
and range_info_of_column ri j =
    ri.ri_of_col j
and range_info_of_row ri i =
    ri.ri_of_row i
;;

(************************************************************************
\label{sec:constr-bounds}

The result of [[generate_range_info]] is, basically, an object with two
methods.
\begin{description}
\item[ [[ri_of_col j]] ] The bounds of $t_j$ in terms of $t_j$.
\item[ [[ri_of_row i]] ] The bounds of $f_i$ in terms of $f_i$.
\end{description}
It also contains a field [[ri_symb_bounds]], which is an expression of the
bounds on the free variables that appeared in the query bounds.
************************************************************************)
let generate_range_info
	({gs_t_var_name_v = t_var_name_v;
	  gs_rel_to_qb_m = rel_to_qb_m;
	  gs_bbm = bbm; gs_q = q}
	    as gs: global_state_type)
	var_of_field ls =
    let rec row_to_constraint field =
	let i = ls.ls_index_of_field field
	and field_var = var_of_field field
	in let field_val = expr_of_row gs.gs_t_var_name_v ls i
	in
	    Expr_op(Op_eq,[Expr_var(field_var);field_val])
    and make_ident var =
	Ident_valid(var, Type_scalar(Type_int),
	    Ident_bind,An_list([]))
    (* The names of all of the $t_j$'s. *)
    in let t_vars =
	((Range.fold_left
	    (fun result i -> result@[t_var_name_v.(i-1)])
	    [] 1 (matrix_cols ls.ls_H)
	: string list))
    in let t_idents = List.map make_ident t_vars
    (* Blah, blah, blah. Same for loops. *)
    in let i_entry =
	let i_vars = collect_loop_vars q
	in let i_fields = List.map (fun x -> Field_loop x) i_vars
	in let i_equates = List.map row_to_constraint i_fields
	in {
	    re_vars = i_vars;
	    re_fields = i_fields;
	    re_equates = make_and i_equates;
	    re_bounds = q.query_bounds
	}
    and a_entries =
	(**************************************************
	For each array access that appears in the query, we need to generate a set
	of constraints to equate the array indices to $t$, and a set of constraints
	from the array bounds.
	**************************************************)
	String_map.fold
	    (fun r' qb entries ->
		let r = qb.qb_rel
		in let bbi = String_map.find r bbm
		in let raw_in_fields = bbi.bb_index_fields
		in let in_fields = List.map (fun f -> Field_rel(r',f)) raw_in_fields
		in let in_vars = List.map var_of_field in_fields
		in let entry = {
		    re_vars = in_vars;
		    re_fields = in_fields;
		    re_equates =
			make_and (List.map row_to_constraint in_fields);
		    re_bounds =
			bbi.bb_bounds_maker (List.map (fun x -> Expr_var x) in_vars)
		} in
		    entry::entries)
	    rel_to_qb_m []
    in let entries = i_entry::a_entries
    in let free_vars =
	(**************************************************
	All of the free variables that occur in the loop and black-box
	generated bounds.
	**************************************************)
        let empty_s = String_set.empty
	in let t_idents_s = 
	  String_set.of_list (List.map bmf_ident_name t_idents)
	and re_vars_s =
	  String_set.of_list
	    (List'.flat_map
	       (fun entry -> entry.re_vars)
	       entries)
	in let scope_s = String_set.union t_idents_s re_vars_s
	in let raw_free_s =
	  List.fold_left
	    (fun s entry -> String_set.union s (expr_fvs entry.re_bounds))
	    empty_s entries
	in String_set.to_list (String_set.diff raw_free_s scope_s)
    in let free_idents = List.map make_ident free_vars
    in let make_R_entry entry =
	parse_expr "{ [$f,$t] | (exist [$vars] |
			   $equates && $bounds) }"
	     [("$f",An_list(List.map (fun x -> An_string x) free_vars));
	      ("$t",An_list(List.map (fun x -> An_string x) t_vars));
	      ("$vars",An_list(List.map (fun x -> An_string x) entry.re_vars));
	      ("$equates", An_expr(entry.re_equates));
	      ("$bounds", An_expr(entry.re_bounds))]
    in let r_t =
	let r_entries = List.map make_R_entry entries
	in let orig_R_t =
	    List.fold_left intersect_sets (List.hd r_entries) (List.tl r_entries)
	in let r_t = simplify_set orig_R_t
	in
	    r_t
    in let (symb_bounds,col_bounds_v) =
	let all_bounds = fm_code_gen r_t
	in let (symb_bounds,t_bounds) =
	    List.fold_left
		(fun (symb_bounds,bounds) free_var ->
		    let free_var_bounds =
			bounds2expr (Expr_var(free_var)) (List.hd bounds)
		    in let symb_bounds' =
			make_and [free_var_bounds;symb_bounds]
		    and bounds' = List.tl bounds
		    in
			(symb_bounds',bounds'))
		(Expr_bool(true),all_bounds) free_vars
	in let col_bounds_v =
	    Array.of_list t_bounds
	in
	    (symb_bounds,col_bounds_v)
    and row_bounds_v =
	let field_to_bounds_m =
	    List.fold_left
		(fun field_to_bounds_m entry ->
		    (* entry_bounds are the bounds for each entry.re_vars *)
		    let r_pred =
			match r_t with
			  Expr_set(_,r_pred) -> r_pred
			| _ -> assert false (* R_t isn't a set *)
		    in let entry_bounds =
			List.map2
			    (fun var field ->
				let vars' = List'.remove var entry.re_vars
				and i = ls.ls_index_of_field field
				in let (scoped_t_idents,unscoped_t_idents) =
				    List.fold_right2
					(fun eo_group group_ident (scoped,unscoped) ->
					    if i < eo_group then
						(scoped,group_ident::unscoped)
					    else
						(group_ident::scoped,unscoped))
					((List.tl ls.ls_old_is)@[ls.ls_i]) t_idents ([],[])
				in let ident = make_ident var
				and idents' = List.map make_ident vars'
				in let scoped = free_idents @ scoped_t_idents
				in let r_t' =
				    Expr_set(Frame(scoped@[ident]),
					Expr_exist(Frame(unscoped_t_idents@idents'),
					    make_and [r_pred;entry.re_equates]))
				and r_entry' =
				    Expr_set(Frame(scoped@[ident]),
					Expr_exist(Frame(unscoped_t_idents@idents'),
					    make_and [entry.re_bounds;entry.re_equates]))
				in let xR_t' = simplify_set r_t'
				and xR_entry' = simplify_set r_entry'
				in let r_t'' = gist_sets xR_t' xR_entry'
				in let all_bounds = fm_code_gen r_t''
				in let just_var_bound =
				    List.fold_left (fun lst _ -> List.tl lst) all_bounds scoped
				in let bound =
				    match just_var_bound with
				      [bound] -> bound
				    | _ -> assert false (* Can't happen *)
				in
				    bound)
			    entry.re_vars entry.re_fields
		    (* Finally, add the bounds for the index fields to
		       "field_to_bounds_m". *)
		    in
			List.fold_right2 Fn_map.add entry.re_fields
		      	    entry_bounds field_to_bounds_m)
		Fn_map.empty entries
	in
	    Misc.build_vector
		(fun i ->
		    (try
			Fn_map.find (ls.ls_index_to_field_v.(i))
			    field_to_bounds_m
		    with Not_found ->
			([],[])))
		(matrix_rows ls.ls_H)
    in {
	ri_symb_bounds = symb_bounds;
	ri_of_col = (fun j -> col_bounds_v.(j-1));
	ri_of_row = (fun i -> row_bounds_v.(i-1))
    }
;;

(************************************************************************
Types for High-level plans
************************************************************************)

type hlp_type =
    global_state_type * stage_type

and stage_type =
  | Stage_constant of constant_stage_type * stage_type
  | Stage_invariant of invariant_stage_type * stage_type
  | Stage_join of join_stage_type * stage_type
  | Stage_determined of determined_stage_type * stage_type
  | Stage_unjoinable of unjoinable_stage_type * stage_type
  | Stage_body of body_stage_type

and constant_stage_type =
    { const_stage_ls: linear_state_type;
      const_stage_ls': linear_state_type }

and invariant_stage_type =
    { invar_stage_ls: linear_state_type;
      invar_stage_ls': linear_state_type;
      invar_stage_equiv_class: int list } 

and join_stage_type =
    { join_stage_ls: linear_state_type;
      join_stage_ls': linear_state_type;
      join_stage_equiv_class: int list;
      join_stage_hterms: hterm_kind list } 

and determined_stage_type =
    { deter_stage_ls: linear_state_type;
      deter_stage_hterm: hterm_kind }

and unjoinable_stage_type = 
    { unjoin_stage_hterm: hterm_kind }

and body_stage_type =
    { body_stage_ls: linear_state_type }

(************************************************************************)

let rec hlp2bmf (gs,stages) =
  let (range_info,body) = 
    stage2bmf gs Fn_map.empty 
      (make_var_of_field ())
      stages
  in
  Stmt_directive(An_list(
		 [An_string("bo-query");
		   An_expr(range_info_of_symbolics range_info)
		 ]),
		 body)

and stage2bmf gs bindings var_of_field stages =
  match stages with
    Stage_constant(stage,stages') ->  
      construct_constant_stage gs bindings var_of_field stage stages'
  | Stage_invariant(stage,stages') -> 
      construct_invariant_stage gs bindings var_of_field stage stages'
  | Stage_join(stage,stages') -> 
      construct_join_stage gs bindings var_of_field stage stages'
  | Stage_determined(stage,stages') -> 
      construct_determined_stage gs bindings var_of_field stage stages'
  | Stage_unjoinable(stage,stages') -> 
      construct_unjoinable_stage gs bindings var_of_field stage stages'
  | Stage_body(stage) -> 
      construct_body_stage gs bindings var_of_field stage

(************************************************************************
Place all of the zero rows (ie, constant accesses) at the very top of $H$.
************************************************************************)

and make_constant_stage
	(gs : global_state_type)
        ({ ss_ls = ls } as ss: step_state_type) =
    let { ls_equiv_classes = equiv_classes; ls_H = h } = ls
    in let _ =
	zassert (ls.ls_i = 1 && ls.ls_j = 1)
	    (Failure("How come stuff is already scheduled?"))
    in let (zero_rows_p,zero_equiv_class) =
	List.fold_left
	    (fun (zero_rows_p,zero_equiv_class) equiv_class ->
		(* are these rows zero? *)
		let i = List.hd equiv_class
		in let equiv_class_zero_p =
		    Range.fold_left
			(fun zero_p j ->
			    zero_p && (matrix_ref h i j = 0))
			true 1 (matrix_cols h)
		in
		    if (not (equiv_class_zero_p)) then
			(zero_rows_p,zero_equiv_class)
		    else begin
			zassert ((not (zero_rows_p)))
			    (Failure("Multiple zero equiv_classes?"));
			(true,equiv_class)
		    end)
	    (false,[]) equiv_classes
    in let (ls',ss') =
      if zero_rows_p then
	let ls' =
	  advance_linear_state Linear_initial ls zero_equiv_class
	in let ss' = update_ss_ls ls' ss
	in (ls',ss')
      else
	(ls,ss)
    in let stage =
      { const_stage_ls = ls;
      	const_stage_ls' = ls' }
    in (ss',stage)

and construct_constant_stage gs bindings var_of_field
    { const_stage_ls = ls;
      const_stage_ls' = ls' } 
    stages' = 
  let zero_rows_p = not (ls == ls')
  in let bindings' =
    if zero_rows_p then
      Range.fold_left
	(fun bindings' k ->
	  let field = ls'.ls_index_to_field_v.(k-1)
	  in
	  add_linear_binding gs bindings' ls' field)
	bindings ls.ls_i (ls'.ls_i-1)
    else
      bindings
  in let (range_info, body) = stage2bmf gs bindings' var_of_field stages'
  in (range_info, body)

(************************************************************************
Schedule a single equiv class that contains an invariant.
************************************************************************)

and make_invariant_stage
	({gs_t_var_name_v = t_var_name_v}
	    as gs: global_state_type)
        ({ss_bound = bound; ss_ls = ls}
	    as ss: step_state_type)
        equiv_class =

    (* fields in the equiv class *)
    let equiv_fields =
      List.map (fun i -> ls.ls_index_to_field_v.(i-1)) equiv_class
    in
    (* The new linear state *)
    let ls' =
      advance_linear_state Linear_initial ls equiv_class
    in 
    (* new set of bound identifiers  *)
    let newly_bound = 
      List'.remove_if_not
	(fun f ->
	  match f with
	    (Field_inv(_)|Field_loop(_)) -> true
	  | _ -> false)
	equiv_fields
    in let bound' = add_bound bound (bound_of_list newly_bound)
    in 
    (* new step state *)
    let ss' =
      update_ss_bound gs bound'
	(update_ss_ls ls' ss)
    in let stage =
      { invar_stage_ls = ls;
      	invar_stage_ls' = ls';
      	invar_stage_equiv_class = equiv_class }
    in (ss',stage)

and construct_invariant_stage
    ({gs_t_var_name_v = t_var_name_v} as gs)
    bindings var_of_field
    { invar_stage_ls = ls;
      invar_stage_ls' = ls';
      invar_stage_equiv_class = equiv_class } stages' =
    (* fields in the equiv class *)
    let equiv_fields =
      List.map (fun i -> ls.ls_index_to_field_v.(i-1)) equiv_class
    (* new bindings *)
    in let bindings' =
      List.fold_left
	(fun bindings' f ->
	  if invariant_field_p f then
	    add_self_binding var_of_field bindings' f
	  else
	    add_linear_binding gs bindings' ls' f)
	bindings equiv_fields
    in let (range_info,body) = stage2bmf gs bindings' var_of_field stages'

    (* routine to generate the binding of the "t" variable *)
    in let gen_t_binding f body =
      let k = ls'.ls_index_of_field f in
      Stmt_let(Frame(
	       [Ident_valid(
		 t_var_name_v.(ls.ls_j-1),
		 Type_scalar(Type_int),
		 Ident_var(solve_for_a_t gs ls' k ls.ls_j
			     (var_of_field f)),
		 An_list([]))]),
	       body)
    (* Routine to test that an invariant is satisfies its constraint with
       the "t" variable. *)
    and gen_t_test f body =
      let k = ls'.ls_index_of_field f in
      let n = var_of_field f
      and v = expr_of_row gs.gs_t_var_name_v ls' k
      in
      Stmt_if(Expr_op(Op_eq,[Expr_var(n);v]),
	      body, Stmt_nop)
    in 
    (* invariant files in the equiv class *)
    let invariant_equiv_fields = 
      let invariant_equiv_fields =
 	List'.remove_if_not invariant_field_p equiv_fields 
      in
      zassert (invariant_equiv_fields <> [])
	(Failure("Can't happen. Must be at least one inv field."));
      invariant_equiv_fields
    in
    (* Add the "t" binding and constraint tests to the body. *)
    let body' =
      List.fold_right gen_t_test (List.tl invariant_equiv_fields)
	(gen_t_binding (List.hd invariant_equiv_fields) body)
    in
    (range_info,body')
      
(* true iff field is an invariant field *)
and invariant_field_p f =
  match f with
    Field_inv(_) -> true
  | _ -> false


(************************************************************************
Do the same, except return an expression for $t_j$ in terms of [[var]].
************************************************************************)
and solve_for_a_t gs ls i j var =
    let {ls_H = h} = ls
    in let h0 = matrix_ref h i j
    in let _ = matrix_set h i j 0
    in let numerator =
	make_plus [Expr_var(var); make_neg
		    (expr_of_row gs.gs_t_var_name_v ls i)]
    in let result =
	match h0 with
	  0 -> assert false (* Can't happen *)
	| 1 -> numerator
	| -1 -> make_neg numerator
	| _ -> Expr_op(Op_div,[numerator;Expr_int(h0)])
    in
	matrix_set h i j h0;
	result

(************************************************************************
Generate a join stage.
************************************************************************)

and make_join_stage
	({gs_t_var_name_v = t_var_name_v}
	    as gs: global_state_type)
	({ss_bound = bound; ss_ls = ls }
	    as ss: step_state_type)
	equiv_class hterms = (
    let ls' =
	(* Permute and cancel in H *)
	advance_linear_state Linear_subseq ls equiv_class
    in let new_bounds =
	let bound_by_hterms =
	    List'.flat_map
		(fun hterm ->
		    list_of_bound (bound_by_hterm hterm))
		hterms
	and bound_as_loops =
	    Range.fold_left
		(fun results k ->
		    match ls'.ls_index_to_field_v.(k-1) with
		      (Field_loop(var) as f) -> f::results
		    | _ -> results)
		[] ls.ls_i (ls'.ls_i-1)
	in
	    bound_by_hterms @ bound_as_loops
    in let bound' =
	add_bound bound (bound_of_list new_bounds)
    in let ss' =
	update_ss_bound gs bound' (update_ss_ls ls' ss)
    in let stage =
      { join_stage_ls = ls;
      	join_stage_ls' = ls';
	join_stage_equiv_class = equiv_class;
      	join_stage_hterms = hterms }
    in (ss',stage)
)

and construct_join_stage
    ({ gs_t_var_name_v = t_var_name_v} as gs)
    bindings var_of_field
    { join_stage_ls = ls;
      join_stage_ls' = ls';
      join_stage_equiv_class = equiv_class;
      join_stage_hterms = hterms } stages' =
  let bindings' =
    let field_to_hterm_m =
      List.fold_left
	(fun m hterm ->
	  Fn_map.add (primary_field_of_hterm hterm) hterm m)
	Fn_map.empty hterms
    in let hterm_of_field field =
      Fn_map.find field field_to_hterm_m
    in
    Range.fold_left
      (fun bindings' k ->
	let field = ls'.ls_index_to_field_v.(k-1)
	in try
	  let hterm = hterm_of_field field
	  in let fields =
	    list_of_bound (bound_by_hterm hterm)
	  in
	  List.fold_left
	    (fun bindings' field ->
	      add_self_binding var_of_field bindings'
		field)
	    bindings' fields
	with Not_found ->
	  add_linear_binding gs
	    bindings' ls' field)
      bindings ls.ls_i (ls'.ls_i-1)
  in let (range_info,body) = stage2bmf gs bindings' var_of_field stages'
  in let name_of_hterm = make_name_of_hterm hterms
  in let terms =
    List.map (gen_join_term gs ls' var_of_field bindings' range_info
		name_of_hterm) 
      hterms
  in let join = {
    join_t_var = t_var_name_v.(ls.ls_j-1);
    join_bounds = range_info.ri_of_col ls.ls_j;
    join_terms = terms;
    join_pred = compute_join_pred gs name_of_hterm hterms;
    join_assertion = compute_assertion_pred gs name_of_hterm hterms;
    join_body = body
  } 
  in
  (range_info, join2stmt join)

(************************************************************************
Generate a determined term stage.
************************************************************************)

and make_determined_stage
	({ gs_bbm = bbm;
	   gs_rel_to_qb_m = rel_to_qb_m } as gs)
	({ ss_bound = bound } as ss)
	hterm =
    let new_bound = bound_by_hterm hterm
    in let bound' = add_bound bound new_bound
    in let ss' = update_ss_bound gs bound' ss
    in let stage =
    { deter_stage_ls = ss.ss_ls;
      deter_stage_hterm = hterm }
    in (ss',stage)

and construct_determined_stage 
    ({ gs_bbm = bbm;
       gs_rel_to_qb_m = rel_to_qb_m } as gs)
    bindings var_of_field
    { deter_stage_ls = ls;
      deter_stage_hterm = hterm } 
    stages' =
  let new_bound = bound_by_hterm hterm
  in let bindings' =
    List.fold_left
      (fun bindings' field ->
	try
	  Fn_map.find field bindings';
	  bindings'
	with Not_found ->
	  add_self_binding var_of_field bindings' field)
      bindings (list_of_bound new_bound)
  in let (range_info,body) = stage2bmf gs bindings' var_of_field stages'
  in let name_of_hterm = make_name_of_hterm [hterm]
  in let determined = {
    determined_term 
      = (gen_join_term gs ls var_of_field bindings' range_info 
	   name_of_hterm hterm);
    determined_pred 
      = compute_join_pred gs name_of_hterm [hterm];
    determined_assertion 
      = compute_assertion_pred gs name_of_hterm [hterm];
    determined_body = body
    } 
  in
  (range_info, determined2stmt determined)


(************************************************************************
Generate a unjoinable term stage.
************************************************************************)

and make_unjoinable_stage
    ({ gs_bbm = bbm;
       gs_rel_to_qb_m = rel_to_qb_m
     } as gs)
    ({ ss_bound = bound } as ss)
    hterm =
    let mthd =
	(* This should really be hidden in scheduler_hier *)
	match hterm with
	  Hterm_enum_simple(enum_m) -> enum_m
	| Hterm_lookup(lookup_m) -> lookup_m
	| _ ->
	    (* Can't happen. Any other kind of hterm contains a join 
	       field. *) 
	    assert false
    in let ss' =
      update_ss_bound gs
	(add_bound bound
	   (bound_of_list [mthd.nam_out_field]))
	ss
    in let stage =
      { unjoin_stage_hterm = hterm }
    in (ss',stage)

and construct_unjoinable_stage 
    ({ gs_bbm = bbm;
       gs_rel_to_qb_m = rel_to_qb_m } as gs)
    bindings var_of_field
    { unjoin_stage_hterm = hterm } stages' =
  let (gencode_f,bindings') =
    (* This should really be hidden in scheduler_hier *)
    match hterm with
      Hterm_enum_simple(enum_m) ->
	let gencode_f body =
	  let r' = enum_m.nam_rel
	  and enum_s = enum_m.nam_unnamed.am_name
	  and in_field_names = enum_m.nam_in_fields
	  and out_field_name = enum_m.nam_out_field
	  in let out_field_var = var_of_field out_field_name
	  in
	  parse_stmt
	    "do {[$var]|%op(in)($var,$R[[$name,$args]])} {
	       $body;
	     }"
	    [("$var",An_string(out_field_var));
	     ("$R",An_string((String_map.find r' rel_to_qb_m).qb_rel));
	      ("$name",An_expr(Expr_string(enum_s)));
	      ("$args",An_list(
	       List.map (fun in_field_name ->
		 An_expr(find_binding bindings
			   in_field_name))
		 in_field_names));
	      ("$body",An_stmt(body))]
	in let bindings' =
	  add_self_binding var_of_field bindings
	    enum_m.nam_out_field
	in
	(gencode_f,bindings')
    | Hterm_lookup(lookup_m) ->
	let bindings' =
	  add_lookup_binding gs bindings lookup_m
	in
	(identity,bindings')
    | _ ->
	    (* Can't happen. Any other kind of hterm contains a join 
	       field. *) 
	assert false
  in let (range_info,body) = stage2bmf gs bindings' var_of_field stages'
  in 
  (range_info, gencode_f body)


(************************************************************************
Assign names to each join term.
************************************************************************)
and make_name_of_hterm hterms =
    let m =
	List.fold_left
	    (fun m hterm ->
		let r = rel_of_hterm hterm
		in let name = Misc.gensym (r ^ "_jt")
		in
		    Hterm_map.add hterm name m)
	    Hterm_map.empty hterms
    in
	(fun hterm -> Hterm_map.find hterm m)

(************************************************************************
Generate a join_term_type
************************************************************************)
and gen_join_term
	(gs: global_state_type)
	(ls': linear_state_type)
	var_of_field bindings' range_info name_of_hterm hterm =
    let field = primary_field_of_hterm hterm
    in let k = ls'.ls_index_of_field field
    in {
	jt_flag = name_of_hterm hterm;
	jt_kind =
	    gen_join_term_kind hterm 
              (gen_jtk_equality gs var_of_field bindings');
	jt_constr = 
	    bmf_expr_simplify
		(expr_of_row gs.gs_t_var_name_v ls' k);
	jt_bounds = range_info.ri_of_row(k)
    }

(************************************************************************
Generating an equality for a [[join_term_kind]]
************************************************************************)
and gen_jtk_equality
	({ gs_rel_to_qb_m = rel_to_qb_m } as gs: global_state_type)
        var_of_field bindings m =
    let r' = m.nam_rel
    and s = m.nam_unnamed.am_name
    and var = var_of_field m.nam_out_field
    in let r = (String_map.find r' rel_to_qb_m).qb_rel
    in
	Jtk_eq(var,r,s,
	    List.map (find_binding bindings)
		m.nam_in_fields)

(************************************************************************
Computing the join and assertion predicates
************************************************************************)

(* The join predicate is arrived at by conjuncting each hterm that can be
   shown to be true from the query predicate. *)

and compute_join_pred gs name_of_hterm hterms =
    let q = global_state_query gs
    in let flag_of_hterm = make_flag_of_hterm gs hterms
    in let hterm_conjuncts =
	List.map
	    (fun hterm ->
		if WS.tautology_p
			(WS.Implies(gs.gs_q.query_pred,
			    WS.Atom(flag_of_hterm hterm))) then
		    WS.Atom(name_of_hterm hterm)
		else
		    WS.Constant(true))
	    hterms
    in
	WS.from_conjuncts hterm_conjuncts

(* The assertion predicate describe all possible truth assignments of the
   hterms that are feasible, with the query assertions as constraints. The
   result is the _disjunction_ of all of these truth assignments. 

   This procedure attempts to find the smallest set of disjuncts that is
   implied by the query assertion predicate. *)

and compute_assertion_pred gs name_of_hterm hterms =
    let flag_of_hterm = make_flag_of_hterm gs hterms
    in 
    (* Given a function, "f", and a list of identifiers, "vars", return
       a list of the results of apply "f" to all possible truth assignments
       to the identifiers in "vars". *)
    let map_for_all_truth_asgns f vars =
      let rec recurs f vals vars =
	match vars with
	  [] -> [f vals]
	| x::vars' ->
	    (recurs f ((x,true)::vals) vars')
	    @(recurs f ((x,false)::vals) vars')
      in
	recurs f [] vars
    (* fixme *)
    and vals_to_wff f vals =
	let val_to_conj (hterm,valu) =
	    let x = WS.Atom(f hterm)
	    in
		if valu then
		    x
		else
		    WS.Negation(x)
	in 
	    WS.from_conjuncts (List.map val_to_conj vals)
    in let all_flag_wffs =
	map_for_all_truth_asgns (vals_to_wff flag_of_hterm) hterms
    and all_name_wffs =
	map_for_all_truth_asgns (vals_to_wff name_of_hterm) hterms
    in let rec iter deleted_p tested_name_wffs tested_flag_wffs
	    remaining_name_wffs remaining_flag_wffs =
	match (remaining_name_wffs,remaining_flag_wffs) with
	  ([],[]) -> (deleted_p,tested_name_wffs,tested_flag_wffs)
	| (name_wff::remaining_name_wffs',
		flag_wff::remaining_flag_wffs') ->
	    let target_wff =
		WS.Implies(gs.gs_q.query_assertion,
		    WS.from_disjuncts
			(tested_flag_wffs@remaining_flag_wffs'))
	    in
		if WS.tautology_p target_wff then
		    iter true tested_name_wffs tested_flag_wffs
			remaining_name_wffs' remaining_flag_wffs'
		else
		    iter deleted_p (name_wff::tested_name_wffs)
			(flag_wff::tested_flag_wffs)
			remaining_name_wffs' remaining_flag_wffs'
	| _ -> assert false (* Can't happen *)
    in let (deleted_p,tested_name_wffs,tested_flag_wffs) =
	iter false [] [] all_name_wffs all_flag_wffs
    in
	if (not (deleted_p)) then
	    WS.Constant(true)
	else
	    WS.from_disjuncts tested_name_wffs

(* Construct a map from an hterm -> a flag. The flag is a string and is the
   boolean flag of the hterm's relation. *)
and make_flag_of_hterm gs hterms =
    let m =
	List.fold_left
	    (fun m hterm ->
		let r' = rel_of_hterm hterm
		in let qb = global_state_qb_of_rel gs r'
		in
		    Hterm_map.add hterm qb.qb_flag m)
	    Hterm_map.empty hterms
    in
	(fun hterm -> Hterm_map.find hterm m)

(************************************************************************
Generating the body of the query.
************************************************************************)

and make_body_stage gs ss =
  let stage = { body_stage_ls = ss.ss_ls }
  in stage

and construct_body_stage gs bindings var_of_field
    { body_stage_ls = ls } =
  let range_info =
    generate_range_info gs var_of_field ls
  and body =
    substr_in_body gs bindings
  in
  (range_info,body)

(************************************************************************
Doing substitutions in the loop body and bounds
************************************************************************)
and substr_in_body
	({gs_rel_to_qb_m = rel_to_qb_m;
	  gs_q = q; gs_bbm = bbm}
	    as gs: global_state_type)
	bindings =
    let var_to_binding_map =
        let var_to_binding_map =
	    String_map.empty
	in let
	    (* First, the loop vars *)
	    var_to_binding_map =
		List.fold_left
		    (fun m var ->
			let field = Field_loop(var)
			in let binding = find_binding bindings field
			in
			    String_map.add var binding m)
		    var_to_binding_map (collect_loop_vars q)
	in let
	    (* Second, the value field vars *)
	    var_to_binding_map =
		String_map.fold
		    (fun r' qb m ->
			let var = qb.qb_value
			in let bbi = String_map.find qb.qb_rel bbm
			in let field = Field_rel(r',bbi.bb_value_field)
			in let binding = find_binding bindings field
			in
			    String_map.add var binding m)
		    rel_to_qb_m var_to_binding_map
	in let
	    (* Third, the presence flag vars *)
	    (* *todo*. this only works when all refs are strong. *)
	    var_to_binding_map =
	  	String_map.fold
	            (fun r' qb m ->
		        let var = qb.qb_flag
			in
			    String_map.add var (Expr_bool(true)) m)
	            rel_to_qb_m var_to_binding_map
	in
	    var_to_binding_map
    in let rec s st =
		   match st with
		     Stmt_assign(lhs,rhs) -> Stmt_assign(e lhs,e rhs)
		   | Stmt_seq(first,second) -> Stmt_seq(s first,s second)
		   | Stmt_call(f,lhs,args) -> Stmt_call(f,e lhs,List.map e args)
		   | Stmt_for(i,lb,ub,step,body) ->
		       Stmt_for(i,e lb,e ub,e step,s body)
		   | Stmt_while(pred,body) -> Stmt_while(e pred,s body)
		   | Stmt_if(pred,thn,els) -> Stmt_if(e pred,s thn,s els)
		   | Stmt_let(frm,body) -> Stmt_let(frm,s body)
		   | Stmt_do(pred,body) -> Stmt_do(e pred,s body)
		   | Stmt_nop -> Stmt_nop
		   | Stmt_directive(ann,s') -> Stmt_directive(ann,s s')
    and e ex =
	    match ex with
	      (Expr_int(_)|Expr_real(_)|Expr_string(_)|Expr_bool(_)|Expr_void) ->
		ex
	    | Expr_var(v) ->
		(try
		    String_map.find v var_to_binding_map
		with Not_found ->
		    ex)
	    | Expr_op(op,args) -> Expr_op(op,List.map e args)
	    | Expr_forall(frm,pred) -> Expr_forall(frm,e pred)
	    | Expr_exist(frm,pred) -> Expr_exist(frm,e pred)
	    | Expr_access(acc,a,name,args) ->
		Expr_access(acc,e a,name,List.map e args)
	    | Expr_set(frm,pred) -> Expr_set(frm,e pred)
    in
	 s (Stmt_directive(An_list([An_string("eo-query")]),
		q.query_body))
;;
