(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
let big_env_create e' global_p =
  Frame({
	fr_bindings=BigBindings(Hashtbl.create 101);
	fr_next=e';
	fr_global_p=global_p
      })
let small_env_create e' global_p =
  Frame({
	fr_bindings=SmallBindings(ref []);
	fr_next=e';
	fr_global_p=global_p
      })

let env_empty_p e =
  match e with
    NoEnv -> true
  | Frame _ -> false

let env_parent e =
  match e with
    NoEnv -> raise (Invalid_argument("env_parent"))
  | Frame({fr_next=e'}) -> e'

let env_frame_iter e f =
  match e with
    NoEnv -> raise (Invalid_argument("env_define"))
  | Frame({fr_bindings=bindings}) ->
      (match bindings with
      | BigBindings(x) -> Hashtbl.iter f x
      | SmallBindings(x) ->
	  List.iter (fun (s,v) -> f s v) !x)

let rec env_define e s b =
  match e with
    NoEnv -> raise (Invalid_argument("env_define"))
  | Frame({fr_bindings=bindings}) ->
      (match bindings with
      | BigBindings(x) -> Hashtbl.add x s b
      | SmallBindings(x) -> x := (s,b)::!x)

and env_lookup_once e s =
  match e with
    NoEnv -> raise Not_found
  | Frame({fr_bindings=bindings}) ->
      (match bindings with
      | BigBindings(x) -> Hashtbl.find x s
      | SmallBindings(x) -> List.assoc s !x)

and env_lookup e s =
  try
    env_lookup_once e s
  with Not_found ->
    (match e with
      NoEnv -> raise Not_found
    | Frame({fr_next=e'}) -> env_lookup e' s)

and env_global_p e =
  match e with
    NoEnv -> raise (Invalid_argument("env_global_p"))
  | Frame({fr_global_p=global_p}) -> global_p

(************************************************************************)

exception Not_a_var of string

(* Throws Not_found or Not_a_var *)
let lookup_var env name =
  match env_lookup env (intern name) with
    Variable(valu) -> valu
  | _ -> raise (Not_a_var(name))

let define_var env name valu =
  env_define env (intern name) (Variable(ref valu))
and define_special env name evaluator =
  env_define env (intern name) (Special(evaluator))
and define_macro env name expander =
  env_define env (intern name) (Macro(expander))

let define_prim0 env name f =
  define_var env name
    (tag_procedure (Primitive(
	      {prim_name=name; prim_func=Prim0(f)})))
and define_prim1 env name f =
  define_var env name
    (tag_procedure (Primitive(
	      {prim_name=name; prim_func=Prim1(f)})))
and define_prim2 env name f =
  define_var env name
    (tag_procedure (Primitive(
	      {prim_name=name; prim_func=Prim2(f)})))
and define_prim3 env name f =
  define_var env name
    (tag_procedure (Primitive(
	      {prim_name=name; prim_func=Prim3(f)})))
and define_prim4 env name f =
  define_var env name
    (tag_procedure (Primitive(
	      {prim_name=name; prim_func=Prim4(f)})))
and define_prim5 env name f =
  define_var env name
    (tag_procedure (Primitive(
	      {prim_name=name; prim_func=Prim5(f)})))
and define_primn env name f =
  define_var env name
    (tag_procedure (Primitive(
	      {prim_name=name; prim_func=Primn(f)})))

(************************************************************************)

let make_global_env () = big_env_create NoEnv true
let make_closure_env env' = small_env_create env' false

(************************************************************************)

let rec copy_env env =
  match env with
    NoEnv -> NoEnv
  | Frame({fr_bindings=bindings; fr_next=env'; fr_global_p=global_p}) ->
      Frame({fr_bindings = copy_bindings bindings;
	      fr_next = copy_env env';
	      fr_global_p = global_p})

and copy_bindings x =
  match x with
    SmallBindings(x) -> SmallBindings(copy_small_bindings x)
  | BigBindings(x) -> BigBindings(copy_big_bindings x)

and copy_big_bindings htbl =
  let htbl' = Hashtbl.create 101 in
  Hashtbl.iter
    (fun key valu -> 
      Hashtbl.add htbl key (copy_binding valu))
    htbl;
  htbl'
    

and copy_small_bindings x =
  let lst = !x in
  let lst' =
    List.fold_left
      (fun lst' (key,valu) -> (key,copy_binding valu)::lst')
      [] lst
  in
  ref lst'

and copy_binding b =
  match b with
    (Special(_)|Macro(_)) -> b
  | Variable(x) -> Variable(ref !x)

(************************************************************************)

let null_environment () = (big_env_create NoEnv true)

let scheme_env = (copy_env (null_environment ()))
let scheme_environment () = copy_env scheme_env

let top_env = ref (big_env_create scheme_env true)
let interaction_environment () = !top_env

let with_top_level_env new_env thunk =
  let old_env = !top_env in
  top_env := new_env;
  try
    let result = thunk () in
    top_env := old_env;
    result
  with exn ->
    top_env := old_env;
    raise exn
