(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
exception Invalid_form of t
exception Unbound of symbol
exception Arguments_mismatch of string

let begin_sym = intern "begin"

let rec eval (expr: t) (env: env_type) = 
  let eval_sub (expr': t) = ((eval expr' env) : t) in
  match expr with
    (* Self-evaluating *)
    (Bool(_)|Char(_)|Float(_)|Int(_)|String(_)) -> expr
    (* Variable lookup *)
  | Symbol(_) -> !(variable_lookup env expr)
    (* Improper lists *)
  | Pair(_) when not (list_p expr) ->
      raise (Invalid_form(expr))
    (* (symbol ...) *) 
  | Pair(_) when symbol_p (car expr) ->
      let sym = (car expr) in
      let b = binding_lookup env sym in 
      (match b with
	Special(eval') -> 
	  (try
	    eval' expr env
	  with (Bad_arg(_,_)|Invalid_argument(_)) ->
	    raise (Invalid_form(expr)))
      | Macro(expander) ->
	    (* Expand the macro *)
	  let expr' = expander expr in
	  let expr'' =
	    (match expr' with
	      Pair(_) -> expr'
	    | _ -> (cons begin_sym (cons expr' nil)))
	  in
	    (* Over-write expr with the expansion *)
	  set_car expr (car expr'');
	  set_cdr expr (cdr expr'');
	    (* Re-evaluation expr. *)
	  eval expr env
      | Variable(valu) ->
	  apply !valu (map eval_sub (cdr expr)))
  | Pair(_) ->
      apply (eval_sub (car expr)) (map eval_sub (cdr expr))
  | _ -> raise (Invalid_form(expr))

and apply (f: t) (args: t) = 
  ((match untag_procedure f with
    Primitive(pr) -> apply_primitive pr args
  | Closure(cl) -> apply_closure cl args
	): t)

and variable_define env sym valu expr =
  if not (symbol_p sym) then
    raise (Invalid_form(expr));
  env_define env sym (Variable(ref valu))

and variable_lookup env sym =
  try
    (match env_lookup env sym with
      Variable(valu) -> valu
    | _ -> raise (Invalid_form(sym)))
  with Not_found ->
    raise (Unbound(symbol_to_string sym))

and binding_lookup env sym =
  try
    env_lookup env sym
  with Not_found ->
    raise (Unbound(symbol_to_string sym))

(************************************************************************)

and apply_primitive pr args =
  let name = pr.prim_name in
  try
    (match pr.prim_func with
      Prim0(f') ->
      	(match to_list args with
	  [] -> f' ()
      	|	_ -> raise (Arguments_mismatch(name)))
    | Prim1(f') ->
      	(match to_list args with
	  [a0] -> f' a0
      	|	_ -> raise (Arguments_mismatch(name)))
    | Prim2(f') ->
      	(match to_list args with
	  [a0;a1] -> f' a0 a1
      	|	_ -> raise (Arguments_mismatch(name)))
    | Prim3(f') ->
      	(match to_list args with
	  [a0;a1;a2] -> f' a0 a1 a2
      	|	_ -> raise (Arguments_mismatch(name)))
    | Prim4(f') ->
      	(match to_list args with
	  [a0;a1;a2;a3] -> f' a0 a1 a2 a3
      	|	_ -> raise (Arguments_mismatch(name)))
    | Prim5(f') ->
      	(match to_list args with
	  [a0;a1;a2;a3;a4] -> f' a0 a1 a2 a3 a4
      	|	_ -> raise (Arguments_mismatch(name)))
    | Primn(f') -> f' args)
  with 
    (* Translate the ML function name to the Scheme function name. *)
    Invalid_argument(name') -> raise (Invalid_argument(name^":"^name'))
  | Bad_arg(name',arg) -> raise (Bad_arg(name^":"^name',arg))

(************************************************************************)

and make_closure formals env body =
  tag_procedure (Closure({closure_formals = formals;
		      closure_env = env;
		      closure_body = body}))

and apply_closure cl actuals =
  let rec process_args env' formals actuals =
    (match (formals,actuals) with
      (Pair(_),Pair(_)) ->
	variable_define env' (car formals) (car actuals)
	  (tag_string "*shouldn't happen*");
	process_args env' (cdr formals) (cdr actuals)
    | (Symbol(_),_) ->
	variable_define env' formals actuals
	  (tag_string "*shouldn't happen*")
    | (Nil,Nil) -> ()
    | ((Nil,_)|(_,Nil)) -> 
	let (out_p,getter) = make_output_port_to_string () in
	let msg = 
	  with_output_to_port out_p
	    (fun () -> write cl.closure_formals)
	in
	raise (Arguments_mismatch(getter ()))
    | _ -> failwith "Sim.apply shouldn't happen.")
  in
  let env' = make_closure_env cl.closure_env in
  process_args env' cl.closure_formals actuals;
  eval cl.closure_body env'

(************************************************************************)

let funcall0 f = apply f nil
let funcall1 f a1 = apply f (list1 a1)
let funcall2 f a1 a2 = apply f (list2 a1 a2)
let funcall3 f a1 a2 a3 = apply f (list3 a1 a2 a3)
let funcall4 f a1 a2 a3 a4 = apply f (list4 a1 a2 a3 a4)
let funcall5 f a1 a2 a3 a4 a5 = apply f (list5 a1 a2 a3 a4 a5)
let funcall6 f a1 a2 a3 a4 a5 a6 = apply f (list6 a1 a2 a3 a4 a5 a6)
let funcall7 f a1 a2 a3 a4 a5 a6 a7 = 
  apply f (list7 a1 a2 a3 a4 a5 a6 a7)
let funcall8 f a1 a2 a3 a4 a5 a6 a7 a8 = 
  apply f (list8 a1 a2 a3 a4 a5 a6 a7 a8)

let load_aux (name : string) (verbose_p : bool) =
  ((
   with_input_from_file name
     (fun () ->
       let rec iter () =
	 try
	   let expr = (read ()) in
	   if verbose_p then
	     begin
	       write_string ";;; ";
	       write expr;
	       newline ()
	     end;
	   (eval expr (interaction_environment ()));
	   iter ()
	 with End_of_file ->
	   ()
       in
       (* This magic incantation will cause the lexer to ignore the first
	  line of a file, if the first two characters of that file are
	  "#!". Guess what this is good for. *)
       set_lex_state Lex_bof;
       iter ())
     ) : unit)

let load (name : string) (verbose_p : bool) = ((
  let cwd = Sys.getcwd () in
  try 
    (Sys.chdir (Filename.dirname name);
     load_aux (Filename.basename name) verbose_p; 
     Sys.chdir cwd)
  with 
    (* catch all exceptions and restore the current
       directory *)
    x -> (Sys.chdir cwd); raise x) : unit)


(************************************************************************)
(* Special forms for scheme						*)
(************************************************************************)

(* (quote x) *)
let spec_quote expr env =
  if length expr <> 2 then
    raise (Invalid_form(expr));
  second expr
let _ = define_special scheme_env "quote" spec_quote

(* (lambda args body ...) *)
let spec_lambda expr env =
  make_closure (cadr expr) env 
    (cons begin_sym (cddr expr))
let _ = define_special scheme_env "lambda" spec_lambda

(* (if test pred)
   (if test pred else) *)
let spec_if expr env =
  let len = length expr in
  if (eval (second expr) env) <> sharp_f then
    eval (third expr) env
  else if len = 3 then
    unspecified
  else if len = 4 then
    eval (fourth expr) env
  else
    raise (Invalid_form(expr))
let _ = define_special scheme_env "if" spec_if

(* (define var x)
   (define (f args ...) body ...) *)
let spec_define expr env =
  if (pair_p (cadr expr)) then
    begin
      let f = car (cadr expr) and
	  args = cdr (cadr expr) and
	  body = cddr expr in
      variable_define env f
      	(make_closure args env (cons begin_sym body))
	expr
    end
  else
    begin
      let var = second expr and
	  x = third expr in
      variable_define env var (eval x env) expr
    end;
  unspecified
let _ = define_special scheme_env "define" spec_define
    
(* (%%define-macro id expander) *)
let spec_define_macro expr env =
  if length expr <> 3 then
    raise (Invalid_form(expr));
  let id = second expr and
      expander = third expr in
  let expander_x = eval expander env in
  if not (symbol_p id) or not (procedure_p expander_x) then
    raise (Invalid_form(expr));
  env_define env id 
    (Macro(fun forms -> funcall1 expander_x forms));
  unspecified
let _ = define_special scheme_env "%%define-macro" spec_define_macro

(* (set! var x) *)
let spec_set expr env =
  if length expr <> 3 then
    raise (Invalid_form(expr));
  let var = second expr and
      x = third expr in
  let val_x = eval x env in
  (try
    (variable_lookup env var) := val_x
  with Unbound(_) as exn ->
    if env_global_p env then
      variable_define env var val_x expr
    else
      raise exn);
  unspecified
let _ = define_special scheme_env "set!" spec_set

(* (begin)
   (begin x)
   (begin x y ...) *)

let spec_begin expr env =
  let rec iter exprs result =
    if null_p exprs then
      result
    else if pair_p exprs then
      iter (cdr exprs) (eval (car exprs) env)
    else
      raise (Invalid_form(expr))
  in
  iter (cdr expr) unspecified
let _ = define_special scheme_env "begin" spec_begin


(* (letrec ((var x) ...) body ...) *)
let spec_letrec expr env =
  let rec create_bindings env' bindings =
    if null_p bindings then
      ()
    else
      begin
      	let binding = car bindings in
      	if length binding <> 2 then
	  raise (Invalid_form(expr));
      	variable_define env' (car binding)
	  (eval (cadr binding) env') expr;
      	create_bindings env' (cdr bindings)
      end
  in
  let env' = make_closure_env env in
  create_bindings env' (cadr expr);
  spec_begin (cons begin_sym (cddr expr)) env'
let _ = define_special scheme_env "letrec" spec_letrec
    
