(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
exception Syntax_error
exception Lexical_error

(************************************************************************)

let create_soc () = ref []
let access_soc ch =
  let s = List.fold_right (^) (List.rev !ch) "" in
  ch := [s];
  String.copy s

let soc_output_string ch str = (ch := str::(!ch))
let soc_output_char ch c =
  soc_output_string ch (String.make 1 c)
let soc_output_int ch i =
  soc_output_string ch (string_of_int i)
let soc_output_float ch f =
  soc_output_string ch (string_of_float f)
let soc_output_newline ch =
  soc_output_string ch "\n"
  

(************************************************************************)

let make_input_port_from_channel in_ch = 
  InPortChannel(in_ch,Lexing.from_channel(in_ch))
let make_input_port_from_string str =
  InPortString(str,Lexing.from_string(str))

let input_port_to_lexbuf in_port =
  match in_port with
    InPortChannel(in_ch,lex) ->  lex
  | InPortString(str,lex) -> lex

let close_input_port in_port =
  match in_port with
    InPortChannel(in_ch,_) -> close_in in_ch
  | InPortString(_,_) -> ()

(************************************************************************)

let make_output_port_to_channel out_ch =
  OutPortChannel(out_ch)
let make_output_port_to_string () =
  let ch = create_soc () in
  (OutPortString(ch), (fun () -> access_soc ch))

let close_output_port out_port =
  match out_port with
    OutPortChannel(out_ch) -> close_out out_ch
  | OutPortString(_) -> ()


(************************************************************************)

let call_with_input_file name proc =
  let in_chan = open_in name in
  let in_port = 
    make_input_port_from_channel in_chan
  in let result = proc in_port
  in
  close_in in_chan;
  result      

let call_with_output_file name proc =
  let out_ch = open_out name in
  let out_port = 
    make_output_port_to_channel out_ch
  in let result = proc out_port
  in
  close_out out_ch;
  result      

(************************************************************************)

let cur_inp_port_ = ref (make_input_port_from_channel stdin)
let cur_out_port_ = ref (make_output_port_to_channel stdout)
let cur_err_port_ = ref (make_output_port_to_channel stderr)

let current_input_port () = !cur_inp_port_
let current_output_port () = !cur_out_port_
let current_error_port () = !cur_err_port_

let with_input_from_port new_port thunk =
  let old_port = !cur_inp_port_ in
  cur_inp_port_ := new_port;
  try
    let result = thunk () in
    cur_inp_port_ := old_port;
    result
  with exn ->
    cur_inp_port_ := old_port;
    raise exn

let with_output_to_port new_port thunk =
  let old_port = !cur_out_port_ in
  cur_out_port_ := new_port;
  try
    let result = thunk () in
    cur_out_port_ := old_port;
    result
  with exn ->
    cur_out_port_ := old_port;
    raise exn

let with_error_to_port new_port thunk =
  let old_port = !cur_err_port_ in
  cur_err_port_ := new_port;
  try
    let result = thunk () in
    cur_err_port_ := old_port;
    result
  with exn ->
    cur_err_port_ := old_port;
    raise exn

let with_input_from_file name thunk =
  call_with_input_file name
    (fun in_port -> with_input_from_port in_port thunk)
let with_output_to_file name thunk =
  call_with_output_file name
    (fun out_port -> with_output_to_port out_port thunk)
let with_error_to_file name thunk =
  call_with_output_file name
    (fun out_port -> with_error_to_port out_port thunk)

(************************************************************************)

type parser_result =
    Valid of t
  | E_o_f

#include "sim_parser.ml"
#include "sim_lexer.ml"

let read () =
  try
    let result = 
      top_datum get_token (input_port_to_lexbuf (current_input_port ())) 
    in
    match result with
      Valid(x) -> x
    | E_o_f -> raise End_of_file
  with exn ->
    (match exn with
      Parsing.Parse_error -> raise Syntax_error
    | Lexing_error(_) -> raise Lexical_error
    | _ -> raise exn)

(************************************************************************)

let write_char c =
  match (current_output_port ()) with
    OutPortChannel(ch) -> output_char ch c
  | OutPortString(ch) -> soc_output_char ch c

let write_string s =
  match (current_output_port ()) with
    OutPortChannel(ch) -> output_string ch s
  | OutPortString(ch) -> soc_output_string ch s

let newline () = write_char '\n'

let output buff ofs len =
  match (current_output_port ()) with
    OutPortChannel(ch) -> Pervasives.output ch buff ofs len
  | OutPortString(ch) ->
      soc_output_string ch (String.sub buff ofs (ofs + len))
  
let flush () =
  match (current_output_port ()) with
    OutPortChannel(ch) -> Pervasives.flush ch
  | OutPortString(ch) -> ()

(************************************************************************)
(* Writing s-exprs							*)
(************************************************************************)

(* type used to control what action is taken by the pretty printer
   which printing a list. *)

type action_type =
    Action_list
  | Action_specform of int
  | Action_func of (Format.formatter -> bool -> t -> unit)


(* pretty_print_map is used to hold the declared actions for
   keywords. set_pretty_printing and get_pretty_printing are used to
   access the map. *)

let pretty_printing_map = ref String_map.empty

let set_pretty_printing sym (x:action_type) =
  pretty_printing_map := 
     String_map.add (symbol_to_string sym) x !pretty_printing_map

let get_pretty_printing sym = (
  try
    String_map.find (symbol_to_string sym) !pretty_printing_map
  with Not_found ->
    Action_list
  : action_type)

(* A symbol whose printed representation is ".". Used by
   improper_to_proper to convert an improper list into a proper list
   whose printed representation is equivalent.

   This is a hack: This only works because "." is not a legal symbol
   name. *)
let dot_sym = intern "."

let rec improper_to_proper e =
  if null_p e then
    []
  else if pair_p e then
    (car e)::(improper_to_proper (cdr e))
  else
    [dot_sym;e]


(* write and pretty_print. Entry points for output. *)
(* todo: why does write pretty print? It shouldn't... *)
let rec write e =    format_s_expr false false e
and display e =      format_s_expr true  false e
and pretty_print e = format_s_expr false true  e

(* format_s_expr is the top-level of the pretty printer. *)
and format_s_expr display_p pretty_p e =
  begin
    (* Construct the appropriate formatter. *)
    let fmt = Format.make_formatter output flush
    in

    (* If not pretty printing, just output on a _really_ big page. *)
    if not pretty_p then begin
      Format.pp_set_margin fmt (999999999-1);
      Format.pp_set_max_boxes fmt 999999999;
    end;

    (* Format e on fmt. *)
    format fmt display_p e;
    (* And flush. *)
    Format.pp_print_flush fmt ()
  end

(* Routine for pp'ing e on fmt. *)
and format fmt display_p e =
  Format.pp_open_hbox fmt ();
  (match e with
  | (Nil|Pair(_)) -> 
      (let action = 
	match e with
	  Nil -> Action_list
	| Pair({hd=(Symbol(_) as x)}) -> get_pretty_printing x
	| _ -> Action_list
      in
      match action with
	Action_func(f) -> f fmt display_p e
      | Action_list -> format_list fmt display_p e 
      | Action_specform(i) -> format_specform fmt display_p e i)
  | Vector(v) ->
      Format.pp_open_hovbox fmt 1;
      Format.pp_print_string fmt "#(";
      for i = 0 to ((Array.length v)-1) do
	if i > 0 then Format.pp_print_space fmt ();
	format fmt display_p v.(i)
      done;
      Format.pp_print_string fmt ")";
      Format.pp_close_box fmt ()
  | Bool(b) ->
      Format.pp_print_string fmt (if b then "#t" else "#f")
  | Char(c) ->
      if display_p then
	Format.pp_print_char fmt c
      else if c = ' ' then
	Format.pp_print_string fmt "#\\space"
      else if c = '\n' then
	Format.pp_print_string fmt "#\\newline"
      else begin
	Format.pp_print_string fmt "#\\";
	  Format.pp_print_char fmt c;
      end
  | String(s) ->
      if display_p then
	Format.pp_print_string fmt s
      else
	begin
	  Format.pp_print_char fmt '"';
	  Format.pp_print_string fmt (String.escaped s);
	  Format.pp_print_char fmt '"'
	end
  | Symbol(s) -> Format.pp_print_string fmt (symbol_to_string e)
  | Int(i) -> Format.pp_print_int fmt i
  | Float(f) -> 
      (* fixme: 1.0 is printed as "1", which is wrong! *)
      Format.pp_print_float fmt f
  | Procedure(_) -> Format.pp_print_string fmt "#<procedure>"
  | Input_port(_) ->Format.pp_print_string fmt "#<input port>"
  | Output_port(_) ->Format.pp_print_string fmt "#<output port>"
  | Eof -> Format.pp_print_string fmt "#<eof>"
  | Unspecified -> Format.pp_print_string fmt "#<unspecified>");
  Format.pp_close_box fmt ()

(* Routine for formatting a list using the Action_list rule. *)
and format_list fmt display_p e =
  match e with
    Nil -> Format.pp_print_string fmt "()"
  | Pair({hd=e1;tl=e2}) ->
      let indent =
	if symbol_p e1 then
	  2 + String.length (symbol_to_string e1)
	else
	  1
      in
      Format.pp_open_box fmt indent;
      Format.pp_print_string fmt "(";
      format fmt display_p e1;
      List.iter
	(fun e' ->
	    Format.pp_print_space fmt ();
	    format fmt display_p e')
	(improper_to_proper e2);
      Format.pp_print_string fmt ")";
      Format.pp_close_box fmt ()
  | _ -> raise (Invalid_argument("format_list"))

(* Routine for formatting a list using the Action_specform(i) rule. *)
and format_specform fmt display_p e i =
  let e' = improper_to_proper e in
  Format.pp_open_vbox fmt 2;
  Format.pp_print_string fmt "(";
  let rec iter i e' =
    if i = 0 then
      e'
    else if e' = [] then
      []
    else
      begin
	Format.pp_print_space fmt ();
	format fmt display_p (List.hd e');
	iter (i-1) (List.tl e')
      end
  in
  Format.pp_open_hovbox fmt 4;
  format fmt display_p (List.hd e');
  let rest = iter i (List.tl e') in
  Format.pp_close_box fmt ();
  List.iter
    (fun x ->
      Format.pp_print_space fmt ();
      format fmt display_p x)
    rest;
  Format.pp_print_string fmt ")";
  Format.pp_close_box fmt ()

(* What follows is code to set up the pretty-printer to format Scheme
   code correctly. *)

let set_abbrev sym_name str =
  set_pretty_printing (intern sym_name)
    (Action_func(
     fun fmt display_p e ->
       match e with
	 Pair({hd=Symbol(_);tl=Pair({hd=e';tl=Nil})}) ->
	   Format.pp_print_string fmt str;
	   format fmt display_p e'
       | _ -> format_list fmt display_p e))

let _ = set_abbrev "quote" "'"
let _ = set_abbrev "quasiquote" "`"
let _ = set_abbrev "unquote" ","
let _ = set_abbrev "unquote-splicing" ",@"

let set_specform str i =
  set_pretty_printing (intern str)
    (Action_specform(i))

let _ = set_specform "begin" 0
let _ = set_specform "let" 1
let _ = set_specform "do" 2
let _ = set_specform "case" 1
let _ = set_specform "lambda" 1
let _ = set_specform "let*" 1
let _ = set_specform "letrec" 1
let _ = set_specform "define" 1

