(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
exception Syntax_error
exception Lexical_error

(************************************************************************)

(* Routines for converting ML values to Scheme I/O ports. *)

val make_input_port_from_channel : in_channel -> input_port
val make_input_port_from_string : string -> input_port
val make_output_port_to_channel : out_channel -> output_port
(* The result is a tuple of two values: the new output port, and a thunk
   for retrieving a string containing everything written to that port. *)
val make_output_port_to_string :  unit -> (output_port * (unit -> string))

val close_input_port : input_port -> unit
val close_output_port : output_port -> unit

(* These routines behave similarly, but not identical, to their Scheme
   equivalents. A port to the file "name" is created and passed to "proc".
   This port is closed if control leaves "proc", either by returning
   normally or by throwing an exception. *)
val call_with_input_file : (*name*)string -> (*proc*)(input_port -> 'a) -> 'a
val call_with_output_file : (*name*)string -> (*proc*)(output_port -> 'a) -> 'a

(************************************************************************)

(* The "default" ports. *)
val current_input_port : unit -> input_port
val current_output_port : unit -> output_port
val current_error_port : unit -> output_port

(* Routines to change the "default" port for the duration of the execution
   of "thunk". *)
val with_input_from_port : input_port -> (*thunk*)(unit -> 'a) -> 'a
val with_output_to_port : output_port -> (*thunk*)(unit -> 'a) -> 'a
val with_error_to_port : output_port -> (*thunk*)(unit -> 'a) -> 'a

(************************************************************************)

(* Read an S-expr from the default input port. *)
val read : unit -> t

(************************************************************************)

(* Write text to the default output port. *)
val write_string : string -> unit
val write_char : char -> unit
val newline : unit -> unit

(************************************************************************)

(* Write S-expr's to the default output port. Consult R4RS for the
   differences between write and display. pretty_print is write with fancy
   indenting. *)
val write : t -> unit
val display : t -> unit
val pretty_print : t -> unit

(* Controlling Pretty printing *)

(* type used to control what action is taken by the pretty printer
   which printing a list.

   Action_list - Use the list format.

       (symbol-name adsf adf asdf
		    asdf asdf asdf asdf)

       ((another list) asdf asdf asdf
	adf asdf asdf)

   Action_specform(i) - Print as a special form. The first i forms
   after the keyword will be indented 4 spaces. The rest of the forms
   will be indented 2.

       ; Action_specform(0)
       (begin
	 a
	 b)

       ; Action_specform(1)
       (let
	   a
	b
	c)

       ; Action_specform(2)
       (do
	   a
	   b
	c
	d)


   Action_func(f) - Calls routine f to do the pretty printing.
   *)

type action_type =
    Action_list
  | Action_specform of int
  | Action_func of (Format.formatter -> bool -> t -> unit)

(* Routine for setting and getting the pp action for keywords. *)
val set_pretty_printing : t -> action_type -> unit
val get_pretty_printing : t -> action_type

(* The following routines might be useful when writing a Action_func
   function. *)

(* Format an s-expr. pretty_print is a wrapper to this routine. 
   The "bool" indictes whether or not "display" rules are being used. *)
val format : Format.formatter -> bool -> t -> unit

(* Format a list using the Action_list rule.
   The "bool" indictes whether or not "display" rules are being used. *)
val format_list : Format.formatter -> bool -> t -> unit

(* Format a list using the Action_specform(i) rule.
   The "bool" indictes whether or not "display" rules are being used. *)
val format_specform  : Format.formatter -> bool -> t -> int -> unit

