(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
{

type lexer_state = Lex_bof | Lex_prefix | Lex_infix;;
let current_lex_state = ref Lex_prefix;;
let set_lex_state st = current_lex_state := st


exception Lexing_error of int

(* String parsing swiped from CL71's scanner.mll *)

let initial_string_buffer = String.create 256
let string_buff = ref initial_string_buffer
let string_index = ref 0

let reset_string_buffer () =
  string_buff := initial_string_buffer;
  string_index := 0

let store_string_char c =
  (if !string_index >= String.length (!string_buff) then
    let new_buff = String.create (String.length (!string_buff) * 2) in
      String.blit new_buff 0 (!string_buff) 0 (String.length (!string_buff));
      string_buff := new_buff;
      ());
  String.set (!string_buff) (!string_index) c;
  incr string_index

let get_stored_string () =
  let s = String.sub (!string_buff) 0 (!string_index) in
    string_buff := initial_string_buffer;
    s

} 

rule token = parse
  [ ' ' '\t' '\n' ]+	{ token lexbuf }
| ';' [^ '\n']*		{ token lexbuf }
| '('			{ BO_LIST }
| "#("			{ BO_VECT }
| ')'			{ EO_SEQ }
| '\''			{ QUOTE }
| '`'			{ QUASIQUOTE }
| ','			{ UNQUOTE }
| ",@"			{ UNQUOTE_SPLICING }
| '.'			{ DOT }
| ['a'-'z''A'-'Z''!''$''%''&''*''/'':''<''=''>''?''~''_''^']
    ['a'-'z''A'-'Z''0'-'9''!''$''%''&''*''/'':''<''=''>''?''~''_''^'
	'.''+''-']*	{ SYMBOL(Lexing.lexeme lexbuf) }
| ('+'|'-'|"...")	{ SYMBOL(Lexing.lexeme lexbuf) }
| "#\\"_		{ CHAR(Lexing.lexeme_char lexbuf 2) }
| "#\\"['a'-'z''A'-'Z']+ {
    let name = StdCase.f (Lexing.lexeme lexbuf) in
    CHAR(
      if name = "#\\space" then
        ' '
      else if name = "#\\newline" then
	'\n'
      else if name = "#\\tab" then
	'\t'
      else
	raise (Lexing_error(0))
	  ) }
| '"' {
    reset_string_buffer();
    string lexbuf;
    STRING(get_stored_string()) 
    }
| '#' ['t' 'T' 'f' 'F']	{
  let name = StdCase.f (Lexing.lexeme lexbuf) in
  if name = "#t" then
    BOOL(true)
  else if name = "#f" then
    BOOL(false)
  else
    raise (Lexing_error(1))
} 
| '-'?['0'-'9']+	{ INT(int_of_string (Lexing.lexeme lexbuf)) }
| '-'?['0'-'9']+('.'['0'-'9']*)?(['e''E']['+''-']?['0'-'9']+)?
    			{ FLOAT(float_of_string (Lexing.lexeme lexbuf)) }
| eof			{ EOF }
| '{'                   { set_lex_state Lex_infix; INF_START }
| _			{ raise (Lexing_error(2)) }


and infix_token = parse
 [ ' ' '\t' '\n' ]+	{ infix_token lexbuf }
| ';' [^ '\n']*		{ infix_token lexbuf }
| '}'                   { set_lex_state Lex_prefix; INF_END }
| '('			{ OPENP }
| ')'                   { CLOSEP }
| '['			{ OPENB }
| ']'			{ CLOSEB }
| '+'                   { PLUS }
| '-'                   { MINUS }
| '*'                   { TIMES }
| '/'                   { DIV }
| ','                   { COMMA }
| '<'                   { LT }
| '>'                   { GT }
| "<="                  { LE }
| ">="                  { GE }
| '='                   { EQ }
| "&&"                  { AND }
| "||"                  { OR }
| '!'                   { NOT }
| ['a'-'z''A'-'Z''_''$']
    ['a'-'z''A'-'Z''0'-'9''_''$''\'''.']*
	                { SYMBOL(Lexing.lexeme lexbuf) }
| "#\\"_		{ CHAR(Lexing.lexeme_char lexbuf 2) }
| "#\\"['a'-'z''A'-'Z']+ {
    let name = StdCase.f (Lexing.lexeme lexbuf) in
    CHAR(
      if name = "#\\space" then
        ' '
      else if name = "#\\newline" then
	'\n'
      else
	raise (Lexing_error(0))
	  ) }
| '"' {
    reset_string_buffer();
    string lexbuf;
    STRING(get_stored_string()) 
    }
| '#' ['t' 'T' 'f' 'F']	{
  let name = StdCase.f (Lexing.lexeme lexbuf) in
  if name = "#t" then
    BOOL(true)
  else if name = "#f" then
    BOOL(false)
  else
    raise (Lexing_error(1))
} 
| ['0'-'9']+	{ INT(int_of_string (Lexing.lexeme lexbuf)) }
| ['0'-'9']+('.'['0'-'9']*)?(['e''E']['+''-']?['0'-'9']+)?
    			{ FLOAT(float_of_string (Lexing.lexeme lexbuf)) }
| eof                   { EOF }
| _			{ raise (Lexing_error(2)) }
    

and string = parse
  '"' { () }
| '\\' ['\\' '"' 't' 'b' 'r' 'n']
	   {
	 let c = (Lexing.lexeme_char lexbuf 1)
	 in let c' =
	   match c with
	     't' -> '\t'
	   | 'b' -> '\b'
	   | 'r' -> '\r'
	   | 'n' -> '\n'
	   | _ -> c
	 in
	 store_string_char(c');
	 string lexbuf
       }
| '\\' ['0'-'9'] ['0'-'9'] ['0'-'9'] {
  store_string_char(Char.chr
		      (64*(Char.code(Lexing.lexeme_char lexbuf 1))
			 + 8*(Char.code(Lexing.lexeme_char lexbuf 2))
			 +   (Char.code(Lexing.lexeme_char lexbuf 3))));
	 string lexbuf
}
| '\\' { raise (Lexing_error(3)) }
| eof { raise (Lexing_error(4)) }
| _ {
  store_string_char(Lexing.lexeme_char lexbuf 0);
  string lexbuf
       }
    
and bof_token = parse
    ( "#!" [^ '\n']* )? { 
  	set_lex_state Lex_prefix; 
  	token lexbuf
    }

{

let get_token lexbuf =
  match !current_lex_state with
    Lex_prefix -> token lexbuf
  | Lex_bof -> bof_token lexbuf
  | Lex_infix -> infix_token lexbuf

} 
