(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
let gensym_count = ref 0
let gensym tag =
  let i = !gensym_count in
  gensym_count := !gensym_count + 1;
  intern ("*" ^ tag ^ "-" ^ (string_of_int i) ^ "*")

let tagify_bool1 f x = tag_bool (f x)
let tagify_bool2 f x y = tag_bool (f x y)
let tagify_unit2 f x y = f x y; unspecified

(************************************************************************)
(* Essential macros							*)
(************************************************************************)

(* todo: quasiquote, do *)

let and_ = intern "and"
let begin_ = intern "begin"
let cond_ = intern "cond"
let else_ = intern "else"
let if_ = intern "if"
let key_ = intern "key"
let lambda_ = intern "lambda"
let let_ = intern "let"
let letrec_ = intern "letrec"
let memv_ = intern "memv"
let or_ = intern "or"
let quote_ = intern "quote"
let thunk_ = intern "thunk"
let x_ = intern "x"

let protect_expansion expr f =
  try
    f ()
  with Invalid_argument(_) ->
    raise (Invalid_form(expr))

let expander_cond expr =
  protect_expansion expr
    (fun () ->
      let rec expand_clauses clauses =
      	if null_p clauses then
  	  (list1 begin_)
      	else if (car (car clauses)) = else_ then
  	  (cons begin_ (cdr (car clauses)))
      	else
  	  (list4 if_ (car (car clauses))
	     (cons begin_ (cdr (car clauses)))
	     (expand_clauses (cdr clauses)))
      in
      expand_clauses (cdr expr))
let _ = define_macro scheme_env "cond" expander_cond

let expander_case expr =
  protect_expansion expr
    (fun () ->
      let key_val = (cadr expr) and
	  clauses = (cddr expr) in
      let clause_bindings =
	(map
	   (fun clause ->
	     (list2 (gensym "thunk")
		(cons lambda_ (cons nil (cdr clause)))))
	   clauses)
      in
      (list3 let_ (cons (list2 key_ key_val) clause_bindings)
	 (cons cond_
	    (map2
	       (fun clause binding ->
		 let test = (car clause) and
		     thunk_ = (car binding) in
		 if test = else_ then
		   (list2 else_ (list1 thunk_))
		 else
		   (list2 (list3 memv_ key_ (list2 quote_ test))
		      (list1 thunk_)))
	       clauses clause_bindings))))
let _ = define_macro scheme_env "case" expander_case

let expander_and expr =
  protect_expansion expr
    (fun () ->
      let rec expand_clauses clauses =
	if null_p clauses then
	  sharp_t
	else if null_p (cdr clauses) then
	  (car clauses)
	else
	  let test = (car clauses) and
	      rest_test = (cdr clauses) in
	  (list3 let_ (list2 
			    (list2 x_ test)
			    (list2 thunk_
			       (list3 lambda_ nil 
				  (cons and_ rest_test))))
	     (list4 if_ x_ (list1 thunk_) x_))
      in
      expand_clauses (cdr expr))
let _ = define_macro scheme_env "and" expander_and

let expander_or expr =
  protect_expansion expr
    (fun () ->
      let rec expand_clauses clauses =
	if null_p clauses then
	  sharp_f
	else if null_p (cdr clauses) then
	  (car clauses)
	else
	  let test = (car clauses) and
	      rest_test = (cdr clauses) in
	  (list3 let_ (list2 
			    (list2 x_ test)
			    (list2 thunk_
			       (list3 lambda_ nil 
				  (cons or_ rest_test))))
	     (list4 if_ x_ (list1 thunk_) x_))
      in
      expand_clauses (cdr expr))
let _ = define_macro scheme_env "or" expander_or

let expander_let expr =
  protect_expansion expr
    (fun () ->
      (* named let *)
      if symbol_p (cadr expr) then
	begin
	  let name = (cadr expr) and
	      bindings = (caddr expr) and
	      body_forms = (cdddr expr) in
	  let names = (map car bindings) and
	      values = (map cadr bindings) in
	  (cons (list3 letrec_ 
		   (list1 
		      (list2 name 
			 (cons lambda_ (cons names body_forms))))
		   name)
	     values)
	end
      (* unnamed let *)
      else
	begin
	  let bindings = (cadr expr) and
	      body_forms = (cddr expr) in
	  let names = (map car bindings) and
	      values = (map cadr bindings) in
	  (cons (cons lambda_ (cons names body_forms)) values)
	end)
let _ = define_macro scheme_env "let" expander_let

let expander_let_star expr =
  protect_expansion expr
    (fun () ->
      let rec recurs bindings body_forms =
      	if null_p bindings then
	  (cons begin_ body_forms)
	else
	  (list2 (list3 lambda_ (list1 (car (car bindings)))
		    (recurs (cdr bindings) body_forms))
	     (cadr (car bindings)))
      in
      recurs (cadr expr) (cddr expr))
let _ = define_macro scheme_env "let*" expander_let_star

(************************************************************************)
(* 6.1 Booleans								*)
(************************************************************************)

let prim_not e =
  if e = sharp_f then
    sharp_t
  else
    sharp_f
let _ = define_prim1 scheme_env "not" prim_not

(************************************************************************)
(* 6.2 Equivalence Predicates						*)
(************************************************************************)

(* todo: eqv? *)

let _ = define_prim2 scheme_env "eq?" (tagify_bool2 eq_p)
let _ = define_prim2 scheme_env "eqv?" (tagify_bool2 eqv_p)
let _ = define_prim2 scheme_env "equal?" (tagify_bool2 equal_p)

(************************************************************************)
(* 6.3 Pairs and lists							*)
(************************************************************************)

let _ = define_prim1 scheme_env "pair?" (tagify_bool1 pair_p)
let _ = define_prim2 scheme_env "cons" cons
let _ = define_prim1 scheme_env "car" car
let _ = define_prim1 scheme_env "cdr" cdr
let _ = define_prim2 scheme_env "set-car!" (tagify_unit2 set_car)
let _ = define_prim2 scheme_env "set-cdr!" (tagify_unit2 set_cdr)
let _ = define_prim1 scheme_env "cadr" cadr
let _ = define_prim1 scheme_env "cddr" cddr
let _ = define_prim1 scheme_env "caddr" caddr
let _ = define_prim1 scheme_env "cdddr" cdddr
let _ = define_prim1 scheme_env "cadddr" cadddr
let _ = define_prim1 scheme_env "cddddr" cddddr
let _ = define_prim1 scheme_env "caddddr" caddddr
let _ = define_prim1 scheme_env "cdddddr" cdddddr
let _ = define_prim1 scheme_env "null?" (tagify_bool1 null_p)
let _ = define_prim1 scheme_env "list?" (tagify_bool1 list_p)
let _ = define_primn scheme_env "list" (fun lst -> lst)

let prim_length x =
  let len = length x in
  if (not (null_p x) & not (pair_p x)) or (len < 0) then
    raise (Bad_arg("length",x))
  else
    tag_int (len)
let _ = define_prim1 scheme_env "length" prim_length

let rec prim_append args =
  if (null_p args) then
      nil
  else
    (append (car args) (prim_append (cdr args)))
let _ = define_primn scheme_env "append" prim_append

let _ = define_primn scheme_env "reverse" reverse
let _ = define_prim2 scheme_env "list-tail"
	(fun x k -> list_tail x (untag_int k))
let _ = define_prim2 scheme_env "list-ref"
	(fun x k -> list_ref x (untag_int k))
let _ = define_prim2 scheme_env "memq" memq
let _ = define_prim2 scheme_env "memv" memv
let _ = define_prim2 scheme_env "member" member
let _ = define_prim2 scheme_env "assq" assq
let _ = define_prim2 scheme_env "assv" assv
let _ = define_prim2 scheme_env "assoc" assoc

(************************************************************************)
(* 6.4 Symbols								*)
(************************************************************************)

let _ = define_prim1 scheme_env "symbol?" (tagify_bool1 symbol_p)
let _ = define_prim1 scheme_env "symbol->string"
    (fun sym -> tag_string (symbol_to_string sym))
let _ = define_prim1 scheme_env "string->symbol"
    (fun x -> string_to_symbol (untag_string x))

(************************************************************************)
(* 6.5 Numbers								*)
(************************************************************************)

let number_p x = int_p x or float_p x
let _ = define_prim1 scheme_env "number?" (tagify_bool1 number_p)
let _ = define_prim1 scheme_env "real?" (tagify_bool1 number_p)

let integer_p x =
  if (int_p x) then
    true
  else if (float_p x) then
    let f = (untag_float x) in
    (ceil f) = f
  else
    false
let _ = define_prim1 scheme_env "integer?" (tagify_bool1 integer_p)

let exact_p x = 
  match x with
    Int(_) -> true
  | Float(_) -> false
  | _ -> raise (Bad_arg("exact_p",x))
let _ = define_prim1 scheme_env "exact?" (tagify_bool1 exact_p)

let inexact_p x = not (exact_p x)
let _ = define_prim1 scheme_env "inexact?" (tagify_bool1 inexact_p)

let nary_comp_op name int_op float_op args =
  let comp x y =
    match (x,y) with
      (Int(a),Int(b)) -> int_op a b
    | (Float(a),Int(b)) -> float_op a (float b)
    | (Int(a),Float(b)) -> float_op (float a) b
    | (Float(a),Float(b)) -> float_op a b
    | _ -> raise (Bad_arg(name, list2 x y))
  in
  if null_p args then
    sharp_t
  else
    let rec iter x lst =
      if null_p lst then
	sharp_t
      else
      	let y = (car lst) in
      	if comp x y then
	  iter y (cdr lst)
      	else
	  sharp_f
    in
    iter (car args) (cdr args)
let _ = define_primn scheme_env "=" (nary_comp_op "=" (=) (=))
let _ = define_primn scheme_env "<" (nary_comp_op "<" (<) (<))
let _ = define_primn scheme_env ">" (nary_comp_op ">" (>) (>))
let _ = define_primn scheme_env "<=" (nary_comp_op "<=" (<=) (<=))
let _ = define_primn scheme_env ">=" (nary_comp_op ">=" (>=) (>=))

let unary_arith_op name int_op float_op x =
  match x with
    Int(a) -> tag_int (int_op a)
  | Float(a) -> tag_float (float_op a)
  | _ -> raise (Bad_arg(name, x))

let binary_arith_op name int_op float_op x y =
  match (x,y) with
    (Int(a),Int(b)) -> tag_int (int_op a b)
  | (Float(a),Int(b)) -> tag_float (float_op a (float b))
  | (Int(a),Float(b)) -> tag_float (float_op (float a)  b)
  | (Float(a),Float(b)) -> tag_float (float_op a b)
  | _ -> raise (Bad_arg(name, list2 x y))

let prim_plus args = 
  fold_left (binary_arith_op "+" (+) (+.)) (tag_int (0)) args
let _ = define_primn scheme_env "+" prim_plus

let prim_times args = 
  fold_left (binary_arith_op "*" ( * ) ( *. )) (tag_int (1)) args
let _ = define_primn scheme_env "*" prim_times

let prim_minus args =
  match length args with
    0 -> tag_int (0)
  | 1 -> unary_arith_op "-" (~-) (~-.) (car args)
  | _ ->
      fold_left (binary_arith_op "-" (-) (-.)) (car args) (cdr args)
let _ = define_primn scheme_env "-" prim_minus

let prim_divide args =
  try 
    match length args with
      0 -> tag_int (1)
    | 1 -> unary_arith_op "/" ((/) 1) ((/.) 1.0) (car args)
    | _ ->
      	fold_left (binary_arith_op "/" (/) (/.)) (car args) (cdr args)
  with Division_by_zero ->
    raise (Bad_arg("/",tag_int (0)))
let _ = define_primn scheme_env "/" prim_divide

let prim_abs x =
  match x with
    Int(a) -> tag_int (abs a)
  | Float(a) -> tag_float (abs_float a)
  | _ -> raise (Bad_arg("abs",x))
let _ = define_prim1 scheme_env "abs" prim_abs

(* todo: fix these so that they work for negative args. *)
let prim_quotient a b = 
  binary_arith_op "quotient" (/) (fun a b -> floor (a /. b)) a b
let _ = define_prim2 scheme_env "quotient" prim_quotient

let prim_remainder a b = 
  binary_arith_op "remainder" (mod) mod_float a b
let _ = define_prim2 scheme_env "remainder" prim_remainder

let prim_modulo a b = 
  binary_arith_op "modulo" (mod) mod_float a b
let _ = define_prim2 scheme_env "modulo" prim_modulo

let identity x = x
let ftruncate x = float (truncate x)

let _ = define_prim1 scheme_env "floor"
 	(unary_arith_op "floor" identity floor)
let _ = define_prim1 scheme_env "ceiling"
 	(unary_arith_op "ceiling" identity ceil)
(* fix the next two *)
let _ = define_prim1 scheme_env "truncate" 
	(unary_arith_op "truncate" identity ftruncate)
let _ = define_prim1 scheme_env "round" 
	(unary_arith_op "round" identity ftruncate)

let unary_inexact_arith_op name op x =
  match x with
    Int(a) -> tag_float (op (float a))
  | Float(a) -> tag_float (op a)
  | _ -> raise (Bad_arg(name,x))

(* fix all of these *)
let _ = define_prim1 scheme_env "exp" (unary_inexact_arith_op "exp" exp)
let _ = define_prim1 scheme_env "log" (unary_inexact_arith_op "log" log)
let _ = define_prim1 scheme_env "sin" (unary_inexact_arith_op "sin" sin)
let _ = define_prim1 scheme_env "cos" (unary_inexact_arith_op "cos" cos)
let _ = define_prim1 scheme_env "tan" (unary_inexact_arith_op "tan" tan)
let _ = define_prim1 scheme_env "asin" (unary_inexact_arith_op "asin" asin)
let _ = define_prim1 scheme_env "acos" (unary_inexact_arith_op "acos" acos)

let prim_atan args =
  let rec f lst =
    match lst with
      (Int(a))::rest -> f ((tag_float (float a))::rest)
    |	[a;Int(b)] -> f [a;tag_float (float b)]
    |	[Float(a)] -> tag_float (atan a)
    |	[Float(a);Float(b)] -> tag_float (atan2 a b)
    |	 _ -> raise (Bad_arg("atan",args))
  in
  f (to_list args)
let _ = define_primn scheme_env "atan" prim_atan

let _ = define_prim1 scheme_env "sqrt" (unary_inexact_arith_op "sqrt" sqrt)

let prim_exact_to_inexact x =
  match x with
    Int(a) -> tag_float (float a)
  | Float(_) -> x
  | _ -> raise (Bad_arg("exact->inexact",x))
let _ = define_prim1 scheme_env "exact->inexact" prim_exact_to_inexact

let prim_inexact_to_exact x =
  match x with
    Int(_) -> x
  | Float(a) -> tag_int (truncate a)
  | _ -> raise (Bad_arg("exact->inexact",x))
let _ = define_prim1 scheme_env "inexact->exact" prim_inexact_to_exact

(* todo: fix this. radix argument. *)
let prim_number_to_string x =
  match x with
    Int(a) -> tag_string (string_of_int a)
  | Float(a) -> tag_string (string_of_float a)
  | _ -> raise (Bad_arg("number->string",x))
let _ = define_prim1 scheme_env "number->string" prim_number_to_string

(* todo: fix this. do real parsing? *)
let prim_string_to_number x =
  let s = untag_string x in
  try
    tag_int (int_of_string s)
  with Failure("int_of_string") ->
    tag_float (float_of_string s)
let _ = define_prim1 scheme_env "string->number" prim_string_to_number

(************************************************************************)
(* 6.6 Characters							*)
(************************************************************************)

let _ = define_prim1 scheme_env "char?" (tagify_bool1 char_p)

let char_compare a b = 
  (Char.code (untag_char a)) - (Char.code (untag_char b))

let _ = define_prim2 scheme_env "char=?" 
    (fun a b -> tag_bool (0 = (char_compare a b)))
let _ = define_prim2 scheme_env "char<?" 
    (fun a b -> tag_bool (0 < (char_compare a b)))
let _ = define_prim2 scheme_env "char>?" 
    (fun a b -> tag_bool (0 > (char_compare a b)))
let _ = define_prim2 scheme_env "char<=?" 
    (fun a b -> tag_bool (0 <= (char_compare a b)))
let _ = define_prim2 scheme_env "char>=?" 
    (fun a b -> tag_bool (0 >= (char_compare a b)))
  
let prim_char_to_integer c =
  tag_int (Char.code (untag_char c))
let _ = define_prim1 scheme_env "char->integer" prim_char_to_integer

let prim_integer_to_char c =
  tag_char (Char.chr (untag_int c))
let _ = define_prim1 scheme_env "integer->char" prim_integer_to_char

let prim_char_upcase c = 
  tag_char (Char.uppercase (untag_char c))
let _ = define_prim1 scheme_env "char-upcase" prim_char_upcase

let prim_char_downcase c = 
  tag_char (Char.lowercase (untag_char c))
let _ = define_prim1 scheme_env "char-downcase" prim_char_downcase

(************************************************************************)
(* 6.7 Strings								*)
(************************************************************************)

let _ = define_prim1 scheme_env "string?" (tagify_bool1 string_p)

let prim_make_string args =
  match to_list args with
    [Int(k)] -> tag_string (String.make k ' ')
  | [Int(k);Char(c)] -> tag_string (String.make k c)
  | _ -> raise (Bad_arg("make-string",args))
let _ = define_primn scheme_env "make-string" prim_make_string

let _ = define_prim1 scheme_env "string-length"
	(fun s -> tag_int (String.length (untag_string s)))
let _ = define_prim2 scheme_env "string-ref"
	(fun s k -> 
	  tag_char (String.get (untag_string s)
		 (untag_int k)))
let _ = define_prim3 scheme_env "string-set!"
	(fun s k c -> 
	  String.set (untag_string s)
	    (untag_int k) (untag_char c);
	  unspecified)

let string_compare a b =
  let a' = untag_string a and
      b' = untag_string b in
  if a == b or a' == b' or a' = b' then
    0
  else if a < b then
    -1
  else
    1   
let _ = define_prim2 scheme_env "string=?" 
    (fun a b -> tag_bool (0 = (string_compare a b)))
let _ = define_prim2 scheme_env "string<?" 
    (fun a b -> tag_bool (0 < (string_compare a b)))
let _ = define_prim2 scheme_env "string>?" 
    (fun a b -> tag_bool (0 > (string_compare a b)))
let _ = define_prim2 scheme_env "string<=?" 
    (fun a b -> tag_bool (0 <= (string_compare a b)))
let _ = define_prim2 scheme_env "string>=?" 
    (fun a b -> tag_bool (0 >= (string_compare a b)))

let _ = define_prim3 scheme_env "substring"
	(fun s st en ->
	  tag_string (String.sub (untag_string s) (untag_int st)
		   (untag_int en)))
let _ = define_primn scheme_env "string-append"
	(fun args ->
	  tag_string (String.concat "" (List.map untag_string (to_list args))))
let _ = define_prim1 scheme_env "string-copy"
	(fun s -> tag_string (String.copy (untag_string s)))
let _ = define_prim2 scheme_env "string-fill!"
	(fun s c ->
	  let s' = untag_string s and c' = untag_char c in
	  String.fill s' 0 (String.length s') c';
	  unspecified)

(************************************************************************)
(* 6.8 Vectors								*)
(************************************************************************)

let _ = define_prim1 scheme_env "vector?" (tagify_bool1 vector_p)

let prim_make_vector args =
  match to_list args with
    [Int(k)] -> tag_vector (Array.create k unspecified)
  | [Int(k);v] -> tag_vector (Array.create k v)
  | _ -> raise (Bad_arg("make-vector",args))
let _ = define_primn scheme_env "make-vector" prim_make_vector

let _ = define_prim1 scheme_env "vector-length"
	(fun s -> tag_int (Array.length (untag_vector s)))
let _ = define_prim2 scheme_env "vector-ref"
	(fun s k -> Array.get (untag_vector s) (untag_int k))
let _ = define_prim3 scheme_env "vector-set!"
	(fun s k v -> 
	  Array.set (untag_vector s) (untag_int k) v;
	  unspecified)
let _ = define_prim2 scheme_env "vector-fill!"
	(fun s v ->
	  let s' = untag_vector s in
	  Array.fill s' 0 (Array.length s') v;
	  unspecified)

(************************************************************************)
(* 6.9 Control features							*)
(************************************************************************)

let _ = define_prim1 scheme_env "procedure?" (tagify_bool1 procedure_p)

let prim_apply args =
  let f = car args and args' = cdr args in
  let rec cons_star lst =
    if null_p (cdr lst) then
      (car lst)
    else
      (cons (car lst) (cons_star (cdr lst)))
  in
  apply f (cons_star args')
let _ = define_primn scheme_env "apply" prim_apply

let prim_map args =
  let f = car args and lsts = (cdr args)  in
  let rec recurs lsts =
    if null_p (first lsts) then
      nil
    else
      (cons (apply f (map car lsts))
	 (recurs (map cdr lsts)))
  in
  recurs lsts
let _ = define_primn scheme_env "map" prim_map

let prim_for_each args =
  let f = car args and lsts = (cdr args)  in
  let rec iter lsts =
    if null_p (first lsts) then
      unspecified
    else
      begin
      	(apply f (map car lsts));
    	(iter (map cdr lsts))
      end
  in
  iter lsts
let _ = define_primn scheme_env "for_each" prim_for_each

exception Scheme_error of string

let prim_error str =
  raise (Scheme_error(untag_string str))

let _ = define_prim1 scheme_env "error" prim_error

(* todo: force, delay, call-with-current-continuation *)

(************************************************************************)
(* 6.10 Input and output						*)
(************************************************************************)

let _ = define_prim2 scheme_env "call-with-input-file"
    (fun name proc ->
      call_with_input_file (untag_string name)
	(fun in_p -> funcall1 proc (tag_input_port (in_p))))
let _ = define_prim2 scheme_env "call-with-output-file"
    (fun name proc ->
      call_with_output_file (untag_string name)
	(fun out_p -> funcall1 proc (tag_output_port(out_p))))

let _ = define_prim1 scheme_env "input-port?" (tagify_bool1 input_port_p)
let _ = define_prim1 scheme_env "output-port?" (tagify_bool1 output_port_p)

let _ = define_prim0 scheme_env "current-input-port"
    (fun () -> tag_input_port(current_input_port ()))
let _ = define_prim0 scheme_env "current-output-port"
    (fun () -> tag_output_port(current_output_port ()))

let _ = define_prim2 scheme_env "with-input-from-file"
    (fun name thunk ->
      with_input_from_file (untag_string name)
	(fun () -> funcall0 thunk))
let _ = define_prim2 scheme_env "with-output-to-file"
    (fun name thunk ->
      with_output_to_file (untag_string name)
	(fun () -> funcall0 thunk))

let _ = define_prim1 scheme_env "open-input-file"
    (fun name ->
      tag_input_port(make_input_port_from_channel 
		       (open_in (untag_string name))))
let _ = define_prim1 scheme_env "open-output-file"
    (fun name ->
      tag_output_port(make_output_port_to_channel 
		    	(open_out (untag_string name))))

let _ = define_prim1 scheme_env "close-input-port"
    (fun in_port ->
      close_input_port (untag_input_port in_port); unspecified)
let _ = define_prim1 scheme_env "close-output-port"
    (fun out_port ->
      close_output_port (untag_output_port out_port); unspecified)

let prim_read args =
  match to_list args with
    [] -> read ()
  | [p] -> with_input_from_port (untag_input_port p) read
  | _ -> raise (Bad_arg("read",args))
let _ = define_primn scheme_env "read" prim_read

(* todo: read-char peek-char *)

let _ = define_prim1 scheme_env "eof-object?" 
    (tagify_bool1 (fun x -> eq_p x eof))

(* todo: char-ready? *)

let prim_write args =
  (match to_list args with
    [x] -> write x
  | [x;p] -> 
      with_output_to_port (untag_output_port p) 
	(fun () -> write x)
  |  _ -> raise (Bad_arg("write",args)));
  unspecified
let _ = define_primn scheme_env "write" prim_write

let prim_display args =
  (match to_list args with
    [x] -> display x
  | [x;p] -> 
      with_output_to_port (untag_output_port p) 
	(fun () -> display x)
  |  _ -> raise (Bad_arg("display",args)));
  unspecified
let _ = define_primn scheme_env "display" prim_display

let prim_pretty args =
  (match to_list args with
    [x] -> pretty_print x; newline ()
  | [x;p] -> 
      with_output_to_port (untag_output_port p) 
	(fun () -> pretty_print x; newline ())
  |  _ -> raise (Bad_arg("pretty_print",args)));
  unspecified
let _ = define_primn scheme_env "pretty" prim_pretty

let prim_newline args =
  (match to_list args with
    [] -> newline ()
  | [p] -> 
      with_output_to_port (untag_output_port p) newline
  |  _ -> raise (Bad_arg("newline",args)));
  unspecified
let _ = define_primn scheme_env "newline" prim_newline

let prim_write_char args =
  (match to_list args with
    [c] -> write_char (untag_char c)
  | [c;p] -> 
      with_output_to_port (untag_output_port p) 
	(fun () -> write_char (untag_char c))
  |  _ -> raise (Bad_arg("write-char",args)));
  unspecified
let _ = define_primn scheme_env "write-char" prim_write_char

let prim_load args =
  (match to_list args with
    [name] -> load (untag_string name) false
  | [name;verbose_p] -> 
      load (untag_string name) (untag_bool verbose_p)
  |  _ -> raise (Bad_arg("load",args)));
  unspecified
let _ = define_primn scheme_env "load" prim_load

(* todo: transcript-on transcript-off *)
