(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
exception Bad_arg of string * t

(************************************************************************)
(* Tag testing								*)
(************************************************************************)

let bool_p x = match x with Bool(_) -> true | _ -> false
let char_p x = match x with Char(_) -> true | _ -> false
let float_p x = match x with Float(_) -> true | _ -> false
let int_p x = match x with Int(_) -> true | _ -> false
let string_p x = match x with String(_) -> true | _ -> false
let pair_p x = match x with Pair(_) -> true | _ -> false
let symbol_p x = match x with Symbol(_) -> true | _ -> false
let vector_p x = match x with Vector(_) -> true | _ -> false
let procedure_p x = match x with Procedure(_) -> true | _ -> false
let input_port_p x = match x with Input_port(_) -> true | _ -> false
let output_port_p x = match x with Output_port(_) -> true | _ -> false

(************************************************************************)
(* Routines to produce tagged values					*)
(************************************************************************)

let nil = Nil
let sharp_t = Bool(true)
let sharp_f = Bool(false)
let tag_bool b = if b then sharp_t else sharp_f
let tag_char c = Char(c)
let tag_float f = Float(f)
let tag_int i = Int(i)
let tag_string s = String(s)
let tag_vector v = Vector(v)
let tag_procedure p = Procedure(p)
let tag_input_port p = Input_port(p)
let tag_output_port p = Output_port(p)
let eof = Eof
let unspecified = Unspecified

(************************************************************************)
(* Routines for untagging values					*)
(************************************************************************)

let untag_bool x =
  match x with
    Bool(y) -> y
  | _ -> raise (Bad_arg("untag_bool",x))

let untag_char x =
  match x with
    Char(y) -> y
  | _ -> raise (Bad_arg("untag_char",x))

let untag_float x =
  match x with
    Float(y) -> y
  | _ -> raise (Bad_arg("untag_float",x))

let untag_int x =
  match x with
    Int(y) -> y
  | _ -> raise (Bad_arg("untag_int",x))

let untag_string x =
  match x with
    String(y) -> y
  | _ -> raise (Bad_arg("untag_string",x))

let untag_vector x =
  match x with
    Vector(y) -> y
  | _ -> raise (Bad_arg("untag_vector",x))

let untag_procedure x =
  match x with
    Procedure(y) -> y
  | _ -> raise (Bad_arg("untag_procedure",x))

let untag_input_port x =
  match x with
    Input_port(y) -> y
  | _ -> raise (Bad_arg("untag_input_port",x))

let untag_output_port x =
  match x with
    Output_port(y) -> y
  | _ -> raise (Bad_arg("untag_output_port",x))


(************************************************************************)
(* Equality test							*)
(************************************************************************)

let eq_p x y = 
  match (x,y) with
    ((Bool(true),Bool(true))
  | (Bool(false),Bool(false))
  | (Nil,Nil)
  | (Unspecified,Unspecified)
  | (Eof,Eof)) ->
      if not (x == y) then
	failwith("eq_p: == invariance is broken! Someone broke abstraction!");
      true
  | _ -> (x == y)

let eqv_p x y =
  match (x,y) with
    (Int(a),Int(b)) -> a = b
  | (Float(a),Float(b)) -> a = b
  | (Char(a),Char(b)) -> a = b
  |  _ -> eq_p x y

let equal_p x y = (x = y)

(************************************************************************)
(* Primitive pairs							*)
(************************************************************************)

let untag_pair x =
  match x with
    Pair(y) -> y
  | _ -> raise (Bad_arg("untag_pair",x))

let cons e1 e2 = Pair({ hd = e1; tl = e2 })
let car e = (untag_pair e).hd
let cdr e = (untag_pair e).tl
let set_car e v = (untag_pair e).hd <- v
let set_cdr e v = (untag_pair e).tl <- v

(************************************************************************)
(* Primitive Symbols							*)
(************************************************************************)

module String_map = Map.Make(
  (* StringOrdering from Ssisl *)
  struct
    type t = string
    let compare = Pervasives.compare
  end)

let untag_symbol x =
  match x with
    Symbol(y) -> y
  | _ -> raise (Bad_arg("untag_symbol",x))

let symbol_to_string e = String.copy (untag_symbol e)
(* Symbol_to_string neither converts case nor maintains pointer invariance.
   YOU ALMOST CERTAINLY WANT TO USE intern INSTEAD OF symbol_to_string. 
   This routine is provided only as a means of implementing the Scheme
   primitive string->symbol. *)
let string_to_symbol s = Symbol(s)

(* Intern converts to the canonical case and maintains the pointer
   invariance of symbols. That is, (eq? (intern "a") (intern "a")) will
   always be true. *)
let symbols_ = ref String_map.empty
let intern s =
  try
    String_map.find s !symbols_
  with Not_found ->
    let sym = string_to_symbol (StdCase.f s) in
    symbols_ := String_map.add s sym !symbols_;
    sym

(************************************************************************)
(* Lists								*)
(************************************************************************)

let null_p x = eq_p x nil

let rec list_p e =
  if null_p e then
    true
  else if pair_p e then
    list_p (cdr e)
  else
    false

let rec to_list e =
  if null_p e then
    []
  else if pair_p e then
    (car e)::(to_list (cdr e))
  else
    raise (Invalid_argument("to_list"))

let rec from_list lst =
  match lst with
    [] -> nil
  | hd::tl -> cons hd (from_list tl)

let cadr x = car (cdr x)
let cddr x = cdr (cdr x)
let caddr x = car (cdr (cdr x))
let cdddr x = cdr (cdr (cdr x))
let cadddr x = car (cdr (cdr (cdr x)))
let cddddr x = cdr (cdr (cdr (cdr x)))
let caddddr x = car (cdr (cdr (cdr (cdr x))))
let cdddddr x = cdr (cdr (cdr (cdr (cdr x))))

let rec list_ref lst i =
  if i < 0 then
    raise (Invalid_argument("list_ref"))
  else if i = 0 then
    car lst
  else
    list_ref (cdr lst) (i - 1)

let rec list_tail lst i =
  if i < 0 then
    raise (Invalid_argument("list_tail"))
  else if i = 0 then
    lst
  else
    list_tail (cdr lst) (i - 1)


let list_ref lst i =
  try
    (car (list_tail lst i))
  with (Invalid_argument(_)) ->
    raise (Invalid_argument("list_tail"))

let rec map f lst =
  if null_p lst then
    nil
  else if pair_p lst then
    cons (f (car lst)) (map f (cdr lst))
  else
    raise (Invalid_argument("map"))

let rec map2 f lst1 lst2 =
  if (null_p lst1) & (null_p lst2) then
    nil
  else if (pair_p lst1) & (pair_p lst2) then
    (cons (f (car lst1) (car lst2)) 
       (map2 f (cdr lst1) (cdr lst2)))
  else
    raise (Invalid_argument("map2"))

let rec fold_left f init lst =
  if null_p lst then
    init
  else if pair_p lst then
    fold_left f (f init (car lst)) (cdr lst)
  else
    raise (Invalid_argument("fold_left"))

let rec fold_right f lst init =
  if null_p lst then
    init
  else if pair_p lst then
    f (car lst) (fold_right f (cdr lst) init)
      else
    raise (Invalid_argument("fold_right"))

let rec length lst =
  let rec iter lst len =
    if null_p lst then
      len
    else if pair_p lst then
      iter (cdr lst) (len + 1)
    else
      -len
  in
  iter lst 0

let append lst1 lst2 =
  try
    let rec recurs lst1 =
      if null_p lst1 then
    	lst2
      else
	(cons (car lst1) (recurs (cdr lst1)))
    in
    recurs lst1
  with Invalid_argument(_) ->
    raise (Invalid_argument("append"))

let reverse lst =
  let rec iter lst result =
    if null_p lst then
      result
    else 
      (iter (cdr lst) (cons (car lst) result))
  in
  (iter lst nil)

let rec memq x lst =
  if null_p lst then
    sharp_f
  else if eq_p x (car lst) then
    lst
  else
    memq x (cdr lst)

let rec memv x lst =
  if null_p lst then
    sharp_f
  else if eqv_p x (car lst) then
    lst
  else
    memv x (cdr lst)

let rec member x lst =
  if null_p lst then
    sharp_f
  else if x = (car lst) then
    lst
  else
    memq x (cdr lst)

let rec assq x lst =
  if null_p lst then
    sharp_f
  else if eq_p x (car (car lst)) then
    (car lst)
  else
    assq x (cdr lst)

let rec assv x lst =
  if null_p lst then
    sharp_f
  else if eqv_p x (car (car lst)) then
    (car lst)
  else
    assv x (cdr lst)

let rec assoc x lst =
  if null_p lst then
    sharp_f
  else if x = (car (car lst)) then
    (car lst)
  else
    assoc x (cdr lst)

let first e   = list_ref e 0
let second e  = list_ref e 1
let third e   = list_ref e 2
let fourth e  = list_ref e 3
let fifth e   = list_ref e 4
let sixth e   = list_ref e 5
let seventh e = list_ref e 6
let eighth e  = list_ref e 7
let nineth e  = list_ref e 8
let tenth e   = list_ref e 9

let list1 x1 = (cons x1 nil)
let list2 x1 x2 = (cons x1 (cons x2 nil))
let list3 x1 x2 x3 = (cons x1 (cons x2 (cons x3 nil)))
let list4 x1 x2 x3 x4 = (cons x1 (cons x2 (cons x3 (cons x4 nil))))
let list5 x1 x2 x3 x4 x5 = 
  (cons x1 (cons x2 (cons x3 (cons x4 (cons x5 nil)))))
let list6 x1 x2 x3 x4 x5 x6 = 
  (cons x1 (cons x2 (cons x3 (cons x4 (cons x5 (cons x6 nil))))))
let list7 x1 x2 x3 x4 x5 x6 x7 = 
  (cons x1 (cons x2 (cons x3 (cons x4 (cons x5 (cons x6 (cons x7 nil)))))))
let list8 x1 x2 x3 x4 x5 x6 x7 x8 = 
  (cons x1 
     (cons x2
 	(cons x3 
	   (cons x4 
	      (cons x5
		 (cons x6 (cons x7 (cons x8 nil))))))))
