(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
(* Exception throw by primitive functions. *)
exception Bad_arg of 
  string * 	(* Name of primitive function in which error was detected. *)
    t		(* Argument value that caused the error. *)

(* Routines for determining the "type" of a Scheme value. *)
val bool_p : t -> bool
val char_p : t -> bool
val float_p : t -> bool
val int_p : t -> bool
val string_p : t -> bool
val pair_p : t -> bool
val symbol_p : t -> bool
val vector_p : t -> bool
val procedure_p : t -> bool
val input_port_p : t -> bool
val output_port_p : t -> bool

(* Variables and routines to tag ML values to produce Scheme values *)
val nil : t
val sharp_t : t
val sharp_f : t
val tag_bool : bool -> t
val tag_char : char -> t
val tag_float : float -> t
val tag_int : int -> t
val tag_string : string -> t
(* There is purposely no tag_pair. Use cons instead. *)
(* There is purposely no tag_symbol. Use string_to_symbol or intern
   instead. *)
val tag_vector : t array -> t
(* todo: make this obsolete *)
val tag_procedure : procedure -> t
val tag_input_port : input_port -> t
val tag_output_port : output_port -> t
val eof : t
val unspecified : t

(* Routines to perform run-time type checks and to untag values. That is,
   routines to convert from Scheme values to ML values. *)
val untag_bool : t -> bool
val untag_char : t -> char
val untag_float : t -> float
val untag_int : t -> int
val untag_string : t -> string
(* There is purposely no untag_pair. Use car and cdr instead. *)
(* There is purposely no untag_symbol. Use symbol_to_string instead. *)
val untag_vector : t -> t array
(* todo: make this obsolete *)
val untag_procedure : t -> procedure
val untag_input_port : t -> input_port
val untag_output_port : t -> output_port

(* Equality tests. NOTE: these routines should be used instead of the (=)
   and (==) operators. *)
(* eq_p: the strongest test. "pointer" equality. roughly corrisponds to the
   (==) operator. *)
val eq_p : t -> t -> bool
(* eqv_p: slightly weaker. structural equality for characters and numbers.
   "pointer" equality for all other values. *)
val eqv_p : t -> t -> bool
(* equal_p: the weakest test. structural equality. roughly corrisponds to
   the (=) operator. *)
val equal_p : t -> t -> bool

(* Primitive pairs *)
val car : t -> t
val cdr : t -> t
val cons : t -> t -> t
val set_car : t -> t -> unit
val set_cdr : t -> t -> unit

(* Primitive Symbols *)
val symbol_to_string : t -> string
(* Intern converts to the canonical case and maintains the pointer
   invariance of symbols. That is, (eq? (intern "a") (intern "a")) will
   always be true. *)
val intern : string -> t
(* Symbol_to_string neither converts case nor maintains pointer invariance.
   YOU ALMOST CERTAINLY WANT TO USE intern INSTEAD OF symbol_to_string. 
   This routine is provided only as a means of implementing the Scheme
   primitive string->symbol. *)
val string_to_symbol : string -> t

(* Pairs and Lists *)
(* A "list" is a sequence of cons pairs terminated by a nil. *)
val list_p : t -> bool
val to_list : t -> t list
val from_list : t list -> t

val cadr : t -> t
val cddr : t -> t
val caddr : t -> t
val cdddr : t -> t
val cadddr : t -> t
val cddddr : t -> t
val caddddr : t -> t
val cdddddr : t -> t

(* returns the nth cons pair of a list, starting from 0. *)
val list_tail : t -> int -> t
(* returns the nth element of a list. 
   (list_ref l i) = (car (list_tail l i)) *)
val list_ref : t -> int -> t

val map : (t -> t) -> t -> t
val map2 : (t -> t -> t) -> t -> t -> t
val fold_left : (t -> t -> t) -> t -> t -> t
val fold_right : (t -> t -> t) -> t -> t -> t
(* (length '()) -> 0
   (length '(a b)) -> 2
   (length '(a b . c)) -> -2 *)
val length : t -> int
val append : t -> t -> t
val reverse : t -> t
val memq : t -> t -> t
val member : t -> t -> t
val assq : t -> t -> t
val assoc : t -> t -> t

(* Returns the value from a particular location with a list. *)
val first : t -> t
val second : t -> t
val third : t -> t
val fourth : t -> t
val fifth : t -> t
val sixth : t -> t
val seventh : t -> t
val eighth : t -> t
val nineth : t -> t
val tenth : t -> t

(* each listN will construct an N element list whose values are given by
   its arguments. *)
val list1 : t -> t
val list2 : t -> t -> t
val list3 : t -> t -> t -> t
val list4 : t -> t -> t -> t -> t
val list5 : t -> t -> t -> t -> t -> t
val list6 : t -> t -> t -> t -> t -> t -> t
val list7 : t -> t -> t -> t -> t -> t -> t -> t
val list8 : t -> t -> t -> t -> t -> t -> t -> t -> t
