(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
(********************************************************
 *			   SPMD translation
 *********************************************************)

(*
  ************************************************************
  Important reminders:
  
  - most of the lambdas has symbolic constants as the first
    argument
  - changing the structure of the lambdas might break the code
    in a bunch of places, because the positions of things like
    global indices in alignement are hard-codes (I did not quite see
    a way around it).  On the other hand, I do not see a good
    reason why the structure of things like alignment should be 
    changed at all

  *************************************************************
*)


open Printf
open Ssisl
open Bsct_base
open Expr_string
module E = Expr_string.E   (* expressions *)
module A = Expr_string.A   (* affine forms and maps *)
module R = Rel_expr (* relational expressions *)

let debug = ref false;;
let debug_on () = debug := true;;
let debug_off () = debug := false;;

let lbda_to_string lbda = (Sx.to_string_pretty (L.to_sexpr lbda));;
let rel_to_string q = (Sx.to_string_pretty (R.to_sexpr q));;
let parse_error msg sexpr = 
  failwith(msg ^ ":" ^ (Sx.to_string sexpr));;

let verify_keys2 lbda1 lbda2 lst msg =
  (List.iter
     (fun (k1,k2) ->
       (zassert ((L.var_list_size lbda1 k1)=
		 (L.var_list_size lbda2 k2))
	  (Failure(msg ^ ":verify_keys2: "^k1^","^k2))))
     lst)

let analyze2_in_key = "$in"
and analyze2_out_key = "$out"
and analyze2_t_key = "$t"
and analyze2_u_key = "$u"

let analyze2 lbda keys_in keys_out =
  (zassert ([] = (List'.intersection keys_in keys_out))
     (Failure("Spmd.analyze2 : keys_in,out are not disjoint: "^
	      (String.concat " " (keys_in @ (","::keys_out))))));
  let lbda' = 
    (L.concat_vars 
       (L.concat_vars lbda keys_in analyze2_in_key)
       keys_out analyze2_out_key) in
  ((L_affine.analyze lbda' 
      analyze2_out_key 
      analyze2_in_key 
      analyze2_t_key
      analyze2_u_key).L_affine.desc)

let analyze_full lbda keys_in keys_out =
  (zassert ([] = (List'.intersection keys_in keys_out))
     (Failure("Spmd.summary2 : keys_in,out are not disjoint: "^
	      (String.concat " " (keys_in @ (","::keys_out))))));
  let lbda' = 
    (L.concat_vars 
       (L.concat_vars lbda keys_in analyze2_in_key)
       keys_out analyze2_out_key) in
  (L_affine.analyze lbda' 
     analyze2_out_key 
     analyze2_in_key 
     analyze2_t_key
     analyze2_u_key)
     

(*********************************************************)
(*********************************************************)

(*********************************************************
  Lambda expressions are used extensively. Here are 
  the common keys:
  symb -- symbolic constants
  proc -- processor numbers
  t_global -- global template index
  t_local -- local template index
  iter -- global iteration space
  global -- global array index
  local -- local array index
  array -- array index (when we do not discriminate
                        between global or local)
*********************************************************)

module Template =
  struct
    type t =
    	{ name : string;
	  global_bounds : L.t; (* Lsymb Lt Bounds *)
	  local_bounds : L.t; (* Lsymb Lp Lt' Bounds *)
	  distribution : R.t (* Lsymb Lt Lp Lt' ... *)
	} 

    (*********************************************************)
    let n_global templ = (L.var_list_size templ.global_bounds "t_global")
    and n_local templ = (L.var_list_size templ.local_bounds "t_local")
    and n_procs templ = (L.var_list_size templ.local_bounds "proc")

    (*********************************************************)
    let verify templ =
      (zassert 
	 (L.check_keys_p templ.global_bounds 
	    ["symb"; "t_global"])
	 (Failure "template: global bounds, nvars"));
      (zassert 
	 (L.check_keys_p templ.local_bounds 
	    ["symb"; "t_local"; "proc" ])
	 (Failure "template: local bounds, nvars"));
      (zassert
	 (((n_global templ)+(n_local templ)+(n_procs templ)) =
	  (List.length (R.get_out_vars templ.distribution)))
	 (Failure "template: n_global+n_local+n_procs"));
      ()
  end

(*********************************************************)
(*********************************************************)
module Relation =
  struct
    type t =
	{ name : string;
	  bounds : L.t;
	  desc : relation_desc
	} 

    and relation_desc = 
	Global of global_description
      |	Local

    (* bounds are of the form Lsymb Larray ... *)
    and global_description =
	{ template : Template.t;
	  local : t;
	  alignment : L.t (* Lsymb La La' Lt Lt' .... *)
	} 

    (*********************************************************)
    let get_global r =
      match r.desc with
      |	Global(d) -> d
      |	Local -> failwith "Spmd.Relation: expecting global"

    let get_local r =
      match r.desc with
      |	Local -> ()
      |	Global(_) ->  failwith "Spmd.Relation: expecting local"

    let local_p r =
      match r.desc with
      |	Local -> true
      |	Global(_) -> false

    let get_alignment r =
      let d = (get_global r) in d.alignment

    let get_template r =
      (get_global r).template

    let get_template_name r =
      (get_global r).template.Template.name

    let n_dim r = (L.var_list_size r.bounds "array")

    let n_global r = 
      let _ = (get_global r) in (n_dim r)

    let n_local r =
      let d = (get_global r) in
      (n_dim d.local)

    let get_local4global r =
      (get_global r).local

    (*********************************************************)
    let sane_alignment_p r = begin
      (* let Ls La La' Lt Lt' F(s,a,a',t,t') be the 
	 alignment constraint. The following must hold:
	 (1) For every valid combination of s,a',t 
	         there must be unique a.
	 (2) For every valid combination of s,a,t' 
                 there must be unique a'.
	 (3) For every s,a there must be t
	 (4) For every s,a' there must be t'
	 (These are necessary conditions)
	 ******)
      let algn = (get_alignment r) in

      (* condition (3) *)
      let sane_global_p = (
	match (analyze2 algn ["symb";"global"] ["t_global"]) with
	  L_affine.Universal_unique(_) -> true
	| L_affine.Universal(_) -> true
	| _ -> false)

      (* condition (4) *)
      and sane_local_p = (
	match (analyze2 algn ["symb";"local"] ["t_local"]) with
	  L_affine.Universal_unique(_) -> true
	| L_affine.Universal(_) -> true
	| _ -> false)

      (* condition (1) *)
      and global_determined_p = (
      	match (analyze2 algn ["symb"; "local"; "t_global"] ["global"]) with
	  L_affine.Universal_unique(_) -> true
	| L_affine.Bound_unique(_) -> true
	| _ -> false)
	  
      (* condition (2) *)
      and local_determined_p = (
      	match (analyze2 algn ["symb"; "global"; "t_local"] ["local"]) with
	  L_affine.Universal_unique(_) -> true
	| L_affine.Bound_unique(_) -> true
	| _ -> false) in

      (sane_global_p && 
       sane_local_p &&
       global_determined_p && 
       local_determined_p)
    end (* sane_alignment_p *)

    (*********************************************************)
    let verify r = begin
      match r with
	{bounds=bnds; desc=Local} -> ()
      |	{name=nm; bounds=bnds; 
	  desc=Global { template=templ; 
			local=local_r; 
			alignment=algn }
	} ->
	  let fail msg = (Failure(msg ^ " : " ^ nm)) in
	  (verify_keys2 bnds algn 
	     [ ("array","global"); ("symb", "symb") ]
	     "Relation.verify: global ");
	  (verify_keys2 local_r.bounds algn 
	     [ ("array","local"); ("symb", "symb") ]
	     "Relation.verify: local ");
	  (zassert (sane_alignment_p r)
	     (fail "Spmd.Relation.verify: alignment"));
	  ()
    end


    (*********************************************************)
    let replicated_p r = begin
      let algn = (get_alignment r) in
      match (analyze2 algn ["symb"; "global"] ["t_global"]) with
	L_affine.Universal_unique(_) -> false
      |	L_affine.Universal(_) -> true
      |	_ -> failwith "Spmd.Relation: replicated_p bad alignment"
    end (* replicated_p *)

    (*********************************************************)
    let fully_replicated_p r = begin
      let algn = (get_alignment r) in
      
      match (analyze2 algn ["symb"; "global"] ["t_global"]) with
      	L_affine.Universal_unique(_) -> false
      |	L_affine.Universal(lbda) -> 
	  (* lbda = Ls a Lt Lparam . t = f(s,a,param) 
	     we want to determine if, in fact, 
	     f() does not depend on s and a *)
	  let f_args = (E.collect_atoms (L.get_predicate lbda)) in
	  ([] = (List'.intersection 
		   (L.get_var_list lbda analyze2_in_key) 
		   f_args))
	    
      |	_ -> failwith "Spmd.Relation: fully replicated_p bad alignment"
    end (* fully_replicated_p *)
  end

(*********************************************************)
module Env =
  struct

    type symbol = 
	Template_symbol of Template.t
      |	Relation_symbol of Relation.t

    (* symbol table is an association list *)
    type t = (string * symbol) list
	  
    (*********************************************************)
    let empty = []

    let defined_p env name = (List.mem_assoc name env)

    let get env name = (
      try (List.assoc name env) with
	Not_found -> (failwith("Spmd.Env: not found "^name)))

    let add env name sym = (name,sym)::env

    let rec remove env name = ( 
      (* only remove the "top" binding *)
      match env with
	(name',sym)::rest ->
	  if (name' = name) then rest
	  else (name',sym)::(remove rest name)
      |	[] -> [])

    let remove_all env name = ( 
      (* remove all bindings *)
      (List'.remove_if (fun (name',sym) -> (name' = name)) env))

    let rec replace env name sym = (
      (* only replaces the top binding *)
      match env with
	(name',sym')::rest ->
	  if (name' = name) then (name,sym)::rest
	  else (name',sym')::(replace rest name sym)
      |	[] -> [])

    (*********************************************************)
    let get_template env name = (
      match (get env name) with
	Template_symbol(templ) -> templ
      |	_ -> failwith("Spmd.Env: not a template"^name))

    let get_relation env name = (
      match (get env name) with
	Relation_symbol(r) -> r
      |	_ -> failwith("Spmd.Env: not a local relation"^name))

    let get_local_relation env name = (
      let r = (get_relation env name) in
      ((Relation.get_local r); r))

    let get_global_relation env name = (
      let r = (get_relation env name) in
      ((Relation.get_global r); r))

    (*********************************************************)
    let verify_symbol sym = (
      match sym with
	Template_symbol(templ) -> (Template.verify templ)
      |	Relation_symbol(r) -> (Relation.verify r))

    let verify_env env = (
      (List.iter 
	 (fun (name,sym) -> verify_symbol sym)
	 env);
      env)

    (*********************************************************)
    let parse_template env sexpr = begin
      match (Sx.to_list sexpr) with
	[n;g;l;d] -> 
	  (add env (Sx.untag_symbol n)
	     (Template_symbol 
		{ Template.name = (Sx.untag_symbol n);
		  Template.global_bounds = (L_convert.parse g);
		  Template.local_bounds = (L_convert.parse l);
		  Template.distribution = (R.from_sexpr d); }))
      |	_ -> (parse_error "Spmd.parse_template" sexpr)
    end (* parse_template *)
      
    let parse_local env sexpr = begin
      match (Sx.to_list sexpr) with
	[n;bnd] -> 
	  (add env (Sx.untag_symbol n)
	     (Relation_symbol
		{ Relation.name = (Sx.untag_symbol n);
		  Relation.bounds = (L_convert.parse bnd);
		  Relation.desc = Relation.Local }))
      |	_ -> (parse_error "Spmd.parse_local" sexpr)
    end (* parse local *)

    let parse_global env sexpr = begin
      match (Sx.to_list sexpr) with
      	[n;n_loc;n_templ;bnd;algn] ->
	  (add env (Sx.untag_symbol n)
	     (Relation_symbol
		{ Relation.name = (Sx.untag_symbol n);
		  Relation.bounds = (L_convert.parse bnd);
		  Relation.desc = 
		  (Relation.Global 
		     { Relation.template = (get_template env 
					      (Sx.untag_symbol n_templ));
		       Relation.local = (get_local_relation env
					   (Sx.untag_symbol n_loc));
		       Relation.alignment = (L_convert.parse algn) }
		     )}))
      |	_ -> (parse_error "Spmd.parse_global" sexpr)
    end (* parse_global *)

    let parse_symbol env sexpr = begin
      match sexpr with 
	Sx.Pair(lookahead,rest) 
	  when (Sx.untag_symbol lookahead) = "template" ->
	    (parse_template env rest)
      |	Sx.Pair(lookahead,rest) 
	  when (Sx.untag_symbol lookahead) = "local" ->
	    (parse_local env rest)
      |	Sx.Pair(lookahead,rest) 
	  when (Sx.untag_symbol lookahead) = "global" ->
	    (parse_global env rest)
      |	_ -> (parse_error "Spmd.parse_symbol" sexpr)
    end (* parse_symbol *)

    let parse_env sexpr =
      (verify_env
	 (List.fold_left parse_symbol [] (Sx.to_list sexpr)))

    (*********************************************************)
    let unparse_template templ = begin
      (Sx.from_list
	 [ (Sx.tag_symbol templ.Template.name);
	   (L_convert.unparse templ.Template.global_bounds);
	   (L_convert.unparse templ.Template.local_bounds);
	   (R.to_sexpr templ.Template.distribution)
	 ] )
    end (* unparse_template *)

    let unparse_relation r = begin
      match r.Relation.desc with
	Relation.Local -> (Sx.from_list
			     [ (Sx.tag_symbol "local");
			       (Sx.tag_symbol r.Relation.name);
			       (L_convert.unparse r.Relation.bounds)
			     ] )
      |	Relation.Global(g) -> (Sx.from_list
				 [ (Sx.tag_symbol "global");
				   (Sx.tag_symbol r.Relation.name);
				   (Sx.tag_symbol g.Relation.local.Relation.name);
				   (Sx.tag_symbol 
				      g.Relation.template.Template.name);
				   (L_convert.unparse r.Relation.bounds);
				   (L_convert.unparse g.Relation.alignment)
				 ] )
    end (* unparse_relation *)

    let unparse_symbol sym = begin
      match sym with
	Template_symbol(templ) -> unparse_template templ
      |	Relation_symbol(r) -> unparse_relation r
      
    end (* unparse_symbol *)

    let unparse_env env = 
      (Sx.from_list
	 (List.map (fun (name,sym) -> unparse_symbol sym)
	    (List.rev env)))


    (*********************************************************)
  end (* module Env *)


(*********************************************************)
module Solver =
  struct

    (* this is the input type for Spmd translation *)
    (* (environment, computation template, 
           computation alignment, query) *)
    type input = Input of Env.t * string * L.t * R.t

    let parse_input sexpr = begin
      match (Sx.to_list sexpr) with
	[e;t;c;q] ->
	  (Input 
	     ((Env.parse_env e),
	      (Sx.untag_symbol t),
	      (L_convert.parse c),
	      (R.from_sexpr q)))
      |	_ -> (parse_error "Spmd.parse_input " sexpr)
    end (* parse_input *)

    let unparse_input (Input(env,t,c,q)) = begin
      (Sx.from_list
	 [ (Env.unparse_env env);
	   (Sx.tag_symbol t);
	   (L_convert.unparse c);
	   (R.to_sexpr q) ])
    end (* unparse_input *)


    (*********************************************************)
    let local_acc_name acc_name = (acc_name ^ "$local")

    (*********************************************************)
    type find_aligned_result =
	Aligned of (string list) * L.t 
      |	Not_aligned

    let find_aligned env templ_name comp_align query = begin
      let access_name_lst = (R.get_binding_names query) in
      let candidate_groups = 
	(* sorted in descending order of group size *)
	(Sort.list
	   (fun l1 l2 -> ((List.length l1) > (List.length l2)))
	   (List'.gen_subsets access_name_lst))
      and access_eq = (L.rename_unique (R.get_eqs query)) in
      (verify_keys2 comp_align access_eq
	 [ ("symb","symb"); ("iter","iter")]
	 "Spmd.Solver.find_aligned : access_eq and comp_align");

      let access_and_comp_align =
	(* Ls Li Lt Lt' Laccess1 Laccess2 .... *)
	(* t' is not constrained, yet *)
	(L.apply access_eq 
	   (L.rename_unique (L.add_key 
			       comp_align "t_local" 
			       (Template.n_local 
				  (Env.get_template env templ_name))))
	   ((List.map (fun name -> L.A(name,name))
	       (L.get_all_keys access_eq)) @
	    [ L.B("t_global","t_global"); 
	      L.B("t_local","t_local");
	      L.Sub("symb", "symb");
	      L.Sub("iter", "iter") ])) in

      let get_global_relation_access access_name = (
	let binding = (R.get_binding query access_name) in
	(Env.get_global_relation env binding.R.relation)) in

      let group_aligned_p lst = begin
	(if !debug then begin
	  (printf "Spmd.group_aligned_p, testing: %s"
	     (String.concat " " lst));
	  (print_newline ())
	end else ());

	let lst' = (* access name, template name, alignment *)
	  (List.map
	     (fun accs ->
	       let r = (get_global_relation_access accs) in
	       (accs,
		(Relation.get_template_name r),
		(L.rename_unique 
		   (Relation.get_alignment r))))
	     lst)

	and same_template_p templ_all (_,t_name,_) = (t_name = templ_all) in

	if (List.for_all (same_template_p templ_name) lst')
	then begin
	  let absorb lbda (accs_name,t_name,algn) = (
	    (* lbda has keys "symb", "iter", "t_global", "t_local" + 
	       some crud. algn has keys "symb", "t_global", "t_local",
	       "global", "local".
	       "global" corresponds to accs_name in lbda *)
	    (verify_keys2 lbda algn
	       [ ("symb","symb"); (accs_name,"global") ]
	       "Spmd.Solver.group_aligned_p.absorb ");
	    (L.apply lbda algn
	       ((List.map (fun k -> L.A(k,k)) (L.get_all_keys lbda)) @
		[ L.Sub("symb","symb");
		  L.Sub("t_global","t_global");
		  L.Sub("t_local","t_local");
		  L.Sub("global",accs_name);
		  L.B("local",(local_acc_name accs_name)) ]))
	      ) in
	  let test_lbda = 
	    (List.fold_left absorb access_and_comp_align lst') in
	  (if !debug then begin
	    (printf "group_aligned_p test_lbda=%s"
	       (lbda_to_string test_lbda));
	    (print_newline ())
	  end else ());
	  (match (analyze2 test_lbda ["symb";"iter"] ["t_global"]) with
	    L_affine.Universal_unique(_) -> Aligned(lst,test_lbda)
	  | L_affine.Universal(_) -> Aligned(lst,test_lbda)
	  | _ -> Not_aligned)
	      
	end
	else
	  Not_aligned (* not the same template *)
      end (* group_aligned_p *) in

      let rec find lst = (
	match lst with
	  [] -> Not_aligned
	| g::lst' ->
	    (match (group_aligned_p g) with
	      Not_aligned -> (find lst')
	    | Aligned(g',lbda) -> 
		(zassert (g'=g) (Failure("Spmd.find_aligned: hmmm")));
		Aligned(g',lbda))) in

      (find candidate_groups)

    end (* find_aligned *)
	  
    (*********************************************************)
    let data_centric_by_aligned_p aligned_lst access_lbda = begin
      (* we have to check that
	 (1)each feasible combination of local indices of
	    the aligned arrays with the global template point 
	    uniquelly determines an iteration
	 (2)each feasible combination of an iteration and 
	    a local template point uniquelly determines 
	    local indices
	 (3)each feasible combination of local indices of the accesses
	    in the list aligned_lst uniquelly determines all
	    other accesses
	 *)
      let aligned_lst_local = 
	(List.map local_acc_name aligned_lst) in
      
      let cond_1_p = (
	match (analyze2 access_lbda 
		 ("symb"::"t_global"::aligned_lst_local)
		 ["iter"])
	with
	  L_affine.Universal_unique(_) -> true
	| L_affine.Bound_unique(_) -> true
	| _ -> false)

      and cond_2_p = (
	match (analyze2 access_lbda ["symb";"iter";"t_local"]
		 aligned_lst_local) with
	  L_affine.Universal_unique(_) -> true
	| L_affine.Bound_unique(_) -> true
	| _ -> false)

      and cond_3_p = (
	let other_accesses = 
	  (List'.difference (L.get_all_keys access_lbda)
	     (["symb";"iter";"t_local";"t_global"]
	      @ aligned_lst 
	      @ aligned_lst_local)) in
	match (analyze2 access_lbda ("symb"::aligned_lst_local)
		 other_accesses) 
	with
	  L_affine.Universal_unique(_) -> true
	| L_affine.Bound_unique(_) -> true
	| _ -> false)

      in (cond_1_p && cond_2_p && cond_3_p)
    end

    (*********************************************************)
    let localize_aligned_data_centric env query aligned_lst access_lbda = begin
      (* we need to:
	 - replace the bindings in the query
	 - use access_lbda to replace equalities in the query and
	   the loop variables in the query
	 *)
      let replace1_binding query access_name = begin
	let bind = (R.get_binding query access_name) in
	let global_r = (Env.get_global_relation env bind.R.relation) in
	let local_r = (Relation.get_local4global global_r) in
	let bind' = { R.relation = local_r.Relation.name;
		      R.index_vars = (L.get_var_list access_lbda
					(local_acc_name access_name));
		      R.value_var = bind.R.value_var;
		      R.flag_var = bind.R.flag_var } in
	let query' = (R.rename_binding 
			(R.set_binding query
			   access_name
			   bind')
			access_name
			(local_acc_name access_name)) in
	if (false && !debug) then begin
	  (print_string "-------trace of replace1 ----------------------");
	  (print_newline ());
	  (printf "query before %s" (rel_to_string query));
	  (print_newline ());
	  (printf "access_name %s" access_name);
	  (print_newline ());
	  (printf "query after %s" (rel_to_string query'));
	  (print_newline ());
	  (print_string "-------end of replace1-------------------------");
	  (print_newline ());
	end else ();

	query' (* this is the result of replace1_binding *)

      end (* replace1_binding *) in

      let query' = (List.fold_left replace1_binding query aligned_lst) in
      let new_keys =
	(List'.difference
	   (L.get_all_keys access_lbda)
	   ("symb"::"iter"::"t_global"::"t_local"::aligned_lst)) in

      let new_bindings =
	(List.map (fun k -> (k, (L.get_var_list access_lbda k)))
	   new_keys) in

      match (analyze2 access_lbda ["symb"] new_keys) with
	L_affine.Universal(lbda) -> begin
	    (* Lout Lin Lt . out = f(in,t)
	       out is the concatenation of all the keys that we
	       want to keep. in="symb". t should be renamed into
	       "iter" *)
	  if !debug then begin
	    (printf "new keys: %s\n" (String.concat " " new_keys));
	    (printf "old access:\n%s\n" (lbda_to_string access_lbda));
	    (printf "new access:\n%s\n" (lbda_to_string lbda));
	    (print_newline ());
	  end else ();
	  let new_symb = (L.get_var_list lbda analyze2_in_key)
	  and new_iter = (L.get_var_list lbda analyze2_u_key) in
	  let new_bindings' = (
	    [ ("iter",new_iter); ("symb",new_symb) ] @ new_bindings) in
	  let new_eqs = (L.make { L.bindings=new_bindings';
				  L.predicate = (L.get_predicate lbda) })
	  and new_ineqs = (L.make { L.bindings=new_bindings';
				    L.predicate = E.Const(E.Bool(true)) }) in
	  (R.verify
	     (R.set_iter
	     	(R.set_eqs (R.set_ineqs query' new_ineqs) new_eqs)
	     	new_iter)
	     "Spmd.Solver.localize_aligned_data_centric: bad output query")

	end (* L_affine.Universal match *)
      | _ -> failwith("Spmd: I am not sure how to handle this, yet")

    end (* localize_aligned_data_centric *)

    (*********************************************************)
    let select_localized_bindings env query =
      (* returns (binding name, binding) pairs *)
      (List'.remove_if_not
	 (fun (name,bind) ->
	   (Relation.local_p (Env.get_relation env bind.R.relation)))
	 (R.get_all_bindings query))

    (*********************************************************)
    let project_access_lbda lbda lst = begin
      (* project the lambda_expression into the
	 access names in the lst *)
      let { L_affine.summary_lbda=lbda';
	    L_affine.desc=descriptor } = 
 	(analyze_full lbda ["symb"; "iter"] lst) in
      match descriptor with
	L_affine.Universal_unique(lbda') ->
	  (* lbda' has the form
	     Lin Lout out=f(in), where in is the concatenation of
	     symb and iter vars and out is the concatenation of
	     variables in the lst 

	     we want the output keys to be of the form
	     Lsymb Liter La1 .. Lan , where ai are members of the lst 
	     *)
	  let new_bindings = (List.map
				(fun name -> (name, (L.get_var_list lbda name)))
				("symb"::"iter"::lst)) in
	  let lbda'' = (L.key_blit lbda' 
			  [analyze2_in_key; analyze2_out_key]
			  new_bindings) in
	  if !debug then begin
	    (printf "project_access_lbda:\nin %s\nlst: %s\nintermed: %s\nout: %s"
	       (lbda_to_string lbda) (String.concat " " lst)
	       (lbda_to_string lbda')
	       (lbda_to_string lbda''));
	    (print_newline ())
	  end else ();
	  lbda''
      |	_ -> failwith("Spmd.Solver.project_access_lbda: "^
		      (lbda_to_string lbda))
    end (* project_access_lbda *)

    (*********************************************************)
    let project_query query lst = begin
      (*
	we have to:
	(1) remove the bindings with the names not in the lst
	(2) project away the variables in the removed bindings 
	    in the equalities and inequalities
	(3) remove the variables in the removed bindings from the 
	    out-vars and the predicate
	*)
      let vars_to_remove = 
	(List.map Pervasives.fst
	   (List'.remove_if
	      (fun (name,tpe) -> 
	     	(match tpe with 
		  R.Loop -> false | 
		  R.Bound(bvt) -> 
		    (List.mem (R.get_binding_name4var_type bvt) lst)))
	      (R.get_inverse_binding_list query)))
      and new_eqs = (project_access_lbda
		       (R.get_eqs query) lst) in
      let new_ineqs = (L.set_predicate new_eqs (E.Const(E.Bool(true)))) in

      let flags_in_pred = 
	(List.map
	   (fun e -> 
	     (match e with 
	       E.Atom(v)->v
	     | _ -> failwith("Spmd.Solver.project_query: conjunct is expected")))
	   (E.args_by_op E.And (L.get_predicate (R.get_pred query)))) in

      let new_flags_in_pred = (List'.difference flags_in_pred vars_to_remove) in
      let new_pred_lbda = 
	(L.make { L.bindings = [("flags",new_flags_in_pred)];
		  L.predicate = (E.make_and
				   (List.map (fun v -> E.Atom(v))
				      new_flags_in_pred)) }) in
      let new_out_vars = 
	(List'.difference (R.get_out_vars query) vars_to_remove) in

      let query' = 
	(R.verify (R.make 
		     (R.get_iter query)
		     new_out_vars
		     (List'.remove_if_not
			(fun (name,bind) -> (List.mem name lst))
			(R.get_all_bindings query))
		     new_eqs
		     new_ineqs
		     new_pred_lbda) 
	   "Spmd.Solver.project_query") in

      if !debug then begin
	(printf "project query:\nin %s\nlst: %s\nout: %s"
	   (rel_to_string query)
	   (String.concat " " lst)
	   (rel_to_string query'));
      	(print_newline ())
      end else ();
      query'

    end (* project_query *)

    (*********************************************************)
    let local_iteration_set_query env query = begin
      let local_bind_names =
	(List.map Pervasives.fst
	   (List'.remove_if_not
	      (fun (name,bind) -> (Relation.local_p
				     (Env.get_relation env bind.R.relation)))
	      (R.get_all_bindings query))) in
      (project_query query local_bind_names)	   
    end (* local_iteration_set_query *)
  end


(*********************************************************)
(*               Scheme interface is here                *)
(*********************************************************)

module Calc =
  struct
    module Lscm = Lambda_scheme

    (*********************************************************)
    let debug form_flag =
      let f = Scheme.untag_bool form_flag in
      if f then debug_on () else debug_off ();
      Scheme.nil

    (*********************************************************)
    let copy_input form =
      (Sx.to_scheme
	 (Solver.unparse_input
	    (Solver.parse_input 
	       (Sx.from_scheme form))))

    (*********************************************************)
    let find_aligned spmd_input_form = (
      match (Solver.parse_input (Sx.from_scheme spmd_input_form))
      with
	Solver.Input(env,templ_name,algn,query) ->
	  (match (Solver.find_aligned env templ_name algn query) 
	  with
	    Solver.Not_aligned -> Scheme.nil
	  | Solver.Aligned(lst,lbda) ->
	      (Sx.to_scheme
		 (Sx.Pair
		    ((Sx.tag_symbol_list lst),
		     (L_convert.unparse lbda))
		    ))
		)
	    )

    (*********************************************************)
    let data_centric_by_aligned_p form = (
      let aligned_lst = 
	(Sx.untag_symbol_list
	   (Sx.from_scheme (Scheme.car form)))
      and access_lbda =
	(Lscm.scheme_to_lbda (Scheme.cdr form)) in
      let p = (Solver.data_centric_by_aligned_p 
		 aligned_lst
		 access_lbda) in
      (Scheme.tag_bool p)
	)

    (*********************************************************)
    let localize_aligned_data_centric form_input form_aligned = (
      let aligned_lst = 
	(Sx.untag_symbol_list
	   (Sx.from_scheme (Scheme.car form_aligned)))
      and access_lbda =
	(Lscm.scheme_to_lbda (Scheme.cdr form_aligned)) in
      match (Solver.parse_input (Sx.from_scheme form_input))
      with
	Solver.Input(env,templ_name,algn,query) ->
	  let query' = 
	    (Solver.localize_aligned_data_centric 
	       env query aligned_lst access_lbda) in
	  (Sx.to_scheme
	     (R.to_sexpr query'))
	    )

    (*********************************************************)
    let local_iteration_set_query form_input form_query = (
      let query = (R.from_sexpr (Sx.from_scheme form_query)) in
      match (Solver.parse_input (Sx.from_scheme form_input)) with
	Solver.Input(env,_,_,_) ->
	  let query' = 
	    (Solver.local_iteration_set_query env query) in
	  (Sx.to_scheme (R.to_sexpr query'))
      )

    (*********************************************************)
    let attach env prefix =
      (Scheme.define_prim1 env (prefix ^ "debug") debug);
      (Scheme.define_prim1 env (prefix ^ "copy-input") copy_input);
      (Scheme.define_prim1 env (prefix ^ "find-aligned") find_aligned);
      (Scheme.define_prim1 env (prefix ^ "dc-aligned-p")
	 data_centric_by_aligned_p);
      (Scheme.define_prim2 env (prefix ^ "localize-aligned")
	 localize_aligned_data_centric);
      (Scheme.define_prim2 env (prefix ^ "local-iter-set")
	 local_iteration_set_query);
      ()
  end








  
