(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
type 'b find_result =
    Found of 'b
  | Unfound
      ;;

let find_value thunk =
  try Found(thunk ())
  with Not_found -> Unfound

(************************************************************************)

let identity x = x;;
    
let complement1 p a = not (p a)
let complement2 p a b = not (p a b)

let zassert v exn = if ((not (v))) then raise exn else ();;
    
let rec recurs x f =
  f (fun x' -> recurs x' f) x 
;;

(************************************************************************)
(* Dynamic-wind and friends.						*)
(************************************************************************)

let dynamic_wind before during after =
  before ();
  let r =
    try 
      during ()
    with exn ->
      after();
      raise exn
  in
  after ();
  r

let call_with_dynamic_binding r v thunk =
  let old_v = !r in
  dynamic_wind
    (fun () -> r := v)
    thunk
    (fun () -> r := old_v)

(************************************************************************)

let compare_ints x y = x - y;;
    
let compare_pair f1 f2 (a1,a2) (b1,b2) =
  let c1 = (f1 a1 b1) in
  if (c1 <> 0) then
    c1
  else
    (f2 a2 b2)
      ;;
    

(* Some utilities to deal with hashing *)
let 
    compare_hash =
  fun a b -> 
    let ha = (Hashtbl.hash a) and hb = (Hashtbl.hash b)
    in if (ha <> hb) then (ha - hb)
    else (Pervasives.compare a b)
	;;      
      
module type OrderedType =
  sig
    type t
    val compare: t -> t -> int
  end

(************************************************************************)
(* A module for instrumenting code for tracing and printing backtraces  *)
(* on errors.								*)
(************************************************************************)

module Tracing =
  struct

    (**************************************************)
    (* swiped from printexc.ml			      *)
    (**************************************************)

    let print_exn = function
    	Out_of_memory ->
	  prerr_string "Out of memory\n"
      | Match_failure(file, first_char, last_char) ->
	  prerr_string "Pattern matching failed, file ";
	  prerr_string file;
	  prerr_string ", chars "; prerr_int first_char;
	  prerr_char '-'; prerr_int last_char; prerr_char '\n'
      | x ->
	  prerr_string "Uncaught exception: ";
	  prerr_string (Obj.magic(Obj.field (Obj.field (Obj.repr x) 0) 0));
	  if Obj.size (Obj.repr x) > 1 then begin
            prerr_char '(';
            for i = 1 to Obj.size (Obj.repr x) - 1 do
              if i > 1 then prerr_string ", ";
              let arg = Obj.field (Obj.repr x) i in
              if not (Obj.is_block arg) then
            	prerr_int (Obj.magic arg : int)
              else if Obj.tag arg = 252 then begin
            	prerr_char '"';
            	prerr_string (Obj.magic arg : string);
            	prerr_char '"'
              end else
            	prerr_char '_'
            done;
            prerr_char ')'
	  end;
	  prerr_char '\n'

    (**************************************************)
    (* Global state				      *)
    (**************************************************)

    exception Pending of exn

    let push x r = r := x::(!r)
    let pop r = r := List.tl (!r)
    let peek r = List.hd (!r)
	
    let verbose_flags = (ref [] : bool list ref)
    let handlers = (ref [] : (exn -> bool) list ref)
    let depth = ref 0


    (**************************************************)
    (* Top of a tracing region			      *)
    (**************************************************)

    let top verbose_p thunk =
      push verbose_p verbose_flags;
      try
	let result = thunk () in
	pop verbose_flags;
	result
      with ex ->
	pop verbose_flags;
	(match ex with
	  Pending(ex) -> Printf.fprintf stderr "\n"; raise ex
      	|	ex -> raise ex)

    (**************************************************)
    (* A tracing extent				      *)
    (**************************************************)

    let wrap thunk name entry_printer exit_printer =
      let _ =
	if peek verbose_flags then
	  begin
	    Printf.fprintf stderr "*** (+%d) %s " (!depth) name;
	    entry_printer stderr;
	    Printf.fprintf stderr "\n";
	    flush stderr
	  end
      in let result =
	try
	  let _ = (depth := !depth + 1) in 
	  let result = thunk () in 
	  let _ = (depth := !depth - 1) in
	  result
	with ex ->
	  depth := (!depth - 1);
	  (match ex with
	    Pending(_) ->
	      Printf.fprintf stderr "*** (!%d) %s " (!depth) name;
	      entry_printer stderr;
	      Printf.fprintf stderr "\n";
	      flush stderr;
	      raise ex
	  | _ when List.exists (fun p -> p ex) (!handlers) ->
	      if peek verbose_flags then
		begin
		  Printf.fprintf stderr "*** (^%d) %s: " (!depth) name;
		  print_exn ex;
		  flush stderr
		end;
	      raise ex
	  | _ ->
	      Printf.fprintf stderr "\n*** ";
	      print_exn ex;
	      Printf.fprintf stderr "*** Backtrace follows:\n";
	      Printf.fprintf stderr "*** (!%d) %s " (!depth) name;
	      entry_printer stderr;
	      Printf.fprintf stderr "\n";
	      flush stderr;	      
	      raise (Pending(ex)))
      in let _ =
	if peek verbose_flags then
	  begin
	    Printf.fprintf stderr "*** (-%d) %s " (!depth) name;
	    exit_printer stderr result;
	    Printf.fprintf stderr "\n";
	    flush stderr
	  end
      in
      result

    (**************************************************)
    (* Which exns to ignore			      *)
    (**************************************************)

    let pass thunk pred =
      push pred handlers;
      try
	let result = thunk() in
	pop handlers;
	result
      with ex ->
	pop handlers;
	raise ex	
  end

