(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
module OrigSet     = Set
module OrigMap     = Map
module OrigHashtbl = Hashtbl

(************************************************************************)

module type ELEMENT =
  sig
    type t
    val equal: t -> t -> bool
  end

module type ORDERED_ELEMENT =
  sig
    type t
    val compare: t -> t -> int
  end

module type HASHED_ELEMENT =
  sig
    type t
    val equal: t -> t -> bool
    val hash: t -> int
  end

module type COMPLETE_ELEMENT =
  sig
    type t
    val compare: t -> t -> int
    val equal: t -> t -> bool
    val hash: t -> int
  end

module Int_element =
  struct
    type t = int
    let equal = (=)
    let compare = (-)
    let hash = OrigHashtbl.hash
  end

module String_element =
  struct
    type t = string
    let equal = (=)
    let compare = Pervasives.compare
    let hash = OrigHashtbl.hash
  end

(************************************************************************)

module type VECTOR =
  sig
    type elt
    type t
    val create: unit -> t
    val make: int -> t
    val capacity: t -> int
    val set_capacity: t -> int -> unit
    val capacity_incr: t -> int
    val set_capacity_incr: t -> int -> unit
    val trim_to_size: t -> unit
    val erase: t -> unit
    val ensure_capacity: t -> int -> unit
    val len: t -> int
    val set_len: t -> int -> unit
    exception Uninitialized
    val at: t -> int -> elt
    val set_at: t -> int -> elt -> unit
    val insert_at: t -> int -> elt -> unit
    val add_first: t -> elt -> unit
    val add_last: t -> elt -> unit
    val remove_at: t -> int -> unit
    val remove_first: t -> unit
    val remove_last: t -> unit
    val index_from: t -> elt -> int -> int
    val index: t -> elt -> int
    val rindex_from: t -> elt -> int -> int
    val rindex: t -> elt -> int
    val remove: t -> elt -> unit
  end

module Vector(E : ELEMENT) =
  struct
    (* *)
    type elt = E.t
    type t = {
	mutable v_elems : elt option array;
	mutable v_len : int;
	mutable v_incr : int;
      }	
    (* *)
    let create () = { v_elems = [| |]; v_len = 0; v_incr = 64; }
    let make i = { v_elems = Array.make i None; v_len = i; v_incr = 64; }
    (* *)
    let capacity v = Array.length v.v_elems
    let set_capacity v n =
      let a = v.v_elems in
      if n <> Array.length a then
	begin
	  let a' = Array.make n None in
	  Array.blit a 0 a' 0
	    (min (Array.length a) (Array.length a'));
	  v.v_elems <- a'
	end;
      if n < v.v_len then v.v_len <- n
    let capacity_incr v = v.v_incr
    let set_capacity_incr v n = v.v_incr <- n
    let trim_to_size v = set_capacity v v.v_len
    let erase v = set_capacity v 0
    let ensure_capacity v n =
      if Array.length v.v_elems < n then
	begin
	  let num_chunks =
	    if n mod v.v_incr = 0 then
	      n / v.v_incr
	    else
	      n / v.v_incr + 1
	  in
	  set_capacity v (num_chunks*v.v_incr)
	end
    (* *)
    let len v = v.v_len
    let set_len v n =
      if n > v.v_len then
	ensure_capacity v n
      else if n < v.v_len then
	for j = n to v.v_len - 1 do
	  v.v_elems.(j) <- None
	done;
      v.v_len <- n
    (* *)
    let bounds_check c i =
      if not (0 <= i && i < c) then raise Not_found
    exception Uninitialized
    let at v i =
      bounds_check v.v_len i;
      match v.v_elems.(i) with
	None -> raise Uninitialized
      |	Some(x) -> x
    let set_at v i x =
      bounds_check v.v_len i;
      v.v_elems.(i) <- Some(x)
    (* *)
    let insert_at v i x =
      bounds_check (v.v_len+1) i;
      set_len v (v.v_len + 1);
      for j = v.v_len-1 downto i+1 do
	v.v_elems.(j) <- v.v_elems.(j-1)
      done;
      v.v_elems.(i) <- Some(x)
    let add_first v x = insert_at v 0 x
    let add_last v x = insert_at v (len v) x
    let remove_at v i =
      bounds_check v.v_len i;
      for j = i+1 to v.v_len-1 do
      	v.v_elems.(j-1) <- v.v_elems.(j)
      done;
      set_len v (v.v_len - 1)
    let remove_first v = remove_at v 0
    let remove_last v = remove_at v ((len v)-1)
    (* *)
    let index_from v key i =
      bounds_check v.v_len i;
      recurs i
	(fun recurs i ->
	  if i = v.v_len then
	    raise Not_found
	  else 
	    match v.v_elems.(i) with
	      None -> recurs (i+1)
	    | Some(x) ->
		if E.equal key x then
		  i
		else
		  recurs (i+1))
    let index v key = index_from v key 0
    let rindex_from v key i =
      bounds_check v.v_len i;
      recurs i
	(fun recurs i ->
	  if i = -1 then
	    raise Not_found
	  else 
	    match v.v_elems.(i) with
	      None -> recurs (i-1)
	    | Some(x) ->
		if E.equal key x then
		  i
		else
		  recurs (i-1))
    let rindex v key = rindex_from v key (v.v_len -1)
    let remove v x =
      try remove_at v (index v x)
      with Not_found -> ()
  end

(************************************************************************)

module type SET =
  sig
    type elt
    type t
    val empty: t
    val is_empty: t -> bool
    val mem: elt -> t -> bool
    val add: elt -> t -> t
    val remove: elt -> t -> t
    val union: t -> t -> t
    val inter: t -> t -> t
    val diff: t -> t -> t
    val compare: t -> t -> int
    val equal: t -> t -> bool
    val subset: t -> t -> bool
    val iter: (elt -> unit) -> t -> unit
    val fold: (elt -> 'data -> 'data) -> t -> 'data -> 'data
    val cardinal: t -> int
    val elements: t -> elt list
    val choose: t -> elt
    val to_list: t -> elt list
    val of_list: elt list -> t
    val absorb: t -> elt list -> t
  end

module Set(Ord: ORDERED_ELEMENT) =
  struct
    module T = OrigSet.Make(Ord)
    type elt = T.elt
    type t = T.t
    let empty = T.empty
    let is_empty = T.is_empty
    let mem = T.mem
    let add = T.add
    let remove = T.remove
    let union = T.union
    let inter = T.inter
    let diff = T.diff
    let compare = T.compare
    let equal = T.equal
    let subset = T.subset
    let iter = T.iter
    let fold = T.fold
    let cardinal = T.cardinal
    let elements = T.elements
    let choose = T.choose	  
    let to_list = T.elements
    let rec absorb s lst =
      if lst = [] then
	s
      else
	absorb (add (List.hd lst) s) (List.tl lst)
    let of_list lst = absorb empty lst
  end

module Int_set = Set(Int_element)
module String_set = Set(String_element)

(************************************************************************)

module type MAP =
  sig
    type key
    type 'data t
    val empty: 'data t
    val add: key -> 'data -> 'data t -> 'data t
    val find: key -> 'data t -> 'data
    val remove: key -> 'data t -> 'data t
    val iter: (key -> 'data -> unit) -> 'data t -> unit
    val map: ('a -> 'b) -> 'a t -> 'b t
    val fold: (key -> 'data -> 'a -> 'a) -> 'data t -> 'a -> 'a
    val update: 'data t -> key -> 'data -> ('data -> 'data) -> 'data t
    val replace: 'data t -> key -> 'data -> 'data t
    val keys: 'data t -> key list
    val size: 'data t -> int
  end

module Map(Ord: ORDERED_ELEMENT) =
  struct
    module T = OrigMap.Make(Ord)
    type key = T.key
    type 'data t = 'data T.t
    let empty = T.empty
    let add = T.add
    let find = T.find
    let remove = T.remove
    let iter = T.iter
    let map = T.map
    let fold = T.fold
    (* fixme: Better implementation. *)
    let update m k init_d f =
      let (m',d) =
	try
	  let d = find k m in
	  (remove k m, d)
	with Not_found ->
	  (m, init_d)
      in
      add k (f d) m'
    let replace m k d = update m k d (fun _ -> d)
    (* *)
    let keys m = fold (fun key _ lst -> key::lst) m []
    let size m = fold (fun _ _ n -> n+1) m 0 
  end

module String_map = Map(String_element)

(************************************************************************)

module type HASHTBL =
  sig
    type key
    type 'data t
    val create: int -> 'data t
    val clear: 'data t -> unit
    val add: 'data t -> key -> 'data -> unit
    val remove: 'data t -> key -> unit
    val find: 'data t -> key -> 'data
    val find_all: 'data t -> key -> 'data list
    val iter: (key -> 'data -> unit) -> 'data t -> unit
    val fold: (key -> 'data -> 'a -> 'a) -> 'data t -> 'a -> 'a
    val update: 'data t -> key -> 'data -> ('data -> 'data) -> unit
    val replace: 'data t -> key -> 'data -> unit
    val keys: 'data t -> key list
    val size: 'data t -> int
  end

module Hashtbl(H: HASHED_ELEMENT) =
  struct
    module T = OrigHashtbl.Make(H)
    type key = T.key
    type 'data t = 'data T.t
    let create = T.create
    let clear = T.clear
    let add = T.add
    let remove = T.remove
    let find = T.find
    let find_all = T.find_all
    let iter = T.iter
    (* *)
    let fold f ht x =
      let y = ref x in
      iter (fun k d -> y := f k d !y) ht;
      !y
    (* fixme: Better implementation. *)
    let update ht k init_d f =
      let d =
	try
	  let d = find ht k in
	  remove ht k;
	  d
	with Not_found ->
	  init_d
      in
      add ht k (f d)
    let replace ht k d = update ht k d (fun _ -> d)
    (* *)
    let keys ht = fold (fun key _ lst -> key::lst) ht []
    let size ht = fold (fun _ _ n -> n+1) ht 0 
  end

module String_hashtbl = Hashtbl(String_element)

let hash = OrigHashtbl.hash
let hash_param = OrigHashtbl.hash_param

(************************************************************************)

module type UNIONFIND =
  sig
    type elt
    type t
    val make : elt list -> t
    val find_rep : t -> elt -> elt
    val union : t -> elt -> elt -> unit
    val finalize : t -> unit
    val find_class : t -> elt -> elt list
  end

module Unionfind(Ord: ORDERED_ELEMENT) =
  struct
    let equal x y = Ord.compare x y = 0
    module H = 
      Hashtbl
	(struct 
	  type t = Ord.t
	  let equal = equal
	  let hash = OrigHashtbl.hash
	end)
    type elt = Ord.t
    type t = {
	uf_universe : elt list;
	mutable uf_finalized_p : bool;
	uf_links : elt ref H.t;
	uf_classes : elt list H.t;
      }
    let make univ =
      let links =
	let ht = H.create 101 in
	List.iter
	  (fun e -> H.add ht e (ref e))
	  univ;
	ht
      in
      { uf_universe = univ;
	uf_finalized_p = false;
	uf_links = links;
	uf_classes = H.create 101 }
    let rec find_rep uf e =
      let r = H.find uf.uf_links e in
      if equal e !r then
	e
      else
	begin
	  let e' = find_rep uf !r in
	  r := e';
	  e'
	end
    let union uf e1 e2 =
      if uf.uf_finalized_p then failwith "union: already finalized";
      let e1' = find_rep uf e1 in
      let r1 = H.find uf.uf_links e1' in
      assert (equal !r1 e1');
      r1 := e2
    let finalize uf =
      if uf.uf_finalized_p then failwith "finalized: already finalized";
      uf.uf_finalized_p <- true;
      List.iter
	(fun e ->
	  let e' = find_rep uf e in
	  H.update uf.uf_classes e' [] (fun lst -> e::lst))
	uf.uf_universe

    let find_class uf e =
      if not uf.uf_finalized_p then
 	failwith "finalized: not yet finalized";
      H.find uf.uf_classes e
  end
	

module String_unionfind = Unionfind(String_element)

(************************************************************************)

module type ORDINATOR =
    sig
      type elt
      type t
      val create: unit -> t
      val to_int: t -> elt -> int
      val of_int: t -> int -> elt
      val size: t -> int
      val compare: t -> elt -> elt -> int
    end

module Ordinator(E: ELEMENT) =
  struct
    module EVector = Vector(E)
    module EHashtbl = 
      Hashtbl
	(struct
	  type t = E.t
	  let equal = E.equal
	  let hash = OrigHashtbl.hash
      	end)
    type elt = E.t
    type t = {
	ord_v : EVector.t;
	ord_ht : int EHashtbl.t;
      }
    let create () = { 
      ord_v = EVector.create (); 
      ord_ht = EHashtbl.create 101;
    }
    let to_int o x =
      try EHashtbl.find o.ord_ht x
      with Not_found ->
	let i = EVector.len o.ord_v in
	EVector.add_last o.ord_v x;
	EHashtbl.add o.ord_ht x i;
	i
    let of_int o i = 
      try EVector.at o.ord_v i with Not_found -> assert false
    let size o = EVector.len o.ord_v
    let compare o x y = (to_int o x) - (to_int o y)
  end
      
