(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
(************************************************************************)

module type ELEMENT =
  sig
    type t
    val equal: t -> t -> bool
  end

module type ORDERED_ELEMENT =
  sig
    type t
    val compare: t -> t -> int
  end

module type HASHED_ELEMENT =
  sig
    type t
    val equal: t -> t -> bool
    val hash: t -> int
  end

module type COMPLETE_ELEMENT =
  sig
    type t
    val compare: t -> t -> int
    val equal: t -> t -> bool
    val hash: t -> int
  end

module Int_element : (COMPLETE_ELEMENT with type t = int)
module String_element : (COMPLETE_ELEMENT with type t = string)

(************************************************************************)

module type VECTOR =
  sig
    type elt
    type t
    val create: unit -> t
    val make: int -> t
    val capacity: t -> int
    val set_capacity: t -> int -> unit
    val capacity_incr: t -> int
    val set_capacity_incr: t -> int -> unit
    val trim_to_size: t -> unit
    val erase: t -> unit
    val ensure_capacity: t -> int -> unit
    val len: t -> int
    val set_len: t -> int -> unit
    exception Uninitialized
    val at: t -> int -> elt
    val set_at: t -> int -> elt -> unit
    val insert_at: t -> int -> elt -> unit
    val add_first: t -> elt -> unit
    val add_last: t -> elt -> unit
    val remove_at: t -> int -> unit
    val remove_first: t -> unit
    val remove_last: t -> unit
    val index_from: t -> elt -> int -> int
    val index: t -> elt -> int
    val rindex_from: t -> elt -> int -> int
    val rindex: t -> elt -> int
    val remove: t -> elt -> unit
  end

module Vector(E: ELEMENT): (VECTOR with type elt = E.t)

(************************************************************************)

module type SET =
  sig
    (* Present in O'Caml's Set module. *)
    type elt
    type t
    val empty: t
    val is_empty: t -> bool
    val mem: elt -> t -> bool
    val add: elt -> t -> t
    val remove: elt -> t -> t
    val union: t -> t -> t
    val inter: t -> t -> t
    val diff: t -> t -> t
    val compare: t -> t -> int
    val equal: t -> t -> bool
    val subset: t -> t -> bool
    val iter: (elt -> unit) -> t -> unit
    val fold: (elt -> 'a -> 'a) -> t -> 'a -> 'a
    val cardinal: t -> int
    val elements: t -> elt list
    val choose: t -> elt
    (* new functions. *)
    val to_list: t -> elt list
    val of_list: elt list -> t
    val absorb: t -> elt list -> t
  end

module Set(Ord: ORDERED_ELEMENT): (SET with type elt = Ord.t)

module Int_set : (SET with type elt = int)
module String_set : (SET with type elt = string)


(************************************************************************)

module type MAP =
  sig
    (* Present in O'Caml's Map module. *)
    type key
    type 'data t
    val empty: 'data t
    val add: key -> 'data -> 'data t -> 'data t
    val find: key -> 'data t -> 'data
    val remove: key -> 'data t -> 'data t
    val iter: (key -> 'data -> unit) -> 'data t -> unit
    val map: ('a -> 'b) -> 'a t -> 'b t
    val fold: (key -> 'data -> 'a -> 'a) -> 'data t -> 'a -> 'a
    val update: 'data t -> key -> 'data -> ('data -> 'data) -> 'data t
    val replace: 'data t -> key -> 'data -> 'data t
    val keys: 'data t -> key list
    val size: 'data t -> int
  end

module Map(Ord: ORDERED_ELEMENT): (MAP with type key = Ord.t)

module String_map : (MAP with type key = string)

(************************************************************************)

module type HASHTBL =
  sig
    type key
    type 'data t
    val create: int -> 'data t
    val clear: 'data t -> unit
    val add: 'data t -> key -> 'data -> unit
    val remove: 'data t -> key -> unit
    val find: 'data t -> key -> 'data
    val find_all: 'data t -> key -> 'data list
    val iter: (key -> 'data -> unit) -> 'data t -> unit
    val fold: (key -> 'data -> 'a -> 'a) -> 'data t -> 'a -> 'a
    val update: 'data t -> key -> 'data -> ('data -> 'data) -> unit
    val replace: 'data t -> key -> 'data -> unit
    val keys: 'data t -> key list
    val size: 'data t -> int
  end

module Hashtbl(H: HASHED_ELEMENT): (HASHTBL with type key = H.t)

module String_hashtbl : (HASHTBL with type key = string)

val hash : 'a -> int
val hash_param : int -> int -> 'a -> int

(************************************************************************)

module type UNIONFIND =
  sig
    type elt
    type t
    val make : elt list -> t
    val find_rep : t -> elt -> elt
    val union : t -> elt -> elt -> unit
    val finalize : t -> unit
    val find_class : t -> elt -> elt list
  end

module Unionfind(Ord: ORDERED_ELEMENT) : (UNIONFIND with type elt = Ord.t)
module String_unionfind : (UNIONFIND with type elt = string)

(************************************************************************)

module type ORDINATOR =
    sig
      type elt
      type t
      val create: unit -> t
      val to_int: t -> elt -> int
      val of_int: t -> int -> elt
      val size: t -> int
      val compare: t -> elt -> elt -> int
    end

module Ordinator(E: ELEMENT) : (ORDINATOR with type elt = E.t)
