(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
(* flat_map *)
let flat_map f lst =
  List.fold_right (fun x result -> (f x)@result) lst []

(* map [a;b;c;....] into
   [(f 0 a); (f 1 b); (f 2 c); ....]
   *)
let mapi f lst = (
  let rec mapi_aux f i lst = (
    match lst with
    | x::lst' -> (f i x)::(mapi_aux f (i+1) lst')
    | [] -> []) in
  mapi_aux f 0 lst
)
    ;;


let match0 lst =
    match lst with
      [] -> ()
    | _ -> failwith "Expected a length 0 list."
;;

let match1 lst =
    match lst with
      [i] -> i
    | _ -> failwith "Expected a length 1 list."
;;

let match2 lst =
    match lst with
      [i;j] -> (i,j)
    | _ -> failwith "Expected a length 2 list."
;;

let match3 lst =
    match lst with
      [i;j;k] -> (i,j,k)
    | _ -> failwith "Expected a length 3 list."
;;

let match4 lst =
    match lst with
      [i;j;k;l] -> (i,j,k,l)
    | _ -> failwith "Expected a length 4 list."
;;

let match5 lst =
    match lst with
      [i;j;k;l;m] -> (i,j,k,l,m)
    | _ -> failwith "Expected a length 5 list."
;;


let range frm too step =
  let rec recurs_up frm =
    if frm >= too then [] else frm::(recurs_up (frm+step))
  and iter_up frm result =
    if frm >= too then result else iter_up (frm+step) (frm::result)
  and recurs_down frm =
    if frm <= too then [] else frm::(recurs_down (frm+step))
  and iter_down frm result =
    if frm <= too then result else iter_down (frm+step) (frm::result)
  in
  let n = (too - frm) / step in
  (* Use recursion for small lists and iteration for large lists.
     This is in order to avoid blowing out the stack. *)
  if n < 100 then
    if step > 0 then recurs_up frm else recurs_down frm
  else
    List.rev (if step > 0 then iter_up frm [] else iter_down frm [])

let from_to lb ub = range lb ub 1
let from_downto ub lb = range ub lb (-1)

(* nthcdr *)
let rec nthcdr lst i =
  match (lst,i) with
    (lst,0) -> lst
  | (_::lst',i) when i > 0 -> nthcdr lst' (i-1)
  | _ -> raise (Invalid_argument("nthcdr"))

(* sublist *)
let rec sublist lst start finish =
  match (start,finish) with
    (0,finish) when 0 <= finish ->
      let rec iter i lst result  =
	match (i,lst) with
	  (0,_) -> List.rev result
	| (i,x::lst') -> iter (i-1) lst' (x::result)
	| _ -> raise (Invalid_argument("sublist"))
      in
      iter finish lst []
  | (start,finish) when 0 <= start & start <= finish ->
      sublist (nthcdr lst start) 0 (finish-start)
  | _ -> raise (Invalid_argument("sublist"))

(* remove *)

let long_remove_if key pred lst =
  let rec iter lst result =
    match lst with
      [] -> List.rev result
    | x::lst' ->
	if pred (key x) then
	  iter lst' result
	else
	  iter lst' (x::result)
  in
  iter lst []

let long_remove test key item lst =
  long_remove_if key (fun z -> test z item) lst
let long_remove_if_not key pred lst =
  long_remove_if key (complement1 pred) lst

let remove_if pred lst = long_remove_if identity pred lst
let remove item lst = long_remove (=) identity item lst
let remove_if_not pred lst = long_remove_if_not identity pred lst

(* member *)

let rec long_member_if key pred lst =
  match lst with
    [] -> []
  | x::_ when pred (key x) -> lst
  | _::lst' -> long_member_if key pred lst'

let long_member test key item lst =
  long_member_if key (fun z -> test z item) lst
let long_member_if_not key pred lst =
  long_member_if key (complement1 pred) lst

let member_if pred lst = long_member_if identity pred lst
let member item lst = long_member (=) identity item lst
let member_if_not pred lst = long_member_if_not identity pred lst

(* remove_duplicates *)

let long_remove_duplicates test key lst =
  let rec iter lst result =
    match lst with
      [] -> List.rev result
    | x::lst' when (long_member test key (key x) lst') <> [] ->
 	iter lst' result
    | x::lst' -> iter lst' (x::result)
  in
  iter lst []

let long_remove_duplicates_from_end test key lst =
  List.rev (long_remove_duplicates test key (List.rev lst))

let remove_duplicates lst = long_remove_duplicates (=) identity lst
let remove_duplicates_from_end lst = long_remove_duplicates_from_end (=) identity lst

(* substitute *)

let long_substitute_if key newitem pred lst =
  let rec iter lst result =
    match lst with
      [] -> List.rev result
    | x::lst' when pred (key x) -> iter lst' (newitem::result)
    | x::lst' -> iter lst' (x::result)
  in
  iter lst []

let long_substitute test key newitem olditem lst =
  long_substitute_if key newitem (fun z -> test z (key olditem)) lst
let long_substitute_if_not key newitem pred lst =
  long_substitute_if key newitem (complement1 pred) lst

let substitute_if newitem pred lst = 
  long_substitute_if identity newitem pred lst
let substitute newitem olditem lst = 
  long_substitute (=) identity newitem olditem lst
let substitute_if_not newitem pred lst = 
  long_substitute_if_not identity newitem pred lst

(* find *)

let rec long_find_if key pred lst =
  match lst with
    [] -> raise Not_found
  | x::_ when pred (key x) -> x
  | _::lst' -> long_find_if key pred lst'

let long_find test key item lst =
  long_find_if key (fun z -> test z item) lst
let long_find_if_not key pred lst =
  long_find_if key (complement1 pred) lst

let find_if pred lst = long_find_if identity pred lst
let find item lst = long_find (=) identity item lst
let find_if_not pred lst = long_find_if_not identity pred lst

(* position *)

let long_position_if key pred lst =
  let rec iter lst i =
    match lst with
      [] -> raise Not_found
    | x::_ when pred (key x) -> i
    | _::lst' -> iter lst' (i+1)
  in
  iter lst 0

let long_position test key item lst =
  long_position_if key (fun z -> test z item) lst
let long_position_if_not key pred lst =
  long_position_if key (complement1 pred) lst

let position_if pred lst = long_position_if identity pred lst
let position item lst = long_position (=) identity item lst
let position_if_not pred lst = long_position_if_not identity pred lst

(* count *)

let long_count_if key pred lst =
  let rec iter lst n =
    match lst with
      [] -> n
    | x::lst' when pred (key x) -> iter lst' (n+1)
    | _::lst' -> iter lst' n
  in
  iter lst 0

let long_count test key item lst =
  long_count_if key (fun z -> test z item) lst
let long_count_if_not key pred lst =
  long_count_if key (complement1 pred) lst

let count_if pred lst = long_count_if identity pred lst
let count item lst = long_count (=) identity item lst
let count_if_not pred lst = long_count_if_not identity pred lst

(* mismatch *)

let long_mismatch test key lst1 lst2 = 0 (* ... *)

let mismatch lst1 lst2 = long_mismatch (=) identity lst1 lst2

(* search *)

let long_search test key lst1 lst2 = 0 (* ... *)

let search lst1 lst2 = long_search (=) identity lst1 lst2

(* adjoin *)

let long_adjoin test key item lst =
  if (long_member test key (key item) lst) <> [] then
    lst
  else
    item::lst

let adjoin item lst = long_adjoin (=) identity item lst

(* union *)

let long_union test key lst1 lst2 =
  let rec tester lst1 result =
    match lst1 with
      [] -> rev_app result lst2
    | x::lst1' ->
	if long_member test key (key x) result <> [] or
	  long_member test key (key x) lst2 <> [] then
	  tester lst1' result
	else
	  tester lst1' (x::result)
  and rev_app lst result =
      match lst with
	[] -> result
      |	x::lst' -> rev_app lst' (x::result)
  in
  tester lst1 []

let union lst1 lst2 = long_union (=) identity lst1 lst2

(* intersection *)

let long_intersection test key lst1 lst2 =
  let rec iter lst result =
    match lst with
      [] -> List.rev result
    | x::lst' ->
	if (long_member test key (key x) lst2) <> [] &
	  not (long_member test key (key x) result <> [] ) then
	  iter lst' (x::result)
	else
	  iter lst' result
  in
  iter lst1 []

let intersection lst1 lst2 = long_intersection (=) identity lst1 lst2

(* difference *)

let long_difference test key lst1 lst2 =
  let rec iter lst result =
      match lst with
	[] -> List.rev result
      |	x::lst' ->
	  if long_member test key (key x) lst2 <> [] or
	    long_member test key (key x) result <> [] then
	    iter lst' result
	  else
	    iter lst' (x::result)
  in
  iter lst1 []

let difference lst1 lst2 = long_difference (=) identity lst1 lst2

(* exclusive_or *)

let long_exclusive_or test key lst1 lst2 =
  (long_difference test key lst1 lst2)
    @ (long_difference test key lst2 lst1)

let exclusive_or lst1 lst2 = long_exclusive_or (=) identity lst1 lst2

(* subsetp *)

let long_subsetp test key lst1 lst2 =
  let rec iter lst =
    match lst with
      [] -> true
    | x::_ when not (long_member test key (key x) lst2 <> []) -> false
    | _::lst' -> iter lst'
  in
  iter lst1

let subsetp lst1 lst2 = long_subsetp (=) identity lst1 lst2

(* samep *)
let samep lst1 lst2 =
  ((subsetp lst1 lst2) && (subsetp lst2 lst1))
    


(* test if all are unique *)

let long_uniqp compare_keys key lst =
  let compare_elt x1 x2 =
    (compare_keys (key x1) (key x2)) in
  let order_elt x1 x2 = (0 < (compare_elt x1 x2))
  and equal_p x1 x2 = (0 = (compare_elt x1 x2)) in
  let rec no_dup lst' = (
    match lst' with
      [] -> true
    | [x] -> true
    | x::y::rest -> (not (equal_p x y)) && (no_dup (y::rest))) in
  (no_dup (Sort.list order_elt lst))

let uniqp lst = long_uniqp Pervasives.compare identity lst

(* generate subsets *)
let rec gen_subsets lst = (
  match lst with
    [] -> []
  | x::lst' -> 
      let g = (gen_subsets lst') in
      ([x]::(List.map (fun l -> x::l) g)) @ g)
    
(* make a cross-product of two lists *)
let rec cross lst1 lst2 = (
  match lst1 with
    [] -> []
  | x::lst1' -> 
      (List.map (fun y -> (x,y)) lst2) @ (cross lst1' lst2))
	
