(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
(************************************************************************)
(* Useful functions from CL 0.71					*)
(************************************************************************)

(* flat_map f [a;b;c] = (f a)@(f b)@(f c) *)
val flat_map : ('a -> 'b list) -> 'a list -> 'b list;;

(************************************************************************)
(* Other useful functions						*)
(************************************************************************)

val mapi : (int -> 'a -> 'b) -> 'a list -> 'b list;;


(* List destructors *)

val match0 : ('a list) -> unit;;
val match1 : ('a list) -> 'a;;
val match2 : ('a list) -> 'a * 'a ;;
val match3 : ('a list) -> 'a * 'a * 'a;;
val match4 : ('a list) -> 'a * 'a * 'a * 'a;;
val match5 : ('a list) -> 'a * 'a * 'a * 'a * 'a;;

(* Ranges of integers *)

(* range from to step -> [from,from+step,...,to) *)
val range : int -> int -> int -> int list
(* from_to lb ub -> [lb,lb+1,...,ub) *)
val from_to : int -> int -> int list
(* from_to ub lb -> [ub,ub-1,...,lb) *)
val from_downto : int -> int -> int list

(************************************************************************)
(* Common Lisp inspired list functions					*)
(************************************************************************)

(* The functions in this module are inspired by the routines in Chps 14 and
   15 of CLtl2. The functionality is meant to as similar as possible to
   that described in CLtl2, so any differences are probably bugs. There are
   a few exceptions to this, however:

   o Since ML does not provide keyword arguments, these arguments must
     always be present. Only the :test and :key arguments are used here.

   o The :test and :key arguments are placed as the beginning of the
     argument list. This is facilitate currying.

   o If a lisp function, xxx, takes keyword arguments, it will appear here
     with the name, long_xxx. The name, xxx, will be used here for the
     result of currying long_xxx with ":test (=) :key identity".

   Notes:
   o If there is a function,
	foo_if key pred lst
     then there will also be the followings routines,
	foo test key item lst == foo_if key (fun z -> test z item) lst
	foo_if_not pred lst == foo_if (complement1 pred) lst
   o THESE ROUTINES MAINTAIN ORIGINAL LIST ORDER.

   *)

(************************************************************************)
(* Misc routines							*)
(************************************************************************)

(* Returns the nth cdr of the list. Throws Not_found if the list is too
   short. *)

val nthcdr : 'a list -> int -> 'a list

(************************************************************************)
(* Lists as sequences							*)
(************************************************************************)

(* The elements with indices [begin,end) *)
val sublist: 'a list -> int(*begin*) -> int(*end*) -> 'a list

(* Remove each element, x, from lst for which (pred (key x)) is true. *)
val long_remove_if: ('a -> 'b)(*key*) -> ('b -> bool)(*pred*) -> 'a list -> 'a list
val long_remove: ('b -> 'b -> bool)(*test*) -> ('a -> 'b)(*key*) -> 'b -> 'a list -> 'a list
val long_remove_if_not: ('a -> 'b)(*key*) -> ('b -> bool)(*pred*) -> 'a list -> 'a list
val remove_if: ('a -> bool)(*pred*) -> 'a list -> 'a list
val remove: 'a -> 'a list -> 'a list
val remove_if_not: ('a -> bool)(*pred*) -> 'a list -> 'a list

(* Remove duplicate elements from a list. All but the last occurence of
   each duplicate are removed. *)
val long_remove_duplicates: 
    ('b -> 'b -> bool)(*test*) -> ('a -> 'b)(*key*) -> 'a list -> 'a list
(* Remove duplicate elements from a list. All but the first occurence of
   each duplicate are removed. *)
val long_remove_duplicates_from_end:
    ('b -> 'b -> bool)(*test*) -> ('a -> 'b)(*key*) -> 'a list -> 'a list
val remove_duplicates: 'a list -> 'a list
val remove_duplicates_from_end: 'a list -> 'a list

(* Replace each element, x, with newitem, if (test (key x) (key new_item))
   is true. *)
val long_substitute_if:
    ('a -> 'b)(*key*) -> 'a(*newitem*) -> ('b -> bool)(*pred*) -> 'a list -> 'a list
val long_substitute:
    ('b -> 'b -> bool)(*test*) -> ('a -> 'b)(*key*) -> 'a(*newitem*) -> 'a(*olditem*) -> 'a list -> 'a list
val long_substitute_if_not:
    ('a -> 'b)(*key*) -> 'a(*newitem*) -> ('b -> bool)(*pred*) -> 'a list -> 'a list
val substitute_if:
    'a(*newitem*) -> ('a -> bool)(*pred*) -> 'a list -> 'a list
val substitute:
    'a(*newitem*) -> 'a(*olditem*) -> 'a list -> 'a list
val substitute_if_not:
    'a(*newitem*) -> ('a -> bool)(*pred*) -> 'a list -> 'a list

(* The leftmost element, x, for which (pred (key x)) is true. 
   Throws Not_found otherwise. *)
val long_find_if: ('a -> 'b)(*key*) -> ('b -> bool)(*pred*) -> 'a list -> 'a
val long_find: ('b -> 'b -> bool)(*test*) -> ('a -> 'b)(*key*) -> 'b -> 'a list -> 'a
val long_find_if_not: ('a -> 'b)(*key*) -> ('b -> bool)(*pred*) -> 'a list -> 'a
val find_if: ('a -> bool)(*pred*) -> 'a list -> 'a
val find: 'a -> 'a list -> 'a
val find_if_not: ('a -> bool)(*pred*) -> 'a list -> 'a

(* The index (from 0) of the leftmost element, x, for which
   (pred (key x)) is true. 
   Throws Not_found otherwise. *)
val long_position_if: ('a -> 'b)(*key*) -> ('b -> bool)(*pred*) -> 'a list -> int
val long_position: ('b -> 'b -> bool)(*test*) -> ('a -> 'b)(*key*) -> 'b -> 'a list -> int
val long_position_if_not: ('a -> 'b)(*key*) -> ('b -> bool)(*pred*) -> 'a list -> int
val position_if: ('a -> bool)(*pred*) -> 'a list -> int
val position: 'a -> 'a list -> int
val position_if_not: ('a -> bool)(*pred*) -> 'a list -> int

(* The number of elements, x, for which (pred (key x)) is true. 
   Throws Not_found otherwise. *)
val long_count_if: ('a -> 'b)(*key*) -> ('b -> bool)(*pred*) -> 'a list -> int
val long_count: ('b -> 'b -> bool)(*test*) -> ('a -> 'b)(*key*) -> 'b -> 'a list -> int
val long_count_if_not: ('a -> 'b)(*key*) -> ('b -> bool)(*pred*) -> 'a list -> int
val count_if: ('a -> bool)(*pred*) -> 'a list -> int
val count: 'a -> 'a list -> int
val count_if_not: ('a -> bool)(*pred*) -> 'a list -> int

(* The index within lst1 (from 0) at which the first (leftmost) mismatch
   between the elements of lst1 and lst2 occurs. If lst1 and lst2 match,
   but lst2 is longer, then returns length of lst1.
   Throws Not_found otherwise. *)
val long_mismatch: ('b -> 'b -> bool)(*test*) -> ('a -> 'b)(*key*) -> 'a list(*lst1*) ->
  'a list(*lst2*) -> int
val mismatch: 'a list(*lst1*) -> 'a list(*lst2*) -> int

(* The index within lst2 at which the subsequence lst1 appears.
   Throws Not_found otherwise. *)
val long_search: ('b -> 'b -> bool)(*test*) -> ('a -> 'b)(*key*) -> 'a list(*lst1*) ->
  'a list(*lst2*) -> int
val search: 'a list(*lst1*) -> 'a list(*lst2*) -> int

(************************************************************************)
(* Lists as sets							*)
(************************************************************************)

(* The sublist starting with the leftmost element, x, for which
   (pred (key x)) is true. Returns [] if no such x is found. *)
val long_member_if: ('a -> 'b)(*key*) -> ('b -> bool)(*pred*) -> 'a list -> 'a list
val long_member: ('b -> 'b -> bool)(*test*) -> ('a -> 'b)(*key*) -> 'b -> 'a list -> 'a list
val long_member_if_not: ('a -> 'b)(*key*) -> ('b -> bool)(*pred*) -> 'a list -> 'a list
val member_if: ('a -> bool)(*pred*) -> 'a list -> 'a list
val member: 'a -> 'a list -> 'a list
val member_if_not: ('a -> bool)(*pred*) -> 'a list -> 'a list

(* Add an item to a list, only if it isn't already there. *)
val long_adjoin: ('b -> 'b -> bool)(*test*) -> ('a -> 'b)(*key*) -> 'a -> 'a list -> 'a list
val adjoin: 'a -> 'a list -> 'a list

(* Prepend all elements of lst1 not in lst2 to lst2. Order within lst1 and
   lst2 are maintained. NOTE: this is more than CLtl2 requires. *)
val long_union: ('b -> 'b -> bool)(*test*) -> ('a -> 'b)(*key*) -> 'a list -> 'a list -> 'a list
val union: 'a list -> 'a list -> 'a list

(* The first occurance of each element of lst1 that appears in lst2. The 
   order of result is the same as in lst1. NOTE: this is more than CLtl2
   requires. *)
val long_intersection: 
    ('b -> 'b -> bool)(*test*) -> ('a -> 'b)(*key*) -> 'a list -> 'a list -> 'a list
val intersection: 'a list -> 'a list -> 'a list

(* The first occurance of each elements of lst1 that do not appear in lst2.
   NOTE: this is more than CLtl2 requires. *)
val long_difference: 
    ('b -> 'b -> bool)(*test*) -> ('a -> 'b)(*key*) -> 'a list -> 'a list -> 'a list
val difference: 'a list -> 'a list -> 'a list

(* Append the first occurance of each elements of lst1 that do not appear
   in lst2 to the first occurance of each element that does not appear in
   lst1. The order of these elements is preserved. NOTE: this is more than
   CLtl2 requires. *)
val long_exclusive_or: 
    ('b -> 'b -> bool)(*test*) -> ('a -> 'b)(*key*) -> 'a list -> 'a list -> 'a list
val exclusive_or: 'a list -> 'a list -> 'a list

(* True iff lst1 is a subset of lst2 *)
val long_subsetp:
    ('b -> 'b -> bool)(*test*) -> ('a -> 'b)(*key*) -> 'a list -> 'a list -> bool
val subsetp: 'a list -> 'a list -> bool

(* True iff lst1 and lst2 contain the same elts
   (not necessarily in order) *)
val samep : 'a list -> 'a list -> bool

(* True if lst has no duplicates *)
val long_uniqp:
    ('a -> 'a -> int)(*comparator*) -> ('b -> 'a)(*key*) -> 'b list -> bool

val uniqp: (* uses Pervasives.compare *)
    'a list -> bool

(* Generates all non-empty subsets *)
(* elements are assumed to be unique *)
val gen_subsets: 'a list -> 'a list list

(* generate a cross product of two lists *)
val cross : 'a list -> 'b list -> ('a * 'b) list
