(************************************************************************)
(*	Copyright (C) 1998. Bernoulli Project, Cornell University.	*)
(************************************************************************)
open Ssisl
open Bsct_base

module type S =
  sig

    type atom

    module ATOM : AtomType with type t = atom

    type t =
    	Conjunct of t * t
      | Disjunct of t * t
      | Implies  of t * t
      | Equality of t * t
      | Negation of t
      | Atom of atom
      | Constant of bool

    val compare: t -> t -> int
    exception Invalid_s_expr of Sx.t
    val from_sexpr: Sx.t -> t
    val to_sexpr: t -> Sx.t

    (* Return a list of the conjuncts or disjuncts that appear at the top
       level of the expression e *)
    val to_conjuncts: (*e*) t -> t list
    val to_disjuncts: (*e*) t -> t list

    (* Make a wff from a list of conjuncts or disjuncts *)
    val from_conjuncts: t list -> t
    val from_disjuncts: t list -> t

    (* Return true iff e is a tautology *)
    val tautology_p: (*e*) t -> bool

    (* gist w1 w2 -> w1', where w2 && w1' => w1. Hopefully, w1' is
       "minimal." *)
    val gist: (*w1*) t -> (*w2*) t -> (*w1'*) t

    val simplify: t -> t

    val subst: (*bindings*)(atom * t) list -> (*w*)t -> t

  end

module Make(A: AtomType) = struct

  module ATOM = A
  type atom = A.t

  module AtomSet = Set.Make(A)
  module AtomMap = Map.Make(A)

  type t =
      Conjunct of t * t
    | Disjunct of t * t
    | Implies  of t * t
    | Equality of t * t
    | Negation of t
    | Atom of A.t
    | Constant of bool
	  
  (* Return a list of the conjuncts that appear at the top level of the
     expression e *)
  let rec to_conjuncts e =
    match e with
      Conjunct(e1,e2) -> (to_conjuncts e1)@(to_conjuncts e2)
    | _ -> [e]
	  
  (* Return a list of the disjuncts that appear at the top level of the
     expression e *)
  let rec to_disjuncts e =
    match e with
      Disjunct(e1,e2) -> (to_disjuncts e1)@(to_disjuncts e2)
    | _ -> [e]
	  
  (* Return a wff from a list of conjuncts *)
  let rec from_conjuncts lst =
    match lst with
      [] -> Constant(true) (* *todo*: is this right? *)
    | [x] -> x
    | x::lst' -> Conjunct(x, from_conjuncts lst')

  (* Return a wff from a list of disjuncts *)
  let rec from_disjuncts lst =
    match lst with
      [] -> Constant(true)
    | [x] -> x
    | x::lst' -> Disjunct(x, from_disjuncts lst')

  let compare = Pervasives.compare
  exception Invalid_s_expr of Sx.t
  let rec from_sexpr x =
    match x with
      Sx.Bool(b) -> Constant(b)
    | Sx.Pair(Sx.Symbol("and"),rest) ->
	from_conjuncts (List.map from_sexpr (Sx.to_list rest))
    | Sx.Pair(Sx.Symbol("or"),rest) ->
	from_disjuncts (List.map from_sexpr (Sx.to_list rest))
    | Sx.Pair(Sx.Symbol("=>"),rest) ->
	(match Sx.to_list rest with
	  [a;b] -> Implies(from_sexpr a,from_sexpr b)
	| _ -> raise (Invalid_s_expr(x)))
    | Sx.Pair(Sx.Symbol("="),rest) ->
	(match Sx.to_list rest with
	  [a;b] -> Equality(from_sexpr a,from_sexpr b)
	| _ -> raise (Invalid_s_expr(x)))
    | Sx.Pair(Sx.Symbol("not"),rest) ->
	(match Sx.to_list rest with
	  [a] -> Negation(from_sexpr a)
	| _ -> raise (Invalid_s_expr(x)))
    | _ -> Atom(A.from_sexpr x)

  let rec to_sexpr x =
    match x with
      Conjunct(_,_) -> 
	Sx.Pair(Sx.Symbol("and"),
		Sx.from_list (List.map to_sexpr (to_conjuncts x)))
    | Disjunct(_,_) -> 
	Sx.Pair(Sx.Symbol("or"),
		Sx.from_list (List.map to_sexpr (to_disjuncts x)))
    | Implies(a,b) ->
	Sx.from_list [Sx.Symbol("=>"); to_sexpr a; to_sexpr b]
    | Equality(a,b) ->
	Sx.from_list [Sx.Symbol("="); to_sexpr a; to_sexpr b]
    | Negation(a) ->
	Sx.from_list [Sx.Symbol("not"); to_sexpr a]
    | Atom(a) -> A.to_sexpr a
    | Constant(b) -> Sx.Bool(b)

  (* Return the set of free variables in e *)
  let rec free_vars e =
    match e with
      Conjunct(e1,e2) -> AtomSet.union (free_vars e1) (free_vars e2)
    | Disjunct(e1,e2) -> AtomSet.union (free_vars e1) (free_vars e2)
    | Implies(e1,e2) -> AtomSet.union (free_vars e1) (free_vars e2)
    | Equality(e1,e2) -> AtomSet.union (free_vars e1) (free_vars e2)
    | Negation(e1) -> free_vars e1
    | Atom(s) -> AtomSet.add s AtomSet.empty
    | Constant(_) -> AtomSet.empty
	  
  (* Return true iff e is a tautology *)
  let rec tautology_p e =
    let fvs = AtomSet.elements (free_vars e)
    and e' = compile_expr e
    in
    for_all_truth_assignments fvs e'
      
  and compile_expr e =
    match e with
      Conjunct(e1,e2) ->
	let e1' = compile_expr e1
	and e2' = compile_expr e2
	in
	(fun asgn -> (e1' asgn) && (e2' asgn))
    | Disjunct(e1,e2) ->
	let e1' = compile_expr e1
	and e2' = compile_expr e2
	in
	(fun asgn -> (e1' asgn) || (e2' asgn))
    | Implies(e1,e2) ->
	let e1' = compile_expr e1
	and e2' = compile_expr e2
	in
	(fun asgn -> (not ((e1' asgn))) || (e2' asgn))
    | Equality(e1,e2) ->
	let e1' = compile_expr e1
	and e2' = compile_expr e2
	in
	(fun asgn -> (e1' asgn) = (e2' asgn))
    | Negation(e1) ->
	let e1' = compile_expr e1
	in
	(fun asgn -> (not ((e1' asgn))))
    | Atom(s) ->
	(fun asgn -> AtomMap.find s asgn)
    | Constant(b) ->
	(fun asgn -> b)
	  
  and for_all_truth_assignments fvs e' =
    for_all_truth_assignments' fvs AtomMap.empty e'
      
  and for_all_truth_assignments' fvs asgn e' =
    match fvs with
      [] -> e' asgn
    | fv::fvs' ->
	let f x =
	  for_all_truth_assignments' fvs' (AtomMap.add fv x asgn) e'
	in
	(f true)&&(f false)

  let rec simplify x =
    let true_wff = Constant(true)
    and false_wff = Constant(false)
    in
    let rec make_and terms =
      let terms = List'.flat_map to_conjuncts terms in
      let terms = List'.remove_duplicates terms in
      let terms = List'.remove true_wff terms in
      if List.mem false_wff terms then
	false_wff
      else
	from_conjuncts terms
    and make_or terms =
      let terms = List'.flat_map to_disjuncts terms in
      let terms = List'.remove_duplicates terms in
      let terms = List'.remove false_wff terms in
      if List.mem true_wff terms then
	true_wff
      else
	from_conjuncts terms
    and make_implies lhs rhs = Implies(lhs,rhs)
    and make_equality lhs rhs = Equality(lhs,rhs)
    and make_negation term =
      match term with
	Negation(term') -> term'
      |	_ -> Negation(term)
    in
    match x with
      Conjunct(_,_) -> make_and (List.map simplify (to_conjuncts x))
    | Disjunct(_,_) -> make_or (List.map simplify (to_disjuncts x))
    | Implies(y,z) -> make_implies (simplify y) (simplify z)
    | Equality(y,z) -> make_equality (simplify y) (simplify z)
    | Negation(y) -> make_negation (simplify y)
    | _ -> x

  (* gist w1 w2 -> w1', where w2 && w1' => w1. Hopefully, w1' is
     "minimal." *)
  let gist w1 w2 =
    let w1 = simplify w1
    in let terms = to_conjuncts w1
    in let terms' =
      List.fold_right
	(fun term terms' ->
	  if tautology_p (Implies(w2,term)) then
	    (* term is redundent wrt w2 *)
	    terms'
	  else
	    term::terms')
	terms []
    in
    from_conjuncts terms'

    let rec subst bindings w =
      let r = subst bindings in
      match w with
	Conjunct(w1,w2) -> Conjunct(r w1,r w2)
      |	Disjunct(w1,w2) -> Disjunct(r w1,r w2)
      |	Implies(w1,w2) -> Implies(r w1,r w2)
      |	Equality(w1,w2) -> Equality(r w1,r w2)
      |	Negation(w1) -> Negation(r w1)
      |	Constant _ -> w
      |	Atom(v) ->
	  (try List.assoc v bindings
	  with Not_found -> w)

end

module Simple = Make(SimpleAtom)
